/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any 32-bit protected mode environment
*
* Description:  Header file to include all Nucleus internal headers.
*
****************************************************************************/

#ifndef __SNAP_AGP_COMMON_H
#define __SNAP_AGP_COMMON_H

#include "snap/agp.h"
#include "snap/agp/version.h"
#include "snap/common/internal.h"
#include "snap/common/io.h"
#include "snap/common/mem.h"
#include "snap/common/list.h"
#include "clib/peloader.h"
#include "clib/os/os.h"
#include "clib/attrib.h"
#ifndef __WIN32_VXD__
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#endif

/*--------------------- Macros and type definitions -----------------------*/

#pragma pack(1)

/****************************************************************************
REMARKS:
Structure defining all the Nucleus graphics driver functions as imported
into the Binary Portable DLL.
{secret}
****************************************************************************/
typedef struct {
    long        dwSize;
    void        (NAPIP N_log)(const char *fmt,...);
    N_uint32    (NAPIP PCI_accessReg)(int index,N_uint32 value,int func,PCIDeviceInfo *info);
    void        (NAPIP usDelay)(N_uint32 microSeconds);
    ibool       (NAPIP N_createBlockList)(N_blockList *l,int objectSize);
    ibool       (NAPIP N_resizeBlockList)(N_blockList *l);
    void        (NAPIP N_freeBlockList)(N_blockList *l);
    void *      (NAPIP N_buildFreeList)(N_blockList *l);
    void        (NAPIP N_initList)(N_list *l);
    void        (NAPIP N_destroyList)(N_list *l,void (*freeNode)(void *node));
    void        (NAPIP N_addAfter)(N_list *l,void *node,void *after);
    void        (NAPIP N_addToHead)(N_list *l,void *node);
    void        (NAPIP N_addToTail)(N_list *l,void *node);
    void *      (NAPIP N_removeNext)(N_list *l,void *node);
    void *      (NAPIP N_firstNode)(N_list *l);
    void *      (NAPIP N_nextNode)(void *prev);
    void *      (NAPIP N_findNode)(N_list *l,void *n);
    } DRV_imports;

/****************************************************************************
REMARKS:
Function pointer type for the Binary Portable DLL initialisation entry point
for graphics device drivers.
{secret}
****************************************************************************/
typedef AGP_devCtx * (NAPIP AGP_initDriver_t)(
    const char *path,
    N_int32 deviceIndex,
    PCIDeviceInfo *AGPDevice,
    PCIDeviceInfo *AGPBridge,
    PCIDeviceInfo *hostBridge,
    N_imports *nImp,
    PM_imports *pmImp,
    DRV_imports *drvImp,
    AGP_globalOptions *globalOpt);

#ifndef __DRIVER__
#define _CEXPORT __cdecl
#else
#define _CEXPORT __cdecl __export
#endif

#pragma pack()

/*-------------------------- Function Prototypes --------------------------*/

#ifdef  __cplusplus
extern "C" {                        /* Use "C" linkage when in C++ mode */
#endif

/* Device driver exported entry points. Ideally these would only be listed in
 * the snap/agp/driver.h header file, however because we need to call these
 * functions from the generic loader code directly when compiling in TEST_HARNESS
 * mode, they need to be listed here.
 */

AGP_devCtx * _CEXPORT AGP_initDriver(const char *path,N_int32 deviceIndex,PCIDeviceInfo *AGPDevice,PCIDeviceInfo *AGPBridge,PCIDeviceInfo *hostBridge,N_imports *nImp,PM_imports *pmImp,DRV_imports *drvImp,AGP_globalOptions *globalOpt);

#ifdef  __cplusplus
}                                   /* End of "C" linkage for C++       */
#endif

#endif  /* __SNAP_AGP_COMMON_H */

