/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  IBM PC (MS DOS)
*
* Description:  Header file for internal driver neutral SNAP code. This
*               module is common to all SciTech SNAP device driver types.
*
****************************************************************************/

#ifndef __SNAP_COMMON_INTERNAL_H
#define __SNAP_COMMON_INTERNAL_H

#include "snap/common.h"
#ifdef  __WIN32_VXD__
#include "vxdfile.h"
#else
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#endif

/*---------------------- Macros and type definitions ----------------------*/

#pragma pack(1)

// Macros for special function pointer vectors

#if defined(__IBMC__) || defined(__IBMCPP__)
typedef void (*__code_ptr_t)();
typedef short (*__scode_ptr_t)();
typedef int (*__icode_ptr_t)();
typedef long (*__lcode_ptr_t)();
typedef ushort (*__uscode_ptr_t)();
typedef uint (*__uicode_ptr_t)();
typedef ulong (*__ulcode_ptr_t)();
typedef void (_ASMAPIP __c_code_ptr_t)();
typedef long (_ASMAPIP __lc_code_ptr_t)();
#define     CNOT_IMPLEMENTED    (__c_code_ptr_t)0xFFFFFFFF
#define     NOT_IMPLEMENTED     (__code_ptr_t)0xFFFFFFFF
#define     CNOT_SUPPORTED      (__c_code_ptr_t)0xFFFFFFFE
#define     LCNOT_SUPPORTED     (__lc_code_ptr_t)0xFFFFFFFE
#define     NOT_SUPPORTED       (__code_ptr_t)0xFFFFFFFE
#define     CNEED_SPECS         (__c_code_ptr_t)0xFFFFFFFD
#define     NEED_SPECS          (__code_ptr_t)0xFFFFFFFD
#else
#define     CNOT_IMPLEMENTED    (void*)0xFFFFFFFF
#define     NOT_IMPLEMENTED     (void*)0xFFFFFFFF
#define     CNOT_SUPPORTED      (void*)0xFFFFFFFE
#define     LCNOT_SUPPORTED     (void*)0xFFFFFFFE
#define     NOT_SUPPORTED       (void*)0xFFFFFFFE
#define     CNEED_SPECS         (void*)0xFFFFFFFD
#define     NEED_SPECS          (void*)0xFFFFFFFD
#endif

/****************************************************************************
REMARKS:
Structure defining the directory of flat files attached to the DLL.
{secret}
****************************************************************************/
typedef struct {
    char    filename[40];
    ulong   start;
    ulong   length;
    } FLAT_DIR;

/****************************************************************************
REMARKS:
Structure defining a flat file such that we can do all file I/O to the
flat file appended to the end of the core binary driver file.
{secret}
****************************************************************************/
typedef struct {
    FILE    *f;
    ulong   start;
    ulong   offset;
    ulong   length;
    } N_FILE;

// Macro to find the size of a static array

#define ARR_SIZE(a)         (sizeof(a)/sizeof((a)[0]))

// Macro to clear a structure, excluding the dwSize member

#define CLEAR_STRUCTURE(d)                                      \
    memset((uchar*)d+sizeof((d)->dwSize),0,                     \
        (d)->dwSize-sizeof((d)->dwSize))

// Macros to copy a structure that includes dwSize members

#define COPY_STRUCTURE(d,s)                                     \
{                                                               \
    if (d != s) {                                               \
        memset((uchar*)d+sizeof((d)->dwSize),0,                 \
            (d)->dwSize-sizeof((d)->dwSize));                   \
        (d)->dwSize = MIN((s)->dwSize,(d)->dwSize);             \
        memcpy(d,s,(d)->dwSize);                                \
        }                                                       \
}

#define COPY_STRUCTURE2(d,s,dSize)                              \
{                                                               \
    if (d != s) {                                               \
        memset((uchar*)d+sizeof(dSize),0,dSize-sizeof(dSize));  \
        *(N_uint32*)(d) = MIN((s)->dwSize,dSize);               \
        memcpy(d,s,*(N_uint32*)(d));                            \
        }                                                       \
}

#pragma pack()

/*------------------------- Global Variables ------------------------------*/

extern int status;
extern int disableLogging;

/*-------------------------- Function Prototypes --------------------------*/

#ifdef  __cplusplus
extern "C" {                        /* Use "C" linkage when in C++ mode */
#endif

/* File I/O functions for opening and reading files from a flat file
 * structure appended to the end of the
 */

ibool       FLAT_init(const char *bpdname);
FLAT_DIR *  FLAT_openDirectory(int *count);
void        FLAT_closeDirectory(void);
N_FILE *    FLAT_openFile(const char *filename);
long        FLAT_readFile(N_FILE *f,void *buf,long size);
void        FLAT_closeFile(N_FILE *f);
void        N_logCPUInfo(void);
ibool       _PCI_findDriverName(PCIDeviceInfo *info,char *filename,char *tabFile);

#define _log N_log

#ifdef  __cplusplus
}                                   /* End of "C" linkage for C++       */
#endif

#endif  /* __SNAP_COMMON_INTERNAL_H */

