/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Common header for regular and memory mapped I/O access
*               from C code. This module is common to all SciTech SNAP
*               device driver types.
*
****************************************************************************/

#ifndef __SNAP_COMMON_IO_H
#define __SNAP_COMMON_IO_H

#include "snap/common.h"
#include "snap/common/internal.h"
#ifdef  __WIN32_VXD__
#undef  _outpb
#undef  _outpw
#undef  _outpd
#undef  _inpb
#undef  _inpw
#undef  _inpd
#undef  outpb
#undef  outpw
#undef  outpd
#undef  inpb
#undef  inpw
#undef  inpd
#endif

/*---------------------- Macros and type definitions ----------------------*/

// Macro to declare a size of 1Mb of

#define MB                  (1024UL * 1024UL)

// Macro to declare a value in MHz

#define MHz                 (1000000)

// Macros to declare variables used in functions

#define	MMIO_DECLARE										\
	volatile uchar *mmioBase = (volatile uchar*)DC.IOMemMaps[0]
#define	MMIO_DECLARE_1										\
	volatile uchar *mmioBase = (volatile uchar*)DC.IOMemMaps[1]
#define	MMIO_DECLARE_2										\
	volatile uchar *mmioBase = (volatile uchar*)DC.IOMemMaps[2]
#define	MMIO_DECLARE_3										\
	volatile uchar *mmioBase = (volatile uchar*)DC.IOMemMaps[3]

#ifdef __INTEL__

// Macros to read and write MMIO registers. All macros assume the local
// register variable mmioBase is set to the base of the MMIO registers.
// To avoid problems with compiler bugs relate to 'volatile' pointers,
// we use inline assembler to do the actual memory mapped I/O access.

#define	outmb(reg,val)	            (*((volatile uchar*)(mmioBase+(reg)))) = val
#define	outmw(reg,val)	            (*((volatile ushort*)(mmioBase+(reg)))) = val
#define	outmd(reg,val)	            (*((volatile ulong*)(mmioBase+(reg)))) = val
#define	outmdn(reg,val)	            (*((volatile ulong*)(mmioBase+(reg)))) = val
#define	inmb(reg)		            (*((volatile uchar*)(mmioBase+(reg))))
#define	inmw(reg)		            (*((volatile ushort*)(mmioBase+(reg))))
#define	inmd(reg)		            (*((volatile ulong*)(mmioBase+(reg))))
#define modmb(reg,mask,value)       outmb(((reg)),(inmb((reg)) & ~((mask))) | (((value)) & ((mask))))
#define setmb(reg,mask)             outmb(((reg)),inmb(((reg))) | ((mask)))
#define clrmb(reg,mask)             outmb(((reg)),inmb(((reg))) & ~((mask)))
#define modmw(reg,mask,value)       outmw(((reg)),(inmw((reg)) & ~((mask))) | (((value)) & ((mask))))
#define setmw(reg,mask)             outmw(((reg)),inmw(((reg))) | ((mask)))
#define clrmw(reg,mask)             outmw(((reg)),inmw(((reg))) & ~((mask)))
#define modmd(reg,mask,value)       outmd(((reg)),(inmd((reg)) & ~((mask))) | (((value)) & ((mask))))
#define setmd(reg,mask)             outmd(((reg)),inmd(((reg))) | ((mask)))
#define clrmd(reg,mask)             outmd(((reg)),inmd(((reg))) & ~((mask)))
#define rep_movsd_le(dst,src,count) rep_movsd(dst,src,count)

#elif defined(__PPC__)

// Inline assembler functions to perform memory mapped I/O on PowerPC.
// We hide the inline assembler function calls via macros that match
// the same semantics used for x86 boxes

static __inline__ uchar mm_inpb(
    __volatile__ void *base,
    const ulong offset)
{
    register uchar val;
    __asm__ __volatile__(
        "lbzx %0,%1,%2\n\t"
        "eieio"
        : "=r" (val)
        : "b" (base), "r" (offset),
        "m" (*((volatile uchar *)base+offset)));
    return val;
}

static __inline__ ushort mm_inpw(
    __volatile__ void *base,
    const ulong offset)
{
    register ushort val;
    __asm__ __volatile__(
        "lhbrx %0,%1,%2\n\t"
        "eieio"
        : "=r" (val)
        : "b" (base), "r" (offset),
        "m" (*((volatile uchar *)base+offset)));
    return val;
}

static __inline__ uint mm_inpd(
    __volatile__ void *base,
    const ulong offset)
{
    register uint val;
    __asm__ __volatile__(
        "lwbrx %0,%1,%2\n\t"
        "eieio"
        : "=r" (val)
        : "b" (base), "r" (offset),
        "m" (*((volatile uchar *)base+offset)));
    return val;
}

static __inline__ void mm_outpb(
    __volatile__ void *base,
    const ulong offset,
    const uchar val)
{
    __asm__ __volatile__(
        "stbx %1,%2,%3\n\t"
        "eieio"
        : "=m" (*((volatile uchar *)base+offset))
        : "r" (val), "b" (base), "r" (offset));
}

static __inline__ void mm_outpw(
    __volatile__ void *base,
    const ulong offset,
    const ushort val)
{
    __asm__ __volatile__(
        "sthbrx %1,%2,%3\n\t"
        "eieio"
        : "=m" (*((volatile uchar *)base+offset))
        : "r" (val), "b" (base), "r" (offset));
}

static __inline__ void mm_outpd(
    __volatile__ void *base,
    const ulong offset,
    const uint val)
{
    __asm__ __volatile__(
        "stwbrx %1,%2,%3\n\t"
        "eieio"
        : "=m" (*((volatile uchar *)base+offset))
        : "r" (val), "b" (base), "r" (offset));
}

static __inline__ void mm_outpdn(
    __volatile__ void *base,
    const ulong offset,
    const uint val)
{
    __asm__ __volatile__(
        "stwx %1,%2,%3\n\t"
        "eieio"
        : "=m" (*((volatile uchar *)base+offset))
        : "r" (val), "b" (base), "r" (offset));
}

#define	outmb(reg,val)  mm_outpb(mmioBase,reg,val)
#define	outmw(reg,val)  mm_outpw(mmioBase,reg,val)
#define	outmd(reg,val)  mm_outpd(mmioBase,reg,val)
#define	outmdn(reg,val) mm_outpdn(mmioBase,reg,val)
#define	inmb(reg)       mm_inpb(mmioBase,reg)
#define	inmw(reg)       mm_inpw(mmioBase,reg)
#define	inmd(reg)       mm_inpd(mmioBase,reg)

#else

#error SNAP has not been ported to this platform yet!

#endif

// Macros to declare binary constants

#define B00000000 0x00
#define B00000001 0x01
#define B00000010 0x02
#define B00000011 0x03
#define B00000100 0x04
#define B00000101 0x05
#define B00000110 0x06
#define B00000111 0x07
#define B00001000 0x08
#define B00001001 0x09
#define B00001010 0x0A
#define B00001011 0x0B
#define B00001100 0x0C
#define B00001101 0x0D
#define B00001110 0x0E
#define B00001111 0x0F
#define B00010000 0x10
#define B00010001 0x11
#define B00010010 0x12
#define B00010011 0x13
#define B00010100 0x14
#define B00010101 0x15
#define B00010110 0x16
#define B00010111 0x17
#define B00011000 0x18
#define B00011001 0x19
#define B00011010 0x1A
#define B00011011 0x1B
#define B00011100 0x1C
#define B00011101 0x1D
#define B00011110 0x1E
#define B00011111 0x1F
#define B00100000 0x20
#define B00100001 0x21
#define B00100010 0x22
#define B00100011 0x23
#define B00100100 0x24
#define B00100101 0x25
#define B00100110 0x26
#define B00100111 0x27
#define B00101000 0x28
#define B00101001 0x29
#define B00101010 0x2A
#define B00101011 0x2B
#define B00101100 0x2C
#define B00101101 0x2D
#define B00101110 0x2E
#define B00101111 0x2F
#define B00110000 0x30
#define B00110001 0x31
#define B00110010 0x32
#define B00110011 0x33
#define B00110100 0x34
#define B00110101 0x35
#define B00110110 0x36
#define B00110111 0x37
#define B00111000 0x38
#define B00111001 0x39
#define B00111010 0x3A
#define B00111011 0x3B
#define B00111100 0x3C
#define B00111101 0x3D
#define B00111110 0x3E
#define B00111111 0x3F
#define B01000000 0x40
#define B01000001 0x41
#define B01000010 0x42
#define B01000011 0x43
#define B01000100 0x44
#define B01000101 0x45
#define B01000110 0x46
#define B01000111 0x47
#define B01001000 0x48
#define B01001001 0x49
#define B01001010 0x4A
#define B01001011 0x4B
#define B01001100 0x4C
#define B01001101 0x4D
#define B01001110 0x4E
#define B01001111 0x4F
#define B01010000 0x50
#define B01010001 0x51
#define B01010010 0x52
#define B01010011 0x53
#define B01010100 0x54
#define B01010101 0x55
#define B01010110 0x56
#define B01010111 0x57
#define B01011000 0x58
#define B01011001 0x59
#define B01011010 0x5A
#define B01011011 0x5B
#define B01011100 0x5C
#define B01011101 0x5D
#define B01011110 0x5E
#define B01011111 0x5F
#define B01100000 0x60
#define B01100001 0x61
#define B01100010 0x62
#define B01100011 0x63
#define B01100100 0x64
#define B01100101 0x65
#define B01100110 0x66
#define B01100111 0x67
#define B01101000 0x68
#define B01101001 0x69
#define B01101010 0x6A
#define B01101011 0x6B
#define B01101100 0x6C
#define B01101101 0x6D
#define B01101110 0x6E
#define B01101111 0x6F
#define B01110000 0x70
#define B01110001 0x71
#define B01110010 0x72
#define B01110011 0x73
#define B01110100 0x74
#define B01110101 0x75
#define B01110110 0x76
#define B01110111 0x77
#define B01111000 0x78
#define B01111001 0x79
#define B01111010 0x7A
#define B01111011 0x7B
#define B01111100 0x7C
#define B01111101 0x7D
#define B01111110 0x7E
#define B01111111 0x7F
#define B10000000 0x80
#define B10000001 0x81
#define B10000010 0x82
#define B10000011 0x83
#define B10000100 0x84
#define B10000101 0x85
#define B10000110 0x86
#define B10000111 0x87
#define B10001000 0x88
#define B10001001 0x89
#define B10001010 0x8A
#define B10001011 0x8B
#define B10001100 0x8C
#define B10001101 0x8D
#define B10001110 0x8E
#define B10001111 0x8F
#define B10010000 0x90
#define B10010001 0x91
#define B10010010 0x92
#define B10010011 0x93
#define B10010100 0x94
#define B10010101 0x95
#define B10010110 0x96
#define B10010111 0x97
#define B10011000 0x98
#define B10011001 0x99
#define B10011010 0x9A
#define B10011011 0x9B
#define B10011100 0x9C
#define B10011101 0x9D
#define B10011110 0x9E
#define B10011111 0x9F
#define B10100000 0xA0
#define B10100001 0xA1
#define B10100010 0xA2
#define B10100011 0xA3
#define B10100100 0xA4
#define B10100101 0xA5
#define B10100110 0xA6
#define B10100111 0xA7
#define B10101000 0xA8
#define B10101001 0xA9
#define B10101010 0xAA
#define B10101011 0xAB
#define B10101100 0xAC
#define B10101101 0xAD
#define B10101110 0xAE
#define B10101111 0xAF
#define B10110000 0xB0
#define B10110001 0xB1
#define B10110010 0xB2
#define B10110011 0xB3
#define B10110100 0xB4
#define B10110101 0xB5
#define B10110110 0xB6
#define B10110111 0xB7
#define B10111000 0xB8
#define B10111001 0xB9
#define B10111010 0xBA
#define B10111011 0xBB
#define B10111100 0xBC
#define B10111101 0xBD
#define B10111110 0xBE
#define B10111111 0xBF
#define B11000000 0xC0
#define B11000001 0xC1
#define B11000010 0xC2
#define B11000011 0xC3
#define B11000100 0xC4
#define B11000101 0xC5
#define B11000110 0xC6
#define B11000111 0xC7
#define B11001000 0xC8
#define B11001001 0xC9
#define B11001010 0xCA
#define B11001011 0xCB
#define B11001100 0xCC
#define B11001101 0xCD
#define B11001110 0xCE
#define B11001111 0xCF
#define B11010000 0xD0
#define B11010001 0xD1
#define B11010010 0xD2
#define B11010011 0xD3
#define B11010100 0xD4
#define B11010101 0xD5
#define B11010110 0xD6
#define B11010111 0xD7
#define B11011000 0xD8
#define B11011001 0xD9
#define B11011010 0xDA
#define B11011011 0xDB
#define B11011100 0xDC
#define B11011101 0xDD
#define B11011110 0xDE
#define B11011111 0xDF
#define B11100000 0xE0
#define B11100001 0xE1
#define B11100010 0xE2
#define B11100011 0xE3
#define B11100100 0xE4
#define B11100101 0xE5
#define B11100110 0xE6
#define B11100111 0xE7
#define B11101000 0xE8
#define B11101001 0xE9
#define B11101010 0xEA
#define B11101011 0xEB
#define B11101100 0xEC
#define B11101101 0xED
#define B11101110 0xEE
#define B11101111 0xEF
#define B11110000 0xF0
#define B11110001 0xF1
#define B11110010 0xF2
#define B11110011 0xF3
#define B11110100 0xF4
#define B11110101 0xF5
#define B11110110 0xF6
#define B11110111 0xF7
#define B11111000 0xF8
#define B11111001 0xF9
#define B11111010 0xFA
#define B11111011 0xFB
#define B11111100 0xFC
#define B11111101 0xFD
#define B11111110 0xFE
#define B11111111 0xFF

// Macro to swap the endianness for a 16-bit short

#define SWAP_ENDIAN_SHORT(s)                \
   (((ushort)(((uchar*)&(s))[0]) << 8)  |   \
    ((ushort)(((uchar*)&(s))[1]) << 0))

// Macro to swap the endianness for a 32-bit long

#define SWAP_ENDIAN_LONG(l)                 \
   (((ulong)(((uchar*)&(l))[0]) << 24) |    \
    ((ulong)(((uchar*)&(l))[1]) << 16) |    \
    ((ulong)(((uchar*)&(l))[2]) << 8)  |    \
    ((ulong)(((uchar*)&(l))[3]) << 0))

// Macros to convert values in specific endian format

#ifdef  __BIG_ENDIAN__
#define CVT_SHORT_LE(s)             SWAP_ENDIAN_SHORT(s)
#define CVT_SHORT_BE(s)             (s)
#define CVT_LONG_LE(s)              SWAP_ENDIAN_LONG(s)
#define CVT_LONG_BE(s)              (s)
#define GET_SHORT_LE(base,index)    SWAP_ENDIAN_SHORT(((ushort*)base)[index])
#define GET_SHORT_BE(base,index)    (((ushort*)base)[index])
#define GET_LONG_LE(base,index)     SWAP_ENDIAN_LONG(((ulong*)base)[index])
#define GET_LONG_BE(base,index)     (((ulong*)base)[index])
#define PUT_SHORT_LE(base,index,v)  (((ushort*)base)[index]) = SWAP_ENDIAN_SHORT(v)
#define PUT_SHORT_BE(base,index,v)  (((ushort*)base)[index]) = (v)
#define PUT_LONG_LE(base,index,v)   (((ulong*)base)[index]) = SWAP_ENDIAN_LONG(v)
#define PUT_LONG_BE(base,index,v)   (((ulong*)base)[index]) = (v)
#else
#define CVT_SHORT_LE(s)             (s)
#define CVT_SHORT_BE(s)             SWAP_ENDIAN_SHORT(s)
#define CVT_LONG_LE(s)              (s)
#define CVT_LONG_BE(s)              SWAP_ENDIAN_LONG(s)
#define GET_SHORT_LE(base,index)    (((ushort*)base)[index])
#define GET_SHORT_BE(base,index)    SWAP_ENDIAN_SHORT(((ushort*)base)[index])
#define GET_LONG_LE(base,index)     (((ulong*)base)[index])
#define GET_LONG_BE(base,index)     SWAP_ENDIAN_LONG(((ulong*)base)[index])
#define PUT_SHORT_LE(base,index,v)  (((ushort*)base)[index]) = (v)
#define PUT_SHORT_BE(base,index,v)  (((ushort*)base)[index]) = SWAP_ENDIAN_SHORT(v)
#define PUT_LONG_LE(base,index,v)   (((ulong*)base)[index]) = (v)
#define PUT_LONG_BE(base,index,v)   (((ulong*)base)[index]) = SWAP_ENDIAN_LONG(v)
#endif

// Macro to determine if the VGA is enabled and responding

#define VGA_NOT_ACTIVE()    ((inpb(0x3CC) == 0xFF) || ((inpb(0x3CC) & B00000010) == 0))

// Union to hold CPUID information.

typedef union {
    struct {
        ulong   eax;
        ulong   ebx;
        ulong   edx;
        ulong   ecx;
        } r;
    struct {
        ulong   numLevels;
        char    str[12];
        } id;
    } CPUID_t;

// Macro to block fill a block of memory, expanding to DWORD's and
// DWORD aligning the destination address for maximum performance.
// The macro assumes that the 8-bit color value has been pre-expanded
// to fill an entire 32-bit DWORD (ie: color is 32-bits wide!)
//
// On return, addr contains the address of the next byte to be processed

#define DWORD_FILL(addr,bytes,color)                \
{                                                   \
    if (bytes >= 8) {                               \
        while (((uint)addr & 3) != 0) {             \
            *addr++ = color;                        \
            bytes--;                                \
            }                                       \
        addr = rep_stosd(addr,bytes >> 2,color);    \
        addr = rep_stosb(addr,bytes & 3,color);     \
        }                                           \
    else                                            \
        addr = rep_stosb(addr,bytes,color);         \
}

/*-------------------------- Function Prototypes --------------------------*/

// Inline assembler functions for Watcom C++

#if defined(__WATCOMC__)

uchar disable(void);
#pragma aux disable =               \
    "cli"                           \
    modify exact [];

uchar enable(void);
#pragma aux enable =                \
    "sti"                           \
    modify exact [];

void _outpb(ulong port,uchar val);
#pragma aux _outpb =                \
    "out    dx,al"                  \
    parm [edx] [eax]                \
    modify exact [];

void _outpw(ulong port,ushort val);
#pragma aux _outpw =                \
    "out    dx,ax"                  \
    parm [edx] [eax]                \
    modify exact [];

void _outpd(ulong port,ulong val);
#pragma aux _outpd =                \
    "out    dx,eax"                 \
    parm [edx] [eax]                \
    modify exact [];

uchar _inpb(ulong port);
#pragma aux _inpb =                 \
    "in     al,dx"                  \
    parm [edx]                      \
    value [al]                      \
    modify exact [al];

ushort _inpw(ulong port);
#pragma aux _inpw =                 \
    "in     ax,dx"                  \
    parm [edx]                      \
    value [ax]                      \
    modify exact [ax];

ulong _inpd(ulong port);
#pragma aux _inpd =                 \
    "in     eax,dx"                 \
    parm [edx]                      \
    value [eax]                     \
    modify exact [eax];

uchar _rdinx(ulong port,uchar index);
#pragma aux _rdinx =                \
    "out    dx,al"                  \
    "inc    edx"                    \
    "in     al,dx"                  \
    parm [edx] [al]                 \
    value [al]                      \
    modify exact [edx al];

void _wrinx(ulong port,uchar index,uchar value);
#pragma aux _wrinx =                \
    "out    dx,ax"                  \
    parm [edx] [al] [ah]            \
    modify exact [];

uchar rolb(uchar val,uchar shift);
#pragma aux rolb =                  \
    "rol    al,cl"                  \
    parm [al] [cl]                  \
    value [al]                      \
    modify exact [al cl];

ushort rolb_w(uchar val,uchar shift);
#pragma aux rolb_w =                \
    "rol    al,cl"                  \
    "mov    ah,al"                  \
    parm [al] [cl]                  \
    value [ax]                      \
    modify exact [ax cl];

ushort rolw(ushort val,uchar shift);
#pragma aux rolw =                  \
    "rol    ax,cl"                  \
    parm [ax] [cl]                  \
    value [ax]                      \
    modify exact [ax cl];

ulong rolw_d(ushort val,uchar shift);
#pragma aux rolw_d =                \
    "rol    ax,cl"                  \
    "mov    edx,eax"                \
    "shl    eax,16"                 \
    "mov    ax,dx"                  \
    parm [ax] [cl]                  \
    value [eax]                     \
    modify exact [ax cl edx];

ulong rold(ulong val,uchar shift);
#pragma aux rold =                  \
    "rol    eax,cl"                 \
    parm [eax] [cl]                 \
    value [eax]                     \
    modify exact [eax cl];

uchar rorb(uchar val,uchar shift);
#pragma aux rorb =                  \
    "ror    al,cl"                  \
    parm [al] [cl]                  \
    value [al]                      \
    modify exact [al cl];

ushort rorb_w(uchar val,uchar shift);
#pragma aux rorb_w =                \
    "ror    al,cl"                  \
    "mov    ah,al"                  \
    parm [al] [cl]                  \
    value [ax]                      \
    modify exact [ax cl];

ushort rorw(ushort val,uchar shift);
#pragma aux rorw =                  \
    "ror    ax,cl"                  \
    parm [ax] [cl]                  \
    value [ax]                      \
    modify exact [ax cl];

ulong rord(ulong val,uchar shift);
#pragma aux rord =                  \
    "ror    eax,cl"                 \
    parm [eax] [cl]                 \
    value [eax]                     \
    modify exact [eax cl];

ulong rorw_d(ushort val,uchar shift);
#pragma aux rorw_d =                \
    "ror    ax,cl"                  \
    "mov    edx,eax"                \
    "shl    eax,16"                 \
    "mov    ax,dx"                  \
    parm [ax] [cl]                  \
    value [eax]                     \
    modify exact [ax cl edx];

void ror_set_pattern(void *addr,void *pattern,int shift);
#pragma aux ror_set_pattern =   \
    "mov    al,[esi+0]"         \
    "mov    ah,[esi+1]"         \
    "mov    bl,[esi+2]"         \
    "mov    bh,[esi+3]"         \
    "ror    al,cl"              \
    "ror    ah,cl"              \
    "ror    bl,cl"              \
    "ror    bh,cl"              \
    "mov    [edi+0],al"         \
    "mov    [edi+1],ah"         \
    "mov    [edi+2],bl"         \
    "mov    [edi+3],bh"         \
    "mov    al,[esi+4]"         \
    "mov    ah,[esi+5]"         \
    "mov    bl,[esi+6]"         \
    "mov    bh,[esi+7]"         \
    "ror    al,cl"              \
    "ror    ah,cl"              \
    "ror    bl,cl"              \
    "ror    bh,cl"              \
    "mov    [edi+4],al"         \
    "mov    [edi+5],ah"         \
    "mov    [edi+6],bl"         \
    "mov    [edi+7],bh"         \
    parm [edi] [esi] [cl]       \
    modify exact [ax bx cl esi edi];

void rep_outsb(ulong port,uchar *src,int count);
#pragma aux rep_outsb =             \
    "rep outsb"                     \
    parm [edx] [esi] [ecx]          \
    modify exact [dx esi ecx];

void rep_outsw(ulong port,uchar *src,int count);
#pragma aux rep_outsw =             \
    "rep outsw"                     \
    parm [edx] [esi] [ecx]          \
    modify exact [dx esi ecx];

void rep_outsd(ulong port,uchar *src,int count);
#pragma aux rep_outsd =             \
    "rep outsd"                     \
    parm [edx] [esi] [ecx]          \
    modify exact [dx esi ecx];

void rep_outsdb(ulong port,void *src,int count);
#pragma aux rep_outsdb =            \
    "mov ecx,eax"                   \
    "shr ecx,2"                     \
    "rep outsd"                     \
    "mov ecx,eax"                   \
    "and ecx,3"                     \
    "jz  Exit"                      \
    "mov al,[esi]"                  \
    "dec ecx"                       \
    "jz  Done"                      \
    "mov ah,[esi+1]"                \
    "dec ecx"                       \
    "jz  Done"                      \
    "rol eax,16"                    \
    "out dx,eax"                    \
    "rol eax,16"                    \
"Done: mov [edi],eax"               \
"Exit:"                             \
    parm [edx] [esi] [eax]          \
    modify exact [dx esi ecx];

void * rep_movsb(void *dst,void *src,int count);
#pragma aux rep_movsb =             \
    "rep movsb"                     \
    parm [edi] [esi] [ecx]          \
    value [edi]                     \
    modify exact [edi esi ecx];

void * rep_movsw(void *dst,void *src,int count);
#pragma aux rep_movsw =             \
    "rep movsw"                     \
    parm [edi] [esi] [ecx]          \
    value [edi]                     \
    modify exact [edi esi ecx];

void * rep_movsd(void *dst,void *src,int count);
#pragma aux rep_movsd =             \
    "rep movsd"                     \
    parm [edi] [esi] [ecx]          \
    value [edi]                     \
    modify exact [edi esi ecx];

void rep_movsdb(void *dst,void *src,int count);
#pragma aux rep_movsdb =            \
    "mov ecx,eax"                   \
    "shr ecx,2"                     \
    "rep movsd"                     \
    "mov ecx,eax"                   \
    "and ecx,3"                     \
    "jz  Exit"                      \
    "mov al,[esi]"                  \
    "dec ecx"                       \
    "jz  Done"                      \
    "mov ah,[esi+1]"                \
    "dec ecx"                       \
    "jz  Done"                      \
    "rol eax,16"                    \
    "mov al,[esi+2]"                \
    "rol eax,16"                    \
"Done: mov [edi],eax"               \
"Exit:"                             \
    parm [edi] [esi] [eax]          \
    modify exact [edi esi eax ecx];

void movsdb(void *dst,void *src,int count);
#pragma aux movsdb =                \
    "and ecx,3"                     \
    "jz  Exit"                      \
    "mov al,[esi]"                  \
    "dec ecx"                       \
    "jz  Done"                      \
    "mov ah,[esi+1]"                \
    "dec ecx"                       \
    "jz  Done"                      \
    "rol eax,16"                    \
    "mov al,[esi+2]"                \
    "rol eax,16"                    \
"Done: mov [edi],eax"               \
"Exit:"                             \
    parm [edi] [esi] [ecx]          \
    modify exact [edi esi ecx];

void * rep_movsdw(void *dst,void *src,int count);
#pragma aux rep_movsdw =            \
    "mov ecx,eax"                   \
    "shr ecx,1"                     \
    "rep movsd"                     \
    "mov ecx,eax"                   \
    "and ecx,1"                     \
    "rep movsw"                     \
    parm [edi] [esi] [eax]          \
    modify exact [edi esi eax ecx];

void * rep_stosb(void *dst,int count,uchar value);
#pragma aux rep_stosb =             \
    "rep stosb"                     \
    parm [edi] [ecx] [al]           \
    value [edi]                     \
    modify exact [edi ecx al];

void * rep_stosw(void *dst,int count,ushort value);
#pragma aux rep_stosw =             \
    "rep stosw"                     \
    parm [edi] [ecx] [ax]           \
    value [edi]                     \
    modify exact [edi ecx ax];

void * rep_stosd(void *dst,int count,ulong value);
#pragma aux rep_stosd =             \
    "rep stosd"                     \
    parm [edi] [ecx] [eax]          \
    value [edi]                     \
    modify exact [edi ecx eax];

void * rep_stosdb(void *dst,int count,ulong value);
#pragma aux rep_stosdb =            \
    "mov ecx,esi"                   \
    "shr ecx,2"                     \
    "rep stosd"                     \
    "mov ecx,esi"                   \
    "and ecx,3"                     \
    "rep stosb"                     \
    parm [edi] [esi] [eax]          \
    modify exact [edi esi eax ecx];

void * rep_stosdw(void *dst,int count,ulong value);
#pragma aux rep_stosdw =            \
    "mov ecx,esi"                   \
    "shr ecx,1"                     \
    "rep stosd"                     \
    "mov ecx,esi"                   \
    "and ecx,1"                     \
    "rep stosw"                     \
    parm [edi] [esi] [eax]          \
    modify exact [edi esi eax ecx];

void rep_mirrorsdb(void *dst,void *src,void *mirror,int count);
#pragma aux rep_mirrorsdb =         \
    "push ecx"                      \
    "xor edx,edx"                   \
    "shr ecx,2"                     \
    "jcxz ShortRun"                 \
"MainLoop:"                         \
    "lodsd"                         \
    "mov dl,al"                     \
    "mov al,[ebx+edx]"              \
    "mov dl,ah"                     \
    "mov ah,[ebx+edx]"              \
    "rol eax,16"                    \
    "mov dl,al"                     \
    "mov al,[ebx+edx]"              \
    "mov dl,ah"                     \
    "mov ah,[ebx+edx]"              \
    "rol eax,16"                    \
    "stosd"                         \
    "dec ecx"                       \
    "jnz MainLoop"                  \
"ShortRun:"                         \
    "pop ecx"                       \
    "and ecx,3"                     \
    "jz  Exit"                      \
    "mov dl,[esi]"                  \
    "mov al,[ebx+edx]"              \
    "dec ecx"                       \
    "jz  Done"                      \
    "mov dl,[esi+1]"                \
    "mov ah,[ebx+edx]"              \
    "dec ecx"                       \
    "jz  Done"                      \
    "rol eax,16"                    \
    "mov dl,[esi+2]"                \
    "mov al,[ebx+edx]"              \
    "rol eax,16"                    \
"Done: mov [edi],eax"               \
"Exit:"                             \
    parm [edi] [esi] [ebx] [ecx]    \
    modify exact [edi esi eax ebx ecx edx];

N_fix32 FXmul(N_fix32 a,N_fix32 b);
#pragma aux FXmul =                 \
    "imul   edx"                    \
    "add    eax,8000h"              \
    "adc    edx,0"                  \
    "shrd   eax,edx,16"             \
    parm [eax] [edx]                \
    value [eax]                     \
    modify exact [eax edx];

N_fix32 FXdiv(N_fix32 dividend,N_fix32 divisor);
#pragma aux FXdiv =                 \
    "xor    eax,eax"                \
    "shrd   eax,edx,16"             \
    "sar    edx,16"                 \
    "idiv   ebx"                    \
    parm [edx] [ebx]                \
    value [eax]                     \
    modify exact [eax edx];

N_fix32 FXmuldiv(N_fix32 a,N_fix32 b,N_fix32 c);
#pragma aux FXmuldiv =              \
    "imul   ebx"                    \
    "idiv   ecx"                    \
    parm [eax] [ebx] [ecx]          \
    value [eax]                     \
    modify exact [eax edx];

N_fix32 FXoneOver(N_fix32 f);
#pragma aux FXoneOver =             \
    "xor    eax,eax"                \
    "mov    edx,1"                  \
    "idiv   ebx"                    \
    parm [ebx]                      \
    value [eax]                     \
    modify exact [eax edx];

N_fix32 FXsquare(N_fix32 f);
#pragma aux FXsquare =              \
    "imul   eax"                    \
    "add    eax,8000h"              \
    "adc    edx,0"                  \
    "shrd   eax,edx,16"             \
    parm [eax]                      \
    value [eax]                     \
    modify exact [edx];

void *memcpy(void *dst,const void *src,size_t n);
#pragma aux memcpy =                \
    "mov    eax,ecx"                \
    "shr    ecx,2"                  \
    "rep    movsd"                  \
    "mov    cl,al"                  \
    "and    cl,3"                   \
    "rep    movsb"                  \
    parm [edi] [esi] [ecx]          \
    value [edi]                     \
    modify exact [eax ecx esi edi];

void *blast_words(void *dst,const void *src,size_t n);
#pragma aux blast_words =           \
    "shr    ecx,1"                  \
    "rep    movsd"                  \
    "rcl    ecx,1"                  \
    "rep    movsw"                  \
    parm [edi] [esi] [ecx]          \
    value [edi]                     \
    modify exact [ecx esi edi];

void CPUID(ulong EAX,CPUID_t *data);
#pragma aux CPUID =                 \
    "db 0Fh"                        \
    "db 0A2h"                       \
    "mov    [esi],eax"              \
    "mov    [esi+4],ebx"            \
    "mov    [esi+8],edx"            \
    "mov    [esi+12],ecx"           \
    parm [eax] [esi]                \
    modify exact [eax ebx ecx edx esi];

ulong mkcolor4(ulong);
#pragma aux mkcolor4 =              \
    "and    al,0Fh"                 \
    "mov    ah,al"                  \
    "shl    ah,4"                   \
    "or     al,ah"                  \
    "or     ah,al"                  \
    "movzx  edx,ax"                 \
    "shl    eax,16"                 \
    "or     edx,eax"                \
    parm [eax]                      \
    value [edx]                     \
    modify exact [eax edx];

ulong mkcolor8(ulong);
#pragma aux mkcolor8 =              \
    "mov    ah,al"                  \
    "movzx  edx,ax"                 \
    "shl    eax,16"                 \
    "or     edx,eax"                \
    parm [eax]                      \
    value [edx]                     \
    modify exact [eax edx];

ulong mkcolor16(ulong);
#pragma aux mkcolor16 =             \
    "movzx  edx,ax"                 \
    "shl    eax,16"                 \
    "or     edx,eax"                \
    parm [eax]                      \
    value [edx]                     \
    modify exact [eax edx];

ushort mkcolor8_w(ulong);
#pragma aux mkcolor8_w =            \
    "mov    ah,al"                  \
    parm [eax]                      \
    modify exact [eax];

ushort swap_bytes(ulong);
#pragma aux swap_bytes =            \
    "xchg   ah,al"                  \
    parm [eax]                      \
    modify exact [eax];

uchar rolb1(uchar *val);
#pragma aux rolb1 =                 \
    "rol    byte ptr [ebx],1"       \
    "setc   al"                     \
    parm [ebx]                      \
    value [al]                      \
    modify exact [al];

uchar rorb1(uchar *val);
#pragma aux rorb1 =                 \
    "ror    byte ptr [ebx],1"       \
    "setc   al"                     \
    parm [ebx]                      \
    value [al]                      \
    modify exact [al];

void rwb(uchar *addr);
#pragma aux rwb =                   \
    "mov    al,[ebx]"               \
    "mov    [ebx],al"               \
    parm [ebx]                      \
    modify exact [ebx al];

void rwbv(uchar *addr,uchar v);
#pragma aux rwbv =                  \
    "mov    al,[ebx]"               \
    "mov    [ebx],ah"               \
    parm [ebx] [ah]                 \
    modify exact [ebx al ah];

#else /* !__WATCOMC__ */

#undef	disable
#undef	enable
uchar 	_ASMAPI disable(void);
uchar 	_ASMAPI enable(void);
void 	_ASMAPI _outpb(ulong port,uchar val);
void 	_ASMAPI _outpw(ulong port,ushort val);
void 	_ASMAPI _outpd(ulong port,ulong val);
uchar 	_ASMAPI _inpb(ulong port);
ushort 	_ASMAPI _inpw(ulong port);
ulong 	_ASMAPI _inpd(ulong port);
uchar	_rdinx(ulong port,uchar index);
void 	_wrinx(ulong port,uchar index,uchar value);
uchar 	_ASMAPI rolb(uchar val,uchar shift);
ushort 	_ASMAPI rolb_w(uchar val,uchar shift);
ushort 	_ASMAPI rolw(ushort val,uchar shift);
ulong 	_ASMAPI rolw_d(ushort val,uchar shift);
ulong 	_ASMAPI rold(ulong val,uchar shift);
uchar 	_ASMAPI rorb(uchar val,uchar shift);
ushort 	_ASMAPI rorb_w(uchar val,uchar shift);
ushort 	_ASMAPI rorw(ushort val,uchar shift);
ulong 	_ASMAPI rord(ulong val,uchar shift);
ulong 	_ASMAPI rorw_d(ushort val,uchar shift);
void 	_ASMAPI ror_set_pattern(void *addr,void *pattern,int shift);
void 	_ASMAPI rep_outsb(ulong port,uchar *src,int count);
void 	_ASMAPI rep_outsw(ulong port,uchar *src,int count);
void 	_ASMAPI rep_outsd(ulong port,uchar *src,int count);
void 	_ASMAPI rep_outsdb(ulong port,void *src,int count);
void * 	_ASMAPI rep_movsb(void *dst,void *src,int count);
void * 	_ASMAPI rep_movsw(void *dst,void *src,int count);
void * 	_ASMAPI rep_movsd(void *dst,void *src,int count);
void * 	_ASMAPI rep_movsd_swhw(void *dst,void *src,int count);
void * 	_ASMAPI rep_movsd_le(void *dst,void *src,int count);
void * 	_ASMAPI rep_movsw_le(void *dst,void *src,int count);
void  	_ASMAPI rep_movsdb(void *dst,void *src,int count);
void  	_ASMAPI rep_movsdw(void *dst,void *src,int count);
void * 	_ASMAPI rep_stosb(void *dst,int count,uchar value);
void * 	_ASMAPI rep_stosw(void *dst,int count,ushort value);
void * 	_ASMAPI rep_stosd(void *dst,int count,ulong value);
void * 	_ASMAPI rep_stosdb(void *dst,int count,ulong value);
void * 	_ASMAPI rep_stosdw(void *dst,int count,ulong value);
void 	_ASMAPI movsdb(void *dst,void *src,int count);
void   	_ASMAPI rep_mirrorsdb(void *dst,void *src,void *mirror,int count);
N_fix32 _ASMAPI FXmul(N_fix32 a,N_fix32 b);
N_fix32 _ASMAPI FXdiv(N_fix32 dividend,N_fix32 divisor);
N_fix32 _ASMAPI FXmuldiv(N_fix32 a,N_fix32 b,N_fix32 c);
N_fix32 _ASMAPI FXoneOver(N_fix32 f);
N_fix32 _ASMAPI FXsquare(N_fix32 f);
void *  _ASMAPI blast_words(void *dst,const void *src,size_t n);
void    _ASMAPI _CPUID(ulong EAX,CPUID_t *data);
#define CPUID   _CPUID
ulong   _ASMAPI mkcolor4(ulong);
ulong   _ASMAPI mkcolor8(ulong);
ulong   _ASMAPI mkcolor16(ulong);
ushort  _ASMAPI mkcolor8_w(ulong);
ushort  _ASMAPI swap_bytes(ulong);
uchar   _ASMAPI rolb1(uchar *val);
uchar   _ASMAPI rorb1(uchar *val);
void    _ASMAPI rwb(uchar *addr);
void    _ASMAPI rwbv(uchar *addr,uchar v);

#endif  /* !__WATCOMC__ */

// Macro to expand 1bpp color values to 32-bit values

#define mkcolor1(c) ((c) ? 0xFFFFFFFF : 0x00000000)

// Helper functions

void    modinx(ulong port,uchar index,uchar mask,uchar value);
void    setinx(ulong port,uchar index,uchar mask);
void    clrinx(ulong port,uchar index,uchar mask);
void    modreg(ulong port,uchar mask,uchar value);
void    setreg(ulong port,uchar mask);
void    clrreg(ulong port,uchar mask);
void    modregw(ulong port,ushort mask,ushort value);
void    setregw(ulong port,ushort mask);
void    clrregw(ulong port,ushort mask);
void    modregd(ulong port,ulong mask,ulong value);
void    setregd(ulong port,ulong mask);
void    clrregd(ulong port,ulong mask);
ibool   tstinx(ulong port,ushort index,ushort mask);
ibool   tstreg(ulong port,ushort mask);
uchar   rdinx_m(ulong reg,uchar index);
void    wrinx_m(ulong reg,uchar index,uchar value);
void    modinx_m(ulong reg,uchar index,uchar mask,uchar value);
void    setinx_m(ulong reg,uchar index,uchar mask);
void    clrinx_m(ulong reg,uchar index,uchar mask);

// Allow for relocation of I/O port writes by letting the device driver
// take over the inp/outp functions.

#ifndef RELOCATE_FIXEDIO
#ifdef VEM_PORTIO
#define outpb(port,val)                 VEM_outpb(port,val)
#define outpw(port,val)                 VEM_outpw(port,val)
#define outpd(port,val)                 VEM_outpd(port,val)
#define inpb(port)                      VEM_inpb(port)
#define inpw(port)                      VEM_inpw(port)
#define inpd(port)                      VEM_inpd(port)
#define rdinx(port,index)               VEM_rdinx(port,index)
#define wrinx(port,index,value)         VEM_wrinx(port,index,value)
#define modinx(port,index,mask,value)   VEM_modinx(port,index,mask,value)
#define setinx(port,index,mask)         VEM_setinx(port,index,mask)
#define clrinx(port,index,mask)         VEM_clrinx(port,index,mask)
#else
#define outpb(port,val)         _outpb(port,val)
#define outpw(port,val)         _outpw(port,val)
#define outpd(port,val)         _outpd(port,val)
#define inpb(port)              _inpb(port)
#define inpw(port)              _inpw(port)
#define inpd(port)              _inpd(port)
#define rdinx(port,index)       _rdinx(port,index)
#define wrinx(port,index,value) _wrinx(port,index,value)
#endif
#else
void    outpb(ulong port,uchar val);
void    outpw(ulong port,ushort val);
void    outpd(ulong port,ulong val);
uchar   inpb(ulong port);
ushort  inpw(ulong port);
ulong   inpd(ulong port);
uchar   rdinx(ulong port,uchar index);
void    wrinx(ulong port,uchar index,uchar value);
#endif

#endif  /* __SNAP_COMMON_IO_H */

