/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Internal header for for SNAP linked list management. This
*               module is common to all SciTech SNAP device driver types.
*
****************************************************************************/

#ifndef __SNAP_COMMON_LIST_H
#define __SNAP_COMMON_LIST_H

#include "snap/common/internal.h"

/*---------------------- Macros and type definitions ----------------------*/

#pragma pack(1)

// Return a pointer to the user space given the address of the header of
// a node.

#define LST_USERSPACE(h)    ((void*)((N_node*)(h) + 1))

// Return a pointer to the header of a node, given the address of the
// user space.

#define LST_HEADER(n)       ((N_node*)(n) - 1)

// Return a pointer to the user space of the list's head node. This user
// space does not actually exist, but it is useful to be able to address
// it to enable insertion at the start of the list.

#define LST_HEAD(l)         LST_USERSPACE((l)->head)

// Return a pointer to the user space of the last node in list.	

#define LST_TAIL(l)         LST_USERSPACE((l)->z->prev)

// Determine if a list is empty

#define LST_EMPTY(l)        ((l)->count == 0)

/****************************************************************************
REMARKS:
Structure for the list header structure for tracking objects.

MEMBERS:
next    - Pointer to next object in the list
prev    - Pointer to previous object in the list
****************************************************************************/
typedef struct N_node {
    struct N_node   *next;
    struct N_node   *prev;
    } N_node;

/****************************************************************************
REMARKS:
Structure for the the linked list itself.

MEMBERS:
count   - Number of elements currently in list
head    - Pointer to head element of list
z       - Pointer to last node of list
hz      - Space for head and z nodes
****************************************************************************/
typedef struct {
    int     count;
    N_node  *head;
    N_node  *z;
    N_node  hz[2];
    } N_list;

#pragma pack()

/*-------------------------- Function Prototypes --------------------------*/

#ifdef  __cplusplus
extern "C" {                        /* Use "C" linkage when in C++ mode */
#endif

void    NAPI N_initList(N_list *l);
void    NAPI N_destroyList(N_list *l,void (*freeNode)(void *node));
void    NAPI N_addAfter(N_list *l,void *node,void *after);
void    NAPI N_addBefore(N_list *l,void *node,void *before);
void    NAPI N_addToHead(N_list *l,void *node);
void    NAPI N_addToTail(N_list *l,void *node);
void *  NAPI N_remove(N_list *l,void *node);
void *  NAPI N_removeNext(N_list *l,void *node);
void *  NAPI N_firstNode(N_list *l);
void *  NAPI N_lastNode(N_list *l);
void *  NAPI N_nextNode(void *prev);
void *  NAPI N_prevNode(void *next);

#ifdef  __cplusplus
}                                   /* End of "C" linkage for C++       */
#endif

#endif  /* __SNAP_COMMON_LIST_H */

