/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Internal header for SNAP block memory management. This
*               module is common to all SciTech SNAP device driver types.
*
****************************************************************************/

#ifndef __SNAP_COMMON_MEM_H
#define __SNAP_COMMON_MEM_H

#include "snap/common/internal.h"

/*---------------------- Macros and type definitions ----------------------*/

#pragma pack(1)

// Macro to determine the size of blocks allocated on the block list.
// We allocate the blocks in large chunks to avoid memory management
// overheads, and make the blocks multiples of a memory page.

#define N_BLOCK_SIZE            4096

// Macro to get the to the next block on the freelist

#define N_FREELIST_NEXT(p)      (*(void**)(p))

/****************************************************************************
REMARKS:
Structure for the header block for blocks on the block list.

MEMBERS:
next    - Pointer to next block in list
****************************************************************************/
typedef struct N_block {
    struct N_block *next;
    } N_block;

/****************************************************************************
REMARKS:
Structure for the header block for blocks on the block list.

MEMBERS:
curBlock    - List of memory pool blocks
blockCount  - Number of pool blocks allocated
max         - Maximum objects in block
objectSize  - Size of objects in block
****************************************************************************/
typedef struct {
    N_block *curBlock;
    int     blockCount;
    int     max;
    int     objectSize;
    } N_blockList;

#pragma pack()

/*-------------------------- Function Prototypes --------------------------*/

#ifdef  __cplusplus
extern "C" {                        /* Use "C" linkage when in C++ mode */
#endif

ibool   NAPI N_createBlockList(N_blockList *l,int objectSize);
ibool   NAPI N_resizeBlockList(N_blockList *l);
void    NAPI N_freeBlockList(N_blockList *l);
void *  NAPI N_buildFreeList(N_blockList *l);

#ifdef  __cplusplus
}                                   /* End of "C" linkage for C++       */
#endif

#endif  /* __SNAP_COMMON_MEM_H */

