/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Common header for C based accelerator code.
*
****************************************************************************/

#ifndef __SNAP_GRAPHICS_ACCEL_H
#define __SNAP_GRAPHICS_ACCEL_H

/*---------------------- Macros and type definitions ----------------------*/

// Source and destination ROP3 codes

#define ROP3_BLACK              0x00UL  // 0
#define ROP3_SRC_NOTMERGESRC    0x11UL  // DSon
#define ROP3_SRC_MASKNOTSRC     0x22UL  // DSna
#define ROP3_SRC_NOTCOPYSRC     0x33UL  // Sn
#define ROP3_SRC_MASKSRCNOT     0x44UL  // SDna
#define ROP3_NOT                0x55UL  // Dn
#define ROP3_SRC_XORSRC         0x66UL  // DSx
#define ROP3_SRC_NOTMASKSRC     0x77UL  // DSan
#define ROP3_SRC_MASKSRC        0x88UL  // DSa
#define ROP3_SRC_NOTXORSRC      0x99UL  // DSxn
#define ROP3_NOP                0xAAUL  // D
#define ROP3_SRC_MERGENOTSRC    0xBBUL  // DSno
#define ROP3_SRC_COPYSRC        0xCCUL  // S
#define ROP3_SRC_MERGESRCNOT    0xDDUL  // SDno
#define ROP3_SRC_MERGESRC       0xEEUL  // DSo
#define ROP3_WHITE              0xFFUL  // 1

// Pattern and destination ROP3 codes

#define ROP3_BLACK              0x00UL  // 0
#define ROP3_PATT_NOTMERGESRC   0x05UL  // DPon
#define ROP3_PATT_MASKNOTSRC    0x0AUL  // DPna
#define ROP3_PATT_NOTCOPYSRC    0x0FUL  // Pn
#define ROP3_PATT_MASKSRCNOT    0x50UL  // PDna
#define ROP3_NOT                0x55UL  // Dn
#define ROP3_PATT_XORSRC        0x5AUL  // DPx
#define ROP3_PATT_NOTMASKSRC    0x5FUL  // DPan
#define ROP3_PATT_MASKSRC       0xA0UL  // DPa
#define ROP3_PATT_NOTXORSRC     0xA5UL  // DPxn
#define ROP3_NOP                0xAAUL  // D
#define ROP3_PATT_MERGENOTSRC   0xAFUL  // DPno
#define ROP3_PATT_COPYSRC       0xF0UL  // P
#define ROP3_PATT_MERGESRCNOT   0xF5UL  // PDno
#define ROP3_PATT_MERGESRC      0xFAUL  // DPo
#define ROP3_WHITE              0xFFUL  // 1

// Macro to determine if a ROP3 code includes a source

#define ROP3_HAS_SOURCE(rop3)   (((rop3 >> 2) ^ rop3) & 0x33)
#define ROP2_HAS_SOURCE(rop)    (((rop >> 2) ^ rop) & 0x03)

// Macros to access the RGB components in a packed 24 bit RGB tuple

#define BYTE0(c)        (((uchar*)&(c))[0])
#define BYTE1(c)        (((uchar*)&(c))[1])
#define BYTE2(c)        (((uchar*)&(c))[2])

// Macros to load the mix type given the current write mode operation.

#define LOADMIX(mix)        (mixTranslate[mix])
#define LOADPATTMIX(mix)    (pattMixTranslate[mix])
#define LOADLINEMIX(mix)    (lineMixTranslate[mix])

// Macros to swap two values

#define SWAP(a,b)   { a^=b; b^=a; a^=b; }

#define SWAPT(v1,v2,t) {    \
    t = v1;                 \
    v1 = v2;                \
    v2 = t;                 \
    }

#define MAKEXY(x,y)     (((ulong)(x) << 16) | (ulong)(y))

// Macros to access color values for 24bpp modes

#define Red(c)          (((uchar*)&(c))[0])
#define Green(c)        (((uchar*)&(c))[1])
#define Blue(c)         (((uchar*)&(c))[2])

// VESA Approved 8514/A Register Definitions

#define DAC_MASK        0x02ea
#define DAC_R_INDEX     0x02eb
#define DAC_W_INDEX     0x02ec
#define DAC_DATA        0x02ed
#define DISP_STAT       0x02e8
#define H_TOTAL         0x02e8
#define H_DISP          0x06e8
#define H_SYNC_STRT     0x0ae8
#define H_SYNC_WID      0x0ee8
#define V_TOTAL         0x12e8
#define V_DISP          0x16e8
#define V_SYNC_STRT     0x1ae8
#define V_SYNC_WID      0x1ee8
#define DISP_CNTL       0x22e8
#define ADVFUNC_CNTL    0x4ae8
#define SUBSYS_STAT     0x42e8
#define SUBSYS_CNTL     0x42e8
#define ROM_PAGE_SEL    0x46e8
#define CUR_Y           0x82e8
#define CUR_X           0x86e8
#define DESTY_AXSTP     0x8ae8
#define DESTX_DIASTP    0x8ee8
#define ERR_TERM        0x92e8
#define MAJ_AXIS_PCNT   0x96e8
#define GP_STAT         0x9ae8
#define CMD             0x9ae8
#define SHORT_STROKE    0x9ee8
#define BKGD_COLOR      0xa2e8
#define FRGD_COLOR      0xa6e8
#define WRT_MASK        0xaae8
#define RD_MASK         0xaee8
#define COLOR_CMP       0xb2e8
#define BKGD_MIX        0xb6e8
#define FRGD_MIX        0xbae8
#define MULTIFUNC_CNTL  0xbee8
#define MIN_AXIS_PCNT   0x0000
#define SCISSORS_T      0x1000
#define SCISSORS_L      0x2000
#define SCISSORS_B      0x3000
#define SCISSORS_R      0x4000
#define PATTERN_L       0x8000
#define PATTERN_H       0x9000
#define PIX_CNTL        0xa000
#define PIX_TRANS       0xe2e8

// Graphics Processor Status Register

#define GPBUSY          0x0200
#define DATDRDY         0x0100

// Command Register

#define CMD_NOP         0x0000
#define CMD_LINE        0x2000
#define CMD_RECT        0x4000
#define CMD_BITBLT      0xc200
#define CMD_OP_MSK      0xf000
#define BYTSEQ          0x1000
#define _16BIT          0x0200
#define PCDATA          0x0100
#define INC_Y           0x0080
#define YMAJAXIS        0x0040
#define INC_X           0x0020
#define DRAW            0x0010
#define LINETYPE        0x0008
#define LASTPIX         0x0004
#define PLANAR          0x0002
#define WRTDATA         0x0001

// Short Stroke Vector Transfer Register
// (The angular Defs also apply to the Command Register)

#define VECDIR_000      0x0000
#define VECDIR_045      0x0020
#define VECDIR_090      0x0040
#define VECDIR_135      0x0060
#define VECDIR_180      0x0080
#define VECDIR_225      0x00a0
#define VECDIR_270      0x00c0
#define VECDIR_315      0x00e0
#define SSVDRAW         0x0010

// Background Mix Register

#define BSS_BKGDCOL     0x0000
#define BSS_FRGDCOL     0x0020
#define BSS_PCDATA      0x0040
#define BSS_BITBLT      0x0060

// Foreground Mix Register

#define FSS_BKGDCOL     0x0000
#define FSS_FRGDCOL     0x0020
#define FSS_PCDATA      0x0040
#define FSS_BITBLT      0x0060

// Standard 8514/A compatible mixes

#define MIX_MASK                        0x001f

#define MIX_BLACK               0x01UL  // 0
#define MIX_NOTMERGESRC         0x0FUL  // DSon
#define MIX_MASKNOTSRC          0x0EUL  // DSna
#define MIX_NOTCOPYSRC          0x04UL  // Sn
#define MIX_MASKSRCNOT          0x0DUL  // SDna
#define MIX_NOT                 0x00UL  // Dn
#define MIX_XORSRC              0x05UL  // DSx
#define MIX_NOTMASKSRC          0x08UL  // DSan
#define MIX_MASKSRC             0x0CUL  // DSa
#define MIX_NOTXORSRC           0x06UL  // DSxn
#define MIX_NOP                 0x03UL  // D
#define MIX_MERGENOTSRC         0x09UL  // DSno
#define MIX_COPYSRC             0x07UL  // S
#define MIX_MERGESRCNOT         0x0AUL  // SDno
#define MIX_MERGESRC            0x0BUL  // DSo
#define MIX_WHITE               0x0002  // 1

// ATI Mach32 extended mix codes

#define MIX_MIN                         0x0010
#define MIX_DST_MINUS_SRC               0x0011
#define MIX_SRC_MINUS_DST               0x0012
#define MIX_PLUS                        0x0013
#define MIX_MAX                         0x0014
#define MIX_HALF__DST_MINUS_SRC         0x0015
#define MIX_HALF__SRC_MINUS_DST         0x0016
#define MIX_AVERAGE                     0x0017
#define MIX_DST_MINUS_SRC_SAT           0x0018
#define MIX_SRC_MINUS_DST_SAT           0x001a
#define MIX_HALF_DST_MINUS_SRC_SAT      0x001c
#define MIX_HALF_SRC_MINUS_DST_SAT      0x001e
#define MIX_AVERAGE_SAT                 0x001f

// Pixel Control Register

#define MIXSEL_FRGDMIX  0x0000
#define MIXSEL_PATT     0x0040
#define MIXSEL_EXPPC    0x0080
#define MIXSEL_EXPBLT   0x00c0
#define PLANEMODE       0x0004

// Wait for the specified number of FIFO entries to become available

#define WAITQ_8514(qEntries)                        \
{                                                   \
    uchar stat;                                     \
    do {                                            \
        stat = inpb(GP_STAT);                       \
        } while (stat & (0x100 >> qEntries));       \
}

// Wait for the Graphics Processor to finish

#define CLRCMD_8514()                               \
{                                                   \
    uint stat;                                      \
    do {                                            \
        stat = inpw(GP_STAT) & 0xffff;              \
        } while (stat & GPBUSY);                    \
}

// Macro to pack 8bpp 332 RGB colors

#define DRV_PACK_332(r,g,b)         \
    ((ulong)(b) >> 6) |             \
    (((ulong)(g) >> 3) & 0x1C) |    \
    ((ulong)(r) & 0xE0)

// Macros to pack 15bpp and 16bpp RGB colors

#define DRV_PACK_15(r,g,b)          \
    ((ulong)(b) >> 3) |             \
    (((ulong)(g) << 2) & 0x3E0) |   \
    (((ulong)(r) << 7) & 0x7C00)

#define DRV_PACK_16(r,g,b)          \
    ((ulong)(b) >> 3) |             \
    (((ulong)(g) << 3) & 0x7E0) |   \
    (((ulong)(r) << 8) & 0xF800)

// Macro to pack 24bpp and 32bpp RGB colors

#define DRV_PACK_888(r,g,b)         \
    ((ulong)(b)) |                  \
    ((ulong)(g) << 8) |             \
    ((ulong)(r) << 16)

// Macro to scale back the bresenham error terms as necessary

#define SCALE_BRESENHAM_ERROR(initialError,majorInc,diagInc,precision)  \
    while (ABS(majorInc) > (0x7FFFFFFFL >> (32-precision))              \
            || ABS(diagInc) > (0x7FFFFFFFL >> (32-precision))) {        \
        majorInc >>= 1;                                                 \
        initialError >>= 1;                                             \
        diagInc >>= 1;                                                  \
        }

// Macro to write DWwidth DWORD's to the FIFO'ed HOST blit port. It is
// very important to check for enough FIFO entries to ensure that
// we don't assert the PCI RETRY pin on the graphcis card (which will
// cause audio DMA operations to be screwed up).

#define FIFO_MOVSD(dst,src,DWcount,name)                        \
{                                                               \
    volatile ulong *d = (volatile ulong*)dst;                   \
    ulong *s = (ulong*)src;                                     \
    int count = DWcount;                                        \
    while (count & ~0x03) {                                     \
        WAITQ_##name(4);                                        \
        d[0] = s[0]; d[1] = s[1]; d[2] = s[2]; d[3] = s[3];     \
        d = (volatile ulong*)dst;                               \
        s += 4;                                                 \
        count -= 4;                                             \
        }                                                       \
    if (count) {                                                \
        WAITQ_##name(count);                                    \
        if (count >= 1) d[0] = s[0];                            \
        if (count >= 2) d[1] = s[1];                            \
        if (count >= 3) d[2] = s[2];                            \
        }                                                       \
}

// Same as the above, but handles the last DWORD with a partial read
// from system memory to ensure we don't violate the original bitmap
// alignment and cause a GPF.

#define FIFO_MOVSDB(dst,src,Bwidth,name)                        \
{                                                               \
    volatile ulong *d = (volatile ulong*)dst;                   \
    ulong *s = (ulong*)src;                                     \
    int DWcount = Bwidth/4;                                     \
    while (DWcount & ~0x03) {                                   \
        WAITQ_##name(4);                                        \
        d[0] = s[0]; d[1] = s[1]; d[2] = s[2]; d[3] = s[3];     \
        d = (volatile ulong*)dst;                               \
        s += 4;                                                 \
        DWcount -= 4;                                           \
        }                                                       \
    if (DWcount) {                                              \
        WAITQ_##name(DWcount);                                  \
        if (DWcount >= 1) d[0] = s[0];                          \
        if (DWcount >= 2) d[1] = s[1];                          \
        if (DWcount >= 3) d[2] = s[2];                          \
        s += DWcount;                                           \
        }                                                       \
    if (Bwidth & 3) {                                           \
        WAITQ_##name(1);                                        \
        movsdb(dst,s,Bwidth & 3);                               \
        }                                                       \
}

#endif  /* __SNAP_GRAPHICS_ACCEL_H */

