/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Function prototypes for generic VGA and driver helper
*               functions.
*
****************************************************************************/

#ifndef __SNAP_GRAPHICS_ANALYSE_H
#define __SNAP_GRAPHICS_ANALYSE_H

/*--------------------- Macros and type definitions -----------------------*/

#pragma pack(1)

/****************************************************************************
REMARKS:
Entry in the main mode table stored in the analysed data tables on disk.
This mode table is used to determine what modes have been analysed and to
find the analysed data for the specific mode.

HEADER:
graphics/analyse.h

MEMBERS:
pitch   - Pitch for the display mode
xRes    - X resolution for the mode
yRes    - Y resolution for the mode
bits    - Bits per pixel for the mode (0 = text mode)
flags   - Mode flags for the mode
****************************************************************************/
typedef struct {
    short   pitch;
    short   xRes;
    short   yRes;
    short   bits;
    ulong   flags;
    } ModeEntry;

#define gaVGACore   0x80000000L

/****************************************************************************
REMARKS:
Structure to define the header for the analysed data files on disk.

HEADER:
graphics/analyse.h

MEMBERS:
size        - Size of data tables for each analysed mode
numModes    - Number of display modes analysed
profileSize - Size of the chipset profile table in bytes
commentSize - Size of the chipset profile table comments in bytes
****************************************************************************/
typedef struct {
    short           size;
    short           numModes;
    long            profileSize;
    long            commentSize;
    //uchar         vgaBuffer[size];
    //ModeEntry     modeTable[numModes];
    //uchar         buffer[size * numModes];
    //ulong         profile[profileSize];
    //char          comments[commentSize];
    } FileHeader;

/****************************************************************************
REMARKS:
Structure to define the information for VGA compatible modes. This is used
in the DOS based analyse program to analyse VGA compatible modes such as
text modes etc.

HEADER:
graphics/analyse.h

MEMBERS:
mode    - VGA BIOS mode number
pitch   - Pitch for the VGA display mode
xRes    - X resolution for the mode
yRes    - Y resolution for the mode
bits    - Bits per pixel for the mode (0 = text mode)
****************************************************************************/
typedef struct {
    int     mode;
    int     pitch;
    int     xRes;
    int     yRes;
    int     bits;
    } VGA_mode;

#pragma pack()

/*-------------------------- Function Prototypes --------------------------*/

#ifdef  __cplusplus
extern "C" {                        /* Use "C" linkage when in C++ mode */
#endif

#ifdef  __cplusplus
}                                   /* End of "C" linkage for C++       */
#endif

#endif /* __SNAP_GRAPHICS_ANALYSE_H */

