/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any 32-bit protected mode environment
*
* Description:  Internal header file for the SciTech SNAP Graphics
*               Architecture device driver loader library. This common
*               header file is included by both the low level device driver
*               code as well as the device neutral loader code. Anything
*               that is internal but shared between the device drivers
*               and the loader code goes into this module.
*
****************************************************************************/

#ifndef __SNAP_GRAPHICS_COMMON_H
#define __SNAP_GRAPHICS_COMMON_H

#include "snap/graphics.h"
#include "snap/graphics/crtc.h"

/*---------------------- Macros and type definitions ----------------------*/

/* Define the virtual desktop multi-head display mode numbers. These
 * definitions are here because they need to be common between the
 * device specific modules and the device neutral code.
 */

#define gaVirtualMode8bpp   ((N_uint16)(gaModeMask-4))
#define gaVirtualMode15bpp  ((N_uint16)(gaModeMask-3))
#define gaVirtualMode16bpp  ((N_uint16)(gaModeMask-2))
#define gaVirtualMode24bpp  ((N_uint16)(gaModeMask-1))
#define gaVirtualMode32bpp  ((N_uint16)(gaModeMask-0))

#pragma pack(1)

/****************************************************************************
REMARKS:
Structure defining all the SNAP graphics driver functions as imported
into the Binary Portable DLL.
{secret}
****************************************************************************/
typedef struct {
    long        dwSize;
    N_uint32    (NAPIP PCI_accessReg)(int index,N_uint32 value,int func,PCIDeviceInfo *info);
    ibool       (NAPIP CDBX_getRefreshListV1)(int device,int attributes,int xRes,int yRes,int tripleScanLowRes,int doubleScan512,GA_monitor *monitor,N_uint32 maxClock,N_uint8 maxRefresh,int size,N_int32 *refreshRateList);
    ibool       (NAPIP CDBX_getCRTCTimingsV2)(int device,int attributes,int xRes,int yRes,int tripleScanLowRes,int doubleScan512,int refresh,GA_monitor *monitor,N_uint32 maxClock,N_uint8 maxRefresh,CDBX_record *crtc);
    void        (NAPIP CDBX_setGlobalRefreshV2)(int device,int refresh);
    void        (NAPIP CDBX_setCRTCTimingsV1)(int device,int xRes,int yRes,GA_CRTCInfo *crtc);
    void        *reserved1;
    void        (NAPIP usDelay)(N_uint32 microSeconds);
    void        (NAPIP msDelay)(N_uint32 milliSeconds);
    void *      (NAPIP mapBIOSImage)(int deviceIndex);
    void        (NAPIP unmapBIOSImage)(int deviceIndex);
    ibool       (NAPIP postControllerObsolete)(N_int32 device);
    void        (NAPIP N_log)(const char *fmt,...);
    ibool       (NAPIP CDBX_getRefreshListV2)(int device,int attributes,int xRes,int yRes,int bitsPerPixel,int tripleScanLowRes,int doubleScan512,GA_monitor *monitor,N_uint32 maxClock,N_uint8 maxRefresh,int size,N_int32 *refreshRateList,GA_GetClosestPixelClock_t findClosestClock);
    ibool       (NAPIP PCI_getPCIList)(PCIDeviceInfo **PCIList,int *count);
    ibool       (NAPIP CDBX_setCRTCTimingsV2)(int device,int xRes,int yRes,GA_CRTCInfo *crtc,ibool addNewTiming,ibool makeDefault);
    int         (NAPIP DDC_quickDetect)(GA_SCIFuncs *sciFuncs,N_int32 iChannel);
    int         (NAPIP DDC_getFlatPanelResolution)(GA_SCIFuncs *sciFuncs,N_int32 iChannel,N_int16 *xRes, N_int16 *yRes);
    ibool       (NAPIP CDBX_getRefreshList)(int device,int head,int attributes,int xRes,int yRes,int bitsPerPixel,int tripleScanLowRes,int doubleScan512,GA_monitor *monitor,N_uint32 maxClock,N_uint8 maxRefresh,int size,N_int32 *refreshRateList,GA_GetClosestPixelClock_t findClosestClock);
    ibool       (NAPIP CDBX_getCRTCTimings)(int device,int head,int attributes,int xRes,int yRes,int tripleScanLowRes,int doubleScan512,int refresh,GA_monitor *monitor,N_uint32 maxClock,N_uint8 maxRefresh,CDBX_record *crtc);
    void        (NAPIP CDBX_setGlobalRefresh)(int device,int head,int refresh);
    ibool       (NAPIP CDBX_setCRTCTimings)(int device,int head,int xRes,int yRes,GA_CRTCInfo *crtc,ibool addNewTiming,ibool makeDefault);
    } DRV_imports;

/****************************************************************************
REMARKS:
Subset of DRV_imports for building detect modules.
{secret}
****************************************************************************/
typedef struct {
    long        dwSize;
    ibool       (NAPIP PCI_getPCIList)(PCIDeviceInfo **PCIList,int *count);
    } DETECT_imports;

/****************************************************************************
REMARKS:
Structure defining all the SNAP OpenGL compatile 3D graphics driver
functions as imported into the Binary Portable DLL.
{secret}
****************************************************************************/
typedef struct {
    long        dwSize;
    void        (NAPIP N_log)(const char *fmt,...);
    void        (NAPIP usDelay)(N_uint32 microSeconds);
    void        (NAPIP msDelay)(N_uint32 milliSeconds);
    ibool       (NAPIP PCI_getPCIList)(PCIDeviceInfo **PCIList,int *count);
    PE_MODULE * (NAPIP GA_loadLibrary)(const char *szBPDName,ulong *size,ibool shared);
    void *      (NAPIP GA_getProcAddress)(PE_MODULE *hModule,const char *szProcName);
    void        (NAPIP GA_freeLibrary)(PE_MODULE *hModule);
    } GL3D_imports;

/****************************************************************************
REMARKS:
Function pointer type for the Binary Portable DLL initialisation entry point
for graphics device drivers.
{secret}
****************************************************************************/
typedef GA_devCtx * (NAPIP GA_initDriver_t)(
    ibool shared,
    const char *path,
    int deviceIndex,
    PCIDeviceInfo *info,
    GA_monitor *monitor,
    N_imports *nImp,
    PM_imports *pmImp,
    DRV_imports *drvImp,
    GA_options *licensedOptions,
    GA_globalOptions *globalOpt,
    ibool sharedIRQ,
    GA_exports *exportTable,
    GA_sharedInfo *sharedInfo,
    int loadInGUI);

/****************************************************************************
REMARKS:
Function pointer type for the Binary Portable DLL initialisation entry point
for graphics device detection modules.
{secret}
****************************************************************************/
typedef ibool (NAPIP GA_detectChipset_t)(
    PM_imports *pmImp,
    DETECT_imports *drvImp,
    PCIDeviceInfo *info,
    char *mfrName,
    char *chipName);

/****************************************************************************
REMARKS:
Function pointer type for the Binary Portable DLL initialisation entry point
for OpenGL compatible 3D graphics device drivers.
{secret}
****************************************************************************/
typedef struct GA_glCtx * (NAPIP GA_gl3DInitDriver_t)(
    ibool shared,
    const char *path,
    GA_devCtx *hwCtx,
    PCIDeviceInfo *info,
    PM_imports *pmImp,
    GL3D_imports *gl3DImp,
    GA_options *licensedOptions,
    GA_globalOptions *globalOpt);
    
struct GA_glFuncs;    

/****************************************************************************
REMARKS:
Function pointer type for the Binary Portable DLL initialisation entry point
for the OpenGL compatible utility library.
{secret}
****************************************************************************/
typedef struct GA_gluFuncs * (NAPIP GA_gluInitDriver_t)(
    PM_imports *pmImp,
    struct GA_glFuncs *glFuncs);

/****************************************************************************
REMARKS:
Function pointer type for the Binary Portable DLL exit entry point
for the OpenGL compatible utility library.
{secret}
****************************************************************************/
typedef void (NAPIP GA_gluExitDriver_t)(void);

/****************************************************************************
REMARKS:
Function pointer type for the Binary Portable DLL initialisation entry point
for graphics device drivers.
{secret}
****************************************************************************/
typedef ibool (NAPIP GA_detectDevice_t)(
    char *driverName,
    N_imports *nImp,
    PM_imports *pmImp);

/****************************************************************************
REMARKS:
Function pointer type for the function to get the list of chipset names.
{secret}
****************************************************************************/
typedef void (NAPIP GA_getChipsetNames_t)(
    char **mfrName,
    char ***chipsetNames);

/****************************************************************************
REMARKS:
Function pointer type for the function to get the list of certification
information.
{secret}
****************************************************************************/
typedef void (NAPIP GA_getCertifyInfo_t)(
    GA_certifyInfo *info);

/****************************************************************************
REMARKS:
Function pointer type for the Binary Portable DLL initialisation entry point.
{secret}
****************************************************************************/
typedef GA_devCtx * (NAPIP GA_initFilter_t)(
    GA_devCtx *dc,
    const char *path,
    N_imports *nImp,
    PM_imports *pmImp,
    DRV_imports *drvImp,
    GA_options *licensedOpt,
    GA_globalOptions *globalOpt);

#ifndef __DRIVER__
#define _CEXPORT __cdecl
#else
#define _CEXPORT __cdecl __export
#endif

#pragma pack()

/*--------------------------- Global variables ----------------------------*/

#ifdef DETECT_ONLY
extern DETECT_imports  _VARAPI _DETECT_imports;
#else
extern DRV_imports  _VARAPI _DRV_imports;
#endif
extern N_imports    _VARAPI _N_imports;
extern GA_imports   _VARAPI _GA_imports;

/*-------------------------- Function Prototypes --------------------------*/

#ifdef  __cplusplus
extern "C" {                        /* Use "C" linkage when in C++ mode */
#endif

/* Device driver exported entry points. Ideally these would only be listed in
 * the snap/graphics/driver.h header file, however because we need to call these
 * functions from the generic loader code directly when compiling in TEST_HARNESS
 * mode, they need to be listed here.
 */

GA_devCtx * _CEXPORT GA_initDriver(ibool shared,const char *path,int deviceIndex,PCIDeviceInfo *info,GA_monitor *monitor,N_imports *nImp,PM_imports *pmImp,DRV_imports *drvImp,GA_options *licensedOptions,GA_globalOptions *globalOpt,ibool sharedIRQ,GA_exports *exportTable,GA_sharedInfo *sharedInfo,int loadInGUI);
void        _CEXPORT GA_getChipsetNames(char **mfrName,char ***chipsetNames);
void        _CEXPORT GA_getCertifyInfo(GA_certifyInfo *info);
ibool       _CEXPORT GA_detectDevice(char *driverName,N_imports *nImp,PM_imports *pmImp);

/* Direct function prototype to load the bugfix01 driver when in TEST_HARNESS mode */

GA_devCtx * NAPI BUGFIX01_loadDriver(GA_devCtx *dc,const char *path,N_imports *nImp,PM_imports *pmImp,DRV_imports *drvImp,GA_options *licensedOpt,GA_globalOptions *globalOpt,PE_MODULE **hModule);

/* Initialise the OpenGL compatible driver module */

struct GA_glCtx *  GL3D_initModule(ibool shared,const char *path,GA_devCtx *hwCtx,PCIDeviceInfo *info,GA_options *licensedOptions,GA_globalOptions *globalOpt);

#ifdef  __cplusplus
}                                   /* End of "C" linkage for C++       */
#endif

#endif  /* __SNAP_GRAPHICS_COMMON_H */

