/****************************************************************************
*
*                         SciTech Display Doctor
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Header file for detecting external RAMDAC's and clock chips.
*
****************************************************************************/

#ifndef __SNAP_GRAPHICS_DACID_H
#define __SNAP_GRAPHICS_DACID_H

/*--------------------- Macros and type definitions -----------------------*/

/* RAMDAC types */

typedef enum {
    DAC_INTERNAL,       /* Chipset specific internal 24 bit DAC     */
    DAC_SC6DAC,         /* Sierra SC11481/6/8 15 bit DAC            */
    DAC_SC3DAC,         /* Sierra SC11482/3/4 15 bit DAC            */
    DAC_SC7DAC,         /* Sierra SC11485/7/9 16 bit DAC            */
    DAC_MU16DAC,        /* Music MU9C4870 16 bit DAC                */
    DAC_ATI30DAC,       /* ATI68830 16 bit DAC                      */
    DAC_UMC17DAC,       /* UMC UM70c178 24 bit DAC                  */
    DAC_ATT08DAC,       /* AT&T 20C408/9 TrueColor 24 bit DAC       */
    DAC_ATT90DAC,       /* AT&T 20C490 TrueColor 24 bit DAC         */
    DAC_ATT91DAC,       /* AT&T 20C491 TrueColor 24 bit DAC         */
    DAC_ATT92DAC,       /* AT&T 20C492/3 24 bit DAC                 */
    DAC_ATT98DAC,       /* AT&T 21C498 TrueColor 24 bit DAC         */
    DAC_ATT99DAC,       /* AT&T 20C499 TrueColor 24 bit DAC         */
    DAC_ATT4DAC,        /* AT&T 21C504 24 bit DAC                   */
    DAC_ATT5DAC,        /* AT&T 21C505 24 bit DAC                   */
    DAC_ATT511DAC,      /* AT&T 21C511 24 bit DAC                   */
    DAC_ADAC1DAC,       /* Acumos ADAC1 24 bit DAC                  */
    DAC_SC20DAC,        /* Sierra SC15020 TrueColor 24 bit DAC      */
    DAC_SC21DAC,        /* Sierra SC15021 TrueColor 24 bit DAC      */
    DAC_SC25DAC,        /* Sierra SC15025 TrueColor 24 bit DAC      */
    DAC_SC26DAC,        /* Sierra SC15026 TrueColor 24 bit DAC      */
    DAC_MU18DAC,        /* Music MU9C1880 TrueColor 24 bit DAC      */
    DAC_MU49DAC,        /* Music MU9C4910 TrueColor 24 bit DAC      */
    DAC_MU99DAC,        /* Music MU9C9910 TrueColor 24 bit DAC      */
    DAC_STG0DAC,        /* Inmos STG1700 24 bit DAC                 */
    DAC_STG2DAC,        /* Inmos STG1702 24 bit DAC                 */
    DAC_STG3DAC,        /* Inmos STG1703 24 bit DAC                 */
    DAC_STG32DAC,       /* STG1732 24 bit DAC                       */
    DAC_STG64DAC,       /* STG1764 24 bit DAC                       */
    DAC_ICS00DAC,       /* ICS5300 GENDAC 24 bit DAC                */
    DAC_ICS01DAC,       /* ICS5301 GENDAC 24 bit DAC                */
    DAC_ICS40DAC,       /* ICS5340 GENDAC 24 bit DAC                */
    DAC_ICS41DAC,       /* ICS5341 GENDAC 24 bit DAC                */
    DAC_ICS42DAC,       /* ICS5342 GENDAC 24 bit DAC                */
    DAC_TR8DAC,         /* Trident 8001 24 bit DAC                  */
    DAC_ALG12DAC,       /* Avance Logic ALG1201 24 bit DAC          */
    DAC_ALG13DAC,       /* Avance Logic ALG1301 24 bit DAC          */
    DAC_UMC18DAC,       /* UMC UM70c188 24 bit DAC                  */
    DAC_CH91DAC,        /* Chrontel CH8391 24 bit DAC               */
    DAC_CH98DAC,        /* Chrontel CH8398 24 bit DAC               */
    DAC_CH98ADAC,       /* Chrontel CH8398A 24 bit DAC              */
    DAC_ICW516DAC,      /* IC Works W30C516 24 bit DAC              */
    DAC_ICW524DAC,      /* IC Works W30C524 24 bit DAC              */
    DAC_BT1DAC,         /* BrookTree Bt481 16 bit DAC               */
    DAC_BT2DAC,         /* BrookTree Bt482 16 bit DAC               */
    DAC_BT4DAC,         /* BrookTree Bt484 24 bit DAC               */
    DAC_BT5DAC,         /* BrookTree Bt485 24 bit DAC               */
    DAC_BT9DAC,         /* BrookTree Bt489 24 bit DAC               */
    DAC_TV0DAC,         /* TI TVP3020 24 bit DAC                    */
    DAC_TV5DAC,         /* TI TVP3025 24 bit DAC                    */
    DAC_TV6DAC,         /* TI TVP3026 24 bit DAC                    */
    DAC_TV30DAC,        /* TI TVP3030 24 bit DAC                    */
    DAC_TL5DAC,         /* TLC34075/6/7,ATI68875,Bt476/8 24 bit DAC */
    DAC_ATI60DAC,       /* ATI68860 24 bit DAC                      */
    DAC_IBM24DAC,       /* IBM RGB524 24 bit DAC                    */
    DAC_IBM25DAC,       /* IBM RGB514/525 24 bit DAC                */
    DAC_IBM26DAC,       /* IBM RGB526 24 bit DAC                    */
    DAC_IBM26DBDAC,     /* IBM RGB526DB 24 bit DAC                  */
    DAC_IBM28DAC,       /* IBM RGB528 24 bit DAC                    */
    DAC_IBM624DAC,      /* IBM RGB624 24 bit DAC                    */
    DAC_OTI088DAC,      /* Oak OTI-088 Integrated 24 bit DAC        */
    DAC_MAX_ID,         /* Maximum DAC ID                           */
    } GA_dacType;

/* The S3 SDAC is equivalent to this one */

#define DAC_S3SDAC  DAC_ICS00DAC

#define STANDARD_DAC_NAMES                                      \
    "Sierra SC11481/6/8 15 bit DAC",                            \
    "Sierra SC11482/3/4 15 bit DAC",                            \
    "Sierra SC11485/7/9 16 bit DAC",                            \
    "Music MU9C4870 16 bit DAC",                                \
    "ATI68830 16 bit DAC",                                      \
    "UMC UM70c178 24 bit DAC",                                  \
    "AT&T ATT 20C408/9,IC Works W30C409 24 bit DAC",            \
    "AT&T ATT 20C490 24 bit DAC",                               \
    "AT&T ATT 20C491 24 bit DAC",                               \
    "AT&T ATT 20C492/3,WinBond 82c490 24 bit DAC",              \
    "AT&T ATT 21C498,IC Works W30C498 24 bit DAC",              \
    "AT&T ATT 20C499,IC Works W30C499 24 bit DAC",              \
    "AT&T ATT 21C504 24 bit DAC",                               \
    "AT&T ATT 21C505 24 bit DAC",                               \
    "AT&T ATT 21C511 24 bit DAC",                               \
    "Acumos ADAC1 24 bit DAC",                                  \
    "Sierra SC15020 24 bit DAC",                                \
    "Sierra SC15021 24 bit DAC",                                \
    "Sierra SC15025 24 bit DAC",                                \
    "Sierra SC15026 24 bit DAC",                                \
    "Music MU9C1880 24 bit DAC",                                \
    "Music MU9C4910 24 bit DAC",                                \
    "Music MU9C9910 24 bit DAC",                                \
    "SGS/Inmos STG1700 24 bit DAC",                             \
    "SGS/Inmos STG1702 24 bit DAC",                             \
    "SGS/Inmos STG1703 24 bit DAC",                             \
    "SGS/Inmos STG1732 24 bit DAC",                             \
    "SGS/Inmos STG1764 24 bit DAC",                             \
    "ICS5300 GENDAC,S3 86c716 SDAC 24 bit DAC",                 \
    "ICS5301 GENDAC 24 bit DAC",                                \
    "ICS5340 GENDAC 24 bit DAC",                                \
    "ICS5341 GENDAC 24 bit DAC",                                \
    "ICS5342 GENDAC,S3 86c708 GENDAC 24 bit DAC",               \
    "Trident TDK8001 24 bit DAC",                               \
    "Avance Logic ALG1201 24 bit DAC",                          \
    "Avance Logic ALG1301 24 bit DAC",                          \
    "UMC UM70c188 24 bit DAC",                                  \
    "Chrontel CH8391 24 bit DAC",                               \
    "Chrontel CH8398 24 bit DAC",                               \
    "Chrontel CH8398A 24 bit DAC",                              \
    "IC Works W30C516 24 bit DAC",                              \
    "IC Works W30C524 24 bit DAC",                              \
    "BrookTree Bt481 24 bit DAC",                               \
    "BrookTree Bt482 24 bit DAC",                               \
    "BrookTree Bt484 24 bit DAC",                               \
    "BrookTree Bt485 24 bit DAC",                               \
    "BrookTree Bt489 24 bit DAC",                               \
    "TI TVP3020 24 bit DAC",                                    \
    "TI TVP3025 24 bit DAC",                                    \
    "TI TVP3026 24 bit DAC",                                    \
    "TI TVP3030 24 bit DAC",                                    \
    "TLC34075/6/7,ATI68875,Bt476/8 24 bit DAC",                 \
    "ATI68860 24 bit DAC",                                      \
    "IBM RGB524 24 bit DAC",                                    \
    "IBM RGB514/525 24 bit DAC",                                \
    "IBM RGB526 24 bit DAC",                                    \
    "IBM RGB526DB 24 bit DAC",                                  \
    "IBM RGB528 24 bit DAC",                                    \
    "IBM RGB624 24 bit DAC",                                    \
    "Oak OTI-088 24 bit DAC"

/* Clock chip types */

typedef enum {
    CLK_INTERNAL,       /* Chipset specific internal clock          */
    CLK_HARD,           /* Discrete clock selection                 */
    CLK_ATI18_0,        /* ATI 18818-0 Clock                        */
    CLK_ATI18_1,        /* ATI 18818-1 Clock                        */
    CLK_ICS9161,        /* ICS9191/ICD2061 Clock                    */
    CLK_ATT08,          /* ATT 20C408/09,ICWorks W30C409 Clock      */
    CLK_ATT99,          /* ATT 20C499,IC_orks W30C499 Clock         */
    CLK_MU99,           /* Music MU9C9910 Integrated Clock          */
    CLK_STG3,           /* SGS/Inmos STG1703 Integrated Clock       */
    CLK_STG64,          /* STG1732/64 Integrated clock              */
    CLK_ICS0x,          /* ICS5300/1 GENDAC Integrated Clock        */
    CLK_ICS4x,          /* ICS5340/1/2 GENDAC Integrated Clock      */
    CLK_CH91,           /* Chrontel CH8391/8 Integrated Clock       */
    CLK_CH98A,          /* Chrontel CH8398A Integrated Clock        */
    CLK_ICW524,         /* IC Works W30C524 Integrated Clock        */
    CLK_TVP5,           /* TI TVP3025 Integrated Clock              */
    CLK_TVP6,           /* TI TVP3026 Integrated Clock              */
    CLK_TVP30,          /* TI TVP3030 Integrated Clock              */
    CLK_IBM24,          /* IBM RGB524/528 Integrated Clock          */
    CLK_IBM25,          /* IBM RGB514/525 Integrated Clock          */
    CLK_IBM26,          /* IBM RGB526 Integrated Clock              */
    CLK_ALG13,          /* Avance 1301 Integrated clock             */
    CLK_OTI088,         /* Oak OTI-088 Integrated clock             */
    CLK_ICW43C94A,      /* IC Works W43C94A                         */
    CLK_MAX_ID,         /* Maximum clock ID                         */
    } GA_clkTypes;

#define STANDARD_CLK_NAMES                                      \
    "Discrete clock selection",                                 \
    "ATI 18818-0 Clock",                                        \
    "ATI 18818-1 Clock",                                        \
    "ICS9161/ICD2061/Diamond DCS2824 Clock",                    \
    "ATT 20C408/409, ICWorks W30C409 Integrated Clock",         \
    "ATT 20C499, ICWorks W30C499 Clock",                        \
    "Music MU9C9910 Integrated Clock",                          \
    "SGS/Inmos STG1703 Integrated Clock",                       \
    "SGS/Inmos STG1732/64 Integrated Clock",                    \
    "ICS5300/1 GENDAC Integrated Clock",                        \
    "ICS5340/1/2 GENDAC Integrated Clock",                      \
    "Chrontel CH8391/8 Integrated Clock",                       \
    "Chrontel CH8398A Integrated Clock",                        \
    "IC Works W30C524 Integrated Clock",                        \
    "TI TVP3025 Integrated Clock",                              \
    "TI TVP3026 Integrated Clock",                              \
    "TI TVP3030 Integrated Clock",                              \
    "IBM RGB524/528 Integrated Clock",                          \
    "IBM RGB514/525 Integrated Clock",                          \
    "IBM RGB526 Integrated Clock",                              \
    "Avance Logic ALG1301 Integrated Clock",                    \
    "Oak OTI-088 Integrated Clock",                             \
    "ICWorks W43C94A Clock"

/*-------------------------- Function Prototypes --------------------------*/

#ifdef  __cplusplus
extern "C" {                        /* Use "C" linkage when in C++ mode */
#endif

/* Function to detect an external RAMDAC */

int     DAC_detect(void);

/* Function to detect an external clock chip */

int     CLK_detect(int dacID);

#ifdef  __cplusplus
}                                   /* End of "C" linkage for C++       */
#endif

#endif  /* __SNAP_GRAPHICS_DACID_H */

