/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Header file for the DDStereo glasses types by vendor name.
*
****************************************************************************/

#ifndef __SNAP_GRAPHICS_GLASSES_H
#define __SNAP_GRAPHICS_GLASSES_H

/****************************************************************************
REMARKS:
This enumeration defines the types of known LC shutter glasses by the
Stereo libraries. Note that these values here are not used internally
by SNAP at all, but are the values stored in the GA_options structure
and used by the stereo control panel applet to configure the SNAP
stereo parameters as necesary.

HEADER:
snap/graphics/glasses.h

MEMBERS:
gaGlasses_Custom                        - User defined custom LC glasses
gaGlasses_StereoGraphics_SimulEyes      - StereoGraphics SimulEyes LC glasses
gaGlasses_StereoGraphics_CrystalEyes    - StereoGraphics CrystalEyes LC glasses
gaGlasses_Woobo_CyberBoy                - Woobo CyberBoy LC glasses
gaGlasses_Chinon_CyberShades            - Chinon CyberShades LC glasses
gaGlasses_VRJoy                         - VRJoy LC glasses
gaGlasses_IArt_VRWorks                  - I-Art VR Works LC glasses
gaGlasses_ISee3D_DVU2000I               - ISee3D DVU2000-I LC glasses
gaGlasses_VRex_VRSurfer                 - VRex VRSurface LC glasses
gaGlasses_NuVision_3DSpex               - NuVision 3D-Spex LC glasses
gaGlasses_H3D_Compatible                - H3D compatible LC glasses
gaGlasses_IODisplay_H3D                 - i-O Display Systems H3D LC glasses
gaGlasses_IODisplay_H3DPlus             - i-O Display Systems H3D Plus LC glasses
gaGlasses_ELSA_3DRevelator              - ELSA 3D Revelator LC glasses
gaGlasses_3DTV_PCP                      - 3DTV PCP Parallel Port LC glasses
gaGlasses_3DTV_PCS                      - 3DTV PCS Serial Port LC glasses
gaGlasses_Anaglyph                      - Anaglyph passive glasses
gaGlasses_SciTech_Anaglyph              - SciTech Anaglyph red/blue passive glasses
gaGlasses_StereoGraphics_StereoEyes     - StereoGraphics StereoEyes LC glasses
gaGlasses_StereoGraphics_SyncDoubler    - StereoGraphics CrystalEyes Sync Doubler
gaHMD_First_HMD                         - First head mounted display ID
gaHMD_Custom                            - User defined custom Head Mounted Display
gaHMD_Forte_VFX1                        - Forte VFX1 Head Mounted Display
gaHMD_IIS_VFX3D                         - Interactive Imaging Systems VFX3D Head Mounted Display
gaHMD_IODisplay_iglasses                - i-O Display Systems i-glasses Head Mounted Display
gaHMD_VirtualIO_iglasses                - Virtual-IO i-glasses Head Mounted Display
gaProjector_First_Projector             - First stereo projector ID
gaProjector_Custom                      - User defined custom stereo projector
gaGlasses_None                          - Indicates the device is not configured
****************************************************************************/
typedef enum {
    gaGlasses_Custom,
    gaGlasses_StereoGraphics_SimulEyes,
    gaGlasses_StereoGraphics_CrystalEyes,
    gaGlasses_Woobo_CyberBoy,
    gaGlasses_Chinon_CyberShades,
    gaGlasses_VRJoy,
    gaGlasses_IArt_VRWorks,
    gaGlasses_ISee3D_DVU2000I,
    gaGlasses_VRex_VRSurfer,
    gaGlasses_NuVision_3DSpex,
    gaGlasses_H3D_Compatible,
    gaGlasses_IODisplay_H3D,
    gaGlasses_IODisplay_H3DPlus,
    gaGlasses_ELSA_3DRevelator,
    gaGlasses_3DTV_PCP,
    gaGlasses_3DTV_PCS,
    gaGlasses_Anaglyph,
    gaGlasses_SciTech_Anaglyph,
    gaGlasses_StereoGraphics_StereoEyes,
    gaGlasses_StereoGraphics_SyncDoubler,
    gaHMD_First_HMD             = 0x1000,
    gaHMD_Custom,
    gaHMD_Forte_VFX1,
    gaHMD_IIS_VFX3D,
    gaHMD_IODisplay_iglasses,
    gaHMD_VirtualIO_iglasses,
    gaProjector_First_Projector = 0x2000,
    gaProjector_Lenticular,
    gaProjector_StereoGraphics_Lenticular,
    gaProjector_Custom,
    gaGlasses_None              = 0x1010101,
    } GA_stereoDeviceType;

#endif  /* __GLASSES_H */

