/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any 32-bit protected mode environment
*
* Description:  Header file for the binary log file data structures.
*
****************************************************************************/

#ifndef __SNAP_GRAPHICS_LOGFILE_H
#define __SNAP_GRAPHICS_LOGFILE_H

/*---------------------- Macros and type definitions ----------------------*/

/* Test information structure */

#pragma pack(4)
typedef struct {
    char            Release[80];
    char            DateTested[80];
    char            TestedBy[80];
    char            PartNo[80];
    char            DriverName[80];
    char            CPU[80];
    char            TransientFailures;
    char            res[100];
    } CF_TestingInfo;

/* Card Specific Information structure */

typedef struct {
    char            ManufacturerName[80];
    char            ChipsetName[80];
    char            DACName[80];
    char            ClockName[80];
    char            VersionInfo[80];
    char            BuildDate[80];
    N_uint32        TotalMemory;
    char            gaHaveDisplayStart;
    char            gaHaveLinearBuffer;
    char            gaHaveAccel2D;
    char            gaHaveHWCursor;
    char            gaHave8BitDAC;
    char            gaHaveInterlaced;
    char            gaHaveTripleBuffer;
    char            gaHaveStereo;
    char            gaHaveAccelVideo;
    char            gaHaveAccel3D;
    char            gaHaveHWColorCursor;
    char            gaHaveLCDOutput;
    char            gaHaveTVOutput;
    N_fix32         memoryClock;            // Version 1.01
    N_fix32         defaultMemoryClock;     // Version 1.01
    N_fix32         maxMemoryClock;         // Version 1.01
    char            res[88];
    } CF_CardInfo;

/* PCI bus configuration information */

typedef struct {
    PCIDeviceInfo   PCI[20];                // Version 1.01
    N_int32         NumPCI;                 // Version 1.01
    N_uint32        PCIID;                  // Version 1.04
    } CF_PCIInfo;

/* Performance Information, one per color depth (8,15,16,24,32) */

typedef struct {
    float           RawGAMark;
    float           GUIGAMark;
    float           Game2DGAMark;
    float           Game3DGAMark;
    } CF_PerformanceInfo;

/* 2D specific information
 * All the following can be:
 *
 *  No              = 0
 *  Yes             = 1
 *  NOT_IMPLEMENTED = 2
 *  NEED_SPECS      = 3
 */

typedef struct {
    char            Use8x8MonoPattern;
    char            Use8x8TransMonoPattern;
    char            Use8x8ColorPattern;
    char            Use8x8TransColorPattern;
    char            SetPlaneMask;
    char            DrawScanList;
    char            DrawPattScanList;
    char            DrawColorPattScanList;
    char            DrawEllipseList;
    char            DrawPattEllipseList;
    char            DrawColorPattEllipseList;
    char            DrawFatEllipseList;
    char            DrawPattFatEllipseList;
    char            DrawColorPattFatEllipseList;
    char            DrawRect;
    char            DrawRectLin;
    char            DrawPattRect;
    char            DrawColorPattRect;
    char            DrawTrap;
    char            DrawPattTrap;
    char            DrawColorPattTrap;
    char            DrawLine;
    char            DrawLineInt;
    char            DrawBresenhamLine;
    char            DrawStippleLine;
    char            DrawStippleLineInt;
    char            DrawBresenhamStippleLine;
    char            PutMonoImageMSBSys;
    char            PutMonoImageMSBLin;
    char            PutMonoImageMSBBM;
    char            PutMonoImageLSBSys;
    char            PutMonoImageLSBLin;
    char            PutMonoImageLSBBM;
    char            ClipMonoImageMSBSys;
    char            ClipMonoImageMSBLin;
    char            ClipMonoImageMSBBM;
    char            ClipMonoImageLSBSys;
    char            ClipMonoImageLSBLin;
    char            ClipMonoImageLSBBM;
    char            BitBlt;
    char            BitBltLin;
    char            BitBltSys;
    char            BitBltBM;
    char            BitBltPatt;
    char            BitBltPattLin;
    char            BitBltPattSys;
    char            BitBltPattBM;
    char            BitBltColorPatt;
    char            BitBltColorPattLin;
    char            BitBltColorPattSys;
    char            BitBltColorPattBM;
    char            SrcTransBlt;
    char            SrcTransBltLin;
    char            SrcTransBltSys;
    char            SrcTransBltBM;
    char            DstTransBlt;
    char            DstTransBltLin;
    char            DstTransBltSys;
    char            DstTransBltBM;
    char            StretchBlt;
    char            StretchBltLin;
    char            StretchBltSys;
    char            StretchBltBM;
    char            ConvertBltSys_Obsolete;
    char            ConvertBltBM_Obsolete;
    char            StretchConvertBltSys_Obsolete;
    char            StretchConvertBltBM_Obsolete;
    char            BitBltFxTest;
    char            BitBltFx;
    char            BitBltFxLin;
    char            BitBltFxSys;
    char            BitBltFxBM;
    char            GetBitmapSys;
    char            GetBitmapBM;
    char            res[100];
    } CF_2DInfo;

/* 3D specific information
 * All the following can be:
 *
 *  No              = 0
 *  Yes             = 1
 *  NOT_IMPLEMENTED = 2
 *  NEED_SPECS      = 3
 */

typedef struct {
    char            res[200];
    } CF_3DInfo;

/* Mode Specific Information, one per (x,y,bpp) resolution */

typedef struct {
    N_uint16        XResolution;
    N_uint16        YResolution;
    N_uint16        BitsPerPixel;
    N_uint16        MaxBuffers;
    N_uint32        MaxPixelClock;
    N_uint32        MaxRefreshRate;
    char            gaHaveDisplayStart;
    char            gaHaveLinearBuffer;
    char            gaHaveAccel2D;
    char            gaHaveHWCursor;
    char            gaHaveInterlaced;
    char            gaHaveTripleBuffer;
    char            gaHaveStereo;
    char            gaHaveAccelVideo;
    char            gaHaveAccel3D;
    char            gaHaveHWColorCursor;
    char            gaHaveGammaCorrect;
    char            gaHaveGetCurrentScanLine;
    char            res[100];
    CF_2DInfo       info2d;
    CF_3DInfo       info3d;
    } CF_ModeInfo;

#define CF_SIGNATURE    "GA CONFORM DATA FILE"

/* Legacy version 1.0 binary file structure */

typedef struct {
    char                Signature[80];  // "GA CONFORM DATA FILE"
    ulong               version;        // Version number (ie: 0x101)
    CF_TestingInfo      testInfo;
    CF_CardInfo         cardInfo;
    CF_PerformanceInfo  perfInfo8;
    CF_PerformanceInfo  perfInfo15;
    CF_PerformanceInfo  perfInfo16;
    CF_PerformanceInfo  perfInfo24;
    CF_PerformanceInfo  perfInfo32;
    N_int32             numberOfModes;
    CF_ModeInfo         modeInfo[1];    // NumberOfModes entries
    } CF_Data_V10;

/* Overall binary file structure for conformance log file */

typedef struct {
    char                Signature[80];  // "GA VISUAL DATA FILE"
    ulong               version;        // Version number (ie: 0x101)
    CF_TestingInfo      testInfo;
    CF_CardInfo         cardInfo;
    CF_PCIInfo          pciInfo;
    CF_PerformanceInfo  perfInfo8;
    CF_PerformanceInfo  perfInfo15;
    CF_PerformanceInfo  perfInfo16;
    CF_PerformanceInfo  perfInfo24;
    CF_PerformanceInfo  perfInfo32;
    N_int32             numberOfModes;
    CF_ModeInfo         modeInfo[1];    // NumberOfModes entries
    } CF_Data;

/* Overall binary file structure for visual.dat file */

typedef struct {
    char                Signature[80];  // "GA CONFORM DATA FILE"
    ulong               version;        // Version number (ie: 0x101)
    ibool               quickTest;
    CF_TestingInfo      testInfo;
    CF_CardInfo         cardInfo;
    } VS_Data;

#define CF_CERTIFY_SIGNATURE    "GA CERTIFY DATA FILE"
#define CF_MAX_PCIID            40

/* Information stored for a single chipset certification log */

typedef struct {
    char            ChipsetName[30];
    char            BuildDate[30];
    N_uint16        CertifyVersion;
    char            CertifiedDate[30];
    N_uint8         CertifyFlags;
    N_int32         NumPCI;
    N_uint32        PCIID[1];

    /* Remainder of PCIID information stored in here */
    } CF_CertifyChipInfo;

/* Structure for a single manufacturer chipset certification log */

typedef struct {
    char            DriverName[20];
    char            ManufacturerName[30];
    N_int32         NumChipsets;

    /* NumChipsets list of CF_CertifyChipInfo structures */
    } CF_CertifyDrvInfo;

/* Structure for a single manufacturer chipset certification log */

typedef struct {
    char            Signature[80];
    ulong           version;
    N_int32         NumDrivers;

    /* NumDrivers list of CF_CertifyDrvInfo structures */
    } CF_CertifyInfo;

#pragma pack()

/* Environment variable to determine where the log files end up */

#define CONFORM_LOGS    "CONFORM_LOGS"

/*---------------------------- Global Variables ---------------------------*/

#ifdef  __cplusplus
extern "C" {            /* Use "C" linkage when in C++ mode */
#endif

/* Global variables provided by calling module */
extern int      numErrors,numWarnings;
extern char     name[],part[],driverName[],cpuName[],passFail[];
extern char     *logfile;
extern char     logsPath[];

/* Error logging functions */

void    logIt(const char *fmt,...);
void    fatal(const char *fmt,...);
void    fail(const char *fmt,...);
void    warn(const char *fmt,...);
void    logFailure(void);
int     CopyBinaryFile(char *sourceName,char *destName);
void    CopyFileToServer(char *filename);
void    CopyBitmapsToServer(void);
void    DetermineProcessorDir(void);
ibool   RemoveFiles(const char *path,const char *fileMask);

#ifdef  __cplusplus
}                       /* End of "C" linkage for C++   */
#endif

#endif  /* __SNAP_GRAPHICS_LOGFILE_H */

