/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Header file describing the refresh rate table structures.
*
****************************************************************************/

#ifndef __SNAP_GRAPHICS_REFRESH_H
#define __SNAP_GRAPHICS_REFRESH_H

/*--------------------- Macros and type definitions -----------------------*/

#pragma pack(1)

/****************************************************************************
REMARKS:
Currently configured monitor information used to restrict available modes,
refresh rates and DPMS capabilities. If maxRes is set to -1 (the default)
then the values are unrestricted (no monitor is currently configured).

HEADER:
graphics/driver.h

MEMBERS:
maxRes      - Maximum resolution (horizontal pels)
maxHScan    - Maximum horizontal scan (in KHz)
maxVScan    - Maximum vertical scan (in Hz)
DPMS        - True if DPMS compliant
aspect      - Aspect ratio of monitor (X / Y * 1000). A 4:3 monitor has a
              value of 1333 while a 16:9 monitor has a value of 1778
****************************************************************************/
typedef struct {
    ulong   dwSize;
    short   maxRes;
    char    maxHScan;
    char    maxVScan;
    char    DPMS;
    short   aspect;
    } GA_monitorInfo;

#define gaAspect4_3     1333
#define gaAspect16_9    1778

/****************************************************************************
REMARKS:
Structure defining the entry for a single resolution and the associated
refresh rate as stored in the CRTC configuration file. We explicitly list
the resolution of the mode so that the files will be upwards compatible
and will continue to work as new modes are added to the core drivers
as the product progresses.

HEADER:
graphics/refresh.h

MEMBERS:
****************************************************************************/
typedef struct {
    short       xRes;
    short       yRes;
    GA_CRTCInfo crtc;
    } GA_refreshEntry;

/****************************************************************************
REMARKS:
Structure of the refresh table configuration file. There will be one of
these files for each display connected to the display adapter. When we
support separate LCD and TV out control we will have separate files on
disk for each of these refresh tables. Note that the refresh table has
an arbitrary number of display modes listed in it, and if a mode is not
found in the table the default CRT timings stored in the driver will
be used (always 60Hz Non-Interlaced).

HEADER:
graphics/refresh.h

MEMBERS:
signature       - 20 character signature ('SDDREFRESH')
monitor         - Configuration information for monitor attached to device
numResolutions  - Number of resolutions in refresh table
modes           - Table of refresh rates for each mode
****************************************************************************/
typedef struct {
    char            signature[20];
    GA_monitorInfo  monitor;
    short           numResolutions;
    GA_refreshEntry modes[1];
    /* Remainder of modes live here */
    } GA_refreshHeader;

#define REFRESH_SIGNATURE   "SDDREFRESH"
#define CRT_FILENAME        "crt%d.dat"

#pragma pack()

#endif  /* __SNAP_GRAPHICS_REFRESH_H */

