/****************************************************************************
*
*                   SciTech SNAP Graphics Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  16-bit Windows driver environment.
*
* Description:  Header file for the 16->32 bit thunks for 16-bit display
*               drivers calling the 32-bit flat model SNAP drivers.
*
****************************************************************************/

#ifndef __SNAP_GRAPHICS_THUNKS_H
#define __SNAP_GRAPHICS_THUNKS_H

#include "snap/graphics.h"

/*---------------------------- Global Variables ---------------------------*/

#ifdef  __WINDOWS16__
#pragma pack(1)
/****************************************************************************
REMARKS:
Thunks structure used to communicate with the ref2d rasteriser.
****************************************************************************/
typedef struct {
    N_uint32            dwSize;
    ibool               (NAPIP QueryFunctions)(N_uint32 id,void *funcs);
    N_int32             (NAPIP SetDrawBuffer)(GA_buffer *drawBuf,void *framebuffer,N_int32 bitsPerPixel,GA_pixelFormat *pf,GA_devCtx *hwCtx,N_int32 softwareOnly);
    void *              (NAPIP RotateBitmap)(void *src,N_int32 bitsPerPixel,N_int32 *stride,N_int32 width,N_int32 height);
    } REF2D_thunks;
#pragma pack()

extern REF2D_thunks     _GA_ref2d;
extern GA_2DStateFuncs  _GA_state2d;
extern GA_2DRenderFuncs _GA_draw2d;
extern GA_cursorFuncs   _GA_cursor;
extern GA_videoFuncs    _GA_video;
#endif

/*-------------------------- Function Prototypes --------------------------*/

#ifdef  __cplusplus
extern "C" {            /* Use "C" linkage when in C++ mode */
#endif

/* Initialisation and helper functions */

#ifdef  __WINDOWS16__
void            SDD_setupThunks(ushort code_sel32,ushort data_sel32,GA_devCtx FAR *driver32);
ibool           SDD_getBankFunc32(int FAR *codeLen,void FAR * FAR *bankFunc);
void _FAR_ *    MapFlat(void _FAR_ *p);

/* 2D reference rasteriser thunks */

ibool           GA_QueryFunctionsRef2d(N_int32 id,void _FAR_ *funcs);
void            GA_SetDrawBufferRef2d(GA_buffer _FAR_ *drawBuf,void _FAR_ *framebuffer,N_int32 bitsPerPixel,GA_pixelFormat _FAR_ *pf,GA_devCtx _FAR_ *hwCtx,N_int32 softwareOnly);
void _FAR_ *    GA_RotateBitmapRef2d(void _FAR_ *src,N_int32 bitsPerPixel,N_int32 _FAR_ *stride,N_int32 width,N_int32 height);

/* Assembler stubs that are not called directly */

ulong           GA_RotateBitmapRef2d_Asm(void _FAR_ *src,N_int32 bitsPerPixel,N_int32 _FAR_ *stride,N_int32 width,N_int32 height);
#endif

/* SNAP loader library functions (calls down to VxD) */

int             GA_enumerateDevices(N_int32 shared);
GA_devCtx _FAR_ * GA_loadDriver(N_int32 deviceIndex,N_int32 shared);
ibool           GA_setActiveDevice(N_int32 deviceIndex);
void            GA_unloadDriver(GA_devCtx _FAR_ *dc);
ibool           GA_QueryFunctions(GA_devCtx _FAR_ *dc,N_uint32 id,N_int32 safetyLevel,void _FAR_ *funcs);
ibool           GA_loadRef2d(GA_devCtx _FAR_ *dc,N_int32 shared,GA_modeInfo _FAR_ *modeInfo,N_int32 transferStart,REF2D_driver _FAR_ *_FAR_ *drv);
void            GA_unloadRef2d(GA_devCtx _FAR_ *dc);

/* GA_initFuncs (calls down into VxD) */

void            GA_GetConfigInfo(GA_devCtx _FAR_ *dc_flat,GA_configInfo _FAR_ *info);
void            GA_SetModeProfile(GA_devCtx FAR *dc_flat,GA_modeProfile _FAR_ *profile);
void            GA_GetOptions(GA_devCtx FAR *dc_flat,GA_options _FAR_ *options);
void            GA_SetOptions(GA_devCtx FAR *dc_flat,GA_options _FAR_ *options);
N_int32         GA_GetVideoModeInfo(GA_devCtx FAR *dc_flat,N_uint32 mode,GA_modeInfo _FAR_ *modeInfo);
N_int32         GA_SetVideoMode(GA_devCtx FAR *dc_flat,N_uint32 mode,N_int32 _FAR_ *virtualX,N_int32 _FAR_ *virtualY,N_int32 _FAR_ *bytesPerLine,N_int32 _FAR_ *maxMem,N_int32 refreshRate,GA_CRTCInfo _FAR_ *crtc);
N_uint32        GA_GetVideoMode(GA_devCtx FAR *dc_flat);
N_int32         GA_GetCustomVideoModeInfo(GA_devCtx FAR *dc_flat,N_int32 xRes,N_int32 yRes,N_int32 virtualX,N_int32 virtualY,N_int32 bitsPerPixel,GA_modeInfo _FAR_ *modeInfo);
N_int32         GA_SetCustomVideoMode(GA_devCtx FAR *dc_flat,N_int32 xRes,N_int32 yRes,N_int32 bitsPerPixel,N_uint32 flags,N_int32 _FAR_ *virtualX,N_int32 _FAR_ *virtualY,N_int32 _FAR_ *bytesPerLine,N_int32 _FAR_ *maxMem,GA_CRTCInfo _FAR_ *crtc);
N_uint32        GA_GetClosestPixelClock(GA_devCtx FAR *dc_flat,N_int32 xRes,N_int32 yRes,N_int32 bitsPerPixel,N_uint32 pixelClock);
void            GA_GetCRTCTimings(GA_devCtx FAR *dc_flat,GA_CRTCInfo _FAR_ *crtc);
void            GA_SetCRTCTimings(GA_devCtx FAR *dc_flat,GA_CRTCInfo _FAR_ *crtc);
void            GA_SetGlobalRefresh(GA_devCtx FAR *dc_flat,N_int32 refresh);
N_int32         GA_SetDisplayOutput(GA_devCtx FAR *dc_flat,N_int32 device);
N_int32         GA_GetDisplayOutput(GA_devCtx FAR *dc_flat);

/* GA_driverFuncs (calls down into VxD) _FAR_ */

void            GA_SetDisplayStart(GA_devCtx FAR *dc_flat,N_int32 offset,N_int32 waitVRT);
void            GA_SetDisplayStartXY(GA_devCtx FAR *dc_flat,N_int32 x,N_int32 y,N_int32 waitVRT);
void            GA_SetStereoDisplayStart(GA_devCtx FAR *dc_flat,N_int32 leftOffset,N_int32 rightOffset,N_int32 waitVRT);
N_int32         GA_GetDisplayStartStatus(GA_devCtx FAR *dc_flat);
void            GA_EnableStereoMode(GA_devCtx FAR *dc_flat,N_int32 enable);
N_int32         GA_IsVSync(GA_devCtx FAR *dc_flat);
void            GA_WaitVSync(GA_devCtx FAR *dc_flat);
void            GA_SetVSyncWidth(GA_devCtx FAR *dc_flat,N_int32 width);
N_int32         GA_GetVSyncWidth(GA_devCtx FAR *dc_flat);
void            GA_SetPaletteData(GA_devCtx FAR *dc_flat,GA_palette _FAR_ *pal,N_int32 num,N_int32 index,N_int32 waitVRT);
void            GA_GetPaletteData(GA_devCtx FAR *dc_flat,GA_palette _FAR_ *pal,N_int32 num,N_int32 index);
void            GA_SetGammaCorrectData(GA_devCtx FAR *dc_flat,GA_palette _FAR_ *pal,N_int32 num,N_int32 index,N_int32 waitVRT);
void            GA_GetGammaCorrectData(GA_devCtx FAR *dc_flat,GA_palette _FAR_ *pal,N_int32 num,N_int32 index);
N_int32         GA_GetCurrentScanLine(GA_devCtx FAR *dc_flat);
void            GA_SetPaletteDataExt(GA_devCtx FAR *dc_flat,GA_paletteExt _FAR_ *pal,N_int32 num,N_int32 index,N_int32 waitVRT);
void            GA_GetPaletteDataExt(GA_devCtx FAR *dc_flat,GA_paletteExt _FAR_ *pal,N_int32 num,N_int32 index);
void            GA_SetGammaCorrectDataExt(GA_devCtx FAR *dc_flat,GA_paletteExt _FAR_ *pal,N_int32 num,N_int32 index,N_int32 waitVRT);
void            GA_GetGammaCorrectDataExt(GA_devCtx FAR *dc_flat,GA_paletteExt _FAR_ *pal,N_int32 num,N_int32 index);

/* GA_cursorFuncs */

void            GA_SetMonoCursor(GA_monoCursor _FAR_ *cursor);
void            GA_SetMonoCursorColor(GA_palette _FAR_ *foreground,GA_palette _FAR_ *background);
void            GA_SetColorCursor(GA_colorCursor _FAR_ *cursor);
void            GA_SetColorCursor256(GA_colorCursor256 _FAR_ *cursor);
void            GA_SetColorCursorRGBA(GA_colorCursorRGBA _FAR_ *cursor);
N_int32         GA_SetCursorPos(N_int32 x,N_int32 y);
void            GA_ShowCursor(N_int32 visible);
void            GA_BeginAccess(N_int32 left,N_int32 top,N_int32 right,N_int32 bottom);
void            GA_EndAccess(void);
N_int32         GA_IsHardwareCursor(void);

/* GA_2DStateFuncs */

N_int32         GA_SetDrawBuffer(GA_buffer _FAR_ *drawBuf);
N_int32         GA_IsIdle(void);
void            GA_WaitTillIdle(void);
void            GA_EnableDirectAccess(void);
void            GA_DisableDirectAccess(void);
void            GA_SetMix(N_int32 mix);
void            GA_SetForeColor(GA_color color);
void            GA_SetBackColor(GA_color color);
void            GA_Set8x8MonoPattern(N_int32 index,GA_pattern _FAR_ *pattern);
void            GA_Use8x8MonoPattern(N_int32 index);
void            GA_Use8x8TransMonoPattern(N_int32 index);
void            GA_Set8x8ColorPattern(N_int32 index,GA_colorPattern _FAR_ *pattern);
void            GA_Use8x8ColorPattern(N_int32 index);
void            GA_Use8x8TransColorPattern(N_int32 index,GA_color transparent);
void            GA_SetLineStipple(GA_stipple stipple);
void            GA_SetLineStippleCount(N_uint32 count);
void            GA_SetPlaneMask(N_uint32 mask);
void            GA_SetSrcBlendFunc(N_int32 blendFunc);
void            GA_SetDstBlendFunc(N_int32 blendFunc);
void            GA_SetAlphaValue(N_uint8 alpha);
void            GA_SetLineStyle(N_uint32 styleMask,N_uint32 styleStep,N_uint32 styleValue);

/* GA_2DRenderFuncs */

GA_color        GA_GetPixel(N_int32 x,N_int32 y);
void            GA_PutPixel(N_int32 x,N_int32 y);
void            GA_DrawScanList(N_int32 y,N_int32 length,N_int16 _FAR_ *scans);
void            GA_DrawPattScanList(N_int32 y,N_int32 length,N_int16 _FAR_ *scans);
void            GA_DrawColorPattScanList(N_int32 y,N_int32 length,N_int16 _FAR_ *scans);
void            GA_DrawEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 _FAR_ *scans);
void            GA_DrawPattEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 _FAR_ *scans);
void            GA_DrawColorPattEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 _FAR_ *scans);
void            GA_DrawFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 _FAR_ *scans);
void            GA_DrawPattFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 _FAR_ *scans);
void            GA_DrawColorPattFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 _FAR_ *scans);
void            GA_DrawRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void            GA_DrawPattRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void            GA_DrawColorPattRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void            GA_DrawTrap(GA_trap _FAR_ *trap);
void            GA_DrawPattTrap(GA_trap _FAR_ *trap);
void            GA_DrawColorPattTrap(GA_trap _FAR_ *trap);
void            GA_DrawLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast);
void            GA_DrawBresenhamLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags);
void            GA_DrawStippleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent);
void            GA_DrawBresenhamStippleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent);
void            GA_DrawEllipse(N_int32 left,N_int32 top,N_int32 A,N_int32 B);
void            GA_ClipEllipse(N_int32 left,N_int32 top,N_int32 A,N_int32 B,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
void            GA_PutMonoImageMSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 _FAR_ *image,N_int32 transparent);
void            GA_PutMonoImageMSBLin(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_int32 imageOfs,N_int32 transparent);
void            GA_PutMonoImageMSBBM(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 _FAR_ *image,N_int32 imagePhysAddr,N_int32 transparent);
void            GA_PutMonoImageLSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 _FAR_ *image,N_int32 transparent);
void            GA_PutMonoImageLSBLin(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_int32 imageOfs,N_int32 transparent);
void            GA_PutMonoImageLSBBM(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 _FAR_ *image,N_int32 imagePhysAddr,N_int32 transparent);
void            GA_ClipMonoImageMSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 _FAR_ *image,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
void            GA_ClipMonoImageMSBLin(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_int32 imageOfs,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
void            GA_ClipMonoImageMSBBM(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 _FAR_ *image,N_int32 imagePhysAddr,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
void            GA_ClipMonoImageLSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 _FAR_ *image,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
void            GA_ClipMonoImageLSBLin(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_int32 imageOfs,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
void            GA_ClipMonoImageLSBBM(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 _FAR_ *image,N_int32 imagePhysAddr,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
void            GA_BitBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void            GA_BitBltLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void            GA_BitBltSys(void _FAR_ *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,N_int32 flipY);
void            GA_BitBltBM(void _FAR_ *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void            GA_BitBltPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void            GA_BitBltPattLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void            GA_BitBltPattSys(void _FAR_ *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void            GA_BitBltPattBM(void _FAR_ *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void            GA_BitBltColorPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void            GA_BitBltColorPattLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void            GA_BitBltColorPattSys(void _FAR_ *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void            GA_BitBltColorPattBM(void _FAR_ *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void            GA_SrcTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void            GA_SrcTransBltLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void            GA_SrcTransBltSys(void _FAR_ *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent,N_int32 flipY);
void            GA_SrcTransBltBM(void _FAR_ *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void            GA_DstTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void            GA_DstTransBltLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void            GA_DstTransBltSys(void _FAR_ *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent,N_int32 flipY);
void            GA_DstTransBltBM(void _FAR_ *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void            GA_StretchBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix);
void            GA_StretchBltLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix);
void            GA_StretchBltSys(void _FAR_ *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix,N_int32 flipY);
void            GA_StretchBltBM(void _FAR_ *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix);
N_int32         GA_ConvertBltSys(void _FAR_ *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 bitsPerPixel,GA_pixelFormat _FAR_ *pixelFormat,GA_palette _FAR_ *dstPal,GA_palette _FAR_ *srcPal,N_int32 dither,N_int32 mix,N_int32 flipY);
N_int32         GA_ConvertBltBM(void _FAR_ *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 bitsPerPixel,GA_pixelFormat _FAR_ *pixelFormat,GA_palette _FAR_ *dstPal,GA_palette _FAR_ *srcPal,N_int32 dither,N_int32 mix);
N_int32         GA_StretchConvertBltSys(void _FAR_ *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 bitsPerPixel,GA_pixelFormat _FAR_ *pixelFormat,GA_palette _FAR_ *dstPal,GA_palette _FAR_ *srcPal,N_int32 dither,N_int32 mix,N_int32 flipY);
N_int32         GA_StretchConvertBltBM(void _FAR_ *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 bitsPerPixel,GA_pixelFormat _FAR_ *pixelFormat,GA_palette _FAR_ *dstPal,GA_palette _FAR_ *srcPal,N_int32 dither,N_int32 mix);
N_int32         GA_BitBltFxTest(GA_bltFx _FAR_ *fx);
N_int32         GA_BitBltFx(N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,GA_bltFx _FAR_ *fx);
N_int32         GA_BitBltFxLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,GA_bltFx _FAR_ *fx);
N_int32         GA_BitBltFxSys(void _FAR_ *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,GA_bltFx _FAR_ *fx);
N_int32         GA_BitBltFxBM(void _FAR_ *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,GA_bltFx _FAR_ *fx);
void            GA_GetBitmapSys(void _FAR_ *dstAddr,N_int32 dstPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void            GA_GetBitmapBM(void _FAR_ *dstAddr,N_int32 dstPhysAddr,N_int32 dstPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void            GA_UpdateScreen(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
N_int32         GA_DrawClippedLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32         GA_DrawClippedBresenhamLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32         GA_DrawClippedStippleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32         GA_DrawClippedBresenhamStippleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
void            GA_BitBltPlaneMasked(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_uint32 planeMask);
void            GA_BitBltPlaneMaskedLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_uint32 planeMask);
void            GA_BitBltPlaneMaskedSys(void _FAR_ *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_uint32 planeMask,N_int32 flipY);
void            GA_BitBltPlaneMaskedBM(void _FAR_ *srcAddr,N_int32 srcPhysAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_uint32 planeMask);
void            GA_DrawRectLin(N_int32 dstOfs,N_int32 dstPitch,N_int32 left,N_int32 top,N_int32 width,N_int32 height,GA_color color,N_int32 mix);
void            GA_DrawRectExt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,GA_color color,N_int32 mix);
void            GA_DrawStyleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent);
void            GA_DrawBresenhamStyleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent);
N_int32         GA_DrawClippedStyleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32         GA_DrawClippedBresenhamStyleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);

#ifdef  __cplusplus
}                                   /* End of "C" linkage for C++       */
#endif

#endif  /* __SNAP_GRAPHICS_THUNKS_H */

