;****************************************************************************
;*
;*                     SciTech SNAP Graphics Architecture
;*
;*  ========================================================================
;*
;*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
;*
;*   This file may be distributed and/or modified under the terms of the
;*   GNU General Public License version 2 as published by the Free
;*   Software Foundation and appearing in the file LICENSE.GPL included
;*   in the packaging of this file.
;*
;*   Licensees holding a valid Commercial License for this product from
;*   SciTech Software, Inc. may use this file in accordance with the
;*   Commercial License Agreement provided with the Software.
;*
;*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
;*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
;*   PURPOSE.
;*
;*   See http://www.scitechsoft.com/license/ for information about
;*   the licensing options available and how to purchase a Commercial
;*   License Agreement.
;*
;*   Contact license@scitechsoft.com if any conditions of this licensing
;*   are not clear to you, or you have questions about licensing options.
;*
;*  ========================================================================
;*
;* Language:    ANSI C
;* Environment: 16-bit Windows driver environment.
;*
;* Description: Header file for the 16->32 bit thunks for 16-bit display
;*              drivers calling the 32-bit flat model SNAP Graphics drivers.
;*
;****************************************************************************

;----------------------------------------------------------------------------
; Fundamental types
;----------------------------------------------------------------------------

typedef N_int8      byte
typedef N_int16     word
typedef N_int32     dword
typedef N_uint8     byte
typedef N_uint16    word
typedef N_uint32    dword
typedef N_fix32     dword
typedef N_flt       dword

;----------------------------------------------------------------------------
; SNAP Graphics Architecture ASM functions structure
;----------------------------------------------------------------------------

;----------------------------------------------------------------------------
; REF2D_driver functions structure
;----------------------------------------------------------------------------

struc   REF2D_thunks_s
dwSize                          ulong       ?
QueryFunctionsRef2d             cptr        ?
SetDrawBufferRef2d              cptr        ?
RotateBitmapRef2d_Asm           cptr        ?
ends    REF2D_thunks_s

;----------------------------------------------------------------------------
; Device Driver Hardware 2D State Management functions structure
;----------------------------------------------------------------------------

struc   GA_2DStateFuncs_s
dwSize                          ulong       ?
SetDrawBuffer                   cptr        ?
IsIdle                          cptr        ?
WaitTillIdle                    cptr        ?
EnableDirectAccess              cptr        ?
DisableDirectAccess             cptr        ?
SetMix                          cptr        ?
SetForeColor                    cptr        ?
SetBackColor                    cptr        ?
Set8x8MonoPattern               cptr        ?
Use8x8MonoPattern               cptr        ?
Use8x8TransMonoPattern          cptr        ?
Set8x8ColorPattern              cptr        ?
Use8x8ColorPattern              cptr        ?
Use8x8TransColorPattern         cptr        ?
SetLineStipple                  cptr        ?
SetLineStippleCount             cptr        ?
SetPlaneMask                    cptr        ?
SetSrcBlendFunc                 cptr        ?
SetDstBlendFunc                 cptr        ?
SetAlphaValue                   cptr        ?
SetupForWindows                 cptr        ?
ResetForWindows                 cptr        ?
SetLineStyle                    cptr        ?
ends    GA_2DStateFuncs_s

;----------------------------------------------------------------------------
; Device Driver Hardware 2D Rendering functions structure
;----------------------------------------------------------------------------

struc   GA_2DRenderFuncs_s
dwSize                          ulong       ?
GetPixel                        cptr        ?
PutPixel                        cptr        ?
DrawScanList                    cptr        ?
DrawPattScanList                cptr        ?
DrawColorPattScanList           cptr        ?
DrawEllipseList                 cptr        ?
DrawPattEllipseList             cptr        ?
DrawColorPattEllipseList        cptr        ?
DrawFatEllipseList              cptr        ?
DrawPattFatEllipseList          cptr        ?
DrawColorPattFatEllipseList     cptr        ?
DrawRect                        cptr        ?
res1                            cptr        ?
DrawPattRect                    cptr        ?
DrawColorPattRect               cptr        ?
DrawTrap                        cptr        ?
DrawPattTrap                    cptr        ?
DrawColorPattTrap               cptr        ?
DrawLineInt                     cptr        ?
DrawBresenhamLine               cptr        ?
DrawStippleLineInt              cptr        ?
DrawBresenhamStippleLine        cptr        ?
DrawEllipse                     cptr        ?
ClipEllipse                     cptr        ?
PutMonoImageMSBSys              cptr        ?
PutMonoImageMSBLin              cptr        ?
PutMonoImageMSBBM               cptr        ?
PutMonoImageLSBSys              cptr        ?
PutMonoImageLSBLin              cptr        ?
PutMonoImageLSBBM               cptr        ?
ClipMonoImageMSBSys             cptr        ?
ClipMonoImageMSBLin             cptr        ?
ClipMonoImageMSBBM              cptr        ?
ClipMonoImageLSBSys             cptr        ?
ClipMonoImageLSBLin             cptr        ?
ClipMonoImageLSBBM              cptr        ?
BitBlt                          cptr        ?
BitBltLin                       cptr        ?
BitBltSys                       cptr        ?
BitBltBM                        cptr        ?
BitBltPatt                      cptr        ?
BitBltPattLin                   cptr        ?
BitBltPattSys                   cptr        ?
BitBltPattBM                    cptr        ?
BitBltColorPatt                 cptr        ?
BitBltColorPattLin              cptr        ?
BitBltColorPattSys              cptr        ?
BitBltColorPattBM               cptr        ?
SrcTransBlt                     cptr        ?
SrcTransBltLin                  cptr        ?
SrcTransBltSys                  cptr        ?
SrcTransBltBM                   cptr        ?
DstTransBlt                     cptr        ?
DstTransBltLin                  cptr        ?
DstTransBltSys                  cptr        ?
DstTransBltBM                   cptr        ?
StretchBlt                      cptr        ?
StretchBltLin                   cptr        ?
StretchBltSys                   cptr        ?
StretchBltBM                    cptr        ?
ConvertBltSys                   cptr        ?
ConvertBltBM                    cptr        ?
StretchConvertBltSys            cptr        ?
StretchConvertBltBM             cptr        ?
BitBltFxTest                    cptr        ?
BitBltFx                        cptr        ?
BitBltFxLin                     cptr        ?
BitBltFxSys                     cptr        ?
BitBltFxBM                      cptr        ?
GetBitmapSys                    cptr        ?
GetBitmapBM                     cptr        ?
UpdateScreen                    cptr        ?
DrawClippedLineInt              cptr        ?
DrawClippedBresenhamLine        cptr        ?
DrawClippedStippleLineInt       cptr        ?
DrawClippedBresenhamStippleLine cptr        ?
BitBltPlaneMasked               cptr        ?
BitBltPlaneMaskedLin            cptr        ?
BitBltPlaneMaskedSys            cptr        ?
BitBltPlaneMaskedBM             cptr        ?
DrawRectLin                     cptr        ?
DrawRectExt                     cptr        ?
DrawStyleLineInt                cptr        ?
DrawBresenhamStyleLine          cptr        ?
DrawClippedStyleLineInt         cptr        ?
DrawClippedBresenhamStyleLine   cptr        ?
ends    GA_2DRenderFuncs_s

;----------------------------------------------------------------------------
; Init functions structure
;----------------------------------------------------------------------------

struc   GA_initFuncs_s
dwSize                  ulong       ?
GetConfigInfo           cptr        ?
res1                    cptr        ?
SetModeProfile          cptr        ?
GetOptions              cptr        ?
SetOptions              cptr        ?
GetVideoModeInfo        cptr        ?
SetVideoMode            cptr        ?
GetVideoMode            cptr        ?
GetCustomVideoModeInfo  cptr        ?
SetCustomVideoMode      cptr        ?
GetClosestPixelClock    cptr        ?
GetCRTCTimings          cptr        ?
SaveCRTCTimings         cptr        ?
SetGlobalRefresh        cptr        ?
SaveRestoreState        cptr        ?
SetDisplayOutput        cptr        ?
GetDisplayOutput        cptr        ?
SetSoftwareRenderFuncs  cptr        ?
GetUniqueFilename       cptr        ?
GetCRTCTable            cptr        ?
GetMonitorInfo          cptr        ?
SetMonitorInfo          cptr        ?
GetCurrentVideoModeInfo cptr        ?
GetCertifyInfo          cptr        ?
SetCRTCTimings          cptr        ?
AlignLinearBuffer       cptr        ?
ends    GA_initFuncs_s

;----------------------------------------------------------------------------
; Driver functions structure
;----------------------------------------------------------------------------

struc   GA_driverFuncs_s
dwSize                  ulong       ?
SetBank                 cptr        ?
SetDisplayStart         cptr        ?
SetDisplayStartXY       cptr        ?
SetStereoDisplayStart   cptr        ?
GetDisplayStartStatus   cptr        ?
EnableStereoMode        cptr        ?
IsVSync                 cptr        ?
WaitVSync               cptr        ?
SetVSyncWidth           cptr        ?
GetVSyncWidth           cptr        ?
SetPaletteData          cptr        ?
GetPaletteData          cptr        ?
SetGammaCorrectData     cptr        ?
GetGammaCorrectData     cptr        ?
GetCurrentScanLine      cptr        ?
SetPaletteDataExt       cptr        ?
GetPaletteDataExt       cptr        ?
SetGammaCorrectDataExt  cptr        ?
GetGammaCorrectDataExt  cptr        ?
ends    GA_driverFuncs_s

;----------------------------------------------------------------------------
; Cursor functions structure
;----------------------------------------------------------------------------

struc   GA_cursorFuncs_s
dwSize                  ulong       ?
SetMonoCursor           cptr        ?
SetMonoCursorColor      cptr        ?
SetColorCursor          cptr        ?
SetCursorPos            cptr        ?
ShowCursor              cptr        ?
BeginAccess             cptr        ?
EndAccess               cptr        ?
SetColorCursor256       cptr        ?
SetColorCursorRGBA      cptr        ?
IsHardwareCursor        cptr        ?
ends    GA_cursorFuncs_s

