/****************************************************************************
*
*                   SciTech SNAP Graphics Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any 32-bit protected mode environment
*
* Description:  Header file for the SciTech SNAP Graphics Architecture,
*               VBE/Core emulation driver. This driver can be used to
*               emulate the VBE/Core BIOS serivces from a 32-bit protected
*               mode device driver on any OS. All you need is the mechanism
*               too hook the real mode Int 10h interrupt from 32-bit
*               protected mode.
*
****************************************************************************/

#ifndef __SNAP_GRAPHICS_VBECORE_H
#define __SNAP_GRAPHICS_VBECORE_H

#include "snap/graphics.h"

/*---------------------- Macros and type definitions ----------------------*/

#pragma pack(1)

/* Register structure used to pass client register information to and
 * from the VBE/Core emulation driver. We only define the subset of
 * registers that are actually needed by the VBE/Core functions to avoid
 * unnecessary data copying. The first part of the strucure is
 * identical to the PMREGS strucure defined in the PMAPI.H header file.
 */

struct _VEMDWORDREGS {
    ulong   eax,ebx,ecx,edx,esi,edi,ebp,cflag;
    };

struct _VEMWORDREGS {
    ushort  ax,ax_hi;
    ushort  bx,bx_hi;
    ushort  cx,cx_hi;
    ushort  dx,dx_hi;
    ushort  si,si_hi;
    ushort  di,di_hi;
    ushort  bp,bp_hi;
    ushort  cflag,cflag_hi;
    };

struct _VEMBYTEREGS {
    uchar   al, ah; ushort ax_hi;
    uchar   bl, bh; ushort bx_hi;
    uchar   cl, ch; ushort cx_hi;
    uchar   dl, dh; ushort dx_hi;
    };

typedef union {
    struct  _VEMDWORDREGS e;
    struct  _VEMWORDREGS  x;
    struct  _VEMBYTEREGS  h;
    } VEM_REGS_t;

typedef struct {
    VEM_REGS_t  r;
    ushort      es;
    ushort      ds;
    } VEM_REGS;

/* Structure containing flags to enable/disable features in the driver */

typedef struct {
    ibool   disableLinear;
    ibool   disablePal;
    ibool   disablePM32;
    ibool   disableVBE20;
    ibool   disableVBE30;
    ibool   disableVBEPM;
    ibool   disableVBESCI;
    ibool   disableVBEDDC;
    } VEM_DisableFlags;

/****************************************************************************
REMARKS:
Data structure to maintain per-VM information for all virtual machines in
the system.

MEMBERS:
cntBank             - Current bank for virtual machine
cntBits             - Current bits per pixel for virtual machine
sharedInfo          - Pointer to the SNAP shared info structure; shared by all!
isTextMode          - True if the current VM is in text mode
xRes                - X resolution for the current mode
yRes                - Y resolution for the current mode
bitsPerPixel        - Color depth for the current mode
bytesPerLine        - Scanline pitch for the current mode
cntMode             - Current mode for VBE BIOS emulator
cntMaxMem           - Current maxMem for VBE BIOS emulator
cntLinearBufAddr    - Current linear buffer address for VBE BIOS emulator
cntDACWidth         - Current RAMDAC width for VBE BIOS emulator
cntStartX           - Current display start X for VBE BIOS emulator
cntStartY           - Current display start Y for VBE BIOS emulator
cntBytesPerLine     - Current bytes per line for for VBE BIOS emulator
PDB                 - Current page directory base for VBE BIOS emulator
PageTable           - Current page table for VBE BIOS emulator
isBankedEmulation   - True if banked framebuffer emulation enabled
numPages            - Number of video pages for the current VGA mode
fbOffset            - Current framebuffer offset
frameBufferBase     - Video framebuffer base address (regardless of paging)
frameBuffer         - Video framebuffer base address adjusted for paging
frameBuffer2        - Video framebuffer base address for CGA modes only (odd lines)
scanBuf             - Temporary scanline buffer for blit operations
****************************************************************************/
typedef struct PerVMInfo {
    N_uint32        cntBank;
    N_uint32        cntBits;
    GA_sharedInfo   *sharedInfo;
    ibool           isTextMode;
    N_int32         xRes;
    N_int32         yRes;
    N_int32         bitsPerPixel;
    N_int32         bytesPerLine;
    ulong           cntMode;
    ulong           cntMaxMem;
    ulong           cntLinearBufAddr;
    int             cntDACWidth;
    int             cntStartX;
    int             cntStartY;
    int             cntBytesPerLine;
    ulong           *PDB;
    ulong           *PageTable;
    ibool           isBankedEmulation;
    uchar           numPages;
    N_int32         fbOffset;
    void            *frameBufferBase;
    void            *frameBuffer;
    void            *frameBuffer2;
    void            *scanBuf;
    void            (NAPIP SetBank)(N_int32 bank);
    void            (*SetFBOffset)(N_int32 offset);
    void            (*PutPixel)(N_int32 x,N_int32 y,N_uint32 color);
    N_uint32        (*GetPixel)(N_int32 x,N_int32 y);
    void            (*DrawChar)(N_int32 left,N_int32 top,N_int32 cellHeight,N_uint8 *bitmap,GA_color foreColor,GA_color backColor);
    void            (*ReadChar)(N_int32 left,N_int32 top,N_int32 cellHeight,N_uint8 *bitmap);
    void            (*Scroll)(N_int32 left,N_int32 top,N_int32 right,N_int32 bottom,N_int32 numLines,N_int32 cellHeight,N_uint8 attribute);
    void            (*BitBlt)(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop);
    void            (*FillAttribute)(N_int32 left,N_int32 top,N_int32 width,N_int32 height,N_uint8 attribute);
    } PerVMInfo;

/****************************************************************************
REMARKS:
Structure defining all the VBE/Core emulation driver export functions.
{secret}
****************************************************************************/
typedef struct {
    ulong           dwSize;
    int             (PMAPIP VEM_getRMBufSize)(void);
    ibool           (PMAPIP VEM_init)(GA_devCtx *dc,VEM_DisableFlags *flags,void *rmBufAddr,uint rmBufSeg,uint rmBufOff,uint dataSel,uint codeSel);
    void            (PMAPIP VEM_createVM)(PerVMInfo *vmInfo);
    void            (PMAPIP VEM_destroyVM)(PerVMInfo *vmInfo);
    ibool           (PMAPIP VEM_onInt10)(VEM_REGS *pRegs);
    void            (PMAPIP VEM_bankSwitch)(uint bank,ibool fastPath);
    void            (PMAPIP VEM_exit)(void);
    } VEM_exports;

/****************************************************************************
REMARKS:
Structure defining all the VBE/Core emulation driver import functions
imported from the OS specific code.
{secret}
****************************************************************************/
typedef struct {
    ulong           dwSize;
    ibool           (PMAPIP VEM_inFullscreenDOSBox)(void);
    ibool           (PMAPIP VEM_isSysVM)(void);
    void            (PMAPIP VEM_simulateVGAIO)(void);
    void            (PMAPIP VEM_enablePortTrapping)(ibool enable);
    ulong           (PMAPIP VEM_getPDB)(void);
    void            (PMAPIP VEM_flushTLB)(void);
    PerVMInfo *     (PMAPIP VEM_enterCriticalSection)(void);
    void            (PMAPIP VEM_leaveCriticalSection)(void);
    int             (NAPIP DDC_init)(GA_devCtx *dc);
    ibool           (NAPIP DDC_readEDID)(N_int32 slaveAddr,uchar *edid,N_int32 length,N_int32 blockNumber,N_int32 monitorPort);
    void            (NAPIP GA_loadModeProfile)(GA_devCtx *dc,GA_modeProfile *profile);
    } VEM_imports;

/****************************************************************************
REMARKS:
Function pointer type for the Binary Portable DLL initialisation entry point.
{secret}
****************************************************************************/
typedef VEM_exports * (NAPIP VEM_initLibrary_t)(PM_imports *pmImp,VEM_imports *vemImp);

#pragma pack()

/*------------------------- Function Prototypes ---------------------------*/

#ifdef  __cplusplus
extern "C" {            /* Use "C" linkage when in C++ mode */
#endif

/* Return the size of the real mode memory buffer */

int     PMAPI VEM_getRMBufSize(void);

/* Initialise the VBE/Core emulation driver */

ibool   PMAPI VEM_init(GA_devCtx *dc,VEM_DisableFlags *flags,void *rmBufAddr,uint rmBufSeg,uint rmBufOff,uint dataSel,uint codeSel);

/* Handle the Int 10h software interrupt */

ibool   PMAPI VEM_onInt10(VEM_REGS *pRegs);

/* Fast path entry point to bank switch function (set bank only) */

void    PMAPI VEM_bankSwitch(uint bank,ibool fastPath);

/* Function to create a new virtual machine and initialise internal variables */

void    PMAPI VEM_createVM(PerVMInfo *vmInfo);

/* Function to destroy the virtual machine and free any internal structures */

void    PMAPI VEM_destroyVM(PerVMInfo *vmInfo);

/* Clean up and exit the VBE/Core emulation driver */

void    PMAPI VEM_exit(void);

/* OS provided service to determine if we are currently in a DOS box. This
 * should return true if the foreground process is a fullscreen DOS box.
 * In non Windows 3.x and 9x environments, this function should always
 * return true.
 */

ibool   PMAPI VEM_inFullscreenDOSBox(void);

/* OS provided service to that should return true if the current VM is the
 * system VM, or false otherwise. For non Windows 3.x and 9x environements,
 * this function should always return false (ie: environments where the
 * emulated VBE/Core is never used by the Windows drivers).
 */

ibool   PMAPI VEM_isSysVM(void);

/* OS provided service to simulate VGA registers to the VDD where necessary.
 * Usually called after modes sets to reflect the ring 0 VGA register
 * state for important registers through to the VDD's virtualised state.
 */

void    PMAPI VEM_simulateVGAIO(void);

/* OS provided service to enable and disable I/O port trapping for the
 * VGA registers. This is called at runtime to enable and disable the
 * port trapping where necessary.
 */

void    PMAPI VEM_enablePortTrapping(ibool enable);

/* OS provided service to ensure that the currently executing virtual
 * machine has exclusive access to SNAP as well as ensuring that
 * the virtual machines's GA_sharedInfo structure is the active one.
 */
PerVMInfo * PMAPI VEM_enterCriticalSection(void);
void        PMAPI VEM_leaveCriticalSection(void);

/* OS provided services for getting the page directory base register and
 * flushing the TLB. These functions are used to implement virtualised
 * banked framebuffer modes for backwards compatibility, and need to run
 * at ring 0. If necessary the OS driver will call down into ring 0
 * to execute them (if the emulation driver itself is not running at ring 0).
 */

ulong   PMAPI VEM_getPDB(void);
void    PMAPI VEM_flushTLB(void);

/* OS provided virtual port I/O functions that perform port I/O in such a
 * manner that it is reflected through the DOS box. Ie: any other drives
 * that are trapping port I/O in a DOS box (such as the Windows VDD etc)
 * will see the port I/O writes so they will remain in sync with the
 * virtual BIOS emulator.
 */
void    PMAPI VEM_outpb(ulong port, uchar value);
uchar   PMAPI VEM_inpb(ulong port);
void    PMAPI VEM_outpw(ulong port, ushort value);
ushort  PMAPI VEM_inpw(ulong port);
void    PMAPI VEM_outpd(ulong port, ulong value);
ulong   PMAPI VEM_inpd(ulong port);
uchar   PMAPI VEM_rdinx(ulong port,uchar index);
void    PMAPI VEM_wrinx(ulong port,uchar index,uchar value);
void    PMAPI VEM_modinx(ulong port,uchar index,uchar mask,uchar value);
void    PMAPI VEM_setinx(ulong port,uchar index,uchar mask);
void    PMAPI VEM_clrinx(ulong port,uchar index,uchar mask);

#ifdef  __cplusplus
}                       /* End of "C" linkage for C++   */
#endif

#endif  /* __SNAP_GRAPHICS_VBECORE_H */

