/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any 32-bit ring-0 protected mode environment
*
* Description:  C module for the AGP GART Driver API. Uses the SciTech PM
*               library for interfacing with OS specific functions.
*
****************************************************************************/

#include "snap/agp/common.h"

/*------------------------- Global Variables ------------------------------*/

char                        SNAPDriverPath[PM_MAX_PATH];
char                        SNAPConfigPath[PM_MAX_PATH];
char                        SNAPLogName[] = "agp.log";
static AGP_globalOptions    globalOpt;
static PCIDeviceInfo        *PCI;
static int                  *BridgeIndex;
static int                  NumPCI = -1;
static int                  NumBridges;
static PCIDeviceInfo        *AGPDevice = NULL;
static PCIDeviceInfo        *AGPBridge = NULL;
static PCIDeviceInfo        *AGPHostBridge = NULL;
static PCIDeviceInfo        *FirstHostBridge = NULL;
static int                  NumDevices;
static PE_MODULE            *hModule = NULL;
static AGP_devCtx           *agp = NULL;
static int                  agpRefCount = 0;

AGP_imports                 _VARAPI _AGP_imports;
PM_imports                  _VARAPI _PM_imports;
N_imports                   _VARAPI _N_imports;

DRV_imports _VARAPI _AGP_drvImports = {
    sizeof(DRV_imports),
    N_log,
    PCI_accessReg,
    _OS_delay,
    N_createBlockList,
    N_resizeBlockList,
    N_freeBlockList,
    N_buildFreeList,
    N_initList,
    N_destroyList,
    N_addAfter,
    N_addToHead,
    N_addToTail,
    N_removeNext,
    N_firstNode,
    N_nextNode,
    N_findNode,
    };

/*-------------------------- Implementation -------------------------------*/

#define OPTIONS_NAME    "agpopt.dat"

/****************************************************************************
REMARKS:
Enumerates the PCI bus and dumps the PCI configuration information to the
log file.
****************************************************************************/
static void EnumeratePCI(void)
{
    int             i;
    PCIDeviceInfo   *info;

    _log("Displaying enumeration of PCI bus (%d devices, %d display devices)\n",
        NumPCI, NumDevices);
    _log("\n");
    _log("Bus Slot Fnc DeviceID  SubSystem Rev Class IRQ Base10h  Base14h  Cmd\n");
    for (i = 0; i < NumPCI; i++) {
        _log("%2d   %2d  %2d  %04X:%04X %04X:%04X %02X  %02X:%02X %02X  %08X %08X %04X   ",
            PCI[i].slot.p.Bus,
            PCI[i].slot.p.Device,
            PCI[i].slot.p.Function,
            PCI[i].VendorID,
            PCI[i].DeviceID,
            PCI[i].u.type0.SubSystemVendorID,
            PCI[i].u.type0.SubSystemID,
            PCI[i].RevID,
            PCI[i].BaseClass,
            PCI[i].SubClass,
            PCI[i].u.type0.InterruptLine,
            PCI[i].u.type0.BaseAddress10,
            PCI[i].u.type0.BaseAddress14,
            PCI[i].Command);
        if (&PCI[i] == AGPDevice)
            _log("<- AGP Card\n");
        else
            _log("\n");
        }
    _log("\n");
    _log("DeviceID  Stat Ifc Cch Lat Hdr BIST Int Base18h  Base1Ch  Base20h  Base24h  \n");
    for (i = 0; i < NumPCI; i++) {
        _log("%04X:%04X %04X  %02X  %02X  %02X  %02X  %02X  %02X  %08X %08X %08X %08X ",
            PCI[i].VendorID,
            PCI[i].DeviceID,
            PCI[i].Status,
            PCI[i].Interface,
            PCI[i].CacheLineSize,
            PCI[i].LatencyTimer,
            PCI[i].HeaderType,
            PCI[i].BIST,
            PCI[i].u.type0.InterruptPin,
            PCI[i].u.type0.BaseAddress18,
            PCI[i].u.type0.BaseAddress1C,
            PCI[i].u.type0.BaseAddress20,
            PCI[i].u.type0.BaseAddress24);
        if (&PCI[i] == AGPDevice)
            _log("<- AGP Card\n");
        else
            _log("\n");
        }
    _log("\n");
    _log("DeviceID  BAR10Len BAR14Len BAR18Len BAR1CLen BAR20Len BAR24Len ROMLen\n");
    for (i = 0; i < NumPCI; i++) {
        _log("%04X:%04X %08X %08X %08X %08X %08X %08X %08X ",
            PCI[i].VendorID,
            PCI[i].DeviceID,
            PCI[i].u.type0.BaseAddress10Len,
            PCI[i].u.type0.BaseAddress14Len,
            PCI[i].u.type0.BaseAddress18Len,
            PCI[i].u.type0.BaseAddress1CLen,
            PCI[i].u.type0.BaseAddress20Len,
            PCI[i].u.type0.BaseAddress24Len,
            PCI[i].u.type0.ROMBaseAddressLen);
        _log("\n");
        }
    _log("\n");
    _log("Displaying enumeration of %d bridge devices\n",NumBridges);
    _log("\n");
    _log("DeviceID  P# S# B# IOB  IOL  MemBase  MemLimit PreBase  PreLimit Ctrl\n");
    for (i = 0; i < NumBridges; i++) {
        info = &PCI[BridgeIndex[i]];
        _log("%04X:%04X %02X %02X %02X %04X %04X %08X %08X %08X %08X %04X ",
            info->VendorID,
            info->DeviceID,
            info->u.type1.PrimaryBusNumber,
            info->u.type1.SecondayBusNumber,
            info->u.type1.SubordinateBusNumber,
            ((u16)info->u.type1.IOBase << 8) & 0xF000,
            info->u.type1.IOLimit ?
                ((u16)info->u.type1.IOLimit << 8) | 0xFFF : 0,
            ((u32)info->u.type1.MemoryBase << 16) & 0xFFF00000,
            info->u.type1.MemoryLimit ?
                ((u32)info->u.type1.MemoryLimit << 16) | 0xFFFFF : 0,
            ((u32)info->u.type1.PrefetchableMemoryBase << 16) & 0xFFF00000,
            info->u.type1.PrefetchableMemoryLimit ?
                ((u32)info->u.type1.PrefetchableMemoryLimit << 16) | 0xFFFFF : 0,
            info->u.type1.BridgeControl);
        if (info == AGPBridge)
            _log("<- AGP Bridge\n");
        else if (info == AGPHostBridge)
            _log("<- AGP Host Bridge\n");
        else
            _log("\n");
        }
    _log("\n");
}

/****************************************************************************
PARAMETERS:
info    - PCI device information to search for AGP capabilities block

RETURNS:
Offset in config space to start of AGP capabilities block, 0 if none found.
****************************************************************************/
uchar PCI_findAGPCaps(
    PCIDeviceInfo *info,
    PCIAGPCapability *caps)
{
    uchar capPtr;

    // Check the Status register to see if this device supports capability
    // lists. If not, it is not an AGP-compliant device.
    if ((info->Status & PCI_statusCapabilitiesList) == 0)
        return 0;

    // The device supports capability lists, find the AGP capabilities
    if ((info->HeaderType & (~PCI_multiFunctionType)) == PCI_bridgeType)
        capPtr = info->u.type1.CapabilitiesPointer;
    else
        capPtr = info->u.type0.CapabilitiesPointer;

    // Search for the AGP capabilities list
    while (capPtr != 0) {
        // Read the start of the capabilities block at this offset
        PCI_readRegBlock(info,capPtr,&caps->h,sizeof(caps->h));
        if (caps->h.capsID == PCI_capsAGP) {
            // Found the AGP capabilities block
            return capPtr;
            }
        else {
            // Some other capability, so keep looking
            capPtr = caps->h.next;
            }
        }

    // If we get here, then no AGP capabilities list was found!
    return 0;
}

/****************************************************************************
PARAMETERS:
info    - PCI device information to get for AGP capabilities block for
caps    - Place to store the returned master device capabilities structure

RETURNS:
True on success, false on failure
****************************************************************************/
ibool PCI_getAGPCaps(
    PCIDeviceInfo *info,
    PCIAGPCapability *caps)
{
    uchar capPtr;

    // Find the header block for the AGP capabilities list
    if ((capPtr = PCI_findAGPCaps(info,caps)) == 0)
        return false;

    // Now read the rest of the AGP capabilities list
    PCI_readRegBlock(info,capPtr+sizeof(caps->h),&caps->h+1,sizeof(*caps)-sizeof(caps->h));
    return true;
}

/****************************************************************************
PARAMETERS:
info    - PCI device information to get for AGP capabilities block for
caps    - Place to store the returned master device capabilities structure

RETURNS:
True on success, false on failure
****************************************************************************/
ibool PCI_setAGPCaps(
    PCIDeviceInfo *info,
    PCIAGPCapability *caps)
{
    uchar capPtr;

    // Find the header block for the AGP capabilities list
    if ((capPtr = PCI_findAGPCaps(info,caps)) == 0)
        return false;

    // Now write the AGP command register
    PCI_writeRegBlock(info,
        capPtr+PM_FIELD_OFFSET(PCIAGPCapability,AGPCommand),
        &caps->AGPCommand,sizeof(caps->AGPCommand));
    return true;
}

/****************************************************************************
RETURNS:
Pointer to the AGP bridge PCI device information structure

REMARKS:
This function enumerates the number of available display devices on the
PCI bus, and returns the associated AGP bridge device if found.
****************************************************************************/
static PCIDeviceInfo *AGP_findDevice(
    N_uint16 vendorID,
    N_uint16 deviceID)
{
    int i;

    // Now build a list of all display class devices
    for (i = 0; i < NumPCI; i++) {
        if (PCI[i].VendorID == vendorID && PCI[i].DeviceID == deviceID)
            return &PCI[i];
        }
    return NULL;
}

/****************************************************************************
RETURNS:
Pointer to the AGP bridge PCI device information structure

REMARKS:
This function enumerates the number of available display devices on the
PCI bus, and returns the associated AGP bridge device if found.
****************************************************************************/
static ibool AGP_findBridge(
    N_int32 deviceIndex)
{
    int                 i,j,chipID;
    PCIDeviceInfo       *info;
    PCIAGPCapability    agpCaps;

    // If this is the first time we have been called, enumerate all
    // devices on the PCI bus.
    if (NumPCI == -1) {
        // First determine how many devices are on the PCI bus, and allocate
        // enough memory to describe all the devices.
        if ((NumPCI = PCI_getNumDevices()) == 0)
            return -1;
        if ((PCI = calloc(1,NumPCI * (sizeof(PCI[0]) + sizeof(BridgeIndex[0])))) == NULL)
            return false;
        BridgeIndex = (int*)(PCI + NumPCI);
        for (i = 0; i < NumPCI; i++)
            PCI[i].dwSize = sizeof(PCI[i]);
        if (PCI_enumerate(PCI) == 0)
            return false;

        // Build a list of all PCI bridge devices
        for (i = 0,NumBridges = 0,BridgeIndex[0] = -1; i < NumPCI; i++) {
            if (PCI[i].BaseClass == PCI_BRIDGE_CLASS) {
                BridgeIndex[NumBridges++] = i;
                if (PCI[i].SubClass == PCI_HOST_BRIDGE_SUBCLASS) {
                    if (!AGPHostBridge && PCI_getAGPCaps(&PCI[i],&agpCaps))
                        AGPHostBridge = &PCI[i];
                    if (!FirstHostBridge)
                        FirstHostBridge = &PCI[i];
                    }
                }
            }

        // Now build a list of all display class devices
        for (i = 0; i < NumPCI; i++) {
            if (PCI_IS_DISPLAY_CLASS(&PCI[i])) {
                if (PCI[i].slot.p.Bus != 0) {
                    // This device is on a different bus than the primary
                    // PCI bus, so it is probably an AGP device. Find the
                    // AGP bus device that controls that bus so we can
                    // control it.
                    for (j = 0; j < NumBridges; j++) {
                        info = &PCI[BridgeIndex[j]];
                        if (info->u.type1.SecondayBusNumber == (int)PCI[i].slot.p.Bus) {
                            AGPDevice = &PCI[i];
                            AGPBridge = info;
                            break;
                            }
                        }
                    }
                }
            }

        // If we did not find an AGP device and this is an Intel i810/815 based
        // system, then we need to handle this special case here. These systems
        // are not really AGP, so we need to match up the display device in
        // here specifically.
        if (!AGPDevice && FirstHostBridge->VendorID == 0x8086) {
            AGPHostBridge = FirstHostBridge;
            switch (AGPHostBridge->DeviceID) {
                case 0x7120: chipID = 0x7121; break;    // Intel i810
                case 0x7122: chipID = 0x7123; break;    // Intel i810/DC100
                case 0x7124: chipID = 0x7125; break;    // Intel i810E
                case 0x1130: chipID = 0x1132; break;    // Intel i815
                case 0x2560: chipID = 0x2562; break;    // Intel i845 Brookdale-G
                    break;
                }
            AGPDevice = AGP_findDevice(AGPHostBridge->VendorID,chipID);
            }

        // TODO: Add special code to check for and find the ATI devices that
        //       support a PCI GART if we did not find a compatible AGP
        //       bridge. Note that we need to fall back on this if we *do*
        //       find an AGP device however it is not one that we have a
        //       driver for!!

        // Now dump all PCI information to the log file
        EnumeratePCI();
        }
    (void)deviceIndex;
    return (AGPDevice != NULL);
}

/****************************************************************************
REMARKS:
Function to log the currently configured monitor information to disk.
****************************************************************************/
static void DumpDriverInfo(
    AGP_devCtx *dc)
{
    AGP_driverFuncs agp;
    AGP_configInfo  info;

    agp.dwSize = sizeof(agp);
    if (!dc->QueryFunctions(AGP_GET_DRIVERFUNCS,&agp))
        PM_fatalError("Unable to get device driver functions!");
    info.dwSize = sizeof(info);
    agp.GetConfigInfo(&info);
    _log("\nAGP GART device configuration:\n");
    _log("  Manufacturer......... %s\n"
         "  Chipset.............. %s\n"
         "  Driver Revision...... %s\n"
         "  Driver Build......... %s\n",
        info.ManufacturerName,
        info.ChipsetName,
        info.VersionInfo,
        info.BuildDate);
    if (info.Certified) {
        _log("  Certified Version.... %d.%02d\n"
             "  Certified Date....... %s\n",
            info.CertifyVersion >> 8,
            info.CertifyVersion & 0xFF,
            info.CertifiedDate);
        }
    else {
        _log("  Certified Version.... Not Certified\n");
        }
}

/****************************************************************************
PARAMETERS:
dc      - Pointer to device context
device  - Device index of the device we are initialising

RETURNS:
Error code

REMARKS:
Performs internal initialisation on the AGP_devCtx driver block, assuming
that it has been loaded correctly.
{secret}
****************************************************************************/
N_int32 _AGP_initInternal(
    AGP_devCtx *agp,
    int device)
{
    /* Verify that the file is the driver file we are expecting */
    if (strcmp(agp->Signature,AGP_SIGNATURE) != 0)
        return status = nCorruptDriver;

    /* Map any memory mapped I/O regions that are necessary */
    if (agp->MMIOBase) {
        if ((agp->IOMemMaps[0] = PM_mapPhysicalAddr(agp->MMIOBase,agp->MMIOLen-1,false)) == NULL) {
            _log("Failed to map IO memory: 0x%08X for 0x%08X bytes\n",agp->MMIOBase,agp->MMIOLen);
            return status = nMemMapError;
            }
        }

    /* Now call the driver to initialise itself */
    return (status = agp->InitDriver());
}

/****************************************************************************
PARAMETERS:
szBPDName   - Name of BPD file to load

REMARKS:
Internal function to load a BPD file from inside the SNAP flat file
for the main device driver. This code also handles loading files from the
regular file system as an override to files in the flat file.
{secret}
****************************************************************************/
static PE_MODULE * AGP_loadLibrary(
    const char *szBPDName)
{
    PE_MODULE   *hMod;
    N_FILE      *f;
    char        path[PM_MAX_PATH];
    ulong       size;

    // First try to load an external binary
    strcpy(path,SNAPDriverPath);
    strcat(path,szBPDName);
    if ((hMod = PE_loadLibrary(path,false)) != NULL)
        return hMod;

    // Next try in the bound binary if not found externally
    if ((f = FLAT_openFile(szBPDName)) == NULL)
        return NULL;
    hMod = PE_loadLibraryExt(f->f,f->start,&size,false);
    FLAT_closeFile(f);
    return hMod;
}

/****************************************************************************
DESCRIPTION:
Loads an AGP GART driver file from disk for the machine.

HEADER:
snap/agp.h

RETURNS:
Pointer to the loaded AGP driver, or NULL on error.

REMARKS:
Loads the driver file from disk and intialises the device context ready for
use. If the driver file cannot be found, or the driver does not detect the
installed hardware, we return NULL and the application can get the status
code with the AGP_status function.

SEE ALSO:
AGP_unloadDriver, GA_status, GA_errorMsg
****************************************************************************/
AGP_devCtx * NAPI AGP_loadDriver(
    N_int32 deviceIndex)
{
    char                filename[PM_MAX_PATH];
    AGP_initDriver_t    AGP_initDriver;

    // See if the driver has already been loaded first, and return it.
    if (agp) {
        agpRefCount++;
        return agp;
        }

    // Load global options to initialise defaults
    AGP_getGlobalOptions(&globalOpt);

    // Find the AGP bridge device on the PCI bus
    status = nDriverNotFound;
    if (!AGP_findBridge(deviceIndex))
        return NULL;
    _log("Loading driver for AGP host bridge %04X:%04X\n", AGPHostBridge->VendorID,AGPHostBridge->DeviceID);
    _log("--------------------------------------------\n");

    // Now find the name of the driver to load from the directory table
    if (!_PCI_findDriverName(AGPHostBridge,filename,"certify.tab")) {
        if (!globalOpt.bAllowNonCertified || !_PCI_findDriverName(AGPHostBridge,filename,"driver.tab")) {
            _log("Driver not found!\n");
            return NULL;
            }
        }

    // Now try to open the driver file
    if ((hModule = AGP_loadLibrary(filename)) == NULL) {
        _log("Unable to load driver '%s'!\n", filename);
        return NULL;
        }
    status = nCorruptDriver;
    if ((AGP_initDriver = PE_getProcAddress(hModule,"_AGP_initDriver")) == NULL)
        goto Error;
    status = nNotDetected;

    // Now try to initialise the driver file
    if ((agp = AGP_initDriver(SNAPConfigPath,deviceIndex,AGPDevice,AGPBridge,AGPHostBridge,&_N_imports,&_PM_imports,&_AGP_drvImports,&globalOpt)) == NULL)
        goto Error;
    if (_AGP_initInternal(agp,deviceIndex) != nOK)
        goto Error;

    // Log the configured AGP bridge information!
    DumpDriverInfo(agp);

    _log("\n");
    _log("Driver for AGP host bridge %04X:%04X loaded successfully!\n", AGPHostBridge->VendorID,AGPHostBridge->DeviceID);
    _log("---------------------------------------------------------\n\n");

    // Success!
    agpRefCount = 1;
    status = nOK;
    return agp;

Error:
    PE_freeLibrary(hModule);
    return NULL;
}

/****************************************************************************
DESCRIPTION:
Unloads the loaded AGP device driver.

HEADER:
snap/agp.h

PARAMETERS:
dc  - Pointer to device context to unload

REMARKS:
This function simply unloads the AGP driver from memory, and frees
any of the internal resources that have been allocated for it.

SEE ALSO:
AGP_loadDriver
****************************************************************************/
void NAPI AGP_unloadDriver(
    AGP_devCtx *dc)
{
    if (--agpRefCount == 0) {
        agp->UnloadDriver();
        PE_freeLibrary(hModule);
        agp = NULL;
        }
}

/****************************************************************************
DESCRIPTION:
Returns the current global options from the AGP device driver.

HEADER:
snap/agp.h

PARAMETERS:
options - Place to store the returned options information

REMARKS:
This function returns a structure which contains global configuration
options effecitve for the installed devices.

Note:   The dwSize member of the profile structure is intended for future
        compatibility, and must be set to the size of the structure before
        calling this function. Only the number of bytes set in the
        dwSize member will be copied into the callers structure.

SEE ALSO:
AGP_setGlobalOptions, AGP_saveGlobalOptions
****************************************************************************/
void NAPI AGP_getGlobalOptions(
    AGP_globalOptions *options)
{
    FILE    *f;
    char    filename[PM_MAX_PATH];

    memset(&globalOpt,0,sizeof(globalOpt));
    globalOpt.dwSize = sizeof(globalOpt);
    globalOpt.bAllowNonCertified = true;
    globalOpt.wCertifiedVersion = 0x100;
    globalOpt.bEnableFastWrite = true;
    globalOpt.maxAGPRate = 8;
    strcpy(filename,SNAPConfigPath);
    PM_backslash(filename);
    strcat(filename,OPTIONS_NAME);
    if ((f = fopen(filename,"rb")) != NULL) {
        fread(&globalOpt,1,sizeof(globalOpt),f);
        fclose(f);
        if (globalOpt.dwSize >= sizeof(globalOpt))
            globalOpt.dwSize = sizeof(globalOpt);
        }
    COPY_STRUCTURE(options,&globalOpt);
}

/****************************************************************************
DESCRIPTION:
Sets the current global options for the AGP device driver.

HEADER:
snap/agp.h

PARAMETERS:
options - Global options to make active.

REMARKS:
This function installs a new set of global device options that control the
operation of all the devices at runtime. A default set of options is
always built into the device driver, but the options can be changed at
any time. The options may be made permanent with a call to the
AGP_saveGlobalOptions function.

Note:   The dwSize member of the profile structure is intended for future
        compatibility, and must be set to the size of the structure before
        calling this function. Only the number of bytes set in the
        dwSize member will be copied from the callers structure.

SEE ALSO:
AGP_getGlobalOptions, AGP_saveGlobalOptions
****************************************************************************/
void NAPI AGP_setGlobalOptions(
    AGP_globalOptions *options)
{
    globalOpt.dwSize = sizeof(globalOpt);
    COPY_STRUCTURE(&globalOpt,options);
}

/****************************************************************************
DESCRIPTION:
Save the global options to disk

HEADER:
snap/agp.h

PARAMETERS:
options - Global options information block to save

REMARKS:
This function saves the global options information to disk.

SEE ALSO:
AGP_getGlobalOptions, AGP_setGlobalOptions
****************************************************************************/
void NAPI AGP_saveGlobalOptions(
    AGP_globalOptions *options)
{
    FILE    *f;
    char    filename[PM_MAX_PATH];

    strcpy(filename,SNAPConfigPath);
    PM_backslash(filename);
    strcat(filename,OPTIONS_NAME);
    if ((f = fopen(filename,"wb")) != NULL) {
        fwrite(options,1,options->dwSize,f);
        fclose(f);
        }
}

/****************************************************************************
DESCRIPTION:
Returns the error code for AGP_loadDriver if the function failed.

HEADER:
snap/agp.h

RETURNS:
Status code for the AGP_loadDriver function.

REMARKS:
Returns the error code for AGP_loadDriver if the function failed.

SEE ALSO:
AGP_errorMsg
****************************************************************************/
int NAPI AGP_status(void)
{ return status; }

/****************************************************************************
DESCRIPTION:
Returns a descriptive string for a device loader error code.

HEADER:
snap/agp.h

PARAMETERS:
status  - Status code to get string for

RETURNS:
String describing error condition.

REMARKS:
Returns a descriptive string for a device error code.

SEE ALSO:
AGP_status
****************************************************************************/
const char * NAPI AGP_errorMsg(
    N_int32 status)
{
    const char *msg[] = {
        "No error",
        "Hardware not detected",
        "Hardware has not been POST'ed by the BIOS",
        "Driver file not found",
        "File loaded was not a driver file",
        "Not enough memory to load driver",
        "Driver file is an older version",
        "Could not map physical memory areas",
        "General I/O error",
        "Could not hook requested hardware IRQ",
        "Driver is not certified",
        "Internal device driver error",
        "Not enough memory to complete operation",
        "Not enough spare resources to complete operation",
        "Invalid parameter passed to function",
        "No AGP support services found",
        };
    if (status >= nOK && status < nMaxError)
        return msg[status];
    return "Unknown error!";
}

/****************************************************************************
REMARKS:
Fatal error handler called when a non-imported function is called by the
driver. We leave this to a runtime error so that older applications and
shell drivers will work with newer bpd drivers provided no newer functions
are required by the driver itself. If they are, the application or shell
driver needs to be recompiled.
****************************************************************************/
static void _PM_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported PM_imports import function called! Please re-compile!\n");
}

/****************************************************************************
REMARKS:
Fatal error handler for non-imported N_imports.
****************************************************************************/
static void _N_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported N_imports import function called! Please re-compile!\n");
}

/****************************************************************************
REMARKS:
Fatal error handler for non-imported AGP_imports.
****************************************************************************/
static void _AGP_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported AGP_imports import function called! Please re-compile!\n");
}

/****************************************************************************
PARAMETERS:
path    - Path to Binary Portable DLL
bpdname - Name of the Binary Portable DLL
pmImp   - PM library imports
nImp    - Generic SNAP imports
gaImp   - Graphics Architecture specific imports

RETURNS:
Pointer to exported function list

REMARKS:
This function initialises the SNAP Graphics Architecture loader library
and returns the list of loader library exported functions.
{secret}
****************************************************************************/
AGP_exports * _CEXPORT AGP_initLibrary(
    const char *path,
    const char *bpdname,
    PM_imports *pmImp,
    N_imports *nImp,
    AGP_imports *gaImp)
{
    static AGP_exports _AGP_exports = {
        sizeof(AGP_exports),
        AGP_status,
        AGP_errorMsg,
        AGP_loadDriver,
        AGP_unloadDriver,
        AGP_getGlobalOptions,
        AGP_setGlobalOptions,
        AGP_saveGlobalOptions,
        };
    int     i,max;
    ulong   *p;
    char    drive[PM_MAX_DRIVE];
    char    dir[PM_MAX_PATH];

    // Initialize all default imports to point to fatal error handler
    // for upwards compatibility.
    max = sizeof(_PM_imports)/sizeof(AGP_initLibrary_t);
    for (i = 0,p = (ulong*)&_PM_imports; i < max; i++)
        *p++ = (ulong)_PM_fatalErrorHandler;
    max = sizeof(_N_imports)/sizeof(AGP_initLibrary_t);
    for (i = 0,p = (ulong*)&_N_imports; i < max; i++)
        *p++ = (ulong)_N_fatalErrorHandler;
    max = sizeof(_AGP_imports)/sizeof(AGP_initLibrary_t);
    for (i = 0,p = (ulong*)&_AGP_imports; i < max; i++)
        *p++ = (ulong)_AGP_fatalErrorHandler;

    // Now copy all our imported functions
    memcpy(&_PM_imports,pmImp,MIN(sizeof(_PM_imports),pmImp->dwSize));
    memcpy(&_N_imports,nImp,MIN(sizeof(_N_imports),nImp->dwSize));
    memcpy(&_AGP_imports,gaImp,MIN(sizeof(_AGP_imports),gaImp->dwSize));
    PM_splitpath(bpdname,drive,dir,NULL,NULL);
    PM_makepath(SNAPDriverPath,drive,dir,NULL,NULL);
    PM_backslash(SNAPDriverPath);
    strcat(SNAPDriverPath,"agp");
    PM_backslash(SNAPDriverPath);
    strcpy(SNAPConfigPath,path);
    PM_backslash(SNAPConfigPath);
#ifndef CHECKED
    if (!FLAT_init(bpdname))
        return false;
#endif
    return &_AGP_exports;
}

