/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Header file containing the internal chipset definitions
*               for the driver.
*
****************************************************************************/

#ifndef __INTERNAL_H
#define __INTERNAL_H

#include "snap/agp/driver.h"

/*--------------------- Macros and type definitions -----------------------*/

// Chip revision types

typedef enum {
    INTEL_440LX,
    INTEL_440BX,
    INTEL_440GX,
    } AGP_chipsets;

// Offsets of control registers for i440 in PCI configuration space

#define APBASE_OFFSET           0x10    // Aperture Base Address
#define APSIZE_OFFSET           0xB4    // Aperture Size Register
#define PACCFG_OFFSET           0x50    // PAC Configuration Register
#define AGPCTRL_OFFSET          0xB0    // AGP Control Register
#define ATTBASE_OFFSET          0xB8    // Aperture Translation Table Base

// Conversions from APSIZE encoding to MB
//
// 0x3F (b 11 1111) =   4MB
// 0x3E (b 11 1110) =   8MB
// 0x3C (b 11 1100) =  16MB
// 0x38 (b 11 1000) =  32MB
// 0x30 (b 11 0000) =  64MB
// 0x20 (b 10 0000) = 128MB
// 0x00 (b 00 0000) = 256MB

#define AP_SIZE_4MB             0x3F
#define AP_SIZE_8MB             0x3E
#define AP_SIZE_16MB            0x3C
#define AP_SIZE_32MB            0x38
#define AP_SIZE_64MB            0x30
#define AP_SIZE_128MB           0x20
#define AP_SIZE_256MB           0x00

// GART Entry states are defined so that all software-only states
// have the Valid bit clear.

#define GART_ENTRY_VALID        1           //  001
#define GART_ENTRY_FREE         0           //  000

#define GART_ENTRY_WC           2           //  010
#define GART_ENTRY_UC           4           //  100

#define GART_ENTRY_RESERVED_WC  GART_ENTRY_WC
#define GART_ENTRY_RESERVED_UC  GART_ENTRY_UC

#define GART_ENTRY_VALID_WC     (GART_ENTRY_VALID)
#define GART_ENTRY_VALID_UC     (GART_ENTRY_VALID)

// Structure to define the GART table entries

typedef struct {
    N_uint32    Valid           : 1;
    N_uint32    Reserved        : 11;
    N_uint32    Page            : 20;
    } GART_entryHW;

typedef struct {
    N_uint32    State           : 3;
    N_uint32    Reserved        : 29;
    } GART_entrySW;

typedef struct {
    union {
        GART_entryHW    Hard;
        N_uint32        AsUlong;
        GART_entrySW    Soft;
        } e;
    } GART_pte;

// Define the layout of the hardware registers

typedef struct {
    N_uint32    Reserved1       : 7;
    N_uint32    GTLB_Enable     : 1;
    N_uint32    Reserved2       : 24;
    } AGPCTRL;

typedef struct {
    N_uint16    Reserved1       : 9;
    N_uint16    GlobalEnable    : 1;
    N_uint16    PCIEnable       : 1;
    N_uint16    Reserved2       : 5;
    } PACCFG;

#if 0

// Define the 440-specific extension
typedef struct {
    ibool           GlobalEnable;
    ibool           PCIEnable;
    N_physAddr      ApertureStart;
    N_uint32        ApertureLength;
    GART_pte        *GartCached;
    GART_pte        *Gart;
    N_uint32        GartLength;
    N_physAddr      GartPhysical;
    } AGP440_EXTENSION;
#endif

/*-------------------------- Function Prototypes --------------------------*/

N_int32 I440_ReserveMemory(AGP_physCtx *ctx);

#endif  /* __INTERNAL_H */

