;****************************************************************************
;*
;*                    SciTech SNAP Graphics Architecture
;*
;*  ========================================================================
;*
;*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
;*
;*   This file may be distributed and/or modified under the terms of the
;*   GNU General Public License version 2 as published by the Free
;*   Software Foundation and appearing in the file LICENSE.GPL included
;*   in the packaging of this file.
;*
;*   Licensees holding a valid Commercial License for this product from
;*   SciTech Software, Inc. may use this file in accordance with the
;*   Commercial License Agreement provided with the Software.
;*
;*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
;*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
;*   PURPOSE.
;*
;*   See http://www.scitechsoft.com/license/ for information about
;*   the licensing options available and how to purchase a Commercial
;*   License Agreement.
;*
;*   Contact license@scitechsoft.com if any conditions of this licensing
;*   are not clear to you, or you have questions about licensing options.
;*
;*  ========================================================================
;*
;* Language:    80386 Assembler
;* Environment: IBM PC (MS DOS)
;*
;* Description: Helper functions for SNAP device drivers.
;*
;****************************************************************************

include "scitech.mac"

header  _ports

begcodeseg  _ports

;----------------------------------------------------------------------------
; void enable(void)
;----------------------------------------------------------------------------
cprocstart  enable

        sti
        ret

cprocend

;----------------------------------------------------------------------------
; void disable(void)
;----------------------------------------------------------------------------
cprocstart  disable

        cli
        ret

cprocend

;----------------------------------------------------------------------------
; int _inpb(int port)
;----------------------------------------------------------------------------
; Reads a byte from the specified port
;----------------------------------------------------------------------------
cprocstart  _inpb

        ARG     port:UINT

        push    ebp
        mov     ebp,esp
        xor     eax,eax
        mov     edx,[port]
        in      al,dx
        pop     ebp
        ret

cprocend

;----------------------------------------------------------------------------
; int _inpw(int port)
;----------------------------------------------------------------------------
; Reads a word from the specified port
;----------------------------------------------------------------------------
cprocstart  _inpw

        ARG     port:UINT

        push    ebp
        mov     ebp,esp
        xor     eax,eax
        mov     edx,[port]
        in      ax,dx
        pop     ebp
        ret

cprocend

;----------------------------------------------------------------------------
; ulong _inpd(int port)
;----------------------------------------------------------------------------
; Reads a word from the specified port
;----------------------------------------------------------------------------
cprocstart  _inpd

        ARG     port:UINT

        push    ebp
        mov     ebp,esp
        xor     eax,eax
        mov     edx,[port]
        in      eax,dx
        pop     ebp
        ret

cprocend

;----------------------------------------------------------------------------
; void _outpb(int port,int value)
;----------------------------------------------------------------------------
; Write a byte to the specified port.
;----------------------------------------------------------------------------
cprocstart  _outpb

        ARG     port:UINT, value:UINT

        push    ebp
        mov     ebp,esp
        mov     edx,[port]
        mov     eax,[value]
        out     dx,al
        pop     ebp
        ret

cprocend

;----------------------------------------------------------------------------
; void _outpw(int port,int value)
;----------------------------------------------------------------------------
; Write a word to the specified port.
;----------------------------------------------------------------------------
cprocstart  _outpw

        ARG     port:UINT, value:UINT

        push    ebp
        mov     ebp,esp
        mov     edx,[port]
        mov     eax,[value]
        out     dx,ax
        pop     ebp
        ret

cprocend

;----------------------------------------------------------------------------
; void _outpd(int port,ulong value)
;----------------------------------------------------------------------------
; Write a word to the specified port.
;----------------------------------------------------------------------------
cprocstart  _outpd

        ARG     port:UINT, value:ULONG

        push    ebp
        mov     ebp,esp
        mov     edx,[port]
        mov     eax,[value]
        out     dx,eax
        pop     ebp
        ret

cprocend

;----------------------------------------------------------------------------
; uchar rolb(uchar val,uchar shift);
;----------------------------------------------------------------------------
cprocstart  rolb

        ARG     val:UCHAR, shift:UCHAR

        push    ebp
        mov     ebp,esp
        mov     al,[val]
        mov     cl,[shift]
        rol     al,cl
        pop     ebp
        ret

cprocend

;----------------------------------------------------------------------------
; ushort rolb_w(uchar val,uchar shift);
;----------------------------------------------------------------------------
cprocstart  rolb_w

        ARG     val:UCHAR, shift:UCHAR

        push    ebp
        mov     ebp,esp
        mov     al,[val]
        mov     cl,[shift]
        rol     al,cl
        mov     ah,al
        pop     ebp
        ret

cprocend

;----------------------------------------------------------------------------
; ushort rolw(ushort val,uchar shift);
;----------------------------------------------------------------------------
cprocstart  rolw

        ARG     val:USHORT, shift:UCHAR

        push    ebp
        mov     ebp,esp
        mov     ax,[val]
        mov     cl,[shift]
        rol     ax,cl
        pop     ebp
        ret

cprocend

;----------------------------------------------------------------------------
; ulong rolw_d(ushort val,uchar shift);
;----------------------------------------------------------------------------
cprocstart  rolw_d

        ARG     val:USHORT, shift:UCHAR

        push    ebp
        mov     ebp,esp
        mov     ax,[val]
        mov     cl,[shift]
        rol     ax,cl
        mov     edx,eax
        shl     eax,16
        mov     ax,dx
        pop     ebp
        ret

cprocend

;----------------------------------------------------------------------------
; ulong rold(ulong val,uchar shift);
;----------------------------------------------------------------------------
cprocstart  rold

        ARG     val:ULONG, shift:UCHAR

        push    ebp
        mov     ebp,esp
        mov     eax,[val]
        mov     cl,[shift]
        rol     eax,cl
        pop     ebp
        ret

cprocend

;----------------------------------------------------------------------------
; uchar rorb(uchar val,uchar shift);
;----------------------------------------------------------------------------
cprocstart  rorb

        ARG     val:UCHAR, shift:UCHAR

        push    ebp
        mov     ebp,esp
        mov     al,[val]
        mov     cl,[shift]
        ror     al,cl
        pop     ebp
        ret

cprocend

;----------------------------------------------------------------------------
; ushort rorb_w(uchar val,uchar shift);
;----------------------------------------------------------------------------
cprocstart  rorb_w

        ARG     val:UCHAR, shift:UCHAR

        push    ebp
        mov     ebp,esp
        mov     al,[val]
        mov     cl,[shift]
        ror     al,cl
        mov     ah,al
        pop     ebp
        ret

cprocend

;----------------------------------------------------------------------------
; ushort rorw(ushort val,uchar shift);
;----------------------------------------------------------------------------
cprocstart  rorw

        ARG     val:USHORT, shift:UCHAR

        push    ebp
        mov     ebp,esp
        mov     ax,[val]
        mov     cl,[shift]
        ror     ax,cl
        pop     ebp
        ret

cprocend

;----------------------------------------------------------------------------
; ulong rorw_d(ushort val,uchar shift);
;----------------------------------------------------------------------------
cprocstart  rorw_d

        ARG     val:USHORT, shift:UCHAR

        push    ebp
        mov     ebp,esp
        mov     ax,[val]
        mov     cl,[shift]
        ror     ax,cl
        mov     edx,eax
        shl     eax,16
        mov     ax,dx
        pop     ebp
        ret

cprocend

;----------------------------------------------------------------------------
; ulong rord(ulong val,uchar shift);
;----------------------------------------------------------------------------
cprocstart  rord

        ARG     val:ULONG, shift:UCHAR

        push    ebp
        mov     ebp,esp
        mov     eax,[val]
        mov     cl,[shift]
        ror     eax,cl
        pop     ebp
        ret

cprocend

;----------------------------------------------------------------------------
; void  ror_set_pattern(void *addr,void *pattern,int shift);
;----------------------------------------------------------------------------
cprocstart  ror_set_pattern

        ARG     addr:DPTR, pattern:DPTR, shift:UCHAR

        enter_c
        mov     edi,[addr]
        mov     esi,[pattern]
        mov     cl,[shift]
        mov     al,[esi+0]
        mov     ah,[esi+1]
        mov     bl,[esi+2]
        mov     bh,[esi+3]
        ror     al,cl
        ror     ah,cl
        ror     bl,cl
        ror     bh,cl
        mov     [edi+0],al
        mov     [edi+1],ah
        mov     [edi+2],bl
        mov     [edi+3],bh
        mov     al,[esi+4]
        mov     ah,[esi+5]
        mov     bl,[esi+6]
        mov     bh,[esi+7]
        ror     al,cl
        ror     ah,cl
        ror     bl,cl
        ror     bh,cl
        mov     [edi+4],al
        mov     [edi+5],ah
        mov     [edi+6],bl
        mov     [edi+7],bh
        leave_c
        ret

cprocend

;----------------------------------------------------------------------------
; void  rep_outsb(ulong port,uchar *src,int count);
;----------------------------------------------------------------------------
cprocstart  rep_outsb

        ARG     port:ULONG, src:DPTR, count:UINT

        enter_c
        mov     edx,[port]
        mov     esi,[src]
        mov     ecx,[count]
        rep     outsb
        leave_c
        ret

cprocend

;----------------------------------------------------------------------------
; void  rep_outsw(ulong port,uchar *src,int count);
;----------------------------------------------------------------------------
cprocstart  rep_outsw

        ARG     port:ULONG, src:DPTR, count:UINT

        enter_c
        mov     edx,[port]
        mov     esi,[src]
        mov     ecx,[count]
        rep     outsw
        leave_c
        ret

cprocend

;----------------------------------------------------------------------------
; void  rep_outsd(ulong port,uchar *src,int count);
;----------------------------------------------------------------------------
cprocstart  rep_outsd

        ARG     port:ULONG, src:DPTR, count:UINT

        enter_c
        mov     edx,[port]
        mov     esi,[src]
        mov     ecx,[count]
        rep     outsd
        leave_c
        ret

cprocend

;----------------------------------------------------------------------------
; void rep_outsdb(ulong port,void *src,int count);
;----------------------------------------------------------------------------
cprocstart  rep_outsdb

        ARG     port:ULONG, src:DPTR, count:UINT

        enter_c
        mov     edx,[port]
        mov     esi,[src]
        mov     eax,[count]
        rep     outsb
        mov     ecx,eax
        shr     ecx,2
        rep     outsd
        mov     ecx,eax
        and     ecx,3
        jz      @@Exit
        mov     al,[esi]
        dec     ecx
        jz      @@Done
        mov     ah,[esi+1]
        dec     ecx
        jz      @@Done
        rol     eax,16
        out     dx,eax
        rol     eax,16
@@Done: mov     [edi],eax
@@Exit: leave_c
        ret

cprocend

;----------------------------------------------------------------------------
; void * rep_movsb(void *dst,void *src,int count);
;----------------------------------------------------------------------------
cprocstart  rep_movsb

        ARG     dst:DPTR, src:DPTR, count:UINT

        enter_c
        mov     edi,[dst]
        mov     esi,[src]
        mov     ecx,[count]
        rep     movsb
        mov     eax,edi
        leave_c
        ret

cprocend

;----------------------------------------------------------------------------
; void * rep_movsw(void *dst,void *src,int count);
;----------------------------------------------------------------------------
cprocstart  rep_movsw

        ARG     dst:DPTR, src:DPTR, count:UINT

        enter_c
        mov     edi,[dst]
        mov     esi,[src]
        mov     ecx,[count]
        rep     movsw
        mov     eax,edi
        leave_c
        ret

cprocend

;----------------------------------------------------------------------------
; void * rep_movsd(void *dst,void *src,int count);
;----------------------------------------------------------------------------
cprocstart  rep_movsd

        ARG     dst:DPTR, src:DPTR, count:UINT

        enter_c
        mov     edi,[dst]
        mov     esi,[src]
        mov     ecx,[count]
        rep     movsd
        mov     eax,edi
        leave_c
        ret

cprocend

;----------------------------------------------------------------------------
; void * rep_movsdb(void *dst,void *src,int count);
;----------------------------------------------------------------------------
cprocstart  rep_movsdb

        ARG     dst:DPTR, src:DPTR, count:UINT

        enter_c
        mov     edi,[dst]
        mov     esi,[src]
        mov     eax,[count]
        mov     ecx,eax
        shr     ecx,2
        rep     movsd
        mov     ecx,eax
        and     ecx,3
        jz      @@Exit
        mov     al,[esi]
        dec     ecx
        jz      @@Done
        mov     ah,[esi+1]
        dec     ecx
        jz      @@Done
        rol     eax,16
        mov     al,[esi+2]
        rol     eax,16
@@Done: mov     [edi],eax
@@Exit: mov     eax,edi
        leave_c
        ret

cprocend

;----------------------------------------------------------------------------
; void movsdb(void *dst,void *src,int count);
;----------------------------------------------------------------------------
cprocstart  movsdb

        ARG     dst:DPTR, src:DPTR, count:UINT

        enter_c
        mov     edi,[dst]
        mov     esi,[src]
        mov     ecx,[count]
        and     ecx,3
        jz      @@Exit
        mov     al,[esi]
        dec     ecx
        jz      @@Done
        mov     ah,[esi+1]
        dec     ecx
        jz      @@Done
        rol     eax,16
        mov     al,[esi+2]
        rol     eax,16
@@Done: mov     [edi],eax
@@Exit: leave_c
        ret

cprocend

;----------------------------------------------------------------------------
; void * rep_movsdw(void *dst,void *src,int count);
;----------------------------------------------------------------------------
cprocstart  rep_movsdw

        ARG     dst:DPTR, src:DPTR, count:UINT

        enter_c
        mov     edi,[dst]
        mov     esi,[src]
        mov     eax,[count]
        mov     ecx,eax
        shr     ecx,1
        rep     movsd
        mov     ecx,eax
        and     ecx,1
        rep     movsw
        mov     eax,edi
        leave_c
        ret

cprocend

;----------------------------------------------------------------------------
; void * rep_stosb(void *dst,int count,uchar value);
;----------------------------------------------------------------------------
cprocstart  rep_stosb

        ARG     dst:DPTR, count:UINT, value:UCHAR

        enter_c
        mov     edi,[dst]
        mov     ecx,[count]
        mov     al,[value]
        rep     stosb
        mov     eax,edi
        leave_c
        ret

cprocend

;----------------------------------------------------------------------------
; void * rep_stosw(void *dst,int count,ushort value);
;----------------------------------------------------------------------------
cprocstart  rep_stosw

        ARG     dst:DPTR, count:UINT, value:USHORT

        enter_c
        mov     edi,[dst]
        mov     ecx,[count]
        mov     ax,[value]
        rep     stosw
        mov     eax,edi
        leave_c
        ret

cprocend

;----------------------------------------------------------------------------
; void * rep_stosd(void *dst,int count,ulong value);
;----------------------------------------------------------------------------
cprocstart  rep_stosd

        ARG     dst:DPTR, count:UINT, value:ULONG

        enter_c
        mov     edi,[dst]
        mov     ecx,[count]
        mov     eax,[value]
        rep     stosd
        mov     eax,edi
        leave_c
        ret

cprocend

;----------------------------------------------------------------------------
; void * rep_stosdb(void *dst,int count,ulong value);
;----------------------------------------------------------------------------
cprocstart  rep_stosdb

        ARG     dst:DPTR, count:UINT, value:ULONG

        enter_c
        mov     edi,[dst]
        mov     esi,[count]
        mov     eax,[value]
        mov     ecx,esi
        shr     ecx,2
        rep     stosd
        mov     ecx,esi
        and     ecx,3
        rep     stosb
        mov     eax,edi
        leave_c
        ret

cprocend

;----------------------------------------------------------------------------
; void * rep_stosdw(void *dst,int count,ulong value);
;----------------------------------------------------------------------------
cprocstart  rep_stosdw

        ARG     dst:DPTR, count:UINT, value:ULONG

        enter_c
        mov     edi,[dst]
        mov     esi,[count]
        mov     eax,[value]
        mov     ecx,esi
        shr     ecx,1
        rep     stosd
        mov     ecx,esi
        and     ecx,1
        rep     stosw
        mov     eax,edi
        leave_c
        ret

cprocend

;----------------------------------------------------------------------------
; N_fix32 FXmul(N_fix32 a,N_fix32 b);
;----------------------------------------------------------------------------
cprocstart  FXmul

        ARG     a:ULONG, b:ULONG

        enter_c
        mov     eax,[a]
        mov     edx,[b]
        imul    edx
        add     eax,8000h
        adc     edx,0
        shrd    eax,edx,16
        leave_c
        ret

cprocend

;----------------------------------------------------------------------------
; N_fix32 FXdiv(N_fix32 dividend,N_fix32 divisor);
;----------------------------------------------------------------------------
cprocstart  FXdiv

        ARG     dividend:ULONG, divisor:ULONG

        enter_c
        mov     edx,[dividend]
        mov     ebx,[divisor]
        xor     eax,eax
        shrd    eax,edx,16
        sar     edx,16
        idiv    ebx
        leave_c
        ret

cprocend

;----------------------------------------------------------------------------
; N_fix32 FXmuldiv(N_fix32 a,N_fix32 b,N_fix32 c);
;----------------------------------------------------------------------------
cprocstart  FXmuldiv

        ARG     a:ULONG, b:ULONG, c:ULONG

        enter_c
        mov     eax,[a]
        mov     ebx,[b]
        mov     ecx,[c]
        imul    ebx
        idiv    ecx
        leave_c
        ret

cprocend

;----------------------------------------------------------------------------
; N_fix32 FXoneOver(N_fix32 f);
;----------------------------------------------------------------------------
cprocstart  FXoneOver

        ARG     f:ULONG

        enter_c
        mov     eax,[f]
        xor     eax,eax
        mov     edx,1
        idiv    ebx
        leave_c
        ret

cprocend

;----------------------------------------------------------------------------
; N_fix32 FXsquare(N_fix32 f);
;----------------------------------------------------------------------------
cprocstart  FXsquare

        ARG     f:ULONG

        enter_c
        mov     eax,[f]
        imul    eax
        add     eax,8000h
        adc     edx,0
        shrd    eax,edx,16
        leave_c
        ret

cprocend

;----------------------------------------------------------------------------
; void * blast_words(void *dst,const void *src,size_t n);
;----------------------------------------------------------------------------
cprocstart  blast_words

        ARG     dst:DPTR, src:DPTR, n:UINT

        enter_c
        mov     edi,[dst]
        mov     esi,[src]
        mov     ecx,[n]
        shr     ecx,1
        rep     movsd
        rcl     ecx,1
        rep     movsw
        mov     eax,edi
        leave_c
        ret

cprocend

;----------------------------------------------------------------------------
; void  _CPUID(ulong EAX,CPUID_t *data);
;----------------------------------------------------------------------------
cprocstart  _CPUID

        ARG     v_eax:ULONG, data:DPTR

        enter_c
        mov     eax,[v_eax]
        mov     esi,[data]
        CPUID
        mov     [esi],eax
        mov     [esi+4],ebx
        mov     [esi+8],edx
        mov     [esi+12],ecx
        leave_c
        ret

cprocend

;----------------------------------------------------------------------------
; ulong mkcolor4(GA_color);
;----------------------------------------------------------------------------
cprocstart  mkcolor4

        ARG     color:ULONG

        push    ebp
        mov     ebp,esp
        mov     eax,[color]
        mov     ah,al
        shl     ah,4
        or      al,ah
        or      ah,al
        movzx   edx,ax
        shl     eax,16
        or      edx,eax
        mov     eax,edx
        pop     ebp
        ret

cprocend

;----------------------------------------------------------------------------
; ulong mkcolor8(GA_color);
;----------------------------------------------------------------------------
cprocstart  mkcolor8

        ARG     color:ULONG

        push    ebp
        mov     ebp,esp
        mov     eax,[color]
        mov     ah,al
        movzx   edx,ax
        shl     eax,16
        or      edx,eax
        mov     eax,edx
        pop     ebp
        ret

cprocend

;----------------------------------------------------------------------------
; ulong mkcolor16(GA_color);
;----------------------------------------------------------------------------
cprocstart  mkcolor16

        ARG     color:ULONG

        push    ebp
        mov     ebp,esp
        mov     eax,[color]
        movzx   edx,ax
        shl     eax,16
        or      edx,eax
        mov     eax,edx
        pop     ebp
        ret

cprocend

;----------------------------------------------------------------------------
; ushort mkcolor8_w(GA_color);
;----------------------------------------------------------------------------
cprocstart  mkcolor8_w

        ARG     color:ULONG

        push    ebp
        mov     ebp,esp
        mov     eax,[color]
        mov     ah,al
        pop     ebp
        ret

cprocend

;----------------------------------------------------------------------------
; ushort swap_bytes(GA_color);
;----------------------------------------------------------------------------
cprocstart  swap_bytes

        ARG     color:ULONG

        push    ebp
        mov     ebp,esp
        mov     eax,[color]
        xchg    ah,al
        pop     ebp
        ret

cprocend

;----------------------------------------------------------------------------
; uchar rolb1(uchar *val);
;----------------------------------------------------------------------------
cprocstart  rolb1

        ARG     val:DPTR

        push    ebp
        mov     ebp,esp
        mov     ecx,[val]
        rol     [BYTE ecx],1
        setc    al
        pop     ebp
        ret

cprocend

;----------------------------------------------------------------------------
; uchar rorb1(uchar *val);
;----------------------------------------------------------------------------
cprocstart  rorb1

        ARG     val:DPTR

        push    ebp
        mov     ebp,esp
        mov     ecx,[val]
        ror     [BYTE ecx],1
        setc    al
        pop     ebp
        ret

cprocend

;----------------------------------------------------------------------------
; void  rwb(uchar *addr);
;----------------------------------------------------------------------------
cprocstart  rwb

        ARG     addr:DPTR

        push    ebp
        mov     ebp,esp
        mov     ecx,[addr]
        mov     al,[ecx]
        mov     [ecx],al
        pop     ebp
        ret

cprocend

;----------------------------------------------------------------------------
; void  rwbv(uchar *addr,uchar v);
;----------------------------------------------------------------------------
cprocstart  rwbv

        ARG     addr:DPTR, val:UCHAR

        push    ebp
        mov     ebp,esp
        mov     ecx,[addr]
        mov     ah,[val]
        mov     al,[ecx]
        mov     [ecx],ah
        pop     ebp
        ret

cprocend

endcodeseg  _ports

        END                         ; End of module
