/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to load and connect to the X86 emulator when compiled
*               into a binary portable DLL.
*
****************************************************************************/

#include "loader.h"
#include "biosemu.h"

/*---------------------------- Global Variables ---------------------------*/

BE_exports  _VARAPI _BE_exports;
static int          loaded = false;
static PE_MODULE    *hModBPD = NULL;
static ulong        hModSize;

#if defined(__DRIVER__) || defined(TEST_HARNESS)
extern PM_imports _VARAPI _PM_imports;
#else
#include "pmimp.h"
#endif

/*----------------------------- Implementation ----------------------------*/

#define DLL_NAME    "biosemu.bpd"

/****************************************************************************
REMARKS:
Fatal error handler for non-exported BE_exports.
****************************************************************************/
static void _BE_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported BIOSEmu export function called!\n");
}

/****************************************************************************
PARAMETERS:
shared  - True to load the driver into shared memory.

REMARKS:
Loads the SNAP binary portable DLL into memory and initilises it.
****************************************************************************/
static ibool LoadDriver(void)
{
    BE_initLibrary_t    BE_initLibrary;
    BE_exports          *beExp;
    int                 i,max;
    ulong               *p;

    /* Check if we have already loaded the driver */
    if (loaded)
        return true;
    if ((hModBPD = GA_loadLibrary(DLL_NAME,&hModSize,false)) == NULL)
        return false;
    if ((BE_initLibrary = (BE_initLibrary_t)PE_getProcAddress(hModBPD,"_BE_initLibrary")) == NULL)
        return false;
    if ((beExp = BE_initLibrary(&_PM_imports)) == NULL)
        PM_fatalError("BE_initLibrary failed!\n");

    /* Initialize all default imports to point to fatal error handler
     * for upwards compatibility, and copy the exported functions.
     */
    max = sizeof(_BE_exports)/sizeof(BE_initLibrary_t);
    for (i = 0,p = (ulong*)&_BE_exports; i < max; i++)
        *p++ = (ulong)_BE_fatalErrorHandler;
    memcpy(&_BE_exports,beExp,MIN(sizeof(_BE_exports),beExp->dwSize));
    loaded = true;
    return true;
}

/* The following are stub entry points that the application calls to
 * initialise the SNAP loader library, and we use this to load our
 * driver DLL from disk and initialise the library using it.
 * {secret}
 */
ibool PMAPI BE_init(
    u32 debugFlags,
    int memSize,
    BE_VGAInfo *info)
{
    if (LoadDriver())
        return _BE_exports.BE_init(debugFlags,memSize,info);
    return false;
}

/* {secret} */
void PMAPI BE_exit(void)
{
    if (loaded) {
        _BE_exports.BE_exit();
        PE_freeLibrary(hModBPD);
        loaded = false;
        }
}

