/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  C support routines for 2D drawing functions.
*
****************************************************************************/

#include "internal.h"
#include "accel.h"

/*--------------------------- Global variables ----------------------------*/

static int              TDFX_fifoEntries;
static GA_buffer        TDFX_drawBuf;
static ulong            TDFX_format;
static ulong            TDFX_foreColor;
static ulong            TDFX_backColor;
static ulong            TDFX_srcMix;
static ulong            TDFX_pattMix;
static ulong            TDFX_bytesPerPixel;
static ulong            TDFX_prevY;
static GA_pattern       TDFX_monoPatt[8];
static ulong            TDFX_cntMonoPatt0;
static ulong            TDFX_cntMonoPatt1;
static ulong            TDFX_monoPattCmd;
static ulong            TDFX_monoPattBltCmd;
static GA_colorPattern  TDFX_colorPatt[8];
static ulong            TDFX_cntColorPatt0;
static ulong            TDFX_cntColorPatt1;
static GA_stipple       TDFX_lineStipple;
static ulong            TDFX_lineStippleCount;

// Offscreen area for blitting scanlines of data from system memory
static uchar            *TDFX_blitBufAddr1;
static ulong            TDFX_blitBufOffset1;
static uchar            *TDFX_blitBufAddr2;
static ulong            TDFX_blitBufOffset2;

static ulong mixTranslate[] = {
    ROP3_BLACK              ,
    ROP3_SRC_NOTMERGESRC    ,
    ROP3_SRC_MASKNOTSRC     ,
    ROP3_SRC_NOTCOPYSRC     ,
    ROP3_SRC_MASKSRCNOT     ,
    ROP3_NOT                ,
    ROP3_SRC_XORSRC         ,       // GA_XOR_MIX
    ROP3_SRC_NOTMASKSRC     ,
    ROP3_SRC_MASKSRC        ,       // GA_AND_MIX
    ROP3_SRC_NOTXORSRC      ,
    ROP3_NOP                ,       // GA_NOP_MIX
    ROP3_SRC_MERGENOTSRC    ,
    ROP3_SRC_COPYSRC        ,       // GA_REPLACE_MIX
    ROP3_SRC_MERGESRCNOT    ,
    ROP3_SRC_MERGESRC       ,       // GA_OR_MIX
    ROP3_WHITE              ,
    };

static ulong pattMixTranslate[] = {
    ROP3_BLACK              ,
    ROP3_PATT_NOTMERGESRC   ,
    ROP3_PATT_MASKNOTSRC    ,
    ROP3_PATT_NOTCOPYSRC    ,
    ROP3_PATT_MASKSRCNOT    ,
    ROP3_NOT                ,
    ROP3_PATT_XORSRC        ,     // GA_XOR_MIX
    ROP3_PATT_NOTMASKSRC    ,
    ROP3_PATT_MASKSRC       ,     // GA_AND_MIX
    ROP3_PATT_NOTXORSRC     ,
    ROP3_NOP                ,     // GA_NOP_MIX
    ROP3_PATT_MERGENOTSRC   ,
    ROP3_PATT_COPYSRC       ,     // GA_REPLACE_MIX
    ROP3_PATT_MERGESRCNOT   ,
    ROP3_PATT_MERGESRC      ,     // GA_OR_MIX
    ROP3_WHITE              ,
    };

/*----------------------------- Implementation ----------------------------*/

#include "../ellipse.c"

void ACCEL2D_init(void)
{
    //-----------------------------------------------------------------------
    // Setup hardware cursor management functions
    //-----------------------------------------------------------------------
    DV.cursor.SetMonoCursor                 = TDFX_SetMonoCursor;
    DV.cursor.SetMonoCursorColor            = TDFX_SetMonoCursorColor;
    DV.cursor.SetColorCursor                = CNOT_SUPPORTED;
    DV.cursor.SetCursorPos                  = TDFX_SetCursorPos;
    DV.cursor.ShowCursor                    = TDFX_ShowCursor;

    //-----------------------------------------------------------------------
    // Setup 2D state management functions
    //-----------------------------------------------------------------------
    DV.state2d.SetDrawBuffer                = TDFX_SetDrawBuffer;
    DV.state2d.IsIdle                       = TDFX_IsIdle;
    DV.state2d.WaitTillIdle                 = TDFX_WaitTillIdle;
    DV.state2d.SetMix                       = TDFX_SetMix;
    DV.state2d.SetForeColor                 = TDFX_SetForeColor;
    DV.state2d.SetBackColor                 = TDFX_SetBackColor;
    DV.state2d.Set8x8MonoPattern            = TDFX_Set8x8MonoPattern;
    DV.state2d.Use8x8MonoPattern            = TDFX_Use8x8MonoPattern;
    DV.state2d.Use8x8TransMonoPattern       = TDFX_Use8x8TransMonoPattern;
    DV.state2d.Set8x8ColorPattern           = TDFX_Set8x8ColorPattern;
    DV.state2d.Use8x8ColorPattern           = TDFX_Use8x8ColorPattern;
    DV.state2d.Use8x8TransColorPattern      = CNOT_SUPPORTED;
    DV.state2d.SetLineStipple               = TDFX_SetLineStipple;
    DV.state2d.SetLineStippleCount          = TDFX_SetLineStippleCount;
    DV.state2d.SetPlaneMask                 = CNOT_SUPPORTED;
    DV.state2d.SetupForWindows              = TDFX_SetupForWindows;

    //-----------------------------------------------------------------------
    // Setup 2D drawing functions
    //-----------------------------------------------------------------------
    Set2DDraw(TDFX,DrawRect);
    Set2DDraw(TDFX,DrawRectExt);
    Set2DDraw(TDFX,DrawRectLin);
    Set2DDraw(TDFX,DrawPattRect);
    Set2DDraw(TDFX,DrawColorPattRect);
    Set2DDraw(TDFX,DrawScanList);
    Set2DDraw(TDFX,DrawPattScanList);
    Set2DDraw(TDFX,DrawColorPattScanList);
    Set2DDraw(TDFX,DrawTrap);
    Set2DDraw(TDFX,DrawPattTrap);
    Set2DDraw(TDFX,DrawColorPattTrap);
    Set2DDraw(TDFX,DrawLineInt);
    NeedSpecs2DDraw(DrawBresenhamLine);
    Set2DDraw(TDFX,DrawStippleLineInt);
    NeedSpecs2DDraw(DrawBresenhamStippleLine);
    Set2DDraw(TDFX,PutMonoImageMSBSys);
    Set2DDraw(TDFX,ClipMonoImageMSBSys);
    Set2DDraw(TDFX,BitBlt);
    Set2DDraw(TDFX,BitBltLin);
    Set2DDraw(TDFX,BitBltSys);
    Set2DDraw(TDFX,SrcTransBlt);
    Set2DDraw(TDFX,SrcTransBltLin);
    Set2DDraw(TDFX,SrcTransBltSys);
    Set2DDraw(TDFX,DstTransBlt);
    Set2DDraw(TDFX,DstTransBltLin);
    Set2DDraw(TDFX,DstTransBltSys);
    Set2DDraw(TDFX,BitBltPatt);
    Set2DDraw(TDFX,BitBltPattLin);
    Set2DDraw(TDFX,BitBltPattSys);
    Set2DDraw(TDFX,BitBltColorPatt);
    Set2DDraw(TDFX,BitBltColorPattLin);
    Set2DDraw(TDFX,BitBltColorPattSys);
    Set2DDraw(DRV,DrawEllipseList);
    Set2DDraw(DRV,DrawPattEllipseList);
    Set2DDraw(DRV,DrawColorPattEllipseList);
    Set2DDraw(DRV,DrawFatEllipseList);
    Set2DDraw(DRV,DrawPattFatEllipseList);
    Set2DDraw(DRV,DrawColorPattFatEllipseList);

    // TODO: Bresenham lines do not work properly at the moment.
}

//---------------------------------------------------------------------------
// TDFX_SetupAccelerator
//---------------------------------------------------------------------------
// Initialises the graphics accelerator after a mode set ready for operation.
//---------------------------------------------------------------------------
void TDFX_setupAccelerator(void)
{
    // Setup the screen format values
    switch (DV.cntMode.bits) {
        case 8:
            TDFX_bytesPerPixel = 1;
            TDFX_format = TDFX_FORMAT_8BPP;
            break;
        case 15:
        case 16:
            TDFX_bytesPerPixel = 2;
            TDFX_format = TDFX_FORMAT_16BPP;
            break;
        case 24:
            TDFX_format = TDFX_FORMAT_24BPP;
            TDFX_bytesPerPixel = 3;
            break;
        case 32:
            TDFX_format = TDFX_FORMAT_32BPP;
            TDFX_bytesPerPixel = 4;
            break;
        }

    // Setup of offscreen memory areas
    TDFX_blitBufOffset1 = DV.cntModeInfo.MaxScanLines * DV.cntMode.bytesPerLine;
    TDFX_blitBufAddr1   = (uchar*)(ulong)DC.LinearMem + TDFX_blitBufOffset1;
    TDFX_blitBufOffset2 = (DV.cntModeInfo.MaxScanLines+1) * DV.cntMode.bytesPerLine;
    TDFX_blitBufAddr2   = (uchar*)(ulong)DC.LinearMem + TDFX_blitBufOffset2;

    // Setup default non-variant register values
    TDFX_SetupForWindows();
}

/****************************************************************************
REMARKS:
Sets up the default register values we require for rendering in windows
on the desktop when IHV drivers are running.
****************************************************************************/
void NAPI TDFX_SetupForWindows(void)
{
    MMIO_DECLARE;

    // Setup initial values for accelerator registers
    CLRCMD();
    WAITQ_TDFX(5);
    outmd(TDFX_CLIP0_MIN,0);
    outmd(TDFX_CLIP0_MAX,0xFFFFFFFFUL);
    outmd(TDFX_CLIP1_MIN,0);
    outmd(TDFX_CLIP1_MAX,0xFFFFFFFFUL);
    outmd(TDFX_COMMANDEXTRA,0); // disable color keying
}

/****************************************************************************
PARAMETERS:
drawBuf - Buffer to make the active drawing buffer.

RETURNS:
0 on success, -1 on failure
****************************************************************************/
N_int32 NAPI TDFX_SetDrawBuffer(
    GA_buffer *drawBuf)
{
    MMIO_DECLARE;

    TDFX_drawBuf = *drawBuf;
    WAITQ_TDFX(4);
    outmd(TDFX_DSTBASEADDR, TDFX_drawBuf.Offset);
    outmd(TDFX_DSTFORMAT,TDFX_format | TDFX_drawBuf.Stride);
    outmd(TDFX_SRCBASEADDR, TDFX_drawBuf.Offset);
    outmd(TDFX_SRCFORMAT,TDFX_format | TDFX_drawBuf.Stride);
    return 0;
}

/****************************************************************************
REMARKS:
This function is the same as the WaitTillIdle function, however it does
not wait for the engine to become idle but instead returns immediately
with the current status.
****************************************************************************/
N_int32 NAPI TDFX_IsIdle(void)
{
    MMIO_DECLARE;
    return !(inmb(TDFX_STATUS) & 0x80);
}

/****************************************************************************
REMARKS:
This function waits until the hardware accelerator has completed all
currently queued rendering operations. The primary purpose of this function
is to provide the application with the ability to ensure all rendering is
complete, before swapping display pages when doing double buffering, or
before directly accessing the framebuffer memory.
****************************************************************************/
void NAPI TDFX_WaitTillIdle(void)
{
    MMIO_DECLARE;
    CLRCMD();
}

/****************************************************************************
PARAMETERS:
mix     - New mix code to set (GA_mixCodesType)

REMARKS:

This function sets the hardware mix operation for subsequent accelerated
rendering primitives.
****************************************************************************/
N_int32 NAPI TDFX_SetMix(
    N_int32 mix)
{
    TDFX_srcMix = LOADMIX(mix) << ROP0_OFFSET_SHIFT;
    TDFX_pattMix = LOADPATTMIX(mix) << ROP0_OFFSET_SHIFT;
    return true;
}

/****************************************************************************
PARAMETERS:
color   - New foreground color to set

REMARKS:
****************************************************************************/
void NAPI TDFX_SetForeColor(
    GA_color color)
{
    MMIO_DECLARE;

    WAITQ_TDFX(1);
    outmd(TDFX_FORECOLOR, TDFX_foreColor = color);
}

/****************************************************************************
PARAMETERS:
color   - New background color to set

REMARKS:
****************************************************************************/
void NAPI TDFX_SetBackColor(
    GA_color color)
{
    MMIO_DECLARE;

    WAITQ_TDFX(1);
    outmd(TDFX_BACKCOLOR, TDFX_backColor = color);
}

/****************************************************************************
PARAMETERS:
index   - Index of the pattern to download
pattern - Pointer to pattern data to download

REMARKS:
This downloads one of 8, 8x8 monochrome patterns for all subsequent monchrome
pattern filled functions. Offscreen memory holds all 8 patterns.

****************************************************************************/
void NAPI TDFX_Set8x8MonoPattern(
    N_int32 index,
    GA_pattern *pattern)
{
    TDFX_monoPatt[index] = *pattern;
}

/****************************************************************************
PARAMETERS:
index   - Index of the pattern to download
pattern - Pointer to pattern data to download

REMARKS:
This chooses one of 8, 8x8 monochrome patterns for all subsequent monchrome
pattern filled functions.
****************************************************************************/
void NAPI TDFX_Use8x8MonoPattern(
    N_int32 index)
{
    TDFX_cntMonoPatt0 = *(((ulong*)&TDFX_monoPatt[index])+0);
    TDFX_cntMonoPatt1 = *(((ulong*)&TDFX_monoPatt[index])+1);
    TDFX_monoPattCmd = TDFX_CMD_RECTFILL | TDFX_CMD_MONOPATT_EN;
    TDFX_monoPattBltCmd = TDFX_CMD_BLT | TDFX_CMD_MONOPATT_EN;
}

/****************************************************************************
PARAMETERS:
index   - Index of the pattern to download
pattern - Pointer to pattern data to download

REMARKS:
This downloads one of 8, 8x8 monochrome patterns for all subsequent monchrome
pattern filled functions.
****************************************************************************/
void NAPI TDFX_Use8x8TransMonoPattern(
    N_int32 index)
{
    TDFX_cntMonoPatt0 = *(((ulong*)&TDFX_monoPatt[index])+0);
    TDFX_cntMonoPatt1 = *(((ulong*)&TDFX_monoPatt[index])+1);
    TDFX_monoPattCmd = TDFX_CMD_RECTFILL | TDFX_CMD_MONOPATT_EN | TDFX_CMD_MONOTRANS_EN;
    TDFX_monoPattBltCmd = TDFX_CMD_BLT | TDFX_CMD_MONOPATT_EN | TDFX_CMD_MONOTRANS_EN;
}

/****************************************************************************
PARAMETERS:
index   - Index of the pattern to download
pattern - Pointer to pattern data to download

REMARKS:
This downloads one of 8, 8x8 color patterns for all subsequent color
pattern filled functions.

All 8 patterns are held in the local colorPattArray (system memory) and
also in offscreen video memory in a contiguous linear format.
Patterns in offscreen memory are always maintained to the proper Y
alignment because adjustColorPattern() downloads them in the adjusted
format. adjustColorPattern is also called when setDrawBuffer changes
the Y alignment.

****************************************************************************/
void NAPI TDFX_Set8x8ColorPattern(
    N_int32 index,
    GA_colorPattern *pat)
{
    TDFX_colorPatt[index] = *pat;
}

/****************************************************************************
PARAMETERS:
index   - Index of the pattern to select

REMARKS:
Selects one of 8 color patterns for subsequent rendering operations.
****************************************************************************/
void NAPI TDFX_Use8x8ColorPattern(
    N_int32 index)
{
    MMIO_DECLARE;
    uchar   *dst,*src;
    int     dwords;

    TDFX_cntColorPatt0 = *(((ulong*)&TDFX_colorPatt[index])+0);
    TDFX_cntColorPatt1 = *(((ulong*)&TDFX_colorPatt[index])+1);
    switch (DV.cntMode.bits) {
        case 8:
            dwords = 16;
            break;
        case 15:
        case 16:
            dwords = 32;
            break;
        case 24:
        case 32:
            dwords = 48;
            break;
        }
    dst = (uchar*)mmioBase+TDFX_PATTERN_BASE;
    src = (uchar*)&TDFX_colorPatt[index];
    while (dwords >= 0) {
        WAITQ_TDFX(16);
        rep_movsd(dst,src,16);
        dst += 16 * 4;
        src += 16 * 4;
        dwords -= 16;
        }
}

/****************************************************************************
PARAMETERS:
stipple - New 16-bit line stipple pattern to set

REMARKS:
This function sets up a 16-bit line stipple for all subsequent stippled line
drawing functions. In the stipple, pixel 0 corresponds to bit 0, pixel 1 =
bit 1, ... pixel 15 = bit 15 etc. If the stippled line is drawn in
transparent mode, where a bit is 0 in the stipple pattern the destination
pixel remains untouched. If the stipple line is drawin in opaque mode,
where a bit is 0 in the stipple pattern the destination pixel is drawn in
the background color. In all cases where a bit in the stipple pattern is 1,
the pixel is drawn in the foreground color.
****************************************************************************/
void NAPI TDFX_SetLineStipple(
    GA_stipple stipple)
{
    MMIO_DECLARE;

    TDFX_lineStipple = stipple;
    TDFX_lineStippleCount = 0;
    WAITQ_TDFX(1);
    outmd(TDFX_LINESTIPPLE,stipple);
}

/****************************************************************************
PARAMETERS:
count   - New 32-bit line stipple count

REMARKS:
This function sets the line stipple count for stippled line drawing. When
a line is drawn, every pixel that is drawn in the line increments the
stipple counter by 1. The stipple counter modules 16 is used to determine
which bit in the stipple pattern should be used for the next pixel drawn
in the line, and because the driver maintains this stipple count across
stippled line drawing functions, it allows a single stipple pattern to
be correctly applied to a number of connected line segments. This function
allows the user application to preset the stipple count to a specified
value before drawing the next stippled line.
****************************************************************************/
void NAPI TDFX_SetLineStippleCount(
    N_uint32 count)
{
    TDFX_lineStippleCount = count;
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the rectangle to draw
top     - Top coordinate of the rectangle to draw
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines

REMARKS:
This function renders a solid rectangle at the specified location and
currently active color and mix.
****************************************************************************/
void NAPI TDFX_DrawRect(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height)
{
    MMIO_DECLARE;

    WAITQ_TDFX(3);
    outmd(TDFX_COMMAND, TDFX_CMD_RECTFILL | TDFX_srcMix);
    outmd(TDFX_DSTSIZE,(height << 16) | width);
    outmd(TDFX_LAUNCH,(top << 16) | left);
    TDFX_prevY = top;
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the rectangle to draw
top     - Top coordinate of the rectangle to draw
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines

REMARKS:
This function renders a solid rectangle at the specified location and
currently active color and mix.
****************************************************************************/
void NAPI TDFX_DrawRectExt(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    GA_color color,
    N_int32 mix)
{
    MMIO_DECLARE;

    mix = LOADMIX(mix) << ROP0_OFFSET_SHIFT;
    WAITQ_TDFX(4);
    outmd(TDFX_FORECOLOR, color);
    outmd(TDFX_COMMAND, TDFX_CMD_RECTFILL | mix);
    outmd(TDFX_DSTSIZE,(height << 16) | width);
    outmd(TDFX_LAUNCH,(top << 16) | left);

    // Reset defaults
    WAITQ_TDFX(1);
    outmd(TDFX_FORECOLOR, TDFX_foreColor);
    TDFX_prevY = top;
}

/****************************************************************************
PARAMETERS:
dstOfs      - Offset of destination rectangle in video memory
dstPitch    - Pitch of destination rectangle in bytes
left        - Left coordinate of the rectangle to draw
top         - Top coordinate of the rectangle to draw
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines

REMARKS:
This function is identical to the DrawRect function, except that it also
takes a destination linear offset and pitch.
****************************************************************************/
void NAPI TDFX_DrawRectLin(
    N_int32 dstOfs,
    N_int32 dstPitch,
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    GA_color color,
    N_int32 mix)
{
    MMIO_DECLARE;

    mix = LOADMIX(mix) << ROP0_OFFSET_SHIFT;
    WAITQ_TDFX(6);
    outmd(TDFX_FORECOLOR, color);
    outmd(TDFX_DSTBASEADDR, dstOfs);
    outmd(TDFX_DSTFORMAT,TDFX_format | dstPitch);
    outmd(TDFX_COMMAND, TDFX_CMD_RECTFILL | mix);
    outmd(TDFX_DSTSIZE,(height << 16) | width);
    outmd(TDFX_LAUNCH,(top << 16) | left);

    // Reset defaults
    WAITQ_TDFX(3);
    outmd(TDFX_FORECOLOR, TDFX_foreColor);
    outmd(TDFX_DSTBASEADDR, TDFX_drawBuf.Offset);
    outmd(TDFX_DSTFORMAT,TDFX_format | TDFX_drawBuf.Stride);
    TDFX_prevY = top;
}

/****************************************************************************
PARAMETERS:
left        - Left coordinate of the rectangle to draw
top         - Top coordinate of the rectangle to draw
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines

REMARKS:
This function renders a monochrome patterned rectangle at the specified
location and in the currently active colors, mix and monochrome pattern.
****************************************************************************/
void NAPI TDFX_DrawPattRect(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height)
{
    MMIO_DECLARE;

    WAITQ_TDFX(5);
    outmd(TDFX_PATTERN0, TDFX_cntMonoPatt0);
    outmd(TDFX_PATTERN1, TDFX_cntMonoPatt1);
    outmd(TDFX_COMMAND, TDFX_monoPattCmd | TDFX_pattMix);
    outmd(TDFX_DSTSIZE,(height << 16) | width);
    outmd(TDFX_LAUNCH,(top << 16) | left);
    TDFX_prevY = top;
}

/****************************************************************************
PARAMETERS:
left        - Left coordinate of the rectangle to draw
top         - Top coordinate of the rectangle to draw
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines

REMARKS:
This function renders a color patterned rectangle at the specified
location in the currently active mix and color pattern.
****************************************************************************/
void NAPI TDFX_DrawColorPattRect(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height)
{
    MMIO_DECLARE;

    WAITQ_TDFX(5);
    outmd(TDFX_PATTERN0, TDFX_cntColorPatt0);
    outmd(TDFX_PATTERN1, TDFX_cntColorPatt1);
    outmd(TDFX_COMMAND, TDFX_CMD_RECTFILL | TDFX_pattMix);
    outmd(TDFX_DSTSIZE,(height << 16) | width);
    outmd(TDFX_LAUNCH,(top << 16) | left);
    TDFX_prevY = top;
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
scans   - Pointer to an array of scanline data

REMARKS:
This function renders a list of solid scanlines starting at the specified
location in the currently active color and mix.
****************************************************************************/
void NAPI TDFX_DrawScanList(
    N_int32 y,
    N_int32 length,
    N_int16 *scans)
{
    MMIO_DECLARE;

    WAITQ_TDFX(1);
    outmd(TDFX_COMMAND, TDFX_CMD_RECTFILL | TDFX_srcMix);
    while (length--) {
        int width;
        int x1 = *scans++;
        int x2 = *scans++;
        if ((width = (x2 - x1)) != 0) {
            if (width < 0) {
                x1 = x2;
                width = -width;
                }
            WAITQ_TDFX(2);
            outmd(TDFX_DSTSIZE,0x10000 | width);
            outmd(TDFX_LAUNCH,(y << 16) | x1);
            }
        y++;
        }
    TDFX_prevY = y-1;
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
scans   - Pointer to an array of scanline data

REMARKS:
This function renders a list of solid scanlines starting at the specified
location in the currently active color and mix.
****************************************************************************/
void NAPI TDFX_DrawPattScanList(
    N_int32 y,
    N_int32 length,
    N_int16 *scans)
{
    MMIO_DECLARE;

    WAITQ_TDFX(3);
    outmd(TDFX_PATTERN0, TDFX_cntMonoPatt0);
    outmd(TDFX_PATTERN1, TDFX_cntMonoPatt1);
    outmd(TDFX_COMMAND, TDFX_monoPattCmd | TDFX_pattMix);
    while (length--) {
        int width;
        int x1 = *scans++;
        int x2 = *scans++;
        if ((width = (x2 - x1)) != 0) {
            if (width < 0) {
                x1 = x2;
                width = -width;
                }
            WAITQ_TDFX(2);
            outmd(TDFX_DSTSIZE,0x10000 | width);
            outmd(TDFX_LAUNCH,(y << 16) | x1);
            }
        y++;
        }
    TDFX_prevY = y-1;
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
scans   - Pointer to an array of scanline data

REMARKS:
This function renders a list of solid scanlines starting at the specified
location in the currently active color and mix.
****************************************************************************/
void NAPI TDFX_DrawColorPattScanList(
    N_int32 y,
    N_int32 length,
    N_int16 *scans)
{
    MMIO_DECLARE;

    WAITQ_TDFX(3);
    outmd(TDFX_PATTERN0, TDFX_cntColorPatt0);
    outmd(TDFX_PATTERN1, TDFX_cntColorPatt1);
    outmd(TDFX_COMMAND, TDFX_CMD_RECTFILL | TDFX_pattMix);
    while (length--) {
        int width;
        int x1 = *scans++;
        int x2 = *scans++;
        if ((width = (x2 - x1)) != 0) {
            if (width < 0) {
                x1 = x2;
                width = -width;
                }
            WAITQ_TDFX(2);
            outmd(TDFX_DSTSIZE,0x10000 | width);
            outmd(TDFX_LAUNCH,(y << 16) | x1);
            }
        y++;
        }
    TDFX_prevY = y-1;
}

/****************************************************************************
PARAMETERS:
trap    - Pointer to the GA_trap structure describing the trapezoid

REMARKS:
This function renders a solid, flat topped and bottomed trapezoid in the
currently active color and mix.
****************************************************************************/
void NAPI TDFX_DrawTrap(
    GA_trap *trap)
{
    MMIO_DECLARE;
    int y,x1,x2,count = trap->count;

    if (!count)
        return;
    y = trap->y;
    x1 = trap->x1;
    x2 = trap->x2;

    WAITQ_TDFX(1);
    outmd(TDFX_COMMAND, TDFX_CMD_RECTFILL | TDFX_srcMix);
    while (count--) {
        int width;
        int ix1 = (x1 + 0x8000) >> 16;
        int ix2 = (x2 + 0x8000) >> 16;
        if ((width = (ix2 - ix1)) != 0) {
            if (width < 0) {
                ix1 = ix2;
                width = -width;
                }
            WAITQ_TDFX(2);
            outmd(TDFX_DSTSIZE,0x10000 | width);
            outmd(TDFX_LAUNCH,(y << 16) | ix1);
            }
        x1 += trap->slope1;
        x2 += trap->slope2;
        y++;
        }
    trap->x1 = x1;
    trap->x2 = x2;
    trap->y = y;
    TDFX_prevY = y-1;
}

/****************************************************************************
PARAMETERS:
trap    - Pointer to the GA_trap structure describing the trapezoid

REMARKS:
This function renders a solid, flat topped and bottomed trapezoid in the
currently active color and mix.
****************************************************************************/
void NAPI TDFX_DrawPattTrap(
    GA_trap *trap)
{
    MMIO_DECLARE;
    int y,x1,x2,count = trap->count;

    if (!count)
        return;
    y = trap->y;
    x1 = trap->x1;
    x2 = trap->x2;

    WAITQ_TDFX(3);
    outmd(TDFX_PATTERN0, TDFX_cntMonoPatt0);
    outmd(TDFX_PATTERN1, TDFX_cntMonoPatt1);
    outmd(TDFX_COMMAND, TDFX_monoPattCmd | TDFX_pattMix);
    while (count--) {
        int width;
        int ix1 = (x1 + 0x8000) >> 16;
        int ix2 = (x2 + 0x8000) >> 16;
        if ((width = (ix2 - ix1)) != 0) {
            if (width < 0) {
                ix1 = ix2;
                width = -width;
                }
            WAITQ_TDFX(2);
            outmd(TDFX_DSTSIZE,0x10000 | width);
            outmd(TDFX_LAUNCH,(y << 16) | ix1);
            }
        x1 += trap->slope1;
        x2 += trap->slope2;
        y++;
        }
    trap->x1 = x1;
    trap->x2 = x2;
    trap->y = y;
    TDFX_prevY = y-1;
}

/****************************************************************************
PARAMETERS:
trap    - Pointer to the GA_trap structure describing the trapezoid

REMARKS:
This function renders a solid, flat topped and bottomed trapezoid in the
currently active color and mix.
****************************************************************************/
void NAPI TDFX_DrawColorPattTrap(
    GA_trap *trap)
{
    MMIO_DECLARE;
    int y,x1,x2,count = trap->count;

    if (!count)
        return;
    y = trap->y;
    x1 = trap->x1;
    x2 = trap->x2;

    WAITQ_TDFX(3);
    outmd(TDFX_PATTERN0, TDFX_cntColorPatt0);
    outmd(TDFX_PATTERN1, TDFX_cntColorPatt1);
    outmd(TDFX_COMMAND, TDFX_CMD_RECTFILL | TDFX_pattMix);
    while (count--) {
        int width;
        int ix1 = (x1 + 0x8000) >> 16;
        int ix2 = (x2 + 0x8000) >> 16;
        if ((width = (ix2 - ix1)) != 0) {
            if (width < 0) {
                ix1 = ix2;
                width = -width;
                }
            WAITQ_TDFX(2);
            outmd(TDFX_DSTSIZE,0x10000 | width);
            outmd(TDFX_LAUNCH,(y << 16) | ix1);
            }
        x1 += trap->slope1;
        x2 += trap->slope2;
        y++;
        }
    trap->x1 = x1;
    trap->x2 = x2;
    trap->y = y;
    TDFX_prevY = y-1;
}

/****************************************************************************
PARAMETERS:
x1          - X1 coordinate
y1          - Y1 coordinate
x2          - X2 coordinate
y2          - Y2 coordinate
drawLast    - 1 to draw last pixel, 0 to skip it

REMARKS:
This function renders a solid line at the specified location and the
currently active color and mix. This routine will render a line from (x1,y1)
to (x2,y2) inclusive.
****************************************************************************/
void NAPI TDFX_DrawLineInt(
    N_int32 x1,
    N_int32 y1,
    N_int32 x2,
    N_int32 y2,
    N_int32 drawLast)
{
    MMIO_DECLARE;
    ulong   cmd;

    if (drawLast)
        cmd = TDFX_CMD_LINE | TDFX_srcMix;
    else
        cmd = TDFX_CMD_POLYLINE | TDFX_srcMix;
    WAITQ_TDFX(4);
    outmd(TDFX_COMMAND, cmd);
    outmd(TDFX_SRCXY,(y1 << 16) | x1);
    outmd(TDFX_BRESERROR0,0);
    outmd(TDFX_LAUNCH,(y2 << 16) | x2);
    TDFX_prevY = y2;
}

/****************************************************************************
PARAMETERS:
x1              - X1 coordinate
y1              - Y1 coordinate
initialError    - Initial error term
majorInc        - Major increment factor for error term
diagInc         - Diagonal increment factor for error term
count           - Number of pixels to draw
flags           - Flags to control drawing
****************************************************************************/
void NAPI TDFX_DrawBresenhamLine(
    N_int32 x1,
    N_int32 y1,
    N_int32 initialError,
    N_int32 majorInc,
    N_int32 diagInc,
    N_int32 count,
    N_int32 flags)
{
    MMIO_DECLARE;
    N_int32 x2,y2,deltaMajor,deltaMinor;
    ulong   cmd;

    // TODO: This does not appear to work as expected. I think there is
    //       something wrong with the code to get the second coordinates,
    //       and it appears to screw up when the doLastPel or doFirstPel
    //       flags are on.
    //       We also need to figure out how to handle doFirstPel

    // Reverse math to get major and minor deltas and (x2,y2)
    deltaMinor = (majorInc/2);
    deltaMajor = (majorInc - diagInc) / 2;
    if (flags & gaLineXMajor) {
        if (flags & gaLineXPositive)
            x2 = x1 + deltaMajor;
        else
            x2 = x1 - deltaMajor;
        if (flags & gaLineYPositive)
            y2 = y1 + deltaMinor;
        else
            y2 = y1 - deltaMinor;
        }
    else {
        if (flags & gaLineXPositive)
            x2 = x1 + deltaMinor;
        else
            x2 = x1 - deltaMinor;
        if (flags & gaLineYPositive)
            y2 = y1 + deltaMajor;
        else
            y2 = y1 - deltaMajor;
        }
    if (flags & gaLineDoLastPel)
        cmd = TDFX_CMD_LINE | TDFX_srcMix;
    else
        cmd = TDFX_CMD_POLYLINE | TDFX_srcMix;

    SCALE_BRESENHAM_ERROR(initialError,majorInc,diagInc,TDFX_BRES_BITS);

    WAITQ_TDFX(4);
    outmd(TDFX_COMMAND, cmd);
    outmd(TDFX_SRCXY,(y1 << 16) | x1);
    outmd(TDFX_BRESERROR0,initialError | 0x80000000);
    outmd(TDFX_LAUNCH,(y2 << 16) | x2);
    TDFX_prevY = y2;
}

/****************************************************************************
PARAMETERS:
x1          - X1 coordinate
y1          - Y1 coordinate
x2          - X2 coordinate
y2          - Y2 coordinate
drawLast    - 1 to draw last pixel, 0 to skip it

REMARKS:
This function renders a solid line at the specified location and the
currently active color and mix. This routine will render a line from (x1,y1)
to (x2,y2) inclusive.
****************************************************************************/
void NAPI TDFX_DrawStippleLineInt(
    N_int32 x1,
    N_int32 y1,
    N_int32 x2,
    N_int32 y2,
    N_int32 drawLast,
    N_int32 transparent)
{
    MMIO_DECLARE;
    ulong   cmd,lineStyle;
    int     absDeltaX,absDeltaY,count;

    if (drawLast)
        cmd = TDFX_CMD_LINE | TDFX_CMD_STIPPLE_LINE | TDFX_srcMix;
    else
        cmd = TDFX_CMD_POLYLINE | TDFX_CMD_STIPPLE_LINE | TDFX_srcMix;
    if (transparent)
        cmd |= TDFX_CMD_MONOTRANS_EN;
    lineStyle = (TDFX_lineStippleCount << 24) | (15 << 8) | 0;
    WAITQ_TDFX(5);
    outmd(TDFX_LINESTYLE,lineStyle);
    outmd(TDFX_COMMAND, cmd);
    outmd(TDFX_SRCXY,(y1 << 16) | x1);
    outmd(TDFX_BRESERROR0,0);
    outmd(TDFX_LAUNCH,(y2 << 16) | x2);
    TDFX_prevY = y2;

    // Cumulative length added to stipple count
    absDeltaX = ABS(x2 - x1);
    absDeltaY = ABS(y2 - y1);
    if (absDeltaY > absDeltaX)
        count = absDeltaY+1;
    else
        count = absDeltaX+1;
    if (drawLast)
        TDFX_lineStippleCount += count;
    else
        TDFX_lineStippleCount += count-1;
}

/****************************************************************************
PARAMETERS:
x1              - X1 coordinate
y1              - Y1 coordinate
initialError    - Initial error term
majorInc        - Major increment factor for error term
diagInc         - Diagonal increment factor for error term
count           - Number of pixels to draw
flags           - Flags to control drawing
****************************************************************************/
void NAPI TDFX_DrawBresenhamStippleLine(
    N_int32 x1,
    N_int32 y1,
    N_int32 initialError,
    N_int32 majorInc,
    N_int32 diagInc,
    N_int32 count,
    N_int32 flags,
    N_int32 transparent)
{
}

/****************************************************************************
PARAMETERS:
x           - Destination X coordinate to draw the bitmap at
y           - Destination Y coordinate to draw the bitmap at
width       - Width of the bitmap in pixels
height      - Height of the bitmap in pixels
byteWidth   - Width of the bitmap in bytes
image       - Pointer to the bitmap image data to draw
transparent - 1 for transparent, 0 for opaque

REMARKS:
This function copies a monochrome bitmap image from a system memory buffer
to video memory using the hardware accelerator, which is used for fast
bitmap masking and font rendering operations. The bitmap is rendered using
the currently active colors and mix.
****************************************************************************/
void NAPI TDFX_PutMonoImageMSBSys(
    N_int32 x,
    N_int32 y,
    N_int32 width,
    N_int32 height,
    N_int32 byteWidth,
    N_uint8 *image,
    N_int32 transparent)
{
    MMIO_DECLARE;
    int     Bwidth,DWwidth;
    uchar   *dst;
    ulong   cmd;

    // Setup the transfer
    Bwidth = (width + 7) >> 3;
    DWwidth = (Bwidth + 3) >> 2;
    cmd = TDFX_CMD_HOSTBLT | TDFX_srcMix;
    if (transparent)
        cmd |= TDFX_CMD_MONOTRANS_EN;
    WAITQ_TDFX(5);
    outmd(TDFX_COMMAND, cmd);
    outmd(TDFX_SRCFORMAT, DWwidth * 4);
    outmd(TDFX_DSTSIZE, (height << 16) | width);
    outmd(TDFX_SRCXY, 0);
    outmd(TDFX_DSTXY, (y << 16) | x);

    // Transfer the data
    dst = (uchar*)mmioBase + TDFX_LAUNCH;
    while (--height) {
        FIFO_MOVSD(dst,image,DWwidth,TDFX);
        image += byteWidth;
        }
    FIFO_MOVSDB(dst,image,Bwidth,TDFX);

    // Clean up and exit
    WAITQ_TDFX(1);
    outmd(TDFX_SRCFORMAT,TDFX_format | TDFX_drawBuf.Stride);
    TDFX_prevY = y;
}

/****************************************************************************
PARAMETERS:
x           - Destination X coordinate to draw the bitmap at
y           - Destination Y coordinate to draw the bitmap at
width       - Width of the bitmap in pixels
height      - Height of the bitmap in pixels
byteWidth   - Width of the bitmap in bytes
image       - Pointer to the bitmap image data to draw
transparent - 1 for transparent, 0 for opaque
clipLeft    - Left edge of clip rectangle (inclusive)
clipTop     - Top edge of clip rectangle (inclusive)
clipRight   - Right edge of clip rectangle (exclusive)
clipBottom  - Bottom edge of clip rectangle (exclusive)

REMARKS:
This function copies a monochrome bitmap image from a system memory buffer
to video memory using the hardware accelerator, which is used for fast
bitmap masking and font rendering operations. The bitmap is rendered using
the currently active colors and mix.
****************************************************************************/
void NAPI TDFX_ClipMonoImageMSBSys(
    N_int32 x,
    N_int32 y,
    N_int32 width,
    N_int32 height,
    N_int32 byteWidth,
    N_uint8 *image,
    N_int32 transparent,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
    MMIO_DECLARE;
    int     bottom,right,Bwidth,DWwidth,skipLeft;
    uchar   *dst;
    ulong   cmd;

    // Check if image needs to be clipped at all
    right = x + width;
    bottom = y + height;
    if ((bottom <= clipTop) || (y >= clipBottom) ||
        (right <= clipLeft) || (x >= clipRight))
        return;

    // Clip the top and bottom of the bitmap image from being drawn
    if (y < clipTop) {
        int skip = (clipTop - y);
        y = clipTop;
        image += skip * byteWidth;
        height -= skip;
        }
    if (bottom > clipBottom)
        height -= (bottom - clipBottom);

    // Now clip the left and right edges so that the clip rectangle lies
    // within the first and last bytes of the image.
    if (x < clipLeft) {
        int skip = (clipLeft - x) / 8;
        image += skip;
        x += skip * 8;
        width -= skip * 8;
        }
    else
        clipLeft = x;
    if (right > clipRight)
        width -= (right - clipRight);
    if (width <= 0)
        return;

    // Setup the transfer
    skipLeft = (clipLeft - x);
    Bwidth = (width + 7) >> 3;
    DWwidth = (Bwidth + 3) >> 2;
    cmd = TDFX_CMD_HOSTBLT | TDFX_srcMix;
    if (transparent)
        cmd |= TDFX_CMD_MONOTRANS_EN;
    WAITQ_TDFX(5);
    outmd(TDFX_COMMAND, cmd);
    outmd(TDFX_SRCFORMAT, DWwidth * 4);
    outmd(TDFX_DSTSIZE, (height << 16) | (width - skipLeft));
    outmd(TDFX_SRCXY, skipLeft);
    outmd(TDFX_DSTXY, (y << 16) | ((x + skipLeft) & 0xFFFF));

    // Transfer the data
    dst = (uchar*)mmioBase + TDFX_LAUNCH;
    while (--height) {
        FIFO_MOVSD(dst,image,DWwidth,TDFX);
        image += byteWidth;
        }
    FIFO_MOVSDB(dst,image,Bwidth,TDFX);

    // Clean up and exit
    WAITQ_TDFX(1);
    outmd(TDFX_SRCFORMAT,TDFX_format | TDFX_drawBuf.Stride);
    TDFX_prevY = y;
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the source rectangle to copy
top     - Top coordinate of the source rectangle to copy
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
dstLeft - Left coordinate of destination
dstTop  - Top coordinate of destination
mix     - Mix code for the copy (GA_mixCodesType)

REMARKS:
This function copies a rectangular region of video memory from one location
to another.
****************************************************************************/
void NAPI TDFX_BitBlt(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix)
{
    MMIO_DECLARE;
    ulong cmd;

    // Compute buffer start addresses
    cmd = TDFX_CMD_BLT | (LOADMIX(mix) << ROP0_OFFSET_SHIFT);
    if (dstLeft > left) {
        left += width-1;
        dstLeft += width-1;
        cmd |= TDFX_CMD_BLT_RIGHT_TO_LEFT;
        }
    if (dstTop > top) {
        top += height-1;
        dstTop += height-1;
        cmd |= TDFX_CMD_BLT_BOTTOM_TO_TOP;
        }

    // This appears to be necessary to avoid problems in the hardware
    TDFX_BLT_BUGFIX(top);

    // Perform the blt
    WAITQ_TDFX(4);
    outmd(TDFX_COMMAND, cmd);
    outmd(TDFX_DSTSIZE,(height << 16) | width);
    outmd(TDFX_DSTXY,(dstTop << 16) | dstLeft);
    outmd(TDFX_LAUNCH,(top << 16) | left);
    TDFX_prevY = dstTop;
}

/****************************************************************************
PARAMETERS:
srcOfs      - Offset of source bitmap in video memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)

REMARKS:
This routine will copy a linear region of video memory from srcOfs from the
start of video memory to the destination rectangle (dstLeft, dstTop,
dstLeft+width-1, dstTop+height-1) with the specified mix.
****************************************************************************/
void NAPI TDFX_BitBltLin(
    N_int32 srcOfs,
    N_int32 srcPitch,
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix)
{
    MMIO_DECLARE;

    // This appears to be necessary to avoid problems in the hardware
    TDFX_BLT_BUGFIX(top);

    // Perform the blt
    mix = LOADMIX(mix) << ROP0_OFFSET_SHIFT;
    WAITQ_TDFX(6);
    outmd(TDFX_SRCBASEADDR, srcOfs);
    outmd(TDFX_SRCFORMAT,TDFX_format | srcPitch);
    outmd(TDFX_COMMAND, TDFX_CMD_BLT | mix);
    outmd(TDFX_DSTSIZE,(height << 16) | width);
    outmd(TDFX_DSTXY,(dstTop << 16) | dstLeft);
    outmd(TDFX_LAUNCH,(top << 16) | left);
    WAITQ_TDFX(2);
    outmd(TDFX_SRCBASEADDR, TDFX_drawBuf.Offset);
    outmd(TDFX_SRCFORMAT,TDFX_format | TDFX_drawBuf.Stride);
    TDFX_prevY = dstTop;
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
flipY       - True if the bitmap should be flipped in Y direction

REMARKS:
This routine will copy a bitmap from system memory with a starting address
of srcAddr to the destination rectangle (dstLeft, dstTop, dstLeft+width-1,
dstTop+height-1) with the specified mix. The mix code will be used tcombine the source bitmap data with the pixels in the destination bitmap.
****************************************************************************/
void NAPI TDFX_BitBltSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    N_int32 flipY)
{
    MMIO_DECLARE;
    uchar   *src,*dst;
    ulong   dstOfs;
    int     yincr,Bwidth = width * TDFX_bytesPerPixel;

    src = (uchar*)srcAddr + (srcTop * srcPitch) + (srcLeft * TDFX_bytesPerPixel);
    mix = LOADMIX(mix) << ROP0_OFFSET_SHIFT;
    if (!flipY && (height*srcPitch) < (2*DV.cntMode.bytesPerLine)) {
        // Entire image fits into 2 scanlines, so do only one linear blit
        CLRCMD_FIFO(6);
        rep_movsdb(TDFX_blitBufAddr1,src,srcPitch*height);
        outmd(TDFX_SRCBASEADDR, TDFX_blitBufOffset1);
        outmd(TDFX_SRCFORMAT,TDFX_format | srcPitch);
        outmd(TDFX_COMMAND, TDFX_CMD_BLT | mix);
        outmd(TDFX_DSTSIZE,(height << 16) | width);
        outmd(TDFX_DSTXY,(dstTop << 16) | dstLeft);
        outmd(TDFX_LAUNCH,0);
        }
    else {
        // Scanline by scanline blit
        if (flipY) {
            dstTop += (height-1);
            yincr = -1;
            }
        else
            yincr = 1;
        CLRCMD_FIFO(2);
        outmd(TDFX_COMMAND, TDFX_CMD_BLT | mix);
        outmd(TDFX_DSTSIZE,0x10000 | width);
        while (height--) {
            if (height & 1) {
                dst = TDFX_blitBufAddr1;
                dstOfs = TDFX_blitBufOffset1;
                }
            else {
                dst = TDFX_blitBufAddr2;
                dstOfs = TDFX_blitBufOffset2;
                }
            rep_movsdb(dst,src,Bwidth);
            CLRCMD_FIFO(3);
            outmd(TDFX_SRCBASEADDR, dstOfs);
            outmd(TDFX_DSTXY,(dstTop << 16) | dstLeft);
            outmd(TDFX_LAUNCH,0);
            src += srcPitch;
            dstTop += yincr;
            }
        }
    WAITQ_TDFX(2);
    outmd(TDFX_SRCBASEADDR, TDFX_drawBuf.Offset);
    outmd(TDFX_SRCFORMAT,TDFX_format | TDFX_drawBuf.Stride);
}

/****************************************************************************
PARAMETERS:
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value

REMARKS:
This function copies a rectangular region of video memory from one location
to another with source transparency.
****************************************************************************/
void NAPI TDFX_SrcTransBlt(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    MMIO_DECLARE;

    // This appears to be necessary to avoid problems in the hardware
    TDFX_BLT_BUGFIX(top);

    // Perform the blt
    mix = LOADMIX(mix) << ROP0_OFFSET_SHIFT;
    WAITQ_TDFX(8);
    outmd(TDFX_SRCCLRKEYMIN, transparent);
    outmd(TDFX_SRCCLRKEYMAX, transparent);
    outmd(TDFX_ROP, (ROP3_NOP << 8));
    outmd(TDFX_COMMANDEXTRA, TDFX_CMDEXT_SOURCETRANS);
    outmd(TDFX_COMMAND, TDFX_CMD_BLT | mix);
    outmd(TDFX_DSTSIZE,(height << 16) | width);
    outmd(TDFX_DSTXY,(dstTop << 16) | dstLeft);
    outmd(TDFX_LAUNCH,(top << 16) | left);

    WAITQ_TDFX(1);
    outmd(TDFX_COMMANDEXTRA, 0);
    TDFX_prevY = dstTop;
}

/****************************************************************************
PARAMETERS:
srcOfs      - Offset of source bitmap in video memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value

REMARKS:
This function copies a rectangular region of video memory from one location
to another with source transparency.
****************************************************************************/
void NAPI TDFX_SrcTransBltLin(
    N_int32 srcOfs,
    N_int32 srcPitch,
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    MMIO_DECLARE;

    // This appears to be necessary to avoid problems in the hardware
    TDFX_BLT_BUGFIX(top);

    // Perform the blt
    mix = LOADMIX(mix) << ROP0_OFFSET_SHIFT;
    WAITQ_TDFX(10);
    outmd(TDFX_SRCCLRKEYMIN, transparent);
    outmd(TDFX_SRCCLRKEYMAX, transparent);
    outmd(TDFX_ROP, (ROP3_NOP << 8));
    outmd(TDFX_COMMANDEXTRA, TDFX_CMDEXT_SOURCETRANS);
    outmd(TDFX_SRCBASEADDR, srcOfs);
    outmd(TDFX_SRCFORMAT,TDFX_format | srcPitch);
    outmd(TDFX_COMMAND, TDFX_CMD_BLT | mix);
    outmd(TDFX_DSTSIZE,(height << 16) | width);
    outmd(TDFX_DSTXY,(dstTop << 16) | dstLeft);
    outmd(TDFX_LAUNCH,(top << 16) | left);

    WAITQ_TDFX(3);
    outmd(TDFX_SRCBASEADDR, TDFX_drawBuf.Offset);
    outmd(TDFX_SRCFORMAT,TDFX_format | TDFX_drawBuf.Stride);
    outmd(TDFX_COMMANDEXTRA, 0);
    TDFX_prevY = dstTop;
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
flipY       - True if the bitmap should be flipped in Y direction

REMARKS:
This routine will copy a bitmap from system memory with a starting address
of srcAddr to the destination rectangle (dstLeft, dstTop, dstLeft+width-1,
dstTop+height-1) with the specified mix and with source transparency. The
mix code will be used to combine the source bitmap data with the pixels in
the destination bitmap. The transparent color passed will be used to /mask
out/ pixels in the source bitmap from being written to the destination area.
Where a pixel in the source bitmap matches the transparent color, the pixel
will not be written to the destination bitmap.
****************************************************************************/
void NAPI TDFX_SrcTransBltSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent,
    N_int32 flipY)
{
    MMIO_DECLARE;
    uchar   *src,*dst;
    ulong   dstOfs;
    int     yincr,Bwidth = width * TDFX_bytesPerPixel;

    src = (uchar*)srcAddr + (srcTop * srcPitch) + (srcLeft * TDFX_bytesPerPixel);
    mix = LOADMIX(mix) << ROP0_OFFSET_SHIFT;
    CLRCMD_FIFO(5);
    outmd(TDFX_SRCCLRKEYMIN, transparent);
    outmd(TDFX_SRCCLRKEYMAX, transparent);
    outmd(TDFX_ROP, (ROP3_NOP << 8));
    outmd(TDFX_COMMANDEXTRA, TDFX_CMDEXT_SOURCETRANS);
    outmd(TDFX_COMMAND, TDFX_CMD_BLT | mix);
    if (!flipY && (height*srcPitch) < (2*DV.cntMode.bytesPerLine)) {
        // Entire image fits into 2 scanlines, so do only one linear blit
        rep_movsdb(TDFX_blitBufAddr1,src,srcPitch*height);
        WAITQ_TDFX(5);
        outmd(TDFX_SRCBASEADDR, TDFX_blitBufOffset1);
        outmd(TDFX_SRCFORMAT,TDFX_format | srcPitch);
        outmd(TDFX_DSTSIZE,(height << 16) | width);
        outmd(TDFX_DSTXY,(dstTop << 16) | dstLeft);
        outmd(TDFX_LAUNCH,0);
        }
    else {
        // Scanline by scanline blit
        if (flipY) {
            dstTop += (height-1);
            yincr = -1;
            }
        else
            yincr = 1;
        WAITQ_TDFX(1);
        outmd(TDFX_DSTSIZE,0x10000 | width);
        while (height--) {
            if (height & 1) {
                dst = TDFX_blitBufAddr1;
                dstOfs = TDFX_blitBufOffset1;
                }
            else {
                dst = TDFX_blitBufAddr2;
                dstOfs = TDFX_blitBufOffset2;
                }
            rep_movsdb(dst,src,Bwidth);
            CLRCMD_FIFO(3);
            outmd(TDFX_SRCBASEADDR, dstOfs);
            outmd(TDFX_DSTXY,(dstTop << 16) | dstLeft);
            outmd(TDFX_LAUNCH,0);
            src += srcPitch;
            dstTop += yincr;
            }
        }
    WAITQ_TDFX(3);
    outmd(TDFX_SRCBASEADDR, TDFX_drawBuf.Offset);
    outmd(TDFX_SRCFORMAT,TDFX_format | TDFX_drawBuf.Stride);
    outmd(TDFX_COMMANDEXTRA, 0);
}

/****************************************************************************
PARAMETERS:
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value

REMARKS:
This function copies a rectangular region of video memory from one location
to another with source transparency.
****************************************************************************/
void NAPI TDFX_DstTransBlt(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    MMIO_DECLARE;

    // This appears to be necessary to avoid problems in the hardware
    TDFX_BLT_BUGFIX(top);

    // Perform the blt
    mix = LOADMIX(mix);
    WAITQ_TDFX(8);
    outmd(TDFX_DSTCLRKEYMIN, transparent);
    outmd(TDFX_DSTCLRKEYMAX, transparent);
    outmd(TDFX_ROP, mix);
    outmd(TDFX_COMMANDEXTRA, TDFX_CMDEXT_DESTTRANS);
    outmd(TDFX_COMMAND, TDFX_CMD_BLT | (ROP3_NOP << ROP0_OFFSET_SHIFT));
    outmd(TDFX_DSTSIZE,(height << 16) | width);
    outmd(TDFX_DSTXY,(dstTop << 16) | dstLeft);
    outmd(TDFX_LAUNCH,(top << 16) | left);

    WAITQ_TDFX(1);
    outmd(TDFX_COMMANDEXTRA, 0);
    TDFX_prevY = dstTop;
}

/****************************************************************************
PARAMETERS:
srcOfs      - Offset of source bitmap in video memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value

REMARKS:
This function copies a rectangular region of video memory from one location
to another with source transparency.
****************************************************************************/
void NAPI TDFX_DstTransBltLin(
    N_int32 srcOfs,
    N_int32 srcPitch,
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    MMIO_DECLARE;

    // This appears to be necessary to avoid problems in the hardware
    TDFX_BLT_BUGFIX(top);

    // Perform the blt
    mix = LOADMIX(mix);
    WAITQ_TDFX(10);
    outmd(TDFX_DSTCLRKEYMIN, transparent);
    outmd(TDFX_DSTCLRKEYMAX, transparent);
    outmd(TDFX_ROP, mix);
    outmd(TDFX_COMMANDEXTRA, TDFX_CMDEXT_DESTTRANS);    outmd(TDFX_SRCBASEADDR, srcOfs);
    outmd(TDFX_SRCFORMAT,TDFX_format | srcPitch);
    outmd(TDFX_COMMAND, TDFX_CMD_BLT | (ROP3_NOP << ROP0_OFFSET_SHIFT));
    outmd(TDFX_DSTSIZE,(height << 16) | width);
    outmd(TDFX_DSTXY,(dstTop << 16) | dstLeft);
    outmd(TDFX_LAUNCH,(top << 16) | left);

    WAITQ_TDFX(3);
    outmd(TDFX_SRCBASEADDR, TDFX_drawBuf.Offset);
    outmd(TDFX_SRCFORMAT,TDFX_format | TDFX_drawBuf.Stride);
    outmd(TDFX_COMMANDEXTRA, 0);
    TDFX_prevY = dstTop;
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
flipY       - True if the bitmap should be flipped in Y direction

REMARKS:
This routine will copy a bitmap from system memory with a starting address
of srcAddr to the destination rectangle (dstLeft, dstTop, dstLeft+width-1,
dstTop+height-1) with the specified mix and with source transparency. The
mix code will be used to combine the source bitmap data with the pixels in
the destination bitmap. The transparent color passed will be used to /mask
out/ pixels in the source bitmap from being written to the destination area.
Where a pixel in the source bitmap matches the transparent color, the pixel
will not be written to the destination bitmap.
****************************************************************************/
void NAPI TDFX_DstTransBltSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent,
    N_int32 flipY)
{
    MMIO_DECLARE;
    uchar   *src,*dst;
    ulong   dstOfs;
    int     yincr,Bwidth = width * TDFX_bytesPerPixel;

    src = (uchar*)srcAddr + (srcTop * srcPitch) + (srcLeft * TDFX_bytesPerPixel);
    mix = LOADMIX(mix);
    CLRCMD_FIFO(5);
    outmd(TDFX_DSTCLRKEYMIN, transparent);
    outmd(TDFX_DSTCLRKEYMAX, transparent);
    outmd(TDFX_ROP, mix);
    outmd(TDFX_COMMANDEXTRA, TDFX_CMDEXT_DESTTRANS);
    outmd(TDFX_COMMAND, TDFX_CMD_BLT | (ROP3_NOP << ROP0_OFFSET_SHIFT));
    if (!flipY && (height*srcPitch) < (2*DV.cntMode.bytesPerLine)) {
        // Entire image fits into 2 scanlines, so do only one linear blit
        rep_movsdb(TDFX_blitBufAddr1,src,srcPitch*height);
        WAITQ_TDFX(5);
        outmd(TDFX_SRCBASEADDR, TDFX_blitBufOffset1);
        outmd(TDFX_SRCFORMAT,TDFX_format | srcPitch);
        outmd(TDFX_DSTSIZE,(height << 16) | width);
        outmd(TDFX_DSTXY,(dstTop << 16) | dstLeft);
        outmd(TDFX_LAUNCH,0);
        }
    else {
        // Scanline by scanline blit
        if (flipY) {
            dstTop += (height-1);
            yincr = -1;
            }
        else
            yincr = 1;
        WAITQ_TDFX(1);
        outmd(TDFX_DSTSIZE,0x10000 | width);
        while (height--) {
            if (height & 1) {
                dst = TDFX_blitBufAddr1;
                dstOfs = TDFX_blitBufOffset1;
                }
            else {
                dst = TDFX_blitBufAddr2;
                dstOfs = TDFX_blitBufOffset2;
                }
            rep_movsdb(dst,src,Bwidth);
            CLRCMD_FIFO(3);
            outmd(TDFX_SRCBASEADDR, dstOfs);
            outmd(TDFX_DSTXY,(dstTop << 16) | dstLeft);
            outmd(TDFX_LAUNCH,0);
            src += srcPitch;
            dstTop += yincr;
            }
        }
    WAITQ_TDFX(3);
    outmd(TDFX_SRCBASEADDR, TDFX_drawBuf.Offset);
    outmd(TDFX_SRCFORMAT,TDFX_format | TDFX_drawBuf.Stride);
    outmd(TDFX_COMMANDEXTRA, 0);
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the source rectangle to copy
top     - Top coordinate of the source rectangle to copy
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
dstLeft - Left coordinate of destination
dstTop  - Top coordinate of destination
mix     - Mix code for the copy (GA_mixCodesType)

REMARKS:
This function copies a rectangular region of video memory from one location
to another.
****************************************************************************/
void NAPI TDFX_BitBltPatt(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    MMIO_DECLARE;
    ulong cmd;

    // Compute buffer start addresses
    cmd = TDFX_monoPattBltCmd | (rop3 << ROP0_OFFSET_SHIFT);
    if (dstLeft > left) {
        left += width-1;
        dstLeft += width-1;
        cmd |= TDFX_CMD_BLT_RIGHT_TO_LEFT;
        }
    if (dstTop > top) {
        top += height-1;
        dstTop += height-1;
        cmd |= TDFX_CMD_BLT_BOTTOM_TO_TOP;
        }

    // This appears to be necessary to avoid problems in the hardware
    TDFX_BLT_BUGFIX(top);

    // Perform the blt
    WAITQ_TDFX(6);
    outmd(TDFX_PATTERN0, TDFX_cntMonoPatt0);
    outmd(TDFX_PATTERN1, TDFX_cntMonoPatt1);
    outmd(TDFX_COMMAND, cmd);
    outmd(TDFX_DSTSIZE,(height << 16) | width);
    outmd(TDFX_DSTXY,(dstTop << 16) | dstLeft);
    outmd(TDFX_LAUNCH,(top << 16) | left);
    TDFX_prevY = dstTop;
}

/****************************************************************************
PARAMETERS:
srcOfs      - Offset of source bitmap in video memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3            - Mix code for the copy (GA_mixCodesType)

REMARKS:
This routine will copy a linear region of video memory from srcOfs from the
start of video memory to the destination rectangle (dstLeft, dstTop,
dstLeft+width-1, dstTop+height-1) with the specified rop3.
****************************************************************************/
void NAPI TDFX_BitBltPattLin(
    N_int32 srcOfs,
    N_int32 srcPitch,
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    MMIO_DECLARE;

    // This appears to be necessary to avoid problems in the hardware
    TDFX_BLT_BUGFIX(top);

    // Perform the blt
    WAITQ_TDFX(8);
    outmd(TDFX_SRCBASEADDR, srcOfs);
    outmd(TDFX_SRCFORMAT,TDFX_format | srcPitch);
    outmd(TDFX_PATTERN0, TDFX_cntMonoPatt0);
    outmd(TDFX_PATTERN1, TDFX_cntMonoPatt1);
    outmd(TDFX_COMMAND, TDFX_monoPattBltCmd | (rop3 << ROP0_OFFSET_SHIFT));
    outmd(TDFX_DSTSIZE,(height << 16) | width);
    outmd(TDFX_DSTXY,(dstTop << 16) | dstLeft);
    outmd(TDFX_LAUNCH,(top << 16) | left);
    WAITQ_TDFX(2);
    outmd(TDFX_SRCBASEADDR, TDFX_drawBuf.Offset);
    outmd(TDFX_SRCFORMAT,TDFX_format | TDFX_drawBuf.Stride);
    TDFX_prevY = dstTop;
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3            - Mix code for the copy (GA_mixCodesType)
flipY       - True if the bitmap should be flipped in Y direction

REMARKS:
This routine will copy a bitmap from system memory with a starting address
of srcAddr to the destination rectangle (dstLeft, dstTop, dstLeft+width-1,
dstTop+height-1) with the specified rop3. The rop3 code will be used to
combine the source bitmap data with the pixels in the destination bitmap.
****************************************************************************/
void NAPI TDFX_BitBltPattSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3,
    N_int32 flipY)
{
    MMIO_DECLARE;
    uchar   *src,*dst;
    ulong   dstOfs;
    int     yincr,Bwidth = width * TDFX_bytesPerPixel;

    src = (uchar*)srcAddr + (srcTop * srcPitch) + (srcLeft * TDFX_bytesPerPixel);
    if (!flipY && (height*srcPitch) < (2*DV.cntMode.bytesPerLine)) {
        // Entire image fits into 2 scanlines, so do only one linear blit
        CLRCMD_FIFO(8);
        rep_movsdb(TDFX_blitBufAddr1,src,srcPitch*height);
        outmd(TDFX_SRCBASEADDR, TDFX_blitBufOffset1);
        outmd(TDFX_SRCFORMAT,TDFX_format | srcPitch);
        outmd(TDFX_PATTERN0, TDFX_cntMonoPatt0);
        outmd(TDFX_PATTERN1, TDFX_cntMonoPatt1);
        outmd(TDFX_COMMAND, TDFX_monoPattBltCmd | (rop3 << ROP0_OFFSET_SHIFT));
        outmd(TDFX_DSTSIZE,(height << 16) | width);
        outmd(TDFX_DSTXY,(dstTop << 16) | dstLeft);
        outmd(TDFX_LAUNCH,0);
        }
    else {
        // Scanline by scanline blit
        if (flipY) {
            dstTop += (height-1);
            yincr = -1;
            }
        else
            yincr = 1;
        CLRCMD_FIFO(4);
        outmd(TDFX_PATTERN0, TDFX_cntMonoPatt0);
        outmd(TDFX_PATTERN1, TDFX_cntMonoPatt1);
        outmd(TDFX_COMMAND, TDFX_monoPattBltCmd | (rop3 << ROP0_OFFSET_SHIFT));
        outmd(TDFX_DSTSIZE,0x10000 | width);
        while (height--) {
            if (height & 1) {
                dst = TDFX_blitBufAddr1;
                dstOfs = TDFX_blitBufOffset1;
                }
            else {
                dst = TDFX_blitBufAddr2;
                dstOfs = TDFX_blitBufOffset2;
                }
            rep_movsdb(dst,src,Bwidth);
            CLRCMD_FIFO(3);
            outmd(TDFX_SRCBASEADDR, dstOfs);
            outmd(TDFX_DSTXY,(dstTop << 16) | dstLeft);
            outmd(TDFX_LAUNCH,0);
            src += srcPitch;
            dstTop += yincr;
            }
        }
    WAITQ_TDFX(2);
    outmd(TDFX_SRCBASEADDR, TDFX_drawBuf.Offset);
    outmd(TDFX_SRCFORMAT,TDFX_format | TDFX_drawBuf.Stride);
}


/****************************************************************************
PARAMETERS:
left    - Left coordinate of the source rectangle to copy
top     - Top coordinate of the source rectangle to copy
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
dstLeft - Left coordinate of destination
dstTop  - Top coordinate of destination
mix     - Mix code for the copy (GA_mixCodesType)

REMARKS:
This function copies a rectangular region of video memory from one location
to another.
****************************************************************************/
void NAPI TDFX_BitBltColorPatt(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    MMIO_DECLARE;
    ulong cmd;

    // Compute buffer start addresses
    cmd = TDFX_CMD_BLT | (rop3 << ROP0_OFFSET_SHIFT);
    if (dstLeft > left) {
        left += width-1;
        dstLeft += width-1;
        cmd |= TDFX_CMD_BLT_RIGHT_TO_LEFT;
        }
    if (dstTop > top) {
        top += height-1;
        dstTop += height-1;
        cmd |= TDFX_CMD_BLT_BOTTOM_TO_TOP;
        }

    // This appears to be necessary to avoid problems in the hardware
    TDFX_BLT_BUGFIX(top);

    // Perform the blt
    WAITQ_TDFX(6);
    outmd(TDFX_PATTERN0, TDFX_cntColorPatt0);
    outmd(TDFX_PATTERN1, TDFX_cntColorPatt1);
    outmd(TDFX_COMMAND, cmd);
    outmd(TDFX_DSTSIZE,(height << 16) | width);
    outmd(TDFX_DSTXY,(dstTop << 16) | dstLeft);
    outmd(TDFX_LAUNCH,(top << 16) | left);
    TDFX_prevY = dstTop;
}

/****************************************************************************
PARAMETERS:
srcOfs      - Offset of source bitmap in video memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3            - Mix code for the copy (GA_mixCodesType)

REMARKS:
This routine will copy a linear region of video memory from srcOfs from the
start of video memory to the destination rectangle (dstLeft, dstTop,
dstLeft+width-1, dstTop+height-1) with the specified rop3.
****************************************************************************/
void NAPI TDFX_BitBltColorPattLin(
    N_int32 srcOfs,
    N_int32 srcPitch,
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    MMIO_DECLARE;

    // This appears to be necessary to avoid problems in the hardware
    TDFX_BLT_BUGFIX(top);

    // Perform the blt
    WAITQ_TDFX(8);
    outmd(TDFX_SRCBASEADDR, srcOfs);
    outmd(TDFX_SRCFORMAT,TDFX_format | srcPitch);
    outmd(TDFX_PATTERN0, TDFX_cntColorPatt0);
    outmd(TDFX_PATTERN1, TDFX_cntColorPatt1);
    outmd(TDFX_COMMAND, TDFX_CMD_BLT | (rop3 << ROP0_OFFSET_SHIFT));
    outmd(TDFX_DSTSIZE,(height << 16) | width);
    outmd(TDFX_DSTXY,(dstTop << 16) | dstLeft);
    outmd(TDFX_LAUNCH,(top << 16) | left);
    WAITQ_TDFX(2);
    outmd(TDFX_SRCBASEADDR, TDFX_drawBuf.Offset);
    outmd(TDFX_SRCFORMAT,TDFX_format | TDFX_drawBuf.Stride);
    TDFX_prevY = dstTop;
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3            - Mix code for the copy (GA_mixCodesType)
flipY       - True if the bitmap should be flipped in Y direction

REMARKS:
This routine will copy a bitmap from system memory with a starting address
of srcAddr to the destination rectangle (dstLeft, dstTop, dstLeft+width-1,
dstTop+height-1) with the specified rop3. The rop3 code will be used to
combine the source bitmap data with the pixels in the destination bitmap.
****************************************************************************/
void NAPI TDFX_BitBltColorPattSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3,
    N_int32 flipY)
{
    MMIO_DECLARE;
    uchar   *src,*dst;
    ulong   dstOfs;
    int     yincr,Bwidth = width * TDFX_bytesPerPixel;

    src = (uchar*)srcAddr + (srcTop * srcPitch) + (srcLeft * TDFX_bytesPerPixel);
    if (!flipY && (height*srcPitch) < (2*DV.cntMode.bytesPerLine)) {
        // Entire image fits into 2 scanlines, so do only one linear blit
        CLRCMD_FIFO(8);
        rep_movsdb(TDFX_blitBufAddr1,src,srcPitch*height);
        outmd(TDFX_SRCBASEADDR, TDFX_blitBufOffset1);
        outmd(TDFX_SRCFORMAT,TDFX_format | srcPitch);
        outmd(TDFX_PATTERN0, TDFX_cntColorPatt0);
        outmd(TDFX_PATTERN1, TDFX_cntColorPatt1);
        outmd(TDFX_COMMAND, TDFX_CMD_BLT | (rop3 << ROP0_OFFSET_SHIFT));
        outmd(TDFX_DSTSIZE,(height << 16) | width);
        outmd(TDFX_DSTXY,(dstTop << 16) | dstLeft);
        outmd(TDFX_LAUNCH,0);
        }
    else {
        // Scanline by scanline blit
        if (flipY) {
            dstTop += (height-1);
            yincr = -1;
            }
        else
            yincr = 1;
        CLRCMD_FIFO(4);
        outmd(TDFX_PATTERN0, TDFX_cntColorPatt0);
        outmd(TDFX_PATTERN1, TDFX_cntColorPatt1);
        outmd(TDFX_COMMAND, TDFX_CMD_BLT | (rop3 << ROP0_OFFSET_SHIFT));
        outmd(TDFX_DSTSIZE,0x10000 | width);
        while (height--) {
            if (height & 1) {
                dst = TDFX_blitBufAddr1;
                dstOfs = TDFX_blitBufOffset1;
                }
            else {
                dst = TDFX_blitBufAddr2;
                dstOfs = TDFX_blitBufOffset2;
                }
            rep_movsdb(dst,src,Bwidth);
            CLRCMD_FIFO(3);
            outmd(TDFX_SRCBASEADDR, dstOfs);
            outmd(TDFX_DSTXY,(dstTop << 16) | dstLeft);
            outmd(TDFX_LAUNCH,0);
            src += srcPitch;
            dstTop += yincr;
            }
        }
    WAITQ_TDFX(2);
    outmd(TDFX_SRCBASEADDR, TDFX_drawBuf.Offset);
    outmd(TDFX_SRCFORMAT,TDFX_format | TDFX_drawBuf.Stride);
}

