/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Function prototypes for 3dfx register acceleration
*               functions.
*
****************************************************************************/

#ifndef __ACCEL_H
#define __ACCEL_H

// PCI baseIO registers

#define TDFX_PCIINIT0                       DV.dc.IOBase+0x4
#define TDFX_SIPMONITOR                     DV.dc.IOBase+0x8
#define TDFX_LFBMEMORYCONFIG                DV.dc.IOBase+0xC
#define TDFX_MISCINIT0                      DV.dc.IOBase+0x10
#define TDFX_MISCINIT1                      DV.dc.IOBase+0x14
#define TDFX_DRAMINIT0                      DV.dc.IOBase+0x18
#define TDFX_DRAMINIT1                      DV.dc.IOBase+0x1C
#define TDFX_AGPINIT                        DV.dc.IOBase+0x20
#define TDFX_TMUGBEINIT                     DV.dc.IOBase+0x24
#define TDFX_VGAINIT0                       DV.dc.IOBase+0x28
#define TDFX_VGAINIT1                       DV.dc.IOBase+0x2c
#define TDFX_DRAMCOMMAND                    DV.dc.IOBase+0x30
#define TDFX_DRAMDATA                       DV.dc.IOBase+0x34
#define TDFX_PLLCTRL0                       DV.dc.IOBase+0x40
#define TDFX_PLLCTRL1                       DV.dc.IOBase+0x44
#define TDFX_PLLCTRL2                       DV.dc.IOBase+0x48
#define TDFX_DACMODE                        DV.dc.IOBase+0x4c
#define TDFX_DACADDR                        DV.dc.IOBase+0x50
#define TDFX_DACDATA                        DV.dc.IOBase+0x54
#define TDFX_RGBMAXDELTA                    DV.dc.IOBase+0x58
#define TDFX_VIDPROCCFG                     DV.dc.IOBase+0x5c
#define TDFX_HWCURPATADDR                   DV.dc.IOBase+0x60
#define TDFX_HWCURLOC                       DV.dc.IOBase+0x64
#define TDFX_HWCURC0                        DV.dc.IOBase+0x68
#define TDFX_HWCURC1                        DV.dc.IOBase+0x6c
#define TDFX_VIDINFORMAT                    DV.dc.IOBase+0x70
#define TDFX_VIDINSTATUS                    DV.dc.IOBase+0x74
#define TDFX_VIDSERIALPARALLELPORT          DV.dc.IOBase+0x78
#define TDFX_VIDINXDECIMDELTAS              DV.dc.IOBase+0x7c
#define TDFX_VIDINDECIMINITERRS             DV.dc.IOBase+0x80
#define TDFX_VIDYDECIMDELTA                 DV.dc.IOBase+0x84
#define TDFX_VIDPXELBUGTHOLD                DV.dc.IOBase+0x88
#define TDFX_VIDCHROMAMIN                   DV.dc.IOBase+0x8c
#define TDFX_VIDCHROMAMAX                   DV.dc.IOBase+0x90
#define TDFX_VIDCURRENTLINE                 DV.dc.IOBase+0x94
#define TDFX_VIDSCREENSIZE                  DV.dc.IOBase+0x98
#define TDFX_VIDOVERLAYSTARTCOORDS          DV.dc.IOBase+0x9c
#define TDFX_VIDOVERLAYENDSCREENCOORDS      DV.dc.IOBase+0xa0
#define TDFX_VIDOVERLAYDUDX                 DV.dc.IOBase+0xa4
#define TDFX_VIDOVERLAYDUDXOFFSETSRCWIDTH   DV.dc.IOBase+0xa8
#define TDFX_VIDOVERLAYDVDY                 DV.dc.IOBase+0xac
#define TDFX_VIDOVERLAYDVDYOFFSET           DV.dc.IOBase+0xe0
#define TDFX_STARTADDR                      DV.dc.IOBase+0xe4
#define TDFX_STRIDE                         DV.dc.IOBase+0xe8
#define TDFX_VIDINADDR0                     DV.dc.IOBase+0xec
#define TDFX_VIDINADDR1                     DV.dc.IOBase+0xf0
#define TDFX_VIDINADDR2                     DV.dc.IOBase+0xf4
#define TDFX_VIDINSTRIDE                    DV.dc.IOBase+0xf8
#define TDFX_VIDCUROVERLAYSTARTADDR         DV.dc.IOBase+0xfc

// Register bit definitions

#define BIT(n)  (1UL<<(n))
#define SST_SGRAM_OFLOP_DEL_ADJ_SHIFT   20
#define SST_SGRAM_CLK_NODELAY           BIT(13)
#define SST_DRAM_REFRESH_EN             BIT(0)
#define SST_DRAM_REFRESH_VALUE_SHIFT    1
#define SST_DRAM_REFRESH_VALUE          (0x1FF<<SST_DRAM_REFRESH_VALUE_SHIFT)
#define SST_SGRAM_TYPE_SHIFT            27
#define SST_SGRAM_TYPE                  (0x1L<<SST_SGRAM_TYPE_SHIFT)
#define SST_SGRAM_NUM_CHIPSETS          BIT(26)
#define SST_SGRAM_TYPE_8MBIT            (0x0L<<SST_SGRAM_TYPE_SHIFT)
#define SST_SGRAM_TYPE_16MBIT           (0x1L<<SST_SGRAM_TYPE_SHIFT)
#define SST_DISABLE_2D_BLOCK_WRITE      BIT(15)
#define SST_MCTL_TYPE_SDRAM             BIT(30)
#define SST_DAC_MODE_2X                 BIT(0)
#define SST_VIDEO_2X_MODE_EN            BIT(26)
#define SST_VGA0_EXTENSIONS             BIT(6)
#define SST_WAKEUP_3C3                  1
#define SST_VGA0_WAKEUP_SELECT_SHIFT    8
#define SST_VGA0_LEGACY_DECODE_SHIFT    9
#define SST_VGA0_LEGACY_DECODE          (1 << SST_VGA0_LEGACY_DECODE_SHIFT)
#define SST_VGA0_ENABLE_DECODE          0
#define SST_ENABLE_ALT_READBACK         0
#define SST_VGA0_CONFIG_READBACK_SHIFT  10
#define SST_VIDEO_PROCESSOR_EN          BIT(0)
#define SST_CURSOR_MODE_SHIFT           1
#define SST_CURSOR_X11                  (1<<SST_CURSOR_MODE_SHIFT)
#define SST_DESKTOP_EN                  BIT(7)
#define SST_DESKTOP_PIXEL_FORMAT_SHIFT  18
#define SST_DESKTOP_CLUT_BYPASS         BIT(10)
#define SST_HALF_MODE                   BIT(4)
#define SST_CURSOR_EN                   BIT(27)
#define SST_BUSY                        BIT(9)
#define SST_RETRACE                     BIT(6)

// 2D register memory offsets

#define TDFX_STATUS                 0x100000
#define TDFX_INTCNTL                0x100004
#define TDFX_CLIP0_MIN              0x100008
#define TDFX_CLIP0_MAX              0x10000C
#define TDFX_DSTBASEADDR            0x100010
#define TDFX_DSTFORMAT              0x100014
#define TDFX_SRCCLRKEYMIN           0x100018
#define TDFX_SRCCLRKEYMAX           0x10001C
#define TDFX_DSTCLRKEYMIN           0x100020
#define TDFX_DSTCLRKEYMAX           0x100024
#define TDFX_BRESERROR0             0x100028
#define TDFX_BRESERROR1             0x10002C
#define TDFX_ROP                    0x100030
#define TDFX_SRCBASEADDR            0x100034
#define TDFX_COMMANDEXTRA           0x100038
#define TDFX_LINESTIPPLE            0x10003C
#define TDFX_LINESTYLE              0x100040
#define TDFX_PATT0ALIAS             0x100044
#define TDFX_PATT1ALIAS             0x100048
#define TDFX_CLIP1_MIN              0x10004C
#define TDFX_CLIP1_MAX              0x100050
#define TDFX_SRCFORMAT              0x100054
#define TDFX_SRCSIZE                0x100058
#define TDFX_SRCXY                  0x10005C
#define TDFX_BACKCOLOR              0x100060
#define TDFX_FORECOLOR              0x100064
#define TDFX_DSTSIZE                0x100068
#define TDFX_DSTXY                  0x10006C
#define TDFX_COMMAND                0x100070
#define TDFX_PATTERN0               0x100100
#define TDFX_PATTERN1               0x100104
#define TDFX_PATTERN_BASE           0x100100

#define TDFX_LAUNCH                 0x100080

// 3D register memory offsets

#define TDFX_3D_COMMAND             0x200120

// 2D command register bit definitions

#define TDFX_CMD_NOP                0x0
#define TDFX_CMD_BLT                0x1
#define TDFX_CMD_STRETCHBLT         0x2
#define TDFX_CMD_HOSTBLT            0x3
#define TDFX_CMD_HOSTSTRETCHBLT     0x4
#define TDFX_CMD_RECTFILL           0x5
#define TDFX_CMD_LINE               0x6
#define TDFX_CMD_POLYLINE           0x7
#define TDFX_CMD_POLYLINEFILL       0x8
#define TDFX_CMD_WR_SGRAMMODE       0xD
#define TDFX_CMD_WR_SGRAMMASK       0xE
#define TDFX_CMD_WR_SGRAMCOLOR      0xF

#define TDFX_CMD_REVERSELINE        0x00000200
#define TDFX_CMD_SHOWDELTA_X        0x00000400
#define TDFX_CMD_SHOWDELTA_Y        0x00000800
#define TDFX_CMD_STIPPLE_LINE       0x00001000
#define TDFX_CMD_MONOPATT_EN        0x00002000
#define TDFX_CMD_BLT_RIGHT_TO_LEFT  0x00004000
#define TDFX_CMD_BLT_BOTTOM_TO_TOP  0x00008000
#define TDFX_CMD_MONOTRANS_EN       0x00010000
#define TDFX_CMD_X_OFFSET           0x000E0000
#define X_OFFSET_SHIFT              17
#define TDFX_CMD_Y_OFFSET           0x00700000
#define Y_OFFSET_SHIFT              20
#define TDFX_CMD_CLIPSELECT         0x00800000
#define TDFX_CMD_ROP0               0xFF000000
#define ROP0_OFFSET_SHIFT           24

// 2D command extra register bit definitions

#define TDFX_CMDEXT_SOURCETRANS     0x00000001
#define TDFX_CMDEXT_DESTTRANS       0x00000002
#define TDFX_CMDEXT_VSYNC           0x00000004
#define TDFX_CMDEXT_FORCEPATT       0x00000008

// Pixel format register values

#define TDFX_FORMAT_1BPP            0x00000
#define TDFX_FORMAT_8BPP            0x10000
#define TDFX_FORMAT_15BPP           0x20000
#define TDFX_FORMAT_16BPP           0x30000
#define TDFX_FORMAT_24BPP           0x40000
#define TDFX_FORMAT_32BPP           0x50000

// Macro to define the number of precision bits in the bresenham error terms

#define TDFX_BRES_BITS              16

// 3D command register definitions

#define TDFX_3D_NOP                 0

// CLRCMD
//
// Waits until there are no more commands active using the memory mapped
// registers. Do three times per spec example.
//
#define CLRCMD()                                    \
{                                                   \
    ulong stat;                                     \
    int count = 3;                                  \
    WAITQ_TDFX(1);                                  \
    outmd(TDFX_3D_COMMAND,TDFX_3D_NOP);             \
    while (count--) {                               \
        do {                                        \
            stat = inmd(TDFX_STATUS);               \
            } while (stat & SST_BUSY);              \
        }                                           \
    TDFX_fifoEntries = stat & 0x3F;                 \
    TDFX_prevY = -1;                                \
}

// CLRCMD_FIFO
//
// Waits until there are no more commands active using the memory mapped
// registers. Do three times per spec example.
//
#define CLRCMD_FIFO(qEntries)                       \
{                                                   \
    CLRCMD();                                       \
    TDFX_fifoEntries -= qEntries;                   \
}

// WAITQ_TDFX
//
// register contains number of free entries - 1. Status of 31(0x1F) has
// 32 entries available
//
#define WAITQ_TDFX(qEntries)                            \
{                                                       \
    while ((TDFX_fifoEntries - qEntries) < 0)           \
        TDFX_fifoEntries = inmd(TDFX_STATUS) & 0x3F;    \
    TDFX_fifoEntries -= qEntries;                       \
}

// Macro to implement what appears to be a necessary bug fix when
// doing screen->screen blit operations. This code was found in the
// XFree86 drivers.

#define TDFX_BLT_BUGFIX(top)                            \
{                                                       \
    if (top >= TDFX_prevY-8 && top <= TDFX_prevY) {     \
        WAITQ_TDFX(1);                                  \
        outmd(TDFX_3D_COMMAND,TDFX_3D_NOP);             \
        }                                               \
}

//---------------------------------------------------------------------------
// 2D state management function prototypes
//---------------------------------------------------------------------------

N_int32     NAPI TDFX_SetDrawBuffer(GA_buffer *drawBuf);
N_int32     NAPI TDFX_IsIdle(void);
void        NAPI TDFX_WaitTillIdle(void);
void        NAPI TDFX_WaitTillIdleW16(void);
void        NAPI TDFX_WaitTillIdleW16_End(void);

//---------------------------------------------------------------------------
// Hardware cursor function prototypes
//---------------------------------------------------------------------------

void        NAPI TDFX_SetMonoCursor(GA_monoCursor *cursor);
void        NAPI TDFX_SetMonoCursorColor(GA_palette *foreground,GA_palette *background);
void        NAPI TDFX_SetColorCursor(GA_colorCursor *cursor);
N_int32     NAPI TDFX_SetCursorPos(N_int32 x,N_int32 y);
N_int32     NAPI TDFX_SetCursorPosW16(void);
void        NAPI TDFX_SetCursorPosW16_End(void);
void        NAPI TDFX_ShowCursor(N_int32 visible);

//---------------------------------------------------------------------------
// 2D state management function prototypes
//---------------------------------------------------------------------------

N_int32     NAPI TDFX_SetMix(N_int32 mix);
void        NAPI TDFX_SetForeColor(GA_color color);
void        NAPI TDFX_SetBackColor(GA_color color);
void        NAPI TDFX_Set8x8MonoPattern(N_int32 index,GA_pattern *pattern);
void        NAPI TDFX_Use8x8MonoPattern(N_int32 index);
void        NAPI TDFX_Use8x8TransMonoPattern(N_int32 index);
void        NAPI TDFX_Set8x8ColorPattern(N_int32 index,GA_colorPattern *pattern);
void        NAPI TDFX_Use8x8ColorPattern(N_int32 index);
void        NAPI TDFX_Use8x8TransColorPattern(N_int32 index,GA_color transparent);
void        NAPI TDFX_SetLineStipple(GA_stipple stipple);
void        NAPI TDFX_SetLineStippleCount(N_uint32 count);
void        NAPI TDFX_SetupForWindows(void);

//---------------------------------------------------------------------------
// 2D drawing function prototypes
//---------------------------------------------------------------------------
void        NAPI TDFX_DrawScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI TDFX_DrawPattScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI TDFX_DrawColorPattScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI TDFX_DrawTrap(GA_trap *trap);
void        NAPI TDFX_DrawPattTrap(GA_trap *trap);
void        NAPI TDFX_DrawColorPattTrap(GA_trap *trap);
void        NAPI TDFX_DrawRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI TDFX_DrawRectExt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,GA_color color,N_int32 mix);
void        NAPI TDFX_DrawRectLin(N_int32 dstOfs,N_int32 dstPitch,N_int32 left,N_int32 top,N_int32 width,N_int32 height,GA_color color,N_int32 mix);
void        NAPI TDFX_DrawPattRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI TDFX_DrawColorPattRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI TDFX_DrawLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast);
void        NAPI TDFX_DrawStippleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent);
void        NAPI TDFX_DrawBresenhamLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags);
void        NAPI TDFX_DrawBresenhamStippleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent);
void        NAPI TDFX_PutMonoImageMSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent);
void        NAPI TDFX_ClipMonoImageMSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
void        NAPI TDFX_BitBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI TDFX_BitBltLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI TDFX_BitBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,N_int32 flipY);
void        NAPI TDFX_SrcTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI TDFX_SrcTransBltLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI TDFX_SrcTransBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent,N_int32 flipY);
void        NAPI TDFX_DstTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI TDFX_DstTransBltLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI TDFX_DstTransBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent,N_int32 flipY);
void        NAPI TDFX_BitBltPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI TDFX_BitBltPattLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI TDFX_BitBltPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void        NAPI TDFX_BitBltColorPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI TDFX_BitBltColorPattLin(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI TDFX_BitBltColorPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);

#endif /* __ACCEL_H */

