/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module containg PLL clock programing functions.
*
****************************************************************************/

#include "internal.h"
#include "accel.h"

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
REMARKS:
Program the selected pixel clock into the PLL
****************************************************************************/
int TDFX_setPLLPixelClock(
    N_fix32 clock)
{
    int N,M,P;

    if (DV.cntMode.flags & gaVGA) {
        if (ABS(clock - N_FIXCONST(25.18)) < CLOSE_CLOCK)
            return 0;
        if (ABS(clock - N_FIXCONST(28.32)) < CLOSE_CLOCK)
            return 1;
        }
    DV.pll.calculateClock(&DV.pixelClock,clock,&M,&N,&P);
    outpd(TDFX_PLLCTRL0, (M << 8) | (N << 2) | P);
    return 3;
}

/****************************************************************************
REMARKS:
Program the selected memory clock into the memory clock PLL
****************************************************************************/
void TDFX_setPLLMemoryClock(
    N_fix32 clock)
{
    // We don't re-program this for now. The default is correct anyway,
    // and doing so screws up our Windows drivers.
}

/****************************************************************************
REMARKS:
Configure the Clock module ready for use.
****************************************************************************/
void CLOCK_init(void)
{
    int     N,M,P;
    ulong   pllCtrl2;
    N_fix32 mClk;

    //------------------------------------------------------------------
    // Configure the pixel clock limits
    //------------------------------------------------------------------
    if (DV.chipID >= TDFX_VOODOO4) {
        DV.pixelClockMaxTxt         = N_FIXCONST( 57);
        DV.pixelClockMax4           = N_FIXCONST(100);
        DV.pixelClockMax8           = N_FIXCONST(350);
        DV.pixelClockMax16          = N_FIXCONST(350);
        DV.pixelClockMax24          = N_FIXCONST(350);
        DV.pixelClockMax32          = N_FIXCONST(350);
        }
    else if (DV.chipID >= TDFX_VOODOO3) {
        DV.pixelClockMaxTxt         = N_FIXCONST( 43);
        DV.pixelClockMax4           = N_FIXCONST(100);
        DV.pixelClockMax8           = N_FIXCONST(330);
        DV.pixelClockMax16          = N_FIXCONST(330);
        DV.pixelClockMax24          = N_FIXCONST(330);
        DV.pixelClockMax32          = N_FIXCONST(330);
        }
    else {
        DV.pixelClockMaxTxt         = N_FIXCONST( 37);
        DV.pixelClockMax4           = N_FIXCONST( 76);
        DV.pixelClockMax8           = N_FIXCONST(300);
        DV.pixelClockMax16          = N_FIXCONST(300);
        DV.pixelClockMax24          = N_FIXCONST(300);
        DV.pixelClockMax32          = N_FIXCONST(300);
        }

    //------------------------------------------------------------------
    // Configure the PLL programming parameters using the following
    // formula:
    //              Fout = (M+k1) * Fref * k3
    //                     ------------------
    //                       (N+k2) * 2^R
    //------------------------------------------------------------------
    DV.pixelClock.flags         = gaPLL_favorMaxN | gaPLL_favorMaxM | gaPLL_favorMaxVCO;
    DV.pixelClock.fRef          = N_FIXCONST(14.31818);
    DV.pixelClock.minVCO        = N_FIXCONST(8.0);
    DV.pixelClock.maxVCO        = N_FIXCONST(350.0);
    DV.pixelClock.minM          = 3;
    DV.pixelClock.maxM          = 255;
    DV.pixelClock.minN          = 3;
    DV.pixelClock.maxN          = 64;
    DV.pixelClock.maxR          = 3;
    DV.pixelClock.k1            = 2;
    DV.pixelClock.k2            = 2;
    DV.pixelClock.k3            = 1;
    DV.pixelClock.numClocks     = 0;

    //------------------------------------------------------------------
    // Configure the memory clock values
    //------------------------------------------------------------------
    pllCtrl2 = inpd(TDFX_PLLCTRL2);
    N = (pllCtrl2 >> 2) & 0x3F;
    M = (pllCtrl2 >> 8) & 0xFF;
    P = pllCtrl2 & 0x3;
    mClk = FXdiv(FXmul(DV.pixelClock.fRef,N_TOFIX(M+2)),N_TOFIX(N+2)) >> P;
    DV.opt.defaultMemoryClock   = mClk;
    DV.memoryClockVGA           = mClk;
    DV.memoryClock              = DV.pixelClock;
    DV.opt.maxMemoryClock       = FXmul(mClk,N_FIXCONST(1.2));
    if (DV.opt.memoryClock == 0 || DV.opt.memoryClock > DV.opt.maxMemoryClock)
        DV.opt.memoryClock      = mClk;

    //------------------------------------------------------------------
    // Configure the pixel and memory clock programming functions
    //------------------------------------------------------------------
    DV.pll.calculateClock       = DRV_calculateClock;
    DV.pll.setPLLPixelClock     = TDFX_setPLLPixelClock;
    DV.pll.setPLLMemoryClock    = TDFX_setPLLMemoryClock;
}

