/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  IBM PC (MS DOS)
*
* Description:  Hardware specific device detection code.
*
****************************************************************************/

#include "internal.h"

/*----------------------------- Implementation ----------------------------*/
/****************************************************************************
PARAMETERS:
state   - New DPMS state to enable

REMARKS:
Puts the graphics hardware into the specified DPMS state.
****************************************************************************/
void TDFX_setDPMS(
    uchar state)
{
    TDFX_unlockRegisters();
    switch (state) {
        case DPMS_on:       // HSync active, VSync active
            modregd(TDFX_DACMODE,0x1E,0x00);
            break;
        case DPMS_standby:  // HSync inactive, VSync active
            modregd(TDFX_DACMODE,0x1E,0x18);
            break;
        case DPMS_suspend:  // HSync active, VSync inactive
            modregd(TDFX_DACMODE,0x1E,0x06);
            break;
        case DPMS_off:      // HSync inactive, VSync inactive
            modregd(TDFX_DACMODE,0x1E,0x1E);
            break;
        }
    TDFX_lockRegisters();
}

/****************************************************************************
PARAMETERS:
channel - Monitor port to begin control for

REMARKS:
Puts the graphics hardware into a state ready to do I2C communications.
****************************************************************************/
void NAPI TDFX_SCIbegin(
    N_int32 channel)
{
    TDFX_unlockRegisters();
    setregd(TDFX_VIDSERIALPARALLELPORT,0x40000);
}

/****************************************************************************
PARAMETERS:
channel - Monitor port to begin control for

REMARKS:
Restores the graphics hardware after doing I2C communications.
****************************************************************************/
void NAPI TDFX_SCIend(
    N_int32 channel)
{
    clrregd(TDFX_VIDSERIALPARALLELPORT,0x40000);
    TDFX_lockRegisters();
}

/****************************************************************************
PARAMETERS:
channel - Monitor port to begin control for
bit         - 0 or 1 for the value to set

REMARKS:
Sets the value of the SCL clock line on the I2C channel
****************************************************************************/
void NAPI TDFX_SCIwriteSCL(
    N_int32 channel,
    N_int32 bit)
{
    modregd(TDFX_VIDSERIALPARALLELPORT,0x80000,(bit & 1) << 19);
}

/****************************************************************************
PARAMETERS:
channel - Monitor port to begin control for
bit         - 0 or 1 for the value to set

REMARKS:
Sets the value of the SDA data line on the I2C channel
****************************************************************************/
void NAPI TDFX_SCIwriteSDA(
    N_int32 channel,
    N_int32 bit)
{
    modregd(TDFX_VIDSERIALPARALLELPORT,0x100000,(bit & 1) << 20);
}

/****************************************************************************
PARAMETERS:
channel - Monitor port to begin control for

RETURNS
0 or 1 for the value read from I2C channel

REMARKS:
Reads the value of the SCL clock line on the I2C channel
****************************************************************************/
int NAPI TDFX_SCIreadSCL(
    N_int32 channel)
{
    return (inpd(TDFX_VIDSERIALPARALLELPORT) >> 21) & 1;
}

/****************************************************************************
PARAMETERS:
channel - Monitor port to begin control for

RETURNS
0 or 1 for the value read from I2C channel

REMARKS:
Reads the value of the SDA data line on the I2C channel
****************************************************************************/
int NAPI TDFX_SCIreadSDA(
    N_int32 channel)
{
    return (inpd(TDFX_VIDSERIALPARALLELPORT) >> 22) & 1;
}

