/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  IBM PC (MS DOS)
*
* Description:  Hardware specific device detection code.
*
****************************************************************************/

#include "internal.h"

/*--------------------------- Global variables ----------------------------*/

/* Names of manufacturer and chip revision id's. */

char *DRV_manufacturerName = "3dfx";

char *DRV_chipsetNames[] = {
    "Voodoo Banshee",
    "Voodoo3",
    "Voodoo4",
    "Voodoo5",
    NULL,
    };

#include "../certdata.c"

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
RETURNS:
True if hardware detected correctly, false if not.

REMARKS:
The purpose of this function is to fill in the memory and I/O resource
tables that the driver requires based on the PCI configuration space
information. No other information is detected at this time, since we may
need to access registers via the mapped in resources which are not mapped
in by the OS loader code yet.
****************************************************************************/
ibool PNP_detect(void)
{
    //------------------------------------------------------------------
    // Detect the chipset revision
    //------------------------------------------------------------------
    switch (DV.PCIConfig.DeviceID) {
        case 0x0003: DV.chipID = TDFX_BANSHEE; break;
        case 0x0005: DV.chipID = TDFX_VOODOO3; break;
        case 0x0009:
            switch (DV.PCIConfig.u.type0.SubSystemID) {
                case 0x0001:            // V5-6000
                case 0x0002:            // V5-5500
                case 0x0003:            // V5-5500
                case 0x0006:            // V5-5000
                    DV.chipID = TDFX_VOODOO5;
                    break;
                case 0x0004:
                case 0x0005:
                default:
                    DV.chipID = TDFX_VOODOO4;
                    break;
                }
            break;
        default:
            return false;
        }
    if (DV.chipID >= TDFX_VOODOO4) {
        // The cards actually have 64Meg on board, but it may be 32Meg split
        // up between two chipsets in SLI mode (or so it seems anyway).
        DC.LinearBasePtr = (DV.PCIConfig.u.type0.BaseAddress14 & 0xFFE00000);
        DC.LinearSize = MIN(DV.PCIConfig.u.type0.BaseAddress14Len,32 * MB);
        }
    else {
        DC.LinearBasePtr = (DV.PCIConfig.u.type0.BaseAddress14 & 0xFFE00000);
        DC.LinearSize = MIN(DV.PCIConfig.u.type0.BaseAddress14Len,16 * MB);
        }
    DC.MMIOBase[0] = (DV.PCIConfig.u.type0.BaseAddress10 & 0xFFE00000);
    DC.MMIOLen[0]  = 16 * MB;
    DC.IOBase = DV.PCIConfig.u.type0.BaseAddress18 & ~0xF;
    return true;
}

/****************************************************************************
RETURNS:
True if hardware detected correctly, false if not.

REMARKS:
Detects the presence of one of the supported hardware devices and fills
in the global driver device context with the appropriate information.
****************************************************************************/
ibool CHIP_detect(void)
{
    return true;
}

