/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  IBM PC (MS DOS)
*
* Description:  Header file containing the internal chipset definitions
*               for the driver.
*
****************************************************************************/

#ifndef __INTERNAL_H
#define __INTERNAL_H

#include "chipid.h"
#include "accel.h"

/*--------------------- Macros and type definitions -----------------------*/

/* Maximum CRT offset register value (with extension bits) */

#define TDFX_MAX_OFFSET     0xFFF

/* Structure to maintain the hardware cursor state */
typedef struct {
    long    isDoubleScan;
    long    hotX;
    long    hotY;
    long    curX;
    long    curY;
    long    visible;
    } TDFX_CursorState;

/*-------------------------- Function Prototypes --------------------------*/

/* Register lock/unlock functions in reglock.c */
void    TDFX_unlockRegisters(void);
void    TDFX_lockRegisters(void);

/* Functios to enable and disable memory mapped I/O */

/* DPMS support functions */

void    TDFX_setDPMS(uchar state);

/* Chipset specific helper functions */

int TDFX_findLinearMem(int maxMB);

ulong   TDFX_getTextModeFlags(GA_mode *mode);
ulong   TDFX_getModeFlags(GA_mode *mode);
void    TDFX_setChipsetRegisters(void);
void    TDFX_setMemoryClock(void);
void    TDFX_setCRTCRegisters(GA_CRTCInfo *crtc);
void    TDFX_setPixelClock(N_fix32 clock);
void    TDFX_enableLinear(void);
void    TDFX_disableLinear(void);
void    TDFX_setupAccelerator(void);
void    TDFX_preModeSet(void);
void    TDFX_postModeSet(void);
void    TDFX_get2DAccelInfo(GA_modeInfo *gaInfo,GA_modeEntry *entry,N_int32 virtualX,N_int32 virtualY);
void    TDFX_getVideoAccelInfo(GA_modeInfo *gaInfo,GA_modeEntry *entry);
void    TDFX_get3DAccelInfo(GA_modeInfo *gaInfo,GA_modeEntry *entry);
void    TDFX_lockDownTimings(VGA_hCRTC *h,VGA_vCRTC *v,GA_CRTCInfo *crtc);
N_int32 TDFX_setBytesPerLine(N_int32 bytesPerLine);
N_int32 TDFX_roundBytesPerLine(N_int32 bytesPerLine,N_int32 bitsPerPixel);
void    TDFX_configChipsetFlags(void);

/* Chipset specific API functions */

N_int32 NAPI TDFX_GetCurrentScanLine(void);
void    NAPI TDFX_SetBank(N_int32 bank);
void    NAPI TDFX_SetDisplayStart(N_int32 offset,N_int32 waitVRT);
void    NAPI TDFX_SetPaletteData(GA_palette *pal,N_int32 num,N_int32 index,N_int32 waitVRT);
void    NAPI TDFX_GetPaletteData(GA_palette *pal,N_int32 num,N_int32 index);

/* Functions to program the RAMDAC */

void    TDFX_setDACRegisters(void);

/* Functions to program the PLL clocks */

int     TDFX_setPLLPixelClock(N_fix32 clock);
void    TDFX_setPLLMemoryClock(N_fix32 clock);

/* DDC support functions */
void    NAPI TDFX_SCIbegin(N_int32 channel);
void    NAPI TDFX_SCIend(N_int32 channel);
void    NAPI TDFX_SCIwriteSCL(N_int32 channel,N_int32 bit);
void    NAPI TDFX_SCIwriteSDA(N_int32 channel,N_int32 bit);
int     NAPI TDFX_SCIreadSCL(N_int32 channel);
int     NAPI TDFX_SCIreadSDA(N_int32 channel);

#endif  /* __INTERNAL_H */

