/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module containing RAMDAC programming code.
*
****************************************************************************/

#include "snap/graphics/driver.h"
#include "internal.h"
#include "accel.h"

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
REMARKS:
Program the RAMDAC for the appropriate mode
****************************************************************************/
void TDFX_setDACRegisters(void)
{
    ulong mode,multiplex;

    switch (DV.cntMode.bits) {
        case 15:
        case 16:
            mode = 0x1;
            break;
        case 24:
            mode = 0x2;
            break;
        case 32:
            mode = 0x3;
            break;
        default:
            mode = 0x0;
            break;
        }
    mode <<= 18;
    if (DV.cntMode.flags & gaMultiplex) {
        multiplex = 1;
        mode |= SST_VIDEO_2X_MODE_EN;
        }
    else
        multiplex = 0;
    modregd(TDFX_DACMODE,0x1,multiplex);
    modregd(TDFX_VIDPROCCFG,0x41C0000,mode);
}

/****************************************************************************
REMARKS:
Function to set 8-bit wide RAMDAC mode.
****************************************************************************/
ibool TDFX_set8BitDAC(
    ibool enable)
{
    if (enable)
        setregd(TDFX_VGAINIT0,0x4);
    else
        clrregd(TDFX_VGAINIT0,0x4);
    return true;
}

/****************************************************************************
REMARKS:
Configure the RAMDAC module ready for use.
****************************************************************************/
void RAMDAC_init(void)
{
    DV.dac.setDACRegisters = TDFX_setDACRegisters;
    DV.dac.set8BitDAC = TDFX_set8BitDAC;
}

