/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  C support routines for hardware cursor clipping.
*
****************************************************************************/

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
PARAMETERS:
mask        - Source mask to clip
clippedMask - Place to store the clipped mask values
xAdjust     - Left shift factor in X direction
xRevAdjust  - Reverse shift factor (actually 32-xAdjust)

REMARKS:
This function clips a 64-bit value read as two longs from an array,
and stores the clipped or shifted value into the second array.
****************************************************************************/
static void AdjustMask(
    N_uint8 *mask,
    N_uint8 *clippedMask,
    int xAdjust,
    int xRevAdjust)
{
    ulong lo = GET_LONG_BE(mask,0);
    ulong hi = GET_LONG_BE(mask,1);
    if (xAdjust < 32) {
        lo = (lo << xAdjust) | (hi >> xRevAdjust);
        hi = (hi << xAdjust);
        }
    else {
        lo = (hi << xRevAdjust);
        hi = 0;
        }
    PUT_LONG_BE(clippedMask,0,lo);
    PUT_LONG_BE(clippedMask,1,hi);
}

/****************************************************************************
PARAMETERS:
cursor      - Pointer to an GA_monoCursor structure for the cursor image

REMARKS:
This function downloads the specified monochrome cursor definition from the
application into the hardware cursor. The cursor data is passed by the
application in the GA_monoCursor format, which is similar to the Windows
3.1 cursor file format.
****************************************************************************/
void DRV_ClipCursorImage(
    N_uint8 *XORMask,
    N_uint8 *ANDMask,
    N_uint8 *ClippedXORMask,
    N_uint8 *ClippedANDMask,
    int xAdjust,
    int yAdjust)
{
    int     y,maxy = 64 - yAdjust;
    int     xRevAdjust = 32 - xAdjust;

    XORMask += (yAdjust * 8);
    ANDMask += (yAdjust * 8);
    for (y = 0; y < maxy; y++) {
        if (xAdjust) {
            AdjustMask(XORMask,ClippedXORMask,xAdjust,xRevAdjust);
            AdjustMask(ANDMask,ClippedANDMask,xAdjust,xRevAdjust);
            }
        else {
            memcpy(ClippedXORMask,XORMask,8);
            memcpy(ClippedANDMask,ANDMask,8);
            }
        XORMask += 8;
        ANDMask += 8;
        ClippedXORMask += 8;
        ClippedANDMask += 8;
        }
    memset(ClippedXORMask,0,yAdjust * 8);
    memset(ClippedANDMask,0,yAdjust * 8);
}
