############################################################################
#
#                    SciTech SNAP Graphics Architecture
#
#  ========================================================================
#
#   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
#
#   This file may be distributed and/or modified under the terms of the
#   GNU General Public License version 2 as published by the Free
#   Software Foundation and appearing in the file LICENSE.GPL included
#   in the packaging of this file.
#
#   Licensees holding a valid Commercial License for this product from
#   SciTech Software, Inc. may use this file in accordance with the
#   Commercial License Agreement provided with the Software.
#
#   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
#   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE.
#
#   See http://www.scitechsoft.com/license/ for information about
#   the licensing options available and how to purchase a Commercial
#   License Agreement.
#
#   Contact license@scitechsoft.com if any conditions of this licensing
#   are not clear to you, or you have questions about licensing options.
#
#  ========================================================================
#
# Descripton:   Common makefile script to build all SNAP drivers.
#
#               The parent makefile must define CHIPNAME before including
#               this makefile to define the name of the driver created.
#
#############################################################################

.IMPORT .IGNORE : ANALYSE DEBUG_NEW_DRIVER DETECT_ONLY

#----------------------------------------------------------------------------
# Enabling this flag will turn off some advanced stuff in the drivers to
# make debugging and developing new drivers easier. Turn this off once the
# driver is up and running.
#----------------------------------------------------------------------------

.IF $(DEBUG_NEW_DRIVER)
CFLAGS      += -DDEBUG_NEW_DRIVER
.ENDIF

.IF $(DETECT_ONLY)
CFLAGS      += -DDETECT_ONLY
.ENDIF

#----------------------------------------------------------------------------
# Define the lists of object files
#----------------------------------------------------------------------------

DLL_OBJS    = dllstart$O

DRV_OBJS    = gen_api$O gen_help$O gen_vga$O gen_ddc$O getfunc$O gdetect$O \
              detect$O config$O init$O ports$O reglock$O clock$O ramdac$O \
              ddc$O cursor$O 2d$O $(LOCAL_OBJS)
.IF $(EXTERNAL_DAC)
DRV_OBJS    += finddac$O findclk$O
.ENDIF
.IF $(BUILD_DLL)
DRV_OBJS    += _pm_imp$O
.ENDIF

GEN_OBJS    = ncommon$O pciinit$O ddcread$O ddc2bi$O edid$O belib$O \
              gainit$O crtcdbx$O stereo$O galib$O mondbx$O gtfcalc$O recmode$O \
              findchip$O bugfix01$O _be_exp$O

DETECT_OBJS = detect_name$O detect$O reglock$O _pm_imp$O

.IF $(USE_WIN32)
ANALYSE_OBJS = analyse$O gainit$O crtcdbx$O galib$O _gatimer$O gtfcalc$O \
               recmode$O
.ELSE
ANALYSE_OBJS = analyse$O gainit$O crtcdbx$O vesavbe$O mondbx$O stereo$O \
               galib$O _gatimer$O gtfcalc$O recmode$O
.ENDIF

OPTIMISE_OBJS = optimise$O

DIFF_OBJS   = diff$O

#----------------------------------------------------------------------------
# Define the targets to build the mode analysis and test tools. Note that
# the analyse tools have to link against the TEST_HARNESS version of the
# PM library.
#----------------------------------------------------------------------------

.IF $(ANALYSE)
NO_PMLIB    := 1
CFLAGS      += -D__ANALYSE__ -I$(PRIVATE)\src\snap\graphics
EXELIBS     += $(LL)n_ref2d$(LE) $(LL)pm_test$(LE)
all:        analyse$E optimise$E diff$E
analyse$E:  $(ANALYSE_OBJS) $(GEN_OBJS) $(DRV_OBJS)
optimise$E: $(OPTIMISE_OBJS)
diff$E:     $(DIFF_OBJS)
.ENDIF

#----------------------------------------------------------------------------
# Target to build debug binary drivers (this is our default target)
#----------------------------------------------------------------------------

.PHONY db:
    @build_db ow10-drv $(CHIPNAME).dll BUILD_DLL=1 DBG=1 $(MFLAGS)
    $(CP) $(CHIPNAME).dll $(SCITECH)/drivers/graphics/$(CHIPNAME).drv

.PHONY dbdetect:
    @build_db ow10-drv $(CHIPNAME)_detect.dll BUILD_DLL=1 DBG=1 DETECT_ONLY=1 $(MFLAGS)
    $(CP) $(CHIPNAME)_detect.dll $(SCITECH)/drivers/graphics/detect/$(CHIPNAME).drv

#----------------------------------------------------------------------------
# Target to build binary drivers
#----------------------------------------------------------------------------

.PHONY mkdrv:
    @build ow10-drv $(CHIPNAME).dll -u BUILD_DLL=1 OPT=1
    $(CP) $(CHIPNAME).dll $(PRIVATE)/snap/graphics/$(CHIPNAME).drv
    @dmake cleanexe

.PHONY mkdetect:
    @build ow10-drv $(CHIPNAME)_detect.dll -u BUILD_DLL=1 OPT=1 DETECT_ONLY=1
    $(CP) $(CHIPNAME)_detect.dll $(PRIVATE)/snap/graphics/detect/$(CHIPNAME).drv
    @dmake cleanexe

#----------------------------------------------------------------------------
# Define the targets to build the real Binary Portable DLL driver or
# the test harness driver library.
#----------------------------------------------------------------------------

.IF $(BUILD_DLL)

$(CHIPNAME).dll: $(DLL_OBJS) $(DRV_OBJS)

$(CHIPNAME)_detect.dll: $(DLL_OBJS) $(DETECT_OBJS)

.ELSE

CFLAGS      += -DTEST_HARNESS
ASFLAGS     += -I $(PRIVATE)\src\snap\graphics
LIBFILE     = $(LP)n_driver$L
$(LIBFILE): $(GEN_OBJS) $(DRV_OBJS)

.ENDIF

LIBCLEAN    = *.dll *.lib *.a

#----------------------------------------------------------------------------
# Define the list of object files to create dependency information for
#----------------------------------------------------------------------------

DEPEND_OBJ      = $(DLL_OBJS) $(GEN_OBJS) $(DRV_OBJS) $(ANALYSE_OBJS)
DEPEND_INC      += ;../..;hallib
DEPEND_SRC      = $(PRIVATE)/src/snap/common;$(PRIVATE)/src/snap/graphics/drivers;$(PRIVATE)/src/snap/graphics;$(PRIVATE)/src/snap/graphics/drivers/detect;$(PRIVATE)/src/snap/graphics/drivers/bugfix01

.INCLUDE: "$(SCITECH)/makedefs/common.mk"

