/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  C support routines for 2D drawing functions.
*
****************************************************************************/

#include "internal.h"

/*--------------------------- Global variables ----------------------------*/

// Macro to create a Win32 ROP from a SNAP ROP3 code
#define MKROP(gaRop)    (((DWORD)gaRop) << 16)

// Translation table for source/dest ROP codes
DWORD DD_dwSrcRopCodes[] = {
    BLACKNESS,          // GA_R2_BLACK
    NOTSRCERASE,        // GA_R2_NOTMERGESRC
    MKROP(GA_R3_DSna),  // GA_R2_MASKNOTSRC
    NOTSRCCOPY,         // GA_R2_NOTCOPYSRC
    SRCERASE,           // GA_R2_MASKSRCNOT
    MKROP(GA_R3_Dn),    // GA_R2_NOT
    SRCINVERT,          // GA_R2_XORSRC         (GA_XOR_MIX)
    MKROP(GA_R3_DSan),  // GA_R2_NOTMASKSRC
    SRCAND,             // GA_R2_MASKSRC        (GA_AND_MIX)
    MKROP(GA_R3_DSxn),  // GA_R2_NOTXORSRC
    MKROP(GA_R3_D),     // GA_R2_NOP            (GA_NOP_MIX)
    MERGEPAINT,         // GA_R2_MERGENOTSRC
    SRCCOPY,            // GA_R2_COPYSRC        (GA_REPLACE_MIX)
    MKROP(GA_R3_SDno),  // GA_R2_MERGESRCNOT
    SRCPAINT,           // GA_R2_MERGESRC       (GA_OR_MIX)
    WHITENESS,          // GA_R2_WHITE
    };

// Translation table for pattern/dest ROP codes
DWORD DD_dwPatRopCodes[] = {
    BLACKNESS,          // GA_R2_BLACK
    MKROP(GA_R3_DPon),  // GA_R2_NOTMERGESRC
    MKROP(GA_R3_DPna),  // GA_R2_MASKNOTSRC
    MKROP(GA_R3_Pn),    // GA_R2_NOTCOPYSRC
    MKROP(GA_R3_PDna),  // GA_R2_MASKSRCNOT
    MKROP(GA_R3_Dn),    // GA_R2_NOT
    PATINVERT,          // GA_R2_XORSRC         (GA_XOR_MIX)
    MKROP(GA_R3_DPan),  // GA_R2_NOTMASKSRC
    MKROP(GA_R3_DPa),   // GA_R2_MASKSRC        (GA_AND_MIX)
    MKROP(GA_R3_PDxn),  // GA_R2_NOTXORSRC
    MKROP(GA_R3_D),     // GA_R2_NOP            (GA_NOP_MIX)
    MKROP(GA_R3_DPno),  // GA_R2_MERGENOTSRC
    PATCOPY,            // GA_R2_COPYSRC        (GA_REPLACE_MIX)
    MKROP(GA_R3_PDno),  // GA_R2_MERGESRCNOT
    MKROP(GA_R3_DPo),   // GA_R2_MERGESRC       (GA_OR_MIX)
    WHITENESS,          // GA_R2_WHITE
    };

// Tables of valid ROP's indexed by GA_mixCodesType enumeration
ibool       DD_validSrcRops[16];
ibool       DD_validPatRops[16];
int         DD_mix;
GA_color    DD_foreColor;
GA_color    DD_backColor;

/*----------------------------- Implementation ----------------------------*/

#include "../ellipse.c"

/****************************************************************************
REMARKS:
Initialise the 2D drawing functions to be exported by the driver.
****************************************************************************/
void ACCEL2D_init(void)
{
    //-----------------------------------------------------------------------
    // Setup 2D state management functions
    //-----------------------------------------------------------------------
    DV.state2d.SetDrawBuffer            = DD_SetDrawBuffer;
    DV.state2d.IsIdle                   = DD_IsIdle;
    DV.state2d.WaitTillIdle             = DD_WaitTillIdle;
    DV.state2d.EnableDirectAccess       = DD_EnableDirectAccess;
    DV.state2d.DisableDirectAccess      = DD_DisableDirectAccess;
    DV.state2d.SetMix                   = DD_SetMix;
    DV.state2d.SetForeColor             = DD_SetForeColor;
    DV.state2d.SetBackColor             = DD_SetBackColor;
    DV.state2d.Set8x8MonoPattern        = CNOT_SUPPORTED;
    DV.state2d.Use8x8MonoPattern        = CNOT_SUPPORTED;
    DV.state2d.Use8x8TransMonoPattern   = CNOT_SUPPORTED;
    DV.state2d.Set8x8ColorPattern       = CNOT_SUPPORTED;
    DV.state2d.Use8x8ColorPattern       = CNOT_SUPPORTED;
    DV.state2d.Use8x8TransColorPattern  = CNOT_SUPPORTED;
    DV.state2d.SetLineStipple           = CNOT_SUPPORTED;
    DV.state2d.SetLineStippleCount      = CNOT_SUPPORTED;
    DV.state2d.SetPlaneMask             = CNOT_SUPPORTED;
    DV.state2d.SetupForWindows          = CNOT_SUPPORTED;

    //-----------------------------------------------------------------------
    // Setup 2D drawing functions. Everything is initially unsupported.
    //-----------------------------------------------------------------------
    NotSupported2DDraw(DrawScanList);
    NotSupported2DDraw(DrawPattScanList);
    NotSupported2DDraw(DrawColorPattScanList);
    NotImplemented2DDraw(DrawEllipseList);
    NotImplemented2DDraw(DrawPattEllipseList);
    NotImplemented2DDraw(DrawColorPattEllipseList);
    NotImplemented2DDraw(DrawPattFatEllipseList);
    NotImplemented2DDraw(DrawFatEllipseList);
    NotImplemented2DDraw(DrawColorPattFatEllipseList);
    NotSupported2DDraw(DrawRect);
    NotSupported2DDraw(DrawRectExt);
    NotSupported2DDraw(DrawRectLin);
    NotSupported2DDraw(DrawPattRect);
    NotSupported2DDraw(DrawColorPattRect);
    NotSupported2DDraw(DrawTrap);
    NotSupported2DDraw(DrawPattTrap);
    NotSupported2DDraw(DrawColorPattTrap);
    NotSupported2DDraw(DrawLineInt);
    NotSupported2DDraw(DrawBresenhamLine);
    NotSupported2DDraw(DrawStippleLineInt);
    NotSupported2DDraw(DrawBresenhamStippleLine);
    NotSupported2DDraw(PutMonoImageMSBSys);
    NotSupported2DDraw(ClipMonoImageMSBSys);
    NotSupported2DDraw(BitBlt);
    NotSupported2DDraw(BitBltLin);
    NotSupported2DDraw(BitBltSys);
    NotSupported2DDraw(BitBltPatt);
    NotSupported2DDraw(BitBltPattLin);
    NotSupported2DDraw(BitBltPattSys);
    NotSupported2DDraw(BitBltColorPatt);
    NotSupported2DDraw(BitBltColorPattLin);
    NotSupported2DDraw(BitBltColorPattSys);
    NotSupported2DDraw(SrcTransBlt);
    NotSupported2DDraw(SrcTransBltLin);
    NotSupported2DDraw(SrcTransBltSys);
    NotSupported2DDraw(DstTransBlt);
    NotSupported2DDraw(DstTransBltLin);
    NotSupported2DDraw(DstTransBltSys);
}

/****************************************************************************
RETURNS:
True if the DirectDraw ROP is available, false if not.

REMARKS:
Check the fat 256 bit entry array for the specific ROP to see if it is
supported. The dwRop's array in DirectDraw is encoded as 256 entry ROP array
to determine what ROP's are supported by the driver. Presently only those
ROP's defined in the WINGDI.H header file as supported via this interface.
****************************************************************************/
static ibool DD_checkForROP(
    DWORD dwRop)
{
    int index = (dwRop >> 21);
    int mask = (1UL << ((dwRop >> 16) & 0x1F));
    return (DD_caps.dwRops[index] & mask) != 0;
}

/****************************************************************************
REMARKS:
Called to initialise the accelerator during the mode initialisation process.
****************************************************************************/
void DD_setupAccelerator(void)
{
    int i;

    // Re-read the DirectDraw capabilities after the mode has been changed
    // so that we can get the actual capabilities of the display hardware
    // in this mode.
    DD_caps.dwSize = sizeof(DDCAPS);
    DD_caps.dwCaps = 0;
    IDirectDraw7_GetCaps(DD_lpDD,&DD_caps,NULL);
    if (DD_caps.dwCaps & DDCAPS_BLT)
        DC.Attributes |= gaHaveAccel2D;
    if (DD_caps.dwCaps & DDCAPS_3D)
        DC.Attributes |= gaHaveAccel3D;
    if (DD_caps.dwCaps & DDCAPS_OVERLAY)
        DC.Attributes |= gaHaveAccelVideo;
    if (DD_caps.dwCaps2 & DDCAPS2_FLIPNOVSYNC)
        DC.Attributes |= gaHaveTripleBuffer;
    DV.cntModeInfo.Attributes |= DC.Attributes;

    // Now determine what 2D acceleration functions are available
    ACCEL2D_init();
    if (DD_caps.dwCaps & DDCAPS_BLTCOLORFILL) {
        Set2DDraw(DD,DrawScanList);
        Set2DDraw(DD,DrawRect);
        Set2DDraw(DD,DrawRectExt);
        Set2DDraw(DD,DrawTrap);
        Set2DDraw(DRV,DrawEllipseList);
        Set2DDraw(DRV,DrawFatEllipseList);
        }
    if (DD_caps.dwCaps & DDCAPS_BLT) {
        Set2DDraw(DD,BitBlt);
        if (DD_caps.dwCaps & (DDCAPS_BLTSTRETCH | DDCAPS_3D)) {
            Set2DDraw(DD,BitBltFxTest);
            Set2DDraw(DD,BitBltFx);
            Set2DDraw(DD,BitBltFxSys);
            }
        }
    if (DD_caps.dwCaps & DDCAPS_COLORKEY) {
        if (DD_caps.dwCKeyCaps & DDCKEYCAPS_SRCBLT) {
            Set2DDraw(DD,SrcTransBlt);
            }
        if (DD_caps.dwCKeyCaps & DDCKEYCAPS_DESTBLT) {
            Set2DDraw(DD,DstTransBlt);
            }
        }

    // Initialise the array of pre-defined available ROP's
    for (i = GA_R2_BLACK; i <= GA_R2_WHITE; i++) {
        DD_validSrcRops[i] = DD_checkForROP(DD_dwSrcRopCodes[i]);
        DD_validPatRops[i] = DD_checkForROP(DD_dwPatRopCodes[i]);
        }

    // Replace mix is always available!
    DD_validSrcRops[GA_REPLACE_MIX] = true;
    DD_validPatRops[GA_REPLACE_MIX] = true;
}

/****************************************************************************
PARAMETERS:
drawBuf - Buffer to make the active drawing buffer.

RETURNS:
0 on success, -1 on failure
****************************************************************************/
N_int32 NAPI DD_SetDrawBuffer(
    GA_buffer *drawBuf)
{
    // This function can't be used directly with the DirectDraw driver.
    // Instead the app must go through the buffer manager API's.
    (void)drawBuf;
    return -1;
}

/****************************************************************************
REMARKS:
This function is the same as the WaitTillIdle function, however it does
not wait for the engine to become idle but instead returns immediately
with the current status.
****************************************************************************/
N_int32 NAPI DD_IsIdle(void)
{
    HRESULT ret;

    do {
        if ((ret = IDirectDrawSurface7_GetBltStatus(DD_lpVisible,DDGBS_ISBLTDONE)) == DDERR_SURFACELOST)
            PM_doSuspendApp();
        if (ret == DDERR_UNSUPPORTED)
            PM_fatalError("Unable to get blt status!");
        } while (ret != DD_OK && ret != DDERR_WASSTILLDRAWING);
    return (ret == DD_OK);
}

/****************************************************************************
REMARKS:
This function waits until the hardware accelerator has completed all
currently queued rendering operations. The primary purpose of this function
is to provide the application with the ability to ensure all rendering is
complete, before swapping display pages when doing double buffering, or
before directly accessing the framebuffer memory.
****************************************************************************/
void NAPI DD_WaitTillIdle(void)
{
    while (!DD_IsIdle())
        ;
}

/****************************************************************************
REMARKS:
Enables direct access to the currently active display surface for
software rendering etc.
****************************************************************************/
void NAPI DD_EnableDirectAccess(void)
{
    // Use the DDLOCK_NOSYSLOCK flag to avoid taking the Win16Mutex if we
    // can avoid it. This improves performance somewhat. Note that we don't
    // do this when we attempt to determine the actual surface
    if (DD_activeBuf) {
        if (DD_activeBuf->surfLock++ == 0) {
            while (IDirectDrawSurface7_Lock(DD_lpActive,NULL,&DD_lockBuf,LOCK_FLAGS | DDLOCK_NOSYSLOCK,NULL) == DDERR_SURFACELOST)
                PM_doSuspendApp();
            DD_activeBuf->p.Surface = (uchar*)DC.LinearMem + DD_activeBuf->p.Offset;
            }
        }
    else if (DD_lpActive) {
        while (IDirectDrawSurface7_Lock(DD_lpActive,NULL,&DD_lockBuf,DDLOCK_WAIT | DDLOCK_NOSYSLOCK,NULL) == DDERR_SURFACELOST)
            PM_doSuspendApp();
        }

}

/****************************************************************************
REMARKS:
Disables direct access to the currently active display surface for
software rendering etc.
****************************************************************************/
void NAPI DD_DisableDirectAccess(void)
{
    if (DD_activeBuf) {
        if (--DD_activeBuf->surfLock == 0) {
            while (IDirectDrawSurface7_Unlock(DD_lpActive,NULL) == DDERR_SURFACELOST)
                PM_doSuspendApp();
            DD_activeBuf->p.Surface = NULL;
            }
        if (DD_activeBuf->surfLock < 0)
            DD_activeBuf->surfLock = 0;
        }
    else if (DD_lpActive) {
        while (IDirectDrawSurface7_Unlock(DD_lpActive,NULL) == DDERR_SURFACELOST)
            PM_doSuspendApp();
        }
}

/****************************************************************************
PARAMETERS:
mix     - New mix code to set (GA_mixCodesType)

REMARKS:
This function sets the hardware mix operation for subsequent accelerated
rendering primitives.
****************************************************************************/
N_int32 NAPI DD_SetMix(
    N_int32 mix)
{
    DD_mix = mix;
    return true;
}

/****************************************************************************
PARAMETERS:
color   - New foreground color to set
****************************************************************************/
void NAPI DD_SetForeColor(
    GA_color color)
{
    DD_foreColor = color;
}

/****************************************************************************
PARAMETERS:
color   - New foreground color to set
****************************************************************************/
void NAPI DD_SetBackColor(
    GA_color color)
{
    DD_backColor = color;
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the rectangle to draw
top     - Top coordinate of the rectangle to draw
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines

REMARKS:
This function renders a solid rectangle at the specified location and
currently active color and mix.
****************************************************************************/
void NAPI DD_DrawRect(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height)
{
    RECT    dst;

    if (!DD_validPatRops[DD_mix]) {
        // Unsupported ROP so do it in software.
        DV.drawSW.DrawRect(left,top,width,height);
        return;
        }
    dst.left = left;
    dst.top = top;
    dst.right = left + width;
    dst.bottom = top + height;
    DD_bltFx.u5.dwFillColor = DD_foreColor;
    UNLOCK_SURFACE();
    if (DD_mix == GA_REPLACE_MIX) {
        while (IDirectDrawSurface7_Blt(DD_lpActive,&dst,NULL,NULL,
                DDBLT_COLORFILL | DDBLT_WAIT,&DD_bltFx) == DDERR_SURFACELOST)
            PM_doSuspendApp();
        }
    else {
        DD_bltFx.dwROP = DD_dwPatRopCodes[DD_mix];
        while (IDirectDrawSurface7_Blt(DD_lpActive,&dst,NULL,NULL,
                DDBLT_COLORFILL | DDBLT_ROP | DDBLT_WAIT,&DD_bltFx) == DDERR_SURFACELOST)
            PM_doSuspendApp();
        }
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the rectangle to draw
top     - Top coordinate of the rectangle to draw
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines

REMARKS:
This function renders a solid rectangle at the specified location and
currently active color and mix.
****************************************************************************/
void NAPI DD_DrawRectExt(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    GA_color color,
    N_int32 mix)
{
    RECT    dst;

    if (!DD_validPatRops[mix]) {
        // Unsupported ROP so do it in software.
        DV.drawSW.DrawRectExt(left,top,width,height,color,mix);
        return;
        }
    dst.left = left;
    dst.top = top;
    dst.right = left + width;
    dst.bottom = top + height;
    DD_bltFx.u5.dwFillColor = color;
    UNLOCK_SURFACE();
    if (mix == GA_REPLACE_MIX) {
        while (IDirectDrawSurface7_Blt(DD_lpActive,&dst,NULL,NULL,
                DDBLT_COLORFILL | DDBLT_WAIT,&DD_bltFx) == DDERR_SURFACELOST)
            PM_doSuspendApp();
        }
    else {
        DD_bltFx.dwROP = DD_dwPatRopCodes[mix];
        while (IDirectDrawSurface7_Blt(DD_lpActive,&dst,NULL,NULL,
                DDBLT_COLORFILL | DDBLT_ROP | DDBLT_WAIT,&DD_bltFx) == DDERR_SURFACELOST)
            PM_doSuspendApp();
        }
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
scans   - Pointer to an array of scanline data

REMARKS:
This function renders a list of solid scanlines starting at the specified
location in the currently active color and mix.
****************************************************************************/
void NAPI DD_DrawScanList(
    N_int32 y,
    N_int32 length,
    N_int16 *scans)
{
    while (length--) {
        int width;
        int x1 = *scans++;
        int x2 = *scans++;
        if ((width = (x2 - x1)) != 0) {
            if (width < 0) {
                x1 = x2;
                width = -width;
                }
            DD_DrawRect(x1,y,width,1);
            }
        y++;
        }
}

/****************************************************************************
PARAMETERS:
trap    - Pointer to the GA_trap structure describing the trapezoid

REMARKS:
This function renders a solid, flat topped and bottomed trapezoid in the
currently active color and mix.
****************************************************************************/
void NAPI DD_DrawTrap(
    GA_trap *trap)
{
    int     y,x1,x2,count = trap->count;

    if (!count)
        return;
    y = trap->y;
    x1 = trap->x1;
    x2 = trap->x2;
    trap->y += count;
    while (count--) {
        int width;
        int ix1 = (x1 + 0x8000) >> 16;
        int ix2 = (x2 + 0x8000) >> 16;
        if ((width = (ix2 - ix1)) != 0) {
            if (width < 0) {
                ix1 = ix2;
                width = -width;
                }
            DD_DrawRect(ix1,y,width,1);
            }
        x1 += trap->slope1;
        x2 += trap->slope2;
        y++;
        }
    trap->x1 = x1;
    trap->x2 = x2;
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the source rectangle to copy
top     - Top coordinate of the source rectangle to copy
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
dstLeft - Left coordinate of destination
dstTop  - Top coordinate of destination
mix     - Mix code for the copy (GA_mixCodesType)

REMARKS:
This function copies a rectangular region of video memory from one location
to another.
****************************************************************************/
void NAPI DD_BitBlt(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix)
{
    RECT    dst,src;

    if (!DD_validSrcRops[mix]) {
        // Unsupported ROP so do it in software.
        DV.drawSW.BitBlt(left,top,width,height,dstLeft,dstTop,mix);
        return;
        }
    src.left = left;
    src.top = top;
    src.right = left + width;
    src.bottom = top + height;
    UNLOCK_SURFACE();
    if (mix == GA_REPLACE_MIX && !DD_activeBuf->clipper) {
        while (IDirectDrawSurface7_BltFast(DD_lpActive,dstLeft,dstTop,DD_lpActive,
                &src,DDBLTFAST_NOCOLORKEY | DDBLTFAST_WAIT) == DDERR_SURFACELOST)
            PM_doSuspendApp();
        }
    else {
        dst.left = dstLeft;
        dst.top = dstTop;
        dst.right = dstLeft + width;
        dst.bottom = dstTop + height;
        DD_bltFx.dwROP = DD_dwSrcRopCodes[mix];
        while (IDirectDrawSurface7_Blt(DD_lpActive,&dst,DD_lpActive,
                &src,DDBLT_ROP | DDBLT_WAIT,&DD_bltFx) == DDERR_SURFACELOST)
            PM_doSuspendApp();
        }
}

/****************************************************************************
PARAMETERS:
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value

REMARKS:
This function copies a rectangular region of video memory from one location
to another with source transparency.
****************************************************************************/
void NAPI DD_SrcTransBlt(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    RECT        src,dst;
    DDCOLORKEY  ddck;

    if (mix != GA_REPLACE_MIX) {
        // ROP's generally don't work with transparent blits so do it in software.
        DV.drawSW.SrcTransBlt(left,top,width,height,dstLeft,dstTop,mix,transparent);
        return;
        }
    src.left = left;
    src.top = top;
    src.right = left + width;
    src.bottom = top + height;
    UNLOCK_SURFACE();
    if (DD_activeBuf->srcColorKey != transparent) {
        // We avoid re-loading the color key whenever possible
        ddck.dwColorSpaceLowValue  = transparent;
        ddck.dwColorSpaceHighValue = transparent;
        while (IDirectDrawSurface7_SetColorKey(DD_lpActive,DDCKEY_SRCBLT,&ddck) == DDERR_SURFACELOST)
            PM_doSuspendApp();
        DD_activeBuf->srcColorKey = transparent;
        }
    if (!DD_activeBuf->clipper) {
        while (IDirectDrawSurface7_BltFast(DD_lpActive,dstLeft,dstTop,DD_lpActive,
                &src,DDBLTFAST_SRCCOLORKEY | DDBLTFAST_WAIT) == DDERR_SURFACELOST)
            PM_doSuspendApp();
        }
    else {
        dst.left = dstLeft;
        dst.top = dstTop;
        dst.right = dstLeft + width;
        dst.bottom = dstTop + height;
        while (IDirectDrawSurface7_Blt(DD_lpActive,&dst,DD_lpActive,
                &src,DDBLT_KEYSRC | DDBLT_WAIT,NULL) == DDERR_SURFACELOST)
            PM_doSuspendApp();
        }
}

/****************************************************************************
PARAMETERS:
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value

REMARKS:
This function copies a rectangular region of video memory from one location
to another with destination transparency.
****************************************************************************/
void NAPI DD_DstTransBlt(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    RECT        src,dst;
    DDCOLORKEY  ddck;

    if (mix != GA_REPLACE_MIX) {
        // ROP's generally don't work with transparent blits so do it in software.
        DV.drawSW.DstTransBlt(left,top,width,height,dstLeft,dstTop,mix,transparent);
        return;
        }
    src.left = left;
    src.top = top;
    src.right = left + width;
    src.bottom = top + height;
    UNLOCK_SURFACE();
    if (DD_activeBuf->dstColorKey != transparent) {
        // We avoid re-loading the color key whenever possible
        ddck.dwColorSpaceLowValue  = transparent;
        ddck.dwColorSpaceHighValue = transparent;
        while (IDirectDrawSurface7_SetColorKey(DD_lpActive,DDCKEY_DESTBLT,&ddck) == DDERR_SURFACELOST)
            PM_doSuspendApp();
        DD_activeBuf->dstColorKey = transparent;
        }
    if (!DD_activeBuf->clipper) {
        while (IDirectDrawSurface7_BltFast(DD_lpActive,dstLeft,dstTop,DD_lpActive,
                &src,DDBLTFAST_DESTCOLORKEY | DDBLTFAST_WAIT) == DDERR_SURFACELOST)
            PM_doSuspendApp();
        }
    else {
        dst.left = dstLeft;
        dst.top = dstTop;
        dst.right = dstLeft + width;
        dst.bottom = dstTop + height;
        while (IDirectDrawSurface7_Blt(DD_lpActive,&dst,DD_lpActive,
                &src,DDBLT_KEYDEST | DDBLT_WAIT,NULL) == DDERR_SURFACELOST)
            PM_doSuspendApp();
        }
}

/****************************************************************************
PARAMETERS:
fx          - Effects to apply to the blit operation

REMARKS:
This function tests if the blit effect is supported in hardware.
****************************************************************************/
N_int32 DD_BitBltFxTestInt(
    GA_bltFx *fx,
    int *ipf)
{
    int i;

    // We always use the 3D engine, so if we do not have 3D we fail this
    if (!(DD_caps.dwCaps & DDCAPS_3D))
        return false;

    // No raster ops are supported via the 3D engine
    if (fx->Flags & gaBltMixEnable)
        return false;

    // We only support this for destination color depths of 8bpp and greater!
    if (DV.cntModeInfo.BitsPerPixel < 8)
        return false;

    // We do not support stretching (yet).
    if (fx->Flags & gaBltAnyStretch)
        return false;

    // Blending is only available in 15bpp and higher color depths
    if (fx->Flags & gaBltBlend) {
        if (fx->Flags & gaBltAnyColorKey)
            return false;
        if (DV.cntModeInfo.BitsPerPixel < 15)
            return false;
        }

    // TODO: Support 8-bit texture formats with color map for color conversion in HW!

    // Check that we can handle the pixel format conversion
    if (fx->Flags & gaBltConvert) {
CHK(    if (fx->BitsPerPixel > 8)
            CHECK(fx->PixelFormat); )
CHK(    if (fx->BitsPerPixel <= 8)
            CHECK(fx->SrcPalette); )
CHK(    if (DV.cntModeInfo.BitsPerPixel <= 8)
            CHECK(fx->DstPalette); )
CHK(    if (fx->Flags & gaBltTranslateVec)
            CHECK(fx->TranslateVec); )
        if (fx->BitsPerPixel < 8)
            return false;
        if (fx->BitsPerPixel == 8 || ((fx->BitsPerPixel == 16) && (fx->PixelFormat->AlphaPosition == 8)) || (fx->BitsPerPixel == 24 && !DD_textureFormats[pfRGB24])) {
            // These source bitmaps are always converted to 32-bit ARGB. If we are
            // not blending, there is no point converting to a 32-bit ARGB texture
            // as direct software will be faster.
            if (!(fx->Flags & gaBltBlend))
                return false;
            if (fx->BitsPerPixel == 8
                    && fx->SrcBlendFunc != gaBlendConstantColor
                    && fx->SrcBlendFunc != gaBlendOneMinusConstantColor
                    && fx->SrcBlendFunc != gaBlendConstantAlpha
                    && fx->SrcBlendFunc != gaBlendOneMinusConstantAlpha)
                return false;
            if (!DD_textureFormats[pfARGB32])
                return false;
            *ipf = pfARGB32;
            }
        else {
            // TODO: If doing blending with constant alpha for any surface that is
            //       15/16/24/32 bit, then we need to use a 32-bit ARGB texture and
            //       do manual color conversion ourselves. Need to create a test to
            //       test this in the MGL with all input bitmap formats and constant
            //       alpha applied.

            // Check if the pixel format is supported as a native texture by the hardware
            for (i = 0; i < DD_numPixelFormats; i++) {
                if (memcmp(&DD_pixelFormats[i],fx->PixelFormat,sizeof(DD_pixelFormats[i])) == 0)
                    break;
                }
            if (i == DD_numPixelFormats || !DD_textureFormats[i])
                return false;
            *ipf = i;
            }
        }
    else {
        *ipf = DD_iPrimaryPF;
        }
    return true;
}

/****************************************************************************
PARAMETERS:
fx          - Effects to apply to the blit operation

REMARKS:
This function tests if the blit effect is supported in hardware.
****************************************************************************/
N_int32 NAPI DD_BitBltFxTest(
    GA_bltFx *fx)
{
    int ipf;

    return DD_BitBltFxTestInt(fx,&ipf);
}

/****************************************************************************
PARAMETERS:
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstWidth    - Width of the destination rectangle in pixels
dstHeight   - Height of the destination rectangle in scanlines
fx          - Effects to apply to the blit operation

REMARKS:
This function copies a rectangular region of video memory from one location
to another with optional blit effects.
****************************************************************************/
void NAPI DD_BitBltFx(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    GA_bltFx *fx)
{
    int ipf;

    // Test if DirectDraw supports the corresponding effects blit caps in hardware,
    // and blit to/from the active draw surface if supported.
    if (DD_BitBltFxTestInt(fx,&ipf)) {
        // Note that the actual implementation is located in DD_BitBltBuf() in order
        // to special case all of the blit effects options with DirectDraw and Direct3D.
        DD_BitBltFxBuf(&DD_activeBuf->p,srcLeft,srcTop,srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,fx);
        }
    else {
        // If we don't support this in hardware with DirectDraw we punt to software.
        DV.drawSW.BitBltFx(srcLeft,srcTop,srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,fx);
        }
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstWidth    - Width of the destination rectangle in pixels
dstHeight   - Height of the destination rectangle in scanlines
fx          - GA_bltFx structure describing the requested effects
****************************************************************************/
void NAPI DD_BitBltFxSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    GA_bltFx *fx)
{
    GA_bufInt   buf;
    int         ipf;

    // Test if DirectDraw supports the corresponding effects blit caps in hardware,
    // and blit to/from the active draw surface if supported.
    if (DD_BitBltFxTestInt(fx,&ipf)) {
        // Create a fake buffer for the system memory surface
        memset(&buf,0,sizeof(buf));
        buf.p.dwSize = sizeof(GA_buf);
        buf.p.Width = srcWidth;
        buf.p.Height = srcHeight;
        buf.p.Flags = gaBufferSysMem;
        buf.p.Format = 0;
        if (fx->Flags & gaBltConvert) {
            buf.p.BitsPerPixel = fx->BitsPerPixel;
            buf.p.PixelFormat = *fx->PixelFormat;
            }
        else
            buf.p.BitsPerPixel = DV.cntModeInfo.BitsPerPixel;
        buf.p.Stride = buf.p.CacheStride = srcPitch;
        buf.p.Surface = buf.p.SurfaceCache = srcAddr;

        // Use DirectDraw surface as source for blit effects
        DD_BitBltFxBuf(&buf.p,srcLeft,srcTop,srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,fx);
        }
    else {
        // If we don't support this in hardware with DirectDraw we punt to software.
        DV.drawSW.BitBltFxSys(srcAddr,srcPitch,srcLeft,srcTop,srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,fx);
        }
}

