/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the generic SNAP Graphics
*               Architecture API functions on top of DirectDraw for Win32.
*
****************************************************************************/

#include "internal.h"
#include "snap/graphics/confver.h"
#include <clib/os/os.h>

/*--------------------------- Global variables ----------------------------*/

static GA_sharedInfo    localInfo = {0,-1};
char                    DRV_path[256];
GA_driver               DV;                     // Global device driver structure
GA_monitor              monitor[GA_MAX_HEADS];  // Global monitor record
LPDIRECTDRAW7           DD_lpDD = NULL;         // DirectDraw7 object
DDCAPS                  DD_caps;                // DirectDraw capabilities
DDSURFACEDESC2          DD_desktop;             // Original desktop display mode
LPDIRECTDRAWSURFACE7    DD_lpPrimary = NULL;
LPDIRECTDRAWSURFACE7    DD_lpActive = NULL;
LPDIRECTDRAWSURFACE7    DD_lpVisible = NULL;
LPDIRECTDRAWPALETTE     DD_lpPalette = NULL;
GA_bufInt               *DD_activeBuf = NULL;
DDSURFACEDESC2          DD_lockBuf;
DDBLTFX                 DD_bltFx;
ulong                   DD_totalMemory;
ulong                   DD_totalAGPMemory;
int                     DD_numModes;            // Number of DirectDraw modes
GA_mdEntry              DD_modeList[GA_MAX_MODES+1];
PM_MODULE               DD_hModNGAW32 = NULL;
PM_MODULE               DD_hModKERNEL32 = NULL;
PM_MODULE               DD_hModUSER32 = NULL;
PM_MODULE               DD_hModADVAPI32 = NULL;
GA_exports              gaExp;
GA_initFuncs            hwInit;
GA_driverFuncs          hwDriver;
ibool                   DD_isSDDActive = false;
LPDIRECT3D7             DD_lpD3D = NULL;        // Direct3D object for blit effects
LPDIRECT3DDEVICE7       DD_lpD3DDevice = NULL;  // Direct3D device interface
D3DDEVICEDESC7          DD_caps3d;              // Direct3D device caps
#ifdef  __DRIVER__
N_imports _VARAPI       _N_imports;             // Imports to OS specific functions
PM_imports _VARAPI      _PM_imports;            // Imports to PM library functions
DRV_imports _VARAPI     _DRV_imports;           // Imports to driver specific functions
#endif

// Function pointers to functions in KERNEL32.DLL, USER32.DLL and ADVAPI32.DLL

VOID (WINAPI *pSleep)(DWORD dwMilliseconds);
int  (WINAPI *pShowCursor)(BOOL bShow);
BOOL (WINAPI *pSetCursorPos)(int X,int Y);
BOOL (WINAPI *pGetCursorPos)(LPPOINT lpPoint);
BOOL (WINAPI *pPeekMessage)(LPMSG lpMsg,HWND hWnd,UINT wMsgFilterMin,UINT wMsgFilterMax,UINT wRemoveMsg);
BOOL (WINAPI *pTranslateMessage)(CONST MSG *lpMsg);
LONG (WINAPI *pDispatchMessage)(CONST MSG *lpMsg);
BOOL (WINAPI *pSetWindowPos)(HWND hWnd,HWND hWndInsertAfter,int X,int Y,int cx,int cy,UINT uFlags);
BOOL (WINAPI *pShowWindow)(HWND hWnd,int nCmdShow);
LONG (APIENTRY *pRegOpenKey)(HKEY hKey,LPCSTR lpSubKey,PHKEY phkResult);
LONG (APIENTRY *pRegOpenKeyEx)(HKEY hKey,LPCSTR lpSubKey,DWORD ulOptions,REGSAM samDesired,PHKEY phkResult);
LONG (APIENTRY *pRegQueryValueEx)(HKEY hKey,LPCSTR lpValueName,LPDWORD lpReserved,LPDWORD lpType,LPBYTE lpData,LPDWORD lpcbData);
LONG (APIENTRY *pRegSetValueEx)(HKEY hKey,LPCSTR lpValueName,DWORD Reserved,DWORD dwType,CONST BYTE* lpData,DWORD cbData);
LONG (APIENTRY *pRegCloseKey)(HKEY hKey);

// Names of manufacturer and chip revision id's.

char *DRV_manufacturerName = "Microsoft";

char *DRV_chipsetNames[] = {
    "DirectX",
    NULL,
    };

// Pre-defined pixel formats for DirectDraw modes

// TODO: We may need to extend this to support 8bpp textures in hardware, so
//       we can do 8bpp HW color conversion.

GA_pixelFormat DD_pixelFormats[] = {
    {0x1F,0x0A,3,0x1F,0x5,3,0x1F,0x0,3,0x01,0x0F,7},    // pfRGB555
    {0x1F,0x0B,3,0x3F,0x5,2,0x1F,0x0,3,0x00,0x00,0},    // pfRGB565
    {0xFF,0x10,0,0xFF,0x8,0,0xFF,0x0,0,0x00,0x00,0},    // pfRGB24
    {0xFF,0x0,0,0xFF,0x8,0,0xFF,0x10,0,0x00,0x00,0},    // pfBGR24
    {0xFF,0x10,0,0xFF,0x8,0,0xFF,0x0,0,0xFF,0x18,0},    // pfARGB32
    {0xFF,0x0,0,0xFF,0x8,0,0xFF,0x10,0,0xFF,0x18,0},    // pfABGR32
    {0xFF,0x18,0,0xFF,0x10,0,0xFF,0x8,0,0xFF,0x0,0},    // pfRGBA32
    {0xFF,0x8,0,0xFF,0x10,0,0xFF,0x18,0,0xFF,0x0,0},    // pfBGRA32
    };
int DD_pixelDepths[ARR_SIZE(DD_pixelFormats)] = {
    15,16,24,24,32,32,32,32
    };
LPDIRECTDRAWSURFACE7    DD_lpTexture[ARR_SIZE(DD_pixelFormats)] = {0};
DWORD                   DD_textureFormats[ARR_SIZE(DD_pixelFormats)];
int                     DD_textureSizeX[ARR_SIZE(DD_pixelFormats)];
int                     DD_textureSizeY[ARR_SIZE(DD_pixelFormats)];
int                     DD_numPixelFormats = ARR_SIZE(DD_pixelFormats);
int                     DD_iPrimaryPF = 0;

// Driver configuration structure

GA_configInfo       DRV_configInfo = {
    sizeof(GA_configInfo),
    CONFIG_SIGNATURE,
    CONFIG_SIGNATURE,
    CONFIG_SIGNATURE,
    CONFIG_SIGNATURE,
    DRIVER_VERSION,
    DRIVER_DATE,
    false,
    "",
    0,
    };

// Declare all DirectDraw GUID's as static variables within this module

#undef  DEFINE_GUID
#define DEFINE_GUID(name, l, w1, w2, b1, b2, b3, b4, b5, b6, b7, b8) \
        const GUID name \
                = { l, w1, w2, { b1, b2,  b3,  b4,  b5,  b6,  b7,  b8 } }
DEFINE_GUID( IID_IDirect3D7,            0xf5049e77,0x4861,0x11d2,0xa4,0x7,0x0,0xa0,0xc9,0x6,0x29,0xa8);
DEFINE_GUID( IID_IDirect3DDevice7,      0xf5049e79, 0x4861, 0x11d2, 0xa4, 0x7, 0x0, 0xa0, 0xc9, 0x6, 0x29, 0xa8);
DEFINE_GUID( IID_IDirect3DHALDevice,    0x84E63dE0,0x46AA,0x11CF,0x81,0x6F,0x00,0x00,0xC0,0x20,0x15,0x6E );
DEFINE_GUID( IID_IDirect3DTnLHalDevice, 0xf5049e78, 0x4861, 0x11d2, 0xa4, 0x7, 0x0, 0xa0, 0xc9, 0x6, 0x29, 0xa8);

#include "../certdata.c"

// Names of DLL's that we interface with

#define BPD_NAME        "graphics.bpd"
#define NGA_W32_NAME    "nga_w32.dll"

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
HEADER:
graphics/driver.h

PARAMETERS:
mfrName         - Place to store the manufacturer name
chipsetNames    - Place to store the list of chipset names

REMARKS:
Returns the chipset driver manufacturer name and list of supported chipsets.
****************************************************************************/
void _CEXPORT GA_getChipsetNames(
    char **mfr,
    char ***chipsetNames)
{
    *mfr = DRV_manufacturerName;
    *chipsetNames = DRV_chipsetNames;
}

/****************************************************************************
HEADER:
graphics/internal.h

PARAMETERS:
info    - Place to return the certification information

REMARKS:
Returns the chipset driver certification information structures.
****************************************************************************/
void _CEXPORT GA_getCertifyInfo(
    GA_certifyInfo *info)
{
    COPY_STRUCTURE(info,&DRV_certifyInfo);
}

#ifdef  __DRIVER__
/****************************************************************************
REMARKS:
Fatal error handler called when a non-imported function is called by the
driver. We leave this to a runtime error so that older applications and
shell drivers will work with newer bpd drivers provided no newer functions
are required by the driver itself. If they are, the application or shell
driver needs to be recompiled.
****************************************************************************/
static void _PM_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported PM_imports import function called! Please re-compile!\n");
}

/****************************************************************************
REMARKS:
Fatal error handler for non-imported N_imports.
****************************************************************************/
static void _N_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported N_imports import function called! Please re-compile!\n");
}

/****************************************************************************
REMARKS:
Fatal error handler for non-imported GA_imports.
****************************************************************************/
static void _DRV_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported DRV_imports import function called! Please re-compile!\n");
}
#endif

/****************************************************************************
PARAMETERS:
hKey        - Key to open registry key to query data from
szValue     - Value to get information for
value       - Place to store the registry key data read
size        - Size of the string buffer to read into

RETURNS:
True if the key was found, false if not.
****************************************************************************/
static ibool REG_queryStringEx(
    HKEY hKey,
    const char *szValue,
    char *value,
    ulong size)
{
    DWORD   type;

    if (pRegQueryValueEx(hKey,(PCHAR)szValue,(PDWORD)NULL,(PDWORD)&type,(PUCHAR)value,(PDWORD)&size) == ERROR_SUCCESS)
        return true;
    return false;
}

/****************************************************************************
PARAMETERS:
szKey       - Key to query (can contain version number formatting)
szValue     - Value to get information for
value       - Place to store the registry key data read
size        - Size of the string buffer to read into

RETURNS:
True if the key was found, false if not.
****************************************************************************/
static ibool REG_queryString(
    const char *szKey,
    const char *szValue,
    char *value,
    DWORD size)
{
    HKEY    hKey;
    ibool   status = false;

    if (value)
        memset(value,0,size);
    if (pRegOpenKey(HKEY_LOCAL_MACHINE,(PCHAR)szKey,&hKey) == ERROR_SUCCESS) {
        status = REG_queryStringEx(hKey,(PCHAR)szValue,value,size);
        pRegCloseKey(hKey);
        }
    return status;
}

/****************************************************************************
REMARKS:
Returns the version number of DirectX installed in the system.
****************************************************************************/
static int DD_getDXVersion(void)
{
    const char  *szDirectXKey = "SOFTWARE\\Microsoft\\DirectX";
    const char  *szDirectXVer = "Version";
    char        dxVerStr[256];
    int         dxVerMaj,dxVerMin;
    char        *p;

    if (REG_queryString(szDirectXKey,szDirectXVer,dxVerStr,sizeof(dxVerStr))) {
        // String is in the form 4.maj.min.XXXX
        p = strchr(dxVerStr,'.')+1;
        if (*p) {
            dxVerMaj = atoi(p);
            p = strchr(p,'.')+1;
            if (*p) {
                dxVerMin = atoi(p);
                return (dxVerMaj << 8) | dxVerMin;
                }
            }
        }
    return 0;
}

/****************************************************************************
REMARKS:
Special thunk to pass through hardware context instead of the
GDI device context to the hardware functions. The thunk itself is expecting
to see a copy of the original hardware context, not the GDI device context.
****************************************************************************/
static int NAPI DDC_initThunk(
    GA_devCtx *dc)
{
    return gaExp.DDC_init(dc->ring0DC);
}

/****************************************************************************
REMARKS:
Special thunk to pass through hardware context instead of the
GDI device context to the hardware functions. The thunk itself is expecting
to see a copy of the original hardware context, not the GDI device context.
****************************************************************************/
static int NAPI DDC_initExtThunk(
    GA_devCtx *dc,
    N_int32 channel)
{
    return gaExp.DDC_initExt(dc->ring0DC,channel);
}

/****************************************************************************
REMARKS:
Special thunk to pass through hardware context instead of the
GDI device context to the hardware functions. The thunk itself is expecting
to see a copy of the original hardware context, not the GDI device context.
****************************************************************************/
static ibool NAPI DDC_writeEDIDThunk(
    GA_devCtx *dc,
    N_int32 slaveAddr,
    uchar *edid,
    N_int32 length,
    N_int32 blockNumber,
    N_int32 channel)
{
    return gaExp.DDC_writeEDID(dc->ring0DC,slaveAddr,edid,length,blockNumber,channel);
}

/****************************************************************************
REMARKS:
Special thunk to pass through hardware context instead of the
GDI device context to the hardware functions. The thunk itself is expecting
to see a copy of the original hardware context, not the GDI device context.
****************************************************************************/
static int NAPI MCS_beginThunk(
    GA_devCtx *dc)
{
    return gaExp.MCS_begin(dc->ring0DC);
}

/****************************************************************************
REMARKS:
Special thunk to pass through hardware context instead of the
GDI device context to the hardware functions. The thunk itself is expecting
to see a copy of the original hardware context, not the GDI device context.
****************************************************************************/
static int NAPI MCS_beginExtThunk(
    GA_devCtx *dc,
    N_int32 channel)
{
    return gaExp.MCS_beginExt(dc->ring0DC,channel);
}

/****************************************************************************
REMARKS:
Special thunk to pass through hardware context instead of the
GDI device context to the hardware functions. The thunk itself is expecting
to see a copy of the original hardware context, not the GDI device context.
****************************************************************************/
static int NAPI GA_detectPnPMonitorThunk(
    GA_devCtx *dc,
    N_int32 outputHead,
    GA_monitor *monitor,
    ibool *hasChanged)
{
    return gaExp.GA_detectPnPMonitor(dc->ring0DC,outputHead,monitor,hasChanged);
}

/****************************************************************************
REMARKS:
Special thunk to pass through hardware context instead of the
GDI device context to the hardware functions. The thunk itself is expecting
to see a copy of the original hardware context, not the GDI device context.
****************************************************************************/
static int NAPI GA_getParsedEDIDThunk(
    GA_devCtx *dc,
    EDID_record *rec)
{
    return gaExp.GA_getParsedEDID(dc->ring0DC,rec);
}

/****************************************************************************
REMARKS:
This function attempts to connect to and load the system shared SNAP
driver if possible. This is used to pass through important chip specific
functions via the DirectX driver up to ring 3 apps, as well as optimise
drawing in fullscreen DirectX modes when SDD is the active display driver.
****************************************************************************/
static ibool GA_loadHWDriver(
    GA_exports *exportTable)
{
    char            filename[PM_MAX_PATH];
    char            bpdpath[PM_MAX_PATH];
    GA_exports *    exp;
    GA_exports *    (NAPIP _GA_getSystemGAExports)(void);

    // TODO: Ignore the hardware driver if it is the VGA or VBE driver!!!

    // Find the path to the BPD file. Note that we do this first
    // *before* calling PM_findBPD() for the nga_w32.dll library
    // because PM_findBPD() will cache the first directory it finds
    // for finding subsequent files. It is important that we always
    // use the directory where the graphics.bpd file we need is
    // located, so we must always search for that file first.
    if (!PM_findBPD(BPD_NAME,bpdpath))
        return false;

    // Open the shared DLL file to see if we can load the system SNAP driver
    if (!PM_findBPD(NGA_W32_NAME,bpdpath))
        return false;
    strcpy(filename,bpdpath);
    strcat(filename,NGA_W32_NAME);
    if ((DD_hModNGAW32 = PM_loadLibrary(filename)) == NULL) {
        _log("Unable to load NGA_W32.DLL!\n");
        return false;
        }
    if ((_GA_getSystemGAExports = PM_getProcAddress(DD_hModNGAW32,"_GA_getSystemGAExports")) == NULL) {
        _log("Unable to get function exports from NGA_W32.DLL!\n");
        return false;
        }
    if ((exp = _GA_getSystemGAExports()) == NULL)
        return false;
    memcpy(&gaExp,exp,MIN(sizeof(gaExp),exp->dwSize));

    // Now try to load the system SNAP driver
    if ((DC.ring0DC = gaExp.GA_loadDriver(0,true)) == NULL) {
        PM_freeLibrary(DD_hModNGAW32);
        DD_hModNGAW32 = NULL;
        return false;
        }

    // Update the bus type for the device
    DC.BusType = DC.ring0DC->BusType;

    // If we have a valid hardware SNAP driver, then we need to pass
    // all the DDC and MCS functions directly through to the ring 0
    // driver. We need to avoid the overheads of toggling the SCL/SDA
    // clock and data lines via ring 0 thunk calls.
    if (exportTable) {
        exportTable->DDC_init                   = DDC_initThunk;
        exportTable->DDC_readEDID               = gaExp.DDC_readEDID;
        exportTable->MCS_begin                  = MCS_beginThunk;
        exportTable->MCS_getCapabilitiesString  = gaExp.MCS_getCapabilitiesString;
        exportTable->MCS_isControlSupported     = gaExp.MCS_isControlSupported;
        exportTable->MCS_enableControl          = gaExp.MCS_enableControl;
        exportTable->MCS_getControlMax          = gaExp.MCS_getControlMax;
        exportTable->MCS_getControlValue        = gaExp.MCS_getControlValue;
        exportTable->MCS_getControlValues       = gaExp.MCS_getControlValues;
        exportTable->MCS_setControlValue        = gaExp.MCS_setControlValue;
        exportTable->MCS_setControlValues       = gaExp.MCS_setControlValues;
        exportTable->MCS_resetControl           = gaExp.MCS_resetControl;
        exportTable->MCS_saveCurrentSettings    = gaExp.MCS_saveCurrentSettings;
        exportTable->MCS_getTimingReport        = gaExp.MCS_getTimingReport;
        exportTable->MCS_getSelfTestReport      = gaExp.MCS_getSelfTestReport;
        exportTable->MCS_end                    = gaExp.MCS_end;
        exportTable->DDC_writeEDID              = DDC_writeEDIDThunk;
        exportTable->DDC_initExt                = DDC_initExtThunk;
        exportTable->MCS_beginExt               = MCS_beginExtThunk;
        exportTable->GA_detectPnPMonitor        = GA_detectPnPMonitorThunk;
        exportTable->GA_getParsedEDID           = GA_getParsedEDIDThunk;
        }

    // Check if SDD is the active display driver and we have a linear
    // framebuffer available.
    DD_isSDDActive = PM_isSDDActive();
    if (!(DC.ring0DC->Attributes & gaHaveLinearBuffer))
        DD_isSDDActive = false;

    // Initialise access to the hardware device driver
    hwInit.dwSize = sizeof(hwInit);
    if (!gaExp.GA_queryFunctions(DC.ring0DC,GA_GET_INITFUNCS,&hwInit)) {
        _log("Unable to get hardware device driver functions!\n");
        return false;
        }
    hwDriver.dwSize = sizeof(hwDriver);
    if (!gaExp.GA_queryFunctions(DC.ring0DC,GA_GET_DRIVERFUNCS,&hwDriver)) {
        _log("Unable to get hardware device driver functions!\n");
        return false;
        }
    return true;
}

/****************************************************************************
REMARKS:
Clears the message queue of all waiting messages.
****************************************************************************/
void DD_clearMessageQueue(void)
{
    MSG msg;

    pSleep(100);
    while (pPeekMessage(&msg,NULL,0,0,PM_REMOVE)) {
        pTranslateMessage(&msg);
        pDispatchMessage(&msg);
        }
    pSleep(100);
}

/****************************************************************************
REMARKS:
Utility function to load DirectDraw and get the caps for the driver.
****************************************************************************/
static ibool DD_loadDDrawAndGetCaps(
    DDCAPS  *ddCaps)
{
    int i;

    // Unload DirectDraw if already loaded
    for (i = 0; i < DD_numPixelFormats; i++)
        SAFE_RELEASE(DD_lpTexture[i]);
    SAFE_RELEASE(DD_lpD3DDevice);
    SAFE_RELEASE(DD_lpD3D);
    SAFE_RELEASE(DD_lpDD);
    PM_unloadDirectDraw(DC.DeviceIndex);

    // Load DirectDraw and get the DirectDraw7 compatible interfaces
    if ((DD_lpDD = PM_loadDirectDraw(DC.DeviceIndex)) == NULL) {
        _log("Unable to load DirectDraw!");
        return false;
        }
    IDirectDraw7_AddRef(DD_lpDD);
    ddCaps->dwSize = sizeof(DDCAPS);
    ddCaps->dwCaps = 0;
    IDirectDraw7_GetCaps(DD_lpDD,ddCaps,NULL);

    // Load Direct3D interface too for blit effects options, if available
    if (ddCaps->dwCaps & DDCAPS_3D) {
        IDirectDraw7_QueryInterface(DD_lpDD, &IID_IDirect3D7, &DD_lpD3D);
        if (DD_lpD3D == NULL) {
            ddCaps->dwCaps &= ~DDCAPS_3D;
            _log("Direct3D unavailable for use with DirectDraw");
            }
        }
    return true;
}

/****************************************************************************
PARAMETERS:
s   - Pointer to surface description for this mode

RETURNS:
Number of bytes of memory contained in the surface

REMARKS:
Determines the pixel format and pixel depth for a DirectDraw surface
****************************************************************************/
ibool DD_GetPixelFormat(
    DDPIXELFORMAT *ddpfPixelFormat,
    int *bits,
    int *pixelFormat)
{
    // Find the pixel depth and pixel format for the mode
    if ((ddpfPixelFormat->dwFlags & DDPF_PALETTEINDEXED8)) {
        *bits = 8;
        *pixelFormat = pfRGB24;
        return true;
        }
    else if (ddpfPixelFormat->dwFlags & DDPF_RGB) {
        switch (ddpfPixelFormat->u1.dwRGBBitCount) {
            case 16:
                if (ddpfPixelFormat->u2.dwRBitMask == 0xF800) {
                    *bits = 16;
                    *pixelFormat = pfRGB565;
                    return true;
                    }
                else if (ddpfPixelFormat->u2.dwRBitMask == 0x7C00) {
                    *bits = 15;
                    *pixelFormat = pfRGB555;
                    return true;
                    }
                break;
            case 24:
                *bits = 24;
                if (ddpfPixelFormat->u2.dwRBitMask == 0x0000FF) {
                    *pixelFormat = pfBGR24;
                    return true;
                    }
                else if (ddpfPixelFormat->u2.dwRBitMask == 0xFF0000) {
                    *pixelFormat = pfRGB24;
                    return true;
                    }
                break;
            case 32:
                *bits = 32;
                if (ddpfPixelFormat->u2.dwRBitMask == 0x000000FFUL) {
                    *pixelFormat = pfABGR32;
                    return true;
                    }
                else if (ddpfPixelFormat->u2.dwRBitMask == 0xFF000000UL) {
                    *pixelFormat = pfRGBA32;
                    return true;
                    }
                else if (ddpfPixelFormat->u2.dwRBitMask == 0x0000FF00UL) {
                    *pixelFormat = pfBGRA32;
                    return true;
                    }
                else if (ddpfPixelFormat->u2.dwRBitMask == 0x00FF0000UL) {
                    *pixelFormat = pfARGB32;
                    return true;
                    }
                break;
            }
        }
    return false;
}

/****************************************************************************
PARAMETERS:
s   - Pointer to surface description for this mode

RETURNS:
Number of bytes of memory contained in the surface

REMARKS:
Determines the size of a particular DirectDraw surface in bytes.
****************************************************************************/
static long GetSurfaceSize(
    LPDDSURFACEDESC2 s)
{
    return s->dwWidth * s->dwHeight * (s->u4.ddpfPixelFormat.u1.dwRGBBitCount / 8);
}

/****************************************************************************
PARAMETERS:
s   - Pointer to surface description for this mode
lpv - Unused callback pointer

REMARKS:
Display mode enumeration callback, called for every display mode that
DirectDraw supports. We save the information for the display mode into our
table for later retrieval. The display mode enumeration process is time
consuming, so we don't want to run through this every time we need to
check for the availability of a specific mode.
****************************************************************************/
static HRESULT CALLBACK ModeEnumCallBack(
    LPDDSURFACEDESC2 s,
    LPVOID lpv)
{
    GA_mdEntry  *md;
    int         i,j,minPitch,bits,pixelFormat;
    ulong       size;

    // Ensure we don't go off the end of our list
    if (DD_numModes >= GA_MAX_MODES)
        return DDENUMRET_CANCEL;

    // Ignore ModeX surfaces
    if (s->ddsCaps.dwCaps & DDSCAPS_MODEX)
        return DDENUMRET_OK;

    // Find the pixel depth and pixel format for the mode
    DD_GetPixelFormat(&s->u4.ddpfPixelFormat,&bits,&pixelFormat);

    // Check to see if the mode has already been enumerated, if so simply
    // add the new refresh rate to the list and exit. Note that the desktop
    // mode is always the first mode, so we skip over that mode.
    for (i = 1; i < DD_numModes; i++) {
        md = &DD_modeList[i];
        if (md->xres == s->dwWidth && md->yres == s->dwHeight && md->bits == bits
                && md->pixelFormat == pixelFormat) {
            // Add the new refresh rate into the list
            for (j = 0; j < MAX_REFRESH_RATES; j++) {
                if (md->RefreshList[j] == -1) {
                    md->RefreshList[j] = s->u2.dwRefreshRate;
                    md->RefreshList[j+1] = -1;
                    break;
                    }
                }
            return DDENUMRET_OK;
            }
        }

    // Save the resolution and pitch of the mode
    md = &DD_modeList[DD_numModes++];
    md->xres = s->dwWidth;
    md->yres = s->dwHeight;
    md->bits = bits;
    md->pixelFormat = pixelFormat;
    md->bytesPerLine = s->u1.lPitch;
    md->dwCaps = s->ddsCaps.dwCaps;
    if (s->u2.dwRefreshRate == 0)
        md->RefreshList[0] = -1;
    else {
        md->RefreshList[0] = s->u2.dwRefreshRate;
        md->RefreshList[1] = -1;
        }

    // Hack for DirectDraw on NT 4.0 that returns invalid pitch values
    minPitch = md->xres * (s->u4.ddpfPixelFormat.u1.dwRGBBitCount / 8);
    while (md->bytesPerLine < minPitch)
        md->bytesPerLine += s->u1.lPitch;

    // Now determine the memory size. Because DirectDraw fucks up
    // the amount of available video memory (under NT it reports
    // 0!), we check the list of available modes and find the
    // video memory size that is required to hold the largest
    // available display mode and up the memory size to that if it
    // is lower.
    size = GetSurfaceSize(s);
    if (size > DC.TotalMemory)
        DC.TotalMemory = size;
    return DDENUMRET_OK;
}

/****************************************************************************
PARAMETERS:
elem1   - First element to sort
elem2   - Second selement to sort

RETURNS:
< 0 if elem1 less than elem2, 0 if equal, > 0 if elem2 greater than elem2

REMARKS:
Quicksort compare function to sort the list of available display modes
as many DirectDraw drivers report these modes is a completely adhoc fashion.
****************************************************************************/
static int cmp_modes(
    const void *elem1,
    const void *elem2)
{
    const GA_mdEntry *md1 = elem1;
    const GA_mdEntry *md2 = elem2;

    if (md1->bits < md2->bits)
        return -1;
    if (md1->bits == md2->bits) {
        if (md1->xres < md2->xres)
            return -1;
        if (md1->xres == md2->xres) {
            if (md1->yres < md2->yres)
                return -1;
            if (md1->yres == md2->yres)
                return 0;
            }
        }
    return 1;
}

/****************************************************************************
HEADER:
graphics/driver.h

PARAMETERS:
path        - Path to Binary Portable DLL
deviceIndex - Index of the display device in use
info        - Pointer to the PCI configuration space info for the device
nImp        - Generic SNAP imports
drvImp      - Driver specific imports
af          - Pointer to optional VBE/AF driver

RETURNS:
Pointer to exported function list

REMARKS:
Does generic Plug and Play device detection and fills in the device resource
allocation information for the device so that the driver loader can
allocate the necessary OS resources for the device before calling the
real detection code.
****************************************************************************/
GA_devCtx * _CEXPORT GA_initDriver(
    ibool shared,
    const char *path,
    int deviceIndex,
    PCIDeviceInfo *info,
    GA_monitor *pMonitor,
    N_imports *nImp,
    PM_imports *pmImp,
    DRV_imports *drvImp,
    GA_options *licensedOptions,
    GA_globalOptions *globalOpt,
    ibool sharedIRQ,
    GA_exports *exportTable,
    GA_sharedInfo *sharedInfo,
    int loadInGUI)
{
    DDSCAPS2        dds;
    DWORD           freeMem;
    int             i,dxVer,bits,pixelFormat;
#ifdef  __DRIVER__
    int             max;
    ulong           *p;

    // Initialize all default imports to point to fatal error handler
    // for upwards compatibility.
    max = sizeof(_N_imports)/sizeof(GA_initLibrary_t);
    for (i = 0,p = (ulong*)&_N_imports; i < max; i++)
        *p++ = (ulong)_N_fatalErrorHandler;
    max = sizeof(_PM_imports)/sizeof(GA_initLibrary_t);
    for (i = 0,p = (ulong*)&_PM_imports; i < max; i++)
        *p++ = (ulong)_PM_fatalErrorHandler;
    max = sizeof(_DRV_imports)/sizeof(GA_initLibrary_t);
    for (i = 0,p = (ulong*)&_DRV_imports; i < max; i++)
        *p++ = (ulong)_DRV_fatalErrorHandler;

    // Now copy all our imported functions
    memcpy(&_N_imports,nImp,MIN(sizeof(_N_imports),nImp->dwSize));
    memcpy(&_PM_imports,pmImp,MIN(sizeof(_PM_imports),pmImp->dwSize));
    memcpy(&_DRV_imports,drvImp,MIN(sizeof(_DRV_imports),drvImp->dwSize));
#endif
    monitor[0] = *pMonitor;
    strcpy(DRV_path,path);

    // Connect to necessary USER32 and ADVAPI32 functions
    if ((DD_hModKERNEL32 = PM_loadLibrary("KERNEL32.DLL")) == NULL) {
        _log("Unable to connect to KERNEL32.DLL!");
        goto Error;
        }
    if ((DD_hModUSER32 = PM_loadLibrary("USER32.DLL")) == NULL) {
        _log("Unable to connect to USER32.DLL!");
        goto Error;
        }
    if ((DD_hModADVAPI32 = PM_loadLibrary("ADVAPI32.DLL")) == NULL) {
        _log("Unable to connect to ADVAPI32.DLL!");
        goto Error;
        }
    if ((pSleep = PM_getProcAddress(DD_hModKERNEL32,"Sleep")) == NULL)
        goto Error;
    if ((pShowCursor = PM_getProcAddress(DD_hModUSER32,"ShowCursor")) == NULL)
        goto Error;
    if ((pSetCursorPos = PM_getProcAddress(DD_hModUSER32,"SetCursorPos")) == NULL)
        goto Error;
    if ((pGetCursorPos = PM_getProcAddress(DD_hModUSER32,"GetCursorPos")) == NULL)
        goto Error;
    if ((pPeekMessage = PM_getProcAddress(DD_hModUSER32,"PeekMessageA")) == NULL)
        goto Error;
    if ((pTranslateMessage = PM_getProcAddress(DD_hModUSER32,"TranslateMessage")) == NULL)
        goto Error;
    if ((pDispatchMessage = PM_getProcAddress(DD_hModUSER32,"DispatchMessageA")) == NULL)
        goto Error;
    if ((pSetWindowPos = PM_getProcAddress(DD_hModUSER32,"SetWindowPos")) == NULL)
        goto Error;
    if ((pShowWindow = PM_getProcAddress(DD_hModUSER32,"ShowWindow")) == NULL)
        goto Error;
    if ((pRegOpenKey = PM_getProcAddress(DD_hModADVAPI32,"RegOpenKeyA")) == NULL)
        goto Error;
    if ((pRegOpenKeyEx = PM_getProcAddress(DD_hModADVAPI32,"RegOpenKeyExA")) == NULL)
        goto Error;
    if ((pRegQueryValueEx = PM_getProcAddress(DD_hModADVAPI32,"RegQueryValueExA")) == NULL)
        goto Error;
    if ((pRegSetValueEx = PM_getProcAddress(DD_hModADVAPI32,"RegSetValueExA")) == NULL)
        goto Error;
    if ((pRegCloseKey = PM_getProcAddress(DD_hModADVAPI32,"RegCloseKey")) == NULL)
        goto Error;

    // Check that a valid version of DirectDraw is found
    dxVer = DD_getDXVersion();
    if (dxVer == 0)
        _log("No DirectX installed on the system!\n");
    else
        _log("DirectX Version %d.%d found!\n", dxVer >> 8, dxVer & 0xFF);
    if (dxVer < 0x700) {
        _log("SciTech SNAP Graphics requires DirectX 7 to be installed.\n");
        goto Error;
        }

    // Initialise default information
    memset(&DV,0,sizeof(DV));
    if (info)
        memcpy(&DV.PCIConfig,info,sizeof(DV.PCIConfig));
    strcpy(DC.Signature,GA_SIGNATURE);
    DC.Version = GA_VERSION;
    DC.DriverRev = (GA_RELEASE_MAJOR << 8) | GA_RELEASE_MINOR;
    DC.DeviceIndex = deviceIndex;
    strcpy(DC.OemVendorName,SCITECH_STR);
    strcpy(DC.OemCopyright,COPYRIGHT_STR);
    DV.s = &localInfo;
    if (sharedInfo)
        DV.s = sharedInfo;

    // Attempt to load the system SNAP driver
    GA_loadHWDriver(exportTable);

    // Load DirectDraw and get the capabilities.
    if (!DD_loadDDrawAndGetCaps(&DD_caps))
        goto Error;
    if (DD_caps.dwCaps & DDCAPS_BANKSWITCHED) {
        _log("SciTech SNAP Graphics requires a hardware linear framebuffer");
        goto Error;
        }

    // Now configure the global header based on the DirectDraw driver.
    DC.AvailableModes = (void*)&DV.AvailableModes;
    DC.Attributes = gaHaveDisplayStart | gaHaveLinearBuffer | gaHave8BitDAC | gaHaveNonVGAMode;
    if (DD_caps.dwCaps & DDCAPS_BLT)
        DC.Attributes |= gaHaveAccel2D;
    if (DD_caps.dwCaps & DDCAPS_3D)
        DC.Attributes |= gaHaveAccel3D;
    if (DD_caps.dwCaps & DDCAPS_OVERLAY)
        DC.Attributes |= gaHaveAccelVideo;
    if (DD_caps.dwCaps2 & DDCAPS2_FLIPNOVSYNC)
        DC.Attributes |= gaHaveTripleBuffer;
    DC.TextSize = 0;
    DC.TextBasePtr = 0;
    DC.BankSize = 0;
    DC.BankedBasePtr = 0;

    // We store the desktop mode as mode 0x03 in the mode list and make
    // it the default mode when the driver is loaded. Calling SetVideoMode
    // with a mode number of 0x03 will cause the desktop mode to be restored.
    DD_desktop.dwSize = sizeof(DD_desktop);
    IDirectDraw7_GetDisplayMode(DD_lpDD,&DD_desktop);
    DV.cntModeNum = 0x100;
    DV.AvailableModes[0] = 0x100;
    DD_modeList[0].xres = DD_desktop.dwWidth;
    DD_modeList[0].yres = DD_desktop.dwHeight;
    DD_modeList[0].bytesPerLine = DD_desktop.u1.lPitch;
    DD_modeList[0].dwCaps = DD_desktop.ddsCaps.dwCaps;
    DD_modeList[0].RefreshList[0] = -1;
    DD_GetPixelFormat(&DD_desktop.u4.ddpfPixelFormat,&bits,&pixelFormat);
    DD_modeList[0].bits = bits;
    DD_modeList[0].pixelFormat = pixelFormat;

    // Enumerate all available display modes and estimate the amount
    // of video memory on board based on the available modes. Note that
    // we only enumerate refresh rates if SDD is not active, since when SDD is
    // active we get the refresh rates directly from the driver.
    DD_numModes = 1;
    if (IDirectDraw7_EnumDisplayModes(DD_lpDD,DDEDM_REFRESHRATES,NULL,NULL,ModeEnumCallBack) != DD_OK)
        goto Error;

    // Sort all the available modes by increasing resolution
    qsort(&DD_modeList[1],DD_numModes-1,sizeof(DD_modeList[0]),cmp_modes);

    // Build the list of available modes
    for (i = 0; i < DD_numModes; i++)
        DV.AvailableModes[i] = 0x100 + i;
    DV.AvailableModes[i] = 0xFFFF;

    // Note that DirectDraw drivers only ever report how much memory is
    // available sans any internal buffers plus the primary surface. Hence
    // to get the actual amount of memory available we use the totalMem
    // value and add on the size of the primary surface.
    memset(&dds,0,sizeof(dds));
    dds.dwCaps = DDSCAPS_LOCALVIDMEM;
    IDirectDraw7_GetAvailableVidMem(DD_lpDD,&dds,&DD_totalMemory,&freeMem);
    dds.dwCaps = DDSCAPS_NONLOCALVIDMEM;
    IDirectDraw7_GetAvailableVidMem(DD_lpDD,&dds,&DD_totalAGPMemory,&freeMem);

    // Round the total memory size up to nearest Mb power of 2 mark
    DC.TotalMemory = DD_totalMemory + GetSurfaceSize(&DD_desktop);
    DC.TotalMemory /= 1024;
    if (DC.TotalMemory > 32768)
        DC.TotalMemory = 65536;
    else if (DC.TotalMemory > 16384)
        DC.TotalMemory = 32768;
    else if (DC.TotalMemory > 8192)
        DC.TotalMemory = 16384;
    else if (DC.TotalMemory > 4096)
        DC.TotalMemory = 8192;
    else if (DC.TotalMemory > 2048)
        DC.TotalMemory = 4096;
    else if (DC.TotalMemory > 1024)
        DC.TotalMemory = 2048;
    else if (DC.TotalMemory > 512)
        DC.TotalMemory = 1024;

    // Initialise list of licensed options
    if (licensedOptions)
        DV.licensed = *licensedOptions;
    else {
        rep_stosb(&DV.licensed,sizeof(DV.licensed),1);
        DV.licensed.dwSize = sizeof(DV.licensed);
        DV.licensed.resolutions[0] = 0xFFFFFFFFU;
        DV.licensed.colorDepths[0] = 0xFF;
        DV.licensed.maxRefresh = 0xFF;
        }
    DV.globalOpt = *globalOpt;

    // Pre-setup structures to avoid doing this constantly in the code
    memset(&DD_lockBuf,0,sizeof(DD_lockBuf));
    DD_lockBuf.dwSize = sizeof(DD_lockBuf);
    memset(&DD_bltFx,0,sizeof(DD_bltFx));
    DD_bltFx.dwSize = sizeof(DD_bltFx);

    // Fill in the driver init function pointers
    DC.loader.InitDriver = DRV_initDriver;
    DC.loader.QueryFunctions = DRV_queryFunctions;
    DC.loader.UnloadDriver = DRV_unloadDriver;
    return &DC;

Error:
    if (DD_hModUSER32)
        PM_freeLibrary(DD_hModUSER32);
    if (DD_hModADVAPI32)
        PM_freeLibrary(DD_hModADVAPI32);
    SAFE_RELEASE(DD_lpD3DDevice);
    SAFE_RELEASE(DD_lpD3D);
    SAFE_RELEASE(DD_lpDD);
    PM_unloadDirectDraw(DC.DeviceIndex);
    return NULL;
}

/****************************************************************************
REMARKS:
Does internal initialisation to initialise the hardware driver functions
that are valid to be called from within the SNAP drivers when the
DirectDraw driver is loaded.
****************************************************************************/
static void DRV_initHWDriver(void)
{
    GA_DPMSFuncs    dpms;
    GA_SCIFuncs     sci;

    // If SDD is the active display driver, we can directly call
    // the SNAP driver and drawing functions once the DirectDraw
    // mode has been initialised. Hence we override the necessary
    // functions in here.
    if (DD_isSDDActive) {
        // TODO!!
#if 0
        // Translate the attributes flags for the driver
        DC.Attributes = DC.ring0DC->Attributes;

        // Pass through all the driver and cursor functions from the chip driver
        gaExp.GA_queryFunctions(DC.ring0DC,GA_GET_DRIVERFUNCS,&DV.driver);
        gaExp.GA_queryFunctions(DC.ring0DC,GA_GET_CURSORFUNCS,&DV.cursor);

        // Disable the buffer manager as we will be using the SNAP
        // ref2d buffer manager instead, direct to the hardware.

        // TODO: Actually we need to pass all this through to the SNAP
        //       system driver buffer manager, as we need to ensure that we
        //       are properly sharing buffers with DirectDraw!
        CLEAR_STRUCTURE(&DV.buf);
#endif
        }

    // Always pass through the DPMS and DDC functions from the chip driver
    dpms.dwSize = sizeof(dpms);
    if (gaExp.GA_queryFunctions(DC.ring0DC,GA_GET_DPMSFUNCS,&dpms))
        DV.dpms = dpms;
    sci.dwSize = sizeof(sci);
    if (gaExp.GA_queryFunctions(DC.ring0DC,GA_GET_SCIFUNCS,&sci))
        DV.sci = sci;
}

/****************************************************************************
DESCRIPTION:
Detects that the device is installed and configures the driver.

HEADER:
graphics/driver.h

RETURNS:
True if hardware detection correctly, false if not.

REMARKS:
Does generic device detection and fills in the generic components of the
SNAP device context header information after calling the chipset
specific detection code.
****************************************************************************/
N_int32 NAPI DRV_initDriver(void)
{
    FILE                *f;
    char                filename[256];
    int                 dxVer;
    DDDEVICEIDENTIFIER2 dddi;

#define INIT_STRUCTURE(p)       \
    memset(&p,0xFF,sizeof(p));  \
    p.dwSize = sizeof(p);

    // Set all device specific vectors to NOT_IMPLEMENTED by default
    INIT_STRUCTURE(DV.init);
    INIT_STRUCTURE(DV.driver);
    INIT_STRUCTURE(DV.cursor);
    INIT_STRUCTURE(DV.video);
    INIT_STRUCTURE(DV.dpms);
    INIT_STRUCTURE(DV.sci);
    INIT_STRUCTURE(DV.vbe);
    INIT_STRUCTURE(DV.buf);
    INIT_STRUCTURE(DV.clip);
    INIT_STRUCTURE(DV.state2d);
    INIT_STRUCTURE(DV.draw_8bpp);
    INIT_STRUCTURE(DV.draw_16bpp);
    INIT_STRUCTURE(DV.draw_24bpp);
    INIT_STRUCTURE(DV.draw_32bpp);
    INIT_STRUCTURE(DV.drawSW);

#undef GetMonitorInfo

    // Configure all the generic functions for the driver.
    DV.init.GetConfigInfo               = DRV_getConfigInfo;
    DV.init.GetOptions                  = DRV_getOptions;
    DV.init.SetOptions                  = DRV_setOptions;
    DV.init.SetModeProfile              = DRV_setModeProfile;
    DV.init.GetVideoModeInfo            = DD_GetVideoModeInfo;
    DV.init.SetVideoMode                = DD_SetVideoMode;
    DV.init.GetVideoMode                = DD_GetVideoMode;
    DV.init.GetCustomVideoModeInfo      = DD_GetCustomVideoModeInfo;
    DV.init.SetCustomVideoMode          = DD_SetCustomVideoMode;
    DV.init.GetClosestPixelClock        = DD_GetClosestPixelClock;
    DV.init.GetCRTCTimings              = DD_GetCRTCTimings;
    DV.init.SetCRTCTimings              = DD_SetCRTCTimings;
    DV.init.SaveCRTCTimings             = DD_SaveCRTCTimings;
    DV.init.SetGlobalRefresh            = DD_SetGlobalRefresh;
    DV.init.SetDisplayOutput            = DD_SetDisplayOutput;
    DV.init.GetDisplayOutput            = DD_GetDisplayOutput;
    DV.init.SetSoftwareRenderFuncs      = DRV_SetSoftwareRenderFuncs;
    DV.init.GetUniqueFilename           = DD_getUniqueFilename;
    DV.init.GetMonitorInfo              = DD_getMonitorInfo;
    DV.init.SetMonitorInfo              = DD_setMonitorInfo;
    DV.init.GetCurrentVideoModeInfo     = DD_GetCurrentVideoModeInfo;
    DV.init.GetCertifyInfo              = DRV_getCertifyInfo;
    DV.init.GetCurrentRefreshRate       = DD_GetCurrentRefreshRate;
    DV.init.GetVideoModeInfoExt         = DD_GetVideoModeInfoExt;
    DV.init.GetCustomVideoModeInfoExt   = DD_GetCustomVideoModeInfoExt;
    DV.init.SwitchPhysicalResolution    = DD_SwitchPhysicalResolution;
    DV.init.GetNumberOfHeads            = DRV_GetNumberOfHeads;
    DV.init.SetActiveHead               = DRV_SetActiveHead;
    DV.init.GetActiveHead               = DRV_GetActiveHead;
    DV.init.SetRef2dPointer             = DRV_SetRef2dPointer;
    DV.init.SetDisplayOutputExt         = DD_SetDisplayOutputExt;
    DV.init.GetDisplayOutputExt         = DD_GetDisplayOutputExt;
    DV.driver.SetDisplayStart           = DD_SetDisplayStart;
    DV.driver.IsVSync                   = DD_IsVSync;
    DV.driver.WaitVSync                 = DD_WaitVSync;
    DV.driver.SetPaletteData            = DD_SetPaletteData;
    DV.driver.GetPaletteData            = DD_GetPaletteData;

    // TODO: These could be supported!!
    DV.driver.SetGammaCorrectData   = CNOT_SUPPORTED;
    DV.driver.GetGammaCorrectData   = CNOT_SUPPORTED;
    DV.driver.GetCurrentScanLine    = LCNOT_SUPPORTED;

    // Configure the DPMS and power management functions
    DV.dpms.DPMSdetect              = DD_DPMSdetect;
    DV.sci.SCIdetect                = DD_SCIdetect;

    // Initialise the 2D hardware drawing functions
    ACCEL2D_init();
    BUFFER_init();

    // Set the configuration options to all enabled by default.
    rep_stosb(&DV.opt,sizeof(DV.opt),1);
    DV.prof.dwSize = sizeof(DV.prof);
    memset(DV.prof.m.modeList,0,sizeof(DV.prof.m.modeList));
    DV.prof.m.numModes = 0;
    DV.opt.dwSize = sizeof(DV.opt);
    DV.opt.memoryClock = 0;
    DV.opt.defaultMemoryClock = 0;
    DV.opt.maxMemoryClock = 0;
    DV.opt.outputDevice = DV.outputDevice[0] = gaOUTPUT_CRT;
    DV.opt.bDebugMode = 0;
    DV.opt.bMultiDisplay = false;
    DV.opt.bPortrait = false;
    DV.opt.bFlipped = false;
    DV.opt.bInvertColors = false;
    DV.opt.RTCFrequency = 2048;
    DV.opt.RTCAdvanceTicks = 2;
    DV.opt.glassesType = gaGlassesBlueCode;
    DV.opt.ioPort = 0;
    DV.opt.ioAndMask = 0;
    DV.opt.ioLeftOrMask = 0;
    DV.opt.ioRightOrMask = 0;
    DV.opt.ioOffOrMask = 0;
    DV.opt.vSyncWidthLeft = 4;
    DV.opt.vSyncWidthRight = 3;
    DV.opt.text80x43Height = 8;
    DV.opt.text80x50Height = 8;
    DV.opt.text80x60Height = 8;
    DV.opt.text100x43Height = 8;
    DV.opt.text100x50Height = 8;
    DV.opt.text100x60Height = 8;
    DV.opt.text132x43Height = 8;
    DV.opt.text132x50Height = 8;
    DV.opt.text132x60Height = 8;
    DV.opt.tripleScanLowRes = 0;
    DV.opt.doubleScan512 = 1;
    DV.opt.bHardwareStereoSync = 0;

    // Load any overridden options values from disk
    DD_getUniqueFilename(filename,2);
    if ((f = fopen(filename,"rb")) != NULL) {
        fread(&DV.opt,1,sizeof(DV.opt),f);
        fclose(f);
        }
    DRV_setOptions(&DV.opt);

    // Construct the detected configuration info block
    DRV_configInfo.dwSize = sizeof(DRV_configInfo);
    if (DC.ring0DC) {
        // Get the hardware driver config info and pass it through
        hwInit.GetConfigInfo(&DRV_configInfo);
        hwInit.GetCertifyInfo(&DRV_certifyInfo);

        // Initialise the hardware driver pass through functions.
        DRV_initHWDriver();
        }
    else {
        strcpy(DRV_configInfo.ManufacturerName,"Microsoft DirectX");
        IDirectDraw7_GetDeviceIdentifier(DD_lpDD,&dddi,DDGDI_GETHOSTIDENTIFIER);
        strcpy(DRV_configInfo.ChipsetName,dddi.szDescription);
        strcpy(DRV_configInfo.DACName,"Unknown");
        strcpy(DRV_configInfo.ClockName,"Unknown");
        }

    // Report the DirectX version in the VersionInfo field.
    dxVer = DD_getDXVersion();
    sprintf(DRV_configInfo.VersionInfo,"Microsoft DirectX %d.%d", dxVer >> 8, dxVer & 0xFF);
    return nOK;
}

/****************************************************************************
DESCRIPTION:
Performs any steps necessary to unload the driver from memory.

HEADER:
graphics/driver.h

REMARKS:
Performs any steps necessary to unload the driver from memory.
****************************************************************************/
void NAPI DRV_unloadDriver(void)
{
    int i;

    // Kill the buffer manager
    DD_exitBufMgr();

    // Unload the DirectDraw objects
    for (i = 0; i < DD_numPixelFormats; i++)
        SAFE_RELEASE(DD_lpTexture[i]);
    SAFE_RELEASE(DD_lpD3DDevice);
    SAFE_RELEASE(DD_lpD3D);
    SAFE_RELEASE(DD_lpDD);
    PM_unloadDirectDraw(DC.DeviceIndex);

    // Unload the system SNAP driver if loaded
    if (DC.ring0DC)
        gaExp.GA_unloadDriver(DC.ring0DC);
    if (DD_hModNGAW32)
        PM_freeLibrary(DD_hModNGAW32);

    // Unload USER32 and ADVAPI instances
    if (DD_hModUSER32)
        PM_freeLibrary(DD_hModUSER32);
    if (DD_hModADVAPI32)
        PM_freeLibrary(DD_hModADVAPI32);
    if (DD_hModKERNEL32)
        PM_freeLibrary(DD_hModKERNEL32);
}

