/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement mode initialisation and non-drawing
*               runtime related functions.
*
****************************************************************************/

#include "internal.h"

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
REMARKS:
Set a new bank for banked framebuffer modes.
****************************************************************************/
void NAPI DD_SetDisplayStart(
    N_int32 offset,
    N_int32 waitVRT)
{
    // This function does not do anything in the DirectDraw driver. Apps
    // must use the buffer manager functions for page flipping.
    (void)offset;
    (void)waitVRT;
}

/****************************************************************************
REMARKS:
Return true if the device is current in the vsync.
****************************************************************************/
N_int32 NAPI DD_IsVSync(void)
{
    BOOL isVSync;
    IDirectDraw7_GetVerticalBlankStatus(DD_lpDD,&isVSync);
    return isVSync;
}

/****************************************************************************
REMARKS:
Wait for the vertical sync to arrive.
****************************************************************************/
void NAPI DD_WaitVSync(void)
{
    IDirectDraw7_WaitForVerticalBlank(DD_lpDD,DDWAITVB_BLOCKBEGIN,NULL);
}

/****************************************************************************
REMARKS:
Program the hardware palette
****************************************************************************/
void NAPI DD_SetPaletteData(
    GA_palette *pal,
    N_int32 num,
    N_int32 index,
    N_int32 waitVRT)
{
	PALETTEENTRY	pe[256];
	int 			i;

    if (DD_lpPalette) {
    	for (i = 0; i < num; i++) {
    		pe[i].peRed = pal[i].Red;
    		pe[i].peGreen = pal[i].Green;
    		pe[i].peBlue = pal[i].Blue;
    		pe[i].peFlags = PC_RESERVED;
    		}
    	if (waitVRT)
            IDirectDraw7_WaitForVerticalBlank(DD_lpDD,DDWAITVB_BLOCKBEGIN,NULL);
        IDirectDrawPalette_SetEntries(DD_lpPalette,0,index,num,pe);
        }
}

/****************************************************************************
REMARKS:
Read the hardware palette.
****************************************************************************/
void NAPI DD_GetPaletteData(
    GA_palette *pal,
    N_int32 num,
    N_int32 index)
{
	PALETTEENTRY	pe[256];
	int 			i;

    if (DD_lpPalette) {
        IDirectDrawPalette_GetEntries(DD_lpPalette,0,index,num,pe);
    	for (i = 0; i < num; i++) {
    		pal[i].Red = pe[i].peRed;
    		pal[i].Green = pe[i].peGreen;
    		pal[i].Blue = pe[i].peBlue;
    		pal[i].Alpha = 0;
    		}
        }
}

