/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Include module to implement accelerated ellipse drawing
*               functions by simply calling the DrawRect functions for the
*               graphics hardware.
*
****************************************************************************/

#define SETUP_SOLID_SCAN \
    void (NAPIP DrawRect)(N_int32 left,N_int32 top,N_int32 width,N_int32 height); \
    switch (DV.cntMode.bits) { \
        case 8:  DrawRect = DV.draw_8bpp.DrawRect; break; \
        case 15: \
        case 16: DrawRect = DV.draw_16bpp.DrawRect; break; \
        case 24: DrawRect = DV.draw_24bpp.DrawRect; break; \
        case 32: DrawRect = DV.draw_32bpp.DrawRect; break; \
        }

#define SETUP_PATT_SCAN \
    void (NAPIP DrawRect)(N_int32 left,N_int32 top,N_int32 width,N_int32 height); \
    switch (DV.cntMode.bits) { \
        case 8:  DrawRect = DV.draw_8bpp.DrawPattRect; break; \
        case 15: \
        case 16: DrawRect = DV.draw_16bpp.DrawPattRect; break; \
        case 24: DrawRect = DV.draw_24bpp.DrawPattRect; break; \
        case 32: DrawRect = DV.draw_32bpp.DrawPattRect; break; \
        }

#define SETUP_COLORPATT_SCAN \
    void (NAPIP DrawRect)(N_int32 left,N_int32 top,N_int32 width,N_int32 height); \
    switch (DV.cntMode.bits) { \
        case 8:  DrawRect = DV.draw_8bpp.DrawColorPattRect; break; \
        case 15: \
        case 16: DrawRect = DV.draw_16bpp.DrawColorPattRect; break; \
        case 24: DrawRect = DV.draw_24bpp.DrawColorPattRect; break; \
        case 32: DrawRect = DV.draw_32bpp.DrawColorPattRect; break; \
        }

#define DrawScan(y,x1,x2)   DrawRect(x1,y,x2-x1,1)

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI DRV_DrawEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int i,j,maxIndex;
    SETUP_SOLID_SCAN;

    // Render the scanlines in the ellipse
    maxIndex = y+length-1;
    for (i = y,j = y+height; i < maxIndex; i++,j--,scans += 2) {
        DrawScan(i,scans[0],scans[1]);
        DrawScan(j,scans[0],scans[1]);
        }
    if (!(height & 1))
        DrawScan(i,scans[0],scans[1]);
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI DRV_DrawPattEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int i,j,maxIndex;
    SETUP_PATT_SCAN;

    // Render the scanlines in the ellipse
    maxIndex = y+length-1;
    for (i = y,j = y+height; i < maxIndex; i++,j--,scans += 2) {
        DrawScan(i,scans[0],scans[1]);
        DrawScan(j,scans[0],scans[1]);
        }
    if (!(height & 1))
        DrawScan(i,scans[0],scans[1]);
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI DRV_DrawColorPattEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int i,j,maxIndex;
    SETUP_COLORPATT_SCAN;

    // Render the scanlines in the ellipse
    maxIndex = y+length-1;
    for (i = y,j = y+height; i < maxIndex; i++,j--,scans += 2) {
        DrawScan(i,scans[0],scans[1]);
        DrawScan(j,scans[0],scans[1]);
        }
    if (!(height & 1))
        DrawScan(i,scans[0],scans[1]);
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis + pen height adjustment
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI DRV_DrawFatEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int i,j,maxIndex;
    SETUP_SOLID_SCAN;

    // Render the scanlines in the ellipse
    maxIndex = y+length;
    for (i = y,j = y+height; i < maxIndex; i++,j--,scans += 4) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            DrawScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            DrawScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            DrawScan(j,scans[I_LEFTL],scans[I_LEFTR]);
            DrawScan(j,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            DrawScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            DrawScan(j,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
    if ((height+1) & 1) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            DrawScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            DrawScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            DrawScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis + pen height adjustment
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI DRV_DrawPattFatEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int i,j,maxIndex;
    SETUP_PATT_SCAN;

    // Render the scanlines in the ellipse
    maxIndex = y+length;
    for (i = y,j = y+height; i < maxIndex; i++,j--,scans += 4) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            DrawScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            DrawScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            DrawScan(j,scans[I_LEFTL],scans[I_LEFTR]);
            DrawScan(j,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            DrawScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            DrawScan(j,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
    if ((height+1) & 1) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            DrawScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            DrawScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            DrawScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis + pen height adjustment
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI DRV_DrawColorPattFatEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int i,j,maxIndex;
    SETUP_COLORPATT_SCAN;

    // Render the scanlines in the ellipse
    maxIndex = y+length;
    for (i = y,j = y+height; i < maxIndex; i++,j--,scans += 4) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            DrawScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            DrawScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            DrawScan(j,scans[I_LEFTL],scans[I_LEFTR]);
            DrawScan(j,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            DrawScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            DrawScan(j,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
    if ((height+1) & 1) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            DrawScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            DrawScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            DrawScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
}
