/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Generic device driver filter module (#included)
*
****************************************************************************/

/*--------------------------- Global variables ----------------------------*/

static GA_options          *DRV_licensedOpt;
static GA_globalOptions    *DRV_globalOpt;
static ibool               (NAPIP ORG_queryFunctions)(N_uint32 id,N_int32 safetyLevel,void _FAR_ *funcs);
static void                (NAPIP ORG_unloadDriver)(void);
#ifdef  __DRIVER__
char                DRV_path[256];
N_imports _VARAPI   _N_imports;     // Imports to OS specific functions
PM_imports _VARAPI  _PM_imports;    // Imports to PM library functions
DRV_imports _VARAPI _DRV_imports;   // Imports to DDC specific functions
#endif

/*----------------------------- Implementation ----------------------------*/

#ifdef  __DRIVER__
/****************************************************************************
REMARKS:
Fatal error handler called when a non-imported function is called by the
driver. We leave this to a runtime error so that older applications and
shell drivers will work with newer bpd drivers provided no newer functions
are required by the driver itself. If they are, the application or shell
driver needs to be recompiled.
****************************************************************************/
static void _PM_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported PM_imports import function called! Please re-compile!\n");
}

/****************************************************************************
REMARKS:
Fatal error handler for non-imported N_imports.
****************************************************************************/
static void _N_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported N_imports import function called! Please re-compile!\n");
}

/****************************************************************************
REMARKS:
Fatal error handler for non-imported GA_imports.
****************************************************************************/
static void _DRV_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported DRV_imports import function called! Please re-compile!\n");
}
#endif

/****************************************************************************
HEADER:
graphics/filter.h

PARAMETERS:
dc          - Device context to insert filter in front of
path        - Path to Binary Portable DLL directory
nImp        - Generic SNAP imports
pmImp       - Generif PM library imports
drvImp      - Driver specific imports
licensedOpt - Licensed options structure
globalOpt   - Global options structure

RETURNS:
Pointer to SNAP device context modified for the filter driver.

REMARKS:
Does generic filter driver initialisation and handling.
****************************************************************************/
#ifndef  __DRIVER__
static
#endif
GA_devCtx * _CEXPORT GA_initFilter(
    GA_devCtx *dc,
    const char *path,
    N_imports *nImp,
    PM_imports *pmImp,
    DRV_imports *drvImp,
    GA_options *licensedOpt,
    GA_globalOptions *globalOpt)
{
#ifdef  __DRIVER__
    int     i,max;
    ulong   *p;

    // Initialize all default imports to point to fatal error handler
    // for upwards compatibility.
    max = sizeof(_N_imports)/sizeof(GA_initLibrary_t);
    for (i = 0,p = (ulong*)&_N_imports; i < max; i++)
        *p++ = (ulong)_N_fatalErrorHandler;
    max = sizeof(_PM_imports)/sizeof(GA_initLibrary_t);
    for (i = 0,p = (ulong*)&_PM_imports; i < max; i++)
        *p++ = (ulong)_PM_fatalErrorHandler;
    max = sizeof(_DRV_imports)/sizeof(GA_initLibrary_t);
    for (i = 0,p = (ulong*)&_DRV_imports; i < max; i++)
        *p++ = (ulong)_DRV_fatalErrorHandler;

    // Now copy all our imported functions
    memcpy(&_N_imports,nImp,MIN(sizeof(_N_imports),nImp->dwSize));
    memcpy(&_PM_imports,pmImp,MIN(sizeof(_PM_imports),pmImp->dwSize));
    memcpy(&_DRV_imports,drvImp,MIN(sizeof(_DRV_imports),drvImp->dwSize));
#else
    (void)nImp;
    (void)pmImp;
    (void)drvImp;
#endif
    strcpy(DRV_path,path);
    DRV_licensedOpt = licensedOpt;
    DRV_globalOpt = globalOpt;

    // Override the device driver function entry points
    ORG_queryFunctions = dc->loader.QueryFunctions;
    if (dc->Version >= 0x101)
        ORG_unloadDriver = dc->loader.UnloadDriver;
    else {
        dc->Version = 0x101;
        ORG_unloadDriver = NULL;
        }
    dc->loader.QueryFunctions = FILT_queryFunctions;
    dc->loader.UnloadDriver = FILT_unloadDriver;

    // Configure the internal filter driver
    if (!FILT_configure(dc))
        return NULL;
    return dc;
}

