/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Hardware specific device detection code.
*
****************************************************************************/

#include "snap/graphics/driver.h"

/*--------------------------- Global variables ----------------------------*/

// Tables of default pixel formats for the supported video modes. This table
// is filled in with the default values, and is copied to the real table
// before the detection code is run.

GA_pixelFormat DRV_pf15 = {
    0x1F,0x0A,3,0x1F,0x05,3,0x1F,0x00,3,0x01,0x0F,7,    // 555 15bpp
    };

GA_pixelFormat DRV_pf16 = {
    0x1F,0x0B,3,0x3F,0x05,2,0x1F,0x00,3,0x00,0x00,0,    // 565 16bpp
    };

GA_pixelFormat DRV_pf24_RGB = {
    0xFF,0x10,0,0xFF,0x08,0,0xFF,0x00,0,0x00,0x00,0,    // RGB 24bpp
    };

GA_pixelFormat DRV_pf24_BGR = {
    0xFF,0x00,0,0xFF,0x08,0,0xFF,0x10,0,0x00,0x00,0,    // BGR 24bpp
    };

GA_pixelFormat DRV_pf32_ARGB = {
    0xFF,0x10,0,0xFF,0x08,0,0xFF,0x00,0,0xFF,0x18,0,    // ARGB 32bpp
    };

GA_pixelFormat DRV_pf32_ABGR = {
    0xFF,0x00,0,0xFF,0x08,0,0xFF,0x10,0,0xFF,0x18,0,    // ABGR 32bpp
    };

GA_pixelFormat DRV_pf32_RGBA = {
    0xFF,0x18,0,0xFF,0x10,0,0xFF,0x08,0,0xFF,0x00,0,    // RGBA 32bpp
    };

GA_pixelFormat DRV_pf32_BGRA = {
    0xFF,0x08,0,0xFF,0x10,0,0xFF,0x18,0,0xFF,0x00,0,    // BGRA 32bpp
    };

/* Declare a table for performing bitmap mirroring */

uchar DRV_mirrorTable[] = {
    0x00,0x80,0x40,0xC0,0x20,0xA0,0x60,0xE0,
    0x10,0x90,0x50,0xD0,0x30,0xB0,0x70,0xF0,
    0x08,0x88,0x48,0xC8,0x28,0xA8,0x68,0xE8,
    0x18,0x98,0x58,0xD8,0x38,0xB8,0x78,0xF8,
    0x04,0x84,0x44,0xC4,0x24,0xA4,0x64,0xE4,
    0x14,0x94,0x54,0xD4,0x34,0xB4,0x74,0xF4,
    0x0C,0x8C,0x4C,0xCC,0x2C,0xAC,0x6C,0xEC,
    0x1C,0x9C,0x5C,0xDC,0x3C,0xBC,0x7C,0xFC,
    0x02,0x82,0x42,0xC2,0x22,0xA2,0x62,0xE2,
    0x12,0x92,0x52,0xD2,0x32,0xB2,0x72,0xF2,
    0x0A,0x8A,0x4A,0xCA,0x2A,0xAA,0x6A,0xEA,
    0x1A,0x9A,0x5A,0xDA,0x3A,0xBA,0x7A,0xFA,
    0x06,0x86,0x46,0xC6,0x26,0xA6,0x66,0xE6,
    0x16,0x96,0x56,0xD6,0x36,0xB6,0x76,0xF6,
    0x0E,0x8E,0x4E,0xCE,0x2E,0xAE,0x6E,0xEE,
    0x1E,0x9E,0x5E,0xDE,0x3E,0xBE,0x7E,0xFE,
    0x01,0x81,0x41,0xC1,0x21,0xA1,0x61,0xE1,
    0x11,0x91,0x51,0xD1,0x31,0xB1,0x71,0xF1,
    0x09,0x89,0x49,0xC9,0x29,0xA9,0x69,0xE9,
    0x19,0x99,0x59,0xD9,0x39,0xB9,0x79,0xF9,
    0x05,0x85,0x45,0xC5,0x25,0xA5,0x65,0xE5,
    0x15,0x95,0x55,0xD5,0x35,0xB5,0x75,0xF5,
    0x0D,0x8D,0x4D,0xCD,0x2D,0xAD,0x6D,0xED,
    0x1D,0x9D,0x5D,0xDD,0x3D,0xBD,0x7D,0xFD,
    0x03,0x83,0x43,0xC3,0x23,0xA3,0x63,0xE3,
    0x13,0x93,0x53,0xD3,0x33,0xB3,0x73,0xF3,
    0x0B,0x8B,0x4B,0xCB,0x2B,0xAB,0x6B,0xEB,
    0x1B,0x9B,0x5B,0xDB,0x3B,0xBB,0x7B,0xFB,
    0x07,0x87,0x47,0xC7,0x27,0xA7,0x67,0xE7,
    0x17,0x97,0x57,0xD7,0x37,0xB7,0x77,0xF7,
    0x0F,0x8F,0x4F,0xCF,0x2F,0xAF,0x6F,0xEF,
    0x1F,0x9F,0x5F,0xDF,0x3F,0xBF,0x7F,0xFF,
    };

#ifdef DEBUG_BLANKS
int hBlankStart,hBlankEnd,vBlankStart,vBlankEnd;
#endif

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
REMARKS:
Sets up the internal device support vectors to point to the VGA compatible
and generic device support versions before the chipset specific code is
run. The chipset specific code will override these vectors as necessary
later in the code.
****************************************************************************/
void VGA_init(void)
{
    //------------------------------------------------------------------
    // Configure the default VGA/Driver support functions
    //------------------------------------------------------------------
    DV.chip.getDefaultOptions   = STUB_VECTOR;
    DV.chip.unlockRegisters     = STUB_VECTOR;
    DV.chip.unlockCRTCRegisters = VGA_unlockCRTCRegisters;
    DV.chip.lockCRTCRegisters   = VGA_lockCRTCRegisters;
    DV.chip.lockRegisters       = STUB_VECTOR;
    DV.chip.blankDisplay        = VGA_blankDisplay;
    DV.chip.unblankDisplay      = VGA_unblankDisplay;
    DV.chip.preModeSet          = VGA_preModeSet;
    DV.chip.setChipsetRegisters = DRV_setChipsetRegisters;
    DV.chip.setMemoryClock      = STUB_VECTOR;
    DV.chip.setCRTCRegisters    = VGA_setCRTCRegisters;
    DV.chip.setPixelClock       = STUB_VECTOR;
    DV.chip.roundBytesPerLine   = VGA_roundBytesPerLine;
    DV.chip.setBytesPerLine     = VGA_setBytesPerLine;
    DV.chip.enableLinear        = STUB_VECTOR;
    DV.chip.disableLinear       = STUB_VECTOR;
    DV.chip.clearMemory         = DRV_clearMemory;
    DV.chip.setupAccelerator    = STUB_VECTOR;
    DV.chip.setOptions          = STUB_VECTOR;
    DV.chip.postModeSet         = VGA_postModeSet;
    DV.chip.setDACPage          = USSTUB_VECTOR;
    DV.chip.clearDACPage        = STUB_VECTOR;
    DV.chip.getDACPort          = VGA_getDACPort;
    DV.chip.rdDACReg            = DRV_rdDACReg;
    DV.chip.wrDACReg            = DRV_wrDACReg;
    DV.chip.rdDACIdx            = DRV_rdDACIdx;
    DV.chip.wrDACIdx            = DRV_wrDACIdx;
    DV.chip.get2DAccelInfo      = STUB_VECTOR;
    DV.chip.getVideoAccelInfo   = STUB_VECTOR;
    DV.chip.get3DAccelInfo      = STUB_VECTOR;
    DV.dac.setDACRegisters      = STUB_VECTOR;
    DV.dac.set8BitDAC           = VGA_set8BitDAC;
    DV.vga.getBlankTimings      = VGA_getBlankTimings;
    DV.vga.lockDownTimings      = VGA_lockDownTimings;
    DV.driver.SetDisplayStartXY = DRV_SetDisplayStartXY;

    //------------------------------------------------------------------
    // Configure the default pixel formats
    //------------------------------------------------------------------
    DV.pf24                         = &DRV_pf24_RGB;
    DV.pf32                         = &DRV_pf32_ARGB;
}

/****************************************************************************
REMARKS:
Unlocks the CRTC registers for writing
****************************************************************************/
void VGA_unlockCRTCRegisters(void)
{
    DV.chip.unlockRegisters();
}

/****************************************************************************
REMARKS:
Locks the CRTC registers after writing
****************************************************************************/
void VGA_lockCRTCRegisters(void)
{
    DV.chip.lockRegisters();
}

/****************************************************************************
REMARKS:
Disables the VGA display for screen output making it blank.
****************************************************************************/
void VGA_blankDisplay(void)
{
    // Turn screen off
    setinx(0x3C4,0x01,0x20);

    // Disable video output
    inpb(0x3DA);
    outpb(0x3C0,0x00);
}

/****************************************************************************
REMARKS:
Enables the VGA display for screen output.
****************************************************************************/
void VGA_unblankDisplay(void)
{
    // Turn screen back on
    clrinx(0x3C4,0x01,0x20);

    // Enable video output
    inpb(0x3DA);
    outpb(0x3C0,0x20);
}

/****************************************************************************
REMARKS:
Sets up for programming a VGA compatible display mode.
****************************************************************************/
void VGA_preModeSet(void)
{
    DV.chip.unlockRegisters();
    DV.chip.blankDisplay();     // Blank the display screen
    clrinx(0x3D4,0x11,0x80);    // Disable write protect on CRTC registers
}

/****************************************************************************
REMARKS:
Programs the VGA CRTC registers to the hardware using the CRTC register
definition tables.
****************************************************************************/
void VGA_setCRTCRegisters(
    GA_CRTCInfo *crtc)
{
    int         misc;

    // Convert the CRTC timings to our VGA structure format and to the
    // format needed by the VGA hardware.
    DV.hCRTC.hTotal = DRV_scaleCRTC(crtc->HorizontalTotal);
    // Adjust sync positions for VGA modes for GTF timings
    if (DV.cntMode.bits == 0) {
        DV.hCRTC.hSyncStart = DRV_scaleCRTC(crtc->HorizontalSyncStart + 32);
        DV.hCRTC.hSyncEnd = DRV_scaleCRTC(crtc->HorizontalSyncEnd + 32);
        }
    else if (DV.cntMode.bits == 4) {
        DV.hCRTC.hSyncStart = DRV_scaleCRTC(crtc->HorizontalSyncStart + 24);
        DV.hCRTC.hSyncEnd = DRV_scaleCRTC(crtc->HorizontalSyncEnd + 24);
        }
    else {
        DV.hCRTC.hSyncStart = DRV_scaleCRTC(crtc->HorizontalSyncStart + 8);
        DV.hCRTC.hSyncEnd = DRV_scaleCRTC(crtc->HorizontalSyncEnd + 8);
        }
    DV.vCRTC.vTotal = crtc->VerticalTotal;
    DV.vCRTC.vSyncStart = crtc->VerticalSyncStart;
    DV.vCRTC.vSyncEnd = crtc->VerticalSyncEnd;

    // Now compute the blank timings and lock down any specific values
    DV.vga.getBlankTimings(&DV.hCRTC,&DV.vCRTC,crtc);
    DV.vga.lockDownTimings(&DV.hCRTC,&DV.vCRTC,crtc);

    // Store external blanks to be hacked by centering program
#ifdef DEBUG_BLANKS
    if (hBlankStart != -1) {
        DV.hCRTC.hBlankStart = hBlankStart;
        DV.hCRTC.hBlankEnd = hBlankEnd;
        DV.vCRTC.vBlankStart = vBlankStart;
        DV.vCRTC.vBlankEnd = vBlankEnd;
        }
    hBlankStart = DV.hCRTC.hBlankStart;
    hBlankEnd = DV.hCRTC.hBlankEnd;
    vBlankStart = DV.vCRTC.vBlankStart;
    vBlankEnd = DV.vCRTC.vBlankEnd;
#endif

    // Now convert the timings to the internal VGA format
    DV.hCRTC.hTotal = (DV.hCRTC.hTotal >> 3) - 5;
    DV.hCRTC.hDisp = (DV.hCRTC.hDisp >> 3) - 1;
    DV.hCRTC.hBlankStart = DV.hCRTC.hBlankStart >> 3;
    DV.hCRTC.hBlankEnd = DV.hCRTC.hBlankEnd >> 3;
    DV.hCRTC.hSyncStart = DV.hCRTC.hSyncStart >> 3;
    DV.hCRTC.hSyncEnd = DV.hCRTC.hSyncEnd >> 3;
    DV.vCRTC.vTotal = DV.vCRTC.vTotal - 2;
    DV.vCRTC.vDisp = DV.vCRTC.vDisp - 1;
    DV.vCRTC.vBlankStart = DV.vCRTC.vBlankStart - 1;
    DV.vCRTC.vBlankEnd = DV.vCRTC.vBlankEnd - 1;

    // Convert the CRTC flags and add to timings
    if (crtc->Flags & gaInterlaced)
        DV.hCRTC.hSyncStart |= vgaInterlace;
    if ((crtc->Flags & gaDoubleScan) && !(DV.cntMode.flags & gaVGA))
        DV.vCRTC.vSyncStart |= vgaDblScan;
    if ((crtc->Flags & gaTripleScan) && !(DV.cntMode.flags & gaVGA))
        DV.vCRTC.vSyncStart |= vgaTrpScan;
    if (crtc->Flags & gaHSyncNeg)
        DV.vCRTC.vSyncStart |= vgaSyncHNeg;
    if (crtc->Flags & gaVSyncNeg)
        DV.vCRTC.vSyncStart |= vgaSyncVNeg;

    // Now program the VGA CRTC timings
    clrinx(0x3D4,0x11,0x80);    // Disable write protect on CRTC registers
    DRV_writeMultiRegisters((uchar*)&DV.hCRTC,DV.HCRTCRegsDesc);
    DRV_writeMultiRegisters((uchar*)&DV.vCRTC,DV.VCRTCRegsDesc);

    // Set the VGA double scan bit
    if ((crtc->Flags & gaDoubleScan) && (DV.cntMode.flags & gaVGA))
        setinx(0x3D4,0x09,B10000000);

    // Set the sync polarities
    misc = inpb(0x3CC);
    outpb(0x3C2,(misc & 0x3F) | ((DV.vCRTC.vSyncStart & vgaSyncPolMask) >> 8));
    setinx(0x3D4,0x11,0x80);

    // Set bit 8&9 of split screen register to get it out of our hair.
    setinx(0x3D4,0x07,B00010000);
    setinx(0x3D4,0x09,B01000000);
}

/****************************************************************************
PARAMETERS:
Rounds the bytesPerLine value for VGA compatible hardware
****************************************************************************/
N_int32 VGA_roundBytesPerLine(
    N_int32 bytesPerLine,
    N_int32 bitsPerPixel)
{
    (void)bitsPerPixel;
    return bytesPerLine;
}

/****************************************************************************
PARAMETERS:
Programs the VGA compatible CRTC offset register
****************************************************************************/
N_int32 VGA_setBytesPerLine(
    N_int32 bytesPerLine)
{
    if (DV.cntMode.bits == 0) {
        int offset = (bytesPerLine + 3) >> 2;
        wrinx(0x3D4,0x13,offset);
        return offset << 2;
        }
    else if (DV.cntMode.bits <= 4) {
        int offset = (bytesPerLine + 1) >> 1;
        wrinx(0x3D4,0x13,offset);
        return offset << 1;
        }
    else {
        int offset = (bytesPerLine + 7) >> 3;
        wrinx(0x3D4,0x13,offset);
        return offset << 3;
        }
}

/****************************************************************************
REMARKS:
Completes the programming of VGA compatible modes after a mode set.
****************************************************************************/
void VGA_postModeSet(void)
{
    DV.chip.unblankDisplay();       // Renable the display output
    DV.chip.lockRegisters();
}

/****************************************************************************
PARAMETERS:
Computes the CRTC blank timings for the VGA compatible timing structures
****************************************************************************/
void VGA_getBlankTimings(
    VGA_hCRTC *h,
    VGA_vCRTC *v,
    GA_CRTCInfo *crtc)
{
    // Start the hBlank interval at hDisp
    h->hDisp = DRV_scaleCRTC(DV.cntModeInfo.PhysicalXResolution);
    h->hBlankStart = h->hDisp;
    h->hBlankEnd = h->hBlankStart + GA_MAX_HB_WIDTH;
    if (h->hBlankEnd > h->hTotal-32)
        h->hBlankEnd = h->hTotal-32;

    // Start the vBlank interval at vDisp
    v->vDisp = DV.cntModeInfo.PhysicalYResolution;
    if (crtc->Flags & gaInterlaced)
        v->vDisp = v->vDisp / 2;
    if (crtc->Flags & gaDoubleScan)
        v->vDisp *= 2;
    else if (crtc->Flags & gaTripleScan)
        v->vDisp *= 3;
    v->vBlankStart = v->vDisp+1;
    v->vBlankEnd = v->vBlankStart + GA_MAX_VB_WIDTH;
    if (v->vBlankEnd > v->vTotal-8)
        v->vBlankEnd = v->vTotal-8;

    // The blank end timings should always fall on an odd 8 pixel
    // boundary to ensure that the 24 bit modes work properly
    if (DV.cntMode.bits == 24) {
        if (h->hTotal & 7)
            h->hTotal &= 0xFFF8;
        if (h->hBlankEnd & 7)
            h->hBlankEnd &= 0xFFF8;
        }
}

/****************************************************************************
PARAMETERS:
Locks down the VGA compatible timings. This does nothing by default, but
some chips have restrictions on the CRTC timings that must be used so this
function is used to override the default values.
****************************************************************************/
void VGA_lockDownTimings(
    VGA_hCRTC *h,
    VGA_vCRTC *v,
    GA_CRTCInfo *crtc)
{
    // Nothing to do for 'good' VGA cards
    (void)h;
    (void)v;
    (void)crtc;
}

/****************************************************************************
REMARKS:
Forces the VGA compatible DAC into PEL mode.
****************************************************************************/
void VGA_dacToPel(void)
{
    inpb(0x3C8);
}

/****************************************************************************
RETURNS:
Last value read from command register

REMARKS:
Triggers the command register by reading the VGA DAC I/O port 4 times.
****************************************************************************/
uchar VGA_triggerDac(void)
{
    inpb(0x3C6);
    inpb(0x3C6);
    inpb(0x3C6);
    return inpb(0x3C6);
}

/****************************************************************************
RETURNS:
Last value read from command register

REMARKS:
Enters the command mode for VGA compatible extended DAC's
****************************************************************************/
uchar VGA_dacToCommand(void)
{
    VGA_dacToPel();
    return VGA_triggerDac();
}

/****************************************************************************
REMARKS:
Programs the hidden command register for a VGA compatible RAMDAC.
****************************************************************************/
void VGA_dacMode(
    uchar value)
{
    VGA_dacToPel();
    VGA_triggerDac();
    outpb(0x3C6,value);
    VGA_dacToPel();
}

/****************************************************************************
REMARKS:
Programs the hidden command register for a VGA compatible RAMDAC, but takes
a mask for the bits to be modified.
****************************************************************************/
void VGA_dacModeMasked(
    uchar value,
    uchar mask)
{
    uint PELMask;

    VGA_dacToPel();
    PELMask = inpb(0x3C6);
    VGA_triggerDac();
    modreg(0x3C6,mask,value);
    VGA_dacToPel();
    outpb(0x3C6,PELMask);       // Restore PEL mask for buggy hardware
}

/****************************************************************************
REMARKS:
Returns the I/O port address of the VGA RAMDAC ports given the RAMDAC
register index.
****************************************************************************/
int VGA_getDACPort(
    int index)
{
    static ushort port[] = {
        0x3C8,0x3C9,0x3C6,0x3C7,
        };
    return port[index & 3];
}

/****************************************************************************
PARAMETERS:
index   - Indexed register to read

RETURNS:
Value read from the register

REMARKS:
This function reads the value from an indexed RAMDAC register.
****************************************************************************/
uchar DRV_rdDACReg(
    uchar index)
{
    int port = DV.chip.setDACPage(index);
    int val = inpb(port);
    DV.chip.clearDACPage();
    return val;
}

/****************************************************************************
PARAMETERS:
index   - Indexed register to write
value   - New value to write

REMARKS:
This function writes a value to an indexed RAMDAC register.
****************************************************************************/
void DRV_wrDACReg(
    uchar index,
    uchar value)
{
    int port = DV.chip.setDACPage(index);
    outpb(port,value);
    DV.chip.clearDACPage();
}

/****************************************************************************
REMARKS:
Return true if the display start address has changed since it was programmed.
****************************************************************************/
N_int32 NAPI VGA_GetDisplayStartStatus(void)
{
    if (!DV.sharedIRQ)
        return (inpb(0x3C2) & 0x80) >> 7;
    return true;
}

/****************************************************************************
REMARKS:
Returns true if currently inside the vertical retrace period.
****************************************************************************/
N_int32 NAPI VGA_IsVSync(void)
{
    return (inpb(0x3DA) & 8) >> 3;
}

/****************************************************************************
REMARKS:
Waits for the start of the vertical retrace period. If the retrace is
already started, this function waits for the next retrace to begin.
****************************************************************************/
void NAPI VGA_WaitVSync(void)
{
    int waitVRT = 1;
    WaitDEVGA();
    WaitVERTVGA();
}

/****************************************************************************
REMARKS:
Programs the VGA compatible sync width
****************************************************************************/
void NAPI VGA_SetVSyncWidth(N_int32 width)
{
    DV.cntCRTC.VerticalSyncEnd = DV.cntCRTC.VerticalSyncStart + width;
    modinx(0x3D4,0x11,0x0F,DV.cntCRTC.VerticalSyncEnd);
}

/****************************************************************************
REMARKS:
Reads the VGA compatible sync width from the hardware CRTC controller
****************************************************************************/
N_int32 NAPI VGA_GetVSyncWidth(void)
{
    return DV.cntCRTC.VerticalSyncEnd - DV.cntCRTC.VerticalSyncStart;
}

/****************************************************************************
REMARKS:
Issue a short bus delay by reading the VGA status I/O register.
****************************************************************************/
void VGA_delay(void)
{
    if (DV.opt.bEnablePaletteDelay) {
        inpb(0x3DA);
        inpb(0x3DA);
        }
}

/****************************************************************************
REMARKS:
Program the RAMDAC data register. We need a delay in here to ensure that
the writes take hold, otherwise they may go too fast for older hardware.
****************************************************************************/
static void writeDAC(
    int index,
    uchar v)
{
    VGA_delay();
    outpb(index,v);
}

/****************************************************************************
REMARKS:
Read the RAMDAC data register. We need a delay in here to ensure that
the writes take hold, otherwise they may go too fast for older hardware.
****************************************************************************/
static uchar readDAC(
    int index)
{
    VGA_delay();
    return inpb(index);
}

/****************************************************************************
REMARKS:
Programs the VGA palette registers.
****************************************************************************/
void NAPI VGA_SetPaletteData(
    GA_palette *pal,
    N_int32 num,
    N_int32 index,
    N_int32 waitVRT)
{
    int pelData = 0x3C9;

    if (waitVRT) {
        WaitDEVGA();
        WaitVERTVGA();
        }
    if (DV.cntDACWidth == 8) {
        // DAC is in 8-bit mode, so program values directly
        while (num--) {
            writeDAC(0x3C8,index++);
            writeDAC(pelData,pal->Red);
            writeDAC(pelData,pal->Green);
            writeDAC(pelData,pal->Blue);
            pal++;
            }
        }
    else {
        // DAC is in 6-bit mode, so convert to 6-bit values on the fly
        while (num--) {
            writeDAC(0x3C8,index++);
            writeDAC(pelData,pal->Red >> 2);
            writeDAC(pelData,pal->Green >> 2);
            writeDAC(pelData,pal->Blue >> 2);
            pal++;
            }
        }
}

/****************************************************************************
REMARKS:
Read the palette registers.
****************************************************************************/
void NAPI VGA_GetPaletteData(
    GA_palette *pal,
    N_int32 num,
    N_int32 index)
{
    int pelData = 0x3C9;

    if (DV.cntDACWidth == 8) {
        // DAC is in 8-bit mode, so program values directly
        while (num--) {
            writeDAC(0x3C7,index++);
            pal->Red = readDAC(pelData);
            pal->Green = readDAC(pelData);
            pal->Blue = readDAC(pelData);
            pal++;
            }
        }
    else {
        // DAC is in 6-bit mode, so convert to 6-bit values on the fly
        while (num--) {
            writeDAC(0x3C7,index++);
            pal->Red = readDAC(pelData) << 2;
            pal->Green = readDAC(pelData) << 2;
            pal->Blue = readDAC(pelData) << 2;
            pal++;
            }
        }
}

/****************************************************************************
REMARKS:
Programs the VGA compatible gamma correction data into the VGA palette
registers.
****************************************************************************/
void NAPI VGA_SetGammaCorrectData(
    GA_palette *pal,
    N_int32 num,
    N_int32 index,
    N_int32 waitVRT)
{
    int pelData = 0x3C9;

    if (waitVRT) {
        WaitDEVGA();
        WaitVERTVGA();
        }
    while (num--) {
        writeDAC(0x3C8,index++);
        writeDAC(pelData,pal->Red);
        writeDAC(pelData,pal->Green);
        writeDAC(pelData,pal->Blue);
        pal++;
        }
}

/****************************************************************************
REMARKS:
Read the palette registers.
****************************************************************************/
void NAPI VGA_GetGammaCorrectData(
    GA_palette *pal,
    N_int32 num,
    N_int32 index)
{
    int pelData = 0x3C9;

    while (num--) {
        writeDAC(0x3C7,index++);
        pal->Red = readDAC(pelData);
        pal->Green = readDAC(pelData);
        pal->Blue = readDAC(pelData);
        pal++;
        }
}

/****************************************************************************
REMARKS:
Function to set 8-bit wide RAMDAC mode. This is not supported on fully
VGA compatible controllers, so the default is to fail this function.
****************************************************************************/
ibool VGA_set8BitDAC(
    ibool enable)
{
    (void)enable;
    return false;
}

/****************************************************************************
REMARKS:
Downloads the VGA font data to the text mode display memory.
****************************************************************************/
void VGA_loadFont(
    uchar *data,
    int cellHeight)
{
    int     i;
    uchar   *vgaMem = DC.BankedMem;
    uchar   save[7];

    /* Enable access to character generator RAM */
    save[0] = rdinx(0x3C4,0x00);
    save[1] = rdinx(0x3C4,0x02);
    save[2] = rdinx(0x3C4,0x04);
    save[3] = rdinx(0x3C4,0x00);
    save[4] = rdinx(0x3CE,0x04);
    save[5] = rdinx(0x3CE,0x05);
    save[6] = rdinx(0x3CE,0x06);
    wrinx(0x3C4,0x00,0x01);
    wrinx(0x3C4,0x02,0x04);
    wrinx(0x3C4,0x04,0x07);
    wrinx(0x3C4,0x00,0x03);
    wrinx(0x3CE,0x04,0x02);
    wrinx(0x3CE,0x05,0x00);
    wrinx(0x3CE,0x06,0x00);

    /* Zero character generator RAM */
    memset(vgaMem,0,32 * 256);

    /* Download the characters to RAM. Note that we need a delay to avoid
     * problems on some graphics hardware.
     */
    for (i = 0; i < 256; i++) {
        usDelay(10);
        memcpy(vgaMem+(i*32),data+(i*cellHeight),cellHeight);
        }

    /* Set the text character font height */
    modinx(0x3D4,0x9,0x1F,cellHeight-1);

    /* Restore registers */
    wrinx(0x3C4,0x00,save[0]);
    wrinx(0x3C4,0x02,save[1]);
    wrinx(0x3C4,0x04,save[2]);
    wrinx(0x3C4,0x00,save[3]);
    wrinx(0x3CE,0x04,save[4]);
    wrinx(0x3CE,0x05,save[5]);
    wrinx(0x3CE,0x06,save[6]);
}
