/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Include module to implement accelerated ROP3 functions for
*               hardware that only has ROP2 functions. We can implement ROP3
*               operations using multiple passes over the framebuffer and
*               optionally using a temporary scanline buffer.
*
****************************************************************************/

/*--------------------------- Global variables ----------------------------*/

#define FROM_SRC    0x010
#define FROM_DEST   0x020
#define FROM_PATT   0x040
#define FROM_TEMP   0x080
#define TO_DEST     0x100
#define TO_TEMP     0x200
#define R2_0        GA_R2_BLACK
#define R2_1        GA_R2_WHITE
#define PASS_MIX    (*pass & 0xF)

#define DSon        GA_R2_NOTMERGESRC | FROM_SRC  | FROM_DEST
#define DSna        GA_R2_MASKNOTSRC  | FROM_SRC  | FROM_DEST
#define Sn          GA_R2_NOTCOPYSRC  | FROM_SRC  | 0
#define SDna        GA_R2_MASKSRCNOT  | FROM_SRC  | FROM_DEST
#define Dn          GA_R2_NOT         | FROM_SRC  | 0
#define DSx         GA_R2_XORSRC      | FROM_SRC  | FROM_DEST
#define DSan        GA_R2_NOTMASKSRC  | FROM_SRC  | FROM_DEST
#define DSa         GA_R2_MASKSRC     | FROM_SRC  | FROM_DEST
#define DSxn        GA_R2_NOTXORSRC   | FROM_SRC  | FROM_DEST
#define D           GA_R2_COPYSRC     | 0         | FROM_DEST
#define DSno        GA_R2_MERGENOTSRC | FROM_SRC  | FROM_DEST
#define S           GA_R2_COPYSRC     | FROM_SRC  | 0
#define SDno        GA_R2_MERGESRCNOT | FROM_SRC  | FROM_DEST
#define DSo         GA_R2_MERGESRC    | FROM_SRC  | FROM_DEST

#define DPon        GA_R2_NOTMERGESRC | FROM_PATT | FROM_DEST
#define DPna        GA_R2_MASKNOTSRC  | FROM_PATT | FROM_DEST
#define Pn          GA_R2_NOTCOPYSRC  | FROM_PATT | 0
#define PDna        GA_R2_MASKSRCNOT  | FROM_PATT | FROM_DEST
#define DPx         GA_R2_XORSRC      | FROM_PATT | FROM_DEST
#define DPan        GA_R2_NOTMASKSRC  | FROM_PATT | FROM_DEST
#define DPa         GA_R2_MASKSRC     | FROM_PATT | FROM_DEST
#define DPxn        GA_R2_NOTXORSRC   | FROM_PATT | FROM_DEST
#define DPno        GA_R2_MERGENOTSRC | FROM_PATT | FROM_DEST
#define P           GA_R2_COPYSRC     | FROM_PATT | 0
#define PDno        GA_R2_MERGESRCNOT | FROM_PATT | FROM_DEST
#define DPo         GA_R2_MERGESRC    | FROM_PATT | FROM_DEST

#define TSon        GA_R2_NOTMERGESRC | FROM_TEMP | FROM_SRC
#define TSna        GA_R2_MASKNOTSRC  | FROM_TEMP | FROM_SRC
#define STna        GA_R2_MASKSRCNOT  | FROM_TEMP | FROM_SRC
#define Tn          GA_R2_NOT         | FROM_TEMP | 0
#define TSx         GA_R2_XORSRC      | FROM_TEMP | FROM_SRC
#define TSan        GA_R2_NOTMASKSRC  | FROM_TEMP | FROM_SRC
#define TSa         GA_R2_MASKSRC     | FROM_TEMP | FROM_SRC
#define TSxn        GA_R2_NOTXORSRC   | FROM_TEMP | FROM_SRC
#define T           GA_R2_COPYSRC     | FROM_TEMP | 0
#define TSno        GA_R2_MERGENOTSRC | FROM_TEMP | FROM_SRC
#define STno        GA_R2_MERGESRCNOT | FROM_TEMP | FROM_SRC
#define TSo         GA_R2_MERGESRC    | FROM_TEMP | FROM_SRC

#define TPon        GA_R2_NOTMERGESRC | FROM_PATT | FROM_TEMP
#define TPna        GA_R2_MASKNOTSRC  | FROM_PATT | FROM_TEMP
#define PTna        GA_R2_MASKSRCNOT  | FROM_PATT | FROM_TEMP
#define TPx         GA_R2_XORSRC      | FROM_PATT | FROM_TEMP
#define TPan        GA_R2_NOTMASKSRC  | FROM_PATT | FROM_TEMP
#define TPa         GA_R2_MASKSRC     | FROM_PATT | FROM_TEMP
#define TPxn        GA_R2_NOTXORSRC   | FROM_PATT | FROM_TEMP
#define TPno        GA_R2_MERGENOTSRC | FROM_PATT | FROM_TEMP
#define PTno        GA_R2_MERGESRCNOT | FROM_PATT | FROM_TEMP
#define TPo         GA_R2_MERGESRC    | FROM_PATT | FROM_TEMP

#define DTon        GA_R2_NOTMERGESRC | FROM_TEMP | FROM_DEST
#define DTna        GA_R2_MASKNOTSRC  | FROM_TEMP | FROM_DEST
#define TDna        GA_R2_MASKSRCNOT  | FROM_TEMP | FROM_DEST
#define DTx         GA_R2_XORSRC      | FROM_TEMP | FROM_DEST
#define DTan        GA_R2_NOTMASKSRC  | FROM_TEMP | FROM_DEST
#define DTa         GA_R2_MASKSRC     | FROM_TEMP | FROM_DEST
#define DTxn        GA_R2_NOTXORSRC   | FROM_TEMP | FROM_DEST
#define DTno        GA_R2_MERGENOTSRC | FROM_TEMP | FROM_DEST
#define TDno        GA_R2_MERGESRCNOT | FROM_TEMP | FROM_DEST
#define DTo         GA_R2_MERGESRC    | FROM_TEMP | FROM_DEST

typedef struct {
    ushort  pass1;
    ushort  pass2;
    ushort  pass3;
    ushort  pass4;
    ushort  pass5;
    ushort  pass6;
    } ROP3_entry;

ROP3_entry rop3_table[256] = {
    /* 00: 0           */ {0},
    /* 01: DPSoon      */ {TO_DEST | DPo, TO_DEST | DSon},
    /* 02: DPSona      */ {TO_TEMP | S, TO_TEMP | TPon, TO_DEST | DTa},
    /* 03: PSon        */ {TO_DEST | S, TO_DEST | DPon},
    /* 04: SDPona      */ {TO_DEST | DPon, TO_DEST | DSa},
    /* 05: DPon        */ {TO_DEST | DPon},
    /* 06: PDSxnon     */ {TO_DEST | DSxn, TO_DEST | DPon},
    /* 07: PDSaon      */ {TO_DEST | DSa, TO_DEST | DPon},
    /* 08: SDPnaa      */ {TO_DEST | DPna, TO_DEST | DSa},
    /* 09: PDSxon      */ {TO_DEST | DSx, TO_DEST | DPon},
    /* 0A: DPna        */ {TO_DEST | DPna},
    /* 0B: PSDnaon     */ {TO_DEST | SDna, TO_DEST | DPon},
    /* 0C: SPna        */ {TO_TEMP | Pn, TO_TEMP | TSa, TO_DEST | T},
    /* 0D: PDSnaon     */ {TO_DEST | DSna, TO_DEST | DPon},
    /* 0E: PDSonon     */ {TO_DEST | DSon, TO_DEST | DPon},
    /* 0F: Pn          */ {TO_DEST | Pn},
    /* 10: PDSona      */ {TO_DEST | DSon, TO_DEST | DPa},
    /* 11: DSon        */ {TO_DEST | DSon},
    /* 12: SDPxnon     */ {TO_DEST | DPxn, TO_DEST | DSon},
    /* 13: SDPaon      */ {TO_DEST | DPa, TO_DEST | DSon},
    /* 14: DPSxnon     */ {TO_TEMP | S, TO_TEMP | TPxn, TO_DEST | DTon},
    /* 15: DPSaon      */ {TO_TEMP | S, TO_TEMP | TPa, TO_DEST | DTon},
    /* 16: PSDPSanaxx  */ {TO_TEMP | S, TO_TEMP | TPan, TO_TEMP | DTa, TO_TEMP | TSx, TO_TEMP | TPx, TO_DEST | T},
    /* 17: SSPxDSxaxn  */ {TO_TEMP | S, TO_TEMP | TPx, TO_DEST | DSx, TO_DEST | DTa, TO_DEST | DSxn},
    /* 18: SPxPDxa     */ {TO_TEMP | S, TO_TEMP | TPx, TO_DEST | DPx, TO_DEST | DTa},
    /* 19: SDPSanaxn   */ {TO_TEMP | S, TO_TEMP | TPan, TO_TEMP | DTa, TO_TEMP | TSxn, TO_DEST | T},
    /* 1A: PDSPaox     */ {TO_TEMP | P, TO_TEMP | TSa, TO_TEMP | DTo, TO_TEMP | TPx, TO_DEST | T},
    /* 1B: SDPSxaxn    */ {TO_TEMP | S, TO_TEMP | TPx, TO_TEMP | DTa, TO_TEMP | TSxn, TO_DEST | T},
    /* 1C: PSDPaox     */ {TO_DEST | DPa, TO_DEST | DSo, TO_DEST | DPx},
    /* 1D: DSPDxaxn    */ {TO_TEMP | D, TO_TEMP | TPx, TO_TEMP | TSa, TO_DEST | DTxn},
    /* 1E: PDSox       */ {TO_DEST | DSo, TO_DEST | DPx},
    /* 1F: PDSoan      */ {TO_DEST | DSo, TO_DEST | DPan},
    /* 20: DPSnaa      */ {TO_TEMP | Sn, TO_TEMP | TPa, TO_DEST | DTa},
    /* 21: SDPxon      */ {TO_DEST | DPx, TO_DEST | DSon},
    /* 22: DSna        */ {TO_DEST | DSna},
    /* 23: SPDnaon     */ {TO_DEST | PDna, TO_DEST | DSon},
    /* 24: SPxDSxa     */ {TO_TEMP | S, TO_TEMP | TPx, TO_DEST | DSx, TO_DEST | DTa},
    /* 25: PDSPanaxn   */ {TO_TEMP | P, TO_TEMP | TSan, TO_TEMP | DTa, TO_TEMP | TPxn, TO_DEST | T},
    /* 26: SDPSaox     */ {TO_TEMP | S, TO_TEMP | TPa, TO_TEMP | DTo, TO_TEMP | TSx, TO_DEST | T},
    /* 27: SDPSxnox    */ {TO_TEMP | S, TO_TEMP | TPxn, TO_TEMP | DTo, TO_TEMP | TSx, TO_DEST | T},
    /* 28: DPSxa       */ {TO_TEMP | S, TO_TEMP | TPx, TO_DEST | DTa},
    /* 29: PSDPSaoxxn  */ {TO_TEMP | S, TO_TEMP | TPa, TO_TEMP | DTo, TO_TEMP | TSx, TO_TEMP | TPxn, TO_DEST | T},
    /* 2A: DPSana      */ {TO_TEMP | S, TO_TEMP | TPan, TO_DEST | DTa},
    /* 2B: SSPxPDxaxn  */ {TO_TEMP | S, TO_TEMP | TPx, TO_DEST | DPx, TO_DEST | DTa, TO_DEST | DSxn},
    /* 2C: SPDSoax     */ {TO_DEST | DSo, TO_DEST | DPa, TO_DEST | DSx},
    /* 2D: PSDnox      */ {TO_DEST | SDno, TO_DEST | DPx},
    /* 2E: PSDPxox     */ {TO_DEST | DPx, TO_DEST | DSo, TO_DEST | DPx},
    /* 2F: PSDnoan     */ {TO_DEST | SDno, TO_DEST | DPan},
    /* 30: PSna        */ {TO_TEMP | Sn, TO_TEMP | TPa, TO_DEST | T},
    /* 31: SDPnaon     */ {TO_DEST | DPna, TO_DEST | DSon},
    /* 32: SDPSoox     */ {TO_TEMP | S, TO_TEMP | TPo, TO_TEMP | DTo, TO_TEMP | TSx, TO_DEST | T},
    /* 33: Sn          */ {TO_DEST | Sn},
    /* 34: SPDSaox     */ {TO_DEST | DSa, TO_DEST | DPo, TO_DEST | DSx},
    /* 35: SPDSxnox    */ {TO_DEST | DSxn, TO_DEST | DPo, TO_DEST | DSx},
    /* 36: SDPox       */ {TO_DEST | DPo, TO_DEST | DSx},
    /* 37: SDPoan      */ {TO_DEST | DPo, TO_DEST | DSan},
    /* 38: PSDPoax     */ {TO_DEST | DPo, TO_DEST | DSa, TO_DEST | DPx},
    /* 39: SPDnox      */ {TO_DEST | PDno, TO_DEST | DSx},
    /* 3A: SPDSxox     */ {TO_DEST | DSx, TO_DEST | DPo, TO_DEST | DSx},
    /* 3B: SPDnoan     */ {TO_DEST | PDno, TO_DEST | DSan},
    /* 3C: PSx         */ {TO_DEST | S, TO_DEST | DPx},
    /* 3D: SPDSonox    */ {TO_DEST | DSon, TO_DEST | DPo, TO_DEST | DSx},
    /* 3E: SPDSnaox    */ {TO_DEST | DSna, TO_DEST | DPo, TO_DEST | DSx},
    /* 3F: PSan        */ {TO_DEST | S, TO_DEST | DPan},
    /* 40: PSDnaa      */ {TO_DEST | SDna, TO_DEST | DPa},
    /* 41: DPSxon      */ {TO_TEMP | S, TO_TEMP | TPx, TO_DEST | DTon},
    /* 42: SDxPDxa     */ {TO_TEMP | D, TO_TEMP | TSx, TO_DEST | DPx, TO_DEST | DTa},
    /* 43: SPDSanaxn   */ {TO_DEST | DSan, TO_DEST | DPa, TO_DEST | DSxn},
    /* 44: SDna        */ {TO_DEST | SDna},
    /* 45: DPSnaon     */ {TO_TEMP | Sn, TO_TEMP | TPa, TO_DEST | DTon},
    /* 46: DSPDaox     */ {TO_TEMP | D, TO_TEMP | TPa, TO_TEMP | TSo, TO_DEST | DTx},
    /* 47: PSDPxaxn    */ {TO_DEST | DPx, TO_DEST | DSa, TO_DEST | DPxn},
    /* 48: SDPxa       */ {TO_DEST | DPx, TO_DEST | DSa},
    /* 49: PDSPDaoxxn  */ {TO_TEMP | D, TO_TEMP | TPa, TO_TEMP | TSo, TO_TEMP | DTx, TO_TEMP | TPxn, TO_DEST | T},
    /* 4A: DPSDoax     */ {TO_TEMP | D, TO_TEMP | TSo, TO_TEMP | TPa, TO_DEST | DTx},
    /* 4B: PDSnox      */ {TO_DEST | DSno, TO_DEST | DPx},
    /* 4C: SDPana      */ {TO_DEST | DPan, TO_DEST | DSa},
    /* 4D: SSPxDSxoxn  */ {TO_TEMP | S, TO_TEMP | TPx, TO_DEST | DSx, TO_DEST | DTo, TO_DEST | DSxn},
    /* 4E: PDSPxox     */ {TO_TEMP | P, TO_TEMP | TSx, TO_TEMP | DTo, TO_TEMP | TPx, TO_DEST | T},
    /* 4F: PDSnoan     */ {TO_DEST | DSno, TO_DEST | DPan},
    /* 50: PDna        */ {TO_DEST | PDna},
    /* 51: DSPnaon     */ {TO_TEMP | Pn, TO_TEMP | TSa, TO_DEST | DTon},
    /* 52: DPSDaox     */ {TO_TEMP | D, TO_TEMP | TSa, TO_TEMP | TPo, TO_DEST | DTx},
    /* 53: SPDSxaxn    */ {TO_DEST | DSx, TO_DEST | DPa, TO_DEST | DSxn},
    /* 54: DPSonon     */ {TO_TEMP | S, TO_TEMP | TPon, TO_DEST | DTon},
    /* 55: Dn          */ {0},
    /* 56: DPSox       */ {TO_TEMP | S, TO_TEMP | TPo, TO_DEST | DTx},
    /* 57: DPSoan      */ {TO_TEMP | S, TO_TEMP | TPo, TO_DEST | DTan},
    /* 58: PDSPoax     */ {TO_TEMP | P, TO_TEMP | TSo, TO_TEMP | DTa, TO_TEMP | TPx, TO_DEST | T},
    /* 59: DPSnox      */ {TO_TEMP | Sn, TO_TEMP | TPo, TO_DEST | DTx},
    /* 5A: DPx         */ {TO_DEST | DPx},
    /* 5B: DPSDonox    */ {TO_TEMP | D, TO_TEMP | TSon, TO_TEMP | TPo, TO_DEST | DTx},
    /* 5C: DPSDxox     */ {TO_TEMP | D, TO_TEMP | TSx, TO_TEMP | TPo, TO_DEST | DTx},
    /* 5D: DPSnoan     */ {TO_TEMP | Sn, TO_TEMP | TPo, TO_DEST | DTan},
    /* 5E: DPSDnaox    */ {TO_TEMP | D, TO_TEMP | STna, TO_TEMP | TPo, TO_DEST | DTx},
    /* 5F: DPan        */ {TO_DEST | DPan},
    /* 60: PDSxa       */ {TO_DEST | DSx, TO_DEST | DPa},
    /* 61: DSPDSaoxxn  */ {TO_TEMP | D, TO_TEMP | TSa, TO_TEMP | TPo, TO_TEMP | TSx, TO_DEST | DTxn},
    /* 62: DSPDoax     */ {TO_TEMP | D, TO_TEMP | TPo, TO_TEMP | TSa, TO_DEST | DTx},
    /* 63: SDPnox      */ {TO_DEST | DPno, TO_DEST | DSx},
    /* 64: SDPSoax     */ {TO_TEMP | S, TO_TEMP | TPo, TO_TEMP | DTa, TO_TEMP | TSx, TO_DEST | T},
    /* 65: DSPnox      */ {TO_TEMP | Pn, TO_TEMP | TSo, TO_DEST | DTx},
    /* 66: DSx         */ {TO_DEST | DSx},
    /* 67: SDPSonox    */ {TO_TEMP | S, TO_TEMP | TPon, TO_TEMP | DTo, TO_TEMP | TSx, TO_DEST | T},
    /* 68: DSPDSonoxxn */ {TO_TEMP | D, TO_TEMP | TSon, TO_TEMP | TPo, TO_TEMP | TSx, TO_DEST | DTxn},
    /* 69: PDSxxn      */ {TO_DEST | DSx, TO_DEST | DPxn},
    /* 6A: DPSax       */ {TO_TEMP | S, TO_TEMP | TPa, TO_DEST | DTx},
    /* 6B: PSDPSoaxxn  */ {TO_TEMP | S, TO_TEMP | TPo, TO_TEMP | DTa, TO_TEMP | TSx, TO_TEMP | TPxn, TO_DEST | T},
    /* 6C: SDPax       */ {TO_DEST | DPa, TO_DEST | DSx},
    /* 6D: PDSPDoaxxn  */ {TO_TEMP | D, TO_TEMP | TPo, TO_TEMP | TSa, TO_TEMP | DTx, TO_TEMP | TPxn, TO_DEST | T},
    /* 6E: SDPSnoax    */ {TO_TEMP | Sn, TO_TEMP | TPo, TO_TEMP | DTa, TO_TEMP | TSx, TO_DEST | T},
    /* 6F: PDSxnan     */ {TO_DEST | DSxn, TO_DEST | DPan},
    /* 70: PDSana      */ {TO_DEST | DSan, TO_DEST | DPa},
    /* 71: SSDxPDxaxn  */ {TO_TEMP | D, TO_TEMP | TSx, TO_DEST | DPx, TO_DEST | DTa, TO_DEST | DSxn},
    /* 72: SDPSxox     */ {TO_TEMP | S, TO_TEMP | TPx, TO_TEMP | DTo, TO_TEMP | TSx, TO_DEST | T},
    /* 73: SDPnoan     */ {TO_DEST | DPno, TO_DEST | DSan},
    /* 74: DSPDxox     */ {TO_TEMP | D, TO_TEMP | TPx, TO_TEMP | TSo, TO_DEST | DTx},
    /* 75: DSPnoan     */ {TO_TEMP | Pn, TO_TEMP | TSo, TO_DEST | DTan},
    /* 76: SDPSnaox    */ {TO_TEMP | Sn, TO_TEMP | TPa, TO_TEMP | DTo, TO_TEMP | TSx, TO_DEST | T},
    /* 77: DSan        */ {TO_DEST | DSan},
    /* 78: PDSax       */ {TO_DEST | DSa, TO_DEST | DPx},
    /* 79: DSPDSoaxxn  */ {TO_TEMP | D, TO_TEMP | TSo, TO_TEMP | TPa, TO_TEMP | TSx, TO_DEST | DTxn},
    /* 7A: DPSDnoax    */ {TO_TEMP | D, TO_TEMP | STno, TO_TEMP | TPa, TO_DEST | DTx},
    /* 7B: SDPxnan     */ {TO_DEST | DPxn, TO_DEST | DSan},
    /* 7C: SPDSnoax    */ {TO_DEST | DSno, TO_DEST | DPa, TO_DEST | DSx},
    /* 7D: DPSxnan     */ {TO_TEMP | S, TO_TEMP | TPxn, TO_DEST | DTan},
    /* 7E: SPxDSxo     */ {TO_TEMP | S, TO_TEMP | TPx, TO_DEST | DSx, TO_DEST | DTo},
    /* 7F: DPSaan      */ {TO_DEST | DPa, TO_DEST | DSan},
    /* 80: DPSaa       */ {TO_DEST | DPa, TO_DEST | DSa},
    /* 81: SPxDSxon    */ {TO_TEMP | S, TO_TEMP | TPx, TO_DEST | DSx, TO_DEST | DTon},
    /* 82: DPSxna      */ {TO_TEMP | S, TO_TEMP | TPxn, TO_DEST | DTa},
    /* 83: SPDSnoaxn   */ {TO_DEST | DSno, TO_DEST | DPa, TO_DEST | DSxn},
    /* 84: SDPxna      */ {TO_DEST | DPxn, TO_DEST | DSa},
    /* 85: PDSPnoaxn   */ {TO_TEMP | Pn, TO_TEMP | TSo, TO_TEMP | DTa, TO_TEMP | TPxn, TO_DEST | T},
    /* 86: DSPDSoaxx   */ {TO_TEMP | D, TO_TEMP | TSo, TO_TEMP | TPa, TO_TEMP | TSx, TO_DEST | DTx},
    /* 87: PDSaxn      */ {TO_DEST | DSa, TO_DEST | DPxn},
    /* 88: DSa         */ {TO_DEST | DSa},
    /* 89: SDPSnaoxn   */ {TO_TEMP | Sn, TO_TEMP | TPa, TO_TEMP | DTo, TO_TEMP | TSxn, TO_DEST | T},
    /* 8A: DSPnoa      */ {TO_TEMP | Pn, TO_TEMP | TSo, TO_DEST | DTa},
    /* 8B: DSPDxoxn    */ {TO_TEMP | D, TO_TEMP | TPx, TO_TEMP | TSo, TO_DEST | DTxn},
    /* 8C: SDPnoa      */ {TO_DEST | DPno, TO_DEST | DSa},
    /* 8D: SDPSxoxn    */ {TO_TEMP | S, TO_TEMP | TPx, TO_TEMP | DTo, TO_TEMP | TSxn, TO_DEST | T},
    /* 8E: SSDxPDxax   */ {TO_TEMP | D, TO_TEMP | TSx, TO_DEST | DPx, TO_DEST | DTa, TO_DEST | DSx},
    /* 8F: PDSanan     */ {TO_DEST | DSan, TO_DEST | DPan},
    /* 90: PDSxna      */ {TO_DEST | DSxn, TO_DEST | DPa},
    /* 91: SDPSnoaxn   */ {TO_TEMP | Sn, TO_TEMP | TPo, TO_TEMP | DTa, TO_TEMP | TSxn, TO_DEST | T},
    /* 92: DPSDPoaxx   */ {TO_TEMP | D, TO_TEMP | TPo, TO_TEMP | TSa, TO_TEMP | TPx, TO_DEST | DTx},
    /* 93: SPDaxn      */ {TO_DEST | DPa, TO_DEST | DSxn},
    /* 94: PSDPSoaxx   */ {TO_TEMP | S, TO_TEMP | TPo, TO_TEMP | DTa, TO_TEMP | TSx, TO_TEMP | TPx, TO_DEST | T},
    /* 95: DPSaxn      */ {TO_TEMP | S, TO_TEMP | TPa, TO_DEST | DTxn},
    /* 96: DPSxx       */ {TO_DEST | DPx, TO_DEST | DSx},
    /* 97: PSDPSonoxx  */ {TO_TEMP | S, TO_TEMP | TPon, TO_TEMP | DTo, TO_TEMP | TSx, TO_TEMP | TPx, TO_DEST | T},
    /* 98: SDPSonoxn   */ {TO_TEMP | S, TO_TEMP | TPon, TO_TEMP | DTo, TO_TEMP | TSxn, TO_DEST | T},
    /* 99: DSxn        */ {TO_DEST | DSxn},
    /* 9A: DPSnax      */ {TO_TEMP | Sn, TO_TEMP | TPa, TO_DEST | DTx},
    /* 9B: SDPSoaxn    */ {TO_TEMP | S, TO_TEMP | TPo, TO_TEMP | DTa, TO_TEMP | TSxn, TO_DEST | T},
    /* 9C: SPDnax      */ {TO_DEST | PDna, TO_DEST | DSx},
    /* 9D: DSPDoaxn    */ {TO_TEMP | D, TO_TEMP | TPo, TO_TEMP | TSa, TO_DEST | DTxn},
    /* 9E: DSPDSaoxx   */ {TO_TEMP | D, TO_TEMP | TSa, TO_TEMP | TPo, TO_TEMP | TSx, TO_DEST | DTx},
    /* 9F: PDSxan      */ {TO_DEST | DSx, TO_DEST | DPan},
    /* A0: DPa         */ {TO_DEST | DPa},
    /* A1: PDSPnaoxn   */ {TO_TEMP | Pn, TO_TEMP | TSa, TO_TEMP | DTo, TO_TEMP | TPxn, TO_DEST | T},
    /* A2: DPSnoa      */ {TO_TEMP | Sn, TO_TEMP | TPo, TO_DEST | DTa},
    /* A3: DPSDxoxn    */ {TO_TEMP | D, TO_TEMP | TSx, TO_TEMP | TPo, TO_DEST | DTxn},
    /* A4: PDSPonoxn   */ {TO_TEMP | P, TO_TEMP | TSon, TO_TEMP | DTo, TO_TEMP | TPxn, TO_DEST | T},
    /* A5: PDxn        */ {TO_DEST | DPxn},
    /* A6: DSPnax      */ {TO_TEMP | Pn, TO_TEMP | TSa, TO_DEST | DTx},
    /* A7: PDSPoaxn    */ {TO_TEMP | P, TO_TEMP | TSo, TO_TEMP | DTa, TO_TEMP | TPxn, TO_DEST | T},
    /* A8: DPSoa       */ {TO_TEMP | S, TO_TEMP | TPo, TO_DEST | DTa},
    /* A9: DPSoxn      */ {TO_TEMP | S, TO_TEMP | TPo, TO_DEST | DTxn},
    /* AA: D           */ {0},
    /* AB: DPSono      */ {TO_TEMP | S, TO_TEMP | TPon, TO_DEST | DTo},
    /* AC: SPDSxax     */ {TO_DEST | DSx, TO_DEST | DPa, TO_DEST | DSx},
    /* AD: DPSDaoxn    */ {TO_TEMP | D, TO_TEMP | TSa, TO_TEMP | TPo, TO_DEST | DTxn},
    /* AE: DSPnao      */ {TO_TEMP | Pn, TO_TEMP | TSa, TO_DEST | DTo},
    /* AF: DPno        */ {TO_DEST | DPno},
    /* B0: PDSnoa      */ {TO_DEST | DSno, TO_DEST | DPa},
    /* B1: PDSPxoxn    */ {TO_TEMP | P, TO_TEMP | TSx, TO_TEMP | DTo, TO_TEMP | TPxn, TO_DEST | T},
    /* B2: SSPxDSxox   */ {TO_TEMP | S, TO_TEMP | TPx, TO_DEST | DSx, TO_DEST | DTo, TO_DEST | DSx},
    /* B3: SDPanan     */ {TO_DEST | DPan, TO_DEST | DSan},
    /* B4: PSDnax      */ {TO_DEST | SDna, TO_DEST | DPx},
    /* B5: DPSDoaxn    */ {TO_TEMP | D, TO_TEMP | TSo, TO_TEMP | TPa, TO_DEST | DTxn},
    /* B6: DPSDPaoxx   */ {TO_TEMP | D, TO_TEMP | TPa, TO_TEMP | TSo, TO_TEMP | TPx, TO_DEST | DTx},
    /* B7: SDPxan      */ {TO_DEST | DPx, TO_DEST | DSan},
    /* B8: PSDPxax     */ {TO_DEST | DPx, TO_DEST | DSa, TO_DEST | DPx},
    /* B9: DSPDaoxn    */ {TO_TEMP | D, TO_TEMP | TPa, TO_TEMP | TSo, TO_DEST | DTxn},
    /* BA: DPSnao      */ {TO_TEMP | Sn, TO_TEMP | TPa, TO_DEST | DTo},
    /* BB: DSno        */ {TO_DEST | DSno},
    /* BC: SPDSanax    */ {TO_DEST | DSan, TO_DEST | DPa, TO_DEST | DSx},
    /* BD: SDxPDxan    */ {TO_TEMP | D, TO_TEMP | TSx, TO_DEST | DPx, TO_DEST | DTan},
    /* BE: DPSxo       */ {TO_TEMP | S, TO_TEMP | TPx, TO_DEST | DTo},
    /* BF: DPSano      */ {TO_TEMP | S, TO_TEMP | TPan, TO_DEST | DTo},
    /* C0: PSa         */ {TO_DEST | S, TO_DEST | DPa},
    /* C1: SPDSnaoxn   */ {TO_DEST | DSna, TO_DEST | DPo, TO_DEST | DSxn},
    /* C2: SPDSonoxn   */ {TO_DEST | DSon, TO_DEST | DPo, TO_DEST | DSxn},
    /* C3: PSxn        */ {TO_DEST | S, TO_DEST | DPxn},
    /* C4: SPDnoa      */ {TO_DEST | PDno, TO_DEST | DSa},
    /* C5: SPDSxoxn    */ {TO_DEST | DSx, TO_DEST | DPo, TO_DEST | DSxn},
    /* C6: SDPnax      */ {TO_DEST | DPna, TO_DEST | DSx},
    /* C7: PSDPoaxn    */ {TO_DEST | DPo, TO_DEST | DSa, TO_DEST | DPxn},
    /* C8: SDPoa       */ {TO_DEST | DPo, TO_DEST | DSa},
    /* C9: SPDoxn      */ {TO_DEST | DPo, TO_DEST | DSxn},
    /* CA: DPSDxax     */ {TO_TEMP | D, TO_TEMP | TSx, TO_TEMP | TPa, TO_DEST | DTx},
    /* CB: SPDSaoxn    */ {TO_DEST | DSa, TO_DEST | DPo, TO_DEST | DSxn},
    /* CC: S           */ {TO_DEST | S},
    /* CD: SDPono      */ {TO_DEST | DPon, TO_DEST | DSo},
    /* CE: SDPnao      */ {TO_DEST | DPna, TO_DEST | DSo},
    /* CF: SPno        */ {TO_TEMP | Pn, TO_TEMP | TSo, TO_DEST | T},
    /* D0: PSDnoa      */ {TO_DEST | SDno, TO_DEST | DPa},
    /* D1: PSDPxoxn    */ {TO_DEST | DPx, TO_DEST | DSo, TO_DEST | DPxn},
    /* D2: PDSnax      */ {TO_DEST | DSna, TO_DEST | DPx},
    /* D3: SPDSoaxn    */ {TO_DEST | DSo, TO_DEST | DPa, TO_DEST | DSxn},
    /* D4: SSPxPDxax   */ {TO_TEMP | S, TO_TEMP | TPx, TO_DEST | DPx, TO_DEST | DTa, TO_DEST | DSx},
    /* D5: DPSanan     */ {TO_TEMP | S, TO_TEMP | TPan, TO_DEST | DTan},
    /* D6: PSDPSaoxx   */ {TO_TEMP | S, TO_TEMP | TPa, TO_TEMP | DTo, TO_TEMP | TSx, TO_TEMP | TPx, TO_DEST | T},
    /* D7: DPSxan      */ {TO_TEMP | S, TO_TEMP | TPx, TO_DEST | DTan},
    /* D8: PDSPxax     */ {TO_TEMP | P, TO_TEMP | TSx, TO_TEMP | DTa, TO_TEMP | TPx, TO_DEST | T},
    /* D9: SDPSaoxn    */ {TO_TEMP | S, TO_TEMP | TPa, TO_TEMP | DTo, TO_TEMP | TSxn, TO_DEST | T},
    /* DA: DPSDanax    */ {TO_TEMP | D, TO_TEMP | TSan, TO_TEMP | TPa, TO_DEST | DTx},
    /* DB: SPxDSxan    */ {TO_TEMP | S, TO_TEMP | TPx, TO_DEST | DSx, TO_DEST | DTan},
    /* DC: SPDnao      */ {TO_DEST | PDna, TO_DEST | DSo},
    /* DD: SDno        */ {TO_DEST | SDno},
    /* DE: SDPxo       */ {TO_DEST | DPx, TO_DEST | DSo},
    /* DF: SDPano      */ {TO_DEST | DPan, TO_DEST | DSo},
    /* E0: PDSoa       */ {TO_DEST | DSo, TO_DEST | DPa},
    /* E1: PDSoxn      */ {TO_DEST | DSo, TO_DEST | DPxn},
    /* E2: DSPDxax     */ {TO_TEMP | D, TO_TEMP | TPx, TO_TEMP | TSa, TO_DEST | DTx},
    /* E3: PSDPaoxn    */ {TO_DEST | DPa, TO_DEST | DSo, TO_DEST | DPxn},
    /* E4: SDPSxax     */ {TO_TEMP | S, TO_TEMP | TPx, TO_TEMP | DTa, TO_TEMP | TSx, TO_DEST | T},
    /* E5: PDSPaoxn    */ {TO_TEMP | P, TO_TEMP | TSa, TO_TEMP | DTo, TO_TEMP | TPxn, TO_DEST | T},
    /* E6: SDPSanax    */ {TO_TEMP | S, TO_TEMP | TPan, TO_TEMP | DTa, TO_TEMP | TSx, TO_DEST | T},
    /* E7: SPxPDxan    */ {TO_TEMP | S, TO_TEMP | TPx, TO_DEST | DPx, TO_DEST | DTan},
    /* E8: SSPxDSxax   */ {TO_TEMP | S, TO_TEMP | TPx, TO_DEST | DSx, TO_DEST | DTa, TO_DEST | DSx},
    /* E9: DSPDSanaxxn */ {TO_TEMP | D, TO_TEMP | TSan, TO_TEMP | TPa, TO_TEMP | TSx, TO_DEST | DTxn},
    /* EA: DPSao       */ {TO_TEMP | S, TO_TEMP | TPa, TO_DEST | DTo},
    /* EB: DPSxno      */ {TO_TEMP | S, TO_TEMP | TPxn, TO_DEST | DTo},
    /* EC: SDPao       */ {TO_DEST | DPa, TO_DEST | DSo},
    /* ED: SDPxno      */ {TO_DEST | DPxn, TO_DEST | DSo},
    /* EE: DSo         */ {TO_DEST | DSo},
    /* EF: SDPnoo      */ {TO_DEST | DPno, TO_DEST | DSo},
    /* F0: P           */ {TO_DEST | P},
    /* F1: PDSono      */ {TO_DEST | DSon, TO_DEST | DPo},
    /* F2: PDSnao      */ {TO_DEST | DSna, TO_DEST | DPo},
    /* F3: PSno        */ {TO_TEMP | Sn, TO_TEMP | TPo, TO_DEST | T},
    /* F4: PSDnao      */ {TO_DEST | SDna, TO_DEST | DPo},
    /* F5: PDno        */ {TO_DEST | PDno},
    /* F6: PDSxo       */ {TO_DEST | DSx, TO_DEST | DPo},
    /* F7: PDSano      */ {TO_DEST | DSan, TO_DEST | DPo},
    /* F8: PDSao       */ {TO_DEST | DSa, TO_DEST | DPo},
    /* F9: PDSxno      */ {TO_DEST | DSxn, TO_DEST | DPo},
    /* FA: DPo         */ {TO_DEST | DPo},
    /* FB: DPSnoo      */ {TO_TEMP | Sn, TO_TEMP | TPo, TO_DEST | DTo},
    /* FC: PSo         */ {TO_DEST | S, TO_DEST | DPo},
    /* FD: PSDnoo      */ {TO_DEST | SDno, TO_DEST | DPo},
    /* FE: DPSoo       */ {TO_DEST | DPo, TO_DEST | DSo},
    /* FF: 1           */ {0},
    };

static void (NAPIP HW_DrawRect)(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
static void (NAPIP HW_DrawPattRect)(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
static void (NAPIP HW_DrawColorPattRect)(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
static void (NAPIP HW_BitBlt)(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
static void (NAPIP HW_BitBltLin)(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
static void (NAPIP HW_BitBltSys)(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,N_int32 flipY);
static GA_ROP3EmuFuncs  HW;

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
REMARKS:
Sets up the ROP3 emulator as nececessary for the current color depth.
****************************************************************************/
void DRV_SetupROP3Emulator(
    GA_ROP3EmuFuncs *emuFuncs)
{
    switch (DV.cntMode.bits) {
        case 8:
            HW_DrawRect             = DV.draw_8bpp.DrawRect;
            HW_DrawPattRect         = DV.draw_8bpp.DrawPattRect;
            HW_DrawColorPattRect    = DV.draw_8bpp.DrawColorPattRect;
            HW_BitBlt               = DV.draw_8bpp.BitBlt;
            HW_BitBltLin            = DV.draw_8bpp.BitBltLin;
            HW_BitBltSys            = DV.draw_8bpp.BitBltSys;
            break;
        case 15:
        case 16:
            HW_DrawRect             = DV.draw_16bpp.DrawRect;
            HW_DrawPattRect         = DV.draw_16bpp.DrawPattRect;
            HW_DrawColorPattRect    = DV.draw_16bpp.DrawColorPattRect;
            HW_BitBlt               = DV.draw_16bpp.BitBlt;
            HW_BitBltLin            = DV.draw_16bpp.BitBltLin;
            HW_BitBltSys            = DV.draw_16bpp.BitBltSys;
            break;
        case 24:
            HW_DrawRect             = DV.draw_24bpp.DrawRect;
            HW_DrawPattRect         = DV.draw_24bpp.DrawPattRect;
            HW_DrawColorPattRect    = DV.draw_24bpp.DrawColorPattRect;
            HW_BitBlt               = DV.draw_24bpp.BitBlt;
            HW_BitBltLin            = DV.draw_24bpp.BitBltLin;
            HW_BitBltSys            = DV.draw_24bpp.BitBltSys;
            break;
        case 32:
            HW_DrawRect             = DV.draw_32bpp.DrawRect;
            HW_DrawPattRect         = DV.draw_32bpp.DrawPattRect;
            HW_DrawColorPattRect    = DV.draw_32bpp.DrawColorPattRect;
            HW_BitBlt               = DV.draw_32bpp.BitBlt;
            HW_BitBltLin            = DV.draw_32bpp.BitBltLin;
            HW_BitBltSys            = DV.draw_32bpp.BitBltSys;
            break;
        }
    HW = *emuFuncs;
}

// Code to handle special case ROP3's that some hardware screws up

#define HANDLE_SPECIAL_ROPS()                               \
    if (pass[0] == 0) {                                     \
        switch (rop3) {                                     \
            case GA_R3_0:                                   \
                DV.state2d.SetMix(GA_R2_BLACK);             \
                HW_DrawRect(dstLeft,dstTop,width,height);   \
                HW.RestoreMix();                            \
                return;                                     \
            case GA_R3_D:                                   \
                HW.RestoreMix();                            \
                return;                                     \
            case GA_R3_Dn:                                  \
                DV.state2d.SetMix(GA_R2_NOT);               \
                HW_DrawRect(dstLeft,dstTop,width,height);   \
                HW.RestoreMix();                            \
                return;                                     \
            case GA_R3_1:                                   \
                DV.state2d.SetMix(GA_R2_WHITE);             \
                HW_DrawRect(dstLeft,dstTop,width,height);   \
                HW.RestoreMix();                            \
                return;                                     \
            }                                               \
        }

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the source rectangle to copy
top     - Top coordinate of the source rectangle to copy
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
dstLeft - Left coordinate of destination
dstTop  - Top coordinate of destination
mix     - Mix code for the copy (GA_mixCodesType)

REMARKS:
This function copies a rectangular region of video memory from one location
to another.
****************************************************************************/
void NAPI DRV_BitBltPatt(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    ushort  *pass = &rop3_table[rop3].pass1;
    int     i,bottom;

    HW.SaveMix();
    HANDLE_SPECIAL_ROPS();
    if ((pass[0] & TO_TEMP) || (pass[1] & TO_TEMP)) {
        // Process a scanline at a time using the temporary offscreen buffer
        for (bottom = top+height; top < bottom; top++, dstTop++) {
            pass = &rop3_table[rop3].pass1;
            for (i = 0; i < 6 && *pass; i++, pass++) {
                if (*pass & TO_DEST) {
                    // Draw this pass to destination buffer
                    if (*pass & FROM_TEMP) {
                        HW.BltFromTemp(dstLeft,width,dstLeft,dstTop,PASS_MIX);
                        }
                    else if (*pass & FROM_SRC) {
                        HW_BitBlt(left,top,width,1,dstLeft,dstTop,PASS_MIX);
                        }
                    else if (*pass & FROM_PATT) {
                        DV.state2d.SetMix(PASS_MIX);
                        HW_DrawPattRect(dstLeft,dstTop,width,1);
                        }
                    }
                else {
                    // Draw this pass to temporary buffer
                    if (*pass & FROM_SRC) {
                        HW.BltToTemp(left,top,width,dstLeft,PASS_MIX);
                        }
                    else if (*pass & FROM_PATT) {
                        DV.state2d.SetMix(PASS_MIX);
                        HW.DrawPattToTemp(dstLeft,dstTop,width);
                        }
                    else if (*pass & FROM_DEST) {
                        HW.BltToTemp(dstLeft,dstTop,width,dstLeft,PASS_MIX);
                        }
                    }
                }
            }
        }
    else {
        // ROP is handled with multiple passes on the destination
        // framebuffer, so we can do this with whole blits at a time.
        for (i = 0; i < 6 && *pass; i++, pass++) {
            if (*pass & FROM_SRC) {
                HW_BitBlt(left,top,width,height,dstLeft,dstTop,PASS_MIX);
                }
            else if (*pass & FROM_PATT) {
                DV.state2d.SetMix(PASS_MIX);
                HW_DrawPattRect(dstLeft,dstTop,width,height);
                }
            }
        }
    HW.RestoreMix();
}

/****************************************************************************
PARAMETERS:
srcOfs      - Offset of source bitmap in video memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3            - Mix code for the copy (GA_mixCodesType)

REMARKS:
This routine will copy a linear region of video memory from srcOfs from the
start of video memory to the destination rectangle (dstLeft, dstTop,
dstLeft+width-1, dstTop+height-1) with the specified rop3.
****************************************************************************/
void NAPI DRV_BitBltPattLin(
    N_int32 srcOfs,
    N_int32 srcPitch,
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    ushort  *pass = &rop3_table[rop3].pass1;
    int     i,bottom;

    HW.SaveMix();
    HANDLE_SPECIAL_ROPS();
    if ((pass[0] & TO_TEMP) || (pass[1] & TO_TEMP)) {
        // Process a scanline at a time using the temporary offscreen buffer
        for (bottom = top+height; top < bottom; top++, dstTop++) {
            pass = &rop3_table[rop3].pass1;
            for (i = 0; i < 6 && *pass; i++, pass++) {
                if (*pass & TO_DEST) {
                    // Draw this pass to destination buffer
                    if (*pass & FROM_TEMP) {
                        HW.BltFromTemp(dstLeft,width,dstLeft,dstTop,PASS_MIX);
                        }
                    else if (*pass & FROM_SRC) {
                        HW_BitBltLin(srcOfs,srcPitch,left,top,width,1,dstLeft,dstTop,PASS_MIX);
                        }
                    else if (*pass & FROM_PATT) {
                        DV.state2d.SetMix(PASS_MIX);
                        HW_DrawPattRect(dstLeft,dstTop,width,1);
                        }
                    }
                else {
                    // Draw this pass to temporary buffer
                    if (*pass & FROM_SRC) {
                        HW.BltToTempLin(srcOfs,srcPitch,left,top,width,dstLeft,PASS_MIX);
                        }
                    else if (*pass & FROM_PATT) {
                        DV.state2d.SetMix(PASS_MIX);
                        HW.DrawPattToTemp(dstLeft,dstTop,width);
                        }
                    else if (*pass & FROM_DEST) {
                        HW.BltToTemp(dstLeft,dstTop,width,dstLeft,PASS_MIX);
                        }
                    }
                }
            }
        }
    else {
        // ROP is handled with multiple passes on the destination
        // framebuffer, so we can do this with whole blits at a time.
        for (i = 0; i < 6 && *pass; i++, pass++) {
            if (*pass & FROM_SRC) {
                HW_BitBltLin(srcOfs,srcPitch,left,top,width,height,dstLeft,dstTop,PASS_MIX);
                }
            else if (*pass & FROM_PATT) {
                DV.state2d.SetMix(PASS_MIX);
                HW_DrawPattRect(dstLeft,dstTop,width,height);
                }
            }
        }
    HW.RestoreMix();
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3            - Mix code for the copy (GA_mixCodesType)
flipY       - True if the bitmap should be flipped in Y direction

REMARKS:
This routine will copy a bitmap from system memory with a starting address
of srcAddr to the destination rectangle (dstLeft, dstTop, dstLeft+width-1,
dstTop+height-1) with the specified rop3. The rop3 code will be used to
combine the source bitmap data with the pixels in the destination bitmap.
****************************************************************************/
void NAPI DRV_BitBltPattSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3,
    N_int32 flipY)
{
    ushort  *pass = &rop3_table[rop3].pass1;
    int     i,bottom,dstInc;

    HW.SaveMix();
    HANDLE_SPECIAL_ROPS();
    if ((pass[0] & TO_TEMP) || (pass[1] & TO_TEMP) || (flipY && PASS_MIX != R2_0 && PASS_MIX != R2_1)) {
        // Process a scanline at a time using the temporary offscreen buffer
        if (flipY) {
            dstTop += (height-1);
            dstInc = -1;
            }
        else
            dstInc = 1;
        for (bottom = top+height; top < bottom; top++, dstTop += dstInc) {
            pass = &rop3_table[rop3].pass1;
            for (i = 0; i < 6 && *pass; i++, pass++) {
                if (*pass & TO_DEST) {
                    // Draw this pass to destination buffer
                    if (*pass & FROM_TEMP) {
                        HW.BltFromTemp(dstLeft,width,dstLeft,dstTop,PASS_MIX);
                        }
                    else if (*pass & FROM_SRC) {
                        HW_BitBltSys(srcAddr,srcPitch,left,top,width,1,dstLeft,dstTop,PASS_MIX,false);
                        }
                    else if (*pass & FROM_PATT) {
                        DV.state2d.SetMix(PASS_MIX);
                        HW_DrawPattRect(dstLeft,dstTop,width,1);
                        }
                    }
                else {
                    // Draw this pass to temporary buffer
                    if (*pass & FROM_SRC) {
                        HW.BltToTempSys(srcAddr,srcPitch,left,top,width,dstLeft,PASS_MIX);
                        }
                    else if (*pass & FROM_PATT) {
                        DV.state2d.SetMix(PASS_MIX);
                        HW.DrawPattToTemp(dstLeft,dstTop,width);
                        }
                    else if (*pass & FROM_DEST) {
                        HW.BltToTemp(dstLeft,dstTop,width,dstLeft,PASS_MIX);
                        }
                    }
                }
            }
        }
    else {
        // ROP is handled with multiple passes on the destination
        // framebuffer, so we can do this with whole blits at a time.
        for (i = 0; i < 6 && *pass; i++, pass++) {
            if (*pass & FROM_SRC) {
                HW_BitBltSys(srcAddr,srcPitch,left,top,width,height,dstLeft,dstTop,PASS_MIX,flipY);
                }
            else if (*pass & FROM_PATT) {
                DV.state2d.SetMix(PASS_MIX);
                HW_DrawPattRect(dstLeft,dstTop,width,height);
                }
            }
        }
    HW.RestoreMix();
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the source rectangle to copy
top     - Top coordinate of the source rectangle to copy
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
dstLeft - Left coordinate of destination
dstTop  - Top coordinate of destination
mix     - Mix code for the copy (GA_mixCodesType)

REMARKS:
This function copies a rectangular region of video memory from one location
to another.
****************************************************************************/
void NAPI DRV_BitBltColorPatt(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    ushort  *pass = &rop3_table[rop3].pass1;
    int     i,bottom;

    HW.SaveMix();
    HANDLE_SPECIAL_ROPS();
    if ((pass[0] & TO_TEMP) || (pass[1] & TO_TEMP)) {
        // Process a scanline at a time using the temporary offscreen buffer
        for (bottom = top+height; top < bottom; top++, dstTop++) {
            pass = &rop3_table[rop3].pass1;
            for (i = 0; i < 6 && *pass; i++, pass++) {
                if (*pass & TO_DEST) {
                    // Draw this pass to destination buffer
                    if (*pass & FROM_TEMP) {
                        HW.BltFromTemp(dstLeft,width,dstLeft,dstTop,PASS_MIX);
                        }
                    else if (*pass & FROM_SRC) {
                        HW_BitBlt(left,top,width,1,dstLeft,dstTop,PASS_MIX);
                        }
                    else if (*pass & FROM_PATT) {
                        DV.state2d.SetMix(PASS_MIX);
                        HW_DrawColorPattRect(dstLeft,dstTop,width,1);
                        }
                    }
                else {
                    // Draw this pass to temporary buffer
                    if (*pass & FROM_SRC) {
                        HW.BltToTemp(left,top,width,dstLeft,PASS_MIX);
                        }
                    else if (*pass & FROM_PATT) {
                        DV.state2d.SetMix(PASS_MIX);
                        HW.DrawColorPattToTemp(dstLeft,dstTop,width);
                        }
                    else if (*pass & FROM_DEST) {
                        HW.BltToTemp(dstLeft,dstTop,width,dstLeft,PASS_MIX);
                        }
                    }
                }
            }
        }
    else {
        // ROP is handled with multiple passes on the destination
        // framebuffer, so we can do this with whole blits at a time.
        for (i = 0; i < 6 && *pass; i++, pass++) {
            if (*pass & FROM_SRC) {
                HW_BitBlt(left,top,width,height,dstLeft,dstTop,PASS_MIX);
                }
            else if (*pass & FROM_PATT) {
                DV.state2d.SetMix(PASS_MIX);
                HW_DrawColorPattRect(dstLeft,dstTop,width,height);
                }
            }
        }
    HW.RestoreMix();
}

/****************************************************************************
PARAMETERS:
srcOfs      - Offset of source bitmap in video memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3            - Mix code for the copy (GA_mixCodesType)

REMARKS:
This routine will copy a linear region of video memory from srcOfs from the
start of video memory to the destination rectangle (dstLeft, dstTop,
dstLeft+width-1, dstTop+height-1) with the specified rop3.
****************************************************************************/
void NAPI DRV_BitBltColorPattLin(
    N_int32 srcOfs,
    N_int32 srcPitch,
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    ushort  *pass = &rop3_table[rop3].pass1;
    int     i,bottom;

    HW.SaveMix();
    HANDLE_SPECIAL_ROPS();
    if ((pass[0] & TO_TEMP) || (pass[1] & TO_TEMP)) {
        // Process a scanline at a time using the temporary offscreen buffer
        for (bottom = top+height; top < bottom; top++, dstTop++) {
            pass = &rop3_table[rop3].pass1;
            for (i = 0; i < 6 && *pass; i++, pass++) {
                if (*pass & TO_DEST) {
                    // Draw this pass to destination buffer
                    if (*pass & FROM_TEMP) {
                        HW.BltFromTemp(dstLeft,width,dstLeft,dstTop,PASS_MIX);
                        }
                    else if (*pass & FROM_SRC) {
                        HW_BitBltLin(srcOfs,srcPitch,left,top,width,1,dstLeft,dstTop,PASS_MIX);
                        }
                    else if (*pass & FROM_PATT) {
                        DV.state2d.SetMix(PASS_MIX);
                        HW_DrawColorPattRect(dstLeft,dstTop,width,1);
                        }
                    }
                else {
                    // Draw this pass to temporary buffer
                    if (*pass & FROM_SRC) {
                        HW.BltToTempLin(srcOfs,srcPitch,left,top,width,dstLeft,PASS_MIX);
                        }
                    else if (*pass & FROM_PATT) {
                        DV.state2d.SetMix(PASS_MIX);
                        HW.DrawColorPattToTemp(dstLeft,dstTop,width);
                        }
                    else if (*pass & FROM_DEST) {
                        HW.BltToTemp(dstLeft,dstTop,width,dstLeft,PASS_MIX);
                        }
                    }
                }
            }
        }
    else {
        // ROP is handled with multiple passes on the destination
        // framebuffer, so we can do this with whole blits at a time.
        for (i = 0; i < 6 && *pass; i++, pass++) {
            if (*pass & FROM_SRC) {
                HW_BitBltLin(srcOfs,srcPitch,left,top,width,height,dstLeft,dstTop,PASS_MIX);
                }
            else if (*pass & FROM_PATT) {
                DV.state2d.SetMix(PASS_MIX);
                HW_DrawColorPattRect(dstLeft,dstTop,width,height);
                }
            }
        }
    HW.RestoreMix();
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3            - Mix code for the copy (GA_mixCodesType)
flipY       - True if the bitmap should be flipped in Y direction

REMARKS:
This routine will copy a bitmap from system memory with a starting address
of srcAddr to the destination rectangle (dstLeft, dstTop, dstLeft+width-1,
dstTop+height-1) with the specified rop3. The rop3 code will be used to
combine the source bitmap data with the pixels in the destination bitmap.
****************************************************************************/
void NAPI DRV_BitBltColorPattSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3,
    N_int32 flipY)
{
    ushort  *pass = &rop3_table[rop3].pass1;
    int     i,bottom,dstInc;

    HW.SaveMix();
    HANDLE_SPECIAL_ROPS();
    if ((pass[0] & TO_TEMP) || (pass[1] & TO_TEMP) || (flipY && PASS_MIX != R2_0 && PASS_MIX != R2_1)) {
        // Process a scanline at a time using the temporary offscreen buffer
        if (flipY) {
            dstTop += (height-1);
            dstInc = -1;
            }
        else
            dstInc = 1;
        for (bottom = top+height; top < bottom; top++, dstTop += dstInc) {
            pass = &rop3_table[rop3].pass1;
            for (i = 0; i < 6 && *pass; i++, pass++) {
                if (*pass & TO_DEST) {
                    // Draw this pass to destination buffer
                    if (*pass & FROM_TEMP) {
                        HW.BltFromTemp(dstLeft,width,dstLeft,dstTop,PASS_MIX);
                        }
                    else if (*pass & FROM_SRC) {
                        HW_BitBltSys(srcAddr,srcPitch,left,top,width,1,dstLeft,dstTop,PASS_MIX,false);
                        }
                    else if (*pass & FROM_PATT) {
                        DV.state2d.SetMix(PASS_MIX);
                        HW_DrawColorPattRect(dstLeft,dstTop,width,1);
                        }
                    }
                else {
                    // Draw this pass to temporary buffer
                    if (*pass & FROM_SRC) {
                        HW.BltToTempSys(srcAddr,srcPitch,left,top,width,dstLeft,PASS_MIX);
                        }
                    else if (*pass & FROM_PATT) {
                        DV.state2d.SetMix(PASS_MIX);
                        HW.DrawColorPattToTemp(dstLeft,dstTop,width);
                        }
                    else if (*pass & FROM_DEST) {
                        HW.BltToTemp(dstLeft,dstTop,width,dstLeft,PASS_MIX);
                        }
                    }
                }
            }
        }
    else {
        // ROP is handled with multiple passes on the destination
        // framebuffer, so we can do this with whole blits at a time.
        for (i = 0; i < 6 && *pass; i++, pass++) {
            if (*pass & FROM_SRC) {
                HW_BitBltSys(srcAddr,srcPitch,left,top,width,height,dstLeft,dstTop,PASS_MIX,flipY);
                }
            else if (*pass & FROM_PATT) {
                DV.state2d.SetMix(PASS_MIX);
                HW_DrawColorPattRect(dstLeft,dstTop,width,height);
                }
            }
        }
    HW.RestoreMix();
}

