/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Include code for transparent pattern drawing with software
*               tiling, using the hardware linear blitter functions.
*
****************************************************************************/

/****************************************************************************
PARAMETERS:
patOfs          - Pattern offscreen memory starting offset address
patPitch        - Pitch for the offscreen pattern in bytes
left            - Left coordinate of the rectangle to draw
top             - Top coordinate of the rectangle to draw
width           - Width of the rectangle in pixels
height          - Height of the rectangle in scanlines
mix             - Mix to draw the rectangle in
transparent     - Transparent color for the pattern fill
SrcTransBltLin  - Pointer to the linear transparent blitter function

REMARKS:
This helper function renders a color patterned rectangle at the specified
location using the 16x16 color pattern stored in offscreen memory starting
at the the offset of patOfs with a pitch of patPitch.
****************************************************************************/
void DRV_DrawTransColorPattRectLin(
    N_int32 patOfs,
    N_int32 patPitch,
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    N_int32 mix,
    GA_color transparent,
    void (NAPIP SrcTransBltLin)(
        N_int32 srcOfs,
        N_int32 srcPitch,
        N_int32 srcLeft,
        N_int32 srcTop,
        N_int32 width,
        N_int32 height,
        N_int32 dstLeft,
        N_int32 dstTop,
        N_int32 mix,
        GA_color transparent))
{
    // Transparent color pattern using software 8x8 tiling
    int x,y,xCount,yCount;
    int srcX = (left & 7);
    int srcY = (top & 7);
    int dstX,dstY,remWidth,remHeight;

    // Determine initial values for inner loops
    xCount = width / 8;
    yCount = height / 8;
    remWidth = width - (xCount * 8);
    remHeight = height - (yCount * 8);

    // Process full Y rectangles first
    dstY = top;
    for (y = 0; y < yCount; y++) {
        dstX = left;
        // Process full X rectangles first
        for (x = 0; x < xCount; x++) {
            SrcTransBltLin(patOfs,patPitch,srcX,srcY,8,8,dstX,dstY,mix,transparent);
            dstX += 8;
            }

        // Now process remaining partial X rectangle
        if (remWidth)
            SrcTransBltLin(patOfs,patPitch,srcX,srcY,remWidth,8,dstX,dstY,mix,transparent);
        dstY += 8;
        }

    // Now process remaining partial Y rectangles
    if (remHeight) {
        dstX = left;
        for (x = 0; x < xCount; x++) {
            SrcTransBltLin(patOfs,patPitch,srcX,srcY,8,remHeight,dstX,dstY,mix,transparent);
            dstX += 8;
            }

        // Now process remaining partial X rectangle
        if (remWidth)
            SrcTransBltLin(patOfs,patPitch,srcX,srcY,remWidth,remHeight,dstX,dstY,mix,transparent);
        }
}
