/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Header file containing the internal VBE/AF emulation layer
*               functions.
*
****************************************************************************/

#ifndef __VBECORE_INTERNAL_H
#define __VBECORE_INTERNAL_H

#include "vesavbe.h"

/*---------------------- Macro and type definitions -----------------------*/

#pragma pack(1)

/* Maximum number of modes supported per device */

#define NUM_DRIVER_MODES    150

/* The following defines a structure used to cache all necessary mode
 * information after the initial detection for a device, so that
 * it can be re-used efficiently to determine the capabilities of all
 * available display modes.
 */

typedef struct {
    ushort  mode;                   // VBE mode number
    ushort  xres;                   // Horizontal resolution
    ushort  yres;                   // Vertical resolution
    uchar   bits;                   // Bits per pixel
    uchar   dualBanks;              // True if we need dual banks
    uchar   bankShift;              // Bank shift adjust factor
    ulong   Attributes;             // Mode attributes
    ushort  BytesPerScanLine;       // Bytes per scanline
    uchar   NumberOfImagePages;     // Number of images pages
    uchar   RedMaskSize;            // Size of direct color red mask
    uchar   RedFieldPosition;       // Bit posn of lsb of red mask
    uchar   GreenMaskSize;          // Size of direct color green mask
    uchar   GreenFieldPosition;     // Bit posn of lsb of green mask
    uchar   BlueMaskSize;           // Size of direct color blue mask
    uchar   BlueFieldPosition;      // Bit posn of lsb of blue mask
    uchar   RsvdMaskSize;           // Size of direct color res mask
    uchar   RsvdFieldPosition;      // Bit posn of lsb of res mask
    ulong   MaxPixelClock;          // Maximum pixel clock
    } GA_mdEntry;

#pragma pack()

/*---------------------------- Global Variables ---------------------------*/

extern VBE_state    VBE_stateBuf;
extern uint         VBE_version;
extern VBE_vgaInfo  VBE_info;
extern GA_mdEntry   VBE_modes[NUM_DRIVER_MODES+1];

/*-------------------------- Function Prototypes --------------------------*/

/* Function to convert mode attribute flags */

N_uint32    DRV_convertAttributes(ulong vbeAttr);

/* Non drawing, mode specific emulation functions */

void        NAPI DRV_SetBank(N_int32 bank);
void        NAPI DRV_SetDisplayStart(N_int32 offset,N_int32 waitVRT);
void        NAPI DRV_SetStereoDisplayStart(N_int32 leftOffset,N_int32 rightOffset,N_int32 waitVRT);
N_int32     NAPI DRV_GetDisplayStartStatus(void);
N_int32     NAPI DRV_IsVSync(void);
void        NAPI DRV_WaitVSync(void);
void        NAPI DRV_SetVSyncWidth(N_int32 width);
N_int32     NAPI DRV_GetVSyncWidth(void);
void        NAPI DRV_EnableStereoMode(N_int32 enable);
void        NAPI DRV_SetPaletteData(GA_palette *pal,N_int32 num,N_int32 index,N_int32 waitVRT);
void        NAPI DRV_GetPaletteData(GA_palette *pal,N_int32 num,N_int32 index);

/* DPMS support functions */

void        DRV_setDPMS(uchar state);

#endif  /* __VBECORE_INTERNAL_H */
