/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  IBM PC (MS DOS)
*
* Description:  Hardware specific device detection code.
*
****************************************************************************/

#include "chipid.h"
#include "snap/graphics/confver.h"

/*--------------------------- Global variables ----------------------------*/

/* Names of manufacturer and chip revision id's.
 *
 * DON'T CHANGE THIS UNLESS YOU CHANGE THE SNAPBUILD PROGRAM!
 */

char *DRV_manufacturerName = "IBM";

char *DRV_chipsetNames[] = {
    "VGA Compatible",
    NULL,
    };

/* The VGA driver is considered auto-certified, so we hard code the
 * certification information below.
 */

#define MAX_CHIPSETS    ((sizeof(DRV_chipsetNames) / sizeof(DRV_chipsetNames[0])) - 1)
#define CONF_VERSION    ((CONF_RELEASE_MAJOR << 8) | CONF_RELEASE_MINOR)

GA_certifyChipInfo  DRV_certifyChipInfo[MAX_CHIPSETS] = {
   {sizeof(GA_certifyChipInfo),"VGA Compatible",CONF_VERSION,__DATE__,0xFF},
   };

GA_certifyInfo      DRV_certifyInfo = {
    sizeof(GA_certifyInfo),
    CERTIFY_SIGNATURE,
    __DATE__ " " __TIME__,
    MAX_CHIPSETS,
    MAX_CHIPSETS,
    &DRV_certifyChipInfo[0],
    };

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
RETURNS:
True if hardware detected correctly, false if not.

REMARKS:
The purpose of this function is to fill in the memory and I/O resource
tables that the driver requires based on the PCI configuration space
information. No other information is detected at this time, since we may
need to access registers via the mapped in resources which are not mapped
in by the OS loader code yet.
****************************************************************************/
ibool PNP_detect(void)
{
    return true;
}

/****************************************************************************
RETURNS:
True if hardware detected correctly, false if not.

REMARKS:
Detects the chipset revision, RAMDAC and clock chip for the controller.
****************************************************************************/
ibool CHIP_detect(void)
{
    DV.chipID = IBM_VGA;
    return true;
}

