/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement mode initialisation and non-drawing
*               runtime related functions.
*
****************************************************************************/

#include "internal.h"

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
REMARKS:
Programs the current bank for banked framebuffer modes.
****************************************************************************/
void NAPI VGA_SetBank(
    N_int32 bank)
{
    // Nothing to do for VGA!
}

/****************************************************************************
REMARKS:
Programs the display start address
****************************************************************************/
void NAPI VGA_SetDisplayStart(
    N_int32 offset,
    N_int32 waitVRT)
{
    DV.cntDisplayStart = offset;
    WaitDEVGA();
    ClearVERTLatch();
    if (DV.cntMode.bits >= 8) {
        // Program the display start for all other modes and controllers
        SetLowStartAddress8(offset);
        }
    else if (DV.cntMode.bits > 0) {
        // Program the display start for all VGA modes
        SetLowStartAddress4(offset);
        }
    else {
        // Program the display start for all VGA text modes
        SetLowStartAddress0(offset);
        }
    if (waitVRT)
        WaitVERTVGA();
}

/****************************************************************************
REMARKS:
Programs the pixel clock
****************************************************************************/
void VGA_setPixelClock(
    N_fix32 clock)
{
    int sel,div = 0;

    if (clock <= N_FIXCONST(26.00))
        sel = 0;
    else
        sel = 1;
    wrinx(0x3C4,0x00,0x01);
    outpb(0x3C2,(inpb(0x3CC) & 0xF3) | (sel << 2));
    modinx(0x3C4,0x1,0x8,div<<3);                   // VGA half clock bit
    wrinx(0x3C4,0x00,0x03);
}
