/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Chipset profile and register data tables for all VGA
*               compatible modes.
*
****************************************************************************/

//--------------------------------------------------------------------------
// Profile of mode and color depth independent registers
//--------------------------------------------------------------------------

ulong VGA_chipsetProfile[] = {
    _FIXIO_I(0x03D4,0x08),      // Reset Row Scan
    _FIXIO_I(0x03D4,0x0C),      // Start Address High
    _FIXIO_I(0x03D4,0x0D),      // Start Address Low
    _FIXIO_I(0x03D4,0x0E),      // Cursor Location Address High
    _FIXIO_I(0x03D4,0x0F),      // Cursor Location Address Low
    _FIXIO_I(0x03D4,0x18),      // Line Compare
    _FIXIO_I(0x03C0,0x00),      // Palette Register 0
    _FIXIO_I(0x03C0,0x11),      // Border Color
    _FIXIO_I(0x03C0,0x14),      // Pixel Padding
    _FIXIO_I(0x03CE,0x00),      // Set/Reset
    _FIXIO_I(0x03CE,0x01),      // Enable Set/Reset
    _FIXIO_I(0x03CE,0x02),      // Color Compare
    _FIXIO_I(0x03CE,0x03),      // Raster Operation/Rotate Counter
    _FIXIO_I(0x03CE,0x04),      // Read Plane Select
    _FIXIO_I(0x03CE,0x08),      // Bit Mask
    _FIXIO_I(0x03C4,0x00),      // Reset
    _FIXIO_I(0x03C4,0x03),      // Character Font Select
    REG_TERMINATOR,
    };

//--------------------------------------------------------------------------
// Profile of registers that are mode dependent
//--------------------------------------------------------------------------

ulong VGA_modeProfile[] = {
    _FIXIO_B(0x03C2),           // Misc output register
    _FIXIO_I(0x03D4,0x00),      // Horizontal Total
    _FIXIO_I(0x03D4,0x01),      // Horizontal Display End
    _FIXIO_I(0x03D4,0x02),      // Start Horizontal Blank
    _FIXIO_I(0x03D4,0x03),      // End Horizontal Blank
    _FIXIO_I(0x03D4,0x04),      // Start Horizontal Sync Position
    _FIXIO_I(0x03D4,0x05),      // End Horizontal Sync Position
    _FIXIO_I(0x03D4,0x06),      // Vertical Total
    _FIXIO_I(0x03D4,0x07),      // CRTC Overflow
    _FIXIO_I(0x03D4,0x09),      // Maximum Scan Line
    _FIXIO_I(0x03D4,0x0A),      // Cursor Start Scan Line
    _FIXIO_I(0x03D4,0x0B),      // Cursor End Scan Line
    _FIXIO_I(0x03D4,0x10),      // Vertical Retrace Start
    _FIXIO_I(0x03D4,0x11),      // Vertical Retrace End
    _FIXIO_I(0x03D4,0x12),      // Vertical Display End
    _FIXIO_I(0x03D4,0x13),      // Offset
    _FIXIO_I(0x03D4,0x14),      // Underline Location
    _FIXIO_I(0x03D4,0x15),      // Start Vertical Blank
    _FIXIO_I(0x03D4,0x16),      // End Vertical Blank
    _FIXIO_I(0x03D4,0x17),      // CRTC Mode Control
    _FIXIO_I(0x03C0,0x01),      // Palette Register 1
    _FIXIO_I(0x03C0,0x02),      // Palette Register 2
    _FIXIO_I(0x03C0,0x03),      // Palette Register 3
    _FIXIO_I(0x03C0,0x04),      // Palette Register 4
    _FIXIO_I(0x03C0,0x05),      // Palette Register 5
    _FIXIO_I(0x03C0,0x06),      // Palette Register 6
    _FIXIO_I(0x03C0,0x07),      // Palette Register 7
    _FIXIO_I(0x03C0,0x08),      // Palette Register 8
    _FIXIO_I(0x03C0,0x09),      // Palette Register 9
    _FIXIO_I(0x03C0,0x0A),      // Palette Register A
    _FIXIO_I(0x03C0,0x0B),      // Palette Register B
    _FIXIO_I(0x03C0,0x0C),      // Palette Register C
    _FIXIO_I(0x03C0,0x0D),      // Palette Register D
    _FIXIO_I(0x03C0,0x0E),      // Palette Register E
    _FIXIO_I(0x03C0,0x0F),      // Palette Register F
    _FIXIO_I(0x03C0,0x10),      // Attribute Mode Control
    _FIXIO_I(0x03C0,0x12),      // Color Plane Enable
    _FIXIO_I(0x03C0,0x13),      // Horizontal Pixel Panning
    _FIXIO_I(0x03CE,0x05),      // Graphics Controller Mode
    _FIXIO_I(0x03CE,0x06),      // Memory Map Mode Control
    _FIXIO_I(0x03CE,0x07),      // Color Dont Care
    _FIXIO_I(0x03C4,0x01),      // Clocking Mode
    _FIXIO_I(0x03C4,0x02),      // Enable Write Plane
    _FIXIO_I(0x03C4,0x04),      // Memory Mode Control
    REG_TERMINATOR,
    };

//--------------------------------------------------------------------------
// Register data for mode and color depth independent registers
//--------------------------------------------------------------------------

uchar VGA_chipsetData[] = {
    _BYTE(0x00),        // 03D4.08      Reset Row Scan
    _BYTE(0x00),        // 03D4.0C      Start Address High
    _BYTE(0x00),        // 03D4.0D      Start Address Low
    _BYTE(0x00),        // 03D4.0E      Cursor Location Address High
    _BYTE(0x00),        // 03D4.0F      Cursor Location Address Low
    _BYTE(0xFF),        // 03D4.18      Line Compare
    _BYTE(0x00),        // 03C0.00      Palette Register 0
    _BYTE(0x00),        // 03C0.11      Border Color
    _BYTE(0x00),        // 03C0.14      Pixel Padding
    _BYTE(0x00),        // 03CE.00      Set/Reset
    _BYTE(0x00),        // 03CE.01      Enable Set/Reset
    _BYTE(0x00),        // 03CE.02      Color Compare
    _BYTE(0x00),        // 03CE.03      Raster Operation/Rotate Counter
    _BYTE(0x00),        // 03CE.04      Read Plane Select
    _BYTE(0xFF),        // 03CE.08      Bit Mask
    _BYTE(0x03),        // 03C4.00      Reset
    _BYTE(0x00),        // 03C4.03      Character Font Select
    DATA_TERMINATOR,
    };

//--------------------------------------------------------------------------
// Register data for registers that are mode dependent
//--------------------------------------------------------------------------

uchar VGA_TextData[] = {
    _BYTE(0x67),        // 03C2         Misc output register
    _BYTE(0x5F),        // 03D4.00      Horizontal Total
    _BYTE(0x4F),        // 03D4.01      Horizontal Display End
    _BYTE(0x50),        // 03D4.02      Start Horizontal Blank
    _BYTE(0x82),        // 03D4.03      End Horizontal Blank
    _BYTE(0x55),        // 03D4.04      Start Horizontal Sync Position
    _BYTE(0x81),        // 03D4.05      End Horizontal Sync Position
    _BYTE(0xBF),        // 03D4.06      Vertical Total
    _BYTE(0x1F),        // 03D4.07      CRTC Overflow
    _BYTE(0x4F),        // 03D4.09      Maximum Scan Line
    _BYTE(0x0D),        // 03D4.0A      Cursor Start Scan Line
    _BYTE(0x0E),        // 03D4.0B      Cursor End Scan Line
    _BYTE(0x9C),        // 03D4.10      Vertical Retrace Start
    _BYTE(0x8E),        // 03D4.11      Vertical Retrace End
    _BYTE(0x8F),        // 03D4.12      Vertical Display End
    _BYTE(0x28),        // 03D4.13      Offset
    _BYTE(0x1F),        // 03D4.14      Underline Location
    _BYTE(0x96),        // 03D4.15      Start Vertical Blank
    _BYTE(0xB9),        // 03D4.16      End Vertical Blank
    _BYTE(0xA3),        // 03D4.17      CRTC Mode Control
    _BYTE(0x01),        // 03C0.01      Palette Register 1
    _BYTE(0x02),        // 03C0.02      Palette Register 2
    _BYTE(0x03),        // 03C0.03      Palette Register 3
    _BYTE(0x04),        // 03C0.04      Palette Register 4
    _BYTE(0x05),        // 03C0.05      Palette Register 5
    _BYTE(0x14),        // 03C0.06      Palette Register 6
    _BYTE(0x07),        // 03C0.07      Palette Register 7
    _BYTE(0x38),        // 03C0.08      Palette Register 8
    _BYTE(0x39),        // 03C0.09      Palette Register 9
    _BYTE(0x3A),        // 03C0.0A      Palette Register A
    _BYTE(0x3B),        // 03C0.0B      Palette Register B
    _BYTE(0x3C),        // 03C0.0C      Palette Register C
    _BYTE(0x3D),        // 03C0.0D      Palette Register D
    _BYTE(0x3E),        // 03C0.0E      Palette Register E
    _BYTE(0x3F),        // 03C0.0F      Palette Register F
    _BYTE(0x0C),        // 03C0.10      Attribute Mode Control
    _BYTE(0x0F),        // 03C0.12      Color Plane Enable
    _BYTE(0x08),        // 03C0.13      Horizontal Pixel Panning
    _BYTE(0x10),        // 03CE.05      Graphics Controller Mode
    _BYTE(0x0E),        // 03CE.06      Memory Map Mode Control
    _BYTE(0x00),        // 03CE.07      Color Dont Care
    _BYTE(0x20),        // 03C4.01      Clocking Mode
    _BYTE(0x03),        // 03C4.02      Enable Write Plane
    _BYTE(0x02),        // 03C4.04      Memory Mode Control
    DATA_TERMINATOR,
    };

uchar VGA_640x480xVGA4Data[] = {
    _BYTE(0xE3),        // 03C2         Misc output register
    _BYTE(0x5F),        // 03D4.00      Horizontal Total
    _BYTE(0x4F),        // 03D4.01      Horizontal Display End
    _BYTE(0x50),        // 03D4.02      Start Horizontal Blank
    _BYTE(0x82),        // 03D4.03      End Horizontal Blank
    _BYTE(0x54),        // 03D4.04      Start Horizontal Sync Position
    _BYTE(0x80),        // 03D4.05      End Horizontal Sync Position
    _BYTE(0x0B),        // 03D4.06      Vertical Total
    _BYTE(0x3E),        // 03D4.07      CRTC Overflow
    _BYTE(0x40),        // 03D4.09      Maximum Scan Line
    _BYTE(0x00),        // 03D4.0A      Cursor Start Scan Line
    _BYTE(0x00),        // 03D4.0B      Cursor End Scan Line
    _BYTE(0xEA),        // 03D4.10      Vertical Retrace Start
    _BYTE(0x8C),        // 03D4.11      Vertical Retrace End
    _BYTE(0xDF),        // 03D4.12      Vertical Display End
    _BYTE(0x28),        // 03D4.13      Offset
    _BYTE(0x00),        // 03D4.14      Underline Location
    _BYTE(0xE7),        // 03D4.15      Start Vertical Blank
    _BYTE(0x04),        // 03D4.16      End Vertical Blank
    _BYTE(0xE3),        // 03D4.17      CRTC Mode Control
    _BYTE(0x01),        // 03C0.01      Palette Register 1
    _BYTE(0x02),        // 03C0.02      Palette Register 2
    _BYTE(0x03),        // 03C0.03      Palette Register 3
    _BYTE(0x04),        // 03C0.04      Palette Register 4
    _BYTE(0x05),        // 03C0.05      Palette Register 5
    _BYTE(0x14),        // 03C0.06      Palette Register 6
    _BYTE(0x07),        // 03C0.07      Palette Register 7
    _BYTE(0x38),        // 03C0.08      Palette Register 8
    _BYTE(0x39),        // 03C0.09      Palette Register 9
    _BYTE(0x3A),        // 03C0.0A      Palette Register A
    _BYTE(0x3B),        // 03C0.0B      Palette Register B
    _BYTE(0x3C),        // 03C0.0C      Palette Register C
    _BYTE(0x3D),        // 03C0.0D      Palette Register D
    _BYTE(0x3E),        // 03C0.0E      Palette Register E
    _BYTE(0x3F),        // 03C0.0F      Palette Register F
    _BYTE(0x01),        // 03C0.10      Attribute Mode Control
    _BYTE(0x0F),        // 03C0.12      Color Plane Enable
    _BYTE(0x00),        // 03C0.13      Horizontal Pixel Panning
    _BYTE(0x00),        // 03CE.05      Graphics Controller Mode
    _BYTE(0x05),        // 03CE.06      Memory Map Mode Control
    _BYTE(0x0F),        // 03CE.07      Color Dont Care
    _BYTE(0x21),        // 03C4.01      Clocking Mode
    _BYTE(0x0F),        // 03C4.02      Enable Write Plane
    _BYTE(0x06),        // 03C4.04      Memory Mode Control
    DATA_TERMINATOR,
    };

uchar VGA_320x200xVGA8Data[] = {
    _BYTE(0x63),        // 03C2         Misc output register
    _BYTE(0x5F),        // 03D4.00      Horizontal Total
    _BYTE(0x4F),        // 03D4.01      Horizontal Display End
    _BYTE(0x50),        // 03D4.02      Start Horizontal Blank
    _BYTE(0x82),        // 03D4.03      End Horizontal Blank
    _BYTE(0x54),        // 03D4.04      Start Horizontal Sync Position
    _BYTE(0x80),        // 03D4.05      End Horizontal Sync Position
    _BYTE(0xBF),        // 03D4.06      Vertical Total
    _BYTE(0x1F),        // 03D4.07      CRTC Overflow
    _BYTE(0x41),        // 03D4.09      Maximum Scan Line
    _BYTE(0x00),        // 03D4.0A      Cursor Start Scan Line
    _BYTE(0x00),        // 03D4.0B      Cursor End Scan Line
    _BYTE(0x9C),        // 03D4.10      Vertical Retrace Start
    _BYTE(0x8E),        // 03D4.11      Vertical Retrace End
    _BYTE(0x8F),        // 03D4.12      Vertical Display End
    _BYTE(0x28),        // 03D4.13      Offset
    _BYTE(0x40),        // 03D4.14      Underline Location
    _BYTE(0x96),        // 03D4.15      Start Vertical Blank
    _BYTE(0xB9),        // 03D4.16      End Vertical Blank
    _BYTE(0xA3),        // 03D4.17      CRTC Mode Control
    _BYTE(0x01),        // 03C0.01      Palette Register 1
    _BYTE(0x02),        // 03C0.02      Palette Register 2
    _BYTE(0x03),        // 03C0.03      Palette Register 3
    _BYTE(0x04),        // 03C0.04      Palette Register 4
    _BYTE(0x05),        // 03C0.05      Palette Register 5
    _BYTE(0x06),        // 03C0.06      Palette Register 6
    _BYTE(0x07),        // 03C0.07      Palette Register 7
    _BYTE(0x08),        // 03C0.08      Palette Register 8
    _BYTE(0x09),        // 03C0.09      Palette Register 9
    _BYTE(0x0A),        // 03C0.0A      Palette Register A
    _BYTE(0x0B),        // 03C0.0B      Palette Register B
    _BYTE(0x0C),        // 03C0.0C      Palette Register C
    _BYTE(0x0D),        // 03C0.0D      Palette Register D
    _BYTE(0x0E),        // 03C0.0E      Palette Register E
    _BYTE(0x0F),        // 03C0.0F      Palette Register F
    _BYTE(0x41),        // 03C0.10      Attribute Mode Control
    _BYTE(0x0F),        // 03C0.12      Color Plane Enable
    _BYTE(0x00),        // 03C0.13      Horizontal Pixel Panning
    _BYTE(0x40),        // 03CE.05      Graphics Controller Mode
    _BYTE(0x05),        // 03CE.06      Memory Map Mode Control
    _BYTE(0x0F),        // 03CE.07      Color Dont Care
    _BYTE(0x21),        // 03C4.01      Clocking Mode
    _BYTE(0x0F),        // 03C4.02      Enable Write Plane
    _BYTE(0x0E),        // 03C4.04      Memory Mode Control
    DATA_TERMINATOR,
    };
