/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  VGA register profile tables. Only used by the
*               analyse and optimise programs.
*
****************************************************************************/

//---------------------------------------------------------------------------
// Register profile for VGA registers.
//---------------------------------------------------------------------------

ulong VGA_regProfile[] = {
    // Miscellaneous registers
    FIXIO_B(0x3C2,"Misc output register"),

    // CRTC registers
    FIXIO_I(0x3D4,0x00,"Horizontal Total"),
    FIXIO_I(0x3D4,0x01,"Horizontal Display End"),
    FIXIO_I(0x3D4,0x02,"Start Horizontal Blank"),
    FIXIO_I(0x3D4,0x03,"End Horizontal Blank"),
    FIXIO_I(0x3D4,0x04,"Start Horizontal Sync Position"),
    FIXIO_I(0x3D4,0x05,"End Horizontal Sync Position"),
    FIXIO_I(0x3D4,0x06,"Vertical Total"),
    FIXIO_I(0x3D4,0x07,"CRTC Overflow"),
    FIXIO_I(0x3D4,0x08,"Reset Row Scan"),
    FIXIO_I(0x3D4,0x09,"Maximum Scan Line"),
    FIXIO_I(0x3D4,0x0A,"Cursor Start Scan Line"),
    FIXIO_I(0x3D4,0x0B,"Cursor End Scan Line"),
    FIXIO_I(0x3D4,0x0C,"Start Address High"),
    FIXIO_I(0x3D4,0x0D,"Start Address Low"),
    FIXIO_I(0x3D4,0x0E,"Cursor Location Address High"),
    FIXIO_I(0x3D4,0x0F,"Cursor Location Address Low"),
    FIXIO_I(0x3D4,0x10,"Vertical Retrace Start"),
    FIXIO_I(0x3D4,0x11,"Vertical Retrace End"),
    FIXIO_I(0x3D4,0x12,"Vertical Display End"),
    FIXIO_I(0x3D4,0x13,"Offset"),
    FIXIO_I(0x3D4,0x14,"Underline Location"),
    FIXIO_I(0x3D4,0x15,"Start Vertical Blank"),
    FIXIO_I(0x3D4,0x16,"End Vertical Blank"),
    FIXIO_I(0x3D4,0x17,"CRTC Mode Control"),
    FIXIO_I(0x3D4,0x18,"Line Compare"),

    // Attribute controller registers
    FIXIO_I(0x3C0,0x00,"Palette Register 0"),
    FIXIO_I(0x3C0,0x01,"Palette Register 1"),
    FIXIO_I(0x3C0,0x02,"Palette Register 2"),
    FIXIO_I(0x3C0,0x03,"Palette Register 3"),
    FIXIO_I(0x3C0,0x04,"Palette Register 4"),
    FIXIO_I(0x3C0,0x05,"Palette Register 5"),
    FIXIO_I(0x3C0,0x06,"Palette Register 6"),
    FIXIO_I(0x3C0,0x07,"Palette Register 7"),
    FIXIO_I(0x3C0,0x08,"Palette Register 8"),
    FIXIO_I(0x3C0,0x09,"Palette Register 9"),
    FIXIO_I(0x3C0,0x0A,"Palette Register A"),
    FIXIO_I(0x3C0,0x0B,"Palette Register B"),
    FIXIO_I(0x3C0,0x0C,"Palette Register C"),
    FIXIO_I(0x3C0,0x0D,"Palette Register D"),
    FIXIO_I(0x3C0,0x0E,"Palette Register E"),
    FIXIO_I(0x3C0,0x0F,"Palette Register F"),
    FIXIO_I(0x3C0,0x10,"Attribute Mode Control"),
    FIXIO_I(0x3C0,0x11,"Border Color"),
    FIXIO_I(0x3C0,0x12,"Color Plane Enable"),
    FIXIO_I(0x3C0,0x13,"Horizontal Pixel Panning"),
    FIXIO_I(0x3C0,0x14,"Pixel Padding"),

    // Graphics controller registers
    FIXIO_I(0x3CE,0x00,"Set/Reset"),
    FIXIO_I(0x3CE,0x01,"Enable Set/Reset"),
    FIXIO_I(0x3CE,0x02,"Color Compare"),
    FIXIO_I(0x3CE,0x03,"Raster Operation/Rotate Counter"),
    FIXIO_I(0x3CE,0x04,"Read Plane Select"),
    FIXIO_I(0x3CE,0x05,"Graphics Controller Mode"),
    FIXIO_I(0x3CE,0x06,"Memory Map Mode Control"),
    FIXIO_I(0x3CE,0x07,"Color Dont Care"),
    FIXIO_I(0x3CE,0x08,"Bit Mask"),

    // Sequencer registers
    FIXIO_I(0x3C4,0x00,"Reset"),
    FIXIO_I(0x3C4,0x01,"Clocking Mode"),
    FIXIO_I(0x3C4,0x02,"Enable Write Plane"),
    FIXIO_I(0x3C4,0x03,"Character Font Select"),
    FIXIO_I(0x3C4,0x04,"Memory Mode Control"),

    REG_TERMINATOR,
    };
