/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  IBM PC (MSDOS) Real Mode and 16/32 bit Protected Mode.
*
* Description:  Module to implement parsing of the EDID monitor information
*               blocks, as well as constructing correct EDID information
*               from passed in information. All information is extracted
*               from the binary EDID information block into the EDID_record
*               structure.
*
****************************************************************************/

#include "loader.h"

/*--------------------------- Global variables ----------------------------*/

/* {secret} */
typedef struct {
    char    mfrID[4];
    char    *mfrName;
    } ManufacturerName;

/* {secret} */
ManufacturerName mfrNames[] = {
    {"ACI","Action Systems, Inc."},
    {"ADI","ADI Corporation."},
    {"AMT","Tandberg Display"},
    {"AOC","AOC International"},
    {"API","Acer"},
    {"APP","Apple Computer, Inc."},
    {"AST","AST"},
    {"BRG","Bridge Information Co.,Ltd."},
    {"CCP","EPSON"},
    {"COE","Tandberg Display"},
    {"CPC","Daewoo"},
    {"CPL","COMPAL Electronic, Inc."},
    {"CPQ","COMPAQ"},
    {"CRN","Cornerstone Peripherals Technology, Inc."},
    {"CTS","Comtec System Co., Ltd"},
    {"CTX","CTX - Chuntex Electronic"},
    {"CYB","CyberVision"},
    {"DEC","Digital Equipment Corp."},
    {"DEL","Dell Computer Corp."},
    {"DPC","Delta Electronics, Inc."},
    {"DWE","Daewoo"},
    {"ECS","ELITEGROUP Computer Systems"},
    {"EHJ","EPSON"},
    {"EIZ","EIZO NANAO"},
    {"ELS","Elsa AG"},
    {"ENC","EIZO NANAO"},
    {"EPI","Envision Peripherals, Inc."},
    {"FCM","Funai Electric Company of Taiwan"},
    {"FPA","Fujitsu"},
    {"FUJ","Fujitsu"},
    {"GSM","LG Electronics Inc."},
    {"GWY","Gateway"},
    {"HEI","Hyundai Electronics Industries Co., Ltd."},
    {"HIT","Hitachi, Ltd."},
    {"HSL","Hansol Electronics"},
    {"HTC","Hitachi, Ltd."},
    {"HWP","Hewlett-Packard"},
    {"IBM","IBM"},
    {"ICL","ICL"},
    {"IMP","Impression"},
    {"IVM","Iiyama"},
    {"JEN","vision Co.,LTD."},
    {"KDS","Korea Data Systems"},
    {"KFC","SMILE"},
    {"KTX","KTX EDGE Technology"},
    {"LKM","Likom Technology Sdn Bhd"},
    {"LNK","Link Computer, Inc."},
    {"LTN","LITE_ON"},
    {"MAG","MAG Technology Co., Ltd."},
    {"MAX","Maxdata Computer GmbH"},
    {"MEA","Mitsubishi Electronics"},
    {"MEI","Panasonic"},
    {"MEL","Mitsubishi Electronics"},
    {"MIR","miroDISPLAYS GmbH"},
    {"MTC","MITAC"},
    {"NAN","EIZO NANAO"},
    {"NEC","NEC Technologies, Inc."},
    {"NOK","Nokia Display Products"},
    {"OEC","DAYTEK"},
    {"OKI","OKI"},
    {"OQI","ViewSonic Corporation"},
    {"PBE","Packard Bell"},
    {"PBN","Packard Bell"},
    {"PGS","Princeton Graphic Systems"},
    {"PHL","Philips Consumer Electronics"},
    {"PLB","Philips Consumer Electronics"},
    {"PTS","Proview Technology Corp."},
    {"REL","Relisys"},
    {"RDS","Radius, Inc."},
    {"SAM","Samsung"},
    {"SDI","Samtron"},
    {"SEC","EPSON"},
    {"SGX","Silicon Graphics"},
    {"SHP","Sharp"},
    {"SIB","SANYO"},
    {"SIE","Siemens Nixdorf"},
    {"SML","SMILE"},
    {"SNI","Siemens Nixdorf"},
    {"SNY","SONY Corporation"},
    {"SPT","SCEPTRE TECHNOLOGIES, INC."},
    {"SRC","Shamrock Technology"},
    {"STC","Sampo Technology Corp."},
    {"STN","Samtron"},
    {"SYL","Sylvania"},
    {"TAT","Tatung"},
    {"TAX","Taxan"},
    {"TDD","Tandberg Display"},
    {"TEA","TEAC"},
    {"TEI","TECO Information Systems, Inc."},
    {"TOS","Toshiba"},
    {"TPV","TOP Victory Electronics Co.,Ltd."},
    {"TRL","Royal Information Company"},
    {"TSB","Toshiba"},
    {"TTK","TOTOKU Electric"},
    {"TVM","TVM Corp."},
    {"UNM","Unisys"},
    {"VSC","ViewSonic Corporation"},
    {"WTC","WEN Technology Corp."},
    {"",NULL},
    };

/*----------------------------- Implementation ----------------------------*/


/****************************************************************************
PARAMETERS:
rec     - Place to store monitor config record
xRes    - X resolution for the timing
yRes    - Y resolution for the timing

REMARKS:
Sets the maximum resolution value based on the input resolution value
****************************************************************************/
static int ComputeMaxRes(
    int xRes)
{
    /* Find the maximum resolution for this mode */
    if (xRes >= 2048)       return MaxRes_2048x1536;
    else if (xRes >= 1920)  return MaxRes_1920x1440;
    else if (xRes >= 1800)  return MaxRes_1800x1350;
    else if (xRes >= 1600)  return MaxRes_1600x1200;
    else if (xRes >= 1280)  return MaxRes_1280x1024;
    else if (xRes >= 1152)  return MaxRes_1152x864;
    else if (xRes >= 1024)  return MaxRes_1024x768;
    else if (xRes >= 800)   return MaxRes_800x600;
    return MaxRes_640x480;
}

/****************************************************************************
PARAMETERS:
rec     - Place to store monitor config record
xRes    - X resolution for the timing
yRes    - Y resolution for the timing
Hz      - Refresh rate for the timing
maxRes  - Maximum resolution id to store in record

REMARKS:
Computes the horizontal and vertical timings required to support the
specified mode, and stores them in the record if they are larger than
the current maximum values.
****************************************************************************/
static void ComputeTimingValues(
    EDID_record *rec,
    int xRes,
    int yRes,
    int Hz,
    int maxScan,
    uchar flags)
{
    int     maxRes,hScan;
    int     clock;

    /* Find the horizontal and vertical scan rates for this mode */
    hScan = (yRes * 105 * Hz + 5000) / 100000;
    clock = (xRes * 125 * hScan + 5000) / 100000;
    if (maxScan > 0) {
        /* Given the maximum scan rate, determine if we can support this
         * resolution or not. If not, then return without setting the
         * maximum resolution field.
         */
        if (hScan > maxScan)
            return;
        }
    else {
        /* Find the maximum scan rate from the display mode and save */
        if (rec->maxHScan < hScan)
            rec->maxHScan = hScan;
        if (rec->maxVScan < Hz)
            rec->maxVScan = Hz;
        if (rec->maxPClk < clock)
            rec->maxPClk = clock;
        }

    /* Find the maximum resolution for this mode */
    maxRes = ComputeMaxRes(xRes);
    if (rec->maxResolution < maxRes)
        rec->maxResolution = maxRes;

    /* Add this mode to our list of supported timings */
    if (maxScan == -1) {
        rec->standardTimings[rec->numStandardTimings].xRes = xRes;
        rec->standardTimings[rec->numStandardTimings].yRes = yRes;
        rec->standardTimings[rec->numStandardTimings].Hz = Hz;
        rec->standardTimings[rec->numStandardTimings].flags = flags;
        rec->numStandardTimings++;
        }
}

/****************************************************************************
PARAMETERS:
start   - Starting offset for standard timing
edid    - EDID data block to parse from
rec     - EDID record to parse into

REMARKS:
Parses the standard timings from the EDID information block, using this
information to extract monitor operational limits.
****************************************************************************/
static void ParseStandardTiming(
    int start,
    uchar *edid,
    EDID_record *rec,
    uchar flags)
{
    int     xRes,yRes,Hz,numer,denom;

    xRes = ((uint)edid[start] + 31) * 8;
    switch (((uint)edid[start+1] >> 6) & 0x3) {
        case 0: numer = 1;  denom = 1;  break;
        case 1: numer = 4;  denom = 3;  break;
        case 2: numer = 5;  denom = 4;  break;
        case 3: numer = 16; denom = 9;  break;
        }
    yRes = (xRes * denom + (numer/2)) / numer;
    Hz = ((uint)edid[start+1] & 0x3F) + 60;
    if (xRes > 256)
        ComputeTimingValues(rec,xRes,yRes,Hz,-1,flags);
}

/****************************************************************************
PARAMETERS:
rec     - EDID record to parse into

REMARKS:
Sorts all of the standard timings in the list of standard timings by
resolution and vertical frequency.
****************************************************************************/
static void SortStandardTimings(
    EDID_record *rec)
{
    EDID_standardTiming temp[MAX_STANDARD_TIMINGS];
    int                 i,j,idx;
    ushort              xRes,yRes,Hz;

    memcpy(temp,rec->standardTimings,sizeof(temp));
    xRes = yRes = Hz = 0xFFFE;
    idx = -2;
    for (i = 0; i < rec->numStandardTimings; ) {
        for (j = 0; j < rec->numStandardTimings; j++) {
            if (temp[j].xRes <= xRes && temp[j].yRes <= yRes && temp[j].Hz <= Hz) {
                xRes = temp[j].xRes;
                yRes = temp[j].yRes;
                Hz = temp[j].Hz;
                idx = j;
                }
            }
        if (idx == -2) {
            Hz = 0xFFFE;
            idx = -1;
            }
        else if (idx == -1) {
            xRes = yRes = Hz = 0xFFFE;
            idx = -2;
            }
        else {
            rec->standardTimings[i++] = temp[idx];
            memset(&temp[idx],0xFF,sizeof(temp[idx]));
            idx = -2;
            }
        }
}

/****************************************************************************
DESCRIPTION:
Parse the binary EDID block from the monitor into useful information.

PARAMETERS:
edid                - EDID information block to parse
rec                 - Place to store monitor config record
requireDescriptor   - True if the the descriptor block is required

RETURNS:
1 if valid EDID, 2 if old EDID, 0 if not found.

REMARKS:
This function parses the information in the EDID information block and
attempts to fill in a structure containing detailed, useful information.
Specifically this function finds information such as the maximum horizontal
and vertical frequencies, maximum resolution and a list of all known
standard and detailed timings. It also extracts the monitor manufacturer
name, model name, serial numbers, manufacture date and other useful information.

If you pass a value of true for requireDescriptor and the EDID block does
not contain valid operational limits information, this function will return
a value of 2. This indicates that the EDID was parsed successfully, however
the operational limits are missing from the data. If you call this
function again with requireDescriptor set to false, the returned operational
limits are derived from the list of standard and detailed timings in the
EDID block, so may not be completely accurate for the attached monitor.

SEE ALSO:
DDC_initExt, DDC_readEDID
****************************************************************************/
int NAPI EDID_parse(
    uchar *edid,
    EDID_record *rec,
    N_int32 requireDescriptor)
{
    int     i,j,start,xRes,yRes,hBlank,vBlank,Hz,hScan,clock,haveMonitorDescriptor;
    uchar   checkSum;
    ulong   W = 256;
    char    mfrTable[] = " ABCDEFGHIJKLMNOPQRSTUVWXYZ";
    uchar   edidHdr[8] = {0x0,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0};
    char    *modelName;

    /* Check that we have a valid EDID block header and checksum */
    if (memcmp(edid,edidHdr,sizeof(edidHdr) != 0))
        return false;
    checkSum = 0;
    for (i = 0; i < 128; i++)
        checkSum += edid[i];
    if (checkSum != 0)
        return false;

    /* Assume worst case values */
    memset(rec,0,sizeof(*rec));
    rec->version = ((uint)edid[0x12] << 8) | edid[0x13];
    rec->maxResolution = MaxRes_640x480;
    strcpy(rec->mfrName,"Unknown");
    strcpy(rec->modelName,"Unknown");
    strcpy(rec->serialNo,"Unknown");

    /* Extract manufacturer ID, serial number and date of manufacture */
    if (edid[0x8] & 0x80)
        rec->flags |= EDID_DDC2AB;
    rec->mfrID[0] = mfrTable[((edid[0x8]>>2) & 31)];
    rec->mfrID[1] = mfrTable[((edid[8] & 3)*8) | (edid[9]>>5)];
    rec->mfrID[2] = mfrTable[(edid[9] & 31)];
    rec->mfrID[3] = 0;
    rec->productID = edid[0xA] + edid[0xB]*W;
    rec->serialID = edid[0xC] + edid[0xD]*W + edid[0xE]*W*W + edid[0xF]*W*W*W;
    rec->mfrWeek = edid[0x10] & 0x3F;
    rec->mfrYear = 1990 + (edid[0x11] & 0x3F);

    /* Determine PNPID from mfrID and productID */
    sprintf(rec->PNPID,"%s%04X",rec->mfrID,rec->productID);

    /* Lookup manufacturer name and model name from our internal database */
    for (i = 0; mfrNames[i].mfrName != NULL; i++) {
        if (strcmp(rec->mfrID,mfrNames[i].mfrID) == 0) {
            strcpy(rec->mfrName,mfrNames[i].mfrName);
            break;
            }
        }

    /* Video input definition */
    if (edid[0x14] & 0x80)
        rec->signalLevel = EDID_Level_Digital;
    else switch((edid[0x14] >> 5) & 3) {
        case 0: rec->signalLevel = EDID_Level_0700_0300_10P;    break;
        case 1: rec->signalLevel = EDID_Level_0714_0286_10P;    break;
        case 2: rec->signalLevel = EDID_Level_1000_0400_14P;    break;
        case 3: rec->signalLevel = EDID_Level_0700_0300_07P;    break;
        }
    if (edid[0x14] & 0x10)
        rec->flags |= EDID_Blank2Blank;
    if (edid[0x14] & 0x8)
        rec->flags |= EDID_SyncSeparate;
    if (edid[0x14] & 0x4)
        rec->flags |= EDID_SyncComposite;
    if (edid[0x14] & 0x2)
        rec->flags |= EDID_SyncOnGreen;
    if (edid[0x14] & 0x1)
        rec->flags |= EDID_NeedSerration;

    /* Maximum image size */
    rec->maxHSize = edid[0x15];
    rec->maxVSize = edid[0x16];

    /* Display transfer characteristic (Gamma) */
    rec->gamma = N_FIX_1 + FXdiv(N_TOFIX(edid[0x17]),N_FIXCONST(100));

    /* Feature support */
    switch ((edid[0x18] >> 3) & 0x3) {
        case 0: rec->displayType = EDID_GrayScale;  break;
        case 3:
        case 1: rec->displayType = EDID_RGBColor;   break;
        case 2: rec->displayType = EDID_NonRGBColor;    break;
        }
    if ((edid[0x18] & 0x80) != 0)
        rec->flags |= EDID_DPMSStandBy;
    if ((edid[0x18] & 0x40) != 0)
        rec->flags |= EDID_DPMSSuspend;
    if ((edid[0x18] & 0x20) != 0)
        rec->flags |= EDID_DPMSOff;
    if ((edid[0x18] & 2) != 0)
        rec->flags |= EDID_PreferredTiming;
    if ((edid[0x18] & 1) != 0)
        rec->flags |= EDID_GTFEnabled;

    /* Chromaticity data */
    rec->Rx = FXdiv(N_TOFIX( ((edid[0x19]>>6) & 3) + (edid[0x1B] << 2) ),N_FIXCONST(1024));
    rec->Ry = FXdiv(N_TOFIX( ((edid[0x19]>>4) & 3) + (edid[0x1C] << 2) ),N_FIXCONST(1024));
    rec->Gx = FXdiv(N_TOFIX( ((edid[0x19]>>2) & 3) + (edid[0x1D] << 2) ),N_FIXCONST(1024));
    rec->Gy = FXdiv(N_TOFIX( ((edid[0x19]   ) & 3) + (edid[0x1E] << 2) ),N_FIXCONST(1024));
    rec->Bx = FXdiv(N_TOFIX( ((edid[0x1A]>>6) & 3) + (edid[0x1F] << 2) ),N_FIXCONST(1024));
    rec->By = FXdiv(N_TOFIX( ((edid[0x1A]>>4) & 3) + (edid[0x20] << 2) ),N_FIXCONST(1024));
    rec->Wx = FXdiv(N_TOFIX( ((edid[0x1A]>>2) & 3) + (edid[0x21] << 2) ),N_FIXCONST(1024));
    rec->Wy = FXdiv(N_TOFIX( ((edid[0x1A]   ) & 3) + (edid[0x22] << 2) ),N_FIXCONST(1024));

    /* Now parse established timings to find the maximum resolution supported by
     * the monitor, and the horizontal and vertical scan rates.
     */
    if (edid[0x23] & 0x20)  ComputeTimingValues(rec,640,480,60,-1,EDID_VGACompatible);
    if (edid[0x23] & 0x10)  ComputeTimingValues(rec,640,480,67,-1,EDID_MacCompatible);
    if (edid[0x23] & 0x08)  ComputeTimingValues(rec,640,480,72,-1,EDID_VESAStandard);
    if (edid[0x23] & 0x04)  ComputeTimingValues(rec,640,480,75,-1,EDID_VESAStandard);
    if (edid[0x23] & 0x80)  ComputeTimingValues(rec,720,400,70,-1,EDID_VGACompatible);
    if (edid[0x23] & 0x40)  ComputeTimingValues(rec,720,400,88,-1,EDID_XGACompatible);
    if (edid[0x23] & 0x02)  ComputeTimingValues(rec,800,600,56,-1,EDID_VESAStandard);
    if (edid[0x23] & 0x01)  ComputeTimingValues(rec,800,600,60,-1,EDID_VESAStandard);
    if (edid[0x24] & 0x80)  ComputeTimingValues(rec,800,600,72,-1,EDID_VESAStandard);
    if (edid[0x24] & 0x40)  ComputeTimingValues(rec,800,600,75,-1,EDID_VESAStandard);
    if (edid[0x24] & 0x20)  ComputeTimingValues(rec,832,624,75,-1,EDID_MacCompatible);
    if (edid[0x24] & 0x10)  ComputeTimingValues(rec,1024,768,44,-1,EDID_XGACompatible);
    if (edid[0x24] & 0x08)  ComputeTimingValues(rec,1024,768,60,-1,EDID_VESAStandard);
    if (edid[0x24] & 0x04)  ComputeTimingValues(rec,1024,768,70,-1,EDID_VESAStandard);
    if (edid[0x24] & 0x02)  ComputeTimingValues(rec,1024,768,75,-1,EDID_VESAStandard);
    if (edid[0x25] & 0x80)  ComputeTimingValues(rec,1152,870,75,-1,EDID_MacCompatible);
    if (edid[0x24] & 0x01)  ComputeTimingValues(rec,1280,1024,75,-1,EDID_VESAStandard);

    /* Now parse standard timings */
    for (i = 0; i < 8; i++)
        ParseStandardTiming(0x26 + (i*2),edid,rec,EDID_VESAStandard);

    /* Check for monitor descriptor blocks to get monitor limits and for
     * detailed timings.
     */
    haveMonitorDescriptor = false;
    modelName = rec->modelName;
    if (rec->version >= 0x101) {
        /* This is a version 1 revision 1 EDID, which has the optional
         * detailed timings and monitor descriptor blocks so we parse these
         * to determine the operational limits of the monitor.
         */
        for (i = 0; i < 4; i++) {
            start = 0x36 + (i*18);
            if ((edid[start] == 0 && edid[start+1] == 0) || (edid[start+2] == 0
                    && edid[start+4] == 0)) {
                /* This is a monitor descriptor block */
                if (edid[start+3] == 0xFF) {
                    /* Monitor serial number */
                    for (j = 0; j < 13 && edid[start+5+j] != 0xA; j++)
                        rec->serialNo[j] = edid[start+5+j];
                    rec->serialNo[j] = 0;
                    }
                else if (edid[start+3] == 0xFD) {
                    /* We have a Monitor Range Limits block, so parse it!! */
                    rec->minVScan = edid[start+5];
                    rec->maxVScan = edid[start+6];
                    rec->minHScan = edid[start+7];
                    rec->maxHScan = edid[start+8];
                    rec->maxPClk = (int)edid[start+9] * 10;
                    if ((edid[start+10] & 1) != 0)
                        rec->flags |= EDID_GTFEnabled;
                    haveMonitorDescriptor = true;
                    }
                else if (edid[start+3] == 0xFC) {
                    /* Monitor name - Can be across multiple blocks! */
                    for (j = 0; j < 13 && edid[start+5+j] != 0xA; j++)
                        *modelName++ = edid[start+5+j];
                    *modelName = 0;
                    }
                else if (edid[start+3] == 0xFB) {
                    /* Color Points - UNUSED FOR NOW */
                    }
                else if (edid[start+3] == 0xFA) {
                    /* Extra standard timing identifiers */
                    ParseStandardTiming(start+5,edid,rec,EDID_VESAStandard);
                    ParseStandardTiming(start+7,edid,rec,EDID_VESAStandard);
                    ParseStandardTiming(start+9,edid,rec,EDID_VESAStandard);
                    ParseStandardTiming(start+11,edid,rec,EDID_VESAStandard);
                    ParseStandardTiming(start+13,edid,rec,EDID_VESAStandard);
                    ParseStandardTiming(start+15,edid,rec,EDID_VESAStandard);
                    }
                }
            else {
                /* This is a regular detailed timing, so extract the info
                 * from it.
                 */
                xRes = ((uint)edid[start+2] | (((uint)edid[start+4] & 0xF0) << 4));
                yRes = ((uint)edid[start+5] | (((uint)edid[start+7] & 0xF0) << 4));
                hBlank = ((uint)edid[start+3] | (((uint)edid[start+4] & 0xF) << 8));
                vBlank = ((uint)edid[start+6] | (((uint)edid[start+7] & 0xF) << 8));
                clock = ((uint)edid[start] | ((uint)edid[start+1] << 8));
                Hz = ((clock * 10000) / ((xRes + hBlank) * (yRes + vBlank)));
                hScan = ((yRes + vBlank) * Hz + 5) / 10;
                ComputeTimingValues(rec,xRes,yRes,Hz,-2,0);

                /* Extract the detailed timing information */
                j = rec->numDetailedTimings;
                rec->detailedTimings[j].pixelClock = clock;
                rec->detailedTimings[j].hActive = xRes;
                rec->detailedTimings[j].hBlank = hBlank;
                rec->detailedTimings[j].hSyncOffset = ((uint)edid[start+8] | (((uint)edid[start+11] & 0xC0) << 2));
                rec->detailedTimings[j].hSyncWidth = ((uint)edid[start+9] | (((uint)edid[start+11] & 0x30) << 4));
                rec->detailedTimings[j].hBorder = edid[start+15];
                rec->detailedTimings[j].vActive = yRes;
                rec->detailedTimings[j].vBlank = vBlank;
                rec->detailedTimings[j].vSyncOffset = (((uint)edid[start+10] >> 4) | (((uint)edid[start+11] & 0x0C) << 2));
                rec->detailedTimings[j].vSyncWidth = (((uint)edid[start+10] & 0xF) | (((uint)edid[start+11] & 0x03) << 4));
                rec->detailedTimings[j].vBorder = edid[start+16];
                rec->detailedTimings[j].hSize = ((uint)edid[start+12] | (((uint)edid[start+14] & 0xF0) << 4));
                rec->detailedTimings[j].vSize = ((uint)edid[start+13] | (((uint)edid[start+14] & 0x0F) << 8));
                rec->detailedTimings[j].hSyncPol = '-';
                rec->detailedTimings[j].vSyncPol = '-';
                if (((edid[start+17] >> 2) & 0x3) == 0x3) {
                    if (edid[start+17] & 0x2)
                        rec->detailedTimings[j].hSyncPol = '+';
                    if (edid[start+17] & 0x4)
                        rec->detailedTimings[j].vSyncPol = '+';
                    }
                rec->detailedTimings[j].hFreq = hScan;
                rec->detailedTimings[j].Hz = Hz;
                rec->numDetailedTimings++;
                }
            }
        }

    /* Fail if the EDID is really screwed */
    if (rec->maxHScan == 0 || rec->maxVScan == 0)
        return 0;

    /* Sort the standard timings */
    SortStandardTimings(rec);

    /* Unfortunately we seem to be seeing really bad EDID's in the field,
     * so we currently ignore the EDID unless it has a monitor descriptor
     * block in it, from which we can compute the information that we
     * need.
     */
    if (!haveMonitorDescriptor && requireDescriptor)
        return 2;

    /* Now compute the maximum resolution depending on the maximum horizontal
     * scan rate that we have detected. We do this simply by comparing it
     * against the values that are required to display each resolution at
     * 60Hz non-interlaced.
     */
    ComputeTimingValues(rec,640,480,60,rec->maxHScan,0);
    ComputeTimingValues(rec,800,600,60,rec->maxHScan,0);
    ComputeTimingValues(rec,1024,768,60,rec->maxHScan,0);
    ComputeTimingValues(rec,1152,864,60,rec->maxHScan,0);
    ComputeTimingValues(rec,1280,1024,60,rec->maxHScan,0);
    ComputeTimingValues(rec,1600,1200,60,rec->maxHScan,0);
    ComputeTimingValues(rec,1800,1350,60,rec->maxHScan,0);
    ComputeTimingValues(rec,1920,1440,60,rec->maxHScan,0);
    ComputeTimingValues(rec,2048,1536,60,rec->maxHScan,0);

    /* Override the monitor maximum resolution for digital flat panel monitors. The
     * panel resolution is always stored as the detailed timing in the monitor.
     */
    if ((rec->signalLevel == EDID_Level_Digital) && rec->numDetailedTimings > 0)
        rec->maxResolution = ComputeMaxRes(rec->detailedTimings[0].hActive);

    /* Fill in default minimum frequencies */
    if (rec->minHScan == 0)
        rec->minHScan = 31;
    if (rec->minVScan == 0)
        rec->minVScan = 50;
    if (rec->maxVScan < 70)
        rec->maxVScan = 70;
    return 1;
}

