/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Description:  This file contains the code specific to initialising the
*               SciTech SNAP OpenGL compatible 3D module.
*
****************************************************************************/

#include "snapmesa.h"
#include "version.h"

/*------------------------- Global Variables ------------------------------*/

GL3D_driver     GL;
#ifdef __DRIVER__
char                    DRV_path[PM_MAX_PATH];
PM_imports _VARAPI      _PM_imports;
GL3D_imports _VARAPI    _GL3D_imports;
#endif

/* Default halftone palette. This palette is only 236 entries long, which allows
 * 20 entries to be used for system color values such as the first 10 and last
 * 10 entries on Windows systems.
 */

GA_palette _snap_halftonePal[256] = {
    {0x00,0x00,0x00,0}, {0x00,0x00,0x33,0}, {0x00,0x00,0x66,0}, {0x00,0x00,0x99,0},
    {0x00,0x00,0xCC,0}, {0x00,0x00,0xFF,0}, {0x00,0x33,0x00,0}, {0x00,0x33,0x33,0},
    {0x00,0x33,0x66,0}, {0x00,0x33,0x99,0}, {0x00,0x33,0xCC,0}, {0x00,0x33,0xFF,0},
    {0x00,0x66,0x00,0}, {0x00,0x66,0x33,0}, {0x00,0x66,0x66,0}, {0x00,0x66,0x99,0},
    {0x00,0x66,0xCC,0}, {0x00,0x66,0xFF,0}, {0x00,0x99,0x00,0}, {0x00,0x99,0x33,0},
    {0x00,0x99,0x66,0}, {0x00,0x99,0x99,0}, {0x00,0x99,0xCC,0}, {0x00,0x99,0xFF,0},
    {0x00,0xCC,0x00,0}, {0x00,0xCC,0x33,0}, {0x00,0xCC,0x66,0}, {0x00,0xCC,0x99,0},
    {0x00,0xCC,0xCC,0}, {0x00,0xCC,0xFF,0}, {0x00,0xFF,0x00,0}, {0x00,0xFF,0x00,0},
    {0x00,0xFF,0x66,0}, {0x00,0xFF,0x99,0}, {0x00,0xFF,0xCC,0}, {0x00,0xFF,0xFF,0},
    {0x33,0x00,0x00,0}, {0x33,0x00,0x33,0}, {0x33,0x00,0x66,0}, {0x33,0x00,0x99,0},
    {0x33,0x00,0xCC,0}, {0x33,0x00,0xFF,0}, {0x33,0x33,0x00,0}, {0x33,0x33,0x33,0},
    {0x33,0x33,0x66,0}, {0x33,0x33,0x99,0}, {0x33,0x33,0xCC,0}, {0x33,0x33,0xFF,0},
    {0x33,0x66,0x00,0}, {0x33,0x66,0x33,0}, {0x33,0x66,0x66,0}, {0x33,0x66,0x99,0},
    {0x33,0x66,0xCC,0}, {0x33,0x66,0xFF,0}, {0x33,0x99,0x00,0}, {0x33,0x99,0x33,0},
    {0x33,0x99,0x66,0}, {0x33,0x99,0x99,0}, {0x33,0x99,0xCC,0}, {0x33,0x99,0xFF,0},
    {0x33,0xCC,0x00,0}, {0x33,0xCC,0x33,0}, {0x33,0xCC,0x66,0}, {0x33,0xCC,0x99,0},
    {0x33,0xCC,0xCC,0}, {0x33,0xCC,0xFF,0}, {0x00,0xFF,0x00,0}, {0x33,0xFF,0x33,0},
    {0x33,0xFF,0x66,0}, {0x33,0xFF,0x99,0}, {0x33,0xFF,0xCC,0}, {0x33,0xFF,0xFF,0},
    {0x66,0x00,0x00,0}, {0x66,0x00,0x33,0}, {0x66,0x00,0x66,0}, {0x66,0x00,0x99,0},
    {0x66,0x00,0xCC,0}, {0x66,0x00,0xFF,0}, {0x66,0x33,0x00,0}, {0x66,0x33,0x33,0},
    {0x66,0x33,0x66,0}, {0x66,0x33,0x99,0}, {0x66,0x33,0xCC,0}, {0x66,0x33,0xFF,0},
    {0x66,0x66,0x00,0}, {0x66,0x66,0x33,0}, {0x66,0x66,0x66,0}, {0x66,0x66,0x99,0},
    {0x66,0x66,0xCC,0}, {0x66,0x66,0xFF,0}, {0x66,0x99,0x00,0}, {0x66,0x99,0x33,0},
    {0x66,0x99,0x66,0}, {0x66,0x99,0x99,0}, {0x66,0x99,0xCC,0}, {0x66,0x99,0xFF,0},
    {0x66,0xCC,0x00,0}, {0x66,0xCC,0x33,0}, {0x66,0xCC,0x66,0}, {0x66,0xCC,0x99,0},
    {0x66,0xCC,0xCC,0}, {0x66,0xCC,0xFF,0}, {0x66,0xFF,0x00,0}, {0x66,0xFF,0x33,0},
    {0x66,0xFF,0x66,0}, {0x66,0xFF,0x99,0}, {0x66,0xFF,0xCC,0}, {0x66,0xFF,0xFF,0},
    {0x99,0x00,0x00,0}, {0x99,0x00,0x33,0}, {0x99,0x00,0x66,0}, {0x99,0x00,0x99,0},
    {0x99,0x00,0xCC,0}, {0x99,0x00,0xFF,0}, {0x99,0x33,0x00,0}, {0x99,0x33,0x33,0},
    {0x99,0x33,0x66,0}, {0x99,0x33,0x99,0}, {0x99,0x33,0xCC,0}, {0x99,0x33,0xFF,0},
    {0x99,0x66,0x00,0}, {0x99,0x66,0x33,0}, {0x99,0x66,0x66,0}, {0x99,0x66,0x99,0},
    {0x99,0x66,0xCC,0}, {0x99,0x66,0xFF,0}, {0x99,0x99,0x00,0}, {0x99,0x99,0x33,0},
    {0x99,0x99,0x66,0}, {0x99,0x99,0x99,0}, {0x99,0x99,0xCC,0}, {0x99,0x99,0xFF,0},
    {0x99,0xCC,0x00,0}, {0x99,0xCC,0x33,0}, {0x99,0xCC,0x66,0}, {0x99,0xCC,0x99,0},
    {0x99,0xCC,0xCC,0}, {0x99,0xCC,0xFF,0}, {0x99,0xFF,0x00,0}, {0x99,0xFF,0x33,0},
    {0x99,0xFF,0x66,0}, {0x99,0xFF,0x99,0}, {0x99,0xFF,0xCC,0}, {0x99,0xFF,0xFF,0},
    {0xCC,0x00,0x00,0}, {0xCC,0x00,0x33,0}, {0xCC,0x00,0x66,0}, {0xCC,0x00,0x99,0},
    {0xCC,0x00,0xCC,0}, {0xCC,0x00,0xFF,0}, {0xCC,0x33,0x00,0}, {0xCC,0x33,0x33,0},
    {0xCC,0x33,0x66,0}, {0xCC,0x33,0x99,0}, {0xCC,0x33,0xCC,0}, {0xCC,0x33,0xFF,0},
    {0xCC,0x66,0x00,0}, {0xCC,0x66,0x33,0}, {0xCC,0x66,0x66,0}, {0xCC,0x66,0x99,0},
    {0xCC,0x66,0xCC,0}, {0xCC,0x66,0xFF,0}, {0xCC,0x99,0x00,0}, {0xCC,0x99,0x33,0},
    {0xCC,0x99,0x66,0}, {0xCC,0x99,0x99,0}, {0xCC,0x99,0xCC,0}, {0xCC,0x99,0xFF,0},
    {0xCC,0xCC,0x00,0}, {0xCC,0xCC,0x33,0}, {0xCC,0xCC,0x66,0}, {0xCC,0xCC,0x99,0},
    {0xCC,0xCC,0xCC,0}, {0xCC,0xCC,0xFF,0}, {0xCC,0xFF,0x00,0}, {0xCC,0xFF,0x33,0},
    {0xCC,0xFF,0x66,0}, {0xCC,0xFF,0x99,0}, {0xCC,0xFF,0xCC,0}, {0xCC,0xFF,0xFF,0},
    {0xFF,0x00,0x00,0}, {0xFF,0x00,0x00,0}, {0xFF,0x00,0x66,0}, {0xFF,0x00,0x99,0},
    {0xFF,0x00,0xCC,0}, {0xFF,0x00,0xFF,0}, {0xFF,0x00,0x00,0}, {0xFF,0x33,0x33,0},
    {0xFF,0x33,0x66,0}, {0xFF,0x33,0x99,0}, {0xFF,0x33,0xCC,0}, {0xFF,0x33,0xFF,0},
    {0xFF,0x66,0x00,0}, {0xFF,0x66,0x33,0}, {0xFF,0x66,0x66,0}, {0xFF,0x66,0x99,0},
    {0xFF,0x66,0xCC,0}, {0xFF,0x66,0xFF,0}, {0xFF,0x99,0x00,0}, {0xFF,0x99,0x33,0},
    {0xFF,0x99,0x66,0}, {0xFF,0x99,0x99,0}, {0xFF,0x99,0xCC,0}, {0xFF,0x99,0xFF,0},
    {0xFF,0xCC,0x00,0}, {0xFF,0xCC,0x33,0}, {0xFF,0xCC,0x66,0}, {0xFF,0xCC,0x99,0},
    {0xFF,0xCC,0xCC,0}, {0xFF,0xCC,0xFF,0}, {0xFF,0xFF,0x00,0}, {0xFF,0xFF,0x33,0},
    {0xFF,0xFF,0x66,0}, {0xFF,0xFF,0x99,0}, {0xFF,0xFF,0xCC,0}, {0xFF,0xFF,0xFF,0},
    };

/* Ordered 8x8 dither matrix for 8 bit to 2.6 bit halftones. */

int _snap_dither8x8[64] = {
     0, 38,  9, 47,  2, 40, 11, 50,
    25, 12, 35, 22, 27, 15, 37, 24,
     6, 44,  3, 41,  8, 47,  5, 43,
    31, 19, 28, 15, 34, 21, 31, 18,
     1, 39, 11, 49,  0, 39, 10, 48,
    27, 14, 36, 23, 26, 13, 35, 23,
     7, 46,  4, 43,  7, 45,  3, 42,
    33, 20, 30, 17, 32, 19, 29, 16,
    };

/* Division lookup tables.  These tables compute 0-255 divided by 51 and
 * modulo 51.  These tables could approximate gamma correction.
 */

int _snap_div51[256] = {
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
    2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
    2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
    3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
    3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
    4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
    4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 5,
    };

int _snap_mod51[256] = {
    0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
    20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37,
    38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 0, 1, 2, 3, 4, 5, 6,
    7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25,
    26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43,
    44, 45, 46, 47, 48, 49, 50, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12,
    13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30,
    31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48,
    49, 50, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17,
    18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35,
    36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 0, 1, 2, 3,
    4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22,
    23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
    41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 0,
    };

/* Multiplication lookup tables. These compute 0-5 times 6 and 36. */

int _snap_mul6[6] = {
    0, 6, 12, 18, 24, 30
    };

int _snap_mul36[6] = {
    0, 36, 72, 108, 144, 180
    };

/* Ordered 4x4 dither matrix for 8 bit to 5 bit halftones. */

int _snap_dither4x4[16] = {
    0, 4, 1, 5,
    6, 2, 7, 3,
    1, 5, 0, 4,
    7, 3, 6, 2,
    };

/* Division lookup tables for 16bpp dithering */

int _snap_mod8[256] = {
    0,1,2,3,4,5,6,7,0,1,2,3,4,5,6,7,
    0,1,2,3,4,5,6,7,0,1,2,3,4,5,6,7,
    0,1,2,3,4,5,6,7,0,1,2,3,4,5,6,7,
    0,1,2,3,4,5,6,7,0,1,2,3,4,5,6,7,
    0,1,2,3,4,5,6,7,0,1,2,3,4,5,6,7,
    0,1,2,3,4,5,6,7,0,1,2,3,4,5,6,7,
    0,1,2,3,4,5,6,7,0,1,2,3,4,5,6,7,
    0,1,2,3,4,5,6,7,0,1,2,3,4,5,6,7,
    0,1,2,3,4,5,6,7,0,1,2,3,4,5,6,7,
    0,1,2,3,4,5,6,7,0,1,2,3,4,5,6,7,
    0,1,2,3,4,5,6,7,0,1,2,3,4,5,6,7,
    0,1,2,3,4,5,6,7,0,1,2,3,4,5,6,7,
    0,1,2,3,4,5,6,7,0,1,2,3,4,5,6,7,
    0,1,2,3,4,5,6,7,0,1,2,3,4,5,6,7,
    0,1,2,3,4,5,6,7,0,1,2,3,4,5,6,7,
    0,1,2,3,4,5,6,7,0,0,0,0,0,0,0,0,
    };

int _snap_mod4[256] = {
    0,2,4,6,0,2,4,6,0,2,4,6,0,2,4,6,
    0,2,4,6,0,2,4,6,0,2,4,6,0,2,4,6,
    0,2,4,6,0,2,4,6,0,2,4,6,0,2,4,6,
    0,2,4,6,0,2,4,6,0,2,4,6,0,2,4,6,
    0,2,4,6,0,2,4,6,0,2,4,6,0,2,4,6,
    0,2,4,6,0,2,4,6,0,2,4,6,0,2,4,6,
    0,2,4,6,0,2,4,6,0,2,4,6,0,2,4,6,
    0,2,4,6,0,2,4,6,0,2,4,6,0,2,4,6,
    0,2,4,6,0,2,4,6,0,2,4,6,0,2,4,6,
    0,2,4,6,0,2,4,6,0,2,4,6,0,2,4,6,
    0,2,4,6,0,2,4,6,0,2,4,6,0,2,4,6,
    0,2,4,6,0,2,4,6,0,2,4,6,0,2,4,6,
    0,2,4,6,0,2,4,6,0,2,4,6,0,2,4,6,
    0,2,4,6,0,2,4,6,0,2,4,6,0,2,4,6,
    0,2,4,6,0,2,4,6,0,2,4,6,0,2,4,6,
    0,2,4,6,0,2,4,6,0,2,4,6,0,0,0,0,
    };

/* List of exported context management functions */

static GA_glInitFuncs GL3D_initFuncs = {
    sizeof(GL3D_initFuncs),
    GL3D_init,
    GL3D_chooseVisual,
    GL3D_checkVisual,
    GL3D_createContext,
    GL3D_createSysMemContext,
    GL3D_deleteContext,
    GL3D_makeCurrent,
    GL3D_getCurrentContext,
    GL3D_swapBuffers,
    GL3D_setPaletteData,
    GL3D_getHalfTonePalette,
    GL3D_getProcAddress,
    };

/* List of exported OpenGL API functions */

static GA_glFuncs GL3D_glFuncs = {
    sizeof(GL3D_glFuncs),

    /* OpenGL 1.0/1.1 compatible functions */
    glAccum,
    glAlphaFunc,
    glAreTexturesResident,
    glArrayElement,
    glBegin,
    glBindTexture,
    glBitmap,
    glBlendFunc,
    glCallList,
    glCallLists,
    glClear,
    glClearAccum,
    glClearColor,
    glClearDepth,
    glClearIndex,
    glClearStencil,
    glClipPlane,
    glColor3b,
    glColor3bv,
    glColor3d,
    glColor3dv,
    glColor3f,
    glColor3fv,
    glColor3i,
    glColor3iv,
    glColor3s,
    glColor3sv,
    glColor3ub,
    glColor3ubv,
    glColor3ui,
    glColor3uiv,
    glColor3us,
    glColor3usv,
    glColor4b,
    glColor4bv,
    glColor4d,
    glColor4dv,
    glColor4f,
    glColor4fv,
    glColor4i,
    glColor4iv,
    glColor4s,
    glColor4sv,
    glColor4ub,
    glColor4ubv,
    glColor4ui,
    glColor4uiv,
    glColor4us,
    glColor4usv,
    glColorMask,
    glColorMaterial,
    glColorPointer,
    glCopyPixels,
    glCopyTexImage1D,
    glCopyTexImage2D,
    glCopyTexSubImage1D,
    glCopyTexSubImage2D,
    glCullFace,
    glDeleteLists,
    glDeleteTextures,
    glDepthFunc,
    glDepthMask,
    glDepthRange,
    glDisable,
    glDisableClientState,
    glDrawArrays,
    glDrawBuffer,
    glDrawElements,
    glDrawPixels,
    glEdgeFlag,
    glEdgeFlagPointer,
    glEdgeFlagv,
    glEnable,
    glEnableClientState,
    glEnd,
    glEndList,
    glEvalCoord1d,
    glEvalCoord1dv,
    glEvalCoord1f,
    glEvalCoord1fv,
    glEvalCoord2d,
    glEvalCoord2dv,
    glEvalCoord2f,
    glEvalCoord2fv,
    glEvalMesh1,
    glEvalMesh2,
    glEvalPoint1,
    glEvalPoint2,
    glFeedbackBuffer,
    glFinish,
    glFlush,
    glFogf,
    glFogfv,
    glFogi,
    glFogiv,
    glFrontFace,
    glFrustum,
    glGenLists,
    glGenTextures,
    glGetBooleanv,
    glGetClipPlane,
    glGetDoublev,
    glGetError,
    glGetFloatv,
    glGetIntegerv,
    glGetLightfv,
    glGetLightiv,
    glGetMapdv,
    glGetMapfv,
    glGetMapiv,
    glGetMaterialfv,
    glGetMaterialiv,
    glGetPixelMapfv,
    glGetPixelMapuiv,
    glGetPixelMapusv,
    glGetPointerv,
    glGetPolygonStipple,
    glGetString,
    glGetTexEnvfv,
    glGetTexEnviv,
    glGetTexGendv,
    glGetTexGenfv,
    glGetTexGeniv,
    glGetTexImage,
    glGetTexLevelParameterfv,
    glGetTexLevelParameteriv,
    glGetTexParameterfv,
    glGetTexParameteriv,
    glHint,
    glIndexMask,
    glIndexPointer,
    glIndexd,
    glIndexdv,
    glIndexf,
    glIndexfv,
    glIndexi,
    glIndexiv,
    glIndexs,
    glIndexsv,
    glIndexub,
    glIndexubv,
    glInitNames,
    glInterleavedArrays,
    glIsEnabled,
    glIsList,
    glIsTexture,
    glLightModelf,
    glLightModelfv,
    glLightModeli,
    glLightModeliv,
    glLightf,
    glLightfv,
    glLighti,
    glLightiv,
    glLineStipple,
    glLineWidth,
    glListBase,
    glLoadIdentity,
    glLoadMatrixd,
    glLoadMatrixf,
    glLoadName,
    glLogicOp,
    glMap1d,
    glMap1f,
    glMap2d,
    glMap2f,
    glMapGrid1d,
    glMapGrid1f,
    glMapGrid2d,
    glMapGrid2f,
    glMaterialf,
    glMaterialfv,
    glMateriali,
    glMaterialiv,
    glMatrixMode,
    glMultMatrixd,
    glMultMatrixf,
    glNewList,
    glNormal3b,
    glNormal3bv,
    glNormal3d,
    glNormal3dv,
    glNormal3f,
    glNormal3fv,
    glNormal3i,
    glNormal3iv,
    glNormal3s,
    glNormal3sv,
    glNormalPointer,
    glOrtho,
    glPassThrough,
    glPixelMapfv,
    glPixelMapuiv,
    glPixelMapusv,
    glPixelStoref,
    glPixelStorei,
    glPixelTransferf,
    glPixelTransferi,
    glPixelZoom,
    glPointSize,
    glPolygonMode,
    glPolygonOffset,
    glPolygonStipple,
    glPopAttrib,
    glPopClientAttrib,
    glPopMatrix,
    glPopName,
    glPrioritizeTextures,
    glPushAttrib,
    glPushClientAttrib,
    glPushMatrix,
    glPushName,
    glRasterPos2d,
    glRasterPos2dv,
    glRasterPos2f,
    glRasterPos2fv,
    glRasterPos2i,
    glRasterPos2iv,
    glRasterPos2s,
    glRasterPos2sv,
    glRasterPos3d,
    glRasterPos3dv,
    glRasterPos3f,
    glRasterPos3fv,
    glRasterPos3i,
    glRasterPos3iv,
    glRasterPos3s,
    glRasterPos3sv,
    glRasterPos4d,
    glRasterPos4dv,
    glRasterPos4f,
    glRasterPos4fv,
    glRasterPos4i,
    glRasterPos4iv,
    glRasterPos4s,
    glRasterPos4sv,
    glReadBuffer,
    glReadPixels,
    glRectd,
    glRectdv,
    glRectf,
    glRectfv,
    glRecti,
    glRectiv,
    glRects,
    glRectsv,
    glRenderMode,
    glRotated,
    glRotatef,
    glScaled,
    glScalef,
    glScissor,
    glSelectBuffer,
    glShadeModel,
    glStencilFunc,
    glStencilMask,
    glStencilOp,
    glTexCoord1d,
    glTexCoord1dv,
    glTexCoord1f,
    glTexCoord1fv,
    glTexCoord1i,
    glTexCoord1iv,
    glTexCoord1s,
    glTexCoord1sv,
    glTexCoord2d,
    glTexCoord2dv,
    glTexCoord2f,
    glTexCoord2fv,
    glTexCoord2i,
    glTexCoord2iv,
    glTexCoord2s,
    glTexCoord2sv,
    glTexCoord3d,
    glTexCoord3dv,
    glTexCoord3f,
    glTexCoord3fv,
    glTexCoord3i,
    glTexCoord3iv,
    glTexCoord3s,
    glTexCoord3sv,
    glTexCoord4d,
    glTexCoord4dv,
    glTexCoord4f,
    glTexCoord4fv,
    glTexCoord4i,
    glTexCoord4iv,
    glTexCoord4s,
    glTexCoord4sv,
    glTexCoordPointer,
    glTexEnvf,
    glTexEnvfv,
    glTexEnvi,
    glTexEnviv,
    glTexGend,
    glTexGendv,
    glTexGenf,
    glTexGenfv,
    glTexGeni,
    glTexGeniv,
    glTexImage1D,
    glTexImage2D,
    glTexParameterf,
    glTexParameterfv,
    glTexParameteri,
    glTexParameteriv,
    glTexSubImage1D,
    glTexSubImage2D,
    glTranslated,
    glTranslatef,
    glVertex2d,
    glVertex2dv,
    glVertex2f,
    glVertex2fv,
    glVertex2i,
    glVertex2iv,
    glVertex2s,
    glVertex2sv,
    glVertex3d,
    glVertex3dv,
    glVertex3f,
    glVertex3fv,
    glVertex3i,
    glVertex3iv,
    glVertex3s,
    glVertex3sv,
    glVertex4d,
    glVertex4dv,
    glVertex4f,
    glVertex4fv,
    glVertex4i,
    glVertex4iv,
    glVertex4s,
    glVertex4sv,
    glVertexPointer,
    glViewport,

    /* OpenGL 1.2/1.2.1 compatible functions */
    glBlendColor,
    glBlendEquation,
    glDrawRangeElements,
    glColorTable,
    glColorTableParameterfv,
    glColorTableParameteriv,
    glCopyColorTable,
    glGetColorTable,
    glGetColorTableParameterfv,
    glGetColorTableParameteriv,
    glColorSubTable,
    glCopyColorSubTable,
    glConvolutionFilter1D,
    glConvolutionFilter2D,
    glConvolutionParameterf,
    glConvolutionParameterfv,
    glConvolutionParameteri,
    glConvolutionParameteriv,
    glCopyConvolutionFilter1D,
    glCopyConvolutionFilter2D,
    glGetConvolutionFilter,
    glGetConvolutionParameterfv,
    glGetConvolutionParameteriv,
    glGetSeparableFilter,
    glSeparableFilter2D,
    glGetHistogram,
    glGetHistogramParameterfv,
    glGetHistogramParameteriv,
    glGetMinmax,
    glGetMinmaxParameterfv,
    glGetMinmaxParameteriv,
    glHistogram,
    glMinmax,
    glResetHistogram,
    glResetMinmax,
    glTexImage3D,
    glTexSubImage3D,
    glCopyTexSubImage3D,

    /* OpenGL 1.3 compatible functions */
    glActiveTexture,
    glClientActiveTexture,
    glMultiTexCoord1d,
    glMultiTexCoord1dv,
    glMultiTexCoord1f,
    glMultiTexCoord1fv,
    glMultiTexCoord1i,
    glMultiTexCoord1iv,
    glMultiTexCoord1s,
    glMultiTexCoord1sv,
    glMultiTexCoord2d,
    glMultiTexCoord2dv,
    glMultiTexCoord2f,
    glMultiTexCoord2fv,
    glMultiTexCoord2i,
    glMultiTexCoord2iv,
    glMultiTexCoord2s,
    glMultiTexCoord2sv,
    glMultiTexCoord3d,
    glMultiTexCoord3dv,
    glMultiTexCoord3f,
    glMultiTexCoord3fv,
    glMultiTexCoord3i,
    glMultiTexCoord3iv,
    glMultiTexCoord3s,
    glMultiTexCoord3sv,
    glMultiTexCoord4d,
    glMultiTexCoord4dv,
    glMultiTexCoord4f,
    glMultiTexCoord4fv,
    glMultiTexCoord4i,
    glMultiTexCoord4iv,
    glMultiTexCoord4s,
    glMultiTexCoord4sv,
    glLoadTransposeMatrixf,
    glLoadTransposeMatrixd,
    glMultTransposeMatrixf,
    glMultTransposeMatrixd,
    glSampleCoverage,
    glCompressedTexImage3D,
    glCompressedTexImage2D,
    glCompressedTexImage1D,
    glCompressedTexSubImage3D,
    glCompressedTexSubImage2D,
    glCompressedTexSubImage1D,
    glGetCompressedTexImage,

    /* OpenGL 1.4 compatible functions */
    glBlendFuncSeparate,
    glFogCoordf,
    glFogCoordfv,
    glFogCoordd,
    glFogCoorddv,
    glFogCoordPointer,
    glMultiDrawArrays,
    glMultiDrawElements,
    glPointParameterf,
    glPointParameterfv,
    glPointParameteri,
    glPointParameteriv,
    glSecondaryColor3b,
    glSecondaryColor3bv,
    glSecondaryColor3d,
    glSecondaryColor3dv,
    glSecondaryColor3f,
    glSecondaryColor3fv,
    glSecondaryColor3i,
    glSecondaryColor3iv,
    glSecondaryColor3s,
    glSecondaryColor3sv,
    glSecondaryColor3ub,
    glSecondaryColor3ubv,
    glSecondaryColor3ui,
    glSecondaryColor3uiv,
    glSecondaryColor3us,
    glSecondaryColor3usv,
    glSecondaryColorPointer,
    glWindowPos2d,
    glWindowPos2dv,
    glWindowPos2f,
    glWindowPos2fv,
    glWindowPos2i,
    glWindowPos2iv,
    glWindowPos2s,
    glWindowPos2sv,
    glWindowPos3d,
    glWindowPos3dv,
    glWindowPos3f,
    glWindowPos3fv,
    glWindowPos3i,
    glWindowPos3iv,
    glWindowPos3s,
    glWindowPos3sv,
    };

#define GLU3D_NAME  "glu3d.bpd"

/*-------------------------- Implementation -------------------------------*/

/****************************************************************************
PARAMETERS:
funcs       - Pointer to the function list to make valid
size        - Number of functions in the function list

RETURNS:
True if the functions are valid, false if not.

REMARKS:
This function validates the function list and returns a NULL pointer if
the function list is empty and does not contain any functions. Internally
in our drivers we list functions that have not yet been implemented on
a particular hardware device with a value of 0xFFFFFFFF for the function
pointer.
****************************************************************************/
static ibool MakeValid(
    ulong *funcs,
    int size)
{
    int     i,empty = true;

    for (i = 0; i < size; i++) {
        if ((funcs[i] == 0xFFFFFFFF || funcs[i] == 0xFFFFFFFE || funcs[i] == 0xFFFFFFFD))
            funcs[i] = 0;
        if (funcs[i] != 0)
            empty = false;
        }
    return !empty;
}

#define RETURN_VALID(p) \
{ \
    COPY_STRUCTURE2(funcs,&(p),dwSize); \
    return MakeValid(((ulong*)funcs)+1,(dwSize-sizeof(ulong))/4); \
}

/****************************************************************************
DESCRIPTION:
Returns the function pointers for the specified function group.

PARAMETERS:
id      - Identifier for the function group to get pointers for
funcs       - Pointer to the function block to fill in

RETURNS:
True if the functions are supported, false if not.

REMARKS:
This function is the main function that is used to request the function
groups from the driver.
****************************************************************************/
static ibool NAPI GL3D_queryFunctions(
    N_uint32 id,
    void *funcs)
{
    ulong               dwSize = *((ulong*)funcs);
    GA_gluInitDriver_t  GA_gluInitDriver;
    ulong               size;
    GA_gluFuncs         *gluFuncs = NULL;

    memset(funcs,0,dwSize);
    switch (id) {
        case GA_GET_GLINITFUNCS:
            RETURN_VALID(GL.init);
        case GA_GET_GLFUNCS:
            RETURN_VALID(GL.gl);
        case GA_GET_GLUFUNCS:
            /* Dynamically load the GLU library if necessary */
            if (GL.glu.dwSize == 0) {
                if ((GL.hModGLU = GA_loadLibrary(GLU3D_NAME,&size,GL.shared)) == NULL) {
                    _log("Unable to load OpenGL GLU library!\n");
                    return false;
                    }
                if ((GA_gluInitDriver = (GA_gluInitDriver_t)GA_getProcAddress(GL.hModGLU,"_GA_gluInitDriver")) == NULL) {
                    _log("Unable to get OpenGL GLU library entry point!\n");
                    return false;
                    }
                if ((gluFuncs = GA_gluInitDriver(&_PM_imports,&GL.gl)) == NULL) {
                    GA_freeLibrary(GL.hModGLU);
                    GL.hModGLU = NULL;
                    return false;
                    }
                GL.glu.dwSize = sizeof(GL.glu);
                COPY_STRUCTURE(&GL.glu,gluFuncs);
                }
            RETURN_VALID(GL.glu);
        }
    return false;
}

/****************************************************************************
REMARKS:
Driver unload function. This is called prior to the OpenGL rendering
engine being unloaded from memory.
****************************************************************************/
static void NAPI GL3D_unload(void)
{
    GA_gluExitDriver_t  GA_gluExitDriver;

    if (GL.hModGLU) {
        if ((GA_gluExitDriver = (GA_gluExitDriver_t)GA_getProcAddress(GL.hModGLU,"_GA_gluExitDriver")) != NULL)
            GA_gluExitDriver();
        GA_freeLibrary(GL.hModGLU);
        GL.hModGLU = NULL;
        }
}

/****************************************************************************
RETURNS:
Pointer to OpenGL compatible 3D driver interface

REMARKS:
This function initialises the 3D library and returns a pointer to the
OpenGL compatible 3D driver interface.
****************************************************************************/
GA_glCtx * GL3D_initModule(
    ibool shared,
    const char *path,
    GA_devCtx *hwCtx,
    PCIDeviceInfo *info,
    GA_options *licensedOptions,
    GA_globalOptions *globalOpt)
{
    // Save off the passed in information
    GL.shared = shared;
    GL.hwCtx = hwCtx;
    memcpy(&GL.PCIConfig,info,sizeof(GL.PCIConfig));
    GL.licensed = licensedOptions;
    GL.globalOpt = globalOpt;
    strcpy(DRV_path,path);

    // Now initialise the public driver header block
    strcpy(GL.ctx.Signature,GL3D_SIGNATURE);
    GL.ctx.Version = GL3D_VERSION;
    GL.ctx.DriverRev = MESA_VERSION_CODE;
    GL.ctx.OpenGLRev = OPENGL_VERSION_CODE;
    strcpy(GL.ctx.OemVendorName,SCITECH_STR);
    strcpy(GL.ctx.OemCopyright,COPYRIGHT_STR);

    // Setup the entry point functions
    GL.ctx.QueryFunctions   = GL3D_queryFunctions;
    GL.ctx.Unload           = GL3D_unload;

    // Configure all the OpenGL compatible API functions exported
    memcpy(&GL.init,&GL3D_initFuncs,sizeof(GL3D_initFuncs));
    memcpy(&GL.gl,&GL3D_glFuncs,sizeof(GL3D_glFuncs));

    // Set up OpenGL utility functions as empty initially, as we will
    // dynamically load them if they are requested.
    GL.glu.dwSize = 0;
    GL.hModGLU = NULL;

    // Now initialise the internal OpenGL engine
    return &GL.ctx;
}

#ifdef __DRIVER__

/****************************************************************************
REMARKS:
Fatal error handler called when a non-imported function is called by the
driver. We leave this to a runtime error so that older applications and
shell drivers will work with newer bpd drivers provided no newer functions
are required by the driver itself. If they are, the application or shell
driver needs to be recompiled.
****************************************************************************/
static void _PM_fatalErrorHandler(void)
{
    _PM_imports.PM_fatalError("Unsupported PM_imports import function called! Please re-compile!\n");
}

/****************************************************************************
REMARKS:
Fatal error handler called when a non-imported function is called by the
driver. We leave this to a runtime error so that older applications and
shell drivers will work with newer bpd drivers provided no newer functions
are required by the driver itself. If they are, the application or shell
driver needs to be recompiled.
****************************************************************************/
static void _GL3D_fatalErrorHandler(void)
{
    _PM_imports.PM_fatalError("Unsupported GL3D_imports import function called! Please re-compile!\n");
}

/****************************************************************************
RETURNS:
Pointer to OpenGL compatible 3D driver interface

REMARKS:
This function initialises the 3D library and returns a pointer to the
OpenGL compatible 3D driver interface.
{secret}
****************************************************************************/
GA_glCtx * _CEXPORT GA_gl3DInitDriver(
    ibool shared,
    const char *path,
    GA_devCtx *hwCtx,
    PCIDeviceInfo *info,
    PM_imports *pmImp,
    GL3D_imports *gl3DImp,
    GA_options *licensedOptions,
    GA_globalOptions *globalOpt)
{
    int     i,max;
    ulong   *p;

    // Initialize all default imports to point to fatal error handler
    // for upwards compatibility.
    max = sizeof(_PM_imports)/sizeof(GA_gl3DInitDriver_t);
    for (i = 0,p = (ulong*)&_PM_imports; i < max; i++)
        *p++ = (ulong)_PM_fatalErrorHandler;
    max = sizeof(_GL3D_imports)/sizeof(GA_gl3DInitDriver_t);
    for (i = 0,p = (ulong*)&_GL3D_imports; i < max; i++)
        *p++ = (ulong)_GL3D_fatalErrorHandler;

    // Now copy all our imported functions
    memcpy(&_PM_imports,pmImp,MIN(sizeof(_PM_imports),pmImp->dwSize));
    memcpy(&_GL3D_imports,gl3DImp,MIN(sizeof(_GL3D_imports),gl3DImp->dwSize));

    // Now initialise the driver and return the driver context stuff
    return GL3D_initModule(shared,path,hwCtx,info,licensedOptions,globalOpt);
}

#endif /* __DRIVER__ */

