/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any 32-bit protected mode environment
*
* Description:  Module containing optimised line rendering code
*
****************************************************************************/

#include "snapmesa.h"

/*------------------------- Implementation --------------------------------*/

/*
 * Draw a flat-shaded, Z-less, color index 8-bit line
 */
#define NAME flat_z_line_CI8
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    GLubyte         pixel = vert1->index;
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_TYPE          GLubyte
#define BYTES_PER_ROW       (snap->stride)
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_8(snap,X,Y)
#define CLIP_HACK           1
#define PLOT(X,Y)                                                       \
    if (Z < *zPtr) {                                                    \
        *zPtr = Z;                                                      \
        *pixelPtr = pixel;                                              \
        }
#include "swrast/s_linetemp.h"

/*
 * Draw a flat-shaded, color index 8-bit line.
 */
#define NAME flat_line_CI8
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    GLubyte         pixel = vert1->index;
#define PIXEL_TYPE          GLubyte
#define BYTES_PER_ROW       (snap->stride)
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_8(snap,X,Y)
#define CLIP_HACK           1
#define PLOT(X,Y)           *pixelPtr = pixel;
#include "swrast/s_linetemp.h"

/*
 * Draw a flat-shaded, Z-less, RGB 555 line.
 */
#define NAME flat_z_line_RGB_555
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    const GLubyte   *c = vert1->color;                                  \
    GLushort        pixel = PACK_COLOR_555(c[RCOMP],c[GCOMP],c[BCOMP]);
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_TYPE          GLushort
#define BYTES_PER_ROW       (snap->stride)
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_16(snap,X,Y)
#define CLIP_HACK           1
#define PLOT(X,Y)                                                       \
    if (Z < *zPtr) {                                                    \
        *zPtr = Z;                                                      \
        *pixelPtr = pixel;                                              \
        }
#include "swrast/s_linetemp.h"

/*
 * Draw a flat-shaded, RGB 555 line.
 */
#define NAME flat_line_RGB_555
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    const GLubyte   *c = vert1->color;                                  \
    GLushort        pixel = PACK_COLOR_555(c[RCOMP],c[GCOMP],c[BCOMP]);
#define PIXEL_TYPE          GLushort
#define BYTES_PER_ROW       (snap->stride)
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_16(snap,X,Y)
#define CLIP_HACK           1
#define PLOT(X,Y)           *pixelPtr = pixel;
#include "swrast/s_linetemp.h"

/*
 * Draw a flat-shaded, Z-less, RGB 555 dithered line.
 */
#define NAME flat_z_line_DITHER_555
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    GLuint          r = vert1->color[RCOMP];                            \
    GLuint          g = vert1->color[GCOMP];                            \
    GLuint          b = vert1->color[BCOMP];
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_TYPE          GLushort
#define BYTES_PER_ROW       (snap->stride)
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_16(snap,X,Y)
#define CLIP_HACK           1
#define PLOT(X,Y)                                                       \
    if (Z < *zPtr) {                                                    \
        *zPtr = Z;                                                      \
        PACK_COLOR_DITHER555(*pixelPtr,X,Y,r,g,b);                      \
        }
#include "swrast/s_linetemp.h"

/*
 * Draw a flat-shaded, RGB 555 dithered line.
 */
#define NAME flat_line_DITHER_555
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    GLuint          r = vert1->color[RCOMP];                            \
    GLuint          g = vert1->color[GCOMP];                            \
    GLuint          b = vert1->color[BCOMP];
#define PIXEL_TYPE          GLushort
#define BYTES_PER_ROW       (snap->stride)
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_16(snap,X,Y)
#define CLIP_HACK           1
#define PLOT(X,Y)           PACK_COLOR_DITHER555(*pixelPtr,X,Y,r,g,b);
#include "swrast/s_linetemp.h"

/*
 * Draw a flat-shaded, Z-less, RGB 565 line.
 */
#define NAME flat_z_line_RGB_565
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    const GLubyte   *c = vert1->color;                                  \
    GLushort        pixel = PACK_COLOR_565(c[RCOMP],c[GCOMP],c[BCOMP]);
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_TYPE          GLushort
#define BYTES_PER_ROW       (snap->stride)
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_16(snap,X,Y)
#define CLIP_HACK           1
#define PLOT(X,Y)           \
    if (Z < *zPtr) {        \
        *zPtr = Z;          \
        *pixelPtr = pixel;  \
        }
#include "swrast/s_linetemp.h"

/*
 * Draw a flat-shaded, RGB 565 line.
 */
#define NAME flat_line_RGB_565
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    const GLubyte   *c = vert1->color;                                  \
    GLushort        pixel = PACK_COLOR_565(c[RCOMP],c[GCOMP],c[BCOMP]);
#define PIXEL_TYPE          GLushort
#define BYTES_PER_ROW       (snap->stride)
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_16(snap,X,Y)
#define CLIP_HACK           1
#define PLOT(X,Y)           *pixelPtr = pixel;
#include "swrast/s_linetemp.h"

/*
 * Draw a flat-shaded, Z-less, RGB 565 dithered line.
 */
#define NAME flat_z_line_DITHER_565
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    GLuint          r = vert1->color[RCOMP];                            \
    GLuint          g = vert1->color[GCOMP];                            \
    GLuint          b = vert1->color[BCOMP];
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_TYPE          GLushort
#define BYTES_PER_ROW       (snap->stride)
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_16(snap,X,Y)
#define CLIP_HACK           1
#define PLOT(X,Y)                                                       \
    if (Z < *zPtr) {                                                    \
        *zPtr = Z;                                                      \
        PACK_COLOR_DITHER565(*pixelPtr,X,Y,r,g,b);                      \
        }
#include "swrast/s_linetemp.h"

/*
 * Draw a flat-shaded, RGB 565 dithered line.
 */
#define NAME flat_line_DITHER_565
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    GLuint          r = vert1->color[RCOMP];                            \
    GLuint          g = vert1->color[GCOMP];                            \
    GLuint          b = vert1->color[BCOMP];
#define PIXEL_TYPE          GLushort
#define BYTES_PER_ROW       (snap->stride)
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_16(snap,X,Y)
#define CLIP_HACK           1
#define PLOT(X,Y)           PACK_COLOR_DITHER565(*pixelPtr,X,Y,r,g,b);
#include "swrast/s_linetemp.h"

/*
 * Draw a flat-shaded, Z-less, 24-bit RGB line.
 */
#define NAME flat_z_line_RGB
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    GLubyte         r = vert1->color[RCOMP];                            \
    GLubyte         g = vert1->color[GCOMP];                            \
    GLubyte         b = vert1->color[BCOMP];
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_TYPE          rgb_t
#define BYTES_PER_ROW       (snap->stride)
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_24(snap,X,Y)
#define CLIP_HACK           1
#define PLOT(X,Y)                                                       \
    if (Z < *zPtr) {                                                    \
        *zPtr = Z;                                                      \
        pixelPtr->c[0] = b;                                             \
        pixelPtr->c[1] = g;                                             \
        pixelPtr->c[2] = r;                                             \
        }
#include "swrast/s_linetemp.h"

/*
 * Draw a flat-shaded, 24-bit RGB line.
 */
#define NAME flat_line_RGB
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    GLubyte         r = vert1->color[RCOMP];                            \
    GLubyte         g = vert1->color[GCOMP];                            \
    GLubyte         b = vert1->color[BCOMP];
#define PIXEL_TYPE          rgb_t
#define BYTES_PER_ROW       (snap->stride)
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_24(snap,X,Y)
#define CLIP_HACK           1
#define PLOT(X,Y)                                                       \
    {                                                                   \
        pixelPtr->c[0] = b;                                             \
        pixelPtr->c[1] = g;                                             \
        pixelPtr->c[2] = r;                                             \
    }
#include "swrast/s_linetemp.h"

/*
 * Draw a flat-shaded, Z-less, 24-bit BGR line.
 */
#define NAME flat_z_line_BGR
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    GLubyte         r = vert1->color[RCOMP];                            \
    GLubyte         g = vert1->color[GCOMP];                            \
    GLubyte         b = vert1->color[BCOMP];
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_TYPE          rgb_t
#define BYTES_PER_ROW       (snap->stride)
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_24(snap,X,Y)
#define CLIP_HACK           1
#define PLOT(X,Y)                                                       \
    if (Z < *zPtr) {                                                    \
        *zPtr = Z;                                                      \
        pixelPtr->c[0] = r;                                             \
        pixelPtr->c[1] = g;                                             \
        pixelPtr->c[2] = b;                                             \
        }
#include "swrast/s_linetemp.h"

/*
 * Draw a flat-shaded, 24-bit BGR line.
 */
#define NAME flat_line_BGR
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    GLubyte         r = vert1->color[RCOMP];                            \
    GLubyte         g = vert1->color[GCOMP];                            \
    GLubyte         b = vert1->color[BCOMP];
#define PIXEL_TYPE          rgb_t
#define BYTES_PER_ROW       (snap->stride)
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_24(snap,X,Y)
#define CLIP_HACK           1
#define PLOT(X,Y)                                                       \
    {                                                                   \
        pixelPtr->c[0] = r;                                             \
        pixelPtr->c[1] = g;                                             \
        pixelPtr->c[2] = b;                                             \
    }
#include "swrast/s_linetemp.h"

/*
 * Draw a flat-shaded, Z-less, 32-bit ARGB line.
 */
#define NAME flat_z_line_ARGB
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    const GLubyte   *c = vert1->color;                                  \
    GLuint          pixel = PACK_COLOR_ARGB(c[RCOMP],c[GCOMP],c[BCOMP],c[ACOMP]);
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_TYPE          GLuint
#define BYTES_PER_ROW       (snap->stride)
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_32(snap,X,Y)
#define CLIP_HACK           1
#define PLOT(X,Y)                                                       \
    if (Z < *zPtr) {                                                    \
        *zPtr = Z;                                                      \
        *pixelPtr = pixel;                                              \
        }
#include "swrast/s_linetemp.h"

/*
 * Draw a flat-shaded, 32-bit ARGB line.
 */
#define NAME flat_line_ARGB
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    const GLubyte   *c = vert1->color;                                  \
    GLuint          pixel = PACK_COLOR_ARGB(c[RCOMP],c[GCOMP],c[BCOMP],c[ACOMP]);
#define PIXEL_TYPE          GLuint
#define BYTES_PER_ROW       (snap->stride)
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_32(snap,X,Y)
#define CLIP_HACK           1
#define PLOT(X,Y)           *pixelPtr = pixel;
#include "swrast/s_linetemp.h"

/*
 * Draw a flat-shaded, Z-less, 32-bit ABGR line.
 */
#define NAME flat_z_line_ABGR
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    const GLubyte   *c = vert1->color;                                  \
    GLuint          pixel = PACK_COLOR_ABGR(c[RCOMP],c[GCOMP],c[BCOMP],c[ACOMP]);
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_TYPE          GLuint
#define BYTES_PER_ROW       (snap->stride)
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_32(snap,X,Y)
#define CLIP_HACK           1
#define PLOT(X,Y)                                                       \
    if (Z < *zPtr) {                                                    \
        *zPtr = Z;                                                      \
        *pixelPtr = pixel;                                              \
        }
#include "swrast/s_linetemp.h"

/*
 * Draw a flat-shaded, 32-bit ABGR line.
 */
#define NAME flat_line_ABGR
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    const GLubyte   *c = vert1->color;                                  \
    GLuint          pixel = PACK_COLOR_ABGR(c[RCOMP],c[GCOMP],c[BCOMP],c[ACOMP]);
#define PIXEL_TYPE          GLuint
#define BYTES_PER_ROW       (snap->stride)
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_32(snap,X,Y)
#define CLIP_HACK           1
#define PLOT(X,Y)           *pixelPtr = pixel;
#include "swrast/s_linetemp.h"

/*
 * Draw a flat-shaded, Z-less, 32-bit RGBA line.
 */
#define NAME flat_z_line_RGBA
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    const GLubyte   *c = vert1->color;                                  \
    GLuint          pixel = PACK_COLOR_RGBA(c[RCOMP],c[GCOMP],c[BCOMP],c[ACOMP]);
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_TYPE          GLuint
#define BYTES_PER_ROW       (snap->stride)
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_32(snap,X,Y)
#define CLIP_HACK           1
#define PLOT(X,Y)                                                       \
    if (Z < *zPtr) {                                                    \
        *zPtr = Z;                                                      \
        *pixelPtr = pixel;                                              \
        }
#include "swrast/s_linetemp.h"

/*
 * Draw a flat-shaded, 32-bit RGBA line.
 */
#define NAME flat_line_RGBA
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    const GLubyte   *c = vert1->color;                                  \
    GLuint          pixel = PACK_COLOR_RGBA(c[RCOMP],c[GCOMP],c[BCOMP],c[ACOMP]);
#define PIXEL_TYPE          GLuint
#define BYTES_PER_ROW       (snap->stride)
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_32(snap,X,Y)
#define CLIP_HACK           1
#define PLOT(X,Y)           *pixelPtr = pixel;
#include "swrast/s_linetemp.h"

/*
 * Draw a flat-shaded, Z-less, 32-bit BGRA line.
 */
#define NAME flat_z_line_BGRA
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    const GLubyte   *c = vert1->color;                                  \
    GLuint          pixel = PACK_COLOR_BGRA(c[RCOMP],c[GCOMP],c[BCOMP],c[ACOMP]);
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_TYPE          GLuint
#define BYTES_PER_ROW       (snap->stride)
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_32(snap,X,Y)
#define CLIP_HACK           1
#define PLOT(X,Y)                                                       \
    if (Z < *zPtr) {                                                    \
        *zPtr = Z;                                                      \
        *pixelPtr = pixel;                                              \
        }
#include "swrast/s_linetemp.h"

/*
 * Draw a flat-shaded, 32-bit BGRA line.
 */
#define NAME flat_line_BGRA
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    const GLubyte   *c = vert1->color;                                  \
    GLuint          pixel = PACK_COLOR_BGRA(c[RCOMP],c[GCOMP],c[BCOMP],c[ACOMP]);
#define PIXEL_TYPE          GLuint
#define BYTES_PER_ROW       (snap->stride)
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_32(snap,X,Y)
#define CLIP_HACK           1
#define PLOT(X,Y)           *pixelPtr = pixel;
#include "swrast/s_linetemp.h"

/****************************************************************************
REMARKS:
Determine which line function we can hook depending on the current
rendering state.
****************************************************************************/
static swrast_line_func get_line_func(
    GLcontext *ctx,
    SWcontext *swrast)
{
    GA_glContext    *snap = SNAP_CONTEXT(ctx);

    /* Ignore everything we cannot render directly */
    if (ctx->RenderMode != GL_RENDER)           return NULL;
    if (ctx->Line.SmoothFlag)                   return NULL;
    if (ctx->Texture._EnabledUnits)             return NULL;
    if (ctx->Light.ShadeModel != GL_FLAT)       return NULL;
    if (ctx->Line.StippleFlag)                  return NULL;
    if (ctx->Line.Width != 1.0F)                return NULL;
    if (swrast->_RasterMask & MULTI_DRAW_BIT)   return NULL;

    /* Now hook out those functions we can support */
    if (swrast->_RasterMask == DEPTH_BIT
        && ctx->Depth.Func == GL_LESS
        && ctx->Depth.Mask == GL_TRUE
        && ctx->Visual.depthBits == DEFAULT_SOFTWARE_DEPTH_BITS) {
        switch (snap->pixelFormat) {
            case PF_INDEX:
                return flat_z_line_CI8;
            case PF_RGB555:
                return flat_z_line_RGB_555;
            case PF_DITHER555:
                return flat_z_line_DITHER_555;
            case PF_RGB565:
                return flat_z_line_RGB_565;
            case PF_DITHER565:
                return flat_z_line_DITHER_565;
            case PF_RGB888:
                return flat_z_line_RGB;
            case PF_BGR888:
                return flat_z_line_BGR;
            case PF_ARGB8888:
                return flat_z_line_ARGB;
            case PF_ABGR8888:
                return flat_z_line_ABGR;
            case PF_RGBA8888:
                return flat_z_line_RGBA;
            case PF_BGRA8888:
                return flat_z_line_BGRA;
            default:
                return NULL;
            }
        }
    if (swrast->_RasterMask == 0) {     /* No depth test */
        switch (snap->pixelFormat) {
            case PF_INDEX:
                return flat_line_CI8;
            case PF_RGB555:
                return flat_line_RGB_555;
            case PF_DITHER555:
                return flat_line_DITHER_555;
            case PF_RGB565:
                return flat_line_RGB_565;
            case PF_DITHER565:
                return flat_line_DITHER_565;
            case PF_RGB888:
                return flat_line_RGB;
            case PF_BGR888:
                return flat_line_BGR;
            case PF_ARGB8888:
                return flat_line_ARGB;
            case PF_ABGR8888:
                return flat_line_ABGR;
            case PF_RGBA8888:
                return flat_line_RGBA;
            case PF_BGRA8888:
                return flat_line_BGRA;
            default:
                return NULL;
            }
        }
    return  NULL;
}

/****************************************************************************
REMARKS:
Override for the swrast line-selection function.  Try to use one of our
internal line functions, otherwise fall back to the standard swrast functions.
****************************************************************************/
void _snap_choose_line(
    GLcontext *ctx)
{
    SWcontext *swrast = SWRAST_CONTEXT(ctx);

    if ((swrast->Line = get_line_func(ctx,swrast)) == NULL)
        _swrast_choose_line(ctx);
}

