/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any 32-bit protected mode environment
*
* Description:  Module containing optimised span rendering code
*
****************************************************************************/

#include "snapmesa.h"

/*------------------------- Implementation --------------------------------*/

/* 8-bit color index */
#define NAME(PREFIX) PREFIX##_CI8
#define SPAN_VARS \
    const GA_glContext *snap = SNAP_CONTEXT(ctx);
#define INIT_PIXEL_PTR(P, X, Y) \
    GLubyte *P = PIXEL_ADDR_8(snap,X,Y)
#define INC_PIXEL_PTR(P) P++
#define STORE_CI_PIXEL(P, CI) \
    P[0] = CI
#define FETCH_CI_PIXEL(CI, P) \
    CI = P[0]
#include "swrast/s_spantemp.h"

/* 8-bit dithered */
#define NAME(PREFIX) PREFIX##_DITHER332
#define SPAN_VARS \
    const GA_glContext  *snap = SNAP_CONTEXT(ctx);  \
    N_uint8             *ditherVec = snap->ditherVec;
#define INIT_PIXEL_PTR(P, X, Y) \
    GLubyte *P = PIXEL_ADDR_8(snap,X,Y)
#define INC_PIXEL_PTR(P) P++
#define STORE_RGB_PIXEL(P, X, Y, R, G, B) \
    PACK_COLOR_DITHER332(P[0],X,Y,R,G,B)
#define STORE_RGBA_PIXEL(P, X, Y, R, G, B, A) \
    PACK_COLOR_DITHER332(P[0],X,Y,R,G,B)
#define FETCH_RGBA_PIXEL(R, G, B, A, P)             \
{                                                   \
    GLuint C = *P;                                  \
    UNPACK_COLOR_332(C,R,G,B,A);                    \
    (void)ditherVec;                                \
}
#include "swrast/s_spantemp.h"

/* 15-bit undithered */
#define NAME(PREFIX) PREFIX##_RGB_555
#define SPAN_VARS \
    const GA_glContext *snap = SNAP_CONTEXT(ctx);
#define INIT_PIXEL_PTR(P, X, Y) \
    GLushort *P = PIXEL_ADDR_16(snap,X,Y)
#define INC_PIXEL_PTR(P) P++
#define STORE_RGB_PIXEL(P, X, Y, R, G, B) \
    P[0] = PACK_COLOR_555(R,G,B)
#define STORE_RGBA_PIXEL(P, X, Y, R, G, B, A) \
    P[0] = PACK_COLOR_555(R,G,B)
#define FETCH_RGBA_PIXEL(R, G, B, A, P)             \
{                                                   \
    GLuint C = *P;                                  \
    UNPACK_COLOR_555(C,R,G,B,A);                    \
}
#include "swrast/s_spantemp.h"

/* 15-bit dithered */
#define NAME(PREFIX) PREFIX##_RGB_DITHER555
#define SPAN_VARS \
    const GA_glContext *snap = SNAP_CONTEXT(ctx);
#define INIT_PIXEL_PTR(P, X, Y) \
    GLushort *P = PIXEL_ADDR_16(snap,X,Y)
#define INC_PIXEL_PTR(P) P++
#define STORE_RGB_PIXEL(P, X, Y, R, G, B) \
    PACK_COLOR_DITHER555(P[0],X,Y,R,G,B)
#define STORE_RGBA_PIXEL(P, X, Y, R, G, B, A) \
    PACK_COLOR_DITHER555(P[0],X,Y,R,G,B)
#define FETCH_RGBA_PIXEL(R, G, B, A, P)             \
{                                                   \
    GLuint C = *P;                                  \
    UNPACK_COLOR_555(C,R,G,B,A);                    \
}
#include "swrast/s_spantemp.h"

/* 16-bit undithered */
#define NAME(PREFIX) PREFIX##_RGB_565
#define SPAN_VARS \
    const GA_glContext *snap = SNAP_CONTEXT(ctx);
#define INIT_PIXEL_PTR(P, X, Y) \
    GLushort *P = PIXEL_ADDR_16(snap,X,Y)
#define INC_PIXEL_PTR(P) P++
#define STORE_RGB_PIXEL(P, X, Y, R, G, B) \
    P[0] = PACK_COLOR_565(R,G,B)
#define STORE_RGBA_PIXEL(P, X, Y, R, G, B, A) \
    P[0] = PACK_COLOR_565(R,G,B)
#define FETCH_RGBA_PIXEL(R, G, B, A, P)             \
{                                                   \
    GLuint C = *P;                                  \
    UNPACK_COLOR_565(C,R,G,B,A);                    \
}
#include "swrast/s_spantemp.h"

/* 16-bit dithered */
#define NAME(PREFIX) PREFIX##_RGB_DITHER565
#define SPAN_VARS \
    const GA_glContext *snap = SNAP_CONTEXT(ctx);
#define INIT_PIXEL_PTR(P, X, Y) \
    GLushort *P = PIXEL_ADDR_16(snap,X,Y)
#define INC_PIXEL_PTR(P) P++, x++
#define STORE_RGB_PIXEL(P, X, Y, R, G, B) \
    PACK_COLOR_DITHER565(P[0],X,Y,R,G,B)
#define STORE_RGBA_PIXEL(P, X, Y, R, G, B, A) \
    PACK_COLOR_DITHER565(P[0],X,Y,R,G,B)
#define FETCH_RGBA_PIXEL(R, G, B, A, P)             \
{                                                   \
    GLuint C = *P;                                  \
    UNPACK_COLOR_555(C,R,G,B,A);                    \
}
#include "swrast/s_spantemp.h"

/* 24-bit RGB */
#define NAME(PREFIX) PREFIX##_RGB
#define SPAN_VARS \
    const GA_glContext *snap = SNAP_CONTEXT(ctx);
#define INIT_PIXEL_PTR(P, X, Y) \
    rgb_t *P = PIXEL_ADDR_24(snap,X,Y)
#define INC_PIXEL_PTR(P) P += 3
#define STORE_RGB_PIXEL(P, X, Y, R, G, B) \
    P->c[0] = B; P->c[1] = G; P->c[2] = R
#define STORE_RGBA_PIXEL(P, X, Y, R, G, B, A) \
    P->c[0] = B; P->c[1] = G; P->c[2] = R
#define FETCH_RGBA_PIXEL(R, G, B, A, P) \
    B = P->c[0]; G = P->c[1]; R = P->c[2]; A = CHAN_MAX
#include "swrast/s_spantemp.h"

/* 24-bit BGR */
#define NAME(PREFIX) PREFIX##_BGR
#define SPAN_VARS \
    const GA_glContext *snap = SNAP_CONTEXT(ctx);
#define INIT_PIXEL_PTR(P, X, Y) \
    rgb_t *P = PIXEL_ADDR_24(snap,X,Y)
#define INC_PIXEL_PTR(P) P += 3
#define STORE_RGB_PIXEL(P, X, Y, R, G, B) \
    P->c[0] = R; P->c[1] = G; P->c[2] = B
#define STORE_RGBA_PIXEL(P, X, Y, R, G, B, A) \
    P->c[0] = R; P->c[1] = G; P->c[2] = B
#define FETCH_RGBA_PIXEL(R, G, B, A, P) \
    R = P->c[0]; G = P->c[1]; B = P->c[2]; A = CHAN_MAX
#include "swrast/s_spantemp.h"

/* 32-bit ARGB */
#define NAME(PREFIX) PREFIX##_ARGB
#define SPAN_VARS \
    const GA_glContext *snap = SNAP_CONTEXT(ctx);
#define INIT_PIXEL_PTR(P, X, Y) \
    GLuint *P = PIXEL_ADDR_32(snap,X,Y)
#define INC_PIXEL_PTR(P) P++
#define STORE_RGB_PIXEL(P, X, Y, R, G, B) \
   P[0] = PACK_COLOR_ARGB(R,G,B,CHAN_MAX)
#define STORE_RGBA_PIXEL(P, X, Y, R, G, B, A) \
   P[0] = PACK_COLOR_ARGB(R,G,B,A)
#define FETCH_RGBA_PIXEL(R, G, B, A, P) \
{                                                   \
    GLuint C = *P;                                  \
    UNPACK_COLOR_RGBA(C,R,G,B,A);                   \
}
#include "swrast/s_spantemp.h"

/* 32-bit ABGR */
#define NAME(PREFIX) PREFIX##_ABGR
#define SPAN_VARS \
    const GA_glContext *snap = SNAP_CONTEXT(ctx);
#define INIT_PIXEL_PTR(P, X, Y) \
    GLuint *P = PIXEL_ADDR_32(snap,X,Y)
#define INC_PIXEL_PTR(P) P++
#define STORE_RGB_PIXEL(P, X, Y, R, G, B) \
   P[0] = PACK_COLOR_ARGB(R,G,B,CHAN_MAX)
#define STORE_RGBA_PIXEL(P, X, Y, R, G, B, A) \
   P[0] = PACK_COLOR_ARGB(R,G,B,A)
#define FETCH_RGBA_PIXEL(R, G, B, A, P) \
{                                                   \
    GLuint C = *P;                                  \
    UNPACK_COLOR_ABGR(C,R,G,B,A);                   \
}
#include "swrast/s_spantemp.h"

/* 32-bit RGBA */
#define NAME(PREFIX) PREFIX##_RGBA
#define SPAN_VARS \
    const GA_glContext *snap = SNAP_CONTEXT(ctx);
#define INIT_PIXEL_PTR(P, X, Y) \
    GLuint *P = PIXEL_ADDR_32(snap,X,Y)
#define INC_PIXEL_PTR(P) P++
#define STORE_RGB_PIXEL(P, X, Y, R, G, B) \
   P[0] = PACK_COLOR_RGBA(R,G,B,CHAN_MAX)
#define STORE_RGBA_PIXEL(P, X, Y, R, G, B, A) \
   P[0] = PACK_COLOR_RGBA(R,G,B,A)
#define FETCH_RGBA_PIXEL(R, G, B, A, P) \
{                                                   \
    GLuint C = *P;                                  \
    UNPACK_COLOR_RGBA(C,R,G,B,A);                   \
}
#include "swrast/s_spantemp.h"

/* 32-bit BGRA */
#define NAME(PREFIX) PREFIX##_BGRA
#define SPAN_VARS \
    const GA_glContext *snap = SNAP_CONTEXT(ctx);
#define INIT_PIXEL_PTR(P, X, Y) \
    GLuint *P = PIXEL_ADDR_32(snap,X,Y)
#define INC_PIXEL_PTR(P) P++
#define STORE_RGB_PIXEL(P, X, Y, R, G, B) \
   P[0] = PACK_COLOR_BGRA(R,G,B,CHAN_MAX)
#define STORE_RGBA_PIXEL(P, X, Y, R, G, B, A) \
   P[0] = PACK_COLOR_BGRA(R,G,B,A)
#define FETCH_RGBA_PIXEL(R, G, B, A, P) \
{                                                   \
    GLuint C = *P;                                  \
    UNPACK_COLOR_RGBA(C,R,G,B,A);                   \
}
#include "swrast/s_spantemp.h"

/****************************************************************************
REMARKS:
Hook all of color depth specific rendering functions into the rendering
context
****************************************************************************/
void _snap_hook_span_funcs(
    GLcontext *ctx)
{
    GA_glContext    *snap = SNAP_CONTEXT(ctx);
    swrast_dd       *swdd = _swrast_GetDeviceDriverReference(ctx);

    /* Optimized pixel span reading/writing functions */
    switch (snap->pixelFormat) {
        case PF_INDEX:
            swdd->WriteCI32Span         = write_index32_span_CI8;
            swdd->WriteCI8Span          = write_index8_span_CI8;
            swdd->WriteMonoCISpan       = write_monoindex_span_CI8;
            swdd->WriteCI32Pixels       = write_index_pixels_CI8;
            swdd->WriteMonoCIPixels     = write_monoindex_pixels_CI8;
            swdd->ReadCI32Span          = read_index_span_CI8;
            swdd->ReadCI32Pixels        = read_index_pixels_CI8;
            break;
        case PF_DITHER332:
            swdd->WriteRGBASpan         = write_rgba_span_DITHER332;
            swdd->WriteRGBSpan          = write_rgb_span_DITHER332;
            swdd->WriteMonoRGBASpan     = write_monorgba_span_DITHER332;
            swdd->WriteRGBAPixels       = write_rgba_pixels_DITHER332;
            swdd->WriteMonoRGBAPixels   = write_monorgba_pixels_DITHER332;
            swdd->ReadRGBASpan          = read_rgba_span_DITHER332;
            swdd->ReadRGBAPixels        = read_rgba_pixels_DITHER332;
            break;
        case PF_RGB555:
            swdd->WriteRGBASpan         = write_rgba_span_RGB_555;
            swdd->WriteRGBSpan          = write_rgb_span_RGB_555;
            swdd->WriteMonoRGBASpan     = write_monorgba_span_RGB_555;
            swdd->WriteRGBAPixels       = write_rgba_pixels_RGB_555;
            swdd->WriteMonoRGBAPixels   = write_monorgba_pixels_RGB_555;
            swdd->ReadRGBASpan          = read_rgba_span_RGB_555;
            swdd->ReadRGBAPixels        = read_rgba_pixels_RGB_555;
            break;
        case PF_DITHER555:
            swdd->WriteRGBASpan         = write_rgba_span_RGB_DITHER555;
            swdd->WriteRGBSpan          = write_rgb_span_RGB_DITHER555;
            swdd->WriteMonoRGBASpan     = write_monorgba_span_RGB_DITHER555;
            swdd->WriteRGBAPixels       = write_rgba_pixels_RGB_DITHER555;
            swdd->WriteMonoRGBAPixels   = write_monorgba_pixels_RGB_DITHER555;
            swdd->ReadRGBASpan          = read_rgba_span_RGB_DITHER555;
            swdd->ReadRGBAPixels        = read_rgba_pixels_RGB_DITHER555;
            break;
        case PF_RGB565:
            swdd->WriteRGBASpan         = write_rgba_span_RGB_565;
            swdd->WriteRGBSpan          = write_rgb_span_RGB_565;
            swdd->WriteMonoRGBASpan     = write_monorgba_span_RGB_565;
            swdd->WriteRGBAPixels       = write_rgba_pixels_RGB_565;
            swdd->WriteMonoRGBAPixels   = write_monorgba_pixels_RGB_565;
            swdd->ReadRGBASpan          = read_rgba_span_RGB_565;
            swdd->ReadRGBAPixels        = read_rgba_pixels_RGB_565;
            break;
        case PF_DITHER565:
            swdd->WriteRGBASpan         = write_rgba_span_RGB_DITHER565;
            swdd->WriteRGBSpan          = write_rgb_span_RGB_DITHER565;
            swdd->WriteMonoRGBASpan     = write_monorgba_span_RGB_DITHER565;
            swdd->WriteRGBAPixels       = write_rgba_pixels_RGB_DITHER565;
            swdd->WriteMonoRGBAPixels   = write_monorgba_pixels_RGB_DITHER565;
            swdd->ReadRGBASpan          = read_rgba_span_RGB_DITHER565;
            swdd->ReadRGBAPixels        = read_rgba_pixels_RGB_DITHER565;
            break;
        case PF_RGB888:
            swdd->WriteRGBASpan         = write_rgba_span_RGB;
            swdd->WriteRGBSpan          = write_rgb_span_RGB;
            swdd->WriteMonoRGBASpan     = write_monorgba_span_RGB;
            swdd->WriteRGBAPixels       = write_rgba_pixels_RGB;
            swdd->WriteMonoRGBAPixels   = write_monorgba_pixels_RGB;
            swdd->ReadRGBASpan          = read_rgba_span_RGB;
            swdd->ReadRGBAPixels        = read_rgba_pixels_RGB;
            break;
        case PF_BGR888:
            swdd->WriteRGBASpan         = write_rgba_span_BGR;
            swdd->WriteRGBSpan          = write_rgb_span_BGR;
            swdd->WriteMonoRGBASpan     = write_monorgba_span_BGR;
            swdd->WriteRGBAPixels       = write_rgba_pixels_BGR;
            swdd->WriteMonoRGBAPixels   = write_monorgba_pixels_BGR;
            swdd->ReadRGBASpan          = read_rgba_span_BGR;
            swdd->ReadRGBAPixels        = read_rgba_pixels_BGR;
            break;
        case PF_ARGB8888:
            swdd->WriteRGBASpan         = write_rgba_span_ARGB;
            swdd->WriteRGBSpan          = write_rgb_span_ARGB;
            swdd->WriteMonoRGBASpan     = write_monorgba_span_ARGB;
            swdd->WriteRGBAPixels       = write_rgba_pixels_ARGB;
            swdd->WriteMonoRGBAPixels   = write_monorgba_pixels_ARGB;
            swdd->ReadRGBASpan          = read_rgba_span_ARGB;
            swdd->ReadRGBAPixels        = read_rgba_pixels_ARGB;
            break;
        case PF_ABGR8888:
            swdd->WriteRGBASpan         = write_rgba_span_ABGR;
            swdd->WriteRGBSpan          = write_rgb_span_ABGR;
            swdd->WriteMonoRGBASpan     = write_monorgba_span_ABGR;
            swdd->WriteRGBAPixels       = write_rgba_pixels_ABGR;
            swdd->WriteMonoRGBAPixels   = write_monorgba_pixels_ABGR;
            swdd->ReadRGBASpan          = read_rgba_span_ABGR;
            swdd->ReadRGBAPixels        = read_rgba_pixels_ABGR;
            break;
        case PF_RGBA8888:
            swdd->WriteRGBASpan         = write_rgba_span_RGBA;
            swdd->WriteRGBSpan          = write_rgb_span_RGBA;
            swdd->WriteMonoRGBASpan     = write_monorgba_span_RGBA;
            swdd->WriteRGBAPixels       = write_rgba_pixels_RGBA;
            swdd->WriteMonoRGBAPixels   = write_monorgba_pixels_RGBA;
            swdd->ReadRGBASpan          = read_rgba_span_RGBA;
            swdd->ReadRGBAPixels        = read_rgba_pixels_RGBA;
            break;
        case PF_BGRA8888:
            swdd->WriteRGBASpan         = write_rgba_span_BGRA;
            swdd->WriteRGBSpan          = write_rgb_span_BGRA;
            swdd->WriteMonoRGBASpan     = write_monorgba_span_BGRA;
            swdd->WriteRGBAPixels       = write_rgba_pixels_BGRA;
            swdd->WriteMonoRGBAPixels   = write_monorgba_pixels_BGRA;
            swdd->ReadRGBASpan          = read_rgba_span_BGRA;
            swdd->ReadRGBAPixels        = read_rgba_pixels_BGRA;
            break;
        }
}

