/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any 32-bit protected mode environment
*
* Description:  Module containing optimised line rendering code
*
****************************************************************************/

#include "snapmesa.h"

/*------------------------- Implementation --------------------------------*/

/*
 * Flat, NON-depth-buffered, 8-bit CI8 triangle.
 */
#define NAME flat_triangle_CI8
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    GLubyte         pixel = v2->index;
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_8(snap,X,Y)
#define PIXEL_TYPE          GLubyte
#define BYTES_PER_ROW       (snap->stride)
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLubyte *d = pRow;                                                  \
    for (i = 0; i < span.end; i++)                                      \
        *d++ = pixel;
#include "swrast/s_tritemp.h"

/*
 * Flat, depth-buffered, 8-bit CI8 triangle.
 */
#define NAME flat_z_triangle_CI8
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_8(snap,X,Y)
#define PIXEL_TYPE          GLubyte
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    GLubyte         pixel = v2->index;
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLubyte *d = pRow;                                                  \
    for (i = 0; i < span.end; i++, d++) {                               \
        const DEPTH_TYPE z = FixedToDepth(span.z);                      \
        if (z < zRow[i]) {                                              \
            *d = pixel;                                                 \
            zRow[i] = z;                                                \
            }                                                           \
        span.z += span.zStep;                                           \
        }
#include "swrast/s_tritemp.h"

/*
 * Flat, NON-depth-buffered, 16-bit RGB 555 triangle.
 */
#define NAME flat_triangle_RGB_555
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_16(snap,X,Y)
#define PIXEL_TYPE          GLushort
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    const GLubyte   *c = v2->color;                                     \
    GLushort        pixel = PACK_COLOR_555(c[RCOMP],c[GCOMP],c[BCOMP]);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLushort *d = pRow;                                                 \
    for (i = 0; i < span.end; i++)                                      \
        *d++ = pixel;
#include "swrast/s_tritemp.h"

/*
 * Flat, depth-buffered, 16-bit RGB 555 triangle.
 */
#define NAME flat_z_triangle_RGB_555
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_16(snap,X,Y)
#define PIXEL_TYPE          GLushort
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    const GLubyte   *c = v2->color;                                     \
    GLushort        pixel = PACK_COLOR_555(c[RCOMP],c[GCOMP],c[BCOMP]);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLushort *d = pRow;                                                 \
    for (i = 0; i < span.end; i++, d++) {                               \
        const DEPTH_TYPE z = FixedToDepth(span.z);                      \
        if (z < zRow[i]) {                                              \
            *d = pixel;                                                 \
            zRow[i] = z;                                                \
            }                                                           \
        span.z += span.zStep;                                           \
        }
#include "swrast/s_tritemp.h"

/*
 * Smooth, NON-depth-buffered, 16-bit RGB 555 triangle.
 */
#define NAME smooth_triangle_RGB_555
#define INTERP_RGB          1
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_16(snap,X,Y)
#define PIXEL_TYPE          GLushort
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLushort *d = pRow;                                                 \
    for (i = 0; i < span.end; i++, d++) {                               \
        *d = PACK_COLOR_555(                                            \
            FixedToInt(span.red),                                       \
            FixedToInt(span.green),                                     \
            FixedToInt(span.blue));                                     \
        span.red += span.redStep;                                       \
        span.green += span.greenStep;                                   \
        span.blue += span.blueStep;                                     \
        }
#include "swrast/s_tritemp.h"

/*
 * Smooth, depth-buffered, 16-bit RGB 555 triangle.
 */
#define NAME smooth_z_triangle_RGB_555
#define INTERP_RGB          1
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_16(snap,X,Y)
#define PIXEL_TYPE          GLushort
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLushort *d = pRow;                                                 \
    for (i = 0; i < span.end; i++, d++) {                               \
        const DEPTH_TYPE z = FixedToDepth(span.z);                      \
        if (z < zRow[i]) {                                              \
            *d = PACK_COLOR_555(                                        \
                FixedToInt(span.red),                                   \
                FixedToInt(span.green),                                 \
                FixedToInt(span.blue));                                 \
            zRow[i] = z;                                                \
            }                                                           \
        span.red += span.redStep;                                       \
        span.green += span.greenStep;                                   \
        span.blue += span.blueStep;                                     \
        span.z += span.zStep;                                           \
        }
#include "swrast/s_tritemp.h"

/*
 * Flat, NON-depth-buffered, 16-bit dithered RGB 555 triangle.
 */
#define NAME flat_triangle_DITHER_555
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_16(snap,X,Y)
#define PIXEL_TYPE          GLushort
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    GLuint          r = v2->color[RCOMP];                               \
    GLuint          g = v2->color[GCOMP];                               \
    GLuint          b = v2->color[BCOMP];
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLushort *d = pRow;                                                 \
    for (i = 0; i < span.end; i++, d++)                                 \
        PACK_COLOR_DITHER555(*d,span.x+i,span.y,r,g,b);
#include "swrast/s_tritemp.h"

/*
 * Flat, depth-buffered, 16-bit dithered RGB 555 triangle.
 */
#define NAME flat_z_triangle_DITHER_555
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_16(snap,X,Y)
#define PIXEL_TYPE          GLushort
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    GLuint          r = v2->color[RCOMP];                               \
    GLuint          g = v2->color[GCOMP];                               \
    GLuint          b = v2->color[BCOMP];
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLushort *d = pRow;                                                 \
    for (i = 0; i < span.end; i++, d++) {                               \
        const DEPTH_TYPE z = FixedToDepth(span.z);                      \
        if (z < zRow[i]) {                                              \
            PACK_COLOR_DITHER555(*d,span.x+i,span.y,r,g,b);             \
            zRow[i] = z;                                                \
            }                                                           \
        span.z += span.zStep;                                           \
        }
#include "swrast/s_tritemp.h"

/*
 * Smooth, NON-depth-buffered, 16-bit dithered RGB 555 triangle.
 */
#define NAME smooth_triangle_DITHER_555
#define INTERP_RGB          1
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_16(snap,X,Y)
#define PIXEL_TYPE          GLushort
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLushort *d = pRow;                                                 \
    for (i = 0; i < span.end; i++, d++) {                               \
        GLuint r = FixedToInt(span.red);                                \
        GLuint g = FixedToInt(span.green);                              \
        GLuint b = FixedToInt(span.blue);                               \
        PACK_COLOR_DITHER555(*d,span.x+i,span.y,r,g,b);                 \
        span.red += span.redStep;                                       \
        span.green += span.greenStep;                                   \
        span.blue += span.blueStep;                                     \
        }
#include "swrast/s_tritemp.h"

/*
 * Smooth, depth-buffered, 16-bit dithered RGB 555 triangle.
 */
#define NAME smooth_z_triangle_DITHER_555
#define INTERP_RGB          1
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_16(snap,X,Y)
#define PIXEL_TYPE          GLushort
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLushort *d = pRow;                                                 \
    for (i = 0; i < span.end; i++, d++) {                               \
        const DEPTH_TYPE z = FixedToDepth(span.z);                      \
        if (z < zRow[i]) {                                              \
            GLuint r = FixedToInt(span.red);                            \
            GLuint g = FixedToInt(span.green);                          \
            GLuint b = FixedToInt(span.blue);                           \
            PACK_COLOR_DITHER555(*d,span.x+i,span.y,r,g,b);             \
            zRow[i] = z;                                                \
            }                                                           \
        span.red += span.redStep;                                       \
        span.green += span.greenStep;                                   \
        span.blue += span.blueStep;                                     \
        span.z += span.zStep;                                           \
        }
#include "swrast/s_tritemp.h"

/*
 * Flat, NON-depth-buffered, 16-bit RGB 565 triangle.
 */
#define NAME flat_triangle_RGB_565
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_16(snap,X,Y)
#define PIXEL_TYPE          GLushort
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    const GLubyte   *c = v2->color;                                     \
    GLushort        pixel = PACK_COLOR_565(c[RCOMP],c[GCOMP],c[BCOMP]);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLushort *d = pRow;                                                 \
    for (i = 0; i < span.end; i++)                                      \
        *d++ = pixel;
#include "swrast/s_tritemp.h"

/*
 * Flat, depth-buffered, 16-bit RGB 565 triangle.
 */
#define NAME flat_z_triangle_RGB_565
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_16(snap,X,Y)
#define PIXEL_TYPE          GLushort
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    const GLubyte   *c = v2->color;                                     \
    GLushort        pixel = PACK_COLOR_565(c[RCOMP],c[GCOMP],c[BCOMP]);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLushort *d = pRow;                                                 \
    for (i = 0; i < span.end; i++, d++) {                               \
        const DEPTH_TYPE z = FixedToDepth(span.z);                      \
        if (z < zRow[i]) {                                              \
            *d = pixel;                                                 \
            zRow[i] = z;                                                \
            }                                                           \
        span.z += span.zStep;                                           \
        }
#include "swrast/s_tritemp.h"

/*
 * Smooth, NON-depth-buffered, 16-bit RGB 565 triangle.
 */
#define NAME smooth_triangle_RGB_565
#define INTERP_RGB          1
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_16(snap,X,Y)
#define PIXEL_TYPE          GLushort
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLushort *d = pRow;                                                 \
    for (i = 0; i < span.end; i++, d++) {                               \
        *d = PACK_COLOR_565(                                            \
            FixedToInt(span.red),                                       \
            FixedToInt(span.green),                                     \
            FixedToInt(span.blue));                                     \
        span.red += span.redStep;                                       \
        span.green += span.greenStep;                                   \
        span.blue += span.blueStep;                                     \
        }
#include "swrast/s_tritemp.h"

/*
 * Smooth, depth-buffered, 16-bit RGB 565 triangle.
 */
#define NAME smooth_z_triangle_RGB_565
#define INTERP_RGB          1
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_16(snap,X,Y)
#define PIXEL_TYPE          GLushort
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLushort *d = pRow;                                                 \
    for (i = 0; i < span.end; i++, d++) {                               \
        const DEPTH_TYPE z = FixedToDepth(span.z);                      \
        if (z < zRow[i]) {                                              \
            *d = PACK_COLOR_565(                                        \
                FixedToInt(span.red),                                   \
                FixedToInt(span.green),                                 \
                FixedToInt(span.blue));                                 \
            zRow[i] = z;                                                \
            }                                                           \
        span.red += span.redStep;                                       \
        span.green += span.greenStep;                                   \
        span.blue += span.blueStep;                                     \
        span.z += span.zStep;                                           \
        }
#include "swrast/s_tritemp.h"

/*
 * Flat, NON-depth-buffered, 16-bit dithered RGB 565 triangle.
 */
#define NAME flat_triangle_DITHER_565
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_16(snap,X,Y)
#define PIXEL_TYPE          GLushort
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    GLuint          r = v2->color[RCOMP];                               \
    GLuint          g = v2->color[GCOMP];                               \
    GLuint          b = v2->color[BCOMP];
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLushort *d = pRow;                                                 \
    for (i = 0; i < span.end; i++, d++)                                 \
        PACK_COLOR_DITHER565(*d,span.x+i,span.y,r,g,b);
#include "swrast/s_tritemp.h"

/*
 * Flat, depth-buffered, 16-bit dithered RGB 565 triangle.
 */
#define NAME flat_z_triangle_DITHER_565
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_16(snap,X,Y)
#define PIXEL_TYPE          GLushort
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    GLuint          r = v2->color[RCOMP];                               \
    GLuint          g = v2->color[GCOMP];                               \
    GLuint          b = v2->color[BCOMP];
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLushort *d = pRow;                                                 \
    for (i = 0; i < span.end; i++, d++) {                               \
        const DEPTH_TYPE z = FixedToDepth(span.z);                      \
        if (z < zRow[i]) {                                              \
            PACK_COLOR_DITHER565(*d,span.x+i,span.y,r,g,b);             \
            zRow[i] = z;                                                \
            }                                                           \
        span.z += span.zStep;                                           \
        }
#include "swrast/s_tritemp.h"

/*
 * Smooth, NON-depth-buffered, 16-bit dithered RGB 565 triangle.
 */
#define NAME smooth_triangle_DITHER_565
#define INTERP_RGB          1
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_16(snap,X,Y)
#define PIXEL_TYPE          GLushort
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLushort *d = pRow;                                                 \
    for (i = 0; i < span.end; i++, d++) {                               \
        GLuint r = FixedToInt(span.red);                                \
        GLuint g = FixedToInt(span.green);                              \
        GLuint b = FixedToInt(span.blue);                               \
        PACK_COLOR_DITHER565(*d,span.x+i,span.y,r,g,b);                 \
        span.red += span.redStep;                                       \
        span.green += span.greenStep;                                   \
        span.blue += span.blueStep;                                     \
        }
#include "swrast/s_tritemp.h"

/*
 * Smooth, depth-buffered, 16-bit dithered RGB 565 triangle.
 */
#define NAME smooth_z_triangle_DITHER_565
#define INTERP_RGB          1
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_16(snap,X,Y)
#define PIXEL_TYPE          GLushort
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLushort *d = pRow;                                                 \
    for (i = 0; i < span.end; i++, d++) {                               \
        const DEPTH_TYPE z = FixedToDepth(span.z);                      \
        if (z < zRow[i]) {                                              \
            GLuint r = FixedToInt(span.red);                            \
            GLuint g = FixedToInt(span.green);                          \
            GLuint b = FixedToInt(span.blue);                           \
            PACK_COLOR_DITHER565(*d,span.x+i,span.y,r,g,b);             \
            zRow[i] = z;                                                \
            }                                                           \
        span.red += span.redStep;                                       \
        span.green += span.greenStep;                                   \
        span.blue += span.blueStep;                                     \
        span.z += span.zStep;                                           \
        }
#include "swrast/s_tritemp.h"

/*
 * Flat, NON-depth-buffered, 24-bit RGB triangle.
 */
#define NAME flat_triangle_RGB
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_24(snap,X,Y)
#define PIXEL_TYPE          rgb_t
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    GLubyte         r = v2->color[RCOMP];                               \
    GLubyte         g = v2->color[GCOMP];                               \
    GLubyte         b = v2->color[BCOMP];
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    rgb_t   *d = pRow;                                                  \
    for (i = 0; i < span.end; i++, d++) {                               \
        d->c[0] = b;                                                    \
        d->c[1] = g;                                                    \
        d->c[2] = r;                                                    \
        }
#include "swrast/s_tritemp.h"

/*
 * Flat, depth-buffered, 24-bit RGB triangle.
 */
#define NAME flat_z_triangle_RGB
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_24(snap,X,Y)
#define PIXEL_TYPE          rgb_t
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    GLubyte         r = v2->color[RCOMP];                               \
    GLubyte         g = v2->color[GCOMP];                               \
    GLubyte         b = v2->color[BCOMP];
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    rgb_t   *d = pRow;                                                  \
    for (i = 0; i < span.end; i++, d++) {                               \
        const DEPTH_TYPE z = FixedToDepth(span.z);                      \
        if (z < zRow[i]) {                                              \
            d->c[0] = b;                                                \
            d->c[1] = g;                                                \
            d->c[2] = r;                                                \
            zRow[i] = z;                                                \
            }                                                           \
        span.z += span.zStep;                                           \
        }
#include "swrast/s_tritemp.h"

/*
 * Smooth, NON-depth-buffered, 24-bit RGB triangle.
 */
#define NAME smooth_triangle_RGB
#define INTERP_RGB          1
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_24(snap,X,Y)
#define PIXEL_TYPE          rgb_t
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    rgb_t   *d = pRow;                                                  \
    for (i = 0; i < span.end; i++, d++) {                               \
        d->c[0] = (GLubyte)FixedToInt(span.blue);                       \
        d->c[1] = (GLubyte)FixedToInt(span.green);                      \
        d->c[2] = (GLubyte)FixedToInt(span.red);                        \
        span.red += span.redStep;                                       \
        span.green += span.greenStep;                                   \
        span.blue += span.blueStep;                                     \
        }
#include "swrast/s_tritemp.h"

/*
 * Smooth, depth-buffered, 24-bit RGB triangle.
 */
#define NAME smooth_z_triangle_RGB
#define INTERP_RGB          1
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_24(snap,X,Y)
#define PIXEL_TYPE          rgb_t
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    rgb_t   *d = pRow;                                                  \
    for (i = 0; i < span.end; i++, d++) {                               \
        const DEPTH_TYPE z = FixedToDepth(span.z);                      \
        if (z < zRow[i]) {                                              \
            d->c[0] = (GLubyte)FixedToInt(span.blue);                   \
            d->c[1] = (GLubyte)FixedToInt(span.green);                  \
            d->c[2] = (GLubyte)FixedToInt(span.red);                    \
            zRow[i] = z;                                                \
            }                                                           \
        span.red += span.redStep;                                       \
        span.green += span.greenStep;                                   \
        span.blue += span.blueStep;                                     \
        span.z += span.zStep;                                           \
        }
#include "swrast/s_tritemp.h"

/*
 * Flat, NON-depth-buffered, 24-bit BGR triangle.
 */
#define NAME flat_triangle_BGR
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_24(snap,X,Y)
#define PIXEL_TYPE          rgb_t
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    GLubyte         r = v2->color[RCOMP];                               \
    GLubyte         g = v2->color[GCOMP];                               \
    GLubyte         b = v2->color[BCOMP];
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    rgb_t   *d = pRow;                                                  \
    for (i = 0; i < span.end; i++, d++) {                               \
        d->c[0] = r;                                                    \
        d->c[1] = g;                                                    \
        d->c[2] = b;                                                    \
        }
#include "swrast/s_tritemp.h"

/*
 * Flat, depth-buffered, 24-bit BGR triangle.
 */
#define NAME flat_z_triangle_BGR
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_24(snap,X,Y)
#define PIXEL_TYPE          rgb_t
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    GLubyte         r = v2->color[RCOMP];                               \
    GLubyte         g = v2->color[GCOMP];                               \
    GLubyte         b = v2->color[BCOMP];
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    rgb_t   *d = pRow;                                                  \
    for (i = 0; i < span.end; i++, d++) {                               \
        const DEPTH_TYPE z = FixedToDepth(span.z);                      \
        if (z < zRow[i]) {                                              \
            d->c[0] = r;                                                \
            d->c[1] = g;                                                \
            d->c[2] = b;                                                \
            zRow[i] = z;                                                \
            }                                                           \
        span.z += span.zStep;                                           \
        }
#include "swrast/s_tritemp.h"

/*
 * Smooth, NON-depth-buffered, 24-bit BGR triangle.
 */
#define NAME smooth_triangle_BGR
#define INTERP_RGB          1
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_24(snap,X,Y)
#define PIXEL_TYPE          rgb_t
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    rgb_t   *d = pRow;                                                  \
    for (i = 0; i < span.end; i++, d++) {                               \
        d->c[0] = (GLubyte)FixedToInt(span.red);                        \
        d->c[1] = (GLubyte)FixedToInt(span.green);                      \
        d->c[2] = (GLubyte)FixedToInt(span.blue);                       \
        span.red += span.redStep;                                       \
        span.green += span.greenStep;                                   \
        span.blue += span.blueStep;                                     \
        }
#include "swrast/s_tritemp.h"

/*
 * Smooth, depth-buffered, 24-bit BGR triangle.
 */
#define NAME smooth_z_triangle_BGR
#define INTERP_RGB          1
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_24(snap,X,Y)
#define PIXEL_TYPE          rgb_t
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    rgb_t   *d = pRow;                                                  \
    for (i = 0; i < span.end; i++, d++) {                               \
        const DEPTH_TYPE z = FixedToDepth(span.z);                      \
        if (z < zRow[i]) {                                              \
            d->c[0] = (GLubyte)FixedToInt(span.red);                    \
            d->c[1] = (GLubyte)FixedToInt(span.green);                  \
            d->c[2] = (GLubyte)FixedToInt(span.blue);                   \
            zRow[i] = z;                                                \
            }                                                           \
        span.red += span.redStep;                                       \
        span.green += span.greenStep;                                   \
        span.blue += span.blueStep;                                     \
        span.z += span.zStep;                                           \
        }
#include "swrast/s_tritemp.h"

/*
 * Flat, NON-depth-buffered, 32-bit ARGB triangle.
 */
#define NAME flat_triangle_ARGB
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_32(snap,X,Y)
#define PIXEL_TYPE          GLuint
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    const GLubyte   *c = v2->color;                                     \
    GLuint          pixel = PACK_COLOR_ARGB(c[RCOMP],c[GCOMP],c[BCOMP],c[ACOMP]);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLuint  *d = pRow;                                                  \
    for (i = 0; i < span.end; i++)                                      \
        *d++ = pixel;
#include "swrast/s_tritemp.h"

/*
 * Flat, depth-buffered, 32-bit ARGB triangle.
 */
#define NAME flat_z_triangle_ARGB
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_32(snap,X,Y)
#define PIXEL_TYPE          GLuint
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    const GLubyte   *c = v2->color;                                     \
    GLuint          pixel = PACK_COLOR_ARGB(c[RCOMP],c[GCOMP],c[BCOMP],c[ACOMP]);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLuint  *d = pRow;                                                  \
    for (i = 0; i < span.end; i++, d++) {                               \
        const DEPTH_TYPE z = FixedToDepth(span.z);                      \
        if (z < zRow[i]) {                                              \
            *d = pixel;                                                 \
            zRow[i] = z;                                                \
            }                                                           \
        span.z += span.zStep;                                           \
        }
#include "swrast/s_tritemp.h"

/*
 * Smooth, NON-depth-buffered, 32-bit ARGB triangle.
 */
#define NAME smooth_triangle_ARGB
#define INTERP_RGB          1
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_32(snap,X,Y)
#define PIXEL_TYPE          GLuint
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLuint  *d = pRow;                                                  \
    for (i = 0; i < span.end; i++, d++) {                               \
        *d = PACK_COLOR_ARGB(                                           \
            FixedToInt(span.red),                                       \
            FixedToInt(span.green),                                     \
            FixedToInt(span.blue),                                      \
            0);                                                         \
        span.red += span.redStep;                                       \
        span.green += span.greenStep;                                   \
        span.blue += span.blueStep;                                     \
        }
#include "swrast/s_tritemp.h"

/*
 * Smooth, depth-buffered, 32-bit ARGB triangle.
 */
#define NAME smooth_z_triangle_ARGB
#define INTERP_RGB          1
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_32(snap,X,Y)
#define PIXEL_TYPE          GLuint
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLuint  *d = pRow;                                                  \
    for (i = 0; i < span.end; i++, d++) {                               \
        const DEPTH_TYPE z = FixedToDepth(span.z);                      \
        if (z < zRow[i]) {                                              \
            *d = PACK_COLOR_ARGB(                                       \
                FixedToInt(span.red),                                   \
                FixedToInt(span.green),                                 \
                FixedToInt(span.blue),                                  \
                0);                                                     \
            zRow[i] = z;                                                \
            }                                                           \
        span.red += span.redStep;                                       \
        span.green += span.greenStep;                                   \
        span.blue += span.blueStep;                                     \
        span.z += span.zStep;                                           \
        }
#include "swrast/s_tritemp.h"

/*
 * Flat, NON-depth-buffered, 32-bit ABGR triangle.
 */
#define NAME flat_triangle_ABGR
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_32(snap,X,Y)
#define PIXEL_TYPE          GLuint
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    const GLubyte   *c = v2->color;                                     \
    GLuint          pixel = PACK_COLOR_ABGR(c[RCOMP],c[GCOMP],c[BCOMP],c[ACOMP]);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLuint  *d = pRow;                                                  \
    for (i = 0; i < span.end; i++)                                      \
        *d++ = pixel;
#include "swrast/s_tritemp.h"

/*
 * Flat, depth-buffered, 32-bit ABGR triangle.
 */
#define NAME flat_z_triangle_ABGR
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_32(snap,X,Y)
#define PIXEL_TYPE          GLuint
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    const GLubyte   *c = v2->color;                                     \
    GLuint          pixel = PACK_COLOR_ABGR(c[RCOMP],c[GCOMP],c[BCOMP],c[ACOMP]);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLuint  *d = pRow;                                                  \
    for (i = 0; i < span.end; i++, d++) {                               \
        const DEPTH_TYPE z = FixedToDepth(span.z);                      \
        if (z < zRow[i]) {                                              \
            *d = pixel;                                                 \
            zRow[i] = z;                                                \
            }                                                           \
        span.z += span.zStep;                                           \
        }
#include "swrast/s_tritemp.h"

/*
 * Smooth, NON-depth-buffered, 32-bit ABGR triangle.
 */
#define NAME smooth_triangle_ABGR
#define INTERP_RGB          1
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_32(snap,X,Y)
#define PIXEL_TYPE          GLuint
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLuint  *d = pRow;                                                  \
    for (i = 0; i < span.end; i++, d++) {                               \
        *d = PACK_COLOR_ABGR(                                           \
            FixedToInt(span.red),                                       \
            FixedToInt(span.green),                                     \
            FixedToInt(span.blue),                                      \
            0);                                                         \
        span.red += span.redStep;                                       \
        span.green += span.greenStep;                                   \
        span.blue += span.blueStep;                                     \
        }
#include "swrast/s_tritemp.h"

/*
 * Smooth, depth-buffered, 32-bit ABGR triangle.
 */
#define NAME smooth_z_triangle_ABGR
#define INTERP_RGB          1
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_32(snap,X,Y)
#define PIXEL_TYPE          GLuint
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLuint  *d = pRow;                                                  \
    for (i = 0; i < span.end; i++, d++) {                               \
        const DEPTH_TYPE z = FixedToDepth(span.z);                      \
        if (z < zRow[i]) {                                              \
            *d = PACK_COLOR_ABGR(                                       \
                FixedToInt(span.red),                                   \
                FixedToInt(span.green),                                 \
                FixedToInt(span.blue),                                  \
                0);                                                     \
            zRow[i] = z;                                                \
            }                                                           \
        span.red += span.redStep;                                       \
        span.green += span.greenStep;                                   \
        span.blue += span.blueStep;                                     \
        span.z += span.zStep;                                           \
        }
#include "swrast/s_tritemp.h"

/*
 * Flat, NON-depth-buffered, 32-bit RGBA triangle.
 */
#define NAME flat_triangle_RGBA
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_32(snap,X,Y)
#define PIXEL_TYPE          GLuint
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    const GLubyte   *c = v2->color;                                     \
    GLuint          pixel = PACK_COLOR_RGBA(c[RCOMP],c[GCOMP],c[BCOMP],c[ACOMP]);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLuint  *d = pRow;                                                  \
    for (i = 0; i < span.end; i++)                                      \
        *d++ = pixel;
#include "swrast/s_tritemp.h"

/*
 * Flat, depth-buffered, 32-bit RGBA triangle.
 */
#define NAME flat_z_triangle_RGBA
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_32(snap,X,Y)
#define PIXEL_TYPE          GLuint
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    const GLubyte   *c = v2->color;                                     \
    GLuint          pixel = PACK_COLOR_RGBA(c[RCOMP],c[GCOMP],c[BCOMP],c[ACOMP]);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLuint  *d = pRow;                                                  \
    for (i = 0; i < span.end; i++, d++) {                               \
        const DEPTH_TYPE z = FixedToDepth(span.z);                      \
        if (z < zRow[i]) {                                              \
            *d = pixel;                                                 \
            zRow[i] = z;                                                \
            }                                                           \
        span.z += span.zStep;                                           \
        }
#include "swrast/s_tritemp.h"

/*
 * Smooth, NON-depth-buffered, 32-bit RGBA triangle.
 */
#define NAME smooth_triangle_RGBA
#define INTERP_RGB          1
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_32(snap,X,Y)
#define PIXEL_TYPE          GLuint
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLuint  *d = pRow;                                                  \
    for (i = 0; i < span.end; i++, d++) {                               \
        *d = PACK_COLOR_RGBA(                                           \
            FixedToInt(span.red),                                       \
            FixedToInt(span.green),                                     \
            FixedToInt(span.blue),                                      \
            0);                                                         \
        span.red += span.redStep;                                       \
        span.green += span.greenStep;                                   \
        span.blue += span.blueStep;                                     \
        }
#include "swrast/s_tritemp.h"

/*
 * Smooth, depth-buffered, 32-bit RGBA triangle.
 */
#define NAME smooth_z_triangle_RGBA
#define INTERP_RGB          1
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_32(snap,X,Y)
#define PIXEL_TYPE          GLuint
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLuint  *d = pRow;                                                  \
    for (i = 0; i < span.end; i++, d++) {                               \
        const DEPTH_TYPE z = FixedToDepth(span.z);                      \
        if (z < zRow[i]) {                                              \
            *d = PACK_COLOR_RGBA(                                       \
                FixedToInt(span.red),                                   \
                FixedToInt(span.green),                                 \
                FixedToInt(span.blue),                                  \
                0);                                                     \
            zRow[i] = z;                                                \
            }                                                           \
        span.red += span.redStep;                                       \
        span.green += span.greenStep;                                   \
        span.blue += span.blueStep;                                     \
        span.z += span.zStep;                                           \
        }
#include "swrast/s_tritemp.h"

/*
 * Flat, NON-depth-buffered, 32-bit BGRA triangle.
 */
#define NAME flat_triangle_BGRA
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_32(snap,X,Y)
#define PIXEL_TYPE          GLuint
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    const GLubyte   *c = v2->color;                                     \
    GLuint          pixel = PACK_COLOR_BGRA(c[RCOMP],c[GCOMP],c[BCOMP],c[ACOMP]);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLuint  *d = pRow;                                                  \
    for (i = 0; i < span.end; i++)                                      \
        *d++ = pixel;
#include "swrast/s_tritemp.h"

/*
 * Flat, depth-buffered, 32-bit BGRA triangle.
 */
#define NAME flat_z_triangle_BGRA
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_32(snap,X,Y)
#define PIXEL_TYPE          GLuint
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);                          \
    const GLubyte   *c = v2->color;                                     \
    GLuint          pixel = PACK_COLOR_BGRA(c[RCOMP],c[GCOMP],c[BCOMP],c[ACOMP]);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLuint  *d = pRow;                                                  \
    for (i = 0; i < span.end; i++, d++) {                               \
        const DEPTH_TYPE z = FixedToDepth(span.z);                      \
        if (z < zRow[i]) {                                              \
            *d = pixel;                                                 \
            zRow[i] = z;                                                \
            }                                                           \
        span.z += span.zStep;                                           \
        }
#include "swrast/s_tritemp.h"

/*
 * Smooth, NON-depth-buffered, 32-bit BGRA triangle.
 */
#define NAME smooth_triangle_BGRA
#define INTERP_RGB          1
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_32(snap,X,Y)
#define PIXEL_TYPE          GLuint
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLuint  *d = pRow;                                                  \
    for (i = 0; i < span.end; i++, d++) {                               \
        *d = PACK_COLOR_BGRA(                                           \
            FixedToInt(span.red),                                       \
            FixedToInt(span.green),                                     \
            FixedToInt(span.blue),                                      \
            0);                                                         \
        span.red += span.redStep;                                       \
        span.green += span.greenStep;                                   \
        span.blue += span.blueStep;                                     \
        }
#include "swrast/s_tritemp.h"

/*
 * Smooth, depth-buffered, 32-bit BGRA triangle.
 */
#define NAME smooth_z_triangle_BGRA
#define INTERP_RGB          1
#define INTERP_Z            1
#define DEPTH_TYPE          DEFAULT_SOFTWARE_DEPTH_TYPE
#define PIXEL_ADDRESS(X,Y)  PIXEL_ADDR_32(snap,X,Y)
#define PIXEL_TYPE          GLuint
#define BYTES_PER_ROW       (snap->stride)
#define SETUP_CODE                                                      \
    GA_glContext    *snap = SNAP_CONTEXT(ctx);
#define RENDER_SPAN(span)                                               \
    GLuint  i;                                                          \
    GLuint  *d = pRow;                                                  \
    for (i = 0; i < span.end; i++, d++) {                               \
        const DEPTH_TYPE z = FixedToDepth(span.z);                      \
        if (z < zRow[i]) {                                              \
            *d = PACK_COLOR_BGRA(                                       \
                FixedToInt(span.red),                                   \
                FixedToInt(span.green),                                 \
                FixedToInt(span.blue),                                  \
                0);                                                     \
            zRow[i] = z;                                                \
            }                                                           \
        span.red += span.redStep;                                       \
        span.green += span.greenStep;                                   \
        span.blue += span.blueStep;                                     \
        span.z += span.zStep;                                           \
        }
#include "swrast/s_tritemp.h"

/****************************************************************************
REMARKS:
Determine which triangle function we can hook depending on the current
rendering state.
****************************************************************************/
static swrast_tri_func get_triangle_func(
    GLcontext *ctx,
    SWcontext *swrast)
{
    GA_glContext    *snap = SNAP_CONTEXT(ctx);

    /* Ignore everything we cannot render directly */
    if (ctx->RenderMode != GL_RENDER)           return NULL;
    if (ctx->Polygon.SmoothFlag)                return NULL;
    if (ctx->Texture._EnabledUnits)             return NULL;
    if (swrast->_RasterMask & MULTI_DRAW_BIT)   return NULL;
    if (ctx->Line.StippleFlag)                  return NULL;
    if (ctx->Polygon.CullFlag && ctx->Polygon.CullFaceMode == GL_FRONT_AND_BACK)
        return NULL;

    /* Now hook out those functions we can support */
    if (ctx->Light.ShadeModel == GL_SMOOTH
        && swrast->_RasterMask == DEPTH_BIT
        && ctx->Depth.Func == GL_LESS
        && ctx->Depth.Mask == GL_TRUE
        && ctx->Visual.depthBits == DEFAULT_SOFTWARE_DEPTH_BITS) {
        /* Flat or smooth triangles with GL_LESS depth buffering enabled */
        if (ctx->Light.ShadeModel == GL_SMOOTH) {
            switch (snap->pixelFormat) {
                case PF_RGB555:
                    return smooth_z_triangle_RGB_555;
                case PF_DITHER555:
                    return smooth_z_triangle_DITHER_555;
                case PF_RGB565:
                    return smooth_z_triangle_RGB_565;
                case PF_DITHER565:
                    return smooth_z_triangle_DITHER_565;
                case PF_RGB888:
                    return smooth_z_triangle_RGB;
                case PF_BGR888:
                    return smooth_z_triangle_BGR;
                case PF_ARGB8888:
                    return smooth_z_triangle_ARGB;
                case PF_ABGR8888:
                    return smooth_z_triangle_ABGR;
                case PF_RGBA8888:
                    return smooth_z_triangle_RGBA;
                case PF_BGRA8888:
                    return smooth_z_triangle_BGRA;
                default:
                    return NULL;
                }
            }
        else if (ctx->Light.ShadeModel == GL_FLAT) {
            switch (snap->pixelFormat) {
                case PF_INDEX:
                    return flat_z_triangle_CI8;
                case PF_RGB555:
                    return flat_z_triangle_RGB_555;
                case PF_DITHER555:
                    return flat_z_triangle_DITHER_555;
                case PF_RGB565:
                    return flat_z_triangle_RGB_565;
                case PF_DITHER565:
                    return flat_z_triangle_DITHER_565;
                case PF_RGB888:
                    return flat_z_triangle_RGB;
                case PF_BGR888:
                    return flat_z_triangle_BGR;
                case PF_ARGB8888:
                    return flat_z_triangle_ARGB;
                case PF_ABGR8888:
                    return flat_z_triangle_ABGR;
                case PF_RGBA8888:
                    return flat_z_triangle_RGBA;
                case PF_BGRA8888:
                    return flat_z_triangle_BGRA;
                default:
                    return NULL;
                }
            }
        }
    if (swrast->_RasterMask == 0) {
        /* Flat or smooth shading with no depth testing */
        if (ctx->Light.ShadeModel == GL_SMOOTH) {
            switch (snap->pixelFormat) {
                case PF_RGB555:
                    return smooth_triangle_RGB_555;
                case PF_DITHER555:
                    return smooth_triangle_DITHER_555;
                case PF_RGB565:
                    return smooth_triangle_RGB_565;
                case PF_DITHER565:
                    return smooth_triangle_DITHER_565;
                case PF_RGB888:
                    return smooth_triangle_RGB;
                case PF_BGR888:
                    return smooth_triangle_BGR;
                case PF_ARGB8888:
                    return smooth_triangle_ARGB;
                case PF_ABGR8888:
                    return smooth_triangle_ABGR;
                case PF_RGBA8888:
                    return smooth_triangle_RGBA;
                case PF_BGRA8888:
                    return smooth_triangle_BGRA;
                default:
                    return NULL;
                }
            }
        else if (ctx->Light.ShadeModel == GL_FLAT) {
            switch (snap->pixelFormat) {
                case PF_INDEX:
                    return flat_triangle_CI8;
                case PF_RGB555:
                    return flat_triangle_RGB_555;
                case PF_DITHER555:
                    return flat_triangle_DITHER_555;
                case PF_RGB565:
                    return flat_triangle_RGB_565;
                case PF_DITHER565:
                    return flat_triangle_DITHER_565;
                case PF_RGB888:
                    return flat_triangle_RGB;
                case PF_BGR888:
                    return flat_triangle_BGR;
                case PF_ARGB8888:
                    return flat_triangle_ARGB;
                case PF_ABGR8888:
                    return flat_triangle_ABGR;
                case PF_RGBA8888:
                    return flat_triangle_RGBA;
                case PF_BGRA8888:
                    return flat_triangle_BGRA;
                default:
                    return NULL;
                }
            }
        }
    return NULL;
}

/****************************************************************************
REMARKS:
Override for the swrast tri-selection function.  Try to use one of our
internal tri functions, otherwise fall back to the standard swrast functions.
****************************************************************************/
void _snap_choose_triangle(
    GLcontext *ctx)
{
    SWcontext *swrast = SWRAST_CONTEXT(ctx);

    if ((swrast->Triangle = get_triangle_func(ctx,swrast)) == NULL)
        _swrast_choose_triangle(ctx);
}

