/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Description:	This file contains the code specific to initialising the
*		        SciTech SNAP OpenGL utility library module.
*
****************************************************************************/

#include "snap/graphics/common.h"
#include "GL/gl.h"
#include "GL/glu.h"
#include <stdio.h>
#include <string.h>

/*------------------------- Global Variables ------------------------------*/

PM_imports _VARAPI	_PM_imports;
GA_glFuncs _VARAPI	_GA_glFuncs;

/*-------------------------- Implementation -------------------------------*/

/****************************************************************************
REMARKS:
Fatal error handler called when a non-imported function is called by the
driver. We leave this to a runtime error so that older applications and
shell drivers will work with newer bpd drivers provided no newer functions
are required by the driver itself. If they are, the application or shell
driver needs to be recompiled.
****************************************************************************/
static void _PM_fatalErrorHandler(void)
{
    _PM_imports.PM_fatalError("Unsupported PM_imports import function called! Please re-compile!\n");
}

/****************************************************************************
REMARKS:
Fatal error handler called when a non-imported function is called by the
driver. We leave this to a runtime error so that older applications and
shell drivers will work with newer bpd drivers provided no newer functions
are required by the driver itself. If they are, the application or shell
driver needs to be recompiled.
****************************************************************************/
static void _GL_fatalErrorHandler(void)
{
    _PM_imports.PM_fatalError("Unsupported GA_glFuncs import function called! Please re-compile!\n");
}

/****************************************************************************
PARAMETERS:
pmImp   - Pointer to PM library imports
glFuncs - Pointer to OpenGL library functions

RETURNS:
Pointer to exported function list

REMARKS:
This function initialises the GL image library and returns the list of
loader library exported functions.
{secret}
****************************************************************************/
GA_gluFuncs * _CEXPORT GA_gluInitDriver(
    PM_imports *pmImp,
    GA_glFuncs *glFuncs)
{
    static GA_gluFuncs _GA_gluFuncs = {
    	sizeof(GA_gluFuncs),
    	gluBeginCurve,
    	gluBeginPolygon,
    	gluBeginSurface,
    	gluBeginTrim,
    	gluBuild1DMipmaps,
    	gluBuild2DMipmaps,
    	gluCylinder,
    	gluDeleteNurbsRenderer,
    	gluDeleteQuadric,
    	gluDeleteTess,
    	gluDisk,
    	gluEndCurve,
    	gluEndPolygon,
    	gluEndSurface,
    	gluEndTrim,
    	gluErrorString,
    	gluGetNurbsProperty,
    	gluGetString,
    	gluGetTessProperty,
    	gluLoadSamplingMatrices,
    	gluLookAt,
    	gluNewNurbsRenderer,
    	gluNewQuadric,
    	gluNewTess,
    	gluNextContour,
    	gluNurbsCallback,
    	gluNurbsCurve,
    	gluNurbsProperty,
    	gluNurbsSurface,
    	gluOrtho2D,
    	gluPartialDisk,
    	gluPerspective,
    	gluPickMatrix,
    	gluProject,
    	gluPwlCurve,
    	gluQuadricCallback,
    	gluQuadricDrawStyle,
    	gluQuadricNormals,
    	gluQuadricOrientation,
    	gluQuadricTexture,
    	gluScaleImage,
    	gluSphere,
    	gluTessBeginContour,
    	gluTessBeginPolygon,
    	gluTessCallback,
    	gluTessEndContour,
    	gluTessEndPolygon,
    	gluTessNormal,
    	gluTessProperty,
    	gluTessVertex,
    	gluUnProject,
    	};
    int     i,max;
    ulong   *p;

    // Initialize all default imports to point to fatal error handler
    // for upwards compatibility.
    max = sizeof(_PM_imports)/sizeof(GA_gluInitDriver_t);
    for (i = 0,p = (ulong*)&_PM_imports; i < max; i++)
	    *p++ = (ulong)_PM_fatalErrorHandler;
    max = sizeof(_GA_glFuncs)/sizeof(GA_gluInitDriver_t);
    for (i = 0,p = (ulong*)&_GA_glFuncs; i < max; i++)
	    *p++ = (ulong)_GL_fatalErrorHandler;

    // Now copy all our imported functions
    memcpy(&_PM_imports,pmImp,MIN(sizeof(_PM_imports),pmImp->dwSize));
    memcpy(&_GA_glFuncs,glFuncs,MIN(sizeof(_GA_glFuncs),glFuncs->dwSize));
    return &_GA_gluFuncs;
}

/****************************************************************************
REMARKS:
Function to clean up and exit the GLU library. Nothing to do here for the
moment.
{secret}
****************************************************************************/
void _CEXPORT GA_gluExitDriver(void)
{
}

