/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Simple command line program to create a new C code fragment
*               for a GTF CRTC timings, with a range of frequency values.
*
****************************************************************************/

#include "snap/graphics/crtc.h"
#include "gtf.h"
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/*----------------------------- Implementation ----------------------------*/

int round(float val)
{
    return (int)(val + 0.5);
}

void main(int argc,char *argv[])
{
    double      xPixels,yPixels,freq,minfreq,maxfreq,stepfreq;
    GTF_timings t;
    ibool       doubleScan = false;
    ibool       tripleScan = false;
    ibool       interlaced = false;

    if (argc != 4 && argc != 6 && argc != 7) {
        printf("Usage: MKTIMING <xPixels> <yPixels> <minfreq> [<maxfreq> <stepfreq>] [double | triple | interlaced]\n");
        exit(1);
        }

    /* Get values from command line */
    xPixels = atof(argv[1]);
    yPixels = atof(argv[2]);
    minfreq = atof(argv[3]);
    if (argc >= 6) {
        maxfreq = atof(argv[4]);
        stepfreq = atof(argv[5]);
        }
    else {
        maxfreq = minfreq;
        stepfreq = 1;
        }
    if (argc == 7) {
        if (argv[6][0] == 'd')
            doubleScan = true;
        if (argv[6][0] == 't')
            tripleScan = true;
        if (argv[6][0] == 'i')
            interlaced = true;
        }

    /* Compute the CRTC timings */
    for (freq = minfreq; freq <= maxfreq; freq += stepfreq) {
        if (doubleScan)
            GTF_calcTimings(xPixels,yPixels*2,freq,GTF_lockVF,false,false,&t);
        else if (tripleScan)
            GTF_calcTimings(xPixels,yPixels*3,freq,GTF_lockVF,false,false,&t);
        else
            GTF_calcTimings(xPixels,yPixels,freq,GTF_lockVF,false,interlaced,&t);

        printf("    { %4d, %4d, %4d, %4d, %4d, %4d, %4d, %4d,   // %dx%d @ %d Hz%s\n",
            (int)xPixels,(int)yPixels,
            t.h.hTotal,t.h.hSyncStart,t.h.hSyncEnd,
            t.v.vTotal,t.v.vSyncStart,t.v.vSyncEnd,
            (int)xPixels,(int)yPixels,round(freq),
            interlaced ? " I" : "");
        if (doubleScan) {
            printf("      %8ld, %ld, gaHSyncNeg | gaDoubleScan },\n",
                (ulong)(t.dotClock * 1000000.0),round(t.vFreq) * 100);
            }
        else if (tripleScan) {
            printf("      %8ld, %ld, gaHSyncNeg | gaTripleScan },\n",
                (ulong)(t.dotClock * 1000000.0),round(t.vFreq) * 100);
            }
        else if (interlaced) {
            printf("      %8ld, %ld, gaHSyncNeg | gaInterlaced },\n",
                (ulong)(t.dotClock * 1000000.0),round(t.vFreq) * 100);
            }
        else {
            printf("      %8ld, %ld, gaHSyncNeg },\n",
                (ulong)(t.dotClock * 1000000.0),round(t.vFreq) * 100);
            }
        }
}

