;****************************************************************************
;*
;*                    SciTech SNAP Graphics Architecture
;*
;*  ========================================================================
;*
;*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
;*
;*   This file may be distributed and/or modified under the terms of the
;*   GNU General Public License version 2 as published by the Free
;*   Software Foundation and appearing in the file LICENSE.GPL included
;*   in the packaging of this file.
;*
;*   Licensees holding a valid Commercial License for this product from
;*   SciTech Software, Inc. may use this file in accordance with the
;*   Commercial License Agreement provided with the Software.
;*
;*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
;*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
;*   PURPOSE.
;*
;*   See http://www.scitechsoft.com/license/ for information about
;*   the licensing options available and how to purchase a Commercial
;*   License Agreement.
;*
;*   Contact license@scitechsoft.com if any conditions of this licensing
;*   are not clear to you, or you have questions about licensing options.
;*
;*  ========================================================================
;*
;* Language:    NASM
;* Environment: IBM PC 32 bit Protected Mode.
;*
;* Description: Module to implement the call stubs for all the SNAP
;*              Reference Rasteriser drawing functions. The stubs jump
;*              directly to the appropriate routine pointed to by our
;*              global structure of currently active routines (either
;*              hardware or software).
;*
;****************************************************************************

include "scitech.mac"               ; Memory model macros

GA_REPLACE_MIX  equ 12              ; Replace mix code

;----------------------------------------------------------------------------
; Macro for functions that check to see if the ROP2 mix is one of the
; bogus mixes without any source, and then punts to the software
; rasteriser to handle that mix for consistent results. Some hardware will
; hang with some functions on these mixes, and other hardware will do
; things differently depending on when the mix is applied (ie: pre or post
; transparency etc). Normally the OS shell drivers will handle these
; special case ROPs much earlier to avoid these types of problems, but we
; handle them anyway to ensure our drivers are consistent.
;----------------------------------------------------------------------------
%imacro DECLARE_ROP_STUB    1
%ifdef  __NOU__
        global %1
%1:
%else
        global _%1
_%1:
%endif
        mov     al,[BYTE REF_mix]
        mov     ah,al
        shr     ah,2
        xor     al,ah
        and     al,03h
        jz      @@1
        jmp     [DWORD STUBS_START+off]
@@1:    jmp     [DWORD REF_draw2dSW+off]
%assign off off+4
%endmacro

;----------------------------------------------------------------------------
; Macro for *BitBltSys functions that check for GA_REPLACE_MIX
; since the hardware will be slower than software for this operation.
;----------------------------------------------------------------------------
%imacro DECLARE_ROPPARM_STUB    2
%ifdef  __NOU__
        global %1
%1:
%else
        global _%1
_%1:
%endif
        mov     al,[BYTE esp+%2*4]
        mov     ah,al
        shr     ah,2
        xor     al,ah
        and     al,03h
        jz      @@1
        jmp     [DWORD STUBS_START+off]
@@1:    jmp     [DWORD REF_draw2dSW+off]
%assign off off+4
%endmacro

;----------------------------------------------------------------------------
; Macro for *BitBltSys functions that check for GA_REPLACE_MIX
; since the hardware will be slower than software for this operation.
;----------------------------------------------------------------------------
%imacro DECLARE_SLOW_BLTSYS_STUB    2
%ifdef  __NOU__
        global %1
%1:
%else
        global _%1
_%1:
%endif
        cmp     [BYTE REF_slowBltSys],1
        je      @@1
        cmp     [BYTE esp+%2*4],GA_REPLACE_MIX
        je      @@2
@@1:    mov     al,[BYTE esp+%2*4]
        mov     ah,al
        shr     ah,2
        xor     al,ah
        and     al,03h
        jz      @@2
        jmp     [DWORD STUBS_START+off]
@@2:    jmp     [DWORD REF_draw2dSW+off]
%assign off off+4
%endmacro

begdataseg  _stubs
        cextern REF_draw2dSW,DWORD
        cextern REF_slowBltSys,UCHAR
        cextern REF_mix,DWORD
enddataseg  _stubs

include "_stubs.def"

        END

