/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Blended pixel functions for 32bpp color depths.
*
****************************************************************************/

#include "ref2dint.h"

/*----------------------------- Implementation ----------------------------*/

#define COLOR_TYPE  ushort

/****************************************************************************
REMARKS:
Optimised blended pixel functions for common source alpha and constant
alpha pixel blending functions. Destination alpha is always computed as
a zero for speed.
****************************************************************************/

void REF2D_drawPixelBlendedFast_15(
    uchar *addr DCL_X_ARG,
    GA_color S,
    GA_color A)
{
    GA_color D = *((COLOR_TYPE*)addr);
    if (A) A = (A + 1) / 8;
    S = ((S & 0xFFFF) | (S << 16)) & 0x3E07C1F;
    D = ((D & 0xFFFF) | (D << 16)) & 0x3E07C1F;
    D = ((S - D) * A / 32 + D) & 0x3E07C1F;
    *((COLOR_TYPE*)addr) = (COLOR_TYPE)((D & 0xFFFF) | (D >> 16));
}

void REF2D_drawPixelBlendedFast_16(
    uchar *addr DCL_X_ARG,
    GA_color S,
    GA_color A)
{
    GA_color D = *((COLOR_TYPE*)addr);
    if (A) A = (A + 1) / 8;
    S = ((S & 0xFFFF) | (S << 16)) & 0x7E0F81F;
    D = ((D & 0xFFFF) | (D << 16)) & 0x7E0F81F;
    D = ((S - D) * A / 32 + D) & 0x7E0F81F;
    *((COLOR_TYPE*)addr) = (COLOR_TYPE)((D & 0xFFFF) | (D >> 16));
}

void REF2D_drawPixelBlendedSrcAlphaFast_15(
    uchar *addr DCL_X_ARG,
    GA_color S)
{
    if (UNPACK_A_15(S))
        *((COLOR_TYPE*)addr) = S;
}

void REF2D_drawPixelBlendedSrcAlphaFast_16(
    uchar *addr DCL_X_ARG,
    GA_color S)
{
    (void)addr;
    (void)S;
}

void REF2D_drawPixelBlendedConstAlphaFast_15(
    uchar *addr DCL_X_ARG,
    GA_color S)
{
    GA_color D = *((COLOR_TYPE*)addr);
    GA_color A = DC.alphaValue;
    if (A) A = (A + 1) / 8;
    S = ((S & 0xFFFF) | (S << 16)) & 0x3E07C1F;
    D = ((D & 0xFFFF) | (D << 16)) & 0x3E07C1F;
    D = ((S - D) * A / 32 + D) & 0x3E07C1F;
    *((COLOR_TYPE*)addr) = (COLOR_TYPE)((D & 0xFFFF) | (D >> 16));
}

void REF2D_drawPixelBlendedConstAlphaFast_16(
    uchar *addr DCL_X_ARG,
    GA_color S)
{
    GA_color D = *((COLOR_TYPE*)addr);
    GA_color A = DC.alphaValue;
    if (A) A = (A + 1) / 8;
    S = ((S & 0xFFFF) | (S << 16)) & 0x7E0F81F;
    D = ((D & 0xFFFF) | (D << 16)) & 0x7E0F81F;
    D = ((S - D) * A / 32 + D) & 0x7E0F81F;
    *((COLOR_TYPE*)addr) = (COLOR_TYPE)((D & 0xFFFF) | (D >> 16));
}

/****************************************************************************
REMARKS:
Standard blended pixel functions
****************************************************************************/

#define DRAW_PIXEL_PROTO()  \
    void REF2D_drawPixelBlended_15(uchar *addr DCL_X_ARG,GA_color color)
#define REPACK_COLOR(c)     PACK_COLOR_ARGB_ALPHA(UNPACK_A_15(c),UNPACK_R_15(c),UNPACK_G_15(c),UNPACK_B_15(c))
#define PACK_COLOR(D)       PACK_COLOR_15_ALPHA(UNPACK_A_ARGB(D),UNPACK_R_ARGB(D),UNPACK_G_ARGB(D),UNPACK_B_ARGB(D))
#define CONST_ALPHA         DC.alphaValue
#include "raster/blendpix.c"

#define DRAW_PIXEL_PROTO()  \
    void REF2D_drawPixelBlended_16(uchar *addr DCL_X_ARG,GA_color color)
#define REPACK_COLOR(c)     PACK_COLOR_ARGB_ALPHA(UNPACK_A_16(c),UNPACK_R_16(c),UNPACK_G_16(c),UNPACK_B_16(c))
#define PACK_COLOR(D)       PACK_COLOR_16_ALPHA(UNPACK_A_ARGB(D),UNPACK_R_ARGB(D),UNPACK_G_ARGB(D),UNPACK_B_ARGB(D))
#define CONST_ALPHA         DC.alphaValue
#include "raster/blendpix.c"

#define DRAW_PIXEL_PROTO()  \
    void REF2D_drawPixelBlendedClamped_15(uchar *addr DCL_X_ARG,GA_color color)
#define REPACK_COLOR(c)     PACK_COLOR_ARGB_ALPHA(UNPACK_A_15(c),UNPACK_R_15(c),UNPACK_G_15(c),UNPACK_B_15(c))
#define PACK_COLOR(a,r,g,b) PACK_COLOR_15_ALPHA(a,r,g,b)
#define CONST_ALPHA         DC.alphaValue
#include "raster/clamppix.c"

#define DRAW_PIXEL_PROTO()  \
    void REF2D_drawPixelBlendedClamped_16(uchar *addr DCL_X_ARG,GA_color color)
#define REPACK_COLOR(c)     PACK_COLOR_ARGB_ALPHA(UNPACK_A_16(c),UNPACK_R_16(c),UNPACK_G_16(c),UNPACK_B_16(c))
#define PACK_COLOR(a,r,g,b) PACK_COLOR_16_ALPHA(a,r,g,b)
#define CONST_ALPHA         DC.alphaValue
#include "raster/clamppix.c"

