/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Blended pixel functions for 32bpp color depths.
*
****************************************************************************/

#include "ref2dint.h"

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
REMARKS:
Blended pixel functions
****************************************************************************/

#define xA(c)       UNPACK_A_ARGB(c)
#define xR(c)       UNPACK_R_ARGB(c)
#define xG(c)       UNPACK_G_ARGB(c)
#define xB(c)       UNPACK_B_ARGB(c)

void REF2D_drawPixelBlendedFast_24(
    uchar *addr DCL_X_ARG,
    GA_color S,
    GA_color A)
{
    GA_color    D;

    if (A) A++;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = (((((S & 0x00FF00FF) - (D & 0x00FF00FF)) * A) / 256 + D) & 0x00FF00FF)
      + (((((S & 0x0000FF00) - (D & 0x0000FF00)) * A) / 256 + D) & 0x0000FF00);
    addr[0] = UNPACK_B_ARGB(D);
    addr[1] = UNPACK_G_ARGB(D);
    addr[2] = UNPACK_R_ARGB(D);
}

void REF2D_drawPixelBlendedSrcAlphaFast_24(
    uchar *addr DCL_X_ARG,
    GA_color S)
{
    GA_color    D;
    GA_color    A = UNPACK_A_ARGB(S);

    if (A) A++;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = (((((S & 0x00FF00FF) - (D & 0x00FF00FF)) * A) / 256 + D) & 0x00FF00FF)
      + (((((S & 0x0000FF00) - (D & 0x0000FF00)) * A) / 256 + D) & 0x0000FF00);
    addr[0] = UNPACK_B_ARGB(D);
    addr[1] = UNPACK_G_ARGB(D);
    addr[2] = UNPACK_R_ARGB(D);
}

void REF2D_drawPixelBlendedConstAlphaFast_24(
    uchar *addr DCL_X_ARG,
    GA_color S)
{
    GA_color    D;
    GA_color    A = DC.alphaValue;

    if (A) A++;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = (((((S & 0x00FF00FF) - (D & 0x00FF00FF)) * A) / 256 + D) & 0x00FF00FF)
      + (((((S & 0x0000FF00) - (D & 0x0000FF00)) * A) / 256 + D) & 0x0000FF00);
    addr[0] = UNPACK_B_ARGB(D);
    addr[1] = UNPACK_G_ARGB(D);
    addr[2] = UNPACK_R_ARGB(D);
}

void REF2D_drawPixelBlended_24(
    uchar *addr DCL_X_ARG,
    GA_color S)
{
    GA_color    D = 0;

    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = DC.blendSrc(S,D,DC.foreColor,DC.alphaValue)
      + DC.blendDst(S,D,DC.foreColor,DC.alphaValue);
    addr[0] = UNPACK_B_ARGB(D);
    addr[1] = UNPACK_G_ARGB(D);
    addr[2] = UNPACK_R_ARGB(D);
}

void REF2D_drawPixelBlendedClamped_24(
    uchar *addr DCL_X_ARG,
    GA_color S)
{
    GA_color    sR,sG,sB,T,D = 0;

    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    T = DC.blendSrc(S,D,DC.foreColor,DC.alphaValue);
    D = DC.blendDst(S,D,DC.foreColor,DC.alphaValue);
    sB = UNPACK_B_ARGB(T) + UNPACK_B_ARGB(D);
    sG = UNPACK_G_ARGB(T) + UNPACK_G_ARGB(D);
    sR = UNPACK_R_ARGB(T) + UNPACK_R_ARGB(D);
    addr[0] = CLAMP(sB,255);
    addr[1] = CLAMP(sG,255);
    addr[2] = CLAMP(sR,255);
}

