/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Blended pixel functions for 32bpp color depths.
*
****************************************************************************/

#include "ref2dint.h"

/*----------------------------- Implementation ----------------------------*/

#define COLOR_TYPE  ulong

/****************************************************************************
REMARKS:
Optimised blended pixel functions for common source alpha and constant
alpha pixel blending functions. Destination alpha is always computed as
a zero for speed.
****************************************************************************/

void REF2D_drawPixelBlendedFast_Axxx32(
    uchar *addr DCL_X_ARG,
    GA_color S,
    GA_color A)
{
    GA_color D = *((COLOR_TYPE*)addr);
    if (A) A++;
    *((COLOR_TYPE*)addr) =
        (((((S & 0x00FF00FF) - (D & 0x00FF00FF)) * A) / 256 + D) & 0x00FF00FF)
      + (((((S & 0x0000FF00) - (D & 0x0000FF00)) * A) / 256 + D) & 0x0000FF00);
}

void REF2D_drawPixelBlendedFast_xxxA32(
    uchar *addr DCL_X_ARG,
    GA_color S,
    GA_color A)
{
    GA_color D = *((COLOR_TYPE*)addr) >> 8;
    if (A) A++;
    S >>= 8;
    *((COLOR_TYPE*)addr) =
        (((((((S & 0x00FF00FF) - (D & 0x00FF00FF)) * A) / 256 + D) & 0x00FF00FF)
       + (((((S & 0x0000FF00) - (D & 0x0000FF00)) * A) / 256 + D) & 0x0000FF00)) << 8);
}

void REF2D_drawPixelBlendedSrcAlphaFast_Axxx32(
    uchar *addr DCL_X_ARG,
    GA_color S)
{
    GA_color D = *((COLOR_TYPE*)addr);
    GA_color A = UNPACK_A_ARGB(S);
    if (A) A++;
    *((COLOR_TYPE*)addr) =
        (((((S & 0x00FF00FF) - (D & 0x00FF00FF)) * A) / 256 + D) & 0x00FF00FF)
      + (((((S & 0x0000FF00) - (D & 0x0000FF00)) * A) / 256 + D) & 0x0000FF00);
}

void REF2D_drawPixelBlendedSrcAlphaFast_xxxA32(
    uchar *addr DCL_X_ARG,
    GA_color S)
{
    GA_color D = *((COLOR_TYPE*)addr) >> 8;
    GA_color A = UNPACK_A_RGBA(S);
    if (A) A++;
    S >>= 8;
    *((COLOR_TYPE*)addr) =
        (((((((S & 0x00FF00FF) - (D & 0x00FF00FF)) * A) / 256 + D) & 0x00FF00FF)
       + (((((S & 0x0000FF00) - (D & 0x0000FF00)) * A) / 256 + D) & 0x0000FF00)) << 8);
}

void REF2D_drawPixelBlendedConstAlphaFast_Axxx32(
    uchar *addr DCL_X_ARG,
    GA_color S)
{
    GA_color D = *((COLOR_TYPE*)addr);
    GA_color A = DC.alphaValue;
    if (A) A++;
    *((COLOR_TYPE*)addr) =
        (((((S & 0x00FF00FF) - (D & 0x00FF00FF)) * A) / 256 + D) & 0x00FF00FF)
      + (((((S & 0x0000FF00) - (D & 0x0000FF00)) * A) / 256 + D) & 0x0000FF00);
}

void REF2D_drawPixelBlendedConstAlphaFast_xxxA32(
    uchar *addr DCL_X_ARG,
    GA_color S)
{
    GA_color D = *((COLOR_TYPE*)addr) >> 8;
    GA_color A = DC.alphaValue;
    if (A) A++;
    S >>= 8;
    *((COLOR_TYPE*)addr) =
        (((((((S & 0x00FF00FF) - (D & 0x00FF00FF)) * A) / 256 + D) & 0x00FF00FF)
       + (((((S & 0x0000FF00) - (D & 0x0000FF00)) * A) / 256 + D) & 0x0000FF00)) << 8);
}

/****************************************************************************
REMARKS:
Standard blended pixel functions
****************************************************************************/

#define DRAW_PIXEL_PROTO()  \
    void REF2D_drawPixelBlended_Axxx32(uchar *addr DCL_X_ARG,GA_color color)
#define REPACK_COLOR(c)     (c)
#define PACK_COLOR(c)       (c)
#define CONST_ALPHA         DC.alphaValue
#include "raster/blendpix.c"

#define DRAW_PIXEL_PROTO()  \
    void REF2D_drawPixelBlended_xxxA32(uchar *addr DCL_X_ARG,GA_color color)
#define REPACK_COLOR(c)     ((((c) >> 8) & 0x00FFFFFF) | ((c) << 24))
#define PACK_COLOR(c)       ((((c) << 8) & 0xFFFFFF00) | ((c) >> 24))
#define CONST_ALPHA         DC.alphaValue
#include "raster/blendpix.c"

#define DRAW_PIXEL_PROTO()  \
    void REF2D_drawPixelBlendedClamped_Axxx32(uchar *addr DCL_X_ARG,GA_color color)
#define REPACK_COLOR(c)     (c)
#define PACK_COLOR(a,r,g,b) PACK_COLOR_ARGB_ALPHA(a,r,g,b)
#define CONST_ALPHA         DC.alphaValue
#include "raster/clamppix.c"

#define DRAW_PIXEL_PROTO()  \
    void REF2D_drawPixelBlendedClamped_xxxA32(uchar *addr DCL_X_ARG,GA_color color)
#define REPACK_COLOR(c)     ((((c) >> 8) & 0x00FFFFFF) | ((c) << 24))
#define PACK_COLOR(a,r,g,b) PACK_COLOR_RGBA_ALPHA(a,r,g,b)
#define CONST_ALPHA         DC.alphaValue
#include "raster/clamppix.c"

