/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Header file for the SciTech SNAP Graphics buffer manager
*
****************************************************************************/

#ifndef __BUFINT_H
#define __BUFINT_H

/*---------------------- Macros and type definitions ----------------------*/

#pragma pack(4)

// Values stored in the DC.isStereo variable

#define gaNoStereo      0       // Not running in stereo mode
#define gaSoftStereo    1       // Running in software stereo mode
#define gaHWStereo      2       // Running in hardware stereo mode

/****************************************************************************
REMARKS:
Structure for block list of buffers.

MEMBERS:
bl          - Memory pool containg objects
freeList    - Start of free list for pool
****************************************************************************/
typedef struct {
    N_blockList bl;
    N_node      *freeList;
    } bufferList_t;

/****************************************************************************
REMARKS:
Structure to store the current state of the buffer manager
MEMBERS:
****************************************************************************/
typedef struct {
    REF2D_driver        *ref2d;
    GA_regionFuncs      rgnFuncs;
    GA_devCtx           *hwCtx;
    GA_initFuncs        hwInit;
    GA_2DStateFuncs     hwState;
    GA_2DRenderFuncs    hwDraw;
    GA_driverFuncs      hwDrv;
    N_uint32            gaModeNum;
    GA_modeInfo         gaMode;
    GA_2DStateFuncs     state2d;
    GA_2DRenderFuncs    draw2d;
    N_int32             isStereo;
    N_int32             stereoRunning;
    GA_softStereoFuncs  stereo;
    bufferList_t        bufferList;
    int                 priorityCount;
    N_list              bufferHeap;
    int                 byteShift;
    int                 activeX;
    int                 activeY;
    GA_buf              *activeBuf;
    GA_buf              *visibleBuf;
    GA_buf              *visibleRightBuf;
    N_int32             linearMoveableBottom;
    N_int32             linearMoveableCur;
    N_int32             linearFixedCur;
    ibool               compacted;
    ibool               fixedHeapFragmented;
    ibool               pagedOut;
    int                 surfLock;
    GA_region           *freeSpace;
    int                 bufWidth;
    int                 bufHeight;
    ibool               foundSpace;
    GA_rect             bestBounds;
    GA_rect             testRect;
    GA_region           *rgn;
    GA_buf              **flipBuffers;
    int                 numFlipBuffers;

    // Function to align buffer memory
    ibool               (NAPIP AlignLinearBuffer)(N_int32 height,N_int32 *stride,N_int32 *offset,N_int32 *size,N_int32 growUp);

    // Active drawing functions for current display mode (never changes)
    bufReadScan_t       bufReadScan;
    bufBltScan_t        bufBltScan;
    } BM_state;

/****************************************************************************
REMARKS:
Library initialisation entry point for the buffer manager.
****************************************************************************/
typedef ibool (NAPIP BM_initLibrary_t)(GA_devCtx *hwCtx,REF2D_driver *ref2d,bufReadScan_t bufReadScan,bufBltScan_t bufBltScan,GA_bufferFuncs *funcs);

/****************************************************************************
REMARKS:
Library exit entry point for the buffer manager.
****************************************************************************/
typedef ibool (NAPIP BM_exitLibrary_t)(void);

#pragma pack()

/*---------------------------- Global Variables ---------------------------*/

extern BM_state         BM_devCtx;

#define BM  BM_devCtx

/*-------------------------- Function Prototypes --------------------------*/

#ifdef  __cplusplus
extern "C" {                        /* Use "C" linkage when in C++ mode */
#endif

/* Buffer management functions */

ibool       BM_init(GA_devCtx *hwCtx,REF2D_driver *ref2d,bufReadScan_t bufReadScan,bufBltScan_t bufBltScan,GA_bufferFuncs *funcs);
void        BM_exit(void);
ibool       NAPI BM_InitBuffers(N_int32 numBuffers,N_uint32 flags,GA_softStereoFuncs *stereo);
GA_buf *    NAPI BM_GetPrimaryBuffer(void);
GA_buf *    NAPI BM_GetFlippableBuffer(N_int32 index);
GA_buf *    NAPI BM_AllocBuffer(N_int32 width,N_int32 height,N_int32 flags);
GA_buf *    NAPI BM_AllocBufferExt(N_int32 width,N_int32 height,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,N_int32 flags);
ibool       NAPI BM_FreeBuffer(GA_buf *buf);
N_uint32    NAPI BM_LockBuffer(GA_buf *buf);
void        NAPI BM_UnlockBuffer(GA_buf *buf);
N_int32     NAPI BM_SetActiveBuffer(GA_buf *buf);
void        NAPI BM_FlipToBuffer(GA_buf *buf,N_int32 waitVRT);
void        NAPI BM_FlipToStereoBuffer(GA_buf *left,GA_buf *right,N_int32 waitVRT);
int         NAPI BM_GetFlipStatus(void);
void        NAPI BM_WaitTillFlipped(void);
void        NAPI BM_UpdateCache(GA_buf *buf);
void        NAPI BM_UpdateFromCache(GA_buf *buf);
void        NAPI BM_BitBltBuf(GA_buf *buf,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI BM_BitBltPattBuf(GA_buf *buf,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI BM_BitBltColorPattBuf(GA_buf *buf,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI BM_SrcTransBltBuf(GA_buf *buf,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI BM_DstTransBltBuf(GA_buf *buf,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI BM_StretchBltBuf(GA_buf *buf,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix);
void        NAPI BM_BitBltPlaneMaskedBuf(GA_buf *buf,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_uint32 planeMask);
void        NAPI BM_BitBltFxBuf(GA_buf *buf,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,GA_bltFx *fx);
void        NAPI BM_DrawRectBuf(GA_buf *buf,N_int32 left,N_int32 top,N_int32 width,N_int32 height,GA_color color,N_int32 mix);
GA_buf *    NAPI BM_GetActiveBuffer(void);
GA_buf *    NAPI BM_GetVisibleBuffer(void);
GA_buf *    NAPI BM_GetVisibleRightBuffer(void);

void        NAPI BM_BitBltPattBufSys(GA_buf *buf,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI BM_BitBltColorPattBufSys(GA_buf *buf,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI BM_SrcTransBltBufSys(GA_buf *buf,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI BM_DstTransBltBufSys(GA_buf *buf,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI BM_StretchBltBufSys(GA_buf *buf,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix);
void        NAPI BM_BitBltPlaneMaskedBufSys(GA_buf *buf,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_uint32 planeMask);
N_int32     NAPI BM_BitBltFxBufSys(GA_buf *buf,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,GA_bltFx *fx);
void        NAPI BM_DrawRectBufSys(GA_buf *buf,N_int32 left,N_int32 top,N_int32 width,N_int32 height,GA_color color,N_int32 mix);

#ifdef  __cplusplus
}                                   /* End of "C" linkage for C++       */
#endif

#endif  /* __BUFINT_H */

