/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser functions
*               specifically used to work around bugs in prior binary
*               device drivers. We centralise all the bug fix code here
*               for clarity.
*
****************************************************************************/

#include "ref2dint.h"

/*----------------------------- Implementation ----------------------------*/

/* Special bug fix workarounds for ROP3's in older drivers. Due to lack
 * of proper tests in the conformance tests, some drivers do not properly
 * save and restore the hardware mix across ROP3 functions, so we force
 * the hardware mix to be restored below.
 */

/****************************************************************************
PARAMETERS:
srcLeft - Left coordinate of the source rectangle to copy
srcTop  - Top coordinate of the source rectangle to copy
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
dstLeft - Left coordinate of destination
dstTop  - Top coordinate of destination
rop3    - ROP3 raster operation code for the copy (GA_ROP3CodesType)
****************************************************************************/
void NAPI REF2D_BugFixBitBltPatt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    REF_draw2dHW.BitBltPatt(srcLeft,srcTop,width,height,dstLeft,dstTop,rop3);
    DC.hwState.SetMix(DC.mix);
}

/****************************************************************************
PARAMETERS:
srcOfs      - Offset of source bitmap in video memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3        - Mix code for the copy (GA_mixCodesType)
****************************************************************************/
void NAPI REF2D_BugFixBitBltPattLin(
    N_int32 srcOfs,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    REF_draw2dHW.BitBltPattLin(srcOfs,srcPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,rop3);
    DC.hwState.SetMix(DC.mix);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3        - ROP3 raster operation code for the copy (GA_ROP3CodesType)
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI REF2D_BugFixBitBltPattSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3,
    N_int32 flipY)
{
    REF_draw2dHW.BitBltPattSys(srcAddr,srcPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,rop3,flipY);
    DC.hwState.SetMix(DC.mix);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPhysAddr - Physical address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3        - Mix code for the copy (GA_mixCodesType)
****************************************************************************/
void NAPI REF2D_BugFixBitBltPattBM(
    void *srcAddr,
    N_int32 srcPhysAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    REF_draw2dHW.BitBltPattBM(srcAddr,srcPhysAddr,srcPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,rop3);
    DC.hwState.SetMix(DC.mix);
}

/****************************************************************************
PARAMETERS:
srcLeft - Left coordinate of the source rectangle to copy
srcTop  - Top coordinate of the source rectangle to copy
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
dstLeft - Left coordinate of destination
dstTop  - Top coordinate of destination
rop3    - ROP3 raster operation code for the copy (GA_ROP3CodesType)
****************************************************************************/
void NAPI REF2D_BugFixBitBltColorPatt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    REF_draw2dHW.BitBltColorPatt(srcLeft,srcTop,width,height,dstLeft,dstTop,rop3);
    DC.hwState.SetMix(DC.mix);
}

/****************************************************************************
PARAMETERS:
srcOfs      - Offset of source bitmap in video memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3        - Mix code for the copy (GA_mixCodesType)
****************************************************************************/
void NAPI REF2D_BugFixBitBltColorPattLin(
    N_int32 srcOfs,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    REF_draw2dHW.BitBltColorPattLin(srcOfs,srcPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,rop3);
    DC.hwState.SetMix(DC.mix);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3        - ROP3 raster operation code for the copy (GA_ROP3CodesType)
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI REF2D_BugFixBitBltColorPattSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3,
    N_int32 flipY)
{
    REF_draw2dHW.BitBltColorPattSys(srcAddr,srcPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,rop3,flipY);
    DC.hwState.SetMix(DC.mix);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPhysAddr - Physical address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3        - Mix code for the copy (GA_mixCodesType)
****************************************************************************/
void NAPI REF2D_BugFixBitBltColorPattBM(
    void *srcAddr,
    N_int32 srcPhysAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    REF_draw2dHW.BitBltColorPattBM(srcAddr,srcPhysAddr,srcPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,rop3);
    DC.hwState.SetMix(DC.mix);
}

/****************************************************************************
PARAMETERS:
left        - Left coordinate of the rectangle to draw
top         - Top coordinate of the rectangle to draw
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
color       - Color to fill the rectangle with
mix         - Mix to use for drawing the rectangle
****************************************************************************/
void NAPI REF2D_DrawRectExtHW(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    GA_color color,
    N_int32 mix)
{
    GA_color    oldColor = DC.foreColor;
    N_int32     oldMix = DC.mix;

    // Change the hardware colors if necessary
    if (color != oldColor)
        DC.hwState.SetForeColor(color);
    if (mix != oldMix)
        DC.hwState.SetMix(mix);

    // Draw the rectangle
    REF_draw2dHW.DrawRect(left,top,width,height);

    // Restore the current hardware draw buffer, color and mix
    DC.hwState.SetDrawBuffer(&DC.cntDrawBuf);
    if (color != oldColor)
        DC.hwState.SetForeColor(oldColor);
    if (mix != oldMix)
        DC.hwState.SetMix(oldMix);
}

