/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Header file for the windows compatible HalfToning tables
*               for creating a HalfTone palette and for performing the
*               HalfTone dithering algorithm in 8/15/16 bpp modes.
*
*               The halftone palette is set up to use entries from 20 to
*               226 in the physical palette, leaving the top 20 and bottom
*               20 for operating system specific use.
*
****************************************************************************/

#ifndef __DITHER_H
#define __DITHER_H

/*---------------------- Macro and type definitions -----------------------*/

/* Macros to halfTone dither an 8bpp RGB pixel */

#define DITHER_VARS8                                                    \
    uchar  *__dp

#define SETUP_DITHER8(y)                                                \
{                                                                       \
    __dp = &_REF2D_dither8x8[((y) & 7) << 3];                           \
}

#define PACK_COLOR_DITHER8_EX(dp,p,x,R,G,B)                             \
{                                                                       \
    uchar dither = (dp)[(x) & 7];                                       \
    (p) = _REF2D_ditherVec[                                             \
        _REF2D_div51[R] + (_REF2D_mod51[R] > dither) +                  \
        _REF2D_mul6[_REF2D_div51[G] + (_REF2D_mod51[G] > dither)] +     \
        _REF2D_mul36[_REF2D_div51[B] + (_REF2D_mod51[B] > dither)]];    \
}

#define PACK_COLOR_DITHER8(p,x,R,G,B)   PACK_COLOR_DITHER8_EX(__dp,p,x,R,G,B)

/* Macros to halfTone dither a 16bit 5/5/5 RGB pixel */

#define DITHER_VARS16                                                   \
    uchar  *__dp

#define SETUP_DITHER16(y)                                               \
{                                                                       \
    __dp = &_REF2D_dither4x4[((y) & 3) << 2];                           \
}

#define PACK_COLOR_DITHER15_EX(dp,p,x,R,G,B)                            \
{                                                                       \
    int dither = (dp)[(x) & 3];                                         \
    (p) = (ushort)                                                      \
       ((((R >> 3) + (_REF2D_mod8[R] > dither)) << 10) +                \
        (((G >> 3) + (_REF2D_mod8[G] > dither)) << 5) +                 \
        (((B >> 3) + (_REF2D_mod8[B] > dither)) << 0));                 \
}

#define PACK_COLOR_DITHER16_EX(dp,p,x,R,G,B)                            \
{                                                                       \
    int dither = (dp)[(x) & 3];                                         \
    (p) = (ushort)                                                      \
       ((((R >> 3) + (_REF2D_mod8[R] > dither)) << 11) +                \
        (((G >> 2) + (_REF2D_mod4[G] > dither)) << 5) +                 \
        (((B >> 3) + (_REF2D_mod8[B] > dither)) << 0));                 \
}

#define PACK_COLOR_DITHER15(p,x,R,G,B)  PACK_COLOR_DITHER15_EX(__dp,p,x,R,G,B)
#define PACK_COLOR_DITHER16(p,x,R,G,B)  PACK_COLOR_DITHER16_EX(__dp,p,x,R,G,B)

/*--------------------------- Global Variables ----------------------------*/

// TODO: 1bpp dithering support!

#if defined(SUPPORT_4BPP) || defined(SUPPORT_8BPP) || defined(SUPPORT_16BPP)
extern GA_palette   _REF2D_halftonePal[256];
extern GA_color *   _REF2D_ditherVec;
extern int          _REF2D_div51[256];
extern int          _REF2D_mod51[256];
extern int          _REF2D_mul6[6];
extern int          _REF2D_mul36[6];
extern int          _REF2D_dither8x8[64];
#endif
#ifdef  SUPPORT_16BPP
extern int          _REF2D_mod8[256];
extern int          _REF2D_mod4[256];
extern int          _REF2D_dither4x4[16];
#endif

#endif  /* __DITHER_H */

