/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Simple program to generate emulated ROP3 formulas to
*               emulate ROP3 codes using multiple passes using ROP2 codes.
*
****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "pmapi.h"
#include "tcl/cstack.hpp"
#include "snap/graphics.h"

const char *header_string = "\
#define FROM_SRC    0x010\n\
#define FROM_DEST   0x020\n\
#define FROM_PATT   0x040\n\
#define FROM_TEMP   0x080\n\
#define TO_DEST     0x100\n\
#define TO_TEMP     0x200\n\
#define R2_0        GA_R2_BLACK\n\
#define R2_1        GA_R2_WHITE\n\
\n\
#define DSon        GA_R2_NOTMERGESRC | FROM_SRC  | FROM_DEST\n\
#define DSna        GA_R2_MASKNOTSRC  | FROM_SRC  | FROM_DEST\n\
#define Sn          GA_R2_NOTCOPYSRC  | FROM_SRC\n\
#define SDna        GA_R2_MASKSRCNOT  | FROM_SRC  | FROM_DEST\n\
#define Dn          GA_R2_NOT         | FROM_DEST\n\
#define DSx         GA_R2_XORSRC      | FROM_SRC  | FROM_DEST\n\
#define DSan        GA_R2_NOTMASKSRC  | FROM_SRC  | FROM_DEST\n\
#define DSa         GA_R2_MASKSRC     | FROM_SRC  | FROM_DEST\n\
#define DSxn        GA_R2_NOTXORSRC   | FROM_SRC  | FROM_DEST\n\
#define D           GA_R2_NOP         | FROM_DEST\n\
#define DSno        GA_R2_MERGENOTSRC | FROM_SRC  | FROM_DEST\n\
#define S           GA_R2_COPYSRC     | FROM_SRC\n\
#define SDno        GA_R2_MERGESRCNOT | FROM_SRC  | FROM_DEST\n\
#define DSo         GA_R2_MERGESRC    | FROM_SRC  | FROM_DEST\n\
\n\
#define DPon        GA_R2_NOTMERGESRC | FROM_PATT | FROM_DEST\n\
#define DPna        GA_R2_MASKNOTSRC  | FROM_PATT | FROM_DEST\n\
#define Pn          GA_R2_NOTCOPYSRC  | FROM_PATT\n\
#define PDna        GA_R2_MASKSRCNOT  | FROM_PATT | FROM_DEST\n\
#define DPx         GA_R2_XORSRC      | FROM_PATT | FROM_DEST\n\
#define DPan        GA_R2_NOTMASKSRC  | FROM_PATT | FROM_DEST\n\
#define DPa         GA_R2_MASKSRC     | FROM_PATT | FROM_DEST\n\
#define DPxn        GA_R2_NOTXORSRC   | FROM_PATT | FROM_DEST\n\
#define DPno        GA_R2_MERGENOTSRC | FROM_PATT | FROM_DEST\n\
#define P           GA_R2_COPYSRC     | FROM_PATT\n\
#define PDno        GA_R2_MERGESRCNOT | FROM_PATT | FROM_DEST\n\
#define DPo         GA_R2_MERGESRC    | FROM_PATT | FROM_DEST\n\
\n\
#define TSon        GA_R2_NOTMERGESRC | FROM_TEMP | FROM_SRC\n\
#define TSna        GA_R2_MASKNOTSRC  | FROM_TEMP | FROM_SRC\n\
#define STna        GA_R2_MASKSRCNOT  | FROM_TEMP | FROM_SRC\n\
#define Tn          GA_R2_NOT         | FROM_TEMP\n\
#define TSx         GA_R2_XORSRC      | FROM_TEMP | FROM_SRC\n\
#define TSan        GA_R2_NOTMASKSRC  | FROM_TEMP | FROM_SRC\n\
#define TSa         GA_R2_MASKSRC     | FROM_TEMP | FROM_SRC\n\
#define TSxn        GA_R2_NOTXORSRC   | FROM_TEMP | FROM_SRC\n\
#define T           GA_R2_NOP         | FROM_TEMP\n\
#define TSno        GA_R2_MERGENOTSRC | FROM_TEMP | FROM_SRC\n\
#define STno        GA_R2_MERGESRCNOT | FROM_TEMP | FROM_SRC\n\
#define TSo         GA_R2_MERGESRC    | FROM_TEMP | FROM_SRC\n\
\n\
#define TPon        GA_R2_NOTMERGESRC | FROM_PATT | FROM_TEMP\n\
#define TPna        GA_R2_MASKNOTSRC  | FROM_PATT | FROM_TEMP\n\
#define PTna        GA_R2_MASKSRCNOT  | FROM_PATT | FROM_TEMP\n\
#define TPx         GA_R2_XORSRC      | FROM_PATT | FROM_TEMP\n\
#define TPan        GA_R2_NOTMASKSRC  | FROM_PATT | FROM_TEMP\n\
#define TPa         GA_R2_MASKSRC     | FROM_PATT | FROM_TEMP\n\
#define TPxn        GA_R2_NOTXORSRC   | FROM_PATT | FROM_TEMP\n\
#define TPno        GA_R2_MERGENOTSRC | FROM_PATT | FROM_TEMP\n\
#define PTno        GA_R2_MERGESRCNOT | FROM_PATT | FROM_TEMP\n\
#define TPo         GA_R2_MERGESRC    | FROM_PATT | FROM_TEMP\n\
\n\
#define DTon        GA_R2_NOTMERGESRC | FROM_TEMP | FROM_DEST\n\
#define DTna        GA_R2_MASKNOTSRC  | FROM_TEMP | FROM_DEST\n\
#define TDna        GA_R2_MASKSRCNOT  | FROM_TEMP | FROM_DEST\n\
#define DTx         GA_R2_XORSRC      | FROM_TEMP | FROM_DEST\n\
#define DTan        GA_R2_NOTMASKSRC  | FROM_TEMP | FROM_DEST\n\
#define DTa         GA_R2_MASKSRC     | FROM_TEMP | FROM_DEST\n\
#define DTxn        GA_R2_NOTXORSRC   | FROM_TEMP | FROM_DEST\n\
#define DTno        GA_R2_MERGENOTSRC | FROM_TEMP | FROM_DEST\n\
#define TDno        GA_R2_MERGESRCNOT | FROM_TEMP | FROM_DEST\n\
#define DTo         GA_R2_MERGESRC    | FROM_TEMP | FROM_DEST\n\
\n\
typedef struct {\n\
    ushort  pass1;\n\
    ushort  pass2;\n\
    ushort  pass3;\n\
    } ROP3_entry;\n\
\n";

#define AND_MASK    0x1
#define OR_MASK     0x2
#define XOR_MASK    0x4
#define NOT_MASK    0x8

typedef struct {
    char    s[100];
    } Expr;

typedef struct {
    char *match;
    char *s;
    } Lookup;

TCCStack<Expr>  stack(20);

static Lookup rop2[] = {
    // Dest and source rop's
    {"DSon","DSon"},    {"SDon","DSon"},
    {"DSna","DSna"},
    {"Sn",  "Sn"},
    {"SDna","SDna"},
    {"Dn",  "Dn"},
    {"DSx", "DSx"},     {"SDx", "DSx"},
    {"DSan","DSan"},    {"SDan","DSan"},
    {"DSa", "DSa"},     {"SDa", "DSa"},
    {"DSxn","DSxn"},    {"SDxn","DSxn"},
    {"D",   "D"},
    {"DSno","DSno"},
    {"S",   "S"},
    {"SDno","SDno"},
    {"DSo", "DSo"},     {"SDo", "DSo"},
    // Dest and pattern rop's
    {"DPon","DPon"},    {"PDon","DPon"},
    {"DPna","DPna"},
    {"Pn",  "Pn"},
    {"PDna","PDna"},
    {"Dn",  "Dn"},
    {"DPx", "DPx"},     {"PDx", "DPx"},
    {"DPan","DPan"},    {"PDan","DPan"},
    {"DPa", "DPa"},     {"PDa", "DPa"},
    {"DPxn","DPxn"},    {"PDxn","DPxn"},
    {"D",   "D"},
    {"DPno","DPno"},
    {"P",   "P"},
    {"PDno","PDno"},
    {"DPo", "DPo"},     {"PDo", "DPo"},
    // Dest and temporary rop's
    {"DTon","DTon"},    {"TDon","DTon"},
    {"DTna","DTna"},
    {"Tn",  "Tn"},
    {"TDna","TDna"},
    {"Dn",  "Dn"},
    {"DTx", "DTx"},     {"TDx", "DTx"},
    {"DTan","DTan"},    {"TDan","DTan"},
    {"DTa", "DTa"},     {"TDa", "DTa"},
    {"DTxn","DTxn"},    {"TDxn","DTxn"},
    {"D",   "D"},
    {"DTno","DTno"},
    {"T",   "T"},
    {"TDno","TDno"},
    {"DTo", "DTo"},     {"TDo", "DTo"},
    // Temporary and source rop's
    {"TSon","TSon"},    {"STon","TSon"},
    {"TSna","TSna"},
    {"Sn",  "Sn"},
    {"STna","STna"},
    {"Tn",  "Tn"},
    {"TSx", "TSx"},     {"STx", "TSx"},
    {"TSan","TSan"},    {"STan","TSan"},
    {"TSa", "TSa"},     {"STa", "TSa"},
    {"TSxn","TSxn"},    {"STxn","TSxn"},
    {"T",   "T"},
    {"TSno","TSno"},
    {"S",   "S"},
    {"STno","STno"},
    {"TSo", "TSo"},     {"STo", "TSo"},
    // Temporary and pattern rop's
    {"TPon","TPon"},    {"PTon","TPon"},
    {"TPna","TPna"},
    {"Pn",  "Pn"},
    {"PTna","PTna"},
    {"Tn",  "Tn"},
    {"TPx", "TPx"},     {"PTx", "TPx"},
    {"TPan","TPan"},    {"PTan","TPan"},
    {"TPa", "TPa"},     {"PTa", "TPa"},
    {"TPxn","TPxn"},    {"PTxn","TPxn"},
    {"T",   "T"},
    {"TPno","TPno"},
    {"P",   "P"},
    {"PTno","PTno"},
    {"TPo", "TPo"},     {"PTo", "TPo"},
    {NULL,NULL},
    };

char * findROP2(
    char *expr)
{
    int i;
    for (i = 0; rop2[i].match != NULL; i++) {
        if (strcmp(rop2[i].match,expr) == 0)
            return rop2[i].s;
        }
    return NULL;
}

char *separator;

ibool dumpExpr(
    char *result,
    Expr& expr,
    char *op1,
    char *op2,
    char operand,
    int& haveTemp,
    int numDest)
{
    char    *r2,buf[100],rop[10],*p;

    if (op1[1] != 0)
        return false;
    if (!findROP2(expr.s)) {
        if (op2[1] == 0 || (op2[1] == 'n' && op2[2] == 0)
                || (r2 = findROP2(op2)) == NULL) {
            sprintf(buf,"%sTO_TEMP | %s",separator,op2);
            strcat(result,buf);
            sprintf(expr.s,"%sT%c", op1,operand);
            haveTemp++;
            }
        else if (strchr(r2,'T') != NULL) {
            sprintf(buf,"%sTO_TEMP | %s",separator,r2);
            strcat(result,buf);
            sprintf(expr.s,"%sT%c", op1,operand);
            haveTemp++;
            }
        else if (numDest > 1) {
            strcpy(rop,r2);
            r2 = rop;
            p = r2;
            while (*p) {
                if (*p == 'D') {
                    *p = 'T';
                    break;
                    }
                p++;
                }
            sprintf(buf,"%sTO_TEMP | D, TO_TEMP | %s",separator,r2);
            strcat(result,buf);
            sprintf(expr.s,"%sT%c", op1,operand);
            }
        else {
            sprintf(buf,"%sTO_DEST | %s",separator,r2);
            strcat(result,buf);
            sprintf(expr.s,"%sD%c", op1,operand);
            }
        separator = ", ";
        }
    stack.push(expr);
    return true;
}

//#define   TEST_SINGLE "DPSona"
//#define   TEST_SINGLE "SDPSoox"
//#define   TEST_SINGLE "PSDPSanaxx"
//#define   TEST_SINGLE "PSon"
//#define   TEST_SINGLE "PDSox"

//#define   SOFTWARE_ONLY
//#define   TEMPORARY_ONLY
//#define DEST_ONLY

int main(void)
{
#ifndef TEST_SINGLE
    FILE    *f;
#endif
    char    *p,*p1,*p2,*r2;
    char    *operands = "DSP";
    Expr    expr,op1,op2;
    ibool   allXOR,allAND,allOR;
    char    rop3[100],buf[100],result[100];
    int     i = 0,mask,numOperands,numDest;
    int     haveTemp;

#if !defined(TEST_SINGLE) && !defined(SOFTWARE_ONLY)
    printf("%s", header_string);
#endif
    printf("ROP3_entry rop3_table[256] = {\n");
#ifdef  TEST_SINGLE
    strcpy(rop3,TEST_SINGLE);
#else
    if ((f = fopen("r3.lst","r")) == NULL)
        PM_fatalError("Unable to open r3.lst!");
    while (!feof(f)) {
        if (!fgets(rop3,sizeof(rop3),f))
            break;
        if (rop3[strlen(rop3)-1] == '\n')
            rop3[strlen(rop3)-1] = 0;
        if (rop3[0] == '0') {
#if !defined(SOFTWARE_ONLY) && !defined(TEMPORARY_ONLY) && !defined(DEST_ONLY)
            printf("    /* %02X: 0           */ {TO_DEST | R2_0},\n", i);
#endif
            }
        else if (rop3[0] == '1') {
#if !defined(SOFTWARE_ONLY) && !defined(TEMPORARY_ONLY) && !defined(DEST_ONLY)
            printf("    /* %02X: 1           */ {TO_DEST | R2_1},\n", i);
#endif
            }
        else {
#endif
            separator = "";
            numOperands = haveTemp = numDest = allXOR = allAND = allOR = 0;
            result[0] = 0;
            mask = 0;
            p = rop3;
            while (*p && *p != '\n') {
                if (*p == 'D')
                    numDest++;
                p++;
                }
            p = rop3;
            stack.empty();
            while (*p && *p != '\n') {
                if (strchr(operands,*p) == NULL) {
                    // Handle operator code
                    switch (*p) {
                        case 'a':
                            op2 = stack.pop();
                            op1 = stack.pop();
                            sprintf(expr.s,"%s%sa", op1.s, op2.s);
                            if (!dumpExpr(result,expr,op1.s,op2.s,'a',haveTemp,numDest))
                                goto TooComplicated;
                            mask |= AND_MASK;
                            allAND = false;
                            allOR = false;
                            allXOR = false;
                            if (mask == AND_MASK)
                                allAND = true;
                            break;
                        case 'o':
                            op2 = stack.pop();
                            op1 = stack.pop();
                            sprintf(expr.s,"%s%so", op1.s, op2.s);
                            if (!dumpExpr(result,expr,op1.s,op2.s,'o',haveTemp,numDest))
                                goto TooComplicated;
                            mask |= OR_MASK;
                            allAND = false;
                            allOR = false;
                            allXOR = false;
                            if (mask == OR_MASK)
                                allOR = true;
                            break;
                        case 'x':
                            op2 = stack.pop();
                            op1 = stack.pop();
                            sprintf(expr.s,"%s%sx", op1.s, op2.s);
                            if (!dumpExpr(result,expr,op1.s,op2.s,'x',haveTemp,numDest))
                                goto TooComplicated;
                            mask |= XOR_MASK;
                            allAND = false;
                            allOR = false;
                            allXOR = false;
                            if (mask  == XOR_MASK)
                                allXOR = true;
                            break;
                        case 'n':
                            op1 = stack.pop();
                            sprintf(expr.s,"%sn", op1.s);
                            stack.push(expr);
                            mask |= NOT_MASK;
                            break;
                        default:
                            printf("\nSyntax error parsing ROP3 code!\n");
                            return -1;
                        }
                    }
                else {
                    // Push operand on the stack
                    expr.s[0] = *p;
                    expr.s[1] = 0;
                    stack.push(expr);
                    numOperands++;
                    }
                p++;
                }
            expr = stack.pop();
            if ((r2 = findROP2(expr.s)) == NULL) {
TooComplicated:
#if defined(TEMPORARY_ONLY) || defined(DEST_ONLY) || defined(ROP2_ONLY)
                printf("");
#else
                printf("    /* %02X: %-11s */ {0},\n", i,rop3);
#endif
                }
            else {
#ifndef SOFTWARE_ONLY
                if (numOperands > 1 && !(haveTemp && (allAND || allOR || allXOR)) && strchr(r2,'D') == NULL) {
                    sprintf(buf,"%sTO_TEMP | %s, TO_DEST | T",separator,r2);
                    strcat(result,buf);
                    }
                else {
                    sprintf(buf,"%sTO_DEST | %s",separator,r2);
                    strcat(result,buf);
                    }
                if (haveTemp && (allAND || allOR || allXOR)) {
                    p = result + 16;
                    if (*p == 'T') {
                        p += 8;
                        p1 = p;
                        while (*p != ',')
                            p++;
                        *p = 0;
                        p2 = p + 14;
#ifndef TEMPORARY_ONLY
                        printf("    /* %02X: %-11s */ {TO_DEST | D%s, TO_DEST | DS%s},\n", i,rop3,p1,p2);
#endif
                        }
                    else {
#ifndef TEMPORARY_ONLY
                        printf("    /* %02X: %-11s */ {TO_DEST | S, TO_DEST | DP%s},\n", i,rop3,p+9);
#endif
                        }
                    }
                else if (!haveTemp) {
#ifndef TEMPORARY_ONLY
                    printf("    /* %02X: %-11s */ {%s},\n", i,rop3,result);
#endif
                    }
                else {
#ifndef DEST_ONLY
                    printf("    /* %02X: %-11s */ {%s},\n", i,rop3,result);
#endif
                    }
#endif
                }
#ifndef TEST_SINGLE
            }
        i++;
        }
#endif
    printf("    };\n");
    (void)p1;
    (void)p2;
    (void)buf;
    return 0;
}
