/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the flipped display support
*               routines in the SNAP Reference Rasteriser.
*
****************************************************************************/

#include "flipint.h"

// TODO: Clipped lines!

// TODO: BitBltMasked!

/*--------------------------- Global variables ----------------------------*/

FLIPPED_state   FLIPPED_devCtx;

REF2D_driver FLIPPED_drv = {
    sizeof(FLIPPED_drv),
    FLIPPED_QueryFunctions,
    FLIPPED_SetDrawBufferSW,
    FLIPPED_SetDrawSurface,
    FLIPPED_RotateBitmap,
    FLIPPED_UnloadDriver,
    FLIPPED_InitFilter,
    FLIPPED_DeInitFilter,
    FLIPPED_GlobalOptionsUpdated,
    };

static GA_initFuncs FLIPPED_initPT = {
    sizeof(GA_initFuncs),
    NULL,                       // FLIPPED_GetConfigInfo,
    NULL,                       // Reserved
    NULL,                       // FLIPPED_SetModeProfile,
    NULL,                       // FLIPPED_GetOptions,
    FLIPPED_SetOptions,
    NULL,                       // FLIPPED_GetVideoModeInfo,
    NULL,                       // FLIPPED_SetVideoMode,
    NULL,                       // FLIPPED_GetVideoMode,
    NULL,                       // FLIPPED_GetCustomVideoModeInfo,
    NULL,                       // FLIPPED_SetCustomVideoMode,
    NULL,                       // FLIPPED_GetClosestPixelClock,
    NULL,                       // FLIPPED_GetCRTCTimings,
    NULL,                       // FLIPPED_SetCRTCTimings,
    NULL,                       // FLIPPED_SetGlobalRefresh,
    NULL,                       // FLIPPED_SaveRestoreState,
    NULL,                       // FLIPPED_SetDisplayOutput,
    NULL,                       // FLIPPED_GetDisplayOutput,
    NULL,                       // FLIPPED_SetSoftwareRenderFuncs,
    NULL,                       // FLIPPED_GetUniqueFilename,
    NULL,                       // Reserved
    NULL,                       // FLIPPED_GetMonitorInfo,
    NULL,                       // FLIPPED_SetMonitorInfo,
    NULL,                       // FLIPPED_GetCurrentVideoModeInfo,
    NULL,                       // FLIPPED_GetCertifyInfo,
    NULL,                       // FLIPPED_SetCRTCTimings,
    NULL,                       // FLIPPED_AlignLinearBuffer,
    NULL,                       // FLIPPED_GetIORegisterTraps,
    NULL,                       // FLIPPED_GetCurrentRefreshRate,
    NULL,                       // FLIPPED_PollForDisplaySwitch,
    NULL,                       // FLIPPED_PerformDisplaySwitch,
    NULL,                       // FLIPPED_GetVideoModeInfoExt,
    NULL,                       // FLIPPED_GetCustomVideoModeInfoExt,
    NULL,                       // FLIPPED_SwitchPhysicalResolution
    NULL,                       // FLIPPED_GetNumberOfHeads
    NULL,                       // FLIPPED_SetActiveHead
    NULL,                       // FLIPPED_GetActiveHead
    NULL,                       // FLIPPED_SetRef2dPointer
    };

static GA_cursorFuncs FLIPPED_cursorPT = {
    sizeof(GA_cursorFuncs),
    FLIPPED_SetMonoCursor,
    NULL,                       // FLIPPED_SetMonoCursorColor,
    FLIPPED_SetColorCursor,
    FLIPPED_SetCursorPos,
    NULL,                       // FLIPPED_ShowCursor,
    FLIPPED_BeginAccess,
    NULL,                       // FLIPPED_EndAccess,
    FLIPPED_SetColorCursor256,
    FLIPPED_SetColorCursorRGBA,
    NULL,                       // FLIPPED_IsHardwareCursor
    FLIPPED_SetColorCursorRGB,
    };

static GA_2DStateFuncs FLIPPED_state2dPT = {
    sizeof(GA_2DStateFuncs),
    FLIPPED_SetDrawBuffer,
    NULL,                       // REF2D_IsIdle,
    NULL,                       // REF2D_WaitTillIdle,
    NULL,                       // REF2D_EnableDirectAccess
    NULL,                       // REF2D_DisableDirectAccess
    NULL,                       // REF2D_SetMix,
    NULL,                       // REF2D_SetForeColor,
    NULL,                       // REF2D_SetBackColor,
    FLIPPED_Set8x8MonoPattern,
    NULL,                       // REF2D_Use8x8MonoPattern,
    NULL,                       // REF2D_Use8x8TransMonoPattern,
    FLIPPED_Set8x8ColorPattern,
    NULL,                       // REF2D_Use8x8ColorPattern,
    NULL,                       // REF2D_Use8x8TransColorPattern,
    NULL,                       // REF2D_SetLineStipple,
    NULL,                       // REF2D_SetLineStippleCount,
    NULL,                       // REF2D_SetPlaneMask,
    NULL,                       // REF2D_SetSrcBlendFunc,
    NULL,                       // REF2D_SetDstBlendFunc,
    NULL,                       // REF2D_SetAlphaValue,
    NULL,                       // DRIVER_SetupForWindows
    NULL,                       // DRIVER_ResetWindows
    NULL,                       // REF2D_SetLineStyle,
    NULL,                       // REF2D_BuildTranslateVector,
    };

static GA_2DRenderFuncs FLIPPED_draw2dPT = {
    sizeof(GA_2DRenderFuncs),
    FLIPPED_GetPixel,
    FLIPPED_PutPixel,
    FLIPPED_DrawScanList,
    FLIPPED_DrawPattScanList,
    FLIPPED_DrawColorPattScanList,
    FLIPPED_DrawEllipseList,
    FLIPPED_DrawPattEllipseList,
    FLIPPED_DrawColorPattEllipseList,
    FLIPPED_DrawFatEllipseList,
    FLIPPED_DrawPattFatEllipseList,
    FLIPPED_DrawColorPattFatEllipseList,
    FLIPPED_DrawRect,
    NULL,
    FLIPPED_DrawPattRect,
    FLIPPED_DrawColorPattRect,
    FLIPPED_DrawTrap,
    FLIPPED_DrawPattTrap,
    FLIPPED_DrawColorPattTrap,
    FLIPPED_DrawLineInt,
    FLIPPED_DrawBresenhamLine,
    FLIPPED_DrawStippleLineInt,
    FLIPPED_DrawBresenhamStippleLine,
    FLIPPED_DrawEllipse,
    FLIPPED_ClipEllipse,
    FLIPPED_PutMonoImageMSBSys,
    NULL,                       // FLIPPED_PutMonoImageMSBLin,
    NULL,                       // FLIPPED_PutMonoImageMSBBM,
    FLIPPED_PutMonoImageLSBSys,
    NULL,                       // FLIPPED_PutMonoImageLSBLin,
    NULL,                       // FLIPPED_PutMonoImageLSBBM,
    FLIPPED_ClipMonoImageMSBSys,
    NULL,                       // FLIPPED_ClipMonoImageMSBLin,
    NULL,                       // FLIPPED_ClipMonoImageMSBBM,
    FLIPPED_ClipMonoImageLSBSys,
    NULL,                       // FLIPPED_ClipMonoImageLSBLin,
    NULL,                       // FLIPPED_ClipMonoImageLSBBM,
    FLIPPED_BitBlt,
    NULL,                       // FLIPPED_BitBltLin,
    FLIPPED_BitBltSys,
    NULL,                       // FLIPPED_BitBltBM,
    FLIPPED_BitBltPatt,
    NULL,                       // FLIPPED_BitBltPattLin,
    FLIPPED_BitBltPattSys,
    NULL,                       // FLIPPED_BitBltPattBM,
    FLIPPED_BitBltColorPatt,
    NULL,                       // FLIPPED_BitBltColorPattLin,
    FLIPPED_BitBltColorPattSys,
    NULL,                       // FLIPPED_BitBltColorPattBM,
    FLIPPED_SrcTransBlt,
    NULL,                       // FLIPPED_SrcTransBltLin,
    FLIPPED_SrcTransBltSys,
    NULL,                       // FLIPPED_SrcTransBltBM,
    FLIPPED_DstTransBlt,
    NULL,                       // FLIPPED_DstTransBltLin,
    FLIPPED_DstTransBltSys,
    NULL,                       // FLIPPED_DstTransBltBM,
    FLIPPED_StretchBlt,
    NULL,                       // FLIPPED_StretchBltLin,
    FLIPPED_StretchBltSys,
    NULL,                       // FLIPPED_StretchBltBM,
    FLIPPED_ConvertBltSys_Obsolete,
    NULL,                       // FLIPPED_ConvertBltBM_Obsolete,
    FLIPPED_StretchConvertBltSys_Obsolete,
    NULL,                       // FLIPPED_StretchConvertBltBM_Obsolete,
    NULL,                       // FLIPPED_BitBltFxTest,
    NULL,                       // FLIPPED_BitBltFx,
    NULL,                       // FLIPPED_BitBltFxLin,
    NULL,                       // FLIPPED_BitBltFxSys,
    NULL,                       // FLIPPED_BitBltFxBM,
    FLIPPED_GetBitmapSys,
    NULL,                       // FLIPPED_GetBitmapBM,
    NULL,                       // FLIPPED_UpdateScreen,
    NULL,                       // FLIPPED_DrawClippedLineInt,
    NULL,                       // FLIPPED_DrawClippedBresenhamLine,
    NULL,                       // FLIPPED_DrawClippedStippleLineInt,
    NULL,                       // FLIPPED_DrawClippedBresenhamStippleLine,
    NULL,                       // FLIPPED_BitBltPlaneMasked,
    NULL,                       // FLIPPED_BitBltPlaneMaskedLin,
    NULL,                       // FLIPPED_BitBltPlaneMaskedSys,
    NULL,                       // FLIPPED_BitBltPlaneMaskedBM,
    NULL,                       // FLIPPED_DrawRectLin,
    NULL,                       // FLIPPED_DrawRectExt,
    FLIPPED_DrawStyleLineInt,
    FLIPPED_DrawBresenhamStyleLine,
    NULL,                       // FLIPPED_DrawClippedStyleLineInt,
    NULL,                       // FLIPPED_DrawClippedBresenhamStyleLine,
    };

static GA_2DRenderFuncs FLIPPED_draw2dStub = {
    sizeof(GA_2DRenderFuncs),
    FSTUB_GetPixel,
    FSTUB_PutPixel,
    FSTUB_DrawScanList,
    FSTUB_DrawPattScanList,
    FSTUB_DrawColorPattScanList,
    FSTUB_DrawEllipseList,
    FSTUB_DrawPattEllipseList,
    FSTUB_DrawColorPattEllipseList,
    FSTUB_DrawFatEllipseList,
    FSTUB_DrawPattFatEllipseList,
    FSTUB_DrawColorPattFatEllipseList,
    FSTUB_DrawRect,
    NULL,
    FSTUB_DrawPattRect,
    FSTUB_DrawColorPattRect,
    FSTUB_DrawTrap,
    FSTUB_DrawPattTrap,
    FSTUB_DrawColorPattTrap,
    FSTUB_DrawLineInt,
    FSTUB_DrawBresenhamLine,
    FSTUB_DrawStippleLineInt,
    FSTUB_DrawBresenhamStippleLine,
    FSTUB_DrawEllipse,
    FSTUB_ClipEllipse,
    FSTUB_PutMonoImageMSBSys,
    NULL,                       // FSTUB_PutMonoImageMSBLin,
    NULL,                       // FSTUB_PutMonoImageMSBBM,
    FSTUB_PutMonoImageLSBSys,
    NULL,                       // FSTUB_PutMonoImageLSBLin,
    NULL,                       // FSTUB_PutMonoImageLSBBM,
    FSTUB_ClipMonoImageMSBSys,
    NULL,                       // FSTUB_ClipMonoImageMSBLin,
    NULL,                       // FSTUB_ClipMonoImageMSBBM,
    FSTUB_ClipMonoImageLSBSys,
    NULL,                       // FSTUB_ClipMonoImageLSBLin,
    NULL,                       // FSTUB_ClipMonoImageLSBBM,
    FSTUB_BitBlt,
    NULL,                       // FSTUB_BitBltLin,
    FSTUB_BitBltSys,
    NULL,                       // FSTUB_BitBltBM,
    FSTUB_BitBltPatt,
    NULL,                       // FSTUB_BitBltPattLin,
    FSTUB_BitBltPattSys,
    NULL,                       // FSTUB_BitBltPattBM,
    FSTUB_BitBltColorPatt,
    NULL,                       // FSTUB_BitBltColorPattLin,
    FSTUB_BitBltColorPattSys,
    NULL,                       // FSTUB_BitBltColorPattBM,
    FSTUB_SrcTransBlt,
    NULL,                       // FSTUB_SrcTransBltLin,
    FSTUB_SrcTransBltSys,
    NULL,                       // FSTUB_SrcTransBltBM,
    FSTUB_DstTransBlt,
    NULL,                       // FSTUB_DstTransBltLin,
    FSTUB_DstTransBltSys,
    NULL,                       // FSTUB_DstTransBltBM,
    FSTUB_StretchBlt,
    NULL,                       // FSTUB_StretchBltLin,
    FSTUB_StretchBltSys,
    NULL,                       // FSTUB_StretchBltBM,
    FSTUB_ConvertBltSys_Obsolete,
    NULL,                       // FSTUB_ConvertBltBM_Obsolete,
    FSTUB_StretchConvertBltSys_Obsolete,
    NULL,                       // FSTUB_StretchConvertBltBM_Obsolete,
    NULL,                       // FSTUB_BitBltFxTest,
    NULL,                       // FSTUB_BitBltFx,
    NULL,                       // FSTUB_BitBltFxLin,
    NULL,                       // FSTUB_BitBltFxSys,
    NULL,                       // FSTUB_BitBltFxBM,
    FSTUB_GetBitmapSys,
    NULL,                       // FSTUB_GetBitmapBM,
    NULL,                       // FSTUB_UpdateScreen,
    NULL,                       // FSTUB_DrawClippedLineInt,
    NULL,                       // FSTUB_DrawClippedBresenhamLine,
    NULL,                       // FSTUB_DrawClippedStippleLineInt,
    NULL,                       // FSTUB_DrawClippedBresenhamStippleLine,
    NULL,                       // FSTUB_BitBltPlaneMasked,
    NULL,                       // FSTUB_BitBltPlaneMaskedLin,
    NULL,                       // FSTUB_BitBltPlaneMaskedSys,
    NULL,                       // FSTUB_BitBltPlaneMaskedBM,
    NULL,                       // FSTUB_DrawRectLin,
    NULL,                       // FSTUB_DrawRectExt,
    FSTUB_DrawStyleLineInt,
    FSTUB_DrawBresenhamStyleLine,
    NULL,                       // FSTUB_DrawClippedStyleLineInt,
    NULL,                       // FSTUB_DrawClippedBresenhamStyleLine,
    };

GA_initFuncs        FLIPPED_initREF;
GA_cursorFuncs      FLIPPED_cursorREF;
GA_2DStateFuncs     FLIPPED_state2dREF;
GA_2DRenderFuncs    FLIPPED_draw2dREF;
GA_2DRenderFuncs    _VARAPI FLIPPED_draw2dCnt;

#ifdef  __DRIVER__
PM_imports _VARAPI _PM_imports;
#endif

/* Declare a table for performing mono bitmap 180 degree rotations */

static uchar mirrorTable1BPP[] = {
    0x00,0x80,0x40,0xC0,0x20,0xA0,0x60,0xE0,
    0x10,0x90,0x50,0xD0,0x30,0xB0,0x70,0xF0,
    0x08,0x88,0x48,0xC8,0x28,0xA8,0x68,0xE8,
    0x18,0x98,0x58,0xD8,0x38,0xB8,0x78,0xF8,
    0x04,0x84,0x44,0xC4,0x24,0xA4,0x64,0xE4,
    0x14,0x94,0x54,0xD4,0x34,0xB4,0x74,0xF4,
    0x0C,0x8C,0x4C,0xCC,0x2C,0xAC,0x6C,0xEC,
    0x1C,0x9C,0x5C,0xDC,0x3C,0xBC,0x7C,0xFC,
    0x02,0x82,0x42,0xC2,0x22,0xA2,0x62,0xE2,
    0x12,0x92,0x52,0xD2,0x32,0xB2,0x72,0xF2,
    0x0A,0x8A,0x4A,0xCA,0x2A,0xAA,0x6A,0xEA,
    0x1A,0x9A,0x5A,0xDA,0x3A,0xBA,0x7A,0xFA,
    0x06,0x86,0x46,0xC6,0x26,0xA6,0x66,0xE6,
    0x16,0x96,0x56,0xD6,0x36,0xB6,0x76,0xF6,
    0x0E,0x8E,0x4E,0xCE,0x2E,0xAE,0x6E,0xEE,
    0x1E,0x9E,0x5E,0xDE,0x3E,0xBE,0x7E,0xFE,
    0x01,0x81,0x41,0xC1,0x21,0xA1,0x61,0xE1,
    0x11,0x91,0x51,0xD1,0x31,0xB1,0x71,0xF1,
    0x09,0x89,0x49,0xC9,0x29,0xA9,0x69,0xE9,
    0x19,0x99,0x59,0xD9,0x39,0xB9,0x79,0xF9,
    0x05,0x85,0x45,0xC5,0x25,0xA5,0x65,0xE5,
    0x15,0x95,0x55,0xD5,0x35,0xB5,0x75,0xF5,
    0x0D,0x8D,0x4D,0xCD,0x2D,0xAD,0x6D,0xED,
    0x1D,0x9D,0x5D,0xDD,0x3D,0xBD,0x7D,0xFD,
    0x03,0x83,0x43,0xC3,0x23,0xA3,0x63,0xE3,
    0x13,0x93,0x53,0xD3,0x33,0xB3,0x73,0xF3,
    0x0B,0x8B,0x4B,0xCB,0x2B,0xAB,0x6B,0xEB,
    0x1B,0x9B,0x5B,0xDB,0x3B,0xBB,0x7B,0xFB,
    0x07,0x87,0x47,0xC7,0x27,0xA7,0x67,0xE7,
    0x17,0x97,0x57,0xD7,0x37,0xB7,0x77,0xF7,
    0x0F,0x8F,0x4F,0xCF,0x2F,0xAF,0x6F,0xEF,
    0x1F,0x9F,0x5F,0xDF,0x3F,0xBF,0x7F,0xFF,
    };

// TODO: This table is the same as above, and needs to be generated!!

#if 0
static uchar mirrorTable4BPP[] = {
    0x00,0x80,0x40,0xC0,0x20,0xA0,0x60,0xE0,
    0x10,0x90,0x50,0xD0,0x30,0xB0,0x70,0xF0,
    0x08,0x88,0x48,0xC8,0x28,0xA8,0x68,0xE8,
    0x18,0x98,0x58,0xD8,0x38,0xB8,0x78,0xF8,
    0x04,0x84,0x44,0xC4,0x24,0xA4,0x64,0xE4,
    0x14,0x94,0x54,0xD4,0x34,0xB4,0x74,0xF4,
    0x0C,0x8C,0x4C,0xCC,0x2C,0xAC,0x6C,0xEC,
    0x1C,0x9C,0x5C,0xDC,0x3C,0xBC,0x7C,0xFC,
    0x02,0x82,0x42,0xC2,0x22,0xA2,0x62,0xE2,
    0x12,0x92,0x52,0xD2,0x32,0xB2,0x72,0xF2,
    0x0A,0x8A,0x4A,0xCA,0x2A,0xAA,0x6A,0xEA,
    0x1A,0x9A,0x5A,0xDA,0x3A,0xBA,0x7A,0xFA,
    0x06,0x86,0x46,0xC6,0x26,0xA6,0x66,0xE6,
    0x16,0x96,0x56,0xD6,0x36,0xB6,0x76,0xF6,
    0x0E,0x8E,0x4E,0xCE,0x2E,0xAE,0x6E,0xEE,
    0x1E,0x9E,0x5E,0xDE,0x3E,0xBE,0x7E,0xFE,
    0x01,0x81,0x41,0xC1,0x21,0xA1,0x61,0xE1,
    0x11,0x91,0x51,0xD1,0x31,0xB1,0x71,0xF1,
    0x09,0x89,0x49,0xC9,0x29,0xA9,0x69,0xE9,
    0x19,0x99,0x59,0xD9,0x39,0xB9,0x79,0xF9,
    0x05,0x85,0x45,0xC5,0x25,0xA5,0x65,0xE5,
    0x15,0x95,0x55,0xD5,0x35,0xB5,0x75,0xF5,
    0x0D,0x8D,0x4D,0xCD,0x2D,0xAD,0x6D,0xED,
    0x1D,0x9D,0x5D,0xDD,0x3D,0xBD,0x7D,0xFD,
    0x03,0x83,0x43,0xC3,0x23,0xA3,0x63,0xE3,
    0x13,0x93,0x53,0xD3,0x33,0xB3,0x73,0xF3,
    0x0B,0x8B,0x4B,0xCB,0x2B,0xAB,0x6B,0xEB,
    0x1B,0x9B,0x5B,0xDB,0x3B,0xBB,0x7B,0xFB,
    0x07,0x87,0x47,0xC7,0x27,0xA7,0x67,0xE7,
    0x17,0x97,0x57,0xD7,0x37,0xB7,0x77,0xF7,
    0x0F,0x8F,0x4F,0xCF,0x2F,0xAF,0x6F,0xEF,
    0x1F,0x9F,0x5F,0xDF,0x3F,0xBF,0x7F,0xFF,
    };
#endif

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
REMARKS:
Function to initialise the flipped display filter driver.
****************************************************************************/
static ibool FLIPPED_init(
    GA_devCtx *hwCtx,
    GA_globalOptions *gOpt)
{
    /* Initialise the internal state of the filter driver */
    DC.hwCtx = hwCtx;
    DC.gOpt = gOpt;
    DC.enabled = false;
    FLIPPED_initREF.dwSize = sizeof(FLIPPED_initREF);
    if (!hwCtx->loader.QueryFunctions(GA_GET_INITFUNCS,-2,&FLIPPED_initREF))
        return false;

    // Hook the device context QueryFunctions function so we can return
    // our hooked functions.
    DC.QueryFunctionsDC = hwCtx->loader.QueryFunctions;
    hwCtx->loader.QueryFunctions = FLIPPED_QueryFunctionsDC;

    // Obtain a copy of the options structure so we can determine if we
    // are enabled or not.
    DC.opt.dwSize = sizeof(DC.opt);
    FLIPPED_initREF.GetOptions(&DC.opt);

    // Initialise all the filter driver vectors
    INIT_INIT_VECTORS;
    return true;
}

#ifdef  __DRIVER__
/****************************************************************************
REMARKS:
Fatal error handler called when a non-imported function is called by the
driver. We leave this to a runtime error so that older applications and
shell drivers will work with newer bpd drivers provided no newer functions
are required by the driver itself. If they are, the application or shell
driver needs to be recompiled.
****************************************************************************/
static void _PM_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported PM_imports import function called! Please re-compile!\n");
}

/****************************************************************************
REMARKS:
Function entry point for the reference rasteriser library.
****************************************************************************/
REF2D_driver * _CEXPORT FLIPPED_initLibrary(
    GA_devCtx *hwCtx,
    GA_globalOptions *gOpt,
    PM_imports *pmImp)
{
    int     i,max;
    ulong   *p;

    // Initialize all default imports to point to fatal error handler
    // for upwards compatibility.
    max = sizeof(_PM_imports)/sizeof(GA_initLibrary_t);
    for (i = 0,p = (ulong*)&_PM_imports; i < max; i++)
        *p++ = (ulong)_PM_fatalErrorHandler;
    memcpy(&_PM_imports,pmImp,MIN(sizeof(_PM_imports),pmImp->dwSize));
    if (!FLIPPED_init(hwCtx,gOpt))
        return NULL;
    return &DRV;
}
#else
/****************************************************************************
REMARKS:
Function entry point to fake loading of a driver when linked into app.
****************************************************************************/
ibool NAPI FLIPPED_loadDriver(
    GA_devCtx *hwCtx,
    GA_globalOptions *gOpt,
    REF2D_driver **drv,
    PE_MODULE **hModule,
    ulong *size)
{
    if (!FLIPPED_init(hwCtx,gOpt))
        return false;
    *drv = &DRV;
    *hModule = NULL;
    *size = 0;
    return true;
}
#endif

/****************************************************************************
REMARKS:
Enabled or disabled the invert driver internally.
****************************************************************************/
static void FlippedEnable(void)
{
    if (DC.enabled)
        FLIPPED_draw2dCnt = FLIPPED_draw2dPT;
    else
        FLIPPED_draw2dCnt = FLIPPED_draw2dREF;
}

/****************************************************************************
REMARKS:
Function to initialise the INVERT display filter driver. This function
gets called long after the driver has actually been loaded, at the
time when the reference rasteriser module is loaded and initialsied (ie:
at mode init time).
****************************************************************************/
ibool NAPI FLIPPED_InitFilter(
    GA_modeInfo *modeInfo,
    N_int32 transferStart,
    struct _REF2D_driver *ref2d,
    struct _REF2D_driver **drv)
{
    /* Initialise the internal state of the filter driver */
    DC.mi = *modeInfo;
    DC.ref2d = ref2d;
    DC.refState = ref2d->DriverState;
    DC.enabled = IS_FLIPPED();
    FLIPPED_initREF.dwSize = sizeof(FLIPPED_initREF);
    if (!ref2d->QueryFunctions(GA_GET_INITFUNCS,&FLIPPED_initREF))
        return false;
    FLIPPED_cursorREF.dwSize = sizeof(FLIPPED_cursorREF);
    if (!ref2d->QueryFunctions(GA_GET_CURSORFUNCS,&FLIPPED_cursorREF))
        return false;
    FLIPPED_state2dREF.dwSize = sizeof(FLIPPED_state2dREF);
    if (!ref2d->QueryFunctions(GA_GET_2DSTATEFUNCS,&FLIPPED_state2dREF))
        return false;
    FLIPPED_draw2dREF.dwSize = sizeof(FLIPPED_draw2dREF);
    if (!ref2d->QueryFunctions(GA_GET_2DRENDERFUNCS,&FLIPPED_draw2dREF))
        return false;

    /* Allocate memory for the rotation buffer big enough to hold a color
     * bitmap the size of the display screen for rotations.
     */
    DC.bufferSize = DC.mi.YResolution * DC.mi.BytesPerScanLine;
    if ((DC.sysBuffer = malloc(DC.bufferSize)) == NULL)
        return false;
    DC.buffer = DC.sysBuffer;
    DC.transferStart = transferStart;

    /* Initialise all the filter driver vectors */
    INIT_INIT_VECTORS;
    INIT_CURSOR_VECTORS;
    INIT_FILTER_VECTORS;
    FlippedEnable();

    /* Return address of our filter driver as top of filter chain */
    *drv = &DRV;
    return true;
}

/****************************************************************************
REMARKS:
Function to clean up when unloading the driver.
****************************************************************************/
void NAPI FLIPPED_UnloadDriver(void)
{
}

/****************************************************************************
REMARKS:
Function to clean up when the filter driver needs to de-initialise (ie:
going back to text mode when the ref2d driver is unloaded).
****************************************************************************/
void NAPI FLIPPED_DeInitFilter(void)
{
    // Free our system memory blit buffer
    free(DC.sysBuffer);

    // Unhook our QueryFunctions hook, and restore the init vectors
    DC.hwCtx->loader.QueryFunctions = DC.QueryFunctionsDC;
    FLIPPED_initREF.dwSize = sizeof(FLIPPED_initREF);
    DC.hwCtx->loader.QueryFunctions(GA_GET_INITFUNCS,-2,&FLIPPED_initREF);
    INIT_INIT_VECTORS;
}

/****************************************************************************
PARAMETERS:
funcs       - Pointer to the function list to make valid
size        - Size of the function list in bytes

RETURNS:
True if valid, false if not.
****************************************************************************/
static ibool MakeValid(
    ulong *funcs,
    int size)
{
    int i,empty = true;

    for (i = 0; i < size/4; i++) {
        if (funcs[i] != 0)
            empty = false;
        }
    return !empty;
}

/****************************************************************************
DESCRIPTION:
Returns the device driver entry points for the specific function group.

PARAMETERS:
id          - Identifier for the function group to get pointers for
safetyLevel - Safety level being requested.
funcs       - Place to store the function block

RETURNS:
True if functions available, false if not.

REMARKS:
This function allows the filter driver to hook the device driver functions
that are called before any modes are initialised. This is used to hook
functions such as GetVideoModeInfo and SetVideoMode as necessary in the
filter driver.
****************************************************************************/
ibool NAPI FLIPPED_QueryFunctionsDC(
    N_uint32 id,
    N_int32 safetyLevel,
    void *funcs)
{
    ulong   dwSize = *((ulong*)funcs);

    memset(funcs,0,dwSize);
    if (safetyLevel >= 0) {
        switch (id) {
            case GA_GET_INITFUNCS:      RETURN_VALID(FLIPPED_initPT);
            }
        }
    *((ulong*)funcs) = dwSize;
    return DC.QueryFunctionsDC(id,safetyLevel,funcs);
}

/****************************************************************************
DESCRIPTION:
Returns the reference rasterisers for the specified function group.

PARAMETERS:
id      - Identifier for the function group to get pointers for
funcs   - Place to store the function block

RETURNS:
True if functions available, false if not.

REMARKS:
This function is the main function that is used by the graphics application
code to get a block of function pointers for a specified function group.
The function groups are defined GA_funcGroupsType enumeration, and breaks
up the functions in the device driver API into groups of logically similar
functions.

NOTE: If we get a safetyLevel value of less than zero, it indicates that
      we a being called by some other filter driver and should simply
      pass the call onto the low level hardware driver. Otherwise we may
      end up getting into an infinite loop!
****************************************************************************/
ibool NAPI FLIPPED_QueryFunctions(
    N_uint32 id,
    void *funcs)
{
    ulong   dwSize = *((ulong*)funcs);

    memset(funcs,0,dwSize);
    switch (id) {
        case GA_GET_INITFUNCS:      RETURN_VALID(FLIPPED_initPT);
        case GA_GET_CURSORFUNCS:    RETURN_VALID(FLIPPED_cursorPT);
        case GA_GET_2DSTATEFUNCS:   RETURN_VALID(FLIPPED_state2dPT);
        case GA_GET_2DRENDERFUNCS:  RETURN_VALID(FLIPPED_draw2dStub);
        }
    *((ulong*)funcs) = dwSize;
    return DC.ref2d->QueryFunctions(id,funcs);
}

/****************************************************************************
PARAMETERS:
drawBuf             - Buffer to make the active drawing buffer

RETURNS:
True on success, false on failure.

REMARKS:
This function rotates the draw buffer such that it will work correctly
for the flipped display mode when doing page flipping. We do not support
rendering to offscreen buffers unless the code above is aware that it
needs to do some translation of coordinates.
****************************************************************************/
static ibool FlipDrawBuffer(void)
{
    /* We dont support any draw buffers except full sized pages which allows
     * us to support page flipping in hardware in flipped modes.
     */
    if (DC.buf.Width != DC.mi.XResolution || DC.buf.Height != DC.mi.YResolution)
        return false;
    DC.buf.Stride = DC.mi.BytesPerScanLine;
    DC.maxX = DC.buf.Width-1;
    DC.maxY = DC.buf.Height-1;
    DC.maxXFX = N_TOFIX(DC.maxX);
    DC.maxYFX = N_TOFIX(DC.maxY);
    return true;
}

/****************************************************************************
DESCRIPTION:
Sets a display buffer as the active drawing buffer.

PARAMETERS:
drawBuf             - Buffer to make the active drawing buffer
frameBuffer         - Pointer to the start of frameBuffer memory
bitsPerPixel        - Color depth for the buffer
beginDirectAccess   - Direct frameBuffer arbitration function (NULL if none)
endDirectAccess     - Direct frameBuffer arbitration function (NULL if none)

RETURNS:
0 on success, -1 on failure.

REMARKS:
This function allows the application to tell make a display memory
buffer the active rendering buffer for all subsequent rendering commands.
****************************************************************************/
N_int32 NAPI FLIPPED_SetDrawBufferSW(
    GA_buffer *drawBuf,
    void *frameBuffer,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pf,
    GA_devCtx *hwCtx,
    N_int32 forceSoftwareOnly)
{
    DC.buf = *drawBuf;
    DC.hwCtx = hwCtx;
    DC.enabled = (IS_FLIPPED() && frameBuffer == DC.hwCtx->LinearMem);
    if (DC.enabled && !FlipDrawBuffer())
        return -1;
    if (DC.enabled) {
        DC.VRAMBufferStart = DC.VRAMBufferStartY = 0;
        if (DC2D.hwDraw.BitBltLin) {
            if ((DC.mi.MaxLinearOffset - DC.transferStart) > DC.bufferSize) {
                /* Buffers are all allocated in video memory and blitted
                 * with the linear blit functions for performance.
                 */
                DC.VRAMBufferStart = DC.transferStart;
                DC.VRAMBufferStartY = 0;
                }
            }
        else if (DC2D.hwDraw.BitBlt) {
            int startY = (DC.transferStart - DC.buf.Offset) / DC.mi.BytesPerScanLine;
            int sizeY = DC.bufferSize / DC.mi.BytesPerScanLine;
            if ((int)(DC.mi.MaxScanLines - startY) > sizeY) {
                /* Buffers are all allocated in video memory and blitted
                 * with the rectangular blit functions.
                 */
                DC.VRAMBufferStart = DC.transferStart;
                DC.VRAMBufferStartY = startY;
                }
            }
        if (DC.VRAMBufferStart)
            DC.buffer = (uchar*)frameBuffer + DC.VRAMBufferStart;
        else
            DC.buffer = DC.sysBuffer;
        }
    FlippedEnable();
    return DC.ref2d->SetDrawBuffer(&DC.buf,frameBuffer,bitsPerPixel,pf,hwCtx,forceSoftwareOnly);
}

/****************************************************************************
PARAMETERS:
drawBuf - New hardware draw buffer to make active

RETURNS:
0 on success, -1 on failure.

REMARKS:
This function changes the hardware draw buffer for both the software
rasteriser and the hardware device. This can also be used to change the
draw buffer for a software only buffer, but is generally only used when
using hardware.
****************************************************************************/
N_int32 NAPI FLIPPED_SetDrawBuffer(
    GA_buffer *drawBuf)
{
    DC.buf = *drawBuf;
    if (DC.enabled && !FlipDrawBuffer())
        return -1;
    return STATE2D.SetDrawBuffer(&DC.buf);
}

/****************************************************************************
PARAMETERS:
surface         - Pointer to the start of the memory buffer for drawing
xRes            - X resolution for the surface
yRes            - Y resolution for the surface
bytesPerLine    - Scanline width for the memory buffer for drawing
bitsPerPixel    - New color depth for the surface
pf              - New pixel format for the surface

REMARKS:
This function changes the software draw buffer quickly, to allow the
software renderer to render to bitmaps in different locations in memory.
We do in here for speed, since we want to avoid the overhead of setting
the full draw buffer for the reference rasteriser code. If the color
depth or pixel format changes, this function will switch all the internal
rendering vectors to put to the new vectors for the appropriate functions.
This allows the reference rasteriser to work with all color depths active
in a single driver (and for our debug library to work properly).
****************************************************************************/
void NAPI FLIPPED_SetDrawSurface(
    void *surface,
    N_int32 xRes,
    N_int32 yRes,
    N_int32 bytesPerLine,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pf)
{
    // Determine if we need to disable the flipped features
    DC.enabled = (IS_FLIPPED() && DC.refState->haveHWDraw && surface == (DC.refState->frameBuffer + DC.refState->drawOffset));
    DC.buf.Width = xRes;
    DC.buf.Height = yRes;
    DC.buf.Stride = bytesPerLine;
    if (DC.enabled)
        FlipDrawBuffer();
    DC.ref2d->SetDrawSurface(surface,DC.buf.Width,DC.buf.Height,DC.buf.Stride,bitsPerPixel,pf);
}

/****************************************************************************
REMARKS:
Handles updating of the internals when the options structures are updated.
This allows inversion mode to be toggled automatically at runtime using
either the lobal device driver or global SNAP options functions.
****************************************************************************/
static void UpdateOptions(void)
{
    DC.enabled = IS_FLIPPED();
    FlippedEnable();
}

/****************************************************************************
REMARKS:
Handles overriding of the options in the driver. This allows us to know
when the invert option has been disabled or enabled at runtime.
****************************************************************************/
void NAPI FLIPPED_SetOptions(
    GA_options *options)
{
    FLIPPED_initREF.SetOptions(options);
    FLIPPED_initREF.GetOptions(&DC.opt);
    UpdateOptions();
}

/****************************************************************************
REMARKS:
This function gets called when the global options are changed at runtime,
allowing the filter driver to update the internals based on the changed
global options.
****************************************************************************/
void NAPI FLIPPED_GlobalOptionsUpdated(void)
{
    UpdateOptions();
}

/****************************************************************************
PARAMETERS:
dst         - Destination bitmap buffer
src         - Source bitmap buffer
pitch       - Width of the bitmap in bytes
height      - Height of the bitmap in scanlines

REMARKS:
This function rotates an 1BPP bitmap image, and stores the value in the
destination buffer. The source buffer is not modified.
****************************************************************************/
static void RotateBitmap1BPP(
    uchar *dst,
    uchar *src,
    N_int32 pitch,
    N_int32 height)
{
    int max = height * pitch;

// TODO: This will be a pain in the arse!! We need to process a scanline
//       at a time, but we *also* need to process the bitmaps such that
//       we properly take into account rotating the pixels in the bitmap
//       as necessary.
//
//       Perhaps this can be done simply using the following code to rotate
//       the entire bitmap, but then offset the position as necessary
//       and draw with clipping??

    dst += max;
    while (max--)
        *--dst = mirrorTable1BPP[*src++];
}

/****************************************************************************
PARAMETERS:
dst     - Destination bitmap buffer
src     - Source bitmap buffer
pitch   - Source bitmap pitch in bytes
height  - Height of the bitmap in scanlines

REMARKS:
This function rotates an 4BPP bitmap image, and stores the value in the
destination buffer. The source buffer is not modified.
****************************************************************************/
static void RotateBitmap4BPP(
    uchar *dst,
    uchar *src,
    N_int32 srcPitch,
    N_int32 dstPitch,
    N_int32 width,
    N_int32 height,
    N_int32 flipY)
{
// TODO: This will be a pain in the arse!!
#if 0
    int i,max = height * srcPitch;

    dst += max;
    for (i = 0; i < max; i++)
        *--dst = mirrorTable4BPP[*src++];
#endif
    (void)dst;
    (void)src;
    (void)srcPitch;
    (void)dstPitch;
    (void)width;
    (void)height;
    (void)flipY;
}

/****************************************************************************
PARAMETERS:
dst         - Destination bitmap buffer
src         - Source bitmap buffer
srcPitch    - Source bitmap stride in bytes
dstPitch    - Destination bitmap stride in bytes
height      - Height of the source bitmap in scanlines

REMARKS:
This function rotates an 8bpp bitmap image, and stores the value in the
destination buffer. The source buffer is not modified.
****************************************************************************/
static void RotateBitmap8BPP(
    uchar *dst,
    uchar *src,
    N_int32 srcPitch,
    N_int32 dstPitch,
    N_int32 width,
    N_int32 height,
    N_int32 flipY)
{
    BINC(dst,dstPitch * (height-1) + width-1);
    if (flipY) {
        BINC(src,srcPitch * (height-1));
        srcPitch = -srcPitch;
        }
    while (height--) {
        uchar *psrc = src;
        uchar *pdst = dst;
        int count = width;
        while (count--)
            *pdst-- = *psrc++;
        BDEC(dst,dstPitch);
        BINC(src,srcPitch);
        }
}

/****************************************************************************
PARAMETERS:
dst         - Destination bitmap buffer
src         - Source bitmap buffer
srcPitch    - Source bitmap stride in bytes
dstPitch    - Destination bitmap stride in bytes
height      - Height of the source bitmap in scanlines

REMARKS:
This function rotates an 16BPP bitmap image, and stores the value in the
destination buffer. The source buffer is not modified.
****************************************************************************/
static void RotateBitmap16BPP(
    uchar *dst,
    uchar *src,
    N_int32 srcPitch,
    N_int32 dstPitch,
    N_int32 width,
    N_int32 height,
    N_int32 flipY)
{
    BINC(dst,dstPitch * (height-1) + (width-1)*2);
    if (flipY) {
        BINC(src,srcPitch * (height-1));
        srcPitch = -srcPitch;
        }
    while (height--) {
        ushort *psrc = (ushort*)src;
        ushort *pdst = (ushort*)dst;
        int count = width;
        while (count--)
            *pdst-- = *psrc++;
        BDEC(dst,dstPitch);
        BINC(src,srcPitch);
        }
}

/****************************************************************************
PARAMETERS:
dst         - Destination bitmap buffer
src         - Source bitmap buffer
srcPitch    - Source bitmap stride in bytes
dstPitch    - Destination bitmap stride in bytes
height      - Height of the source bitmap in scanlines

REMARKS:
This function rotates an 24BPP bitmap image, and stores the value in the
destination buffer. The source buffer is not modified.
****************************************************************************/
static void RotateBitmap24BPP(
    uchar *dst,
    uchar *src,
    N_int32 srcPitch,
    N_int32 dstPitch,
    N_int32 width,
    N_int32 height,
    N_int32 flipY)
{
    BINC(dst,dstPitch * (height-1) + (width-1)*3);
    if (flipY) {
        BINC(src,srcPitch * (height-1));
        srcPitch = -srcPitch;
        }
    while (height--) {
        uchar *psrc = src;
        uchar *pdst = dst;
        int count = width;
        while (count--) {
            pdst[0] = psrc[0];
            pdst[1] = psrc[1];
            pdst[2] = psrc[2];
            pdst -= 3;
            psrc += 3;
            }
        BDEC(dst,dstPitch);
        BINC(src,srcPitch);
        }
}

/****************************************************************************
PARAMETERS:
dst         - Destination bitmap buffer
src         - Source bitmap buffer
srcPitch    - Source bitmap stride in bytes
dstPitch    - Destination bitmap stride in bytes
height      - Height of the source bitmap in scanlines

REMARKS:
This function rotates an 32BPP bitmap image, and stores the value in the
destination buffer. The source buffer is not modified.
****************************************************************************/
static void RotateBitmap32BPP(
    uchar *dst,
    uchar *src,
    N_int32 srcPitch,
    N_int32 dstPitch,
    N_int32 width,
    N_int32 height,
    N_int32 flipY)
{
    BINC(dst,dstPitch * (height-1) + (width-1)*4);
    if (flipY) {
        BINC(src,srcPitch * (height-1));
        srcPitch = -srcPitch;
        }
    while (height--) {
        ulong *psrc = (ulong*)src;
        ulong *pdst = (ulong*)dst;
        int count = width;
        while (count--)
            *pdst-- = *psrc++;
        BDEC(dst,dstPitch);
        BINC(src,srcPitch);
        }
}

/****************************************************************************
PARAMETERS:
src             - Pointer to the source bitmap to rotate
bitsPerPixel    - Pixel depth of the source bitmap in bits
pitch           - Stride of the source bitmap in bytes
width           - Width of the source bitmap
height          - Height of the source bitmap
****************************************************************************/
void * NAPI FLIPPED_RotateBitmap(
    void *src,
    N_int32 bitsPerPixel,
    N_int32 *pitch,
    N_int32 width,
    N_int32 height)
{
    if (width > DC.mi.XResolution || height > DC.mi.YResolution)
        return NULL;
    switch (bitsPerPixel) {
        case 1:
            RotateBitmap1BPP(DC.sysBuffer,src,*pitch,height);
            break;
        case 4:
            RotateBitmap4BPP(DC.sysBuffer,src,*pitch,*pitch,width,height,false);
            break;
        case 8:
            RotateBitmap8BPP(DC.sysBuffer,src,*pitch,*pitch,width,height,false);
            break;
        case 15:
        case 16:
            RotateBitmap16BPP(DC.sysBuffer,src,*pitch,*pitch,width,height,false);
            break;
        case 24:
            RotateBitmap24BPP(DC.sysBuffer,src,*pitch,*pitch,width,height,false);
            break;
        case 32:
            RotateBitmap32BPP(DC.sysBuffer,src,*pitch,*pitch,width,height,false);
            break;
        default:
            return NULL;
        }
    return DC.sysBuffer;
}

/****************************************************************************
PARAMETERS:
index   - Index of the pattern to download
pattern - Pointer to pattern data to download
****************************************************************************/
void NAPI FLIPPED_Set8x8MonoPattern(
    N_int32 index,
    GA_pattern *pattern)
{
    if (DC.enabled) {
        GA_pattern  pat;

        /* Rotate the 8x8 mono pattern before downloading to hardware */
        pat.p[0] = mirrorTable1BPP[pattern->p[7]];
        pat.p[1] = mirrorTable1BPP[pattern->p[6]];
        pat.p[2] = mirrorTable1BPP[pattern->p[5]];
        pat.p[3] = mirrorTable1BPP[pattern->p[4]];
        pat.p[4] = mirrorTable1BPP[pattern->p[3]];
        pat.p[5] = mirrorTable1BPP[pattern->p[2]];
        pat.p[6] = mirrorTable1BPP[pattern->p[1]];
        pat.p[7] = mirrorTable1BPP[pattern->p[0]];
        STATE2D.Set8x8MonoPattern(index,&pat);
        }
    else
        STATE2D.Set8x8MonoPattern(index,pattern);
}

/****************************************************************************
PARAMETERS:
index   - Index of the pattern to download
pattern - Pointer to pattern data to download
****************************************************************************/
void NAPI FLIPPED_Set8x8ColorPattern(
    N_int32 index,
    GA_colorPattern *pattern)
{
    if (DC.enabled) {
        int             i;
        GA_colorPattern pat;
        switch (DC.mi.BitsPerPixel) {
            case 8: {
                N_uint8 *pd = (N_uint8*)pat.b8.p;
                N_uint8 *ps = &pattern->b8.p[7][0];
                for (i = 0; i < 8; i++) {
                    pd[0] = ps[7];
                    pd[1] = ps[6];
                    pd[2] = ps[5];
                    pd[3] = ps[4];
                    pd[4] = ps[3];
                    pd[5] = ps[2];
                    pd[6] = ps[1];
                    pd[7] = ps[0];
                    pd += 8;
                    ps -= 8;
                    }
                } break;
            case 15:
            case 16: {
                N_uint16 *pd = (N_uint16*)pat.b16.p;
                N_uint16 *ps = &pattern->b16.p[7][0];
                for (i = 0; i < 8; i++) {
                    pd[0] = ps[7];
                    pd[1] = ps[6];
                    pd[2] = ps[5];
                    pd[3] = ps[4];
                    pd[4] = ps[3];
                    pd[5] = ps[2];
                    pd[6] = ps[1];
                    pd[7] = ps[0];
                    pd += 8;
                    ps -= 8;
                    }
                } break;
            case 24: {
                N_uint8 *pd = &pat.b24.p[0][0][0];
                N_uint8 *ps = &pattern->b24.p[7][0][0];
                for (i = 0; i < 8; i++) {
                    pd[0*3+0] = ps[7*3+0]; pd[0*3+1] = ps[7*3+1]; pd[0*3+2] = ps[7*3+2];
                    pd[1*3+0] = ps[6*3+0]; pd[1*3+1] = ps[6*3+1]; pd[1*3+2] = ps[6*3+2];
                    pd[2*3+0] = ps[5*3+0]; pd[2*3+1] = ps[5*3+1]; pd[2*3+2] = ps[5*3+2];
                    pd[3*3+0] = ps[4*3+0]; pd[3*3+1] = ps[4*3+1]; pd[3*3+2] = ps[4*3+2];
                    pd[4*3+0] = ps[3*3+0]; pd[4*3+1] = ps[3*3+1]; pd[4*3+2] = ps[3*3+2];
                    pd[5*3+0] = ps[2*3+0]; pd[5*3+1] = ps[2*3+1]; pd[5*3+2] = ps[2*3+2];
                    pd[6*3+0] = ps[1*3+0]; pd[6*3+1] = ps[1*3+1]; pd[6*3+2] = ps[1*3+2];
                    pd[7*3+0] = ps[0*3+0]; pd[7*3+1] = ps[0*3+1]; pd[7*3+2] = ps[0*3+2];
                    pd += 24;
                    ps -= 24;
                    }
                }  break;
            case 32: {
                N_uint32 *pd = (N_uint32*)pat.b32.p;
                N_uint32 *ps = &pattern->b32.p[7][0];
                for (i = 0; i < 8; i++) {
                    pd[0] = ps[7];
                    pd[1] = ps[6];
                    pd[2] = ps[5];
                    pd[3] = ps[4];
                    pd[4] = ps[3];
                    pd[5] = ps[2];
                    pd[6] = ps[1];
                    pd[7] = ps[0];
                    pd += 8;
                    ps -= 8;
                    }
                } break;
            }
        STATE2D.Set8x8ColorPattern(index,&pat);
        }
    else
        STATE2D.Set8x8ColorPattern(index,pattern);
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
x1      - First X coordinate of scanline
x2      - Second X coordinate of scanline
****************************************************************************/
void NAPI FLIPPED_DrawScan(
    N_int32 y,
    N_int32 x1,
    N_int32 x2)
{
    int width;

    if ((width = (x2 - x1)) == 0)
        return;
    if (width < 0) {
        x1 = x2;
        width = -width;
        }
    DRAW2D.DrawRect(FLIP_RECTX(x1,width),FLIPY(y),width,1);
}

/****************************************************************************
PARAMETERS:
y           - Y coordinate for scanline
x1          - First X coordinate of scanline
x2          - Second X coordinate of scanline
****************************************************************************/
void NAPI FLIPPED_DrawPattScan(
    N_int32 y,
    N_int32 x1,
    N_int32 x2)
{
    int width;

    if ((width = (x2 - x1)) == 0)
        return;
    if (width < 0) {
        x1 = x2;
        width = -width;
        }
    DRAW2D.DrawPattRect(FLIP_RECTX(x1,width),FLIPY(y),width,1);
}

/****************************************************************************
PARAMETERS:
y           - Y coordinate for scanline
x1          - First X coordinate of scanline
x2          - Second X coordinate of scanline
****************************************************************************/
void NAPI FLIPPED_DrawColorPattScan(
    N_int32 y,
    N_int32 x1,
    N_int32 x2)
{
    int width;

    if ((width = (x2 - x1)) == 0)
        return;
    if (width < 0) {
        x1 = x2;
        width = -width;
        }
    DRAW2D.DrawColorPattRect(FLIP_RECTX(x1,width),FLIPY(y),width,1);
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI FLIPPED_DrawScanList(
    N_int32 y,
    N_int32 length,
    N_int16 *scans)
{
    while (length--) {
        int width;
        int x1 = *scans++;
        int x2 = *scans++;
        if ((width = (x2 - x1)) != 0) {
            if (width < 0) {
                x1 = x2;
                width = -width;
                }
            DRAW2D.DrawRect(FLIP_RECTX(x1,width),FLIPY(y),width,1);
            }
        y++;
        }
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI FLIPPED_DrawPattScanList(
    N_int32 y,
    N_int32 length,
    N_int16 *scans)
{
    while (length--) {
        int width;
        int x1 = *scans++;
        int x2 = *scans++;
        if ((width = (x2 - x1)) != 0) {
            if (width < 0) {
                x1 = x2;
                width = -width;
                }
            DRAW2D.DrawPattRect(FLIP_RECTX(x1,width),FLIPY(y),width,1);
            }
        y++;
        }
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI FLIPPED_DrawColorPattScanList(
    N_int32 y,
    N_int32 length,
    N_int16 *scans)
{
    while (length--) {
        int width;
        int x1 = *scans++;
        int x2 = *scans++;
        if ((width = (x2 - x1)) != 0) {
            if (width < 0) {
                x1 = x2;
                width = -width;
                }
            DRAW2D.DrawColorPattRect(FLIP_RECTX(x1,width),FLIPY(y),width,1);
            }
        y++;
        }
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI FLIPPED_DrawEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int i,j,maxIndex;

    // TODO: This is wrong!
    (void)y;
    maxIndex = length-1;
    for (i = 0,j = height; i < maxIndex; i++,j--,scans += 2) {
        FLIPPED_DrawScan(i,scans[0],scans[1]);
        FLIPPED_DrawScan(j,scans[0],scans[1]);
        }
    if (!(height & 1))
        FLIPPED_DrawScan(i,scans[0],scans[0]);
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI FLIPPED_DrawPattEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int i,j,maxIndex;

    // TODO: This is wrong!
    (void)y;
    maxIndex = length-1;
    for (i = 0,j = height; i < maxIndex; i++,j--,scans += 2) {
        FLIPPED_DrawPattScan(i,scans[0],scans[1]);
        FLIPPED_DrawPattScan(j,scans[0],scans[1]);
        }
    if (!(height & 1))
        FLIPPED_DrawPattScan(i,scans[0],scans[0]);
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI FLIPPED_DrawColorPattEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int i,j,maxIndex;

    // TODO: This is wrong!
    (void)y;
    maxIndex = length-1;
    for (i = 0,j = height; i < maxIndex; i++,j--,scans += 2) {
        FLIPPED_DrawColorPattScan(i,scans[0],scans[1]);
        FLIPPED_DrawColorPattScan(j,scans[0],scans[1]);
        }
    if (!(height & 1))
        FLIPPED_DrawColorPattScan(i,scans[0],scans[0]);
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis + pen height adjustment
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI FLIPPED_DrawFatEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int     i,j;

    // TODO: This is wrong!
    (void)y;
    for (i = 0,j = height; i < length; i++,j--,scans += 4) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            FLIPPED_DrawScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            FLIPPED_DrawScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            FLIPPED_DrawScan(j,scans[I_LEFTL],scans[I_LEFTR]);
            FLIPPED_DrawScan(j,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            FLIPPED_DrawScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            FLIPPED_DrawScan(j,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
    if ((height+1) & 1) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            FLIPPED_DrawScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            FLIPPED_DrawScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            FLIPPED_DrawScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis + pen height adjustment
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI FLIPPED_DrawPattFatEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int     i,j;

    // TODO: This is wrong!
    (void)y;
    for (i = 0,j = height; i < length; i++,j--,scans += 4) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            FLIPPED_DrawPattScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            FLIPPED_DrawPattScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            FLIPPED_DrawPattScan(j,scans[I_LEFTL],scans[I_LEFTR]);
            FLIPPED_DrawPattScan(j,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            FLIPPED_DrawPattScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            FLIPPED_DrawPattScan(j,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
    if ((height+1) & 1) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            FLIPPED_DrawPattScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            FLIPPED_DrawPattScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            FLIPPED_DrawPattScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis + pen height adjustment
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI FLIPPED_DrawColorPattFatEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int     i,j;

    // TODO: This is wrong!
    (void)y;
    for (i = 0,j = height; i < length; i++,j--,scans += 4) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            FLIPPED_DrawColorPattScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            FLIPPED_DrawColorPattScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            FLIPPED_DrawColorPattScan(j,scans[I_LEFTL],scans[I_LEFTR]);
            FLIPPED_DrawColorPattScan(j,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            FLIPPED_DrawColorPattScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            FLIPPED_DrawColorPattScan(j,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
    if ((height+1) & 1) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            FLIPPED_DrawColorPattScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            FLIPPED_DrawColorPattScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            FLIPPED_DrawColorPattScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the rectangle to draw
top     - Top coordinate of the rectangle to draw
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
****************************************************************************/
void NAPI FLIPPED_DrawRect(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height)
{
    DRAW2D.DrawRect(FLIP_RECTX(left,width),FLIP_RECTY(top,height),width,height);
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the rectangle to draw
top     - Top coordinate of the rectangle to draw
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
****************************************************************************/
void NAPI FLIPPED_DrawPattRect(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height)
{
    DRAW2D.DrawPattRect(FLIP_RECTX(left,width),FLIP_RECTY(top,height),width,height);
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the rectangle to draw
top     - Top coordinate of the rectangle to draw
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
****************************************************************************/
void NAPI FLIPPED_DrawColorPattRect(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height)
{
    DRAW2D.DrawColorPattRect(FLIP_RECTX(left,width),FLIP_RECTY(top,height),width,height);
}

/****************************************************************************
PARAMETERS:
trap    - Pointer to the GA_trap structure describing the trapezoid
****************************************************************************/
void NAPI FLIPPED_DrawTrap(
    GA_trap *trap)
{
    int y,x1,x2,count = trap->count;

    if (!count)
        return;
    y = trap->y;
    x1 = trap->x1;
    x2 = trap->x2;
    trap->y += count;
    while (count--) {
        int width;
        int ix1 = (x1 + 0x8000) >> 16;
        int ix2 = (x2 + 0x8000) >> 16;
        if ((width = (ix2 - ix1)) != 0) {
            if (width < 0) {
                ix1 = ix2;
                width = -width;
                }
            DRAW2D.DrawRect(FLIP_RECTX(ix1,width),FLIPY(y),width,1);
            }
        x1 += trap->slope1;
        x2 += trap->slope2;
        y++;
        }
    trap->x1 = x1;
    trap->x2 = x2;
}

/****************************************************************************
PARAMETERS:
trap    - Pointer to the GA_trap structure describing the trapezoid
****************************************************************************/
void NAPI FLIPPED_DrawPattTrap(
    GA_trap *trap)
{
    int y,x1,x2,count = trap->count;

    if (!count)
        return;
    y = trap->y;
    x1 = trap->x1;
    x2 = trap->x2;
    trap->y += count;
    while (count--) {
        int width;
        int ix1 = (x1 + 0x8000) >> 16;
        int ix2 = (x2 + 0x8000) >> 16;
        if ((width = (ix2 - ix1)) != 0) {
            if (width < 0) {
                ix1 = ix2;
                width = -width;
                }
            DRAW2D.DrawPattRect(FLIP_RECTX(ix1,width),FLIPY(y),width,1);
            }
        x1 += trap->slope1;
        x2 += trap->slope2;
        y++;
        }
    trap->x1 = x1;
    trap->x2 = x2;
}

/****************************************************************************
PARAMETERS:
trap    - Pointer to the GA_trap structure describing the trapezoid
****************************************************************************/
void NAPI FLIPPED_DrawColorPattTrap(
    GA_trap *trap)
{
    int y,x1,x2,count = trap->count;

    if (!count)
        return;
    y = trap->y;
    x1 = trap->x1;
    x2 = trap->x2;
    trap->y += count;
    while (count--) {
        int width;
        int ix1 = (x1 + 0x8000) >> 16;
        int ix2 = (x2 + 0x8000) >> 16;
        if ((width = (ix2 - ix1)) != 0) {
            if (width < 0) {
                ix1 = ix2;
                width = -width;
                }
            DRAW2D.DrawColorPattRect(FLIP_RECTX(ix1,width),FLIPY(y),width,1);
            }
        x1 += trap->slope1;
        x2 += trap->slope2;
        y++;
        }
    trap->x1 = x1;
    trap->x2 = x2;
}

/****************************************************************************
PARAMETERS:
x1          - X1 coordinate
y1          - Y1 coordinate
x2          - X2 coordinate
y2          - Y2 coordinate
drawLast    - 1 to draw last pixel, 0 to skip it
****************************************************************************/
void NAPI FLIPPED_DrawLineInt(
    N_int32 x1,
    N_int32 y1,
    N_int32 x2,
    N_int32 y2,
    N_int32 drawLast)
{
    DRAW2D.DrawLineInt(FLIPX(x1),FLIPY(y1),FLIPX(x2),FLIPY(y2),drawLast);
}

/****************************************************************************
PARAMETERS:
x1          - X1 coordinate
y1          - Y1 coordinate
x2          - X2 coordinate
y2          - Y2 coordinate
drawLast    - 1 to draw last pixel, 0 to skip it
transparent - 1 if the line is transparent, 0 if opaque
****************************************************************************/
void NAPI FLIPPED_DrawStippleLineInt(
    N_int32 x1,
    N_int32 y1,
    N_int32 x2,
    N_int32 y2,
    N_int32 drawLast,
    N_int32 transparent)
{
    DRAW2D.DrawStippleLineInt(FLIPX(x1),FLIPY(y1),FLIPX(x2),FLIPY(y2),drawLast,transparent);
}

/****************************************************************************
PARAMETERS:
x1              - X1 coordinate
y1              - Y1 coordinate
initialError    - Initial error term
majorInc        - Major increment factor for error term
diagInc         - Diagonal increment factor for error term
count           - Number of pixels to draw
flags           - Flags to control drawing
****************************************************************************/
void NAPI FLIPPED_DrawBresenhamLine(
    N_int32 x1,
    N_int32 y1,
    N_int32 initialError,
    N_int32 majorInc,
    N_int32 diagInc,
    N_int32 count,
    N_int32 flags)
{
    N_int32 newFlags;

    newFlags = flags & (gaLineXMajor | gaLineDoLastPel);
    if (!(flags & gaLineYPositive))
        newFlags |= gaLineYPositive;
    if (!(flags & gaLineXPositive))
        newFlags |= gaLineXPositive;
    DRAW2D.DrawBresenhamLine(FLIPX(x1),FLIPY(y1),initialError,majorInc,diagInc,count,newFlags);
}

/****************************************************************************
PARAMETERS:
x1              - X1 coordinate
y1              - Y1 coordinate
initialError    - Initial error term
majorInc        - Major increment factor for error term
diagInc         - Diagonal increment factor for error term
count           - Number of pixels to draw
flags           - Flags to control drawing
transparent     - 1 if the line is transparent, 0 if opaque
****************************************************************************/
void NAPI FLIPPED_DrawBresenhamStippleLine(
    N_int32 x1,
    N_int32 y1,
    N_int32 initialError,
    N_int32 majorInc,
    N_int32 diagInc,
    N_int32 count,
    N_int32 flags,
    N_int32 transparent)
{
    N_int32 newFlags;

    newFlags = flags & (gaLineXMajor | gaLineDoLastPel);
    if (!(flags & gaLineYPositive))
        newFlags |= gaLineYPositive;
    if (!(flags & gaLineXPositive))
        newFlags |= gaLineXPositive;
    DRAW2D.DrawBresenhamStippleLine(FLIPX(x1),FLIPY(y1),initialError,majorInc,diagInc,count,newFlags,transparent);
}

/****************************************************************************
PARAMETERS:
x1          - X1 coordinate
y1          - Y1 coordinate
x2          - X2 coordinate
y2          - Y2 coordinate
drawLast    - 1 to draw last pixel, 0 to skip it
transparent - 1 if the line is transparent, 0 if opaque
****************************************************************************/
void NAPI FLIPPED_DrawStyleLineInt(
    N_int32 x1,
    N_int32 y1,
    N_int32 x2,
    N_int32 y2,
    N_int32 drawLast,
    N_int32 transparent)
{
    DRAW2D.DrawStyleLineInt(FLIPX(x1),FLIPY(y1),FLIPX(x2),FLIPY(y2),drawLast,transparent);
}

/****************************************************************************
PARAMETERS:
x1              - X1 coordinate
y1              - Y1 coordinate
initialError    - Initial error term
majorInc        - Major increment factor for error term
diagInc         - Diagonal increment factor for error term
count           - Number of pixels to draw
flags           - Flags to control drawing
transparent     - 1 if the line is transparent, 0 if opaque
****************************************************************************/
void NAPI FLIPPED_DrawBresenhamStyleLine(
    N_int32 x1,
    N_int32 y1,
    N_int32 initialError,
    N_int32 majorInc,
    N_int32 diagInc,
    N_int32 count,
    N_int32 flags,
    N_int32 transparent)
{
    N_int32 newFlags;

    newFlags = flags & (gaLineXMajor | gaLineDoLastPel);
    if (!(flags & gaLineYPositive))
        newFlags |= gaLineYPositive;
    if (!(flags & gaLineXPositive))
        newFlags |= gaLineXPositive;
    DRAW2D.DrawBresenhamStyleLine(FLIPX(x1),FLIPY(y1),initialError,majorInc,diagInc,count,newFlags,transparent);
}

/****************************************************************************
PARAMETERS:
x           - Destination X coordinate to draw the bitmap at
y           - Destination Y coordinate to draw the bitmap at
width       - Width of the bitmap in pixels
height      - Height of the bitmap in pixels
byteWidth   - Width of the bitmap in bytes
image       - Pointer to the bitmap image data to draw
transparent - 1 for transparent, 0 for opaque

REMARKS:
If the transparent parameter is set to 1, where a bit is 0 in the bitmap
image the destination pixel remains untouched. If the transparent
parameter is set to 0, where a bit is 0 in the bitmap image the
destination pixel is drawn in the background color. In all cases where a
bit in the bitmap image is 1, the pixel is drawn in the foreground color.

Note:   This function processes the bitmap data in MSB fashion, in that the
        first pixel drawn corrsponds to bit 7 of the first byte. The second
        pixel is bit 6 of the first byte, ...,the 8th pixel is bit 7 of
        the second byte etc. Both LSB and MSB versions are provided for
        performance.
****************************************************************************/
void NAPI FLIPPED_PutMonoImageMSBSys(
    N_int32 x,
    N_int32 y,
    N_int32 width,
    N_int32 height,
    N_int32 byteWidth,
    N_uint8 *image,
    N_int32 transparent)
{
    RotateBitmap1BPP(DC.sysBuffer,image,byteWidth,height);
    DRAW2D.PutMonoImageMSBSys(FLIP_RECTX(x,width),FLIP_RECTY(y,height),width,height,byteWidth,DC.sysBuffer,transparent);
}

/****************************************************************************
PARAMETERS:
x           - Destination X coordinate to draw the bitmap at
y           - Destination Y coordinate to draw the bitmap at
width       - Width of the bitmap in pixels
height      - Height of the bitmap in pixels
byteWidth   - Width of the bitmap in bytes
image       - Pointer to the bitmap image data to draw
transparent - 1 for transparent, 0 for opaque

REMARKS:
This function is identical to the PutMonoImageMSBSys function, except
that it processes the bitmap data in LSB fashion. This means that the
first pixel drawn corrsponds to bit 0 of the first byte. The second
pixel is bit 1 of the first byte, ...,the 8th pixel is bit 0 of the second
byte etc. Both LSB and MSB versions are provided for performance.
****************************************************************************/
void NAPI FLIPPED_PutMonoImageLSBSys(
    N_int32 x,
    N_int32 y,
    N_int32 width,
    N_int32 height,
    N_int32 byteWidth,
    N_uint8 *image,
    N_int32 transparent)
{
    RotateBitmap1BPP(DC.sysBuffer,image,byteWidth,height);
    DRAW2D.PutMonoImageLSBSys(FLIP_RECTX(x,width),FLIP_RECTY(y,height),width,height,byteWidth,DC.sysBuffer,transparent);
}

/****************************************************************************
PARAMETERS:
x           - Destination X coordinate to draw the bitmap at
y           - Destination Y coordinate to draw the bitmap at
width       - Width of the bitmap in pixels
height      - Height of the bitmap in pixels
byteWidth   - Width of the bitmap in bytes
image       - Pointer to the bitmap image data to draw
transparent - 1 for transparent, 0 for opaque
clipLeft    - Left edge of clip rectangle (inclusive)
clipTop     - Top edge of clip rectangle (inclusive)
clipRight   - Right edge of clip rectangle (exclusive)
clipBottom  - Bottom edge of clip rectangle (exclusive)

REMARKS:
If the transparent parameter is set to 1, where a bit is 0 in the bitmap
image the destination pixel remains untouched. If the transparent
parameter is set to 0, where a bit is 0 in the bitmap image the
destination pixel is drawn in the background color. In all cases where a
bit in the bitmap image is 1, the pixel is drawn in the foreground color.

Note:   This function processes the bitmap data in MSB fashion, in that the
        first pixel drawn corrsponds to bit 7 of the first byte. The second
        pixel is bit 6 of the first byte, ...,the 8th pixel is bit 7 of
        the second byte etc. Both LSB and MSB versions are provided for
        performance.
****************************************************************************/
void NAPI FLIPPED_ClipMonoImageMSBSys(
    N_int32 x,
    N_int32 y,
    N_int32 width,
    N_int32 height,
    N_int32 byteWidth,
    N_uint8 *image,
    N_int32 transparent,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
    int bottom,right;

    // Check if image needs to be clipped at all
    right = x + width;
    bottom = y + height;
    if ((bottom <= clipTop) || (y >= clipBottom) ||
        (right <= clipLeft) || (x >= clipRight))
        return;

    // Clip the top and bottom of the bitmap image from being drawn
    if (y < clipTop) {
        int skip = (clipTop - y);
        y = clipTop;
        image += skip * byteWidth;
        height -= skip;
        }
    if (bottom > clipBottom)
        height -= (bottom - clipBottom);

    // Now clip the left and right edges so that the clip rectangle lies
    // within the first and last bytes of the image.
    if (x < clipLeft) {
        int skip = (clipLeft - x) / 8;
        image += skip;
        x += skip * 8;
        width -= skip * 8;
        }
    else
        clipLeft = x;
    if (right > clipRight)
        width -= (right - clipRight);

    width += (clipLeft-x);
    RotateBitmap1BPP(DC.sysBuffer,image,byteWidth,height);
    DRAW2D.ClipMonoImageMSBSys(FLIP_RECTX(x,width),FLIP_RECTY(y,height),width,height,byteWidth,
        DC.sysBuffer,transparent,
        FLIPX(clipRight),FLIPY(clipBottom),FLIPX(clipLeft),FLIPY(clipTop));
}

/****************************************************************************
PARAMETERS:
x           - Destination X coordinate to draw the bitmap at
y           - Destination Y coordinate to draw the bitmap at
width       - Width of the bitmap in pixels
height      - Height of the bitmap in pixels
byteWidth   - Width of the bitmap in bytes
image       - Pointer to the bitmap image data to draw
transparent - 1 for transparent, 0 for opaque
clipLeft    - Left edge of clip rectangle (inclusive)
clipTop     - Top edge of clip rectangle (inclusive)
clipRight   - Right edge of clip rectangle (exclusive)
clipBottom  - Bottom edge of clip rectangle (exclusive)

REMARKS:
This function is identical to the ClipMonoImageMSBSys function, except
that it processes the bitmap data in LSB fashion. This means that the
first pixel drawn corrsponds to bit 0 of the first byte. The second
pixel is bit 1 of the first byte, ...,the 8th pixel is bit 0 of the second
byte etc. Both LSB and MSB versions are provided for performance.
****************************************************************************/
void NAPI FLIPPED_ClipMonoImageLSBSys(
    N_int32 x,
    N_int32 y,
    N_int32 width,
    N_int32 height,
    N_int32 byteWidth,
    N_uint8 *image,
    N_int32 transparent,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
    int bottom,right;

    // Check if image needs to be clipped at all
    right = x + width;
    bottom = y + height;
    if ((bottom <= clipTop) || (y >= clipBottom) ||
        (right <= clipLeft) || (x >= clipRight))
        return;

    // Clip the top and bottom of the bitmap image from being drawn
    if (y < clipTop) {
        int skip = (clipTop - y);
        y = clipTop;
        image += skip * byteWidth;
        height -= skip;
        }
    if (bottom > clipBottom)
        height -= (bottom - clipBottom);

    // Now clip the left and right edges so that the clip rectangle lies
    // within the first and last bytes of the image.
    if (x < clipLeft) {
        int skip = (clipLeft - x) / 8;
        image += skip;
        x += skip * 8;
        width -= skip * 8;
        }
    else
        clipLeft = x;
    if (right > clipRight)
        width -= (right - clipRight);

    width += (clipLeft-x);
    RotateBitmap1BPP(DC.sysBuffer,image,byteWidth,height);
    DRAW2D.ClipMonoImageLSBSys(FLIP_RECTX(x,width),FLIP_RECTY(y,height),width,height,byteWidth,
        DC.sysBuffer,transparent,
        FLIPX(clipRight),FLIPY(clipBottom),FLIPX(clipLeft),FLIPY(clipTop));
}

/****************************************************************************
PARAMETERS:
srcLeft - Left coordinate of the source rectangle to copy
srcTop  - Top coordinate of the source rectangle to copy
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
dstLeft - Left coordinate of destination
dstTop  - Top coordinate of destination
mix     - Mix code for the copy (GA_mixCodesType)
****************************************************************************/
void NAPI FLIPPED_BitBlt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix)
{
    DRAW2D.BitBlt(FLIP_RECTX(srcLeft,width),FLIP_RECTY(srcTop,height),width,height,FLIP_RECTX(dstLeft,width),FLIP_RECTY(dstTop,height),mix);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI FLIPPED_BitBltSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    N_int32 flipY)
{
    DC2D.beginDirectAccess();
    switch (DC.mi.BitsPerPixel) {
        case 8:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft;
            RotateBitmap8BPP(DC.buffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,width,height,flipY);
            break;
        case 15:
        case 16:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*2;
            RotateBitmap16BPP(DC.buffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,width,height,flipY);
            break;
        case 24:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*3;
            RotateBitmap24BPP(DC.buffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,width,height,flipY);
            break;
        case 32:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*4;
            RotateBitmap32BPP(DC.buffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,width,height,flipY);
            break;
        }
    DC2D.endDirectAccess();
    if (DC.VRAMBufferStartY)
        DRAW2D.BitBlt(0,DC.VRAMBufferStartY,width,height,FLIP_RECTX(dstLeft,width),FLIP_RECTY(dstTop,height),mix);
    else if (DC.VRAMBufferStart)
        DRAW2D.BitBltLin(DC.VRAMBufferStart,DC.mi.BytesPerScanLine,0,0,width,height,FLIP_RECTX(dstLeft,width),FLIP_RECTY(dstTop,height),mix);
    else
        DRAW2D.BitBltSys(DC.buffer,DC.mi.BytesPerScanLine,0,0,width,height,FLIP_RECTX(dstLeft,width),FLIP_RECTY(dstTop,height),mix,false);
}

/****************************************************************************
PARAMETERS:
srcLeft - Left coordinate of the source rectangle to copy
srcTop  - Top coordinate of the source rectangle to copy
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
dstLeft - Left coordinate of destination
dstTop  - Top coordinate of destination
rop3    - ROP3 mix code for the copy
****************************************************************************/
void NAPI FLIPPED_BitBltPatt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    DRAW2D.BitBltPatt(FLIP_RECTX(srcLeft,width),FLIP_RECTY(srcTop,height),width,height,FLIP_RECTX(dstLeft,width),FLIP_RECTY(dstTop,height),rop3);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3        - Mix code for the copy (GA_mixCodesType)
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI FLIPPED_BitBltPattSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3,
    N_int32 flipY)
{
    DC2D.beginDirectAccess();
    switch (DC.mi.BitsPerPixel) {
        case 8:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft;
            RotateBitmap8BPP(DC.buffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,width,height,flipY);
            break;
        case 15:
        case 16:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*2;
            RotateBitmap16BPP(DC.buffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,width,height,flipY);
            break;
        case 24:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*3;
            RotateBitmap24BPP(DC.buffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,width,height,flipY);
            break;
        case 32:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*4;
            RotateBitmap32BPP(DC.buffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,width,height,flipY);
            break;
        }
    DC2D.endDirectAccess();
    if (DC.VRAMBufferStartY)
        DRAW2D.BitBltPatt(0,DC.VRAMBufferStartY,width,height,FLIP_RECTX(dstLeft,width),FLIP_RECTY(dstTop,height),rop3);
    else if (DC.VRAMBufferStart)
        DRAW2D.BitBltPattLin(DC.VRAMBufferStart,DC.mi.BytesPerScanLine,0,0,width,height,FLIP_RECTX(dstLeft,width),FLIP_RECTY(dstTop,height),rop3);
    else
        DRAW2D.BitBltPattSys(DC.buffer,DC.mi.BytesPerScanLine,0,0,width,height,FLIP_RECTX(dstLeft,width),FLIP_RECTY(dstTop,height),rop3,false);
}

/****************************************************************************
PARAMETERS:
srcLeft - Left coordinate of the source rectangle to copy
srcTop  - Top coordinate of the source rectangle to copy
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
dstLeft - Left coordinate of destination
dstTop  - Top coordinate of destination
rop3    - ROP3 mix code for the copy
****************************************************************************/
void NAPI FLIPPED_BitBltColorPatt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    DRAW2D.BitBltColorPatt(FLIP_RECTX(srcLeft,width),FLIP_RECTY(srcTop,height),width,height,FLIP_RECTX(dstLeft,width),FLIP_RECTY(dstTop,height),rop3);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3        - Mix code for the copy (GA_mixCodesType)
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI FLIPPED_BitBltColorPattSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3,
    N_int32 flipY)
{
    DC2D.beginDirectAccess();
    switch (DC.mi.BitsPerPixel) {
        case 8:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft;
            RotateBitmap8BPP(DC.buffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,width,height,flipY);
            break;
        case 15:
        case 16:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*2;
            RotateBitmap16BPP(DC.buffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,width,height,flipY);
            break;
        case 24:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*3;
            RotateBitmap24BPP(DC.buffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,width,height,flipY);
            break;
        case 32:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*4;
            RotateBitmap32BPP(DC.buffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,width,height,flipY);
            break;
        }
    DC2D.endDirectAccess();
    if (DC.VRAMBufferStartY)
        DRAW2D.BitBltColorPatt(0,DC.VRAMBufferStartY,width,height,FLIP_RECTX(dstLeft,width),FLIP_RECTY(dstTop,height),rop3);
    else if (DC.VRAMBufferStart)
        DRAW2D.BitBltColorPattLin(DC.VRAMBufferStart,DC.mi.BytesPerScanLine,0,0,width,height,FLIP_RECTX(dstLeft,width),FLIP_RECTY(dstTop,height),rop3);
    else
        DRAW2D.BitBltColorPattSys(DC.buffer,DC.mi.BytesPerScanLine,0,0,width,height,FLIP_RECTX(dstLeft,width),FLIP_RECTY(dstTop,height),rop3,false);
}

/****************************************************************************
PARAMETERS:
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
****************************************************************************/
void NAPI FLIPPED_SrcTransBlt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    DRAW2D.SrcTransBlt(FLIP_RECTX(srcLeft,width),FLIP_RECTY(srcTop,height),width,height,FLIP_RECTX(dstLeft,width),FLIP_RECTY(dstTop,height),mix,transparent);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI FLIPPED_SrcTransBltSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent,
    N_int32 flipY)
{
    DC2D.beginDirectAccess();
    switch (DC.mi.BitsPerPixel) {
        case 8:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft;
            RotateBitmap8BPP(DC.buffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,width,height,flipY);
            break;
        case 15:
        case 16:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*2;
            RotateBitmap16BPP(DC.buffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,width,height,flipY);
            break;
        case 24:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*3;
            RotateBitmap24BPP(DC.buffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,width,height,flipY);
            break;
        case 32:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*4;
            RotateBitmap32BPP(DC.buffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,width,height,flipY);
            break;
        }
    DC2D.endDirectAccess();
    if (DC.VRAMBufferStartY)
        DRAW2D.SrcTransBlt(0,DC.VRAMBufferStartY,width,height,FLIP_RECTX(dstLeft,width),FLIP_RECTY(dstTop,height),mix,transparent);
    else if (DC.VRAMBufferStart)
        DRAW2D.SrcTransBltLin(DC.VRAMBufferStart,DC.mi.BytesPerScanLine,0,0,width,height,FLIP_RECTX(dstLeft,width),FLIP_RECTY(dstTop,height),mix,transparent);
    else
        DRAW2D.SrcTransBltSys(DC.buffer,DC.mi.BytesPerScanLine,0,0,width,height,FLIP_RECTX(dstLeft,width),FLIP_RECTY(dstTop,height),mix,transparent,false);
}

/****************************************************************************
PARAMETERS:
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
****************************************************************************/
void NAPI FLIPPED_DstTransBlt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    DRAW2D.DstTransBlt(FLIP_RECTX(srcLeft,width),FLIP_RECTY(srcTop,height),width,height,FLIP_RECTX(dstLeft,width),FLIP_RECTY(dstTop,height),mix,transparent);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI FLIPPED_DstTransBltSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent,
    N_int32 flipY)
{
    DC2D.beginDirectAccess();
    switch (DC.mi.BitsPerPixel) {
        case 8:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft;
            RotateBitmap8BPP(DC.buffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,width,height,flipY);
            break;
        case 15:
        case 16:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*2;
            RotateBitmap16BPP(DC.buffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,width,height,flipY);
            break;
        case 24:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*3;
            RotateBitmap24BPP(DC.buffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,width,height,flipY);
            break;
        case 32:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*4;
            RotateBitmap32BPP(DC.buffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,width,height,flipY);
            break;
        }
    DC2D.endDirectAccess();
    if (DC.VRAMBufferStartY)
        DRAW2D.DstTransBlt(0,DC.VRAMBufferStartY,width,height,FLIP_RECTX(dstLeft,width),FLIP_RECTY(dstTop,height),mix,transparent);
    else if (DC.VRAMBufferStart)
        DRAW2D.DstTransBltLin(DC.VRAMBufferStart,DC.mi.BytesPerScanLine,0,0,width,height,FLIP_RECTX(dstLeft,width),FLIP_RECTY(dstTop,height),mix,transparent);
    else
        DRAW2D.DstTransBltSys(DC.buffer,DC.mi.BytesPerScanLine,0,0,width,height,FLIP_RECTX(dstLeft,width),FLIP_RECTY(dstTop,height),mix,transparent,false);
}

/****************************************************************************
PARAMETERS:
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstWidth    - Width of the destination rectangle in pixels
dstHeight   - Height of the destination rectangle in scanlines
doClip      - True if the blit should be clipped
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle
mix         - Mix code for the copy (GA_mixCodesType)
****************************************************************************/
void NAPI FLIPPED_StretchBlt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    N_int32 doClip,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom,
    N_int32 mix)
{
    DRAW2D.StretchBlt(FLIP_RECTX(srcLeft,srcWidth),FLIP_RECTY(srcTop,srcHeight),srcWidth,srcHeight,FLIP_RECTX(dstLeft,dstWidth),FLIP_RECTY(dstTop,dstHeight),dstWidth,dstHeight,doClip,FLIPX(clipRight),FLIPY(clipBottom),FLIPX(clipLeft),FLIPY(clipTop),mix);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstWidth    - Width of the destination rectangle in pixels
dstHeight   - Height of the destination rectangle in scanlines
doClip      - True if the blit should be clipped
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle
mix         - Mix code for the copy (GA_mixCodesType)
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI FLIPPED_StretchBltSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    N_int32 doClip,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom,
    N_int32 mix,
    N_int32 flipY)
{
    switch (DC.mi.BitsPerPixel) {
        case 8:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft;
            RotateBitmap8BPP(DC.sysBuffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,srcWidth,srcHeight,flipY);
            break;
        case 15:
        case 16:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*2;
            RotateBitmap16BPP(DC.sysBuffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,srcWidth,srcHeight,flipY);
            break;
        case 24:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*3;
            RotateBitmap24BPP(DC.sysBuffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,srcWidth,srcHeight,flipY);
            break;
        case 32:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*4;
            RotateBitmap32BPP(DC.sysBuffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,srcWidth,srcHeight,flipY);
            break;
        }
    DRAW2D.StretchBltSys(DC.sysBuffer,DC.mi.BytesPerScanLine,0,0,srcWidth,srcHeight,FLIP_RECTX(dstLeft,dstWidth),FLIP_RECTY(dstTop,dstWidth),dstWidth,dstHeight,doClip,FLIPX(clipRight),FLIPY(clipBottom),FLIPX(clipLeft),FLIPY(clipTop),mix,false);
}

/****************************************************************************
PARAMETERS:
srcAddr         - Address of source bitmap in system memory
srcPitch        - Pitch of source bitmap in bytes
srcLeft         - Left coordinate within source bitmap to copy
srcTop          - Top coordinate within source bitmap to copy
width           - Width of the rectangle in pixels
height          - Height of the rectangle in scanlines
dstLeft         - Left coordinate of destination
dstTop          - Top coordinate of destination
bitsPerPixel    - Pixel depth for the bitmap to copy
pixelFormat     - Pixel format for the bitmap to copy
dstPal          - Color palette for the destination device (if dest is 8bpp)
srcPal          - Color palette for the bitmap to copy (if source is 8bpp)
dither          - True if the output should be dithered
mix             - Mix code for the copy (GA_mixCodesType)
flipY           - True if the bitmap should be flipped in Y

RETURNS:
0 on success, -1 on failure.
****************************************************************************/
N_int32 NAPI FLIPPED_ConvertBltSys_Obsolete(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pixelFormat,
    GA_palette *dstPal,
    GA_palette *srcPal,
    N_int32 dither,
    N_int32 mix,
    N_int32 flipY)
{
    switch (bitsPerPixel) {
        case 4:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft/2;
            RotateBitmap4BPP(DC.sysBuffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,width,height,flipY);
            break;
        case 8:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft;
            RotateBitmap8BPP(DC.sysBuffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,width,height,flipY);
            break;
        case 15:
        case 16:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*2;
            RotateBitmap16BPP(DC.sysBuffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,width,height,flipY);
            break;
        case 24:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*3;
            RotateBitmap24BPP(DC.sysBuffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,width,height,flipY);
            break;
        case 32:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*4;
            RotateBitmap32BPP(DC.sysBuffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,width,height,flipY);
            break;
        }
    return DRAW2D.ConvertBltSys_Obsolete(DC.sysBuffer,DC.mi.BytesPerScanLine,0,0,width,height,
        FLIP_RECTX(dstLeft,width),FLIP_RECTY(dstTop,width),bitsPerPixel,
        pixelFormat,dstPal,srcPal,dither,mix,false);
}

/****************************************************************************
PARAMETERS:
srcAddr         - Address of source bitmap in system memory
srcPitch        - Pitch of source bitmap in bytes
srcLeft         - Left coordinate of the source rectangle to copy
srcTop          - Top coordinate of the source rectangle to copy
srcWidth        - Width of the source rectangle in pixels
srcHeight       - Height of the source rectangle in scanlines
dstLeft         - Left coordinate of destination
dstTop          - Top coordinate of destination
dstWidth        - Width of the destination rectangle in pixels
dstHeight       - Height of the destination rectangle in scanlines
doClip          - True if the blit should be clipped
clipLeft        - Left coordinate of clip rectangle
clipTop         - Top coordinate of clip rectangle
clipRight       - Right coordinate of clip rectangle
clipBottom      - Bottom coordinate of clip rectangle
bitsPerPixel    - Pixel depth for the bitmap to copy
pixelFormat     - Pixel format for the bitmap to copy
dstPal          - Color palette for the destination device (if dest is 8bpp)
srcPal          - Color palette for the bitmap to copy (if source is 8bpp)
dither          - True if the output should be dithered
mix             - Mix code for the copy (GA_mixCodesType)
flipY           - True if the bitmap should be flipped in Y

RETURNS:
0 on success, -1 on failure.
****************************************************************************/
N_int32 NAPI FLIPPED_StretchConvertBltSys_Obsolete(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    N_int32 doClip,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pixelFormat,
    GA_palette *dstPal,
    GA_palette *srcPal,
    N_int32 dither,
    N_int32 mix,
    N_int32 flipY)
{
    switch (bitsPerPixel) {
        case 4:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft/2;
            RotateBitmap4BPP(DC.sysBuffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,srcWidth,srcHeight,flipY);
            break;
        case 8:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft;
            RotateBitmap8BPP(DC.sysBuffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,srcWidth,srcHeight,flipY);
            break;
        case 15:
        case 16:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*2;
            RotateBitmap16BPP(DC.sysBuffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,srcWidth,srcHeight,flipY);
            break;
        case 24:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*3;
            RotateBitmap24BPP(DC.sysBuffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,srcWidth,srcHeight,flipY);
            break;
        case 32:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*4;
            RotateBitmap32BPP(DC.sysBuffer,srcAddr,srcPitch,DC.mi.BytesPerScanLine,srcWidth,srcHeight,flipY);
            break;
        }
    return DRAW2D.StretchConvertBltSys_Obsolete(DC.sysBuffer,DC.mi.BytesPerScanLine,0,0,
        srcWidth,srcHeight,FLIP_RECTX(dstLeft,dstWidth),FLIP_RECTY(dstTop,dstHeight),dstWidth,dstHeight,
        doClip,FLIPX(clipRight),FLIPY(clipBottom),FLIPX(clipLeft),FLIPY(clipTop),bitsPerPixel,pixelFormat,dstPal,srcPal,dither,mix,false);
}

/****************************************************************************
PARAMETERS:
dstAddr     - Address of destination bitmap in system memory
dstPhysAddr - Physical address of destination bitmap in system memory
dstPitch    - Pitch of destination bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
****************************************************************************/
void NAPI FLIPPED_GetBitmapSys(
    void *dstAddr,
    N_int32 dstPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix)
{
//  void    *src;

    // TODO: This is wrong! We also need to handle mixes!!
    (void)srcLeft;
    (void)srcTop;
    (void)dstLeft;
    (void)dstTop;
    switch (DC.mi.BitsPerPixel) {
        case 8:
//          src = (uchar*)DC2D.surface + DC2D.mi.bytesPerLine * srcTop + srcLeft;
            RotateBitmap8BPP(DC.sysBuffer,dstAddr,dstPitch,dstPitch,width,height,false);
            break;
        case 15:
        case 16:
//          src = (uchar*)DC2D.surface + DC2D.mi.bytesPerLine * srcTop + srcLeft*2;
            RotateBitmap16BPP(DC.sysBuffer,dstAddr,dstPitch,dstPitch,width,height,false);
            break;
        case 24:
//          src = (uchar*)DC2D.surface + DC2D.mi.bytesPerLine * srcTop + srcLeft*3;
            RotateBitmap24BPP(DC.sysBuffer,dstAddr,dstPitch,dstPitch,width,height,false);
            break;
        case 32:
//          src = (uchar*)DC2D.surface + DC2D.mi.bytesPerLine * srcTop + srcLeft*4;
            RotateBitmap32BPP(DC.sysBuffer,dstAddr,dstPitch,dstPitch,width,height,false);
            break;
        }
}

/****************************************************************************
PARAMETERS:
ileft   - Left coordinate of ellipse rectangle
itop    - Top coordinate of ellipse rectangle
A       - Major axis dimension
B       - Minor axis dimension
****************************************************************************/
void NAPI FLIPPED_DrawEllipse(
    N_int32 left,
    N_int32 top,
    N_int32 A,
    N_int32 B)
{
    DRAW2D.DrawEllipse(FLIP_RECTX(left,A),FLIP_RECTY(top,B),A,B);
}

/****************************************************************************
PARAMETERS:
left        - Left coordinate of ellipse rectangle
top         - Top coordinate of ellipse rectangle
A           - Major axis dimension
B           - Minor axis dimension
clipLeft    - Left coordinate for clip rectangle
clipTop     - Top coordinate for clip rectangle
clipRight   - Right coordinate for clip rectangle
clipBottom  - Bottom coordinate for clip rectangle
****************************************************************************/
void NAPI FLIPPED_ClipEllipse(
    N_int32 left,
    N_int32 top,
    N_int32 A,
    N_int32 B,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
    DRAW2D.ClipEllipse(FLIP_RECTX(left,A),FLIP_RECTY(top,B),A,B,FLIPX(clipRight),FLIPY(clipBottom),FLIPX(clipLeft),FLIPY(clipTop));
}

/****************************************************************************
PARAMETERS:
x   - X coordinate to read pixel from
y   - Y coordinate to read pixel from

RETURNS:
Color read from the pixel on the screen
****************************************************************************/
GA_color NAPI FLIPPED_GetPixel(
    N_int32 x,
    N_int32 y)
{
    return DRAW2D.GetPixel(FLIPX(x),FLIPY(y));
}

/****************************************************************************
PARAMETERS:
x   - X coordinate to plot pixel at
y   - Y coordinate to plot pixel at
****************************************************************************/
void NAPI FLIPPED_PutPixel(
    N_int32 x,
    N_int32 y)
{
    DRAW2D.PutPixel(FLIPX(x),FLIPY(y));
}

/****************************************************************************
PARAMETERS:
cursor      - Pointer to an GA_monoCursor structure for the cursor image

REMARKS:
This function downloads the specified monochrome cursor definition from the
application into the hardware cursor. The cursor data is passed by the
application in the GA_monoCursor structure.
****************************************************************************/
void NAPI FLIPPED_SetMonoCursor(
    GA_monoCursor *cursor)
{
    if (DC.enabled) {
        DC.cursorHotX = cursor->HotX;
        DC.cursorHotY = cursor->HotY;
        DC.monoCursor.HotX = DC.monoCursor.HotY = 0;
        RotateBitmap1BPP(DC.monoCursor.XORMask,cursor->XORMask,8,64);
        RotateBitmap1BPP(DC.monoCursor.ANDMask,cursor->ANDMask,8,64);
        FLIPPED_cursorREF.SetMonoCursor(&DC.monoCursor);
        }
    else
        FLIPPED_cursorREF.SetMonoCursor(cursor);
}

/****************************************************************************
PARAMETERS:
cursor  - Pointer to an GA_colorCursor structure for the cursor image

REMARKS:
This function downloads the specified colorCurs cursor definition from the
application into the hardware cursor.
****************************************************************************/
void NAPI FLIPPED_SetColorCursor(
    GA_colorCursor *cursor)
{
    if (DC.enabled) {
        DC.cursorHotX = cursor->HotX;
        DC.cursorHotY = cursor->HotY;
        DC.colorCursor.HotX = DC.colorCursor.HotY = 0;
        memcpy(DC.colorCursor.Palette,cursor->Palette,sizeof(DC.colorCursor.Palette));
        RotateBitmap4BPP(DC.colorCursor.ColorData,cursor->ColorData,32,32,64,64,false);
        RotateBitmap1BPP(DC.colorCursor.ANDMask,cursor->ANDMask,8,64);
        FLIPPED_cursorREF.SetColorCursor(&DC.colorCursor);
        }
    else
        FLIPPED_cursorREF.SetColorCursor(cursor);
}

/****************************************************************************
PARAMETERS:
cursor  - Pointer to an GA_colorCursor256 structure for the cursor image

REMARKS:
This function downloads the specified colorCurs cursor definition from the
application into the hardware cursor.
****************************************************************************/
void NAPI FLIPPED_SetColorCursor256(
    GA_colorCursor256 *cursor)
{
    if (DC.enabled) {
        DC.cursorHotX = cursor->HotX;
        DC.cursorHotY = cursor->HotY;
        DC.colorCursor256.HotX = DC.colorCursor256.HotY = 0;
        memcpy(DC.colorCursor256.Palette,cursor->Palette,sizeof(DC.colorCursor256.Palette));
        RotateBitmap8BPP(DC.colorCursor256.ColorData,cursor->ColorData,64,64,64,64,false);
        RotateBitmap1BPP(DC.colorCursor256.ANDMask,cursor->ANDMask,8,64);
        FLIPPED_cursorREF.SetColorCursor256(&DC.colorCursor256);
        }
    else
        FLIPPED_cursorREF.SetColorCursor256(cursor);
}

/****************************************************************************
PARAMETERS:
cursor  - Pointer to an GA_colorCursorRGB structure for the cursor image

REMARKS:
This function downloads the specified colorCurs cursor definition from the
application into the hardware cursor.
****************************************************************************/
void NAPI FLIPPED_SetColorCursorRGB(
    GA_colorCursorRGB *cursor)
{
    if (DC.enabled) {
        DC.cursorHotX = cursor->HotX;
        DC.cursorHotY = cursor->HotY;
        DC.colorCursorRGB.HotX = DC.colorCursorRGB.HotY = 0;
        RotateBitmap24BPP(DC.colorCursorRGB.ColorData,cursor->ColorData,192,192,64,64,false);
        RotateBitmap1BPP(DC.colorCursorRGB.ANDMask,cursor->ANDMask,8,64);
        FLIPPED_cursorREF.SetColorCursorRGB(&DC.colorCursorRGB);
        }
    else
        FLIPPED_cursorREF.SetColorCursorRGB(cursor);
}

/****************************************************************************
PARAMETERS:
cursor  - Pointer to an GA_colorCursorRGBA structure for the cursor image

REMARKS:
This function downloads the specified colorCurs cursor definition from the
application into the hardware cursor.
****************************************************************************/
void NAPI FLIPPED_SetColorCursorRGBA(
    GA_colorCursorRGBA *cursor)
{
    if (DC.enabled) {
        DC.cursorHotX = cursor->HotX;
        DC.cursorHotY = cursor->HotY;
        DC.colorCursorRGBA.HotX = DC.colorCursorRGBA.HotY = 0;
        RotateBitmap32BPP(DC.colorCursorRGBA.ColorData,cursor->ColorData,256,256,64,64,false);
        FLIPPED_cursorREF.SetColorCursorRGBA(&DC.colorCursorRGBA);
        }
    else
        FLIPPED_cursorREF.SetColorCursorRGBA(cursor);
}

/****************************************************************************
PARAMETERS:
x   - X coordinate of the cursor position
y   - Y coordinate of the cursor position

REMARKS:
This function sets the location of the hardware cursor in display
coordinates. This function takes the X and Y coordinates of the new cursor
location. This function will place the cursor so that the hotspot of the
cursor image is located at the specified (X,Y) location, and will correctly
handle special cases where the cursor definition needs to be located off the
edges of the display screen (such as when X=0, Y=0 and HotX,HotY > 0).
****************************************************************************/
N_int32 NAPI FLIPPED_SetCursorPos(
    N_int32 x,
    N_int32 y)
{
    if (DC.enabled) {
        x -= DC.cursorHotX;
        y -= DC.cursorHotY;
        x = FLIP_RECTX(x,64);
        y = FLIP_RECTY(y,64);
        return FLIPPED_cursorREF.SetCursorPos(x,y);
        }
    return FLIPPED_cursorREF.SetCursorPos(x,y);
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of display affected
top     - Top coordinate of display affected
right   - Right coordinate of display affected (exclusive)
bottom  - Bottom coordinate of display affected (exclusive)

REMARKS:
This function is used when the SNAP driver is using software monoCurs and
colorCurs cursor acceleration. The cursor images are stored in offscreen
memory buffers, and the hardware blitter is used to draw the cursor images.
This function is used to inform the SNAP driver that the shell display
driver is about to draw to a particular location of memory, and to remove
the cursor image from the screen temporarily while the drawing is taking
place if the drawing region and the cursor position overlap.
****************************************************************************/
void NAPI FLIPPED_BeginAccess(
    N_int32 left,
    N_int32 top,
    N_int32 right,
    N_int32 bottom)
{
    if (DC.enabled)
        FLIPPED_cursorREF.BeginAccess(FLIPX(right),FLIPY(bottom),FLIPX(left),FLIPY(top));
    else
        FLIPPED_cursorREF.BeginAccess(left,top,right,bottom);
}
