/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Simple program to generate ROP3 pixel plotting functions
*               given a reverse polish ROP3 code pneumonic.
*
****************************************************************************/

//#define BPP_1
#define BPP_4
//#define BPP_24

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tcl/cstack.hpp"

typedef struct {
    char    s[100];
    } Expr;

TCCStack<Expr>  stack(20);

int genr3(char *p)
{
    char    *operands = "DSP";
    Expr    expr,op1,op2;
    ibool   haveD = false;
    ibool   haveS = false;
    ibool   haveP = false;

#if defined(BPP_1)
    printf("void REF2D_drawPixelR3_1_%s(uchar *addr,int x,GA_color S,GA_color P)\n", p);
#elif defined(BPP_4)
    printf("void REF2D_drawPixelR3_4_%s(uchar *addr,int x,GA_color S,GA_color P)\n", p);
#elif defined(BPP_24)
    printf("void REF2D_drawPixelR3_24_%s(uchar *addr DCL_X_ARG,GA_color S,GA_color P)\n", p);
#else
    printf("DRAW_PIXEL_PROTO(_%s)\n", p);
#endif
    printf("{\n");
    while (*p) {
        if (strchr(operands,*p) == NULL) {
            // Handle operator code
            switch (*p) {
                case 'a':
                    op2 = stack.pop();
                    op1 = stack.pop();
                    sprintf(expr.s,"(%s & %s)", op1.s, op2.s);
                    stack.push(expr);
                    break;
                case 'o':
                    op2 = stack.pop();
                    op1 = stack.pop();
                    sprintf(expr.s,"(%s | %s)", op1.s, op2.s);
                    stack.push(expr);
                    break;
                case 'x':
                    op2 = stack.pop();
                    op1 = stack.pop();
                    sprintf(expr.s,"(%s ^ %s)", op1.s, op2.s);
                    stack.push(expr);
                    break;
                case 'n':
                    op1 = stack.pop();
                    sprintf(expr.s,"~%s", op1.s);
                    stack.push(expr);
                    break;
                default:
                    printf("\nSyntax error parsing ROP3 code!\n");
                    return -1;
                }
            }
        else {
            // Push operand on the stack
            if (*p == 'D')
                haveD = true;
            else if (*p == 'S')
                haveS = true;
            else if (*p == 'P')
                haveP = true;
            expr.s[0] = *p;
            expr.s[1] = 0;
            stack.push(expr);
            }
        p++;
        }
#if defined(BPP_1)
    printf("    GA_color mask = MONO_MASK(x);\n");
    printf("    GA_color dest = *addr;\n");
    if (haveD)
        printf("    GA_color D = (dest & mask);\n");
    if (haveS)
        printf("    S = (S ? mask : 0);\n");
    if (haveP)
        printf("    P = (P ? mask : 0);\n");
    printf("    S = ");
    expr = stack.pop();
    printf("%s;\n", expr.s);
    printf("    *addr = (uchar)((dest & ~mask) | (S & mask));\n");
#elif defined(BPP_4)
    printf("    GA_color dest = *addr;\n");
    if (haveD)
        printf("    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);\n");
    printf("    S = ");
    expr = stack.pop();
    printf("%s;\n", expr.s);
    printf("    if (x & 1)\n");
    printf("        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));\n");
    printf("    else\n");
    printf("        *addr = (uchar)((dest & 0x0F) | (S << 4));\n");
#elif defined(BPP_24)
    printf("    GA_color D;\n");
    if (haveD) {
        printf("    BYTE0(D) = addr[0];\n");
        printf("    BYTE1(D) = addr[1];\n");
        printf("    BYTE2(D) = addr[2];\n");
        }
    printf("    D = ");
    expr = stack.pop();
    printf("%s;\n", expr.s);
    printf("    addr[0] = BYTE0(D);\n");
    printf("    addr[1] = BYTE1(D);\n");
    printf("    addr[2] = BYTE2(D);\n");
#else
    if (haveD)
        printf("    GA_color D = *dest;\n");
    printf("    *dest = (COLOR_TYPE)");
    expr = stack.pop();
    printf("%s;\n", expr.s);
#endif
    if (!haveS)
        printf("    (void)S;\n");
    if (!haveP)
        printf("    (void)P;\n");
    printf("}\n\n");
    return 0;
}

int main(int argc,char *argv[])
{
    if (argc != 2) {
        printf("Usage: genr3 <rop_code>\n");
        printf("\n");
        printf("Where <rop_code> is the code to generate the ROP3 code for.\n");
        printf("Use a value 'all' to generate all ROP3 codes.\n");
        return -1;
        }
    if (stricmp(argv[1],"all") == 0) {
        genr3("DPSoon");
        genr3("DPSona");
        genr3("PSon");
        genr3("SDPona");
        genr3("DPon");
        genr3("PDSxnon");
        genr3("PDSaon");
        genr3("SDPnaa");
        genr3("PDSxon");
        genr3("DPna");
        genr3("PSDnaon");
        genr3("SPna");
        genr3("PDSnaon");
        genr3("PDSonon");
        genr3("Pn");
        genr3("PDSona");
        genr3("DSon");
        genr3("SDPxnon");
        genr3("SDPaon");
        genr3("DPSxnon");
        genr3("DPSaon");
        genr3("PSDPSanaxx");
        genr3("SSPxDSxaxn");
        genr3("SPxPDxa");
        genr3("SDPSanaxn");
        genr3("PDSPaox");
        genr3("SDPSxaxn");
        genr3("PSDPaox");
        genr3("DSPDxaxn");
        genr3("PDSox");
        genr3("PDSoan");
        genr3("DPSnaa");
        genr3("SDPxon");
        genr3("DSna");
        genr3("SPDnaon");
        genr3("SPxDSxa");
        genr3("PDSPanaxn");
        genr3("SDPSaox");
        genr3("SDPSxnox");
        genr3("DPSxa");
        genr3("PSDPSaoxxn");
        genr3("DPSana");
        genr3("SSPxPDxaxn");
        genr3("SPDSoax");
        genr3("PSDnox");
        genr3("PSDPxox");
        genr3("PSDnoan");
        genr3("PSna");
        genr3("SDPnaon");
        genr3("SDPSoox");
        genr3("Sn");
        genr3("SPDSaox");
        genr3("SPDSxnox");
        genr3("SDPox");
        genr3("SDPoan");
        genr3("PSDPoax");
        genr3("SPDnox");
        genr3("SPDSxox");
        genr3("SPDnoan");
        genr3("PSx");
        genr3("SPDSonox");
        genr3("SPDSnaox");
        genr3("PSan");
        genr3("PSDnaa");
        genr3("DPSxon");
        genr3("SDxPDxa");
        genr3("SPDSanaxn");
        genr3("SDna");
        genr3("DPSnaon");
        genr3("DSPDaox");
        genr3("PSDPxaxn");
        genr3("SDPxa");
        genr3("PDSPDaoxxn");
        genr3("DPSDoax");
        genr3("PDSnox");
        genr3("SDPana");
        genr3("SSPxDSxoxn");
        genr3("PDSPxox");
        genr3("PDSnoan");
        genr3("PDna");
        genr3("DSPnaon");
        genr3("DPSDaox");
        genr3("SPDSxaxn");
        genr3("DPSonon");
        genr3("Dn");
        genr3("DPSox");
        genr3("DPSoan");
        genr3("PDSPoax");
        genr3("DPSnox");
        genr3("DPx");
        genr3("DPSDonox");
        genr3("DPSDxox");
        genr3("DPSnoan");
        genr3("DPSDnaox");
        genr3("DPan");
        genr3("PDSxa");
        genr3("DSPDSaoxxn");
        genr3("DSPDoax");
        genr3("SDPnox");
        genr3("SDPSoax");
        genr3("DSPnox");
        genr3("DSx");
        genr3("SDPSonox");
        genr3("DSPDSonoxxn");
        genr3("PDSxxn");
        genr3("DPSax");
        genr3("PSDPSoaxxn");
        genr3("SDPax");
        genr3("PDSPDoaxxn");
        genr3("SDPSnoax");
        genr3("PDSxnan");
        genr3("PDSana");
        genr3("SSDxPDxaxn");
        genr3("SDPSxox");
        genr3("SDPnoan");
        genr3("DSPDxox");
        genr3("DSPnoan");
        genr3("SDPSnaox");
        genr3("DSan");
        genr3("PDSax");
        genr3("DSPDSoaxxn");
        genr3("DPSDnoax");
        genr3("SDPxnan");
        genr3("SPDSnoax");
        genr3("DPSxnan");
        genr3("SPxDSxo");
        genr3("DPSaan");
        genr3("DPSaa");
        genr3("SPxDSxon");
        genr3("DPSxna");
        genr3("SPDSnoaxn");
        genr3("SDPxna");
        genr3("PDSPnoaxn");
        genr3("DSPDSoaxx");
        genr3("PDSaxn");
        genr3("DSa");
        genr3("SDPSnaoxn");
        genr3("DSPnoa");
        genr3("DSPDxoxn");
        genr3("SDPnoa");
        genr3("SDPSxoxn");
        genr3("SSDxPDxax");
        genr3("PDSanan");
        genr3("PDSxna");
        genr3("SDPSnoaxn");
        genr3("DPSDPoaxx");
        genr3("SPDaxn");
        genr3("PSDPSoaxx");
        genr3("DPSaxn");
        genr3("DPSxx");
        genr3("PSDPSonoxx");
        genr3("SDPSonoxn");
        genr3("DSxn");
        genr3("DPSnax");
        genr3("SDPSoaxn");
        genr3("SPDnax");
        genr3("DSPDoaxn");
        genr3("DSPDSaoxx");
        genr3("PDSxan");
        genr3("DPa");
        genr3("PDSPnaoxn");
        genr3("DPSnoa");
        genr3("DPSDxoxn");
        genr3("PDSPonoxn");
        genr3("PDxn");
        genr3("DSPnax");
        genr3("PDSPoaxn");
        genr3("DPSoa");
        genr3("DPSoxn");
        genr3("D");
        genr3("DPSono");
        genr3("SPDSxax");
        genr3("DPSDaoxn");
        genr3("DSPnao");
        genr3("DPno");
        genr3("PDSnoa");
        genr3("PDSPxoxn");
        genr3("SSPxDSxox");
        genr3("SDPanan");
        genr3("PSDnax");
        genr3("DPSDoaxn");
        genr3("DPSDPaoxx");
        genr3("SDPxan");
        genr3("PSDPxax");
        genr3("DSPDaoxn");
        genr3("DPSnao");
        genr3("DSno");
        genr3("SPDSanax");
        genr3("SDxPDxan");
        genr3("DPSxo");
        genr3("DPSano");
        genr3("PSa");
        genr3("SPDSnaoxn");
        genr3("SPDSonoxn");
        genr3("PSxn");
        genr3("SPDnoa");
        genr3("SPDSxoxn");
        genr3("SDPnax");
        genr3("PSDPoaxn");
        genr3("SDPoa");
        genr3("SPDoxn");
        genr3("DPSDxax");
        genr3("SPDSaoxn");
        genr3("S");
        genr3("SDPono");
        genr3("SDPnao");
        genr3("SPno");
        genr3("PSDnoa");
        genr3("PSDPxoxn");
        genr3("PDSnax");
        genr3("SPDSoaxn");
        genr3("SSPxPDxax");
        genr3("DPSanan");
        genr3("PSDPSaoxx");
        genr3("DPSxan");
        genr3("PDSPxax");
        genr3("SDPSaoxn");
        genr3("DPSDanax");
        genr3("SPxDSxan");
        genr3("SPDnao");
        genr3("SDno");
        genr3("SDPxo");
        genr3("SDPano");
        genr3("PDSoa");
        genr3("PDSoxn");
        genr3("DSPDxax");
        genr3("PSDPaoxn");
        genr3("SDPSxax");
        genr3("PDSPaoxn");
        genr3("SDPSanax");
        genr3("SPxPDxan");
        genr3("SSPxDSxax");
        genr3("DSPDSanaxxn");
        genr3("DPSao");
        genr3("DPSxno");
        genr3("SDPao");
        genr3("SDPxno");
        genr3("DSo");
        genr3("SDPnoo");
        genr3("P");
        genr3("PDSono");
        genr3("PDSnao");
        genr3("PSno");
        genr3("PSDnao");
        genr3("PDno");
        genr3("PDSxo");
        genr3("PDSano");
        genr3("PDSao");
        genr3("PDSxno");
        genr3("DPo");
        genr3("DPSnoo");
        genr3("PSo");
        genr3("PSDnoo");
        genr3("DPSoo");
        return 0;
        }
    else
        return genr3(argv[1]);
}
