/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Simple program to generate ROP3 pixel plotting functions
*               given a reverse polish ROP3 code pneumonic.
*
****************************************************************************/

#define BPP_4
//#define BPP_24

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tcl/cstack.hpp"

typedef struct {
    char    s[100];
    } Expr;

TCCStack<Expr>  stack(20);

int genr3m(char *p)
{
    char    *operands = "DSP";
    Expr    expr,op1,op2;
    ibool   haveD = false;
    ibool   haveS = false;
    ibool   haveP = false;

#if defined(BPP_4)
    printf("void REF2D_drawPixelR3Masked_4_%s(uchar *addr,int x,GA_color S,GA_color P)\n", p);
#elif defined(BPP_24)
    printf("void REF2D_drawPixelR3Masked_24_%s(uchar *addr DCL_X_ARG,GA_color S,GA_color P)\n", p);
#else
    printf("DRAW_PIXEL_PROTO(_%s)\n", p);
#endif
    printf("{\n");
    while (*p) {
        if (strchr(operands,*p) == NULL) {
            // Handle operator code
            switch (*p) {
                case 'a':
                    op2 = stack.pop();
                    op1 = stack.pop();
                    sprintf(expr.s,"(%s & %s)", op1.s, op2.s);
                    stack.push(expr);
                    break;
                case 'o':
                    op2 = stack.pop();
                    op1 = stack.pop();
                    sprintf(expr.s,"(%s | %s)", op1.s, op2.s);
                    stack.push(expr);
                    break;
                case 'x':
                    op2 = stack.pop();
                    op1 = stack.pop();
                    sprintf(expr.s,"(%s ^ %s)", op1.s, op2.s);
                    stack.push(expr);
                    break;
                case 'n':
                    op1 = stack.pop();
                    sprintf(expr.s,"~%s", op1.s);
                    stack.push(expr);
                    break;
                default:
                    printf("\nSyntax error parsing ROP3 code!\n");
                    return -1;
                }
            }
        else {
            // Push operand on the stack
            if (*p == 'D')
                haveD = true;
            else if (*p == 'S')
                haveS = true;
            else if (*p == 'P')
                haveP = true;
            expr.s[0] = *p;
            expr.s[1] = 0;
            stack.push(expr);
            }
        p++;
        }
#if defined(BPP_4)
    printf("    GA_color dest = *addr;\n");
    printf("    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);\n");
    printf("    D = ((D & DC.planeMaskInv) | (");
    expr = stack.pop();
    printf("%s & DC.planeMask));\n", expr.s);
    printf("    if (x & 1)\n");
    printf("        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));\n");
    printf("    else\n");
    printf("        *addr = (uchar)((dest & 0x0F) | (D << 4));\n");
#elif defined(BPP_24)
    printf("    GA_color D;\n");
    printf("    BYTE0(D) = addr[0];\n");
    printf("    BYTE1(D) = addr[1];\n");
    printf("    BYTE2(D) = addr[2];\n");
    printf("    D = ((D & DC.planeMaskInv) | (");
    expr = stack.pop();
    printf("%s & DC.planeMask));\n", expr.s);
    printf("    addr[0] = BYTE0(D);\n");
    printf("    addr[1] = BYTE1(D);\n");
    printf("    addr[2] = BYTE2(D);\n");
#else
    printf("    GA_color D = *dest;\n");
    printf("    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (");
    expr = stack.pop();
    printf("%s & DC.planeMask));\n", expr.s);
#endif
    if (!haveS)
        printf("    (void)S;\n");
    if (!haveP)
        printf("    (void)P;\n");
    printf("}\n\n");
    return 0;
}

int main(int argc,char *argv[])
{
    if (argc != 2) {
        printf("Usage: genr3m <rop_code>\n");
        printf("\n");
        printf("Where <rop_code> is the code to generate the ROP3 code for.\n");
        printf("Use a value 'all' to generate all ROP3 codes.\n");
        return -1;
        }
    if (stricmp(argv[1],"all") == 0) {
        genr3m("DPSoon");
        genr3m("DPSona");
        genr3m("PSon");
        genr3m("SDPona");
        genr3m("DPon");
        genr3m("PDSxnon");
        genr3m("PDSaon");
        genr3m("SDPnaa");
        genr3m("PDSxon");
        genr3m("DPna");
        genr3m("PSDnaon");
        genr3m("SPna");
        genr3m("PDSnaon");
        genr3m("PDSonon");
        genr3m("Pn");
        genr3m("PDSona");
        genr3m("DSon");
        genr3m("SDPxnon");
        genr3m("SDPaon");
        genr3m("DPSxnon");
        genr3m("DPSaon");
        genr3m("PSDPSanaxx");
        genr3m("SSPxDSxaxn");
        genr3m("SPxPDxa");
        genr3m("SDPSanaxn");
        genr3m("PDSPaox");
        genr3m("SDPSxaxn");
        genr3m("PSDPaox");
        genr3m("DSPDxaxn");
        genr3m("PDSox");
        genr3m("PDSoan");
        genr3m("DPSnaa");
        genr3m("SDPxon");
        genr3m("DSna");
        genr3m("SPDnaon");
        genr3m("SPxDSxa");
        genr3m("PDSPanaxn");
        genr3m("SDPSaox");
        genr3m("SDPSxnox");
        genr3m("DPSxa");
        genr3m("PSDPSaoxxn");
        genr3m("DPSana");
        genr3m("SSPxPDxaxn");
        genr3m("SPDSoax");
        genr3m("PSDnox");
        genr3m("PSDPxox");
        genr3m("PSDnoan");
        genr3m("PSna");
        genr3m("SDPnaon");
        genr3m("SDPSoox");
        genr3m("Sn");
        genr3m("SPDSaox");
        genr3m("SPDSxnox");
        genr3m("SDPox");
        genr3m("SDPoan");
        genr3m("PSDPoax");
        genr3m("SPDnox");
        genr3m("SPDSxox");
        genr3m("SPDnoan");
        genr3m("PSx");
        genr3m("SPDSonox");
        genr3m("SPDSnaox");
        genr3m("PSan");
        genr3m("PSDnaa");
        genr3m("DPSxon");
        genr3m("SDxPDxa");
        genr3m("SPDSanaxn");
        genr3m("SDna");
        genr3m("DPSnaon");
        genr3m("DSPDaox");
        genr3m("PSDPxaxn");
        genr3m("SDPxa");
        genr3m("PDSPDaoxxn");
        genr3m("DPSDoax");
        genr3m("PDSnox");
        genr3m("SDPana");
        genr3m("SSPxDSxoxn");
        genr3m("PDSPxox");
        genr3m("PDSnoan");
        genr3m("PDna");
        genr3m("DSPnaon");
        genr3m("DPSDaox");
        genr3m("SPDSxaxn");
        genr3m("DPSonon");
        genr3m("Dn");
        genr3m("DPSox");
        genr3m("DPSoan");
        genr3m("PDSPoax");
        genr3m("DPSnox");
        genr3m("DPx");
        genr3m("DPSDonox");
        genr3m("DPSDxox");
        genr3m("DPSnoan");
        genr3m("DPSDnaox");
        genr3m("DPan");
        genr3m("PDSxa");
        genr3m("DSPDSaoxxn");
        genr3m("DSPDoax");
        genr3m("SDPnox");
        genr3m("SDPSoax");
        genr3m("DSPnox");
        genr3m("DSx");
        genr3m("SDPSonox");
        genr3m("DSPDSonoxxn");
        genr3m("PDSxxn");
        genr3m("DPSax");
        genr3m("PSDPSoaxxn");
        genr3m("SDPax");
        genr3m("PDSPDoaxxn");
        genr3m("SDPSnoax");
        genr3m("PDSxnan");
        genr3m("PDSana");
        genr3m("SSDxPDxaxn");
        genr3m("SDPSxox");
        genr3m("SDPnoan");
        genr3m("DSPDxox");
        genr3m("DSPnoan");
        genr3m("SDPSnaox");
        genr3m("DSan");
        genr3m("PDSax");
        genr3m("DSPDSoaxxn");
        genr3m("DPSDnoax");
        genr3m("SDPxnan");
        genr3m("SPDSnoax");
        genr3m("DPSxnan");
        genr3m("SPxDSxo");
        genr3m("DPSaan");
        genr3m("DPSaa");
        genr3m("SPxDSxon");
        genr3m("DPSxna");
        genr3m("SPDSnoaxn");
        genr3m("SDPxna");
        genr3m("PDSPnoaxn");
        genr3m("DSPDSoaxx");
        genr3m("PDSaxn");
        genr3m("DSa");
        genr3m("SDPSnaoxn");
        genr3m("DSPnoa");
        genr3m("DSPDxoxn");
        genr3m("SDPnoa");
        genr3m("SDPSxoxn");
        genr3m("SSDxPDxax");
        genr3m("PDSanan");
        genr3m("PDSxna");
        genr3m("SDPSnoaxn");
        genr3m("DPSDPoaxx");
        genr3m("SPDaxn");
        genr3m("PSDPSoaxx");
        genr3m("DPSaxn");
        genr3m("DPSxx");
        genr3m("PSDPSonoxx");
        genr3m("SDPSonoxn");
        genr3m("DSxn");
        genr3m("DPSnax");
        genr3m("SDPSoaxn");
        genr3m("SPDnax");
        genr3m("DSPDoaxn");
        genr3m("DSPDSaoxx");
        genr3m("PDSxan");
        genr3m("DPa");
        genr3m("PDSPnaoxn");
        genr3m("DPSnoa");
        genr3m("DPSDxoxn");
        genr3m("PDSPonoxn");
        genr3m("PDxn");
        genr3m("DSPnax");
        genr3m("PDSPoaxn");
        genr3m("DPSoa");
        genr3m("DPSoxn");
        genr3m("D");
        genr3m("DPSono");
        genr3m("SPDSxax");
        genr3m("DPSDaoxn");
        genr3m("DSPnao");
        genr3m("DPno");
        genr3m("PDSnoa");
        genr3m("PDSPxoxn");
        genr3m("SSPxDSxox");
        genr3m("SDPanan");
        genr3m("PSDnax");
        genr3m("DPSDoaxn");
        genr3m("DPSDPaoxx");
        genr3m("SDPxan");
        genr3m("PSDPxax");
        genr3m("DSPDaoxn");
        genr3m("DPSnao");
        genr3m("DSno");
        genr3m("SPDSanax");
        genr3m("SDxPDxan");
        genr3m("DPSxo");
        genr3m("DPSano");
        genr3m("PSa");
        genr3m("SPDSnaoxn");
        genr3m("SPDSonoxn");
        genr3m("PSxn");
        genr3m("SPDnoa");
        genr3m("SPDSxoxn");
        genr3m("SDPnax");
        genr3m("PSDPoaxn");
        genr3m("SDPoa");
        genr3m("SPDoxn");
        genr3m("DPSDxax");
        genr3m("SPDSaoxn");
        genr3m("S");
        genr3m("SDPono");
        genr3m("SDPnao");
        genr3m("SPno");
        genr3m("PSDnoa");
        genr3m("PSDPxoxn");
        genr3m("PDSnax");
        genr3m("SPDSoaxn");
        genr3m("SSPxPDxax");
        genr3m("DPSanan");
        genr3m("PSDPSaoxx");
        genr3m("DPSxan");
        genr3m("PDSPxax");
        genr3m("SDPSaoxn");
        genr3m("DPSDanax");
        genr3m("SPxDSxan");
        genr3m("SPDnao");
        genr3m("SDno");
        genr3m("SDPxo");
        genr3m("SDPano");
        genr3m("PDSoa");
        genr3m("PDSoxn");
        genr3m("DSPDxax");
        genr3m("PSDPaoxn");
        genr3m("SDPSxax");
        genr3m("PDSPaoxn");
        genr3m("SDPSanax");
        genr3m("SPxPDxan");
        genr3m("SSPxDSxax");
        genr3m("DSPDSanaxxn");
        genr3m("DPSao");
        genr3m("DPSxno");
        genr3m("SDPao");
        genr3m("SDPxno");
        genr3m("DSo");
        genr3m("SDPnoo");
        genr3m("P");
        genr3m("PDSono");
        genr3m("PDSnao");
        genr3m("PSno");
        genr3m("PSDnao");
        genr3m("PDno");
        genr3m("PDSxo");
        genr3m("PDSano");
        genr3m("PDSao");
        genr3m("PDSxno");
        genr3m("DPo");
        genr3m("DPSnoo");
        genr3m("PSo");
        genr3m("PSDnoo");
        genr3m("DPSoo");
        return 0;
        }
    else
        return genr3m(argv[1]);
}
