/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the color inversion support
*               routines in the SNAP Reference Rasteriser.
*
****************************************************************************/

#include "invint.h"

// TODO: Clipped lines!

// TODO: BitBltMasked!

/*--------------------------- Global variables ----------------------------*/

INVERT_state    INVERT_devCtx;

REF2D_driver INVERT_drv = {
    sizeof(INVERT_drv),
    INVERT_QueryFunctions,
    INVERT_SetDrawBufferSW,
    INVERT_SetDrawSurface,
    INVERT_RotateBitmap,
    INVERT_UnloadDriver,
    INVERT_InitFilter,
    INVERT_DeInitFilter,
    INVERT_GlobalOptionsUpdated,

    INVERT_GetDrawBuffer,
    INVERT_DrawRectExtSW,
    INVERT_ForceSoftwareOnly,
    INVERT_SetColorCompareMask,
    INVERT_PostSwitchPhysicalResolution,
    };

static GA_initFuncs INVERT_initPT = {
    sizeof(GA_initFuncs),
    NULL,                       // INVERT_GetConfigInfo,
    NULL,                       // Reserved
    NULL,                       // INVERT_SetModeProfile,
    NULL,                       // INVERT_GetOptions,
    INVERT_SetOptions,
    NULL,                       // INVERT_GetVideoModeInfo,
    NULL,                       // INVERT_SetVideoMode,
    NULL,                       // INVERT_GetVideoMode,
    NULL,                       // INVERT_GetCustomVideoModeInfo,
    NULL,                       // INVERT_SetCustomVideoMode,
    NULL,                       // INVERT_GetClosestPixelClock,
    NULL,                       // INVERT_GetCRTCTimings,
    NULL,                       // INVERT_SetCRTCTimings,
    NULL,                       // INVERT_SetGlobalRefresh,
    NULL,                       // INVERT_SaveRestoreState,
    NULL,                       // INVERT_SetDisplayOutput,
    NULL,                       // INVERT_GetDisplayOutput,
    NULL,                       // INVERT_SetSoftwareRenderFuncs,
    NULL,                       // INVERT_GetUniqueFilename,
    NULL,                       // Reserved
    NULL,                       // INVERT_GetMonitorInfo,
    NULL,                       // INVERT_SetMonitorInfo,
    NULL,                       // INVERT_GetCurrentVideoModeInfo,
    NULL,                       // INVERT_GetCertifyInfo
    NULL,                       // INVERT_SetCRTCTimings
    NULL,                       // INVERT_AlignLinearBuffer
    NULL,                       // Reserved
    NULL,                       // INVERT_GetCurrentRefreshRate
    NULL,                       // INVERT_PollForDisplaySwitch
    NULL,                       // INVERT_PerformDisplaySwitch
    NULL,                       // INVERT_GetVideoModeInfoExt
    NULL,                       // INVERT_GetCustomVideoModeInfoExt
    NULL,                       // INVERT_SwitchPhysicalResolution
    NULL,                       // INVERT_GetNumberOfHeads
    NULL,                       // INVERT_SetActiveHead
    NULL,                       // INVERT_GetActiveHead
    NULL,                       // INVERT_SetRef2dPointer
    };

static GA_driverFuncs INVERT_driverPT = {
    sizeof(GA_driverFuncs),
    NULL,                       // INVERT_SetBank,
    NULL,                       // INVERT_SetDisplayStart,
    NULL,                       // INVERT_SetDisplayStartXY,
    NULL,                       // INVERT_SetStereoDisplayStart,
    NULL,                       // INVERT_GetDisplayStartStatus,
    NULL,                       // INVERT_EnableStereoMode,
    NULL,                       // INVERT_IsVSync,
    NULL,                       // INVERT_WaitVSync,
    NULL,                       // INVERT_SetVSyncWidth,
    NULL,                       // INVERT_GetVSyncWidth,
    INVERT_SetPaletteData,
    INVERT_GetPaletteData,
    INVERT_SetGammaCorrectData,
    INVERT_GetGammaCorrectData,
    NULL,                       // INVERT_GetCurrentScanLine,
    INVERT_SetPaletteDataExt,
    INVERT_GetPaletteDataExt,
    INVERT_SetGammaCorrectDataExt,
    INVERT_GetGammaCorrectDataExt,
    };

static GA_cursorFuncs INVERT_cursorPT = {
    sizeof(GA_cursorFuncs),
    NULL,                       // INVERT_SetMonoCursor,
    INVERT_SetMonoCursorColor,
    INVERT_SetColorCursor,
    NULL,                       // INVERT_SetCursorPos,
    NULL,                       // INVERT_ShowCursor,
    NULL,                       // INVERT_BeginAccess,
    NULL,                       // INVERT_EndAccess,
    INVERT_SetColorCursor256,
    INVERT_SetColorCursorRGBA,
    NULL,                       // INVERT_IsHardwareCursor
    INVERT_SetColorCursorRGB,
    };

static GA_2DStateFuncs INVERT_state2dPT = {
    sizeof(GA_2DStateFuncs),
    INVERT_SetDrawBuffer,
    NULL,                       // INVERT_IsIdle,
    NULL,                       // INVERT_WaitTillIdle,
    NULL,                       // INVERT_EnableDirectAccess
    NULL,                       // INVERT_DisableDirectAccess
    NULL,                       // INVERT_SetMix,
    INVERT_SetForeColor,
    INVERT_SetBackColor,
    NULL,                       // INVERT_Set8x8MonoPattern,
    NULL,                       // INVERT_Use8x8MonoPattern,
    NULL,                       // INVERT_Use8x8TransMonoPattern,
    INVERT_Set8x8ColorPattern,
    NULL,                       // INVERT_Use8x8ColorPattern,
    NULL,                       // INVERT_Use8x8TransColorPattern,
    NULL,                       // INVERT_SetLineStipple,
    NULL,                       // INVERT_SetLineStippleCount,
    NULL,                       // INVERT_SetPlaneMask,
    NULL,                       // INVERT_SetBlendFunc,
    NULL,                       // res1
    NULL,                       // INVERT_SetAlphaValue,
    NULL,                       // DRIVER_SetupForWindows
    NULL,                       // DRIVER_ResetWindows
    NULL,                       // INVERT_SetLineStyle,
    NULL,                       // INVERT_BuildTranslateVector,
    NULL,                       // INVERT_SetBlendFunc
    };

static GA_2DStateFuncs INVERT_state2dStub = {
    sizeof(GA_2DStateFuncs),
    ISTUB_SetDrawBuffer,
    ISTUB_IsIdle,
    ISTUB_WaitTillIdle,
    ISTUB_EnableDirectAccess,
    ISTUB_DisableDirectAccess,
    ISTUB_SetMix,
    ISTUB_SetForeColor,
    ISTUB_SetBackColor,
    ISTUB_Set8x8MonoPattern,
    ISTUB_Use8x8MonoPattern,
    ISTUB_Use8x8TransMonoPattern,
    ISTUB_Set8x8ColorPattern,
    ISTUB_Use8x8ColorPattern,
    ISTUB_Use8x8TransColorPattern,
    ISTUB_SetLineStipple,
    ISTUB_SetLineStippleCount,
    ISTUB_SetPlaneMask,
    ISTUB_SetSrcBlendFunc,
    ISTUB_SetDstBlendFunc,
    ISTUB_SetAlphaValue,
    ISTUB_SetupForWindows,
    ISTUB_ResetWindows,
    ISTUB_SetLineStyle,
    ISTUB_BuildTranslateVector,
    ISTUB_SetBlendFunc,
    };

static GA_2DRenderFuncs INVERT_draw2dPT = {
    sizeof(GA_2DRenderFuncs),
    INVERT_GetPixel,
    NULL,                       // INVERT_PutPixel,
    NULL,                       // INVERT_DrawScanList,
    NULL,                       // INVERT_DrawPattScanList,
    NULL,                       // INVERT_DrawColorPattScanList,
    NULL,                       // INVERT_DrawEllipseList,
    NULL,                       // INVERT_DrawPattEllipseList,
    NULL,                       // INVERT_DrawColorPattEllipseList,
    NULL,                       // INVERT_DrawFatEllipseList,
    NULL,                       // INVERT_DrawPattFatEllipseList,
    NULL,                       // INVERT_DrawColorPattFatEllipseList,
    NULL,                       // INVERT_DrawRect,
    NULL,
    NULL,                       // INVERT_DrawPattRect,
    NULL,                       // INVERT_DrawColorPattRect,
    NULL,                       // INVERT_DrawTrap,
    NULL,                       // INVERT_DrawPattTrap,
    NULL,                       // INVERT_DrawColorPattTrap,
    NULL,                       // INVERT_DrawLineInt,
    NULL,                       // INVERT_DrawBresenhamLine,
    NULL,                       // INVERT_DrawStippleLineInt,
    NULL,                       // INVERT_DrawBresenhamStippleLine,
    NULL,                       // INVERT_DrawEllipse,
    NULL,                       // INVERT_ClipEllipse,
    NULL,                       // INVERT_PutMonoImageMSBSys,
    NULL,                       // INVERT_PutMonoImageMSBLin,
    NULL,                       // INVERT_PutMonoImageMSBBM,
    NULL,                       // INVERT_PutMonoImageLSBSys,
    NULL,                       // INVERT_PutMonoImageLSBLin,
    NULL,                       // INVERT_PutMonoImageLSBBM,
    NULL,                       // INVERT_ClipMonoImageMSBSys,
    NULL,                       // INVERT_ClipMonoImageMSBLin,
    NULL,                       // INVERT_ClipMonoImageMSBBM,
    NULL,                       // INVERT_ClipMonoImageLSBSys,
    NULL,                       // INVERT_ClipMonoImageLSBLin,
    NULL,                       // INVERT_ClipMonoImageLSBBM,
    NULL,                       // INVERT_BitBlt,
    NULL,                       // INVERT_BitBltLin,
    INVERT_BitBltSys,
    NULL,                       // INVERT_BitBltBM,
    NULL,                       // INVERT_BitBltPatt,
    NULL,                       // INVERT_BitBltPattLin,
    INVERT_BitBltPattSys,
    NULL,                       // INVERT_BitBltPattBM,
    NULL,                       // INVERT_BitBltColorPatt,
    NULL,                       // INVERT_BitBltColorPattLin,
    INVERT_BitBltColorPattSys,
    NULL,                       // INVERT_BitBltColorPattBM,
    INVERT_SrcTransBlt,
    INVERT_SrcTransBltLin,
    INVERT_SrcTransBltSys,
    INVERT_SrcTransBltBM,
    INVERT_DstTransBlt,
    INVERT_DstTransBltLin,
    INVERT_DstTransBltSys,
    INVERT_DstTransBltBM,
    NULL,                       // INVERT_StretchBlt,
    NULL,                       // INVERT_StretchBltLin,
    INVERT_StretchBltSys,
    NULL,                       // INVERT_StretchBltBM,
    INVERT_ConvertBltSys_Obsolete,
    NULL,                       // INVERT_ConvertBltBM_Obsolete,
    INVERT_StretchConvertBltSys_Obsolete,
    NULL,                       // INVERT_StretchConvertBltBM_Obsolete,
    NULL,                       // INVERT_BitBltFxTest,
    NULL,                       // INVERT_BitBltFx,
    NULL,                       // INVERT_BitBltFxLin,
    NULL,                       // INVERT_BitBltFxSys,
    NULL,                       // INVERT_BitBltFxBM,
    NULL,                       // INVERT_GetBitmapSys,
    NULL,                       // INVERT_GetBitmapBM,
    NULL,                       // INVERT_UpdateScreen,
    NULL,                       // INVERT_DrawClippedLineInt,
    NULL,                       // INVERT_DrawClippedBresenhamLine,
    NULL,                       // INVERT_DrawClippedStippleLineInt,
    NULL,                       // INVERT_DrawClippedBresenhamStippleLine,
    NULL,                       // INVERT_BitBltPlaneMasked,
    NULL,                       // INVERT_BitBltPlaneMaskedLin,
    NULL,                       // INVERT_BitBltPlaneMaskedSys,
    NULL,                       // INVERT_BitBltPlaneMaskedBM,
    NULL,                       // INVERT_DrawRectLin,
    NULL,                       // INVERT_DrawRectLinExt,
    NULL,                       // INVERT_DrawStyleLineInt,
    NULL,                       // INVERT_DrawBresenhamStyleLine,
    NULL,                       // INVERT_DrawClippedStyleLineInt,
    NULL,                       // INVERT_DrawClippedBresenhamStyleLine,
    };

static GA_2DRenderFuncs INVERT_draw2dStub = {
    sizeof(GA_2DRenderFuncs),
    ISTUB_GetPixel,
    ISTUB_PutPixel,
    ISTUB_DrawScanList,
    ISTUB_DrawPattScanList,
    ISTUB_DrawColorPattScanList,
    ISTUB_DrawEllipseList,
    ISTUB_DrawPattEllipseList,
    ISTUB_DrawColorPattEllipseList,
    ISTUB_DrawFatEllipseList,
    ISTUB_DrawPattFatEllipseList,
    ISTUB_DrawColorPattFatEllipseList,
    ISTUB_DrawRect,
    NULL,
    ISTUB_DrawPattRect,
    ISTUB_DrawColorPattRect,
    ISTUB_DrawTrap,
    ISTUB_DrawPattTrap,
    ISTUB_DrawColorPattTrap,
    ISTUB_DrawLineInt,
    ISTUB_DrawBresenhamLine,
    ISTUB_DrawStippleLineInt,
    ISTUB_DrawBresenhamStippleLine,
    ISTUB_DrawEllipse,
    ISTUB_ClipEllipse,
    ISTUB_PutMonoImageMSBSys,
    ISTUB_PutMonoImageMSBLin,
    ISTUB_PutMonoImageMSBBM,
    ISTUB_PutMonoImageLSBSys,
    ISTUB_PutMonoImageLSBLin,
    ISTUB_PutMonoImageLSBBM,
    ISTUB_ClipMonoImageMSBSys,
    ISTUB_ClipMonoImageMSBLin,
    ISTUB_ClipMonoImageMSBBM,
    ISTUB_ClipMonoImageLSBSys,
    ISTUB_ClipMonoImageLSBLin,
    ISTUB_ClipMonoImageLSBBM,
    ISTUB_BitBlt,
    ISTUB_BitBltLin,
    ISTUB_BitBltSys,
    ISTUB_BitBltBM,
    ISTUB_BitBltPatt,
    ISTUB_BitBltPattLin,
    ISTUB_BitBltPattSys,
    ISTUB_BitBltPattBM,
    ISTUB_BitBltColorPatt,
    ISTUB_BitBltColorPattLin,
    ISTUB_BitBltColorPattSys,
    ISTUB_BitBltColorPattBM,
    ISTUB_SrcTransBlt,
    ISTUB_SrcTransBltLin,
    ISTUB_SrcTransBltSys,
    ISTUB_SrcTransBltBM,
    ISTUB_DstTransBlt,
    ISTUB_DstTransBltLin,
    ISTUB_DstTransBltSys,
    ISTUB_DstTransBltBM,
    ISTUB_StretchBlt,
    ISTUB_StretchBltLin,
    ISTUB_StretchBltSys,
    ISTUB_StretchBltBM,
    ISTUB_ConvertBltSys_Obsolete,
    ISTUB_ConvertBltBM_Obsolete,
    ISTUB_StretchConvertBltSys_Obsolete,
    ISTUB_StretchConvertBltBM_Obsolete,
    ISTUB_BitBltFxTest,
    ISTUB_BitBltFx,
    ISTUB_BitBltFxLin,
    ISTUB_BitBltFxSys,
    ISTUB_BitBltFxBM,
    ISTUB_GetBitmapSys,
    ISTUB_GetBitmapBM,
    NULL,                       // ISTUB_UpdateScreen,
    NULL,                       // ISTUB_DrawClippedLineInt,
    NULL,                       // ISTUB_DrawClippedBresenhamLine,
    NULL,                       // ISTUB_DrawClippedStippleLineInt,
    NULL,                       // ISTUB_DrawClippedBresenhamStippleLine,
    NULL,                       // ISTUB_BitBltPlaneMasked,
    NULL,                       // ISTUB_BitBltPlaneMaskedLin,
    NULL,                       // ISTUB_BitBltPlaneMaskedSys,
    NULL,                       // ISTUB_BitBltPlaneMaskedBM,
    ISTUB_DrawRectLin,
    ISTUB_DrawRectExt,
    ISTUB_DrawStyleLineInt,
    ISTUB_DrawBresenhamStyleLine,
    ISTUB_DrawClippedStyleLineInt,
    ISTUB_DrawClippedBresenhamStyleLine,
    };

GA_initFuncs        INVERT_initREF;
GA_driverFuncs      INVERT_driverREF;
GA_cursorFuncs      INVERT_cursorHW;
GA_cursorFuncs      INVERT_cursorREF;
GA_2DStateFuncs     INVERT_state2dREF;
GA_2DRenderFuncs    INVERT_draw2dREF;
GA_2DStateFuncs     _VARAPI INVERT_state2dCnt;
GA_2DRenderFuncs    _VARAPI INVERT_draw2dCnt;

#ifdef  __DRIVER__
PM_imports _VARAPI _PM_imports;
#endif

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
REMARKS:
Function to initialise the INVERT display filter driver when it is loaded
from disk. We only do a partial init in here, as some of the driver is
initialised when a mode is set.
****************************************************************************/
static ibool INVERT_init(
    GA_devCtx *hwCtx,
    GA_globalOptions *opt)
{
    // Initialise the internal state of the filter driver
    DC.hwCtx = hwCtx;
    DC.gOpt = opt;
    DC.enabled = false;
    INVERT_initREF.dwSize = sizeof(INVERT_initREF);
    if (!hwCtx->loader.QueryFunctions(GA_GET_INITFUNCS,-2,&INVERT_initREF))
        return false;
    INVERT_driverREF.dwSize = sizeof(INVERT_driverREF);
    if (!hwCtx->loader.QueryFunctions(GA_GET_DRIVERFUNCS,-2,&INVERT_driverREF))
        return false;

    // Hook the device context QueryFunctions function so we can return
    // our hooked functions.
    DC.QueryFunctionsDC = hwCtx->loader.QueryFunctions;
    hwCtx->loader.QueryFunctions = INVERT_QueryFunctionsDC;

    // Obtain a copy of the options structure so we can determine if we
    // are enabled or not.
    DC.opt.dwSize = sizeof(DC.opt);
    INVERT_initREF.GetOptions(&DC.opt);

    // Initialise all the filter driver vectors
    INIT_INIT_VECTORS;
    INIT_DRIVER_VECTORS;
    return true;
}

#ifdef  __DRIVER__
/****************************************************************************
REMARKS:
Fatal error handler called when a non-imported function is called by the
driver. We leave this to a runtime error so that older applications and
shell drivers will work with newer bpd drivers provided no newer functions
are required by the driver itself. If they are, the application or shell
driver needs to be recompiled.
****************************************************************************/
static void _PM_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported PM_imports import function called! Please re-compile!\n");
}

/****************************************************************************
REMARKS:
Function entry point for the reference rasteriser library.
****************************************************************************/
REF2D_driver * _CEXPORT INVERT_initLibrary(
    GA_devCtx *hwCtx,
    GA_globalOptions *gOpt,
    PM_imports *pmImp)
{
    int     i,max;
    ulong   *p;

    // Initialize all default imports to point to fatal error handler
    // for upwards compatibility.
    max = sizeof(_PM_imports)/sizeof(GA_initLibrary_t);
    for (i = 0,p = (ulong*)&_PM_imports; i < max; i++)
        *p++ = (ulong)_PM_fatalErrorHandler;
    memcpy(&_PM_imports,pmImp,MIN(sizeof(_PM_imports),pmImp->dwSize));
    if (!INVERT_init(hwCtx,gOpt))
        return NULL;
    return &DRV;
}
#else
/****************************************************************************
REMARKS:
Function entry point to fake loading of a driver when linked into app.
****************************************************************************/
ibool NAPI INVERT_loadDriver(
    GA_devCtx *hwCtx,
    GA_globalOptions *gOpt,
    REF2D_driver **drv,
    PE_MODULE **hModule,
    ulong *size)
{
    if (!INVERT_init(hwCtx,gOpt))
        return false;
    *drv = &DRV;
    *hModule = NULL;
    *size = 0;
    return true;
}
#endif

/****************************************************************************
REMARKS:
Enabled or disabled the invert driver internally.
****************************************************************************/
static void InvertEnable(void)
{
    if (DC.enabled) {
        INVERT_state2dCnt = INVERT_state2dPT;
        INVERT_draw2dCnt = INVERT_draw2dPT;
        }
    else {
        INVERT_state2dCnt = INVERT_state2dREF;
        INVERT_draw2dCnt = INVERT_draw2dREF;
        }
}

/****************************************************************************
REMARKS:
Function to invert the existing hardware palette or gamma ramp for the
device driver.
****************************************************************************/
static void InvertPalette(void)
{
    int i;

    if (DC.mi.BitsPerPixel > 8)
        INVERT_driverREF.GetGammaCorrectData(DC.invPal,256,0);
    else
        INVERT_driverREF.GetPaletteData(DC.invPal,256,0);
    for (i = 0; i < 256; i++) {
        DC.invPal[i].Red = ~DC.invPal[i].Red;
        DC.invPal[i].Green = ~DC.invPal[i].Green;
        DC.invPal[i].Blue = ~DC.invPal[i].Blue;
        }
    if (DC.mi.BitsPerPixel > 8)
        INVERT_driverREF.SetGammaCorrectData(DC.invPal,256,0,false);
    else
        INVERT_driverREF.SetPaletteData(DC.invPal,256,0,false);
}

/****************************************************************************
REMARKS:
Function to initialise the INVERT display filter driver. This function
gets called long after the driver has actually been loaded, at the
time when the reference rasteriser module is loaded and initialsied (ie:
at mode init time).
****************************************************************************/
ibool NAPI INVERT_InitFilter(
    GA_modeInfo *modeInfo,
    N_int32 transferStart,
    struct _REF2D_driver *ref2d,
    struct _REF2D_driver **drv)
{
    /* Initialise the internal state of the filter driver */
    DC.mi = *modeInfo;
    DC.ref2d = ref2d;
    DC.refState = ref2d->DriverState;
    DC.enabled = IS_INVERTED();
    INVERT_initREF.dwSize = sizeof(INVERT_initREF);
    if (!ref2d->QueryFunctions(GA_GET_INITFUNCS,&INVERT_initREF))
        return false;
    INVERT_driverREF.dwSize = sizeof(INVERT_driverREF);
    if (!ref2d->QueryFunctions(GA_GET_DRIVERFUNCS,&INVERT_driverREF))
        return false;
    INVERT_cursorHW.dwSize = sizeof(INVERT_cursorHW);
    if (!DC.hwCtx->loader.QueryFunctions(GA_GET_CURSORFUNCS,0,&INVERT_cursorHW))
        return false;
    INVERT_cursorREF.dwSize = sizeof(INVERT_cursorREF);
    if (!ref2d->QueryFunctions(GA_GET_CURSORFUNCS,&INVERT_cursorREF))
        return false;
    INVERT_state2dREF.dwSize = sizeof(INVERT_state2dREF);
    if (!ref2d->QueryFunctions(GA_GET_2DSTATEFUNCS,&INVERT_state2dREF))
        return false;
    INVERT_draw2dREF.dwSize = sizeof(INVERT_draw2dREF);
    if (!ref2d->QueryFunctions(GA_GET_2DRENDERFUNCS,&INVERT_draw2dREF))
        return false;

    /* Initialise color depth specific variables */
    switch (DC.mi.BitsPerPixel) {
        case 8:
            DC.bytesPerPixel = 1;
            DC.colorMask = 0xFF;
            DC.fullColorMask = 0xFFFFFFFF;
            break;
        case 15:
        case 16:
            DC.bytesPerPixel = 2;
            DC.colorMask = (DC.mi.BitsPerPixel == 15) ? 0x7FFF : 0xFFFF;
            DC.fullColorMask = (DC.mi.BitsPerPixel == 15) ? 0x7FFF7FFF : 0xFFFFFFFF;
            break;
        case 24:
            DC.bytesPerPixel = 3;
            DC.colorMask = 0x00FFFFFF;
            DC.fullColorMask = 0xFFFFFFFF;
            break;
        case 32:
            DC.bytesPerPixel = 4;
            if (modeInfo->PixelFormat.AlphaPosition == 0)
                DC.colorMask = DC.fullColorMask = 0xFFFFFF00;
            else
                DC.colorMask = DC.fullColorMask = 0x00FFFFFF;
            break;
        }

    /* Allocate memory for the rotation buffer big enough to hold a color
     * bitmap the size of the display screen for rotations.
     */
    DC.bufferSize = DC.mi.YResolution * DC.mi.BytesPerScanLine;
    if ((DC.sysBuffer = malloc(DC.bufferSize)) == NULL)
        return false;
    DC.buffer = DC.sysBuffer;
    DC.transferStart = transferStart;

    /* Initialise all the filter driver vectors */
    INIT_INIT_VECTORS;
    INIT_CURSOR_VECTORS;
    if (modeInfo->BitsPerPixel > 8 && !INVERT_driverREF.SetGammaCorrectData) {
        /* Inversion done at drawing time */
        INVERT_driverPT = INVERT_driverREF;
        INIT_FILTER_VECTORS;
        }
    else {
        /* Inversion done during SetPaletteData or SetGammaCorrectData */
        INIT_DRIVER_VECTORS;
        INVERT_state2dPT = INVERT_state2dREF;
        INVERT_draw2dPT = INVERT_draw2dREF;

        /* Now invert the initial palette or gamma ramp */
        if (DC.enabled)
            InvertPalette();
        }
    InvertEnable();

    /* Return address of our filter driver as top of filter chain */
    *drv = &DRV;
    return true;
}

/****************************************************************************
REMARKS:
Function to clean up when unloading the driver.
****************************************************************************/
void NAPI INVERT_UnloadDriver(void)
{
}

/****************************************************************************
REMARKS:
Function to clean up when the filter driver needs to de-initialise (ie:
going back to text mode when the ref2d driver is unloaded).
****************************************************************************/
void NAPI INVERT_DeInitFilter(void)
{
    // Free our system memory blit buffer
    free(DC.sysBuffer);

    // Unhook our QueryFunctions hook, and restore the init and driver vectors
    DC.hwCtx->loader.QueryFunctions = DC.QueryFunctionsDC;
    INVERT_initREF.dwSize = sizeof(INVERT_initREF);
    DC.hwCtx->loader.QueryFunctions(GA_GET_INITFUNCS,-2,&INVERT_initREF);
    INVERT_driverREF.dwSize = sizeof(INVERT_driverREF);
    DC.hwCtx->loader.QueryFunctions(GA_GET_DRIVERFUNCS,-2,&INVERT_driverREF);
    INIT_INIT_VECTORS;
    INIT_DRIVER_VECTORS;
}

/****************************************************************************
PARAMETERS:
funcs       - Pointer to the function list to make valid
size        - Size of the function list in bytes

RETURNS:
True if valid, false if not.
****************************************************************************/
static ibool MakeValid(
    ulong *funcs,
    int size)
{
    int i,empty = true;

    for (i = 0; i < size/4; i++) {
        if (funcs[i] != 0)
            empty = false;
        }
    return !empty;
}

/****************************************************************************
DESCRIPTION:
Returns the device driver entry points for the specific function group.

PARAMETERS:
id          - Identifier for the function group to get pointers for
safetyLevel - Safety level being requested.
funcs       - Place to store the function block

RETURNS:
True if functions available, false if not.

REMARKS:
This function allows the filter driver to hook the device driver functions
that are called before any modes are initialised. This is used to hook
functions such as GetVideoModeInfo and SetVideoMode as necessary in the
filter driver.
****************************************************************************/
ibool NAPI INVERT_QueryFunctionsDC(
    N_uint32 id,
    N_int32 safetyLevel,
    void *funcs)
{
    ulong   dwSize = *((ulong*)funcs);

    memset(funcs,0,dwSize);
    if (safetyLevel >= 0) {
        switch (id) {
            case GA_GET_INITFUNCS:      RETURN_VALID(INVERT_initPT);
            }
        }
    *((ulong*)funcs) = dwSize;
    return DC.QueryFunctionsDC(id,safetyLevel,funcs);
}

/****************************************************************************
DESCRIPTION:
Returns the reference rasterisers for the specified function group.

PARAMETERS:
id      - Identifier for the function group to get pointers for
funcs   - Place to store the function block

RETURNS:
True if functions available, false if not.

REMARKS:
This function is the main function that is used by the graphics application
code to get a block of function pointers for a specified function group.
The function groups are defined GA_funcGroupsType enumeration, and breaks
up the functions in the device driver API into groups of logically similar
functions.

NOTE: If we get a safetyLevel value of less than zero, it indicates that
      we a being called by some other filter driver and should simply
      pass the call onto the low level hardware driver. Otherwise we may
      end up getting into an infinite loop!
****************************************************************************/
ibool NAPI INVERT_QueryFunctions(
    N_uint32 id,
    void *funcs)
{
    ulong   dwSize = *((ulong*)funcs);

    memset(funcs,0,dwSize);
    switch (id) {
        case GA_GET_INITFUNCS:      RETURN_VALID(INVERT_initPT);
        case GA_GET_DRIVERFUNCS:    RETURN_VALID(INVERT_driverPT);
        case GA_GET_CURSORFUNCS:    RETURN_VALID(INVERT_cursorPT);
        case GA_GET_2DSTATEFUNCS:   RETURN_VALID(INVERT_state2dStub);
        case GA_GET_2DRENDERFUNCS:  RETURN_VALID(INVERT_draw2dStub);
        }
    *((ulong*)funcs) = dwSize;
    return DC.ref2d->QueryFunctions(id,funcs);
}

/****************************************************************************
DESCRIPTION:
Sets a display buffer as the active drawing buffer.

PARAMETERS:
drawBuf             - Buffer to make the active drawing buffer
frameBuffer         - Pointer to the start of frameBuffer memory
bitsPerPixel        - Color depth for the buffer
beginDirectAccess   - Direct frameBuffer arbitration function (NULL if none)
endDirectAccess     - Direct frameBuffer arbitration function (NULL if none)

RETURNS:
0 on success, -1 on failure.

REMARKS:
This function allows the application to tell make a display memory
buffer the active rendering buffer for all subsequent rendering commands.
****************************************************************************/
N_int32 NAPI INVERT_SetDrawBufferSW(
    GA_buffer *drawBuf,
    void *frameBuffer,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pf,
    GA_devCtx *hwCtx,
    N_int32 forceSoftwareOnly)
{
    DC.buf = *drawBuf;
    DC.hwCtx = hwCtx;
    DC.enabled = (IS_INVERTED() && frameBuffer == DC.hwCtx->LinearMem);
    if (DC.enabled) {
        DC.VRAMBufferStart = DC.VRAMBufferStartY = 0;
        if (DC2D.hwDraw.BitBltLin) {
            if ((DC.mi.MaxLinearOffset - DC.transferStart) > DC.bufferSize) {
                /* Buffers are all allocated in video memory and blitted
                 * with the linear blit functions for performance.
                 */
                DC.VRAMBufferStart = DC.transferStart;
                DC.VRAMBufferStartY = 0;
                }
            }
        else if (DC2D.hwDraw.BitBlt) {
            int startY = (DC.transferStart - DC.buf.Offset) / DC.mi.BytesPerScanLine;
            int sizeY = DC.bufferSize / DC.mi.BytesPerScanLine;
            if ((int)(DC.mi.MaxScanLines - startY) > sizeY) {
                /* Buffers are all allocated in video memory and blitted
                 * with the rectangular blit functions.
                 */
                DC.VRAMBufferStart = DC.transferStart;
                DC.VRAMBufferStartY = startY;
                }
            }
        if (DC.VRAMBufferStart)
            DC.buffer = (uchar*)frameBuffer + DC.VRAMBufferStart;
        else
            DC.buffer = DC.sysBuffer;
        }
    InvertEnable();
    return DC.ref2d->SetDrawBuffer(&DC.buf,frameBuffer,bitsPerPixel,pf,hwCtx,forceSoftwareOnly);
}

/****************************************************************************
PARAMETERS:
drawBuf - New hardware draw buffer to make active

RETURNS:
0 on success, -1 on failure.

REMARKS:
This function changes the hardware draw buffer for both the software
rasteriser and the hardware device. This can also be used to change the
draw buffer for a software only buffer, but is generally only used when
using hardware.
****************************************************************************/
N_int32 NAPI INVERT_SetDrawBuffer(
    GA_buffer *drawBuf)
{
    DC.buf = *drawBuf;
    if (DC.enabled)
        return STATE2D.SetDrawBuffer(&DC.buf);
    return -1;
}

/****************************************************************************
PARAMETERS:
surface         - Pointer to the start of the memory buffer for drawing
xRes            - X resolution for the surface
yRes            - Y resolution for the surface
bytesPerLine    - Scanline width for the memory buffer for drawing
bitsPerPixel    - New color depth for the surface
pf              - New pixel format for the surface

REMARKS:
This function changes the software draw buffer quickly, to allow the
software renderer to render to bitmaps in different locations in memory.
We do in here for speed, since we want to avoid the overhead of setting
the full draw buffer for the reference rasteriser code. If the color
depth or pixel format changes, this function will switch all the internal
rendering vectors to put to the new vectors for the appropriate functions.
This allows the reference rasteriser to work with all color depths active
in a single driver (and for our debug library to work properly).
****************************************************************************/
void NAPI INVERT_SetDrawSurface(
    void *surface,
    N_int32 xRes,
    N_int32 yRes,
    N_int32 bytesPerLine,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pf)
{
    // Determine if we need to disable the color inversion features
    DC.enabled = (IS_INVERTED() && DC.refState->haveHWDraw && surface == (DC.refState->frameBuffer + DC.refState->drawOffset));
    DC.ref2d->SetDrawSurface(surface,xRes,yRes,bytesPerLine,bitsPerPixel,pf);
}

/****************************************************************************
REMARKS:
Handles updating of the internals when the options structures are updated.
This allows inversion mode to be toggled automatically at runtime using
either the lobal device driver or global SNAP options functions.
****************************************************************************/
static void UpdateOptions(void)
{
    int oldEnable = DC.enabled;

    DC.enabled = IS_INVERTED();
    InvertEnable();
    if (oldEnable != DC.enabled && (DC.mi.BitsPerPixel <= 8 || INVERT_driverREF.SetGammaCorrectData)) {
        /* Now invert the hardware palette or gamma ramp to switch. This
         * will enable or disable the inversion automatically when this
         * function is called.
         */
        InvertPalette();
        }
}

/****************************************************************************
REMARKS:
Handles overriding of the options in the driver. This allows us to know
when the invert option has been disabled or enabled at runtime.
****************************************************************************/
void NAPI INVERT_SetOptions(
    GA_options *options)
{
    INVERT_initREF.SetOptions(options);
    DC.opt.dwSize = sizeof(DC.opt);
    INVERT_initREF.GetOptions(&DC.opt);
    UpdateOptions();
}

/****************************************************************************
REMARKS:
This function gets called when the global options are changed at runtime,
allowing the filter driver to update the internals based on the changed
global options.
****************************************************************************/
void NAPI INVERT_GlobalOptionsUpdated(void)
{
    UpdateOptions();
}

/****************************************************************************
REMARKS:
This returns the address of the currently active draw buffer
****************************************************************************/
GA_buffer * NAPI INVERT_GetDrawBuffer(void)
{
    return DC.ref2d->GetDrawBuffer();
}

/****************************************************************************
PARAMETERS:
left        - Left coordinate of the rectangle to draw
top         - Top coordinate of the rectangle to draw
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
color       - Color to fill the rectangle with
mix         - Mix to use for drawing the rectangle
****************************************************************************/
void NAPI INVERT_DrawRectExtSW(
    void *buffer,
    N_int32 dstPitch,
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    GA_color color,
    N_int32 mix)
{
    DC.ref2d->DrawRectExtSW(buffer, dstPitch, left, top, width, height, color, mix);
}

/****************************************************************************
DESCRIPTION:
Forces software only rendering on or off.
****************************************************************************/
N_int32 NAPI INVERT_ForceSoftwareOnly(
    N_int32 enable)
{
    return DC.ref2d->ForceSoftwareOnly(enable);
}

/****************************************************************************
PARAMETERS:
mask    - New color compare mask

REMARKS:
This function is used to override the internal software rendering color
compare mask used for source and destination transparent blit functions.
****************************************************************************/
N_uint32 NAPI INVERT_SetColorCompareMask(
    N_uint32 mask)
{
    return DC.ref2d->SetColorCompareMask(mask);
}

/****************************************************************************
DESCRIPTION:
Fix up the software rasteriser after a mode switch has occured
****************************************************************************/
void NAPI INVERT_PostSwitchPhysicalResolution(void)
{
    DC.ref2d->PostSwitchPhysicalResolution();
}

/****************************************************************************
REMARKS:
Programs the palette registers.
****************************************************************************/
void NAPI INVERT_SetPaletteData(
    GA_palette *pal,
    N_int32 num,
    N_int32 index,
    N_int32 waitVRT)
{
    int i;

    if (DC.enabled) {
        for (i = 0; i < num; i++) {
            DC.invPal[i].Red = ~pal[i].Red;
            DC.invPal[i].Green = ~pal[i].Green;
            DC.invPal[i].Blue = ~pal[i].Blue;
            }
        INVERT_driverREF.SetPaletteData(DC.invPal,num,index,waitVRT);
        }
    else
        INVERT_driverREF.SetPaletteData(pal,num,index,waitVRT);
}

/****************************************************************************
REMARKS:
Read the palette registers.
****************************************************************************/
void NAPI INVERT_GetPaletteData(
    GA_palette *pal,
    N_int32 num,
    N_int32 index)
{
    int i;

    if (DC.enabled) {
        INVERT_driverREF.GetPaletteData(DC.invPal,num,index);
        for (i = 0; i < num; i++) {
            pal[i].Red = ~DC.invPal[i].Red;
            pal[i].Green = ~DC.invPal[i].Green;
            pal[i].Blue = ~DC.invPal[i].Blue;
            }
        }
    else
        INVERT_driverREF.GetPaletteData(pal,num,index);
}

/****************************************************************************
REMARKS:
Programs the VGA compatible gamma correction data into the VGA palette
registers.
****************************************************************************/
void NAPI INVERT_SetGammaCorrectData(
    GA_palette *pal,
    N_int32 num,
    N_int32 index,
    N_int32 waitVRT)
{
    int i;

    if (DC.enabled) {
        for (i = 0; i < num; i++) {
            DC.invPal[i].Red = ~pal[i].Red;
            DC.invPal[i].Green = ~pal[i].Green;
            DC.invPal[i].Blue = ~pal[i].Blue;
            }
        INVERT_driverREF.SetGammaCorrectData(DC.invPal,num,index,waitVRT);
        }
    else
        INVERT_driverREF.SetGammaCorrectData(pal,num,index,waitVRT);
}

/****************************************************************************
REMARKS:
Read the palette registers.
****************************************************************************/
void NAPI INVERT_GetGammaCorrectData(
    GA_palette *pal,
    N_int32 num,
    N_int32 index)
{
    int i;

    if (DC.enabled) {
        INVERT_driverREF.GetGammaCorrectData(DC.invPal,num,index);
        for (i = 0; i < num; i++) {
            pal[i].Red = ~DC.invPal[i].Red;
            pal[i].Green = ~DC.invPal[i].Green;
            pal[i].Blue = ~DC.invPal[i].Blue;
            }
        }
    else
        INVERT_driverREF.GetGammaCorrectData(pal,num,index);
}

/****************************************************************************
REMARKS:
Programs the palette registers.
****************************************************************************/
void NAPI INVERT_SetPaletteDataExt(
    GA_paletteExt *pal,
    N_int32 num,
    N_int32 index,
    N_int32 waitVRT)
{
    int i;

    if (DC.enabled) {
        for (i = 0; i < num; i++) {
            DC.invPalExt[i].Red = ~pal[i].Red;
            DC.invPalExt[i].Green = ~pal[i].Green;
            DC.invPalExt[i].Blue = ~pal[i].Blue;
            }
        INVERT_driverREF.SetPaletteDataExt(DC.invPalExt,num,index,waitVRT);
        }
    else
        INVERT_driverREF.SetPaletteDataExt(pal,num,index,waitVRT);
}

/****************************************************************************
REMARKS:
Read the palette registers.
****************************************************************************/
void NAPI INVERT_GetPaletteDataExt(
    GA_paletteExt *pal,
    N_int32 num,
    N_int32 index)
{
    int i;

    if (DC.enabled) {
        INVERT_driverREF.GetPaletteDataExt(DC.invPalExt,num,index);
        for (i = 0; i < num; i++) {
            pal[i].Red = ~DC.invPalExt[i].Red;
            pal[i].Green = ~DC.invPalExt[i].Green;
            pal[i].Blue = ~DC.invPalExt[i].Blue;
            }
        }
    else
        INVERT_driverREF.GetPaletteDataExt(pal,num,index);
}

/****************************************************************************
REMARKS:
Programs the VGA compatible gamma correction data into the VGA palette
registers.
****************************************************************************/
void NAPI INVERT_SetGammaCorrectDataExt(
    GA_paletteExt *pal,
    N_int32 num,
    N_int32 index,
    N_int32 waitVRT)
{
    int i;

    if (DC.enabled) {
        for (i = 0; i < num; i++) {
            DC.invPalExt[i].Red = ~pal[i].Red;
            DC.invPalExt[i].Green = ~pal[i].Green;
            DC.invPalExt[i].Blue = ~pal[i].Blue;
            }
        INVERT_driverREF.SetGammaCorrectDataExt(DC.invPalExt,num,index,waitVRT);
        }
    else
        INVERT_driverREF.SetGammaCorrectDataExt(pal,num,index,waitVRT);
}

/****************************************************************************
REMARKS:
Read the palette registers.
****************************************************************************/
void NAPI INVERT_GetGammaCorrectDataExt(
    GA_paletteExt *pal,
    N_int32 num,
    N_int32 index)
{
    int i;

    if (DC.enabled) {
        INVERT_driverREF.GetGammaCorrectDataExt(DC.invPalExt,num,index);
        for (i = 0; i < num; i++) {
            pal[i].Red = ~DC.invPalExt[i].Red;
            pal[i].Green = ~DC.invPalExt[i].Green;
            pal[i].Blue = ~DC.invPalExt[i].Blue;
            }
        }
    else
        INVERT_driverREF.GetGammaCorrectDataExt(pal,num,index);
}

/****************************************************************************
PARAMETERS:
foreground  - Foreground color for the cursor
background  - Background color for the cursor

REMARKS:
This function sets the hardware cursor foreground and background colors.
****************************************************************************/
void NAPI INVERT_SetMonoCursorColor(
    GA_palette *foreground,
    GA_palette *background)
{
    GA_palette  fore,back;

    if (DC.enabled && (INVERT_cursorHW.SetMonoCursor || DC.mi.BitsPerPixel <= 8)) {
        fore.Red = ~foreground->Red;
        fore.Green = ~foreground->Green;
        fore.Blue = ~foreground->Blue;
        back.Red = ~background->Red;
        back.Green = ~background->Green;
        back.Blue = ~background->Blue;
        INVERT_cursorREF.SetMonoCursorColor(&fore,&back);
        }
    else
        INVERT_cursorREF.SetMonoCursorColor(foreground,background);
}

/****************************************************************************
PARAMETERS:
cursor  - Pointer to an GA_colorCursor structure for the cursor image
****************************************************************************/
void NAPI INVERT_SetColorCursor(
    GA_colorCursor *cursor)
{
    int i;

    if (DC.enabled && (INVERT_cursorHW.SetColorCursor || DC.mi.BitsPerPixel <= 8)) {
        DC.colorCursor = *cursor;
        for (i = 0; i < 16; i++) {
            DC.colorCursor.Palette[i].Red = ~DC.colorCursor.Palette[i].Red;
            DC.colorCursor.Palette[i].Green = ~DC.colorCursor.Palette[i].Green;
            DC.colorCursor.Palette[i].Blue = ~DC.colorCursor.Palette[i].Blue;
            }
        INVERT_cursorREF.SetColorCursor(&DC.colorCursor);
        }
    else
        INVERT_cursorREF.SetColorCursor(cursor);
}

/****************************************************************************
PARAMETERS:
cursor  - Pointer to an GA_colorCursor256 structure for the cursor image
****************************************************************************/
void NAPI INVERT_SetColorCursor256(
    GA_colorCursor256 *cursor)
{
    int i;

    if (DC.enabled && (INVERT_cursorHW.SetColorCursor256 || DC.mi.BitsPerPixel <= 8)) {
        DC.colorCursor256 = *cursor;
        for (i = 0; i < 256; i++) {
            DC.colorCursor256.Palette[i].Red = ~DC.colorCursor256.Palette[i].Red;
            DC.colorCursor256.Palette[i].Green = ~DC.colorCursor256.Palette[i].Green;
            DC.colorCursor256.Palette[i].Blue = ~DC.colorCursor256.Palette[i].Blue;
            }
        INVERT_cursorREF.SetColorCursor256(&DC.colorCursor256);
        }
    else
        INVERT_cursorREF.SetColorCursor256(cursor);
}

/****************************************************************************
PARAMETERS:
cursor  - Pointer to an GA_colorCursorRGB structure for the cursor image
****************************************************************************/
void NAPI INVERT_SetColorCursorRGB(
    GA_colorCursorRGB *cursor)
{
    int i,j;

    if (DC.enabled && (INVERT_cursorHW.SetColorCursorRGB || DC.mi.BitsPerPixel <= 8)) {
        DC.colorCursorRGB = *cursor;
        for (i = 0; i < 64; i++) {
            for (j = 0; j < 64; j++) {
                GA_palette *p = (GA_palette*)&DC.colorCursorRGB.ColorData[(i * 64 + j) * 3];
                p->Red = ~p->Red;
                p->Green = ~p->Green;
                p->Blue = ~p->Blue;
                }
            }
        INVERT_cursorREF.SetColorCursorRGB(&DC.colorCursorRGB);
        }
    else
        INVERT_cursorREF.SetColorCursorRGB(cursor);
}

/****************************************************************************
PARAMETERS:
cursor  - Pointer to an GA_colorCursorRGBA structure for the cursor image
****************************************************************************/
void NAPI INVERT_SetColorCursorRGBA(
    GA_colorCursorRGBA *cursor)
{
    int i,j;

    if (DC.enabled && (INVERT_cursorHW.SetColorCursorRGBA || DC.mi.BitsPerPixel <= 8)) {
        DC.colorCursorRGBA = *cursor;
        for (i = 0; i < 64; i++) {
            for (j = 0; j < 64; j++) {
                GA_palette *p = (GA_palette*)&DC.colorCursorRGBA.ColorData[(i * 64 + j) * 4];
                p->Red = ~p->Red;
                p->Green = ~p->Green;
                p->Blue = ~p->Blue;
                }
            }
        INVERT_cursorREF.SetColorCursorRGBA(&DC.colorCursorRGBA);
        }
    else
        INVERT_cursorREF.SetColorCursorRGBA(cursor);
}

/****************************************************************************
PARAMETERS:
dst         - Destination bitmap buffer
src         - Source bitmap buffer
srcPitch    - Source bitmap stride in bytes
dstPitch    - Destination bitmap stride in bytes
byteWidth   - Width of the source bitmap to invert in bytes
height      - Height of the source bitmap in scanlines
flipY       - True if the bitmap should be flipped in Y

REMARKS:
This function inverts a bitmap image, and stores the value in the
destination buffer. The source buffer is not modified.
****************************************************************************/
static void InvertBitmap(
    uchar *dst,
    uchar *src,
    N_int32 srcPitch,
    N_int32 dstPitch,
    N_int32 byteWidth,
    N_int32 height,
    N_int32 flipY)
{
    int Bremainder = byteWidth & 3;
    int DWwidth = byteWidth/4;
    int DWremainder = DWwidth & 7;
    int srcAdjust = srcPitch - byteWidth;
    int dstAdjust = dstPitch - byteWidth;

    DWwidth /= 8;
    if (flipY) {
        srcAdjust = -srcPitch - byteWidth;
        src += (height-1)*srcPitch;
        }
    while (height--) {
        ulong *sd = (ulong*)src;
        ulong *dd = (ulong*)dst;
        ulong mask = DC.fullColorMask;
        int count;

        // Process of all DWORD's
        count = DWwidth;
        while (count--) {
            dd[0] = sd[0] ^ mask;
            dd[1] = sd[1] ^ mask;
            dd[2] = sd[2] ^ mask;
            dd[3] = sd[3] ^ mask;
            dd[4] = sd[4] ^ mask;
            dd[5] = sd[5] ^ mask;
            dd[6] = sd[6] ^ mask;
            dd[7] = sd[7] ^ mask;
            dd += 8;
            sd += 8;
            }
        if (DWremainder) {
            count = DWremainder;
            while (count--)
                *dd++ = *sd++ ^ mask;
            }

        // Process all remaining bytes (if any)
        src = (uchar*)sd;
        dst = (uchar*)dd;
        if (Bremainder) {
            count = Bremainder;
            while (count--) {
                *dst++ = *src++ ^ (uchar)mask;
                mask >>= 8;
                }
            }

        // Move to the next scanline
        src += srcAdjust;
        dst += dstAdjust;
        }
}

/****************************************************************************
PARAMETERS:
src             - Pointer to the source bitmap to rotate
bitsPerPixel    - Pixel depth of the source bitmap in bits
pitch           - Stride of the source bitmap in bytes
width           - Width of the source bitmap
height          - Height of the source bitmap
****************************************************************************/
void * NAPI INVERT_RotateBitmap(
    void *src,
    N_int32 bitsPerPixel,
    N_int32 *pitch,
    N_int32 width,
    N_int32 height)
{
    if (width > DC.mi.XResolution || height > DC.mi.YResolution)
        return NULL;
    switch (bitsPerPixel) {
        case 15:
        case 16:
            InvertBitmap(DC.sysBuffer,src,*pitch,*pitch,width*2,height,false);
            break;
        case 24:
            InvertBitmap(DC.sysBuffer,src,*pitch,*pitch,width*3,height,false);
            break;
        case 32:
            InvertBitmap(DC.sysBuffer,src,*pitch,*pitch,width*4,height,false);
            break;
        default:
            return NULL;
        }
    return DC.sysBuffer;
}

/****************************************************************************
PARAMETERS:
color   - New foreground color to set

REMARKS:
This function sets the hardware foreground color for subsequent accelerated
rendering primitives.
****************************************************************************/
void NAPI INVERT_SetForeColor(
    GA_color color)
{
    STATE2D.SetForeColor(color ^ DC.colorMask);
}

/****************************************************************************
PARAMETERS:
color   - New background color to set

REMARKS:
This function sets the hardware background color for subsequent accelerated
rendering primitives.
****************************************************************************/
void NAPI INVERT_SetBackColor(
    GA_color color)
{
    STATE2D.SetBackColor(color ^ DC.colorMask);
}

/****************************************************************************
PARAMETERS:
index   - Index of the pattern to download
pattern - Pointer to pattern data to download
****************************************************************************/
void NAPI INVERT_Set8x8ColorPattern(
    N_int32 index,
    GA_colorPattern *pattern)
{
    if (DC.enabled) {
        GA_colorPattern pat;
        ulong           *src = (ulong*)pattern->b32.p;
        ulong           *dst = (ulong*)pat.b32.p;
        int             count;
        ulong           mask = DC.fullColorMask;

        // Invert the color pattern
        switch (DC.mi.BitsPerPixel) {
            case 8:     count = 2;  break;
            case 15:
            case 16:    count = 4;  break;
            case 24:    count = 6;  break;
            case 32:    count = 8;  break;
            }
        while (count--) {
            dst[0] = src[0] ^ mask;
            dst[1] = src[1] ^ mask;
            dst[2] = src[2] ^ mask;
            dst[3] = src[3] ^ mask;
            dst[4] = src[4] ^ mask;
            dst[5] = src[5] ^ mask;
            dst[6] = src[6] ^ mask;
            dst[7] = src[7] ^ mask;
            dst += 8;
            src += 8;
            }
        STATE2D.Set8x8ColorPattern(index,&pat);
        }
    else
        STATE2D.Set8x8ColorPattern(index,pattern);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
dstPitch    - Place to store the destination bitmap pitch
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
flipY       - True if the bitmap should be INVERT in Y
****************************************************************************/
static void InvertSysBitmap(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 *dstPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 flipY)
{
    // TODO: Shouldn't we support dstLeft and dstTop??
    (void)dstLeft;
    (void)dstTop;
    DC2D.beginDirectAccess();
    switch (DC.mi.BitsPerPixel) {
        case 15:
        case 16:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*2;
            if (DC.VRAMBufferStartY)
                *dstPitch = DC.mi.BytesPerScanLine;
            else
                *dstPitch = ROUNDUP(width*2,DC.mi.BitmapStridePad);
            InvertBitmap(DC.buffer,srcAddr,srcPitch,*dstPitch,width*2,height,flipY);
            break;
        case 24:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*3;
            if (DC.VRAMBufferStartY)
                *dstPitch = DC.mi.BytesPerScanLine;
            else
                *dstPitch = ROUNDUP(width*3,DC.mi.BitmapStridePad);
            InvertBitmap(DC.buffer,srcAddr,srcPitch,*dstPitch,width*3,height,flipY);
            break;
        case 32:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*4;
            if (DC.VRAMBufferStartY)
                *dstPitch = DC.mi.BytesPerScanLine;
            else
                *dstPitch = ROUNDUP(width*4,DC.mi.BitmapStridePad);
            InvertBitmap(DC.buffer,srcAddr,srcPitch,*dstPitch,width*4,height,flipY);
            break;
        }
    DC2D.endDirectAccess();
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
flipY       - True if the bitmap should be INVERT in Y
****************************************************************************/
void NAPI INVERT_BitBltSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    N_int32 flipY)
{
    N_int32 dstPitch;

    InvertSysBitmap(srcAddr,srcPitch,&dstPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,flipY);
    if (DC.VRAMBufferStartY)
        DRAW2D.BitBlt(0,DC.VRAMBufferStartY,width,height,dstLeft,dstTop,mix);
    else if (DC.VRAMBufferStart)
        DRAW2D.BitBltLin(DC.VRAMBufferStart,dstPitch,0,0,width,height,dstLeft,dstTop,mix);
    else
        DRAW2D.BitBltSys(DC.buffer,dstPitch,0,0,width,height,dstLeft,dstTop,mix,flipY);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3        - Mix code for the copy (GA_mixCodesType)
flipY       - True if the bitmap should be INVERT in Y
****************************************************************************/
void NAPI INVERT_BitBltPattSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3,
    N_int32 flipY)
{
    N_int32 dstPitch;

    InvertSysBitmap(srcAddr,srcPitch,&dstPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,flipY);
    if (DC.VRAMBufferStartY)
        DRAW2D.BitBltPatt(0,DC.VRAMBufferStartY,width,height,dstLeft,dstTop,rop3);
    else if (DC.VRAMBufferStart)
        DRAW2D.BitBltPattLin(DC.VRAMBufferStart,dstPitch,0,0,width,height,dstLeft,dstTop,rop3);
    else
        DRAW2D.BitBltPattSys(DC.buffer,dstPitch,0,0,width,height,dstLeft,dstTop,rop3,false);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3        - Mix code for the copy (GA_mixCodesType)
flipY       - True if the bitmap should be INVERT in Y
****************************************************************************/
void NAPI INVERT_BitBltColorPattSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3,
    N_int32 flipY)
{
    N_int32 dstPitch;

    InvertSysBitmap(srcAddr,srcPitch,&dstPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,flipY);
    if (DC.VRAMBufferStartY)
        DRAW2D.BitBltColorPatt(0,DC.VRAMBufferStartY,width,height,dstLeft,dstTop,rop3);
    else if (DC.VRAMBufferStart)
        DRAW2D.BitBltColorPattLin(DC.VRAMBufferStart,dstPitch,0,0,width,height,dstLeft,dstTop,rop3);
    else
        DRAW2D.BitBltColorPattSys(DC.buffer,dstPitch,0,0,width,height,dstLeft,dstTop,rop3,false);
}

/****************************************************************************
PARAMETERS:
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
****************************************************************************/
void NAPI INVERT_SrcTransBlt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    DRAW2D.SrcTransBlt(srcLeft,srcTop,width,height,dstLeft,dstTop,mix,transparent ^ DC.colorMask);
}

/****************************************************************************
PARAMETERS:
srcOfs      - Offset of source bitmap in video memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
****************************************************************************/
void NAPI INVERT_SrcTransBltLin(
    N_int32 srcOfs,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    DRAW2D.SrcTransBltLin(srcOfs,srcPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,mix,transparent ^ DC.colorMask);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
flipY       - True if the bitmap should be INVERT in Y
****************************************************************************/
void NAPI INVERT_SrcTransBltSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent,
    N_int32 flipY)
{
    N_int32 dstPitch;

    InvertSysBitmap(srcAddr,srcPitch,&dstPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,flipY);
    if (DC.VRAMBufferStartY)
        DRAW2D.SrcTransBlt(0,DC.VRAMBufferStartY,width,height,dstLeft,dstTop,mix,transparent ^ DC.colorMask);
    else if (DC.VRAMBufferStart)
        DRAW2D.SrcTransBltLin(DC.VRAMBufferStart,dstPitch,0,0,width,height,dstLeft,dstTop,mix,transparent ^ DC.colorMask);
    else
        DRAW2D.SrcTransBltSys(DC.buffer,dstPitch,0,0,width,height,dstLeft,dstTop,mix,transparent ^ DC.colorMask,false);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPhysAddr - Physical address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
****************************************************************************/
void NAPI INVERT_SrcTransBltBM(
    void *srcAddr,
    N_int32 srcPhysAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    switch (DC.mi.BitsPerPixel) {
        case 15:
        case 16:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*2;
            InvertBitmap(DC.sysBuffer,srcAddr,srcPitch,srcPitch,width*2,height,false);
            break;
        case 24:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*3;
            InvertBitmap(DC.sysBuffer,srcAddr,srcPitch,srcPitch,width*3,height,false);
            break;
        case 32:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*4;
            InvertBitmap(DC.sysBuffer,srcAddr,srcPitch,srcPitch,width*4,height,false);
            break;
        }
    DRAW2D.SrcTransBltBM(srcAddr,srcPhysAddr,srcPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,mix,transparent ^ DC.colorMask);
}

/****************************************************************************
PARAMETERS:
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
****************************************************************************/
void NAPI INVERT_DstTransBlt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    DRAW2D.DstTransBlt(srcLeft,srcTop,width,height,dstLeft,dstTop,mix,transparent ^ DC.colorMask);
}

/****************************************************************************
PARAMETERS:
srcOfs      - Offset of source bitmap in video memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
****************************************************************************/
void NAPI INVERT_DstTransBltLin(
    N_int32 srcOfs,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    DRAW2D.DstTransBltLin(srcOfs,srcPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,mix,transparent ^ DC.colorMask);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
flipY       - True if the bitmap should be INVERT in Y
****************************************************************************/
void NAPI INVERT_DstTransBltSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent,
    N_int32 flipY)
{
    N_int32 dstPitch;

    InvertSysBitmap(srcAddr,srcPitch,&dstPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,flipY);
    if (DC.VRAMBufferStartY)
        DRAW2D.DstTransBlt(0,DC.VRAMBufferStartY,width,height,dstLeft,dstTop,mix,transparent ^ DC.colorMask);
    else if (DC.VRAMBufferStart)
        DRAW2D.DstTransBltLin(DC.VRAMBufferStart,dstPitch,0,0,width,height,dstLeft,dstTop,mix,transparent ^ DC.colorMask);
    else
        DRAW2D.DstTransBltSys(DC.buffer,dstPitch,0,0,width,height,dstLeft,dstTop,mix,transparent ^ DC.colorMask,false);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPhysAddr - Physical address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
****************************************************************************/
void NAPI INVERT_DstTransBltBM(
    void *srcAddr,
    N_int32 srcPhysAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    switch (DC.mi.BitsPerPixel) {
        case 15:
        case 16:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*2;
            InvertBitmap(DC.sysBuffer,srcAddr,srcPitch,srcPitch,width*2,height,false);
            break;
        case 24:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*3;
            InvertBitmap(DC.sysBuffer,srcAddr,srcPitch,srcPitch,width*3,height,false);
            break;
        case 32:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*4;
            InvertBitmap(DC.sysBuffer,srcAddr,srcPitch,srcPitch,width*4,height,false);
            break;
        }
    DRAW2D.DstTransBltBM(srcAddr,srcPhysAddr,srcPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,mix,transparent ^ DC.colorMask);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstWidth    - Width of the destination rectangle in pixels
dstHeight   - Height of the destination rectangle in scanlines
doClip      - True if the blit should be clipped
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle
mix         - Mix code for the copy (GA_mixCodesType)
flipY       - True if the bitmap should be INVERT in Y
****************************************************************************/
void NAPI INVERT_StretchBltSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    N_int32 doClip,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom,
    N_int32 mix,
    N_int32 flipY)
{
    switch (DC.mi.BitsPerPixel) {
        case 15:
        case 16:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*2;
            InvertBitmap(DC.sysBuffer,srcAddr,srcPitch,srcPitch,srcWidth*2,srcHeight,flipY);
            break;
        case 24:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*3;
            InvertBitmap(DC.sysBuffer,srcAddr,srcPitch,srcPitch,srcWidth*3,srcHeight,flipY);
            break;
        case 32:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*4;
            InvertBitmap(DC.sysBuffer,srcAddr,srcPitch,srcPitch,srcWidth*4,srcHeight,flipY);
            break;
        }
    DRAW2D.StretchBltSys(DC.sysBuffer,srcPitch,0,0,srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,doClip,clipLeft,clipTop,clipRight,clipBottom,mix,false);
}

/****************************************************************************
PARAMETERS:
srcAddr         - Address of source bitmap in system memory
srcPitch        - Pitch of source bitmap in bytes
srcLeft         - Left coordinate within source bitmap to copy
srcTop          - Top coordinate within source bitmap to copy
width           - Width of the rectangle in pixels
height          - Height of the rectangle in scanlines
dstLeft         - Left coordinate of destination
dstTop          - Top coordinate of destination
bitsPerPixel    - Pixel depth for the bitmap to copy
pixelFormat     - Pixel format for the bitmap to copy
dstPal          - Color palette for the destination device (if dest is 8bpp)
srcPal          - Color palette for the bitmap to copy (if source is 1/4/8bpp)
dither          - True if the output should be dithered
mix             - Mix code for the copy (GA_mixCodesType)
flipY           - True if the bitmap should be INVERT in Y

RETURNS:
0 on success, -1 on failure.
****************************************************************************/
N_int32 NAPI INVERT_ConvertBltSys_Obsolete(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pixelFormat,
    GA_palette *dstPal,
    GA_palette *srcPal,
    N_int32 dither,
    N_int32 mix,
    N_int32 flipY)
{
    GA_palette  pal[256],*psrc,*pdst;
    int         count;

    switch (bitsPerPixel) {
        case 4:
            count = 16;
            psrc = srcPal;
            pdst = pal;
            while (count--)
                *((ulong*)pdst++) = *((ulong*)psrc++) ^ 0xFFFFFF;
            break;
        case 8:
            count = 256;
            psrc = srcPal;
            pdst = pal;
            while (count--)
                *((ulong*)pdst++) = *((ulong*)psrc++) ^ 0xFFFFFF;
            break;
        case 15:
        case 16:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*2;
            InvertBitmap(DC.sysBuffer,srcAddr,srcPitch,srcPitch,width*2,height,false);
            break;
        case 24:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*3;
            InvertBitmap(DC.sysBuffer,srcAddr,srcPitch,srcPitch,width*3,height,false);
            break;
        case 32:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*4;
            InvertBitmap(DC.sysBuffer,srcAddr,srcPitch,srcPitch,width*4,height,false);
            break;
        }
    return DRAW2D.ConvertBltSys_Obsolete(DC.sysBuffer,srcPitch,0,0,width,height,
        dstLeft,dstTop,bitsPerPixel,pixelFormat,dstPal,pal,dither,mix,flipY);
}

/****************************************************************************
PARAMETERS:
srcAddr         - Address of source bitmap in system memory
srcPitch        - Pitch of source bitmap in bytes
srcLeft         - Left coordinate of the source rectangle to copy
srcTop          - Top coordinate of the source rectangle to copy
srcWidth        - Width of the source rectangle in pixels
srcHeight       - Height of the source rectangle in scanlines
dstLeft         - Left coordinate of destination
dstTop          - Top coordinate of destination
dstWidth        - Width of the destination rectangle in pixels
dstHeight       - Height of the destination rectangle in scanlines
doClip          - True if the blit should be clipped
clipLeft        - Left coordinate of clip rectangle
clipTop         - Top coordinate of clip rectangle
clipRight       - Right coordinate of clip rectangle
clipBottom      - Bottom coordinate of clip rectangle
bitsPerPixel    - Pixel depth for the bitmap to copy
pixelFormat     - Pixel format for the bitmap to copy
dstPal          - Color palette for the destination device (if dest is 8bpp)
srcPal          - Color palette for the bitmap to copy (if source is 8bpp)
dither          - True if the output should be dithered
mix             - Mix code for the copy (GA_mixCodesType)
flipY           - True if the bitmap should be INVERT in Y

RETURNS:
0 on success, -1 on failure.
****************************************************************************/
N_int32 NAPI INVERT_StretchConvertBltSys_Obsolete(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    N_int32 doClip,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pixelFormat,
    GA_palette *dstPal,
    GA_palette *srcPal,
    N_int32 dither,
    N_int32 mix,
    N_int32 flipY)
{
    GA_palette  pal[256],*psrc,*pdst;
    int         count;

    switch (bitsPerPixel) {
        case 4:
            count = 16;
            psrc = srcPal;
            pdst = pal;
            while (count--)
                *((ulong*)pdst++) = *((ulong*)psrc++) ^ 0xFFFFFF;
            break;
        case 8:
            count = 256;
            psrc = srcPal;
            pdst = pal;
            while (count--)
                *((ulong*)pdst++) = *((ulong*)psrc++) ^ 0xFFFFFF;
            break;
        case 15:
        case 16:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*2;
            InvertBitmap(DC.sysBuffer,srcAddr,srcPitch,srcPitch,srcWidth*2,srcHeight,false);
            break;
        case 24:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*3;
            InvertBitmap(DC.sysBuffer,srcAddr,srcPitch,srcPitch,srcWidth*3,srcHeight,false);
            break;
        case 32:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*4;
            InvertBitmap(DC.sysBuffer,srcAddr,srcPitch,srcPitch,srcWidth*4,srcHeight,false);
            break;
        }
    return DRAW2D.StretchConvertBltSys_Obsolete(DC.sysBuffer,srcPitch,0,0,
        srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,
        doClip,clipLeft,clipTop,clipRight,clipBottom,
        bitsPerPixel,pixelFormat,dstPal,pal,dither,mix,flipY);
}

/****************************************************************************
PARAMETERS:
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstWidth    - Width of the destination rectangle in pixels
dstHeight   - Height of the destination rectangle in scanlines
fx          - GA_bltFx structure describing the requested effects
****************************************************************************/
void NAPI INVERT_BitBltFx(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    GA_bltFx *fx)
{
    // Invert color keys if necessary
    if (fx->Flags & gaBltAnyColorKey) {
        fx->ColorKeyLo = fx->ColorKeyLo ^ DC.colorMask;
        fx->ColorKeyHi = fx->ColorKeyHi ^ DC.colorMask;
        }
    DRAW2D.BitBltFx(srcLeft,srcTop,srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,fx);
}

/****************************************************************************
PARAMETERS:
srcOfs      - Offset of source bitmap in video memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstWidth    - Width of the destination rectangle in pixels
dstHeight   - Height of the destination rectangle in scanlines
fx          - GA_bltFx structure describing the requested effects
****************************************************************************/
void NAPI INVERT_BitBltFxLin(
    N_int32 srcOfs,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    GA_bltFx *fx)
{
    // Invert color keys if necessary
    if (fx->Flags & gaBltAnyColorKey) {
        fx->ColorKeyLo = fx->ColorKeyLo ^ DC.colorMask;
        fx->ColorKeyHi = fx->ColorKeyHi ^ DC.colorMask;
        }
    DRAW2D.BitBltFxLin(srcOfs,srcPitch,srcLeft,srcTop,srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,fx);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstWidth    - Width of the destination rectangle in pixels
dstHeight   - Height of the destination rectangle in scanlines
fx          - GA_bltFx structure describing the requested effects
****************************************************************************/
void NAPI INVERT_BitBltFxSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    GA_bltFx *fx)
{
    GA_palette  pal[256],*psrc,*pdst;
    int         count;

    // Invert color keys if necessary
    if (fx->Flags & gaBltAnyColorKey) {
        fx->ColorKeyLo = fx->ColorKeyLo ^ DC.colorMask;
        fx->ColorKeyHi = fx->ColorKeyHi ^ DC.colorMask;
        }
    if (fx->Flags & gaBltConvert) {
        // Invert bitmap data with a different color depth
        switch (fx->BitsPerPixel) {
            case 4:
                count = 16;
                psrc = fx->SrcPalette;
                pdst = pal;
                while (count--)
                    *((ulong*)pdst++) = *((ulong*)psrc++) ^ 0xFFFFFF;
                fx->SrcPalette = pal;
                break;
            case 8:
                count = 256;
                psrc = fx->SrcPalette;
                pdst = pal;
                while (count--)
                    *((ulong*)pdst++) = *((ulong*)psrc++) ^ 0xFFFFFF;
                fx->SrcPalette = pal;
                break;
            case 15:
            case 16:
                srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*2;
                InvertBitmap(DC.sysBuffer,srcAddr,srcPitch,srcPitch,srcWidth*2,srcHeight,false);
                break;
            case 24:
                srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*3;
                InvertBitmap(DC.sysBuffer,srcAddr,srcPitch,srcPitch,srcWidth*3,srcHeight,false);
                break;
            case 32:
                srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*4;
                InvertBitmap(DC.sysBuffer,srcAddr,srcPitch,srcPitch,srcWidth*4,srcHeight,false);
                break;
            }
        }
    else {
        // Invert bitmap data with a same color depth as display
        switch (DC.mi.BitsPerPixel) {
            case 15:
            case 16:
                srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*2;
                InvertBitmap(DC.sysBuffer,srcAddr,srcPitch,srcPitch,srcWidth*2,srcHeight,false);
                break;
            case 24:
                srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*3;
                InvertBitmap(DC.sysBuffer,srcAddr,srcPitch,srcPitch,srcWidth*3,srcHeight,false);
                break;
            case 32:
                srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*4;
                InvertBitmap(DC.sysBuffer,srcAddr,srcPitch,srcPitch,srcWidth*4,srcHeight,false);
                break;
            }
        }
    DRAW2D.BitBltFxSys(srcAddr,srcPitch,srcLeft,srcTop,srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,fx);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPhysAddr - Physical address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstWidth    - Width of the destination rectangle in pixels
dstHeight   - Height of the destination rectangle in scanlines
fx          - GA_bltFx structure describing the requested effects
****************************************************************************/
void NAPI INVERT_BitBltFxBM(
    void *srcAddr,
    N_int32 srcPhysAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    GA_bltFx *fx)
{
    GA_palette  pal[256],*psrc,*pdst;
    int         count;

    // Invert color keys if necessary
    (void)srcPhysAddr;
    if (fx->Flags & gaBltAnyColorKey) {
        fx->ColorKeyLo = fx->ColorKeyLo ^ DC.colorMask;
        fx->ColorKeyHi = fx->ColorKeyHi ^ DC.colorMask;
        }
    if (fx->Flags & gaBltConvert) {
        // Invert bitmap data with a different color depth
        switch (fx->BitsPerPixel) {
            case 4:
                count = 16;
                psrc = fx->SrcPalette;
                pdst = pal;
                while (count--)
                    *((ulong*)pdst++) = *((ulong*)psrc++) ^ 0xFFFFFF;
                fx->SrcPalette = pal;
                break;
            case 8:
                count = 256;
                psrc = fx->SrcPalette;
                pdst = pal;
                while (count--)
                    *((ulong*)pdst++) = *((ulong*)psrc++) & 0xFFFFFF;
                fx->SrcPalette = pal;
                break;
            case 15:
            case 16:
                srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*2;
                InvertBitmap(DC.sysBuffer,srcAddr,srcPitch,srcPitch,srcWidth*2,srcHeight,false);
                break;
            case 24:
                srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*3;
                InvertBitmap(DC.sysBuffer,srcAddr,srcPitch,srcPitch,srcWidth*3,srcHeight,false);
                break;
            case 32:
                srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*4;
                InvertBitmap(DC.sysBuffer,srcAddr,srcPitch,srcPitch,srcWidth*4,srcHeight,false);
                break;
            }
        }
    else {
        // Invert bitmap data with a same color depth as display
        switch (DC.mi.BitsPerPixel) {
            case 15:
            case 16:
                srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*2;
                InvertBitmap(DC.sysBuffer,srcAddr,srcPitch,srcPitch,srcWidth*2,srcHeight,false);
                break;
            case 24:
                srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*3;
                InvertBitmap(DC.sysBuffer,srcAddr,srcPitch,srcPitch,srcWidth*3,srcHeight,false);
                break;
            case 32:
                srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*4;
                InvertBitmap(DC.sysBuffer,srcAddr,srcPitch,srcPitch,srcWidth*4,srcHeight,false);
                break;
            }
        }
    INVERT_BitBltFxSys(srcAddr,srcPitch,srcLeft,srcTop,srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,fx);
}

/****************************************************************************
PARAMETERS:
dstAddr     - Address of destination bitmap in system memory
dstPhysAddr - Physical address of destination bitmap in system memory
dstPitch    - Pitch of destination bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
****************************************************************************/
void NAPI INVERT_GetBitmapSys(
    void *dstAddr,
    N_int32 dstPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix)
{
    void    *src;

    // TODO: Shouldn't we be using dstLeft and dstTop?
    // TODO: Need to implement mixes!!
    (void)dstLeft;
    (void)dstTop;
    switch (DC.mi.BitsPerPixel) {
        case 15:
        case 16:
            src = (uchar*)DC2D.surface + DC2D.mi.bytesPerLine * srcTop + srcLeft*2;
            InvertBitmap(dstAddr,src,dstPitch,dstPitch,width*2,height,false);
            break;
        case 24:
            src = (uchar*)DC2D.surface + DC2D.mi.bytesPerLine * srcTop + srcLeft*3;
            InvertBitmap(dstAddr,src,dstPitch,dstPitch,width*3,height,false);
            break;
        case 32:
            src = (uchar*)DC2D.surface + DC2D.mi.bytesPerLine * srcTop + srcLeft*4;
            InvertBitmap(dstAddr,src,dstPitch,dstPitch,width*4,height,false);
            break;
        }
}

/****************************************************************************
PARAMETERS:
x   - X coordinate to read pixel from
y   - Y coordinate to read pixel from

RETURNS:
Color read from the pixel on the screen
****************************************************************************/
GA_color NAPI INVERT_GetPixel(
    N_int32 x,
    N_int32 y)
{
    return ~DRAW2D.GetPixel(x,y) & DC.colorMask;
}
