/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the multi-controller support
*               routines in the SNAP Reference Rasteriser.
*
****************************************************************************/

#include "mulint.h"

// TODO: Line clipping is not working properly, specifically lines
//       clipped by the MGL (ie: clipBresenhamLine).

// TODO: BitBltMasked!

/*--------------------------- Global variables ----------------------------*/

MULDISP_state   MULDISP_devCtx;

REF2D_driver MULDISP_drv = {
    sizeof(MULDISP_drv),
    MULDISP_QueryFunctions,
    MULDISP_SetDrawBufferSW,
    MULDISP_SetDrawSurface,
    NULL,                       // MULDISP_RotateBitmap,
    MULDISP_UnloadDriver,
    MULDISP_InitFilter,
    MULDISP_DeInitFilter,
    MULDISP_GlobalOptionsUpdated,
    };

static GA_initFuncs MULDISP_initPT = {
    sizeof(GA_initFuncs),
    MULDISP_GetConfigInfo,
    NULL,                       // Reserved
    MULDISP_SetModeProfile,
    MULDISP_GetOptions,
    MULDISP_SetOptions,
    MULDISP_GetVideoModeInfo,
    MULDISP_SetVideoMode,
    MULDISP_GetVideoMode,
    MULDISP_GetCustomVideoModeInfo,
    MULDISP_SetCustomVideoMode,
    MULDISP_GetClosestPixelClockV1,
    MULDISP_GetCRTCTimings,
    MULDISP_SetCRTCTimings,
    MULDISP_SetGlobalRefresh,
    MULDISP_SaveRestoreState,
    MULDISP_SetDisplayOutput,
    MULDISP_GetDisplayOutput,
    MULDISP_SetSoftwareRenderFuncs,
    MULDISP_GetUniqueFilename,
    NULL,                       // Reserved
    MULDISP_GetMonitorInfo,
    MULDISP_SetMonitorInfo,
    MULDISP_GetCurrentVideoModeInfo,
    NULL,                       // TODO: MULDISP_GetCertifyInfo,
    NULL,                       // TODO: MULDISP_SetCRTCTimings,
    NULL,                       // TODO: MULDISP_AlignLinearBuffer,
    NULL,                       // TODO: MULDISP_GetIORegisterTraps,
    NULL,                       // TODO: MULDISP_GetCurrentRefreshRate,
    NULL,                       // TODO: MULDISP_PollForDisplaySwitch,
    NULL,                       // TODO: MULDISP_PerformDisplaySwitch,
    NULL,                       // TODO: MULDISP_GetVideoModeInfoExt,
    NULL,                       // TODO: MULDISP_GetCustomVideoModeInfoExt,
    NULL,                       // TODO: MULDISP_SwitchPhysicalResolution
    NULL,                       // TODO: MULDISP_GetNumberOfHeads
    NULL,                       // TODO: MULDISP_SetActiveHead
    NULL,                       // TODO: MULDISP_GetActiveHead
    NULL,                       // TODO: MULDISP_SetRef2dPointer
    MULDISP_SetDisplayOutputExt,
    MULDISP_GetDisplayOutputExt,
    MULDISP_GetClosestPixelClock,
    };

static GA_driverFuncs MULDISP_driverPT = {
    sizeof(GA_driverFuncs),
    MULDISP_SetBank,
    MULDISP_SetDisplayStart,
    MULDISP_SetDisplayStartXY,
    MULDISP_SetStereoDisplayStart,
    MULDISP_GetDisplayStartStatus,
    MULDISP_EnableStereoMode,
    MULDISP_IsVSync,
    MULDISP_WaitVSync,
    MULDISP_SetVSyncWidth,
    MULDISP_GetVSyncWidth,
    MULDISP_SetPaletteData,
    MULDISP_GetPaletteData,
    MULDISP_SetGammaCorrectData,
    MULDISP_GetGammaCorrectData,
    NULL,                       // MULDISP_GetCurrentScanLine,
    MULDISP_SetPaletteDataExt,
    MULDISP_GetPaletteDataExt,
    MULDISP_SetGammaCorrectDataExt,
    MULDISP_GetGammaCorrectDataExt,
    };

static GA_cursorFuncs MULDISP_cursorPT = {
    sizeof(GA_cursorFuncs),
    MULDISP_SetMonoCursor,
    MULDISP_SetMonoCursorColor,
    MULDISP_SetColorCursor,
    MULDISP_SetCursorPos,
    MULDISP_ShowCursor,
    MULDISP_BeginAccess,
    MULDISP_EndAccess,
    MULDISP_SetColorCursor256,
    MULDISP_SetColorCursorRGBA,
    MULDISP_IsHardwareCursor,
    MULDISP_SetColorCursorRGB,
    };

static GA_2DStateFuncs MULDISP_state2dPT = {
    sizeof(GA_2DStateFuncs),
    MULDISP_SetDrawBuffer,
    MULDISP_IsIdle,
    MULDISP_WaitTillIdle,
    NULL,
    NULL,
    MULDISP_SetMix,
    MULDISP_SetForeColor,
    MULDISP_SetBackColor,
    MULDISP_Set8x8MonoPattern,
    MULDISP_Use8x8MonoPattern,
    MULDISP_Use8x8TransMonoPattern,
    MULDISP_Set8x8ColorPattern,
    MULDISP_Use8x8ColorPattern,
    MULDISP_Use8x8TransColorPattern,
    MULDISP_SetLineStipple,
    MULDISP_SetLineStippleCount,
    MULDISP_SetPlaneMask,
    MULDISP_SetSrcBlendFunc,
    MULDISP_SetDstBlendFunc,
    MULDISP_SetAlphaValue,
    NULL,                       // DRIVER_SetupForWindows
    NULL,                       // DRIVER_ResetWindows
    MULDISP_SetLineStyle,
    NULL,                       // REF2D_BuildTranslateVector,
    MULDISP_SetBlendFunc,
    };

static GA_2DRenderFuncs MULDISP_draw2dPT = {
    sizeof(GA_2DRenderFuncs),
    MULDISP_GetPixel,
    MULDISP_PutPixel,
    MULDISP_DrawScanList,
    MULDISP_DrawPattScanList,
    MULDISP_DrawColorPattScanList,
    MULDISP_DrawEllipseList,
    MULDISP_DrawPattEllipseList,
    MULDISP_DrawColorPattEllipseList,
    MULDISP_DrawFatEllipseList,
    MULDISP_DrawPattFatEllipseList,
    MULDISP_DrawColorPattFatEllipseList,
    MULDISP_DrawRect,
    NULL,
    MULDISP_DrawPattRect,
    MULDISP_DrawColorPattRect,
    MULDISP_DrawTrap,
    MULDISP_DrawPattTrap,
    MULDISP_DrawColorPattTrap,
    MULDISP_DrawLineInt,
    MULDISP_DrawBresenhamLine,
    MULDISP_DrawStippleLineInt,
    MULDISP_DrawBresenhamStippleLine,
    MULDISP_DrawEllipse,
    MULDISP_ClipEllipse,
    MULDISP_PutMonoImageMSBSys,
    NULL,                       // MULDISP_PutMonoImageMSBLin,
    NULL,                       // MULDISP_PutMonoImageMSBBM,
    NULL,                       // MULDISP_PutMonoImageLSBSys,
    NULL,                       // MULDISP_PutMonoImageLSBLin,
    NULL,                       // MULDISP_PutMonoImageLSBBM,
    MULDISP_ClipMonoImageMSBSys,
    NULL,                       // MULDISP_ClipMonoImageMSBLin,
    NULL,                       // MULDISP_ClipMonoImageMSBBM,
    NULL,                       // MULDISP_ClipMonoImageLSBSys,
    NULL,                       // MULDISP_ClipMonoImageLSBLin,
    NULL,                       // MULDISP_ClipMonoImageLSBBM,
    MULDISP_BitBlt,
    NULL,                       // MULDISP_BitBltLin,
    MULDISP_BitBltSys,
    NULL,                       // MULDISP_BitBltBM,
    MULDISP_BitBltPatt,
    NULL,                       // MULDISP_BitBltPattLin,
    MULDISP_BitBltPattSys,
    NULL,                       // MULDISP_BitBltPattBM,
    MULDISP_BitBltColorPatt,
    NULL,                       // MULDISP_BitBltColorPattLin,
    MULDISP_BitBltColorPattSys,
    NULL,                       // MULDISP_BitBltColorPattBM,
    MULDISP_SrcTransBlt,
    NULL,                       // MULDISP_SrcTransBltLin,
    MULDISP_SrcTransBltSys,
    NULL,                       // MULDISP_SrcTransBltBM,
    MULDISP_DstTransBlt,
    NULL,                       // MULDISP_DstTransBltLin,
    MULDISP_DstTransBltSys,
    NULL,                       // MULDISP_DstTransBltBM,
    NULL,                       // MULDISP_StretchBlt,
    NULL,                       // MULDISP_StretchBltLin,
    MULDISP_StretchBltSys,
    NULL,                       // MULDISP_StretchBltBM,
    MULDISP_ConvertBltSys_Obsolete,
    NULL,                       // MULDISP_ConvertBltBM_Obsolete,
    MULDISP_StretchConvertBltSys_Obsolete,
    NULL,                       // MULDISP_StretchConvertBltBM_Obsolete,
    NULL,                       // MULDISP_BitBltFxTest,
    NULL,                       // MULDISP_BitBltFx,
    NULL,                       // MULDISP_BitBltFxLin,
    NULL,                       // MULDISP_BitBltFxSys,
    NULL,                       // MULDISP_BitBltFxBM,
    MULDISP_GetBitmapSys,
    NULL,                       // MULDISP_GetBitmapBM,
    NULL,                       // MULDISP_UpdateScreen,
    MULDISP_DrawClippedLineInt,
    MULDISP_DrawClippedBresenhamLine,
    MULDISP_DrawClippedStippleLineInt,
    MULDISP_DrawClippedBresenhamStippleLine,
    NULL,                       // MULDISP_BitBltPlaneMasked,
    NULL,                       // MULDISP_BitBltPlaneMaskedLin,
    NULL,                       // MULDISP_BitBltPlaneMaskedSys,
    NULL,                       // MULDISP_BitBltPlaneMaskedBM,
    NULL,                       // MULDISP_DrawRectLin,
    NULL,                       // MULDISP_DrawRectExt,
    MULDISP_DrawStyleLineInt,
    MULDISP_DrawBresenhamStyleLine,
    MULDISP_DrawClippedStyleLineInt,
    MULDISP_DrawClippedBresenhamStyleLine,
    };

static GA_2DRenderFuncs MULDISP_draw2dStub = {
    sizeof(GA_2DRenderFuncs),
    MSTUB_GetPixel,
    MSTUB_PutPixel,
    MSTUB_DrawScanList,
    MSTUB_DrawPattScanList,
    MSTUB_DrawColorPattScanList,
    MSTUB_DrawEllipseList,
    MSTUB_DrawPattEllipseList,
    MSTUB_DrawColorPattEllipseList,
    MSTUB_DrawFatEllipseList,
    MSTUB_DrawPattFatEllipseList,
    MSTUB_DrawColorPattFatEllipseList,
    MSTUB_DrawRect,
    NULL,
    MSTUB_DrawPattRect,
    MSTUB_DrawColorPattRect,
    MSTUB_DrawTrap,
    MSTUB_DrawPattTrap,
    MSTUB_DrawColorPattTrap,
    MSTUB_DrawLineInt,
    MSTUB_DrawBresenhamLine,
    MSTUB_DrawStippleLineInt,
    MSTUB_DrawBresenhamStippleLine,
    MSTUB_DrawEllipse,
    MSTUB_ClipEllipse,
    MSTUB_PutMonoImageMSBSys,
    NULL,                       // MSTUB_PutMonoImageMSBLin,
    NULL,                       // MSTUB_PutMonoImageMSBBM,
    NULL,                       // MSTUB_PutMonoImageLSBSys,
    NULL,                       // MSTUB_PutMonoImageLSBLin,
    NULL,                       // MSTUB_PutMonoImageLSBBM,
    MSTUB_ClipMonoImageMSBSys,
    NULL,                       // MSTUB_ClipMonoImageMSBLin,
    NULL,                       // MSTUB_ClipMonoImageMSBBM,
    NULL,                       // MSTUB_ClipMonoImageLSBSys,
    NULL,                       // MSTUB_ClipMonoImageLSBLin,
    NULL,                       // MSTUB_ClipMonoImageLSBBM,
    MSTUB_BitBlt,
    NULL,                       // MSTUB_BitBltLin,
    MSTUB_BitBltSys,
    NULL,                       // MSTUB_BitBltBM,
    MSTUB_BitBltPatt,
    NULL,                       // MSTUB_BitBltPattLin,
    MSTUB_BitBltPattSys,
    NULL,                       // MSTUB_BitBltPattBM,
    MSTUB_BitBltColorPatt,
    NULL,                       // MSTUB_BitBltColorPattLin,
    MSTUB_BitBltColorPattSys,
    NULL,                       // MSTUB_BitBltColorPattBM,
    MSTUB_SrcTransBlt,
    NULL,                       // MSTUB_SrcTransBltLin,
    MSTUB_SrcTransBltSys,
    NULL,                       // MSTUB_SrcTransBltBM,
    MSTUB_DstTransBlt,
    NULL,                       // MSTUB_DstTransBltLin,
    MSTUB_DstTransBltSys,
    NULL,                       // MSTUB_DstTransBltBM,
    NULL,                       // MSTUB_StretchBlt,
    NULL,                       // MSTUB_StretchBltLin,
    MSTUB_StretchBltSys,
    NULL,                       // MSTUB_StretchBltBM,
    MSTUB_ConvertBltSys_Obsolete,
    NULL,                       // MSTUB_ConvertBltBM_Obsolete,
    MSTUB_StretchConvertBltSys_Obsolete,
    NULL,                       // MSTUB_StretchConvertBltBM_Obsolete,
    NULL,                       // MSTUB_BitBltFxTest,
    NULL,                       // MSTUB_BitBltFx,
    NULL,                       // MSTUB_BitBltFxLin,
    NULL,                       // MSTUB_BitBltFxSys,
    NULL,                       // MSTUB_BitBltFxBM,
    MSTUB_GetBitmapSys,
    NULL,                       // MSTUB_GetBitmapBM,
    NULL,                       // MSTUB_UpdateScreen,
    MSTUB_DrawClippedLineInt,
    MSTUB_DrawClippedBresenhamLine,
    MSTUB_DrawClippedStippleLineInt,
    MSTUB_DrawClippedBresenhamStippleLine,
    NULL,                       // MSTUB_BitBltPlaneMasked,
    NULL,                       // MSTUB_BitBltPlaneMaskedLin,
    NULL,                       // MSTUB_BitBltPlaneMaskedSys,
    NULL,                       // MSTUB_BitBltPlaneMaskedBM,
    NULL,                       // MSTUB_DrawRectLin,
    NULL,                       // MSTUB_DrawRectExt,
    MSTUB_DrawStyleLineInt,
    MSTUB_DrawBresenhamStyleLine,
    MSTUB_DrawClippedStyleLineInt,
    MSTUB_DrawClippedBresenhamStyleLine,
    };

GA_2DStateFuncs     _VARAPI MULDISP_state2dCnt;
GA_2DRenderFuncs    _VARAPI MULDISP_draw2dCnt;

#ifdef  __DRIVER__
PM_imports _VARAPI _PM_imports;
#endif

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
REMARKS:
Function to initialise the multi-controller filter driver.
****************************************************************************/
static ibool MULDISP_init(
    GA_devCtx **hwCtx,
    N_int32 numDevices,
    GA_globalOptions *gOpt,
    GA_setActiveDevice_t setActiveDevice)
{
    int         i;
    N_uint16    *p,*q;

    // Initialise the internal state of the filter driver for all devices
    DC.gOpt = gOpt;
    DC.enabled = false;
    DC.numDevices = numDevices;
    DC.setActiveDevice = setActiveDevice;
    DC.cntMode = 0;
    for (i = 0; i < numDevices; i++) {
        DC.hwCtx[i] = hwCtx[i];
        DC.init[i].dwSize = sizeof(DC.init[i]);
        if (!hwCtx[i]->loader.QueryFunctions(GA_GET_INITFUNCS,0,&DC.init[i]))
            return false;
        DC.driver[i].dwSize = sizeof(DC.driver[i]);
        if (!hwCtx[i]->loader.QueryFunctions(GA_GET_DRIVERFUNCS,0,&DC.driver[i]))
            return false;
        DC.AvailableModesDC[i] = hwCtx[i]->AvailableModes;
        }

    // Hook the device context QueryFunctions function so we can return
    // our hooked functions. We only need to hook this for the first
    // device driver, since the others are only managed intenally and never
    // accessed directly by shell drivers.
    DC.QueryFunctionsDC = hwCtx[0]->loader.QueryFunctions;
    hwCtx[0]->loader.QueryFunctions = MULDISP_QueryFunctionsDC;
    hwCtx[0]->AvailableModes = DC.AvailableModes;

    // Create the list of modes, which is identical to the primary
    // controller list, but with a new 'virtual' mode added at the end. We
    // always number our modes from 0x100 up, so we simply make the new
    // mode number one bigger than the last mode in the list.
    for (p = DC.AvailableModesDC[0], q = DC.AvailableModes; *p != 0xFFFF;) {
        *q++ = *p++;
        }
    p--;
    *q++ = DC.virtualMode8 = *p + 1;
    *q++ = DC.virtualMode15 = *p + 2;
    *q++ = DC.virtualMode16 = *p + 3;
    *q++ = DC.virtualMode24 = *p + 4;
    *q++ = DC.virtualMode32 = *p + 5;
    *q = 0xFFFF;
    return true;
}

#ifdef  __DRIVER__
/****************************************************************************
REMARKS:
Fatal error handler called when a non-imported function is called by the
driver. We leave this to a runtime error so that older applications and
shell drivers will work with newer bpd drivers provided no newer functions
are required by the driver itself. If they are, the application or shell
driver needs to be recompiled.
****************************************************************************/
static void _PM_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported PM_imports import function called! Please re-compile!\n");
}

/****************************************************************************
REMARKS:
Function entry point for the reference rasteriser library.
****************************************************************************/
REF2D_driver * _CEXPORT MULDISP_initLibrary(
    GA_devCtx **hwCtx,
    N_int32 numDevices,
    GA_globalOptions *gOpt,
    PM_imports *pmImp,
    GA_setActiveDevice_t setActiveDevice)
{
    int     i,max;
    ulong   *p;

    // Initialize all default imports to point to fatal error handler
    // for upwards compatibility.
    max = sizeof(_PM_imports)/sizeof(GA_initLibrary_t);
    for (i = 0,p = (ulong*)&_PM_imports; i < max; i++)
        *p++ = (ulong)_PM_fatalErrorHandler;
    memcpy(&_PM_imports,pmImp,MIN(sizeof(_PM_imports),pmImp->dwSize));
    if (!MULDISP_init(hwCtx,numDevices,gOpt,setActiveDevice))
        return NULL;
    return &DRV;
}
#else
/****************************************************************************
REMARKS:
Function entry point to fake loading of a driver when linked into app.
****************************************************************************/
ibool NAPI MULDISP_loadDriver(
    GA_devCtx **hwCtx,
    N_int32 numDevices,
    GA_globalOptions *gOpt,
    REF2D_driver **drv,
    PE_MODULE **hModule,
    ulong *size,
    GA_setActiveDevice_t setActiveDevice)
{
    if (!MULDISP_init(hwCtx,numDevices,gOpt,setActiveDevice))
        return false;
    *drv = &DRV;
    *hModule = NULL;
    *size = 0;
    return true;
}
#endif

/****************************************************************************
REMARKS:
Enabled or disabled the invert driver internally.
****************************************************************************/
static void MulDispEnable(void)
{
    if (DC.enabled) {
        MULDISP_state2dCnt = MULDISP_state2dPT;
        MULDISP_draw2dCnt = MULDISP_draw2dPT;
        }
    else {
        MULDISP_state2dCnt = DC.state2d[0];
        MULDISP_draw2dCnt = DC.draw2d[0];
        }
}

/****************************************************************************
REMARKS:
Function to initialise the MULDISP display filter driver. This function
gets called long after the driver has actually been loaded, at the
time when the reference rasteriser module is loaded and initialsied (ie:
at mode init time).
****************************************************************************/
ibool NAPI MULDISP_InitFilter(
    GA_modeInfo *modeInfo,
    N_int32 transferStart,
    struct _REF2D_driver *ref2d,
    struct _REF2D_driver **drv)
{
    // We do nothing in here
    (void)modeInfo;
    (void)transferStart;
    (void)ref2d;
    (void)drv;
    return true;
}

/****************************************************************************
REMARKS:
Function to initialise the MULDISP display filter driver. This function
gets called long after the driver has actually been loaded, at the
time when the reference rasteriser module is loaded and initialsied (ie:
at mode init time).
****************************************************************************/
#ifdef __DRIVER__
ibool _CEXPORT MULDISP_InitFilterExt(
#else
ibool NAPI MULDISP_InitFilterExt(
#endif
    GA_modeInfo *modeInfo,
    N_int32 transferStart,
    struct _REF2D_driver **ref2d,
    struct _REF2D_driver **drv)
{
    int i;

    // Initialise the internal state of the filter driver
    DC.gMi = *modeInfo;
    DC.enabled = IS_MULDISP() && IS_MULDISP_MODE();
    DC.isCursorVisible = false;
    DC.cntCursorX = 0;
    DC.cntCursorY = 0;
    DC.hwCtx[0]->loader.QueryFunctions = MULDISP_QueryFunctionsDC;
    DC.hwCtx[0]->AvailableModes = DC.AvailableModes;
    for (i = 0; i < DC.numDevices; i++) {
        DC.ref2d[i] = ref2d[i];
        DC.refState[i] = ref2d[i]->DriverState;
        DC.init[i].dwSize = sizeof(DC.init[i]);
        if (!ref2d[i]->QueryFunctions(GA_GET_INITFUNCS,&DC.init[i]))
            return false;
        DC.driver[i].dwSize = sizeof(DC.driver[i]);
        if (!ref2d[i]->QueryFunctions(GA_GET_DRIVERFUNCS,&DC.driver[i]))
            return false;
        DC.cursor[i].dwSize = sizeof(DC.cursor[i]);
        if (!ref2d[i]->QueryFunctions(GA_GET_CURSORFUNCS,&DC.cursor[i]))
            return false;
        DC.state2d[i].dwSize = sizeof(DC.state2d[i]);
        if (!ref2d[i]->QueryFunctions(GA_GET_2DSTATEFUNCS,&DC.state2d[i]))
            return false;
        DC.draw2d[i].dwSize = sizeof(DC.draw2d[i]);
        if (!ref2d[i]->QueryFunctions(GA_GET_2DRENDERFUNCS,&DC.draw2d[i]))
            return false;
        DC.mi[i].dwSize = sizeof(DC.mi[i]);
        DC.init[0].GetCurrentVideoModeInfo(&DC.mi[i]);
        DC.cursorVisible[i] = 0;
        }

    // Allocate memory for system memory transfer buffer. This buffer is
    // the same size as the entire virtual screen dimensions, and is used
    // as a temporary blit buffer when copying data between display devices.
    switch (DC.gMi.BitsPerPixel) {
        case 8:
            DC.bufferPitch = DC.gOpt->virtualSize.right;
            break;
        case 15:
        case 16:
            DC.bufferPitch = DC.gOpt->virtualSize.right * 2;
            break;
        case 24:
            DC.bufferPitch = DC.gOpt->virtualSize.right * 3;
            break;
        case 32:
            DC.bufferPitch = DC.gOpt->virtualSize.right * 4;
            break;
        }
    DC.bufferSize = DC.gOpt->virtualSize.bottom * DC.bufferPitch;
    if ((DC.buffer = malloc(DC.bufferSize)) == NULL)
        return false;
    memset(DC.buffer,0,DC.bufferSize);

    /* Initialise all the filter driver vectors */
    MulDispEnable();

    /* Return address of our filter driver as top of filter chain */
    *drv = &DRV;
    return true;
}

/****************************************************************************
REMARKS:
Function to clean up when unloading the driver.
****************************************************************************/
void NAPI MULDISP_UnloadDriver(void)
{
}

/****************************************************************************
REMARKS:
Function to clean up when the filter driver needs to de-initialise (ie:
going back to text mode when the ref2d driver is unloaded).
****************************************************************************/
void NAPI MULDISP_DeInitFilter(void)
{
    int i;

    // Unhook our QueryFunctions hook, and restore the init vectors
    DC.hwCtx[0]->loader.QueryFunctions = DC.QueryFunctionsDC;
    for (i = 0; i < DC.numDevices; i++) {
        DC.init[i].dwSize = sizeof(DC.init[i]);
        DC.hwCtx[i]->loader.QueryFunctions(GA_GET_INITFUNCS,0,&DC.init[i]);
        DC.driver[i].dwSize = sizeof(DC.driver[i]);
        DC.hwCtx[i]->loader.QueryFunctions(GA_GET_DRIVERFUNCS,0,&DC.driver[i]);
        }

    // Free our system memory transfer buffer
    free(DC.buffer);
}

/****************************************************************************
PARAMETERS:
funcs       - Pointer to the function list to make valid
size        - Size of the function list in bytes

RETURNS:
True if valid, false if not.
****************************************************************************/
static ibool MakeValid(
    ulong *funcs,
    int size)
{
    int i,empty = true;

    for (i = 0; i < size/4; i++) {
        if (funcs[i] != 0)
            empty = false;
        }
    return !empty;
}

/****************************************************************************
DESCRIPTION:
Returns the device driver entry points for the specific function group.

PARAMETERS:
id          - Identifier for the function group to get pointers for
safetyLevel - Safety level being requested.
funcs       - Place to store the function block

RETURNS:
True if functions available, false if not.

REMARKS:
This function allows the filter driver to hook the device driver functions
that are called before any modes are initialised. This is used to hook
functions such as GetVideoModeInfo and SetVideoMode as necessary in the
filter driver.

NOTE: If we get a safetyLevel value of less than zero, it indicates that
      we a being called by some other filter driver and should simply
      pass the call onto the low level hardware driver. Otherwise we may
      end up getting into an infinite loop!
****************************************************************************/
ibool NAPI MULDISP_QueryFunctionsDC(
    N_uint32 id,
    N_int32 safetyLevel,
    void *funcs)
{
    ulong   dwSize = *((ulong*)funcs);

    memset(funcs,0,dwSize);
    if (safetyLevel >= 0) {
        switch (id) {
            case GA_GET_INITFUNCS:      RETURN_VALID(MULDISP_initPT);
            case GA_GET_DRIVERFUNCS:    RETURN_VALID(MULDISP_driverPT);
            }
        }
    *((ulong*)funcs) = dwSize;
    return DC.QueryFunctionsDC(id,safetyLevel,funcs);
}

/****************************************************************************
DESCRIPTION:
Returns the reference rasterisers for the specified function group.

PARAMETERS:
id      - Identifier for the function group to get pointers for
funcs   - Place to store the function block

RETURNS:
True if functions available, false if not.

REMARKS:
This function is the main function that is used by the graphics application
code to get a block of function pointers for a specified function group.
The function groups are defined GA_funcGroupsType enumeration, and breaks
up the functions in the device driver API into groups of logically similar
functions.
****************************************************************************/
ibool NAPI MULDISP_QueryFunctions(
    N_uint32 id,
    void *funcs)
{
    ulong   dwSize = *((ulong*)funcs);

    memset(funcs,0,dwSize);
    switch (id) {
        case GA_GET_INITFUNCS:      RETURN_VALID(MULDISP_initPT);
        case GA_GET_DRIVERFUNCS:    RETURN_VALID(MULDISP_driverPT);
        case GA_GET_CURSORFUNCS:    RETURN_VALID(MULDISP_cursorPT);
        case GA_GET_2DSTATEFUNCS:   RETURN_VALID(MULDISP_state2dPT);
        case GA_GET_2DRENDERFUNCS:  RETURN_VALID(MULDISP_draw2dStub);
        }
    *((ulong*)funcs) = dwSize;
    return false;
}

/****************************************************************************
DESCRIPTION:
Sets a display buffer as the active drawing buffer.

PARAMETERS:
drawBuf             - Buffer to make the active drawing buffer
framebuffer         - Pointer to the start of framebuffer memory
bitsPerPixel        - Color depth for the buffer
beginDirectAccess   - Direct framebuffer arbitration function (NULL if none)
endDirectAccess     - Direct framebuffer arbitration function (NULL if none)

RETURNS:
0 on success, -1 on failure.

REMARKS:
This function allows the application to tell make a display memory
buffer the active rendering buffer for all subsequent rendering commands.
****************************************************************************/
N_int32 NAPI MULDISP_SetDrawBufferSW(
    GA_buffer *drawBuf,
    void *framebuffer,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pf,
    GA_devCtx *hwCtx,
    N_int32 forceSoftwareOnly)
{
    int         i;
    GA_buffer   buf;

    DC.enabled = (IS_MULDISP() && IS_MULDISP_MODE() && framebuffer == hwCtx->LinearMem);
    MulDispEnable();
    DC.buf = *drawBuf;
    if (DC.enabled) {
        // We don't support page flipping or arbitrary draw buffers in
        // multi-controller mode, so we chefk for and only handle a full
        // draw buffer.
        if (DC.buf.Offset != 0 ||
            DC.buf.Stride != DC.virtualMi.BytesPerScanLine ||
            DC.buf.Width != DC.virtualMi.XResolution ||
            DC.buf.Height != DC.virtualMi.YResolution)
            return -1;

        // Now set the draw buffer for all devices
        i = DC.numDevices;
        while (i--) {
            buf.dwSize = sizeof(buf);
            buf.Offset = 0;
            buf.Stride = DC.mi[i].BytesPerScanLine;
            buf.Width = DC.bounds[i].right - DC.bounds[i].left;
            buf.Height = DC.bounds[i].bottom - DC.bounds[i].top;
            if (DC.ref2d[i]->SetDrawBuffer(&buf,DC.hwCtx[i]->LinearMem,
                    DC.mi[i].BitsPerPixel,&DC.mi[i].PixelFormat,DC.hwCtx[i],forceSoftwareOnly) != 0)
                return -1;
            }
        return 0;
        }
    else
        return DC.ref2d[0]->SetDrawBuffer(&DC.buf,framebuffer,bitsPerPixel,pf,hwCtx,forceSoftwareOnly);
}

/****************************************************************************
PARAMETERS:
drawBuf - New hardware draw buffer to make active

RETURNS:
0 on success, -1 on failure.

REMARKS:
This function changes the hardware draw buffer for both the software
rasteriser and the hardware device. This can also be used to change the
draw buffer for a software only buffer, but is generally only used when
using hardware.
****************************************************************************/
N_int32 NAPI MULDISP_SetDrawBuffer(
    GA_buffer *drawBuf)
{
    if (!DC.enabled)
        return DC.state2d[0].SetDrawBuffer(&DC.buf);
    return 0;
}

/****************************************************************************
PARAMETERS:
surface         - Pointer to the start of the memory buffer for drawing
xRes            - X resolution for the surface
yRes            - Y resolution for the surface
bytesPerLine    - Scanline width for the memory buffer for drawing
bitsPerPixel    - New color depth for the surface
pf              - New pixel format for the surface

REMARKS:
This function changes the software draw buffer quickly, to allow the
software renderer to render to bitmaps in different locations in memory.
We do in here for speed, since we want to avoid the overhead of setting
the full draw buffer for the reference rasteriser code. If the color
depth or pixel format changes, this function will switch all the internal
rendering vectors to put to the new vectors for the appropriate functions.
This allows the reference rasteriser to work with all color depths active
in a single driver (and for our debug library to work properly).
****************************************************************************/
void NAPI MULDISP_SetDrawSurface(
    void *surface,
    N_int32 xRes,
    N_int32 yRes,
    N_int32 bytesPerLine,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pf)
{
    if (!DC.enabled)
        DC.ref2d[0]->SetDrawSurface(surface,DC.buf.Width,DC.buf.Height,DC.buf.Stride,bitsPerPixel,pf);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_GetConfigInfo(
    GA_configInfo *info)
{
    DC.init[0].GetConfigInfo(info);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_SetModeProfile(
    GA_modeProfile *profile)
{
    DC.init[0].SetModeProfile(profile);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_GetOptions(
    GA_options *options)
{
    DC.init[0].GetOptions(options);
}

/****************************************************************************
REMARKS:
Handles updating of the internals when the options structures are updated.
This allows inversion mode to be toggled automatically at runtime using
either the lobal device driver or global SNAP options functions.
****************************************************************************/
static void UpdateOptions(void)
{
    DC.enabled = IS_MULDISP() && IS_MULDISP_MODE();
    MulDispEnable();
}

/****************************************************************************
REMARKS:
Handles overriding of the options in the driver. This allows us to know
when the invert option has been disabled or enabled at runtime.
****************************************************************************/
void NAPI MULDISP_SetOptions(
    GA_options *options)
{
    DC.init[0].SetOptions(options);
    DC.init[0].GetOptions(&DC.opt);
    UpdateOptions();
}

/****************************************************************************
REMARKS:
This function gets called when the global options are changed at runtime,
allowing the filter driver to update the internals based on the changed
global options.
****************************************************************************/
void NAPI MULDISP_GlobalOptionsUpdated(void)
{
    UpdateOptions();
}

/****************************************************************************
PARAMETERS:
xRes            - Physical pixel X resolution of the display mode
yRes            - Physical line Y resolution of the display mode
virtualX        - Virtual X resolution of the display mode
virtualY        - Virtual Y resolution of the display mode
bitsPerPixel    - Color depth of the display mode
modeInfo        - Structure to fill in with the mode information

RETURNS:
0 on success, -1 on failure

REMARKS:
Filter driver hook to enable multi-controller modes.
****************************************************************************/
N_int32 NAPI MULDISP_GetCustomVideoModeInfo(
    N_int32 xRes,
    N_int32 yRes,
    N_int32 virtualX,
    N_int32 virtualY,
    N_int32 bitsPerPixel,
    GA_modeInfo *modeInfo)
{
    return DC.init[0].GetCustomVideoModeInfo(xRes,yRes,virtualX,virtualY,bitsPerPixel,modeInfo);
}

/****************************************************************************
PARAMETERS:
index   - Index of controller to talk to
xRes    - X resolution of mode to find
yRes    - Y resolution of mode to find
bits    - Color depth of mode to find

RETURNS:
0 on failure, mode number of success

REMARKS:
Searches the mode information list for the controller to find the mode
that matches the desired resolution and color depth.
****************************************************************************/
static N_uint32 FindMode(
    int index,
    int xRes,
    int yRes,
    int bits,
    GA_modeInfo *mi)
{
    N_uint16    *m;

    for (m = DC.AvailableModesDC[index]; *m != 0xFFFF; m++) {
        mi->dwSize = sizeof(*mi);
        if (DC.init[index].GetVideoModeInfo(*m,mi) != 0)
            continue;
        if (mi->XResolution == xRes && mi->YResolution == yRes && mi->BitsPerPixel == bits)
            return *m;
        }
    return 0;
}

/****************************************************************************
PARAMETERS:
mode        - SNAP display mode to get info for
modeInfo    - Structure to fill in with the mode information

RETURNS:
0 on success, -1 on failure

REMARKS:
Filter driver hook to enable multi-controller modes.
****************************************************************************/
N_int32 NAPI MULDISP_GetVideoModeInfo(
    N_uint32 mode,
    GA_modeInfo *modeInfo)
{
    int                 i,bpp,imode = (mode & gaModeMask);
    GA_globalOptions    *g = DC.gOpt;
    GA_modeInfo         mi;
    N_int32             no8bppRGBCursor = gaHave8bppRGBCursor;
    N_int32             noHWCursor = 0;

    if (IS_MULDISP_OPT() && imode >= DC.virtualMode8) {
        // Determine the color depth for this virtual mode
        if (imode == DC.virtualMode8)
            bpp = 8;
        else if (imode == DC.virtualMode15)
            bpp = 15;
        else if (imode == DC.virtualMode16)
            bpp = 16;
        else if (imode == DC.virtualMode24)
            bpp = 24;
        else if (imode == DC.virtualMode32)
            bpp = 32;
        else
            return -1;

        // First get the mode information for all devices based on the
        // configuration information.
        for (i = 0; i < NUM_DEVICES(); i++) {
            // Find the mode for controller based on resolution
            if ((DC.modes[i] = FindMode(i,
                g->resolutions[i].right - g->resolutions[i].left,
                g->resolutions[i].bottom - g->resolutions[i].top,bpp,
                &DC.mi[i])) == 0)
                return -1;

            // Now find the virtual mode information for the controller
            // to ensure that it will fit in the display memory.
            mi.dwSize = sizeof(mi);
            if (DC.init[i].GetCustomVideoModeInfo(
                g->resolutions[i].right - g->resolutions[i].left,
                g->resolutions[i].bottom - g->resolutions[i].top,
                g->bounds[i].right - g->bounds[i].left,
                g->bounds[i].bottom - g->bounds[i].top,
                bpp,&mi) != 0)
                return -1;

            // Enable 8bpp RGB cursor if any card requires this
            if (DC.mi[i].Attributes & gaHave8bppRGBCursor)
                no8bppRGBCursor = 0;

            // Disable hardware cursors if any device does not have one
            if (!(DC.mi[i].Attributes & gaHaveHWCursor))
                noHWCursor          = gaHaveHWCursor;

            // Acceleration must be pure MMIO to support multi-controller!
            if (DC.mi[i].Attributes & gaHaveAccelIOPL)
                return -1;

            // All modes must be linear modes!
            if (!(DC.mi[i].Attributes & gaHaveLinearBuffer))
                return -1;
            }

        // Now copy primary controller mode information block and modify it
        DC.virtualMi = DC.mi[0];
        DC.virtualMi.Attributes         &= ~(gaHaveDisplayStart
                                            | gaHaveTripleBuffer
                                            | gaHaveStereo
                                            | gaHaveAccelVideo
                                            | gaHaveAccel3D
                                            | noHWCursor
                                            | gaHaveLCDOutput
                                            | gaHaveTVOutput
                                            | no8bppRGBCursor);
        DC.virtualMi.Attributes         |= gaIsVirtualMode;
        DC.virtualMi.XResolution        = g->virtualSize.right;
        DC.virtualMi.YResolution        = g->virtualSize.bottom;
        DC.virtualMi.MaxScanLines       = g->virtualSize.bottom;
        DC.virtualMi.MaxBuffers         = 1;
        DC.virtualMi.BitBltCaps         = NULL;
        DC.virtualMi.VideoWindows       = NULL;
        DC.virtualMi.LinearSize         = 0;
        DC.virtualMi.LinearBasePtr      = 0;

        // Now return the mode information block
        COPY_STRUCTURE(modeInfo,&DC.virtualMi);
        return 0;
        }
    else
        return DC.init[0].GetVideoModeInfo(mode,modeInfo);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
N_uint32 NAPI MULDISP_GetVideoMode(void)
{
    if (IS_MULDISP_OPT() && (DC.cntMode & gaModeMask) >= DC.virtualMode8)
        return DC.cntMode;
    else
        return DC.init[0].GetVideoMode();
}

/****************************************************************************
DESCRIPTION:
Sets a custom display mode.

PARAMETERS:
xRes            - X resolution for custom display mode
yRes            - Y resolution for custom display mode
bitsPerPixel    - Color depth for custom display mode (0 = text mode)
virtualX        - Requested virtual display X resolution (-1 to use default)
virtualY        - Returns the addressable display memory Y resolution
bytesPerLine    - Returns the scanline width for the mode
maxMem          - Returns the maximum addressable display memory limit
crtc            - CRTC information block containing CRTC timings (required)

RETURNS:
0 on success, -1 on failure

REMARKS:
This function is used to initialize a custom display mode using the
specified X and Y resolution, color depth and CRTC timings.
****************************************************************************/
N_int32 NAPI MULDISP_SetCustomVideoMode(
    N_int32 xRes,
    N_int32 yRes,
    N_int32 bitsPerPixel,
    N_uint32 flags,
    N_int32 *virtualX,
    N_int32 *virtualY,
    N_int32 *bytesPerLine,
    N_int32 *maxMem,
    GA_CRTCInfo *crtc)
{
    return DC.init[0].SetCustomVideoMode(xRes,yRes,bitsPerPixel,flags,virtualX,virtualY,bytesPerLine,maxMem,crtc);
}

/****************************************************************************
DESCRIPTION:
Sets a specified display mode.

PARAMETERS:
mode            - Mode number to set, including flags
virtualX        - Requesyed virtual display X resolution (-1 to use default)
virtualY        - Returns the addressable display memory Y resolution
bytesPerLine    - Returns the scanline width for the mode
maxMem          - Returns the maximum addressable display memory limit
crtc            - CRTC information block if afRefreshCtrl specified

RETURNS:
0 on success, -1 on failure

REMARKS:
This function is used to initialize a specified display mode.
****************************************************************************/
N_int32 NAPI MULDISP_SetVideoMode(
    N_uint32 mode,
    N_int32 *virtualX,
    N_int32 *virtualY,
    N_int32 *bytesPerLine,
    N_int32 *maxMem,
    N_int32 refreshRate,
    GA_CRTCInfo *crtc)
{
    int                 i;
    GA_globalOptions    *g = DC.gOpt;

    if (IS_MULDISP_OPT() && (mode & gaModeMask) >= DC.virtualMode8) {
        // Populate mode information for all devices
        DC.virtualMi.dwSize = sizeof(DC.virtualMi.dwSize);
        if (MULDISP_GetVideoModeInfo(mode,&DC.virtualMi) != 0)
            return -1;

        // We don't support virtual modes!
        if (*virtualX != -1 || *virtualY != -1 || *bytesPerLine != -1)
            return -1;

        // Now set all the display devices to the correct display modes
        for (i = 0; i < NUM_DEVICES(); i++) {
            DC.bounds[i] = g->bounds[i];
            *virtualX = g->bounds[i].right - g->bounds[i].left;
            *virtualY = g->bounds[i].bottom - g->bounds[i].top;
            *bytesPerLine = DC.mi[i].BytesPerScanLine;
            SET_ACTIVE(i);
            DC.oldMode[i] = DC.init[i].GetVideoMode();
            if (DC.init[i].SetVideoMode(
                DC.modes[i] | (mode & ~(gaModeMask | gaRefreshCtrl)),
                virtualX,virtualY,bytesPerLine,
                maxMem,refreshRate,NULL) != 0) {
                SET_ACTIVE(0);
                return -1;
                }
            }
        SET_MIXED_MODE();

        // Return the virtual screen dimensions
        *bytesPerLine   = DC.mi[0].BytesPerScanLine;
        *virtualX       = g->virtualSize.right;
        *virtualY       = g->virtualSize.bottom;
        DC.cntMode      = mode;
        DC.gMi          = DC.virtualMi;
        DC.enabled      = true;
        return 0;
        }
    else {
        if (DC.cntMode != 0) {
            // Restore text mode for all secondary devices
            for (i = NUM_DEVICES()-1; i > 0; i--) {
                N_int32 vX = -1,vY = -1,bpl = -1;
                SET_ACTIVE(i);
                DC.init[i].SetVideoMode(DC.oldMode[i],&vX,&vY,&bpl,maxMem,0,NULL);
                }
            SET_ACTIVE(0);
            }
        DC.cntMode = 0;
        DC.enabled = IS_MULDISP();
        return DC.init[0].SetVideoMode(mode,virtualX,virtualY,bytesPerLine,maxMem,refreshRate,crtc);
        }
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
N_uint32 NAPI MULDISP_GetClosestPixelClockV1(
    N_int32 xRes,
    N_int32 yRes,
    N_int32 bitsPerPixel,
    N_uint32 pixelClock)
{
    return DC.init[0].GetClosestPixelClockV1(xRes,yRes,bitsPerPixel,pixelClock);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
N_uint32 NAPI MULDISP_GetClosestPixelClock(
    N_int32 outputHead,
    N_int32 xRes,
    N_int32 yRes,
    N_int32 bitsPerPixel,
    N_uint32 pixelClock)
{
    return DC.init[0].GetClosestPixelClock(outputHead,xRes,yRes,bitsPerPixel,pixelClock);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_GetCRTCTimings(
    GA_CRTCInfo *crtc)
{
    DC.init[0].GetCRTCTimings(crtc);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_SetCRTCTimings(
    GA_CRTCInfo *crtc)
{
    DC.init[0].SetCRTCTimings(crtc);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_SetGlobalRefresh(
    N_int32 refresh,
    N_int32 outputHead)
{
    int i;

    for (i = 0; i < DC.numDevices; i++)
        DC.init[i].SetGlobalRefresh(refresh,outputHead);
}

/****************************************************************************
PARAMETERS:
subfunc - Sub-function number
saveBuf - Buffer to save or restore data from.

RETURNS:
Size of the save buffer if subfunc is set to 2.

REMARKS:
This function provides support for saving and restoring the complete
hardware state. This is useful for debuggers and other utility software that
needs to be able to temporarily take over the display and restore it back to
the original state it was in. This function is responsible for saving and
restoring all hardware registers related to the graphics mode.

A buffer large enough to hold the entire hardware state must be allocated by
the calling code and passed in the saveBuf parameter. In order to determine
the size of the hardware state buffer to be allocated, the calling code
should first call this function with subfunc set to 2 to determine the size
of the hardware save/restore state buffer.

The following subfunctions are defined:

subfunction - description
0           - Save hardware state to saveBuf
1           - Restore hardware state from saveBuf
2           - Return state buffer size
****************************************************************************/
N_int32 NAPI MULDISP_SaveRestoreState(
    N_int32 subfunc,
    void *saveBuf)
{
    if (DC.enabled) {
        uchar   *p = saveBuf;
        N_int32 size,i;

        if (subfunc == 2) {
            // Determine the size of the total save buffer
            size = 0;
            for (i = 0; i < DC.numDevices; i++)
                size += DC.init[i].SaveRestoreState(2,NULL);
            return size;
            }

        // Either save or restore the state of all controllers
        for (i = 0; i < DC.numDevices; i++) {
            size = DC.init[i].SaveRestoreState(2,NULL);
            SET_ACTIVE(i);
            DC.init[i].SaveRestoreState(subfunc,p);
            p += size;
            }
        SET_MIXED_MODE();
        return 0;
        }
    else
        return DC.init[0].SaveRestoreState(subfunc,saveBuf);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
N_int32 NAPI MULDISP_SetDisplayOutput(
    N_int32 device)
{
    return DC.init[0].SetDisplayOutput(device);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
N_int32 NAPI MULDISP_SetDisplayOutputExt(
    N_int32 device,
    N_int32 outputHead)
{
    return DC.init[0].SetDisplayOutputExt(device,outputHead);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
N_int32 NAPI MULDISP_GetDisplayOutput(void)
{
    return DC.init[0].GetDisplayOutput();
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
N_int32 NAPI MULDISP_GetDisplayOutputExt(
    N_int32 outputHead)
{
    return DC.init[0].GetDisplayOutputExt(outputHead);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_SetSoftwareRenderFuncs(
    GA_2DRenderFuncs *softwareFuncs)
{
    DC.init[0].SetSoftwareRenderFuncs(softwareFuncs);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_GetUniqueFilename(
    char *filename,
    int type)
{
    DC.init[0].GetUniqueFilename(filename,type);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_GetMonitorInfo(
    GA_monitor *monitor,
    N_int32 outputHead)
{
    DC.init[0].GetMonitorInfo(monitor,outputHead);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_SetMonitorInfo(
    GA_monitor *monitor,
    N_int32 outputHead)
{
    DC.init[0].SetMonitorInfo(monitor,outputHead);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_GetCurrentVideoModeInfo(
    GA_modeInfo *modeInfo)
{
    if (DC.enabled) {
        if (modeInfo->dwSize == -1) {
            modeInfo->dwSize = sizeof(*modeInfo);
            DC.init[modeInfo->Attributes].GetCurrentVideoModeInfo(modeInfo);
            }
        else
            COPY_STRUCTURE(modeInfo,&DC.gMi);
        }
    else
        DC.init[0].GetCurrentVideoModeInfo(modeInfo);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_SetBank(
    N_int32 bank)
{
    DC.driver[0].SetBank(bank);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_SetDisplayStart(
    N_int32 offset,
    N_int32 waitVRT)
{
    if (!DC.enabled)
        DC.driver[0].SetDisplayStart(offset,waitVRT);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_SetDisplayStartXY(
    N_int32 x,
    N_int32 y,
    N_int32 waitVRT)
{
    if (!DC.enabled)
        DC.driver[0].SetDisplayStartXY(x,y,waitVRT);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_SetStereoDisplayStart(
    N_int32 leftOffset,
    N_int32 rightOffset,
    N_int32 waitVRT)
{
    if (!DC.enabled)
        DC.driver[0].SetStereoDisplayStart(leftOffset,rightOffset,waitVRT);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
N_int32 NAPI MULDISP_GetDisplayStartStatus(void)
{
    if (!DC.enabled)
        return DC.driver[0].GetDisplayStartStatus();
    return -1;
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_EnableStereoMode(
    N_int32 enable)
{
    if (!DC.enabled)
        DC.driver[0].EnableStereoMode(enable);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
N_int32 NAPI MULDISP_IsVSync(void)
{
    return DC.driver[0].IsVSync();
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_WaitVSync(void)
{
    DC.driver[0].WaitVSync();
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_SetVSyncWidth(
    N_int32 width)
{
    DC.driver[0].SetVSyncWidth(width);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
N_int32 NAPI MULDISP_GetVSyncWidth(void)
{
    return DC.driver[0].GetVSyncWidth();
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_SetPaletteData(
    GA_palette *pal,
    N_int32 num,
    N_int32 index,
    N_int32 waitVRT)
{
    if (DC.enabled) {
        int i = NUM_DEVICES1();

        while (i--) {
            SET_ACTIVE(i);
            DC.driver[i].SetPaletteData(pal,num,index,waitVRT);
            }
        SET_MIXED_MODE();
        }
    else
        DC.driver[0].SetPaletteData(pal,num,index,waitVRT);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_GetPaletteData(
    GA_palette *pal,
    N_int32 num,
    N_int32 index)
{
    DC.driver[0].GetPaletteData(pal,num,index);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_SetGammaCorrectData(
    GA_palette *pal,
    N_int32 num,
    N_int32 index,
    N_int32 waitVRT)
{
    if (DC.enabled) {
        int i = NUM_DEVICES1();

        while (i--) {
            if (DC.driver[i].SetGammaCorrectData) {
                SET_ACTIVE(i);
                DC.driver[i].SetGammaCorrectData(pal,num,index,waitVRT);
                }
            }
        SET_MIXED_MODE();
        }
    else if (DC.driver[0].SetGammaCorrectData)
        DC.driver[0].SetGammaCorrectData(pal,num,index,waitVRT);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_GetGammaCorrectData(
    GA_palette *pal,
    N_int32 num,
    N_int32 index)
{
    if (DC.driver[0].GetGammaCorrectData)
        DC.driver[0].GetGammaCorrectData(pal,num,index);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_SetPaletteDataExt(
    GA_paletteExt *pal,
    N_int32 num,
    N_int32 index,
    N_int32 waitVRT)
{
    if (DC.enabled) {
        int i = NUM_DEVICES1();

        while (i--) {
            SET_ACTIVE(i);
            DC.driver[i].SetPaletteDataExt(pal,num,index,waitVRT);
            }
        SET_MIXED_MODE();
        }
    else
        DC.driver[0].SetPaletteDataExt(pal,num,index,waitVRT);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_GetPaletteDataExt(
    GA_paletteExt *pal,
    N_int32 num,
    N_int32 index)
{
    DC.driver[0].GetPaletteDataExt(pal,num,index);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_SetGammaCorrectDataExt(
    GA_paletteExt *pal,
    N_int32 num,
    N_int32 index,
    N_int32 waitVRT)
{
    if (DC.enabled) {
        int i = NUM_DEVICES1();

        while (i--) {
            if (DC.driver[i].SetGammaCorrectDataExt) {
                SET_ACTIVE(i);
                DC.driver[i].SetGammaCorrectDataExt(pal,num,index,waitVRT);
                }
            }
        SET_MIXED_MODE();
        }
    else if (DC.driver[0].SetGammaCorrectDataExt)
        DC.driver[0].SetGammaCorrectDataExt(pal,num,index,waitVRT);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_GetGammaCorrectDataExt(
    GA_paletteExt *pal,
    N_int32 num,
    N_int32 index)
{
    if (DC.driver[0].GetGammaCorrectDataExt)
        DC.driver[0].GetGammaCorrectDataExt(pal,num,index);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
N_int32 NAPI MULDISP_IsIdle(void)
{
    int i = NUM_DEVICES2();

    while (i--) {
        if (!DC.state2d[i].IsIdle())
            return false;
        }
    return true;
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_WaitTillIdle(void)
{
    int i = NUM_DEVICES2();

    while (i--)
        DC.state2d[i].WaitTillIdle();
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
N_int32 NAPI MULDISP_SetMix(
    N_int32 mix)
{
    int i = NUM_DEVICES2();

    while (i--)
        DC.state2d[i].SetMix(mix);
    return true;
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_SetForeColor(
    GA_color color)
{
    int i = NUM_DEVICES2();

    while (i--)
        DC.state2d[i].SetForeColor(color);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_SetBackColor(
    GA_color color)
{
    int i = NUM_DEVICES2();

    while (i--)
        DC.state2d[i].SetBackColor(color);
}

/****************************************************************************
PARAMETERS:
index   - Index of the pattern to download
pattern - Pointer to pattern data to download
****************************************************************************/
void NAPI MULDISP_Set8x8MonoPattern(
    N_int32 index,
    GA_pattern *pattern)
{
    int i = NUM_DEVICES2();

    while (i--)
        DC.state2d[i].Set8x8MonoPattern(index,pattern);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_Use8x8MonoPattern(
    N_int32 index)
{
    int i = NUM_DEVICES2();

    while (i--)
        DC.state2d[i].Use8x8MonoPattern(index);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_Use8x8TransMonoPattern(
    N_int32 index)
{
    int i = NUM_DEVICES2();

    while (i--)
        DC.state2d[i].Use8x8TransMonoPattern(index);
}

/****************************************************************************
PARAMETERS:
index   - Index of the pattern to download
pattern - Pointer to pattern data to download
****************************************************************************/
void NAPI MULDISP_Set8x8ColorPattern(
    N_int32 index,
    GA_colorPattern *pattern)
{
    int i = NUM_DEVICES2();

    while (i--)
        DC.state2d[i].Set8x8ColorPattern(index,pattern);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_Use8x8ColorPattern(
    N_int32 index)
{
    int i = NUM_DEVICES2();

    while (i--)
        DC.state2d[i].Use8x8ColorPattern(index);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_Use8x8TransColorPattern(
    N_int32 index,
    GA_color transparent)
{
    int i = NUM_DEVICES2();

    while (i--)
        DC.state2d[i].Use8x8TransColorPattern(index,transparent);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_SetLineStipple(
    GA_stipple stipple)
{
    int i = NUM_DEVICES2();

    while (i--)
        DC.state2d[i].SetLineStipple(stipple);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_SetLineStyle(
    N_uint32 styleMask,
    N_uint32 styleStep,
    N_uint32 styleValue)
{
    int i = NUM_DEVICES2();

    while (i--)
        DC.state2d[i].SetLineStyle(styleMask,styleStep,styleValue);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_SetLineStippleCount(
    N_uint32 count)
{
    int i = NUM_DEVICES2();

    while (i--)
        DC.state2d[i].SetLineStippleCount(count);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_SetPlaneMask(
    N_uint32 mask)
{
    int i = NUM_DEVICES2();

    while (i--)
        DC.state2d[i].SetPlaneMask(mask);
}

/****************************************************************************
PARAMETERS:
blendFunc   - New source blending function to make active

REMARKS:
Obsolete, use SetBlendFunc instead.
****************************************************************************/
void NAPI MULDISP_SetSrcBlendFunc(
    N_int32 blendFunc)
{
    int i = NUM_DEVICES2();

    while (i--)
        DC.state2d[i].SetBlendFunc(DC.srcBlendFunc = blendFunc,DC.dstBlendFunc);
}

/****************************************************************************
PARAMETERS:
blendFunc   - New destination blending function to make active

REMARKS:
Obsolete, use SetBlendFunc instead.
****************************************************************************/
void NAPI MULDISP_SetDstBlendFunc(
    N_int32 blendFunc)
{
    int i = NUM_DEVICES2();

    while (i--)
        DC.state2d[i].SetBlendFunc(DC.srcBlendFunc,DC.dstBlendFunc = blendFunc);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_SetBlendFunc(
    N_int32 srcBlendFunc,
    N_int32 dstBlendFunc)
{
    int i = NUM_DEVICES2();

    while (i--)
        DC.state2d[i].SetBlendFunc(DC.srcBlendFunc = srcBlendFunc,DC.dstBlendFunc = dstBlendFunc);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_SetAlphaValue(
    N_uint8 alpha)
{
    int i = NUM_DEVICES2();

    while (i--)
        DC.state2d[i].SetAlphaValue(alpha);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_SetupForWindows(void)
{
    int i = NUM_DEVICES2();

    while (i--)
        DC.state2d[i].SetupForWindows();
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_ResetForWindows(void)
{
    int i = NUM_DEVICES2();

    while (i--)
        DC.state2d[i].ResetForWindows();
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
x1      - First X coordinate of scanline
x2      - Second X coordinate of scanline
****************************************************************************/
static void DrawScan(
    N_int32 top,
    N_int32 x1,
    N_int32 x2)
{
    int         i = DC.numDevices;
    GA_layout   clip;
    N_int32     left = MIN(x1,x2);
    N_int32     right = MAX(x1,x2);
    N_int32     bottom = top + 1;
    N_int32     cLeft,cTop,cRight,cBottom;

    if (x1 != x2) {
        while (i--) {
            clip = DC.bounds[i];
            if (VISIBLE_RECT(left,top,right,bottom,cLeft,cTop,cRight,cBottom,clip)) {
                DC.draw2d[i].DrawRect(cLeft-clip.left,cTop-clip.top,cRight-cLeft,cBottom-cTop);
                if (EQUAL_RECT(left,top,right,bottom,cLeft,cTop,cRight,cBottom))
                    return;
                }
            }
        }
}

/****************************************************************************
PARAMETERS:
y           - Y coordinate for scanline
x1          - First X coordinate of scanline
x2          - Second X coordinate of scanline
****************************************************************************/
static void DrawPattScan(
    N_int32 top,
    N_int32 x1,
    N_int32 x2)
{
    int         i = DC.numDevices;
    GA_layout   clip;
    N_int32     left = MIN(x1,x2);
    N_int32     right = MAX(x1,x2);
    N_int32     bottom = top + 1;
    N_int32     cLeft,cTop,cRight,cBottom;

    if (x1 != x2) {
        while (i--) {
            clip = DC.bounds[i];
            if (VISIBLE_RECT(left,top,right,bottom,cLeft,cTop,cRight,cBottom,clip)) {
                DC.draw2d[i].DrawPattRect(cLeft-clip.left,cTop-clip.top,cRight-cLeft,cBottom-cTop);
                if (EQUAL_RECT(left,top,right,bottom,cLeft,cTop,cRight,cBottom))
                    return;
                }
            }
        }
}

/****************************************************************************
PARAMETERS:
y           - Y coordinate for scanline
x1          - First X coordinate of scanline
x2          - Second X coordinate of scanline
****************************************************************************/
static void DrawColorPattScan(
    N_int32 top,
    N_int32 x1,
    N_int32 x2)
{
    int         i = DC.numDevices;
    GA_layout   clip;
    N_int32     left = MIN(x1,x2);
    N_int32     right = MAX(x1,x2);
    N_int32     bottom = top + 1;
    N_int32     cLeft,cTop,cRight,cBottom;

    if (x1 != x2) {
        while (i--) {
            clip = DC.bounds[i];
            if (VISIBLE_RECT(left,top,right,bottom,cLeft,cTop,cRight,cBottom,clip)) {
                DC.draw2d[i].DrawColorPattRect(cLeft-clip.left,cTop-clip.top,cRight-cLeft,cBottom-cTop);
                if (EQUAL_RECT(left,top,right,bottom,cLeft,cTop,cRight,cBottom))
                    return;
                }
            }
        }
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI MULDISP_DrawScanList(
    N_int32 y,
    N_int32 length,
    N_int16 *scans)
{
    while (length--) {
        DrawScan(y,scans[0],scans[1]);
        scans += 2;
        y++;
        }
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI MULDISP_DrawPattScanList(
    N_int32 y,
    N_int32 length,
    N_int16 *scans)
{
    while (length--) {
        DrawPattScan(y,scans[0],scans[1]);
        scans += 2;
        y++;
        }
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI MULDISP_DrawColorPattScanList(
    N_int32 y,
    N_int32 length,
    N_int16 *scans)
{
    while (length--) {
        DrawColorPattScan(y,scans[0],scans[1]);
        scans += 2;
        y++;
        }
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI MULDISP_DrawEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int i,j,maxIndex;

    // Render the scanlines in the ellipse
    maxIndex = y+length-1;
    for (i = y,j = y+height; i < maxIndex; i++,j--,scans += 2) {
        DrawScan(i,scans[0],scans[1]);
        DrawScan(j,scans[0],scans[1]);
        }
    if (!(height & 1))
        DrawScan(i,scans[0],scans[1]);
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI MULDISP_DrawPattEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int i,j,maxIndex;

    // Render the scanlines in the ellipse
    maxIndex = y+length-1;
    for (i = y,j = y+height; i < maxIndex; i++,j--,scans += 2) {
        DrawPattScan(i,scans[0],scans[1]);
        DrawPattScan(j,scans[0],scans[1]);
        }
    if (!(height & 1))
        DrawPattScan(i,scans[0],scans[1]);
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI MULDISP_DrawColorPattEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int i,j,maxIndex;

    // Render the scanlines in the ellipse
    maxIndex = y+length-1;
    for (i = y,j = y+height; i < maxIndex; i++,j--,scans += 2) {
        DrawColorPattScan(i,scans[0],scans[1]);
        DrawColorPattScan(j,scans[0],scans[1]);
        }
    if (!(height & 1))
        DrawColorPattScan(i,scans[0],scans[1]);
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis + pen height adjustment
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI MULDISP_DrawFatEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int i,j,maxIndex;

    // Render the scanlines in the ellipse
    maxIndex = y+length;
    for (i = y,j = y+height; i < maxIndex; i++,j--,scans += 4) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            DrawScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            DrawScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            DrawScan(j,scans[I_LEFTL],scans[I_LEFTR]);
            DrawScan(j,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            DrawScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            DrawScan(j,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
    if ((height+1) & 1) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            DrawScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            DrawScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            DrawScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis + pen height adjustment
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI MULDISP_DrawPattFatEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int i,j,maxIndex;

    // Render the scanlines in the ellipse
    maxIndex = y+length;
    for (i = y,j = y+height; i < maxIndex; i++,j--,scans += 4) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            DrawPattScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            DrawPattScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            DrawPattScan(j,scans[I_LEFTL],scans[I_LEFTR]);
            DrawPattScan(j,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            DrawPattScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            DrawPattScan(j,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
    if ((height+1) & 1) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            DrawPattScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            DrawPattScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            DrawPattScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis + pen height adjustment
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI MULDISP_DrawColorPattFatEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int i,j,maxIndex;

    // Render the scanlines in the ellipse
    maxIndex = y+length;
    for (i = y,j = y+height; i < maxIndex; i++,j--,scans += 4) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            DrawColorPattScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            DrawColorPattScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            DrawColorPattScan(j,scans[I_LEFTL],scans[I_LEFTR]);
            DrawColorPattScan(j,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            DrawColorPattScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            DrawColorPattScan(j,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
    if ((height+1) & 1) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            DrawColorPattScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            DrawColorPattScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            DrawColorPattScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the rectangle to draw
top     - Top coordinate of the rectangle to draw
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
****************************************************************************/
void NAPI MULDISP_DrawRect(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height)
{
    int         i = DC.numDevices;
    GA_layout   clip;
    N_int32     right = left + width;
    N_int32     bottom = top + height;
    N_int32     cLeft,cTop,cRight,cBottom;

    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_RECT(left,top,right,bottom,cLeft,cTop,cRight,cBottom,clip)) {
            DC.draw2d[i].DrawRect(cLeft-clip.left,cTop-clip.top,cRight-cLeft,cBottom-cTop);
            if (EQUAL_RECT(left,top,right,bottom,cLeft,cTop,cRight,cBottom))
                return;
            }
        }
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the rectangle to draw
top     - Top coordinate of the rectangle to draw
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
****************************************************************************/
void NAPI MULDISP_DrawPattRect(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height)
{
    int         i = DC.numDevices;
    GA_layout   clip;
    N_int32     right = left + width;
    N_int32     bottom = top + height;
    N_int32     cLeft,cTop,cRight,cBottom;

    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_RECT(left,top,right,bottom,cLeft,cTop,cRight,cBottom,clip)) {
            DC.draw2d[i].DrawPattRect(cLeft-clip.left,cTop-clip.top,cRight-cLeft,cBottom-cTop);
            if (EQUAL_RECT(left,top,right,bottom,cLeft,cTop,cRight,cBottom))
                return;
            }
        }
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the rectangle to draw
top     - Top coordinate of the rectangle to draw
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
****************************************************************************/
void NAPI MULDISP_DrawColorPattRect(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height)
{
    int         i = DC.numDevices;
    GA_layout   clip;
    N_int32     right = left + width;
    N_int32     bottom = top + height;
    N_int32     cLeft,cTop,cRight,cBottom;

    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_RECT(left,top,right,bottom,cLeft,cTop,cRight,cBottom,clip)) {
            DC.draw2d[i].DrawColorPattRect(cLeft-clip.left,cTop-clip.top,cRight-cLeft,cBottom-cTop);
            if (EQUAL_RECT(left,top,right,bottom,cLeft,cTop,cRight,cBottom))
                return;
            }
        }
}

/****************************************************************************
PARAMETERS:
trap    - Pointer to the GA_trap structure describing the trapezoid
****************************************************************************/
void NAPI MULDISP_DrawTrap(
    GA_trap *trap)
{
    int y,x1,x2,count = trap->count;

    if (!count)
        return;
    y = trap->y;
    x1 = trap->x1;
    x2 = trap->x2;
    trap->y += count;
    while (count--) {
        DrawScan(y,(x1 + 0x8000) >> 16,(x2 + 0x8000) >> 16);
        x1 += trap->slope1;
        x2 += trap->slope2;
        y++;
        }
    trap->x1 = x1;
    trap->x2 = x2;
}

/****************************************************************************
PARAMETERS:
trap    - Pointer to the GA_trap structure describing the trapezoid
****************************************************************************/
void NAPI MULDISP_DrawPattTrap(
    GA_trap *trap)
{
    int y,x1,x2,count = trap->count;

    if (!count)
        return;
    y = trap->y;
    x1 = trap->x1;
    x2 = trap->x2;
    trap->y += count;
    while (count--) {
        DrawPattScan(y,(x1 + 0x8000) >> 16,(x2 + 0x8000) >> 16);
        x1 += trap->slope1;
        x2 += trap->slope2;
        y++;
        }
    trap->x1 = x1;
    trap->x2 = x2;
}

/****************************************************************************
PARAMETERS:
trap    - Pointer to the GA_trap structure describing the trapezoid
****************************************************************************/
void NAPI MULDISP_DrawColorPattTrap(
    GA_trap *trap)
{
    int y,x1,x2,count = trap->count;

    if (!count)
        return;
    y = trap->y;
    x1 = trap->x1;
    x2 = trap->x2;
    trap->y += count;
    while (count--) {
        DrawColorPattScan(y,(x1 + 0x8000) >> 16,(x2 + 0x8000) >> 16);
        x1 += trap->slope1;
        x2 += trap->slope2;
        y++;
        }
    trap->x1 = x1;
    trap->x2 = x2;
}

/****************************************************************************
PARAMETERS:
x1          - X1 coordinate
y1          - Y1 coordinate
x2          - X2 coordinate
y2          - Y2 coordinate
drawLast    - 1 to draw last pixel, 0 to skip it
****************************************************************************/
void NAPI MULDISP_DrawLineInt(
    N_int32 x1,
    N_int32 y1,
    N_int32 x2,
    N_int32 y2,
    N_int32 drawLast)
{
    int         i = DC.numDevices;
    GA_layout   clip;

    while (i--) {
        clip = DC.bounds[i];
        if (DC.draw2d[i].DrawClippedLineInt(
                x1-clip.left,y1-clip.top,x2-clip.left,y2-clip.top,drawLast,
                0,0,clip.right-clip.left,clip.bottom-clip.top) == 1)
            return;
        }
}

/****************************************************************************
PARAMETERS:
x1          - X1 coordinate
y1          - Y1 coordinate
x2          - X2 coordinate
y2          - Y2 coordinate
drawLast    - 1 to draw last pixel, 0 to skip it
transparent - 1 if the line is transparent, 0 if opaque
****************************************************************************/
void NAPI MULDISP_DrawStippleLineInt(
    N_int32 x1,
    N_int32 y1,
    N_int32 x2,
    N_int32 y2,
    N_int32 drawLast,
    N_int32 transparent)
{
    int         i = DC.numDevices;
    GA_layout   clip;

    while (i--) {
        clip = DC.bounds[i];
        if (DC.draw2d[i].DrawClippedStippleLineInt(
                x1-clip.left,y1-clip.top,x2-clip.left,y2-clip.top,drawLast,transparent,
                0,0,clip.right-clip.left,clip.bottom-clip.top) == 1)
            return;
        }
}

/****************************************************************************
PARAMETERS:
x1              - X1 coordinate
y1              - Y1 coordinate
initialError    - Initial error term
majorInc        - Major increment factor for error term
diagInc         - Diagonal increment factor for error term
count           - Number of pixels to draw
flags           - Flags to control drawing
****************************************************************************/
void NAPI MULDISP_DrawBresenhamLine(
    N_int32 x1,
    N_int32 y1,
    N_int32 initialError,
    N_int32 majorInc,
    N_int32 diagInc,
    N_int32 count,
    N_int32 flags)
{
    int         i = DC.numDevices;
    GA_layout   clip;

    while (i--) {
        clip = DC.bounds[i];
        if (DC.draw2d[i].DrawClippedBresenhamLine(
                x1-clip.left,y1-clip.top,initialError,majorInc,diagInc,count,flags,
                0,0,clip.right-clip.left,clip.bottom-clip.top) == 1)
            return;
        }
}

/****************************************************************************
PARAMETERS:
x1              - X1 coordinate
y1              - Y1 coordinate
initialError    - Initial error term
majorInc        - Major increment factor for error term
diagInc         - Diagonal increment factor for error term
count           - Number of pixels to draw
flags           - Flags to control drawing
transparent     - 1 if the line is transparent, 0 if opaque
****************************************************************************/
void NAPI MULDISP_DrawBresenhamStippleLine(
    N_int32 x1,
    N_int32 y1,
    N_int32 initialError,
    N_int32 majorInc,
    N_int32 diagInc,
    N_int32 count,
    N_int32 flags,
    N_int32 transparent)
{
    int         i = DC.numDevices;
    GA_layout   clip;

    while (i--) {
        clip = DC.bounds[i];
        if (DC.draw2d[i].DrawClippedBresenhamStippleLine(
                x1-clip.left,y1-clip.top,initialError,majorInc,diagInc,count,flags,transparent,
                0,0,clip.right-clip.left,clip.bottom-clip.top) == 1)
            return;
        }
}

/****************************************************************************
PARAMETERS:
x1          - X1 coordinate
y1          - Y1 coordinate
x2          - X2 coordinate
y2          - Y2 coordinate
drawLast    - 1 to draw last pixel, 0 to skip it
transparent - 1 if the line is transparent, 0 if opaque
****************************************************************************/
void NAPI MULDISP_DrawStyleLineInt(
    N_int32 x1,
    N_int32 y1,
    N_int32 x2,
    N_int32 y2,
    N_int32 drawLast,
    N_int32 transparent)
{
    int         i = DC.numDevices;
    GA_layout   clip;

    while (i--) {
        clip = DC.bounds[i];
        if (DC.draw2d[i].DrawClippedStyleLineInt(
                x1-clip.left,y1-clip.top,x2-clip.left,y2-clip.top,drawLast,transparent,
                0,0,clip.right-clip.left,clip.bottom-clip.top) == 1)
            return;
        }
}

/****************************************************************************
PARAMETERS:
x1              - X1 coordinate
y1              - Y1 coordinate
initialError    - Initial error term
majorInc        - Major increment factor for error term
diagInc         - Diagonal increment factor for error term
count           - Number of pixels to draw
flags           - Flags to control drawing
transparent     - 1 if the line is transparent, 0 if opaque
****************************************************************************/
void NAPI MULDISP_DrawBresenhamStyleLine(
    N_int32 x1,
    N_int32 y1,
    N_int32 initialError,
    N_int32 majorInc,
    N_int32 diagInc,
    N_int32 count,
    N_int32 flags,
    N_int32 transparent)
{
    int         i = DC.numDevices;
    GA_layout   clip;

    while (i--) {
        clip = DC.bounds[i];
        if (DC.draw2d[i].DrawClippedBresenhamStyleLine(
                x1-clip.left,y1-clip.top,initialError,majorInc,diagInc,count,flags,transparent,
                0,0,clip.right-clip.left,clip.bottom-clip.top) == 1)
            return;
        }
}

/****************************************************************************
PARAMETERS:
x1          - X coordinate of first endpoint to clip
y1          - Y coordinate of first endpoint to clip
x2          - X coordinate of second endpoint to clip
y2          - Y coordinate of second endpoint to clip
drawLast    - True if last pixel should be drawn
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle

REMARKS:
Internal function to clip and draw an integer line, such that the clipped
line segment draws the *exact* same set of pixels that the unclipped line
would have drawn. We also correctly handle arbitrarily large lines, while
still allowing the clipped line segment to be drawn in hardware.
****************************************************************************/
N_int32 NAPI MULDISP_DrawClippedLineInt(
    N_int32 x1,
    N_int32 y1,
    N_int32 x2,
    N_int32 y2,
    N_int32 drawLast,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
    // TODO!
    return 0;
}

/****************************************************************************
PARAMETERS:
x1          - X coordinate of first endpoint to clip
y1          - Y coordinate of first endpoint to clip
x2          - X coordinate of second endpoint to clip
y2          - Y coordinate of second endpoint to clip
drawLast    - True if last pixel should be drawn
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle

REMARKS:
Internal function to clip and draw an integer line, such that the clipped
line segment draws the *exact* same set of pixels that the unclipped line
would have drawn. We also correctly handle arbitrarily large lines, while
still allowing the clipped line segment to be drawn in hardware.
****************************************************************************/
N_int32 NAPI MULDISP_DrawClippedBresenhamLine(
    N_int32 x1,
    N_int32 y1,
    N_int32 initialError,
    N_int32 majorInc,
    N_int32 diagInc,
    N_int32 count,
    N_int32 flags,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
    // TODO!
    return 0;
}

/****************************************************************************
PARAMETERS:
x1          - X coordinate of first endpoint to clip
y1          - Y coordinate of first endpoint to clip
x2          - X coordinate of second endpoint to clip
y2          - Y coordinate of second endpoint to clip
drawLast    - True if last pixel should be drawn
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle

REMARKS:
Internal function to clip and draw an integer line, such that the clipped
line segment draws the *exact* same set of pixels that the unclipped line
would have drawn. We also correctly handle arbitrarily large lines, while
still allowing the clipped line segment to be drawn in hardware.
****************************************************************************/
N_int32 NAPI MULDISP_DrawClippedStippleLineInt(
    N_int32 x1,
    N_int32 y1,
    N_int32 x2,
    N_int32 y2,
    N_int32 drawLast,
    N_int32 transparent,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
    // TODO!
    return 0;
}

/****************************************************************************
PARAMETERS:
x1          - X coordinate of first endpoint to clip
y1          - Y coordinate of first endpoint to clip
x2          - X coordinate of second endpoint to clip
y2          - Y coordinate of second endpoint to clip
drawLast    - True if last pixel should be drawn
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle

REMARKS:
Internal function to clip and draw an integer line, such that the clipped
line segment draws the *exact* same set of pixels that the unclipped line
would have drawn. We also correctly handle arbitrarily large lines, while
still allowing the clipped line segment to be drawn in hardware.
****************************************************************************/
N_int32 NAPI MULDISP_DrawClippedBresenhamStippleLine(
    N_int32 x1,
    N_int32 y1,
    N_int32 initialError,
    N_int32 majorInc,
    N_int32 diagInc,
    N_int32 count,
    N_int32 flags,
    N_int32 transparent,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
    // TODO!
    return 0;
}

/****************************************************************************
PARAMETERS:
x1          - X coordinate of first endpoint to clip
y1          - Y coordinate of first endpoint to clip
x2          - X coordinate of second endpoint to clip
y2          - Y coordinate of second endpoint to clip
drawLast    - True if last pixel should be drawn
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle

REMARKS:
Internal function to clip and draw an integer line, such that the clipped
line segment draws the *exact* same set of pixels that the unclipped line
would have drawn. We also correctly handle arbitrarily large lines, while
still allowing the clipped line segment to be drawn in hardware.
****************************************************************************/
N_int32 NAPI MULDISP_DrawClippedStyleLineInt(
    N_int32 x1,
    N_int32 y1,
    N_int32 x2,
    N_int32 y2,
    N_int32 drawLast,
    N_int32 transparent,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
    // TODO!
    return 0;
}

/****************************************************************************
PARAMETERS:
x1          - X coordinate of first endpoint to clip
y1          - Y coordinate of first endpoint to clip
x2          - X coordinate of second endpoint to clip
y2          - Y coordinate of second endpoint to clip
drawLast    - True if last pixel should be drawn
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle

REMARKS:
Internal function to clip and draw an integer line, such that the clipped
line segment draws the *exact* same set of pixels that the unclipped line
would have drawn. We also correctly handle arbitrarily large lines, while
still allowing the clipped line segment to be drawn in hardware.
****************************************************************************/
N_int32 NAPI MULDISP_DrawClippedBresenhamStyleLine(
    N_int32 x1,
    N_int32 y1,
    N_int32 initialError,
    N_int32 majorInc,
    N_int32 diagInc,
    N_int32 count,
    N_int32 flags,
    N_int32 transparent,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
    // TODO!
    return 0;
}

/****************************************************************************
PARAMETERS:
x           - Destination X coordinate to draw the bitmap at
y           - Destination Y coordinate to draw the bitmap at
width       - Width of the bitmap in pixels
height      - Height of the bitmap in pixels
byteWidth   - Width of the bitmap in bytes
image       - Pointer to the bitmap image data to draw
transparent - 1 for transparent, 0 for opaque

REMARKS:
If the transparent parameter is set to 1, where a bit is 0 in the bitmap
image the destination pixel remains untouched. If the transparent
parameter is set to 0, where a bit is 0 in the bitmap image the
destination pixel is drawn in the background color. In all cases where a
bit in the bitmap image is 1, the pixel is drawn in the foreground color.

Note:   This function processes the bitmap data in MSB fashion, in that the
        first pixel drawn corrsponds to bit 7 of the first byte. The second
        pixel is bit 6 of the first byte, ...,the 8th pixel is bit 7 of
        the second byte etc. Both LSB and MSB versions are provided for
        performance.
****************************************************************************/
void NAPI MULDISP_PutMonoImageMSBSys(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    N_int32 byteWidth,
    N_uint8 *image,
    N_int32 transparent)
{
    int         i = DC.numDevices;
    GA_layout   clip;
    N_int32     right = left + width;
    N_int32     bottom = top + height;
    N_int32     cLeft,cTop,cRight,cBottom;

    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_RECT(left,top,right,bottom,cLeft,cTop,cRight,cBottom,clip)) {
            if (EQUAL_RECT(left,top,right,bottom,cLeft,cTop,cRight,cBottom)) {
                DC.draw2d[i].PutMonoImageMSBSys(left-clip.left,top-clip.top,width,height,byteWidth,image,transparent);
                return;
                }
            else {
                DC.draw2d[i].ClipMonoImageMSBSys(left-clip.left,top-clip.top,width,height,byteWidth,image,transparent,
                    cLeft-clip.left,cTop-clip.top,cRight-clip.left,cBottom-clip.top);
                }
            }
        }
}

/****************************************************************************
PARAMETERS:
x           - Destination X coordinate to draw the bitmap at
y           - Destination Y coordinate to draw the bitmap at
width       - Width of the bitmap in pixels
height      - Height of the bitmap in pixels
byteWidth   - Width of the bitmap in bytes
image       - Pointer to the bitmap image data to draw
transparent - 1 for transparent, 0 for opaque
clipLeft    - Left edge of clip rectangle (inclusive)
clipTop     - Top edge of clip rectangle (inclusive)
clipRight   - Right edge of clip rectangle (exclusive)
clipBottom  - Bottom edge of clip rectangle (exclusive)

REMARKS:
If the transparent parameter is set to 1, where a bit is 0 in the bitmap
image the destination pixel remains untouched. If the transparent
parameter is set to 0, where a bit is 0 in the bitmap image the
destination pixel is drawn in the background color. In all cases where a
bit in the bitmap image is 1, the pixel is drawn in the foreground color.

Note:   This function processes the bitmap data in MSB fashion, in that the
        first pixel drawn corrsponds to bit 7 of the first byte. The second
        pixel is bit 6 of the first byte, ...,the 8th pixel is bit 7 of
        the second byte etc. Both LSB and MSB versions are provided for
        performance.
****************************************************************************/
void NAPI MULDISP_ClipMonoImageMSBSys(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    N_int32 byteWidth,
    N_uint8 *image,
    N_int32 transparent,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
    int         i = DC.numDevices;
    GA_layout   clip;
    N_int32     right = left + width;
    N_int32     bottom = top + height;
    N_int32     cLeft,cTop,cRight,cBottom;

    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_RECT(left,top,right,bottom,cLeft,cTop,cRight,cBottom,clip)) {
            if (EQUAL_RECT(left,top,right,bottom,cLeft,cTop,cRight,cBottom)) {
                if (VISIBLE_RECT2(clipLeft,clipTop,clipRight,clipBottom,cLeft,cTop,cRight,cBottom)) {
                    DC.draw2d[i].ClipMonoImageMSBSys(left-clip.left,top-clip.top,width,height,byteWidth,image,transparent,
                        cLeft-clip.left,cTop-clip.top,cRight-clip.left,cBottom-clip.top);
                    }
                return;
                }
            else if (VISIBLE_RECT2(clipLeft,clipTop,clipRight,clipBottom,cLeft,cTop,cRight,cBottom)) {
                DC.draw2d[i].ClipMonoImageMSBSys(left-clip.left,top-clip.top,width,height,byteWidth,image,transparent,
                    cLeft-clip.left,cTop-clip.top,cRight-clip.left,cBottom-clip.top);
                }
            }
        }
}

/****************************************************************************
PARAMETERS:
srcLeft - Left coordinate of the source rectangle to copy
srcTop  - Top coordinate of the source rectangle to copy
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
dstLeft - Left coordinate of destination
dstTop  - Top coordinate of destination
mix     - Mix code for the copy (GA_mixCodesType)
****************************************************************************/
void NAPI MULDISP_BitBlt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix)
{
    int         i = DC.numDevices;
    GA_layout   clip;
    N_int32     srcRight = srcLeft + width;
    N_int32     srcBottom = srcTop + height;
    N_int32     dstRight = dstLeft + width;
    N_int32     dstBottom = dstTop + height;
    N_int32     cLeft,cTop,cRight,cBottom;

    // First determine if the entire blit can be done in hardware on
    // a single device for speed.
    i = DC.numDevices;
    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_RECT(srcLeft,srcTop,srcRight,srcBottom,cLeft,cTop,cRight,cBottom,clip)
            && EQUAL_RECT(srcLeft,srcTop,srcRight,srcBottom,cLeft,cTop,cRight,cBottom)
            && VISIBLE_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom,clip)
            && EQUAL_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom)) {
            DC.draw2d[i].BitBlt(
                srcLeft-clip.left,srcTop-clip.top,
                width,height,
                dstLeft-clip.left,dstTop-clip.top,mix);
            return;
            }
        else
            break;
        }

    // First read source bitmap data from video memory into transfer buffer
    i = DC.numDevices;
    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_RECT(srcLeft,srcTop,srcRight,srcBottom,cLeft,cTop,cRight,cBottom,clip)) {
            if (EQUAL_RECT(srcLeft,srcTop,srcRight,srcBottom,cLeft,cTop,cRight,cBottom)) {
                DC.draw2d[i].GetBitmapSys(DC.buffer,DC.bufferPitch,
                    cLeft-clip.left,cTop-clip.top,
                    width,height,
                    cLeft,cTop,
                    GA_REPLACE_MIX);
                break;
                }
            else {
                DC.draw2d[i].GetBitmapSys(DC.buffer,DC.bufferPitch,
                    cLeft-clip.left,cTop-clip.top,
                    cRight - cLeft,cBottom - cTop,
                    cLeft,cTop,
                    GA_REPLACE_MIX);
                }
            }
        }

    // Now blit the bitmap data from the transfer buffer to video memory
    i = DC.numDevices;
    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom,clip)) {
            if (EQUAL_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom)) {
                DC.draw2d[i].BitBltSys(DC.buffer,DC.bufferPitch,
                    srcLeft,srcTop,width,height,
                    cLeft-clip.left,cTop-clip.top,mix,false);
                break;
                }
            else {
                DC.draw2d[i].BitBltSys(DC.buffer,DC.bufferPitch,
                    srcLeft + (cLeft - dstLeft),srcTop + (cTop - dstTop),
                    cRight - cLeft,cBottom - cTop,
                    cLeft-clip.left,cTop-clip.top,mix,false);
                }
            }
        }
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI MULDISP_BitBltSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    N_int32 flipY)
{
    int         i = DC.numDevices;
    GA_layout   clip;
    N_int32     dstRight = dstLeft + width;
    N_int32     dstBottom = dstTop + height;
    N_int32     cLeft,cTop,cRight,cBottom;

    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom,clip)) {
            if (EQUAL_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom)) {
                DC.draw2d[i].BitBltSys(srcAddr,srcPitch,
                    srcLeft,srcTop,width,height,
                    cLeft-clip.left,cTop-clip.top,mix,flipY);
                return;
                }
            else {
                DC.draw2d[i].BitBltSys(srcAddr,srcPitch,
                    srcLeft + (cLeft - dstLeft),srcTop + (cTop - dstTop),
                    cRight - cLeft,cBottom - cTop,
                    cLeft-clip.left,cTop-clip.top,mix,flipY);
                }
            }
        }
}

/****************************************************************************
PARAMETERS:
srcLeft - Left coordinate of the source rectangle to copy
srcTop  - Top coordinate of the source rectangle to copy
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
dstLeft - Left coordinate of destination
dstTop  - Top coordinate of destination
rop3    - ROP3 mix code for the copy
****************************************************************************/
void NAPI MULDISP_BitBltPatt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    int         i = DC.numDevices;
    GA_layout   clip;
    N_int32     srcRight = srcLeft + width;
    N_int32     srcBottom = srcTop + height;
    N_int32     dstRight = dstLeft + width;
    N_int32     dstBottom = dstTop + height;
    N_int32     cLeft,cTop,cRight,cBottom;

    // First determine if the entire blit can be done in hardware on
    // a single device for speed.
    i = DC.numDevices;
    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_RECT(srcLeft,srcTop,srcRight,srcBottom,cLeft,cTop,cRight,cBottom,clip)
            && EQUAL_RECT(srcLeft,srcTop,srcRight,srcBottom,cLeft,cTop,cRight,cBottom)
            && VISIBLE_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom,clip)
            && EQUAL_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom)) {
            DC.draw2d[i].BitBltPatt(
                srcLeft-clip.left,srcTop-clip.top,
                width,height,
                dstLeft-clip.left,dstTop-clip.top,rop3);
            return;
            }
        else
            break;
        }

    // First read source bitmap data from video memory into transfer buffer
    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_RECT(srcLeft,srcTop,srcRight,srcBottom,cLeft,cTop,cRight,cBottom,clip)) {
            if (EQUAL_RECT(srcLeft,srcTop,srcRight,srcBottom,cLeft,cTop,cRight,cBottom)) {
                DC.draw2d[i].GetBitmapSys(DC.buffer,DC.bufferPitch,
                    cLeft-clip.left,cTop-clip.top,
                    width,height,
                    cLeft,cTop,
                    GA_REPLACE_MIX);
                break;
                }
            else {
                DC.draw2d[i].GetBitmapSys(DC.buffer,DC.bufferPitch,
                    cLeft-clip.left,cTop-clip.top,
                    cRight - cLeft,cBottom - cTop,
                    cLeft,cTop,
                    GA_REPLACE_MIX);
                }
            }
        }

    // Now blit the bitmap data from the transfer buffer to video memory
    i = DC.numDevices;
    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom,clip)) {
            if (EQUAL_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom)) {
                DC.draw2d[i].BitBltPattSys(DC.buffer,DC.bufferPitch,
                    srcLeft,srcTop,width,height,
                    cLeft-clip.left,cTop-clip.top,rop3,false);
                break;
                }
            else {
                DC.draw2d[i].BitBltPattSys(DC.buffer,DC.bufferPitch,
                    srcLeft + (cLeft - dstLeft),srcTop + (cTop - dstTop),
                    cRight - cLeft,cBottom - cTop,
                    cLeft-clip.left,cTop-clip.top,rop3,false);
                }
            }
        }
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3        - Mix code for the copy (GA_mixCodesType)
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI MULDISP_BitBltPattSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3,
    N_int32 flipY)
{
    int         i = DC.numDevices;
    GA_layout   clip;
    N_int32     dstRight = dstLeft + width;
    N_int32     dstBottom = dstTop + height;
    N_int32     cLeft,cTop,cRight,cBottom;

    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom,clip)) {
            if (EQUAL_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom)) {
                DC.draw2d[i].BitBltPattSys(srcAddr,srcPitch,
                    srcLeft,srcTop,width,height,
                    cLeft-clip.left,cTop-clip.top,rop3,flipY);
                return;
                }
            else {
                DC.draw2d[i].BitBltPattSys(srcAddr,srcPitch,
                    srcLeft + (cLeft - dstLeft),srcTop + (cTop - dstTop),
                    cRight - cLeft,cBottom - cTop,
                    cLeft-clip.left,cTop-clip.top,rop3,flipY);
                }
            }
        }
}

/****************************************************************************
PARAMETERS:
srcLeft - Left coordinate of the source rectangle to copy
srcTop  - Top coordinate of the source rectangle to copy
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
dstLeft - Left coordinate of destination
dstTop  - Top coordinate of destination
rop3    - ROP3 mix code for the copy
****************************************************************************/
void NAPI MULDISP_BitBltColorPatt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    int         i = DC.numDevices;
    GA_layout   clip;
    N_int32     srcRight = srcLeft + width;
    N_int32     srcBottom = srcTop + height;
    N_int32     dstRight = dstLeft + width;
    N_int32     dstBottom = dstTop + height;
    N_int32     cLeft,cTop,cRight,cBottom;

    // First determine if the entire blit can be done in hardware on
    // a single device for speed.
    i = DC.numDevices;
    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_RECT(srcLeft,srcTop,srcRight,srcBottom,cLeft,cTop,cRight,cBottom,clip)
            && EQUAL_RECT(srcLeft,srcTop,srcRight,srcBottom,cLeft,cTop,cRight,cBottom)
            && VISIBLE_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom,clip)
            && EQUAL_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom)) {
            DC.draw2d[i].BitBltColorPatt(
                srcLeft-clip.left,srcTop-clip.top,
                width,height,
                dstLeft-clip.left,dstTop-clip.top,rop3);
            return;
            }
        else
            break;
        }

    // First read source bitmap data from video memory into transfer buffer
    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_RECT(srcLeft,srcTop,srcRight,srcBottom,cLeft,cTop,cRight,cBottom,clip)) {
            if (EQUAL_RECT(srcLeft,srcTop,srcRight,srcBottom,cLeft,cTop,cRight,cBottom)) {
                DC.draw2d[i].GetBitmapSys(DC.buffer,DC.bufferPitch,
                    cLeft-clip.left,cTop-clip.top,
                    width,height,
                    cLeft,cTop,
                    GA_REPLACE_MIX);
                break;
                }
            else {
                DC.draw2d[i].GetBitmapSys(DC.buffer,DC.bufferPitch,
                    cLeft-clip.left,cTop-clip.top,
                    cRight - cLeft,cBottom - cTop,
                    cLeft,cTop,
                    GA_REPLACE_MIX);
                }
            }
        }

    // Now blit the bitmap data from the transfer buffer to video memory
    i = DC.numDevices;
    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom,clip)) {
            if (EQUAL_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom)) {
                DC.draw2d[i].BitBltColorPattSys(DC.buffer,DC.bufferPitch,
                    srcLeft,srcTop,width,height,
                    cLeft-clip.left,cTop-clip.top,rop3,false);
                break;
                }
            else {
                DC.draw2d[i].BitBltColorPattSys(DC.buffer,DC.bufferPitch,
                    srcLeft + (cLeft - dstLeft),srcTop + (cTop - dstTop),
                    cRight - cLeft,cBottom - cTop,
                    cLeft-clip.left,cTop-clip.top,rop3,false);
                }
            }
        }
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3        - Mix code for the copy (GA_mixCodesType)
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI MULDISP_BitBltColorPattSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3,
    N_int32 flipY)
{
    int         i = DC.numDevices;
    GA_layout   clip;
    N_int32     dstRight = dstLeft + width;
    N_int32     dstBottom = dstTop + height;
    N_int32     cLeft,cTop,cRight,cBottom;

    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom,clip)) {
            if (EQUAL_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom)) {
                DC.draw2d[i].BitBltColorPattSys(srcAddr,srcPitch,
                    srcLeft,srcTop,width,height,
                    cLeft-clip.left,cTop-clip.top,rop3,flipY);
                return;
                }
            else {
                DC.draw2d[i].BitBltColorPattSys(srcAddr,srcPitch,
                    srcLeft + (cLeft - dstLeft),srcTop + (cTop - dstTop),
                    cRight - cLeft,cBottom - cTop,
                    cLeft-clip.left,cTop-clip.top,rop3,flipY);
                }
            }
        }
}

/****************************************************************************
PARAMETERS:
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
****************************************************************************/
void NAPI MULDISP_SrcTransBlt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    int         i = DC.numDevices;
    GA_layout   clip;
    N_int32     srcRight = srcLeft + width;
    N_int32     srcBottom = srcTop + height;
    N_int32     dstRight = dstLeft + width;
    N_int32     dstBottom = dstTop + height;
    N_int32     cLeft,cTop,cRight,cBottom;

    // First determine if the entire blit can be done in hardware on
    // a single device for speed.
    i = DC.numDevices;
    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_RECT(srcLeft,srcTop,srcRight,srcBottom,cLeft,cTop,cRight,cBottom,clip)
            && EQUAL_RECT(srcLeft,srcTop,srcRight,srcBottom,cLeft,cTop,cRight,cBottom)
            && VISIBLE_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom,clip)
            && EQUAL_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom)) {
            DC.draw2d[i].SrcTransBlt(
                srcLeft-clip.left,srcTop-clip.top,
                width,height,
                dstLeft-clip.left,dstTop-clip.top,mix,transparent);
            return;
            }
        else
            break;
        }

    // First read source bitmap data from video memory into transfer buffer
    i = DC.numDevices;
    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_RECT(srcLeft,srcTop,srcRight,srcBottom,cLeft,cTop,cRight,cBottom,clip)) {
            if (EQUAL_RECT(srcLeft,srcTop,srcRight,srcBottom,cLeft,cTop,cRight,cBottom)) {
                DC.draw2d[i].GetBitmapSys(DC.buffer,DC.bufferPitch,
                    cLeft-clip.left,cTop-clip.top,
                    width,height,
                    cLeft,cTop,
                    GA_REPLACE_MIX);
                break;
                }
            else {
                DC.draw2d[i].GetBitmapSys(DC.buffer,DC.bufferPitch,
                    cLeft-clip.left,cTop-clip.top,
                    cRight - cLeft,cBottom - cTop,
                    cLeft,cTop,
                    GA_REPLACE_MIX);
                }
            }
        }

    // Now blit the bitmap data from the transfer buffer to video memory
    i = DC.numDevices;
    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom,clip)) {
            if (EQUAL_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom)) {
                DC.draw2d[i].SrcTransBltSys(DC.buffer,DC.bufferPitch,
                    srcLeft,srcTop,width,height,
                    cLeft-clip.left,cTop-clip.top,mix,transparent,false);
                break;
                }
            else {
                DC.draw2d[i].SrcTransBltSys(DC.buffer,DC.bufferPitch,
                    srcLeft + (cLeft - dstLeft),srcTop + (cTop - dstTop),
                    cRight - cLeft,cBottom - cTop,
                    cLeft-clip.left,cTop-clip.top,mix,transparent,false);
                }
            }
        }
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI MULDISP_SrcTransBltSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent,
    N_int32 flipY)
{
    int         i = DC.numDevices;
    GA_layout   clip;
    N_int32     dstRight = dstLeft + width;
    N_int32     dstBottom = dstTop + height;
    N_int32     cLeft,cTop,cRight,cBottom;

    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom,clip)) {
            if (EQUAL_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom)) {
                DC.draw2d[i].SrcTransBltSys(srcAddr,srcPitch,
                    srcLeft,srcTop,width,height,
                    cLeft-clip.left,cTop-clip.top,mix,transparent,flipY);
                return;
                }
            else {
                DC.draw2d[i].SrcTransBltSys(srcAddr,srcPitch,
                    srcLeft + (cLeft - dstLeft),srcTop + (cTop - dstTop),
                    cRight - cLeft,cBottom - cTop,
                    cLeft-clip.left,cTop-clip.top,mix,transparent,flipY);
                }
            }
        }
}

/****************************************************************************
PARAMETERS:
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
****************************************************************************/
void NAPI MULDISP_DstTransBlt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    int         i = DC.numDevices;
    GA_layout   clip;
    N_int32     srcRight = srcLeft + width;
    N_int32     srcBottom = srcTop + height;
    N_int32     dstRight = dstLeft + width;
    N_int32     dstBottom = dstTop + height;
    N_int32     cLeft,cTop,cRight,cBottom;

    // First determine if the entire blit can be done in hardware on
    // a single device for speed.
    i = DC.numDevices;
    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_RECT(srcLeft,srcTop,srcRight,srcBottom,cLeft,cTop,cRight,cBottom,clip)
            && EQUAL_RECT(srcLeft,srcTop,srcRight,srcBottom,cLeft,cTop,cRight,cBottom)
            && VISIBLE_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom,clip)
            && EQUAL_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom)) {
            DC.draw2d[i].DstTransBlt(
                srcLeft-clip.left,srcTop-clip.top,
                width,height,
                dstLeft-clip.left,dstTop-clip.top,mix,transparent);
            return;
            }
        else
            break;
        }

    // First read source bitmap data from video memory into transfer buffer
    i = DC.numDevices;
    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_RECT(srcLeft,srcTop,srcRight,srcBottom,cLeft,cTop,cRight,cBottom,clip)) {
            if (EQUAL_RECT(srcLeft,srcTop,srcRight,srcBottom,cLeft,cTop,cRight,cBottom)) {
                DC.draw2d[i].GetBitmapSys(DC.buffer,DC.bufferPitch,
                    cLeft-clip.left,cTop-clip.top,
                    width,height,
                    cLeft,cTop,
                    GA_REPLACE_MIX);
                break;
                }
            else {
                DC.draw2d[i].GetBitmapSys(DC.buffer,DC.bufferPitch,
                    cLeft-clip.left,cTop-clip.top,
                    cRight - cLeft,cBottom - cTop,
                    cLeft,cTop,
                    GA_REPLACE_MIX);
                }
            }
        }

    // Now blit the bitmap data from the transfer buffer to video memory
    i = DC.numDevices;
    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom,clip)) {
            if (EQUAL_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom)) {
                DC.draw2d[i].DstTransBltSys(DC.buffer,DC.bufferPitch,
                    srcLeft,srcTop,width,height,
                    cLeft-clip.left,cTop-clip.top,mix,transparent,false);
                break;
                }
            else {
                DC.draw2d[i].DstTransBltSys(DC.buffer,DC.bufferPitch,
                    srcLeft + (cLeft - dstLeft),srcTop + (cTop - dstTop),
                    cRight - cLeft,cBottom - cTop,
                    cLeft-clip.left,cTop-clip.top,mix,transparent,false);
                }
            }
        }
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI MULDISP_DstTransBltSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent,
    N_int32 flipY)
{
    int         i = DC.numDevices;
    GA_layout   clip;
    N_int32     dstRight = dstLeft + width;
    N_int32     dstBottom = dstTop + height;
    N_int32     cLeft,cTop,cRight,cBottom;

    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom,clip)) {
            if (EQUAL_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom)) {
                DC.draw2d[i].DstTransBltSys(srcAddr,srcPitch,
                    srcLeft,srcTop,width,height,
                    cLeft-clip.left,cTop-clip.top,mix,transparent,flipY);
                return;
                }
            else {
                DC.draw2d[i].DstTransBltSys(srcAddr,srcPitch,
                    srcLeft + (cLeft - dstLeft),srcTop + (cTop - dstTop),
                    cRight - cLeft,cBottom - cTop,
                    cLeft-clip.left,cTop-clip.top,mix,transparent,flipY);
                }
            }
        }
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstWidth    - Width of the destination rectangle in pixels
dstHeight   - Height of the destination rectangle in scanlines
doClip      - True if the blit should be clipped
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle
mix         - Mix code for the copy (GA_mixCodesType)
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI MULDISP_StretchBltSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    N_int32 doClip,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom,
    N_int32 mix,
    N_int32 flipY)
{
    // TODO: Handle this!
}

/****************************************************************************
PARAMETERS:
srcAddr         - Address of source bitmap in system memory
srcPitch        - Pitch of source bitmap in bytes
srcLeft         - Left coordinate within source bitmap to copy
srcTop          - Top coordinate within source bitmap to copy
width           - Width of the rectangle in pixels
height          - Height of the rectangle in scanlines
dstLeft         - Left coordinate of destination
dstTop          - Top coordinate of destination
bitsPerPixel    - Pixel depth for the bitmap to copy
pixelFormat     - Pixel format for the bitmap to copy
dstPal          - Color palette for the destination device (if dest is 8bpp)
srcPal          - Color palette for the bitmap to copy (if source is 8bpp)
dither          - True if the output should be dithered
mix             - Mix code for the copy (GA_mixCodesType)
flipY           - True if the bitmap should be flipped in Y

RETURNS:
0 on success, -1 on failure.
****************************************************************************/
N_int32 NAPI MULDISP_ConvertBltSys_Obsolete(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pixelFormat,
    GA_palette *dstPal,
    GA_palette *srcPal,
    N_int32 dither,
    N_int32 mix,
    N_int32 flipY)
{
    int         i = DC.numDevices;
    GA_layout   clip;
    N_int32     dstRight = dstLeft + width;
    N_int32     dstBottom = dstTop + height;
    N_int32     cLeft,cTop,cRight,cBottom;

    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom,clip)) {
            if (EQUAL_RECT(dstLeft,dstTop,dstRight,dstBottom,cLeft,cTop,cRight,cBottom)) {
                DC.draw2d[i].ConvertBltSys_Obsolete(srcAddr,srcPitch,
                    srcLeft,srcTop,width,height,
                    cLeft-clip.left,cTop-clip.top,
                    bitsPerPixel,pixelFormat,dstPal,srcPal,dither,mix,flipY);
                return 0;
                }
            else {
                DC.draw2d[i].ConvertBltSys_Obsolete(srcAddr,srcPitch,
                    srcLeft + (cLeft - dstLeft),srcTop + (cTop - dstTop),
                    cRight - cLeft,cBottom - cTop,
                    cLeft-clip.left,cTop-clip.top,
                    bitsPerPixel,pixelFormat,dstPal,srcPal,dither,mix,flipY);
                }
            }
        }
    return 0;
}

/****************************************************************************
PARAMETERS:
srcAddr         - Address of source bitmap in system memory
srcPitch        - Pitch of source bitmap in bytes
srcLeft         - Left coordinate of the source rectangle to copy
srcTop          - Top coordinate of the source rectangle to copy
srcWidth        - Width of the source rectangle in pixels
srcHeight       - Height of the source rectangle in scanlines
dstLeft         - Left coordinate of destination
dstTop          - Top coordinate of destination
dstWidth        - Width of the destination rectangle in pixels
dstHeight       - Height of the destination rectangle in scanlines
doClip          - True if the blit should be clipped
clipLeft        - Left coordinate of clip rectangle
clipTop         - Top coordinate of clip rectangle
clipRight       - Right coordinate of clip rectangle
clipBottom      - Bottom coordinate of clip rectangle
bitsPerPixel    - Pixel depth for the bitmap to copy
pixelFormat     - Pixel format for the bitmap to copy
dstPal          - Color palette for the destination device (if dest is 8bpp)
srcPal          - Color palette for the bitmap to copy (if source is 8bpp)
dither          - True if the output should be dithered
mix             - Mix code for the copy (GA_mixCodesType)
flipY           - True if the bitmap should be flipped in Y

RETURNS:
0 on success, -1 on failure.
****************************************************************************/
N_int32 NAPI MULDISP_StretchConvertBltSys_Obsolete(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    N_int32 doClip,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pixelFormat,
    GA_palette *dstPal,
    GA_palette *srcPal,
    N_int32 dither,
    N_int32 mix,
    N_int32 flipY)
{
    // TODO: Handle this!
    return -1;
}

/****************************************************************************
PARAMETERS:
dstAddr     - Address of destination bitmap in system memory
dstPhysAddr - Physical address of destination bitmap in system memory
dstPitch    - Pitch of destination bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
****************************************************************************/
void NAPI MULDISP_GetBitmapSys(
    void *dstAddr,
    N_int32 dstPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix)
{
    int         i = DC.numDevices;
    GA_layout   clip;
    N_int32     srcRight = srcLeft + width;
    N_int32     srcBottom = srcTop + height;
    N_int32     cLeft,cTop,cRight,cBottom;

    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_RECT(srcLeft,srcTop,srcRight,srcBottom,cLeft,cTop,cRight,cBottom,clip)) {
            if (EQUAL_RECT(srcLeft,srcTop,srcRight,srcBottom,cLeft,cTop,cRight,cBottom)) {
                DC.draw2d[i].GetBitmapSys(dstAddr,dstPitch,
                    cLeft-clip.left,cTop-clip.top,
                    width,height,
                    dstLeft,dstTop,
                    mix);
                return;
                }
            else {
                DC.draw2d[i].GetBitmapSys(dstAddr,dstPitch,
                    cLeft-clip.left,cTop-clip.top,
                    cRight - cLeft,cBottom - cTop,
                    dstLeft + (cLeft - srcLeft),dstTop + (cTop - srcTop),
                    mix);
                }
            }
        }
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of ellipse rectangle
top     - Top coordinate of ellipse rectangle
A       - Major axis dimension
B       - Minor axis dimension
****************************************************************************/
void NAPI MULDISP_DrawEllipse(
    N_int32 left,
    N_int32 top,
    N_int32 A,
    N_int32 B)
{
    int         i = DC.numDevices;
    GA_layout   clip;
    N_int32     right = left + A;
    N_int32     bottom = top + B;
    N_int32     cLeft,cTop,cRight,cBottom;

    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_RECT(left,top,right,bottom,cLeft,cTop,cRight,cBottom,clip)) {
            if (EQUAL_RECT(left,top,right,bottom,cLeft,cTop,cRight,cBottom)) {
                DC.draw2d[i].DrawEllipse(left-clip.left,top-clip.top,A,B);
                return;
                }
            else {
                DC.draw2d[i].ClipEllipse(left-clip.left,top-clip.top,A,B,
                    cLeft-clip.left,cTop-clip.top,cRight-clip.left,cBottom-clip.top);
                }
            }
        }
}

/****************************************************************************
PARAMETERS:
left        - Left coordinate of ellipse rectangle
top         - Top coordinate of ellipse rectangle
A           - Major axis dimension
B           - Minor axis dimension
clipLeft    - Left coordinate for clip rectangle
clipTop     - Top coordinate for clip rectangle
clipRight   - Right coordinate for clip rectangle
clipBottom  - Bottom coordinate for clip rectangle
****************************************************************************/
void NAPI MULDISP_ClipEllipse(
    N_int32 left,
    N_int32 top,
    N_int32 A,
    N_int32 B,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
    int         i = DC.numDevices;
    GA_layout   clip;
    N_int32     right = left + A;
    N_int32     bottom = top + B;
    N_int32     cLeft,cTop,cRight,cBottom;

    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_RECT(left,top,right,bottom,cLeft,cTop,cRight,cBottom,clip)) {
            if (EQUAL_RECT(left,top,right,bottom,cLeft,cTop,cRight,cBottom)) {
                if (VISIBLE_RECT2(clipLeft,clipTop,clipRight,clipBottom,cLeft,cTop,cRight,cBottom)) {
                    DC.draw2d[i].ClipEllipse(left-clip.left,top-clip.top,A,B,
                        cLeft-clip.left,cTop-clip.top,cRight-clip.left,cBottom-clip.top);
                    }
                return;
                }
            else {
                if (VISIBLE_RECT2(clipLeft,clipTop,clipRight,clipBottom,cLeft,cTop,cRight,cBottom)) {
                    DC.draw2d[i].ClipEllipse(left-clip.left,top-clip.top,A,B,
                        cLeft-clip.left,cTop-clip.top,cRight-clip.left,cBottom-clip.top);
                    }
                }
            }
        }
}

/****************************************************************************
PARAMETERS:
x   - X coordinate to read pixel from
y   - Y coordinate to read pixel from

RETURNS:
Color read from the pixel on the screen
****************************************************************************/
GA_color NAPI MULDISP_GetPixel(
    N_int32 x,
    N_int32 y)
{
    int         i = DC.numDevices;
    GA_layout   clip;

    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_POINT(x,y,clip))
            return DC.draw2d[i].GetPixel(x-clip.left,y-clip.top);
        }
    return 0;
}

/****************************************************************************
PARAMETERS:
x   - X coordinate to plot pixel at
y   - Y coordinate to plot pixel at
****************************************************************************/
void NAPI MULDISP_PutPixel(
    N_int32 x,
    N_int32 y)
{
    int         i = DC.numDevices;
    GA_layout   clip;

    while (i--) {
        clip = DC.bounds[i];
        if (VISIBLE_POINT(x,y,clip)) {
            DC.draw2d[i].PutPixel(x-clip.left,y-clip.top);
            return;
            }
        }
}

/****************************************************************************
PARAMETERS:
cursor      - Pointer to an GA_monoCursor structure for the cursor image

REMARKS:
This function downloads the specified monochrome cursor definition from the
application into the hardware cursor. The cursor data is passed by the
application in the GA_monoCursor structure.
****************************************************************************/
void NAPI MULDISP_SetMonoCursor(
    GA_monoCursor *cursor)
{
    if (DC.enabled) {
        int i = DC.numDevices;

        while (i--) {
            SET_ACTIVE(i);
            DC.cursor[i].SetMonoCursor(cursor);
            }
        SET_MIXED_MODE();
        }
    else
        DC.cursor[0].SetMonoCursor(cursor);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_SetMonoCursorColor(
    GA_palette *foreground,
    GA_palette *background)
{
    if (DC.enabled) {
        int i = DC.numDevices;

        while (i--) {
            SET_ACTIVE(i);
            DC.cursor[i].SetMonoCursorColor(foreground,background);
            }
        SET_MIXED_MODE();
        }
    else
        DC.cursor[0].SetMonoCursorColor(foreground,background);
}

/****************************************************************************
PARAMETERS:
cursor  - Pointer to an GA_colorCursor structure for the cursor image

REMARKS:
This function downloads the specified colorCurs cursor definition from the
application into the hardware cursor.
****************************************************************************/
void NAPI MULDISP_SetColorCursor(
    GA_colorCursor *cursor)
{
    if (DC.enabled) {
        int i = DC.numDevices;

        while (i--) {
            SET_ACTIVE(i);
            DC.cursor[i].SetColorCursor(cursor);
            }
        SET_MIXED_MODE();
        }
    else
        DC.cursor[0].SetColorCursor(cursor);
}

/****************************************************************************
PARAMETERS:
cursor  - Pointer to an GA_colorCursor256 structure for the cursor image

REMARKS:
This function downloads the specified colorCurs cursor definition from the
application into the hardware cursor.
****************************************************************************/
void NAPI MULDISP_SetColorCursor256(
    GA_colorCursor256 *cursor)
{
    if (DC.enabled) {
        int i = DC.numDevices;

        while (i--) {
            SET_ACTIVE(i);
            DC.cursor[i].SetColorCursor256(cursor);
            }
        SET_MIXED_MODE();
        }
    else
        DC.cursor[0].SetColorCursor256(cursor);
}

/****************************************************************************
PARAMETERS:
cursor  - Pointer to an GA_colorCursorRGB structure for the cursor image

REMARKS:
This function downloads the specified colorCurs cursor definition from the
application into the hardware cursor.
****************************************************************************/
void NAPI MULDISP_SetColorCursorRGB(
    GA_colorCursorRGB *cursor)
{
    if (DC.enabled) {
        int i = DC.numDevices;

        while (i--) {
            SET_ACTIVE(i);
            DC.cursor[i].SetColorCursorRGB(cursor);
            }
        SET_MIXED_MODE();
        }
    else
        DC.cursor[0].SetColorCursorRGB(cursor);
}

/****************************************************************************
PARAMETERS:
cursor  - Pointer to an GA_colorCursorRGBA structure for the cursor image

REMARKS:
This function downloads the specified colorCurs cursor definition from the
application into the hardware cursor.
****************************************************************************/
void NAPI MULDISP_SetColorCursorRGBA(
    GA_colorCursorRGBA *cursor)
{
    if (DC.enabled) {
        int i = DC.numDevices;

        while (i--) {
            SET_ACTIVE(i);
            DC.cursor[i].SetColorCursorRGBA(cursor);
            }
        SET_MIXED_MODE();
        }
    else
        DC.cursor[0].SetColorCursorRGBA(cursor);
}

/****************************************************************************
PARAMETERS:
x   - X coordinate of the cursor position
y   - Y coordinate of the cursor position

REMARKS:
This function sets the location of the hardware cursor in display
coordinates. This function takes the X and Y coordinates of the new cursor
location. This function will place the cursor so that the hotspot of the
cursor image is located at the specified (X,Y) location, and will correctly
handle special cases where the cursor definition needs to be located off the
edges of the display screen (such as when X=0, Y=0 and HotX,HotY > 0).
****************************************************************************/
N_int32 NAPI MULDISP_SetCursorPos(
    N_int32 x,
    N_int32 y)
{
    if (DC.enabled) {
        int         i = DC.numDevices;
        GA_layout   clip;
        N_int32     left,top,right,bottom;
        N_int32     cLeft,cTop,cRight,cBottom;

        DC.cntCursorX = x;
        DC.cntCursorY = y;
        while (i--) {
            SET_ACTIVE(i);
            left = x - DC.refState[i]->cntHotX;
            top = y - DC.refState[i]->cntHotY;
            right = left + DC.refState[i]->cntCursorWidth;
            bottom = top + DC.refState[i]->cntCursorHeight;
            clip = DC.bounds[i];
            if (VISIBLE_RECT(left,top,right,bottom,cLeft,cTop,cRight,cBottom,clip)) {
                if (DC.cursorVisible[i] == 0) {
                    DC.cursor[i].ShowCursor(DC.isCursorVisible);
                    DC.cursorVisible[i] = 1;
                    }
                DC.cursor[i].SetCursorPos(x-clip.left,y-clip.top);
                }
            else {
                if (DC.cursorVisible[i] == 1) {
                    DC.cursor[i].ShowCursor(false);
                    DC.cursorVisible[i] = 0;
                    }
                }
            }
        SET_MIXED_MODE();
        return true;
        }
    else
        return DC.cursor[0].SetCursorPos(x,y);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_ShowCursor(
    N_int32 visible)
{
    if (DC.enabled) {
        int i = DC.numDevices;

        DC.isCursorVisible = visible;
        while (i--)
            DC.cursorVisible[i] = !visible;
        MULDISP_SetCursorPos(DC.cntCursorX,DC.cntCursorY);
        }
    else
        DC.cursor[0].ShowCursor(visible);
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of display affected
top     - Top coordinate of display affected
right   - Right coordinate of display affected (exclusive)
bottom  - Bottom coordinate of display affected (exclusive)

REMARKS:
This function is used when the SNAP driver is using software monoCurs and
colorCurs cursor acceleration. The cursor images are stored in offscreen
memory buffers, and the hardware blitter is used to draw the cursor images.
This function is used to inform the SNAP driver that the shell display
driver is about to draw to a particular location of memory, and to remove
the cursor image from the screen temporarily while the drawing is taking
place if the drawing region and the cursor position overlap.
****************************************************************************/
void NAPI MULDISP_BeginAccess(
    N_int32 left,
    N_int32 top,
    N_int32 right,
    N_int32 bottom)
{
    if (DC.enabled) {
        int         i = DC.numDevices;
        GA_layout   clip;
        N_int32     cLeft,cTop,cRight,cBottom;

        while (i--) {
            clip = DC.bounds[i];
            if (VISIBLE_RECT(left,top,right,bottom,cLeft,cTop,cRight,cBottom,clip))
                DC.cursor[i].BeginAccess(cLeft-clip.left,cTop-clip.top,cRight-clip.left,cBottom-clip.top);
            }
        }
    else
        DC.cursor[0].BeginAccess(left,top,right,bottom);
}

/****************************************************************************
REMARKS:
Handles multi-controller version of function.
****************************************************************************/
void NAPI MULDISP_EndAccess(void)
{
    if (DC.enabled) {
        int i = DC.numDevices;

        while (i--)
            DC.cursor[i].EndAccess();
        }
    else
        DC.cursor[0].EndAccess();
}

/****************************************************************************
REMARKS:
This function return true if the cursor is currently running in hardware,
or false if it is currently running in software.
****************************************************************************/
N_int32 NAPI MULDISP_IsHardwareCursor(void)
{
    if (DC.enabled) {
        int i = DC.numDevices;

        while (i--) {
            if (!DC.cursor[i].IsHardwareCursor())
                return false;
            }
        return true;
        }
    else
        return DC.cursor[0].IsHardwareCursor();
}
