/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Header file for the SciTech SNAP Graphics 2D Multi-Controller
*               filter driver internal functions.
*
****************************************************************************/

#ifndef __MULINT_H
#define __MULINT_H

#include "ref2dint.h"

/*---------------------- Macros and type definitions ----------------------*/

#pragma pack(4)

#define MULDISP_MAX_MODES   255
#define MULDISP_MAX_REFRESH 20

typedef ibool (NAPIP QueryFunc_t)(N_uint32 id,N_int32 safetyLevel,void *funcs);

// TODO: Make this constant dynamic!!

#define MUL_MAX_DEVICES 4

/****************************************************************************
REMARKS:
Structure to store the current state of the 2D portrait filter driver.

MEMBERS:
shared              - True if loaded into shared memory
enabled             - True if multi-controller mode is enabled
hwCtx               - Hardware device for primary driver
mi                  - Mode information block for the current hardware mode
buf                 - Current draw buffer
bytesPerPixel       - Number of bytes in a single pixel
ref2d               - Pointer to the ref2d driver
refState            - Pointer to the ref2d driver internal state
****************************************************************************/
typedef struct {
    ibool               enabled;
    GA_options          opt;
    GA_globalOptions    *gOpt;
    GA_modeInfo         gMi;
    GA_buffer           buf;
    N_int32             numDevices;
    N_int32             bufferSize;
    N_int32             bufferPitch;
    void                *buffer;
    N_uint16            AvailableModes[MULDISP_MAX_MODES];
    N_int32             RefreshRateList[MULDISP_MAX_REFRESH];
    N_uint32            virtualMode8;
    N_uint32            virtualMode15;
    N_uint32            virtualMode16;
    N_uint32            virtualMode24;
    N_uint32            virtualMode32;
    N_int32             srcBlendFunc;
    N_int32             dstBlendFunc;
    N_uint32            cntMode;
    N_uint32            oldMode[MUL_MAX_DEVICES];
    N_int32             cntCursorX;
    N_int32             cntCursorY;
    N_int32             isCursorVisible;
    N_int32             cursorVisible[MUL_MAX_DEVICES];
    GA_layout           bounds[MUL_MAX_DEVICES];
    GA_devCtx           *hwCtx[MUL_MAX_DEVICES];
    REF2D_driver        *ref2d[MUL_MAX_DEVICES];
    REF2D_state         *refState[MUL_MAX_DEVICES];
    GA_modeInfo         virtualMi;
    N_uint32            modes[MUL_MAX_DEVICES];
    GA_modeInfo         mi[MUL_MAX_DEVICES];
    GA_initFuncs        init[MUL_MAX_DEVICES];
    GA_driverFuncs      driver[MUL_MAX_DEVICES];
    GA_cursorFuncs      cursor[MUL_MAX_DEVICES];
    GA_2DStateFuncs     state2d[MUL_MAX_DEVICES];
    GA_2DRenderFuncs    draw2d[MUL_MAX_DEVICES];
    QueryFunc_t         QueryFunctionsDC;
    N_uint16            *AvailableModesDC[MUL_MAX_DEVICES];
    GA_setActiveDevice_t setActiveDevice;
    } MULDISP_state;

/****************************************************************************
REMARKS:
Macro to determine if multi-controller mode is enabled or not.
****************************************************************************/
#define IS_MULDISP_OPT() \
    (DC.gOpt->bVirtualDisplay == 1 || (DC.gOpt->bVirtualDisplay == 0 && DC.opt.bVirtualDisplay == 1))

/****************************************************************************
REMARKS:
Macro to determine if multi-controller mode is enabled or not.
****************************************************************************/
#define IS_MULDISP() \
    (DC.hwCtx[0] != 0 && IS_MULDISP_OPT() && DC.gMi.BitsPerPixel >= 8)

/****************************************************************************
REMARKS:
Macro to determine if the mode is our multi-controller mode.
****************************************************************************/
#define IS_MULDISP_MODE() \
    (DC.cntMode != 0)

/****************************************************************************
REMARKS:
Macro to determine number of devices to iterate over
****************************************************************************/
#define NUM_DEVICES() \
    (IS_MULDISP_OPT() ? DC.numDevices : 1)

/****************************************************************************
REMARKS:
Macro to determine number of devices to iterate over
****************************************************************************/
#define NUM_DEVICES1() \
    ((IS_MULDISP_OPT() && IS_MULDISP_MODE()) ? DC.numDevices : 1)

/****************************************************************************
REMARKS:
Macro to determine number of devices to iterate over
****************************************************************************/
#define NUM_DEVICES2() \
    (DC.enabled ? DC.numDevices : 1)

/****************************************************************************
REMARKS:
Macro to set the active device for exclusive operation (VGA I/O and memory)
****************************************************************************/
#define SET_ACTIVE(i) \
    DC.setActiveDevice(i)

/****************************************************************************
REMARKS:
Macro to switch to mixed mode for MMIO rendering functions
****************************************************************************/
#define SET_MIXED_MODE() \
    DC.setActiveDevice(GA_DEVICE_MIXED_MODE)

/****************************************************************************
REMARKS:
Macro to clip rectangle to the visible rectangle to a clip rectangle. Will
return true if the clipped rectangle is not empty.
****************************************************************************/
#define VISIBLE_RECT(iLeft,iTop,iRight,iBottom,cLeft,cTop,cRight,cBottom,clip)  \
   ((cLeft) = MAX((int)(iLeft),(int)(clip).left),                           \
    (cRight) = MIN((int)(iRight),(int)(clip).right),                        \
    (cTop) = MAX((int)(iTop),(int)(clip).top),                              \
    (cBottom) = MIN((int)(iBottom),(int)(clip).bottom),                     \
    (cBottom) > (cTop) && (cRight) > (cLeft))

/****************************************************************************
REMARKS:
Macro to clip two rectangles. This version puts the result into the second
set of clip rectangle parameters.
****************************************************************************/
#define VISIBLE_RECT2(iLeft,iTop,iRight,iBottom,cLeft,cTop,cRight,cBottom)  \
   ((cLeft) = MAX((int)iLeft,(int)cLeft),                                   \
    (cRight) = MIN((int)iRight,(int)cRight),                                \
    (cTop) = MAX((int)iTop,(int)cTop),                                      \
    (cBottom) = MIN((int)iBottom,(int)cBottom),                             \
    (cBottom) > (cTop) && (cRight) > (cLeft))

/****************************************************************************
REMARKS:
Macro to clip rectangle to the visible rectangle to a clip rectangle. Will
return true if the clipped rectangle is not empty.
****************************************************************************/
#define VISIBLE_POINT(x,y,clip)                                             \
   ((x) >= (clip).left && (x) < (clip).right &&                             \
    (y) >= (clip).top && (y) < (clip).right)

/****************************************************************************
REMARKS:
Macro to determine if two rectangles are equal
****************************************************************************/
#define EQUAL_RECT(left,top,right,bottom,cLeft,cTop,cRight,cBottom)         \
   ((left) == (cLeft) && (top) == (cTop) &&                                 \
    (right) == (cRight) && (bottom) == (cBottom))

#pragma pack()

/*---------------------------- Global Variables ---------------------------*/

extern MULDISP_state    MULDISP_devCtx;
extern REF2D_driver     MULDISP_drv;
extern GA_2DRenderFuncs _VARAPI MULDISP_draw2dCnt;

#undef  DC
#undef  DRV
#define DC      MULDISP_devCtx
#define DC2D    (*MULDISP_devCtx.refState)
#define DRV     MULDISP_drv

/*-------------------------- Function Prototypes --------------------------*/

#ifdef  __cplusplus
extern "C" {                        /* Use "C" linkage when in C++ mode */
#endif

/* Library initialisation entry points */

ibool       NAPI MULDISP_QueryFunctionsDC(N_uint32 id,N_int32 safetyLevel,void *funcs);
ibool       NAPI MULDISP_QueryFunctions(N_uint32 id,void *funcs);
N_int32     NAPI MULDISP_SetDrawBufferSW(GA_buffer *drawBuf,void *framebuffer,N_int32 bitsPerPixel,GA_pixelFormat *pf,GA_devCtx *hwCtx,N_int32 softwareOnly);
void        NAPI MULDISP_SetDrawSurface(void *surface,N_int32 xRes,N_int32 yRes,N_int32 bytesPerLine,N_int32 bitsPerPixel,GA_pixelFormat *pf);
void        NAPI MULDISP_UnloadDriver(void);
ibool       NAPI MULDISP_InitFilter(GA_modeInfo *modeInfo,N_int32 transferStart,struct _REF2D_driver *ref2d,struct _REF2D_driver **drv);
void        NAPI MULDISP_DeInitFilter(void);
void        NAPI MULDISP_GlobalOptionsUpdated(void);

/* Multi-controller init functions */

void        NAPI MULDISP_GetConfigInfo(GA_configInfo *info);
void        NAPI MULDISP_GetModeProfile(GA_modeProfile *profile);
void        NAPI MULDISP_SetModeProfile(GA_modeProfile *profile);
void        NAPI MULDISP_GetOptions(GA_options *options);
void        NAPI MULDISP_SetOptions(GA_options *options);
N_int32     NAPI MULDISP_GetVideoModeInfo(N_uint32 mode,GA_modeInfo *modeInfo);
N_int32     NAPI MULDISP_SetVideoMode(N_uint32 mode,N_int32 *virtualX,N_int32 *virtualY,N_int32 *bytesPerLine,N_int32 *maxMem,N_int32 refreshRate,GA_CRTCInfo *crtc);
N_uint32    NAPI MULDISP_GetVideoMode(void);
N_int32     NAPI MULDISP_GetCustomVideoModeInfo(N_int32 xRes,N_int32 yRes,N_int32 virtualX,N_int32 virtualY,N_int32 bitsPerPixel,GA_modeInfo *modeInfo);
N_int32     NAPI MULDISP_SetCustomVideoMode(N_int32 xRes,N_int32 yRes,N_int32 bitsPerPixel,N_uint32 flags,N_int32 *virtualX,N_int32 *virtualY,N_int32 *bytesPerLine,N_int32 *maxMem,GA_CRTCInfo *crtc);
N_uint32    NAPI MULDISP_GetClosestPixelClockV1(N_int32 xRes,N_int32 yRes,N_int32 bitsPerPixel,N_uint32 pixelClock);
N_uint32    NAPI MULDISP_GetClosestPixelClock(N_int32 outputHead,N_int32 xRes,N_int32 yRes,N_int32 bitsPerPixel,N_uint32 pixelClock);
void        NAPI MULDISP_GetCRTCTimings(GA_CRTCInfo *crtc);
void        NAPI MULDISP_SetCRTCTimings(GA_CRTCInfo *crtc);
void        NAPI MULDISP_SetGlobalRefresh(N_int32 refresh,N_int32 outputHead);
N_int32     NAPI MULDISP_SaveRestoreState(N_int32 subfunc,void *saveBuf);
N_int32     NAPI MULDISP_SetDisplayOutput(N_int32 device);
N_int32     NAPI MULDISP_SetDisplayOutputExt(N_int32 device,N_int32 outputHead);
N_int32     NAPI MULDISP_GetDisplayOutput(void);
N_int32     NAPI MULDISP_GetDisplayOutputExt(N_int32 outputHead);
void        NAPI MULDISP_SetSoftwareRenderFuncs(GA_2DRenderFuncs *softwareFuncs);
void        NAPI MULDISP_GetUniqueFilename(char *filename,int type);
void        NAPI MULDISP_GetMonitorInfo(GA_monitor *monitor,N_int32 outputHead);
void        NAPI MULDISP_SetMonitorInfo(GA_monitor *monitor,N_int32 outputHead);
void        NAPI MULDISP_GetCurrentVideoModeInfo(GA_modeInfo *modeInfo);

/* Multi-controller driver functions */

void        NAPI MULDISP_SetBank(N_int32 bank);
void        NAPI MULDISP_SetDisplayStart(N_int32 offset,N_int32 waitVRT);
void        NAPI MULDISP_SetDisplayStartXY(N_int32 x,N_int32 y,N_int32 waitVRT);
void        NAPI MULDISP_SetStereoDisplayStart(N_int32 leftOffset,N_int32 rightOffset,N_int32 waitVRT);
N_int32     NAPI MULDISP_GetDisplayStartStatus(void);
void        NAPI MULDISP_EnableStereoMode(N_int32 enable);
N_int32     NAPI MULDISP_IsVSync(void);
void        NAPI MULDISP_WaitVSync(void);
void        NAPI MULDISP_SetVSyncWidth(N_int32 width);
N_int32     NAPI MULDISP_GetVSyncWidth(void);
void        NAPI MULDISP_SetPaletteData(GA_palette *pal,N_int32 num,N_int32 index,N_int32 waitVRT);
void        NAPI MULDISP_GetPaletteData(GA_palette *pal,N_int32 num,N_int32 index);
void        NAPI MULDISP_SetGammaCorrectData(GA_palette *pal,N_int32 num,N_int32 index,N_int32 waitVRT);
void        NAPI MULDISP_GetGammaCorrectData(GA_palette *pal,N_int32 num,N_int32 index);
void        NAPI MULDISP_SetPaletteDataExt(GA_paletteExt *pal,N_int32 num,N_int32 index,N_int32 waitVRT);
void        NAPI MULDISP_GetPaletteDataExt(GA_paletteExt *pal,N_int32 num,N_int32 index);
void        NAPI MULDISP_SetGammaCorrectDataExt(GA_paletteExt *pal,N_int32 num,N_int32 index,N_int32 waitVRT);
void        NAPI MULDISP_GetGammaCorrectDataExt(GA_paletteExt *pal,N_int32 num,N_int32 index);

/* Multi-controller cursor functions */

void        NAPI MULDISP_SetMonoCursor(GA_monoCursor *cursor);
void        NAPI MULDISP_SetMonoCursorColor(GA_palette *foreground,GA_palette *background);
void        NAPI MULDISP_SetColorCursor(GA_colorCursor *cursor);
N_int32     NAPI MULDISP_SetCursorPos(N_int32 x,N_int32 y);
void        NAPI MULDISP_ShowCursor(N_int32 visible);
void        NAPI MULDISP_BeginAccess(N_int32 left,N_int32 top,N_int32 right,N_int32 bottom);
void        NAPI MULDISP_EndAccess(void);
void        NAPI MULDISP_SetColorCursor256(GA_colorCursor256 *cursor);
void        NAPI MULDISP_SetColorCursorRGB(GA_colorCursorRGB *cursor);
void        NAPI MULDISP_SetColorCursorRGBA(GA_colorCursorRGBA *cursor);
N_int32     NAPI MULDISP_IsHardwareCursor(void);

/* 2D state reference functions */

N_int32     NAPI MULDISP_SetDrawBuffer(GA_buffer *drawBuf);
N_int32     NAPI MULDISP_IsIdle(void);
void        NAPI MULDISP_WaitTillIdle(void);
void        NAPI MULDISP_EnableDirectAccess(void);
void        NAPI MULDISP_DisableDirectAccess(void);
N_int32     NAPI MULDISP_SetMix(N_int32 mix);
void        NAPI MULDISP_SetForeColor(GA_color color);
void        NAPI MULDISP_SetBackColor(GA_color color);
void        NAPI MULDISP_Set8x8MonoPattern(N_int32 index,GA_pattern *pattern);
void        NAPI MULDISP_Use8x8MonoPattern(N_int32 index);
void        NAPI MULDISP_Use8x8TransMonoPattern(N_int32 index);
void        NAPI MULDISP_Set8x8ColorPattern(N_int32 index,GA_colorPattern *pattern);
void        NAPI MULDISP_Use8x8ColorPattern(N_int32 index);
void        NAPI MULDISP_Use8x8TransColorPattern(N_int32 index,GA_color transparent);
void        NAPI MULDISP_SetLineStipple(GA_stipple stipple);
void        NAPI MULDISP_SetLineStippleCount(N_uint32 count);
void        NAPI MULDISP_SetPlaneMask(N_uint32 mask);
void        NAPI MULDISP_SetSrcBlendFunc(N_int32 blendFunc);
void        NAPI MULDISP_SetDstBlendFunc(N_int32 blendFunc);
void        NAPI MULDISP_SetBlendFunc(N_int32 srcBlendFunc,N_int32 dstBlendFunc);
void        NAPI MULDISP_SetAlphaValue(N_uint8 alpha);
void        NAPI MULDISP_SetupForWindows(void);
void        NAPI MULDISP_ResetForWindows(void);
void        NAPI MULDISP_SetLineStyle(N_uint32 styleMask,N_uint32 styleStep,N_uint32 styleValue);

/* 2D generic reference functions */

void        NAPI MULDISP_DrawScan(N_int32 y,N_int32 x1,N_int32 x2);
void        NAPI MULDISP_DrawPattScan(N_int32 y,N_int32 x1,N_int32 x2);
void        NAPI MULDISP_DrawColorPattScan(N_int32 y,N_int32 x1,N_int32 x2);
void        NAPI MULDISP_DrawScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI MULDISP_DrawPattScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI MULDISP_DrawColorPattScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI MULDISP_DrawEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI MULDISP_DrawPattEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI MULDISP_DrawColorPattEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI MULDISP_DrawFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI MULDISP_DrawPattFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI MULDISP_DrawColorPattFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI MULDISP_DrawRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI MULDISP_DrawPattRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI MULDISP_DrawColorPattRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI MULDISP_DrawRectPatt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,N_int32 rop3);
void        NAPI MULDISP_DrawRectColorPatt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,N_int32 rop3);
void        NAPI MULDISP_DrawTrap(GA_trap *trap);
void        NAPI MULDISP_DrawPattTrap(GA_trap *trap);
void        NAPI MULDISP_DrawColorPattTrap(GA_trap *trap);
void        NAPI MULDISP_DrawLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast);
void        NAPI MULDISP_DrawBresenhamLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags);
void        NAPI MULDISP_DrawStippleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent);
void        NAPI MULDISP_DrawBresenhamStippleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent);
void        NAPI MULDISP_PutMonoImageMSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent);
void        NAPI MULDISP_PutMonoImageLSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent);
void        NAPI MULDISP_ClipMonoImageMSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipWidth,N_int32 clipHeight);
void        NAPI MULDISP_ClipMonoImageLSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipWidth,N_int32 clipHeight);
void        NAPI MULDISP_BitBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI MULDISP_BitBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,N_int32 flipY);
void        NAPI MULDISP_BitBltPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI MULDISP_BitBltPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void        NAPI MULDISP_BitBltColorPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI MULDISP_BitBltColorPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void        NAPI MULDISP_SrcTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI MULDISP_SrcTransBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent,N_int32 flipY);
void        NAPI MULDISP_DstTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI MULDISP_DstTransBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent,N_int32 flipY);
void        NAPI MULDISP_StretchBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix,N_int32 flipY);
N_int32     NAPI MULDISP_ConvertBltSys_Obsolete(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix,N_int32 flipY);
N_int32     NAPI MULDISP_StretchConvertBltSys_Obsolete(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix,N_int32 flipY);
void        NAPI MULDISP_GetBitmapSys(void *dstAddr,N_int32 dstPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI MULDISP_DrawEllipse(N_int32 left,N_int32 top,N_int32 A,N_int32 B);
void        NAPI MULDISP_ClipEllipse(N_int32 left,N_int32 top,N_int32 A,N_int32 B,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
GA_color    NAPI MULDISP_GetPixel(N_int32 x,N_int32 y);
void        NAPI MULDISP_PutPixel(N_int32 x,N_int32 y);
N_int32     NAPI MULDISP_DrawClippedLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32     NAPI MULDISP_DrawClippedBresenhamLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32     NAPI MULDISP_DrawClippedStippleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32     NAPI MULDISP_DrawClippedBresenhamStippleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
void        NAPI MULDISP_DrawStyleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent);
void        NAPI MULDISP_DrawBresenhamStyleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent);
N_int32     NAPI MULDISP_DrawClippedStyleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32     NAPI MULDISP_DrawClippedBresenhamStyleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);

/* Assembler stubs to call the functions via the global REF_draw2d */

void        NAPI MSTUB_DrawScan(N_int32 y,N_int32 x1,N_int32 x2);
void        NAPI MSTUB_DrawPattScan(N_int32 y,N_int32 x1,N_int32 x2);
void        NAPI MSTUB_DrawColorPattScan(N_int32 y,N_int32 x1,N_int32 x2);
void        NAPI MSTUB_DrawScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI MSTUB_DrawPattScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI MSTUB_DrawColorPattScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI MSTUB_DrawEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI MSTUB_DrawPattEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI MSTUB_DrawColorPattEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI MSTUB_DrawFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI MSTUB_DrawPattFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI MSTUB_DrawColorPattFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI MSTUB_DrawRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI MSTUB_DrawPattRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI MSTUB_DrawColorPattRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI MSTUB_DrawRectPatt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,N_int32 rop3);
void        NAPI MSTUB_DrawRectColorPatt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,N_int32 rop3);
void        NAPI MSTUB_DrawTrap(GA_trap *trap);
void        NAPI MSTUB_DrawPattTrap(GA_trap *trap);
void        NAPI MSTUB_DrawColorPattTrap(GA_trap *trap);
void        NAPI MSTUB_DrawLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast);
void        NAPI MSTUB_DrawBresenhamLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags);
void        NAPI MSTUB_DrawStippleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent);
void        NAPI MSTUB_DrawBresenhamStippleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent);
void        NAPI MSTUB_PutMonoImageMSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent);
void        NAPI MSTUB_PutMonoImageLSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent);
void        NAPI MSTUB_ClipMonoImageMSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipWidth,N_int32 clipHeight);
void        NAPI MSTUB_ClipMonoImageLSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipWidth,N_int32 clipHeight);
void        NAPI MSTUB_BitBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI MSTUB_BitBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,N_int32 flipY);
void        NAPI MSTUB_BitBltPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI MSTUB_BitBltPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void        NAPI MSTUB_BitBltColorPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI MSTUB_BitBltColorPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void        NAPI MSTUB_SrcTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI MSTUB_SrcTransBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent,N_int32 flipY);
void        NAPI MSTUB_DstTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI MSTUB_DstTransBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent,N_int32 flipY);
void        NAPI MSTUB_StretchBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix);
void        NAPI MSTUB_StretchBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix,N_int32 flipY);
N_int32     NAPI MSTUB_ConvertBltSys_Obsolete(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix,N_int32 flipY);
N_int32     NAPI MSTUB_StretchConvertBltSys_Obsolete(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix,N_int32 flipY);
void        NAPI MSTUB_GetBitmapSys(void *dstAddr,N_int32 dstPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI MSTUB_DrawEllipse(N_int32 left,N_int32 top,N_int32 A,N_int32 B);
void        NAPI MSTUB_ClipEllipse(N_int32 left,N_int32 top,N_int32 A,N_int32 B,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
GA_color    NAPI MSTUB_GetPixel(N_int32 x,N_int32 y);
void        NAPI MSTUB_PutPixel(N_int32 x,N_int32 y);
N_int32     NAPI MSTUB_DrawClippedLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32     NAPI MSTUB_DrawClippedBresenhamLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32     NAPI MSTUB_DrawClippedStippleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32     NAPI MSTUB_DrawClippedBresenhamStippleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
void        NAPI MSTUB_DrawStyleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent);
void        NAPI MSTUB_DrawBresenhamStyleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent);
N_int32     NAPI MSTUB_DrawClippedStyleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32     NAPI MSTUB_DrawClippedBresenhamStyleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);

#ifdef  __cplusplus
}                                   /* End of "C" linkage for C++       */
#endif

#endif  /* __MULINT_H */
