/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser functions
*               for the SciTech SNAP Graphics Architecture.
*
****************************************************************************/

#include "ref2dint.h"

// NOTE: This module is broken up to get it to compile properly with
//       the IBM VisualAge compiler.

/*----------------------------- Implementation ----------------------------*/

#ifdef SUPPORT_1BPP

void REF2D_drawPixelR3_1_0(uchar *addr,int x,GA_color S,GA_color P)
{
    *addr &= ~MONO_MASK(x);
    (void)S;
    (void)P;
}

void REF2D_drawPixelR3_1_DPSoon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(D | (P | S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSona(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D & ~(P | S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PSon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P | S);
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPona(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S & ~(D | P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    P = (P ? mask : 0);
    S = ~(D | P);
    *addr = (uchar)((dest & ~mask) | (S & mask));
    (void)S;
}

void REF2D_drawPixelR3_1_PDSxnon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P | ~(D ^ S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSaon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P | (D & S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPnaa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S & (D & ~P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSxon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P | (D ^ S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPna(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    P = (P ? mask : 0);
    S = (D & ~P);
    *addr = (uchar)((dest & ~mask) | (S & mask));
    (void)S;
}

void REF2D_drawPixelR3_1_PSDnaon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P | (S & ~D));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPna(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S & ~P);
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSnaon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P | (D & ~S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSonon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P | ~(D | S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_Pn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    P = (P ? mask : 0);
    S = ~P;
    *addr = (uchar)((dest & ~mask) | (S & mask));
    (void)S;
}

void REF2D_drawPixelR3_1_PDSona(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P & ~(D | S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DSon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color D = *addr;
    GA_color mask = MONO_MASK(x);
    *addr = (uchar)((D & ~mask) | ((~D | (S ? mask : 0)) & mask));
    (void)P;
}

void REF2D_drawPixelR3_1_SDPxnon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S | ~(D ^ P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPaon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S | (D & P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSxnon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(D | ~(P ^ S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSaon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(D | (P & S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PSDPSanaxx(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P ^ (S ^ (D & ~(P & S))));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SSPxDSxaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S ^ ((S ^ P) & (D ^ S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPxPDxa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ((S ^ P) & (P ^ D));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPSanaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S ^ (D & ~(P & S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSPaox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P ^ (D | (S & P)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPSxaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S ^ (D & (P ^ S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PSDPaox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P ^ (S | (D & P)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DSPDxaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(D ^ (S & (P ^ D)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P ^ (D | S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSoan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P & (D | S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSnaa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D & (P & ~S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPxon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S | (D ^ P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DSna(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color D = *addr;
    GA_color mask = MONO_MASK(x);
    *addr = (uchar)((D & ~mask) | ((D & ~(S ? mask : 0)) & mask));
    (void)P;
}

void REF2D_drawPixelR3_1_SPDnaon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S | (P & ~D));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPxDSxa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ((S ^ P) & (D ^ S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSPanaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P ^ (D & ~(S & P)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPSaox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S ^ (D | (P & S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPSxnox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S ^ (D | ~(P ^ S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSxa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D & (P ^ S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PSDPSaoxxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P ^ (S ^ (D | (P & S))));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSana(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D & ~(P & S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SSPxPDxaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S ^ ((S ^ P) & (P ^ D)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPDSoax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S ^ (P & (D | S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PSDnox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P ^ (S | ~D));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PSDPxox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P ^ (S | (D ^ P)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PSDnoan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P & (S | ~D));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PSna(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P & ~S);
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPnaon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S | (D & ~P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPSoox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S ^ (D | (P | S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_Sn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color D = *addr;
    GA_color mask = MONO_MASK(x);
    *addr = (uchar)((D & ~mask) | (~(S ? mask : 0) & mask));
    (void)P;
}

void REF2D_drawPixelR3_1_SPDSaox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S ^ (P | (D & S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPDSxnox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S ^ (P | ~(D ^ S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S ^ (D | P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPoan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S & (D | P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PSDPoax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P ^ (S & (D | P)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPDnox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S ^ (P | ~D));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPDSxox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S ^ (P | (D ^ S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPDnoan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S & (P | ~D));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PSx(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P ^ S);
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPDSonox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S ^ (P | ~(D | S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPDSnaox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S ^ (P | (D & ~S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PSan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P & S);
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PSDnaa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P & (S & ~D));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSxon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(D | (P ^ S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDxPDxa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ((S ^ D) & (P ^ D));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPDSanaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S ^ (P & ~(D & S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDna(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color D = *addr;
    GA_color mask = MONO_MASK(x);
    *addr = (uchar)((D & ~mask) | ((~D & (S ? mask : 0)) & mask));
    (void)P;
}

void REF2D_drawPixelR3_1_DPSnaon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(D | (P & ~S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DSPDaox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D ^ (S | (P & D)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PSDPxaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P ^ (S & (D ^ P)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPxa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S & (D ^ P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSPDaoxxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P ^ (D ^ (S | (P & D))));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSDoax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D ^ (P & (S | D)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSnox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P ^ (D | ~S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPana(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S & ~(D & P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SSPxDSxoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S ^ ((S ^ P) | (D ^ S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSPxox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P ^ (D | (S ^ P)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSnoan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P & (D | ~S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDna(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    P = (P ? mask : 0);
    S = (P & ~D);
    *addr = (uchar)((dest & ~mask) | (S & mask));
    (void)S;
}

void REF2D_drawPixelR3_1_DSPnaon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(D | (S & ~P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSDaox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D ^ (P | (S & D)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPDSxaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S ^ (P & (D ^ S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSonon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(D | ~(P | S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_Dn(uchar *addr,int x,GA_color S,GA_color P)
{
    *addr ^= MONO_MASK(x);
    (void)S;
    (void)P;
}

void REF2D_drawPixelR3_1_DPSox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D ^ (P | S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSoan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(D & (P | S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSPoax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P ^ (D & (S | P)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSnox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D ^ (P | ~S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPx(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    P = (P ? mask : 0);
    S = (D ^ P);
    *addr = (uchar)((dest & ~mask) | (S & mask));
    (void)S;
}

void REF2D_drawPixelR3_1_DPSDonox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D ^ (P | ~(S | D)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSDxox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D ^ (P | (S ^ D)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSnoan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(D & (P | ~S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSDnaox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D ^ (P | (S & ~D)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    P = (P ? mask : 0);
    S = ~(D & P);
    *addr = (uchar)((dest & ~mask) | (S & mask));
    (void)S;
}

void REF2D_drawPixelR3_1_PDSxa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P & (D ^ S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DSPDSaoxxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(D ^ (S ^ (P | (D & S))));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DSPDoax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D ^ (S & (P | D)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPnox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S ^ (D | ~P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPSoax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S ^ (D & (P | S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DSPnox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D ^ (S | ~P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DSx(uchar *addr,int x,GA_color S,GA_color P)
{
    *addr ^= S ? MONO_MASK(x) : 0;
    (void)P;
}

void REF2D_drawPixelR3_1_SDPSonox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S ^ (D | ~(P | S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DSPDSonoxxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(D ^ (S ^ (P | ~(D | S))));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSxxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P ^ (D ^ S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D ^ (P & S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PSDPSoaxxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P ^ (S ^ (D & (P | S))));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S ^ (D & P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSPDoaxxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P ^ (D ^ (S & (P | D))));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPSnoax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S ^ (D & (P | ~S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSxnan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P & ~(D ^ S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSana(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P & ~(D & S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SSDxPDxaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S ^ ((S ^ D) & (P ^ D)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPSxox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S ^ (D | (P ^ S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPnoan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S & (D | ~P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DSPDxox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D ^ (S | (P ^ D)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

#endif
