/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser functions
*               for the SciTech SNAP Graphics Architecture.
*
****************************************************************************/

#include "ref2dint.h"

// NOTE: This module is broken up to get it to compile properly with
//       the IBM VisualAge compiler.

/*----------------------------- Implementation ----------------------------*/

#ifdef SUPPORT_4BPP

void REF2D_drawPixelR3_4_0(uchar *addr,int x,GA_color S,GA_color P)
{
    if (x & 1)
        *addr = (uchar)(*addr & 0xF0);
    else
        *addr = (uchar)(*addr & 0x0F);
    (void)S;
    (void)P;
}

void REF2D_drawPixelR3_4_DPSoon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(D | (P | S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DPSona(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (D & ~(P | S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PSon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    S = ~(P | S);
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SDPona(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (S & ~(D | P));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DPon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(D | P);
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
    (void)S;
}

void REF2D_drawPixelR3_4_PDSxnon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(P | ~(D ^ S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PDSaon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(P | (D & S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SDPnaa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (S & (D & ~P));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PDSxon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(P | (D ^ S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DPna(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (D & ~P);
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
    (void)S;
}

void REF2D_drawPixelR3_4_PSDnaon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(P | (S & ~D));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SPna(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    S = (S & ~P);
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PDSnaon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(P | (D & ~S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PDSonon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(P | ~(D | S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_Pn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    S = ~P;
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
    (void)S;
}

void REF2D_drawPixelR3_4_PDSona(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (P & ~(D | S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DSon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(D | S);
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
    (void)P;
}

void REF2D_drawPixelR3_4_SDPxnon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(S | ~(D ^ P));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SDPaon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(S | (D & P));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DPSxnon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(D | ~(P ^ S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DPSaon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(D | (P & S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PSDPSanaxx(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (P ^ (S ^ (D & ~(P & S))));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SSPxDSxaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(S ^ ((S ^ P) & (D ^ S)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SPxPDxa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ((S ^ P) & (P ^ D));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SDPSanaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(S ^ (D & ~(P & S)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PDSPaox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (P ^ (D | (S & P)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SDPSxaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(S ^ (D & (P ^ S)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PSDPaox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (P ^ (S | (D & P)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DSPDxaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(D ^ (S & (P ^ D)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PDSox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (P ^ (D | S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PDSoan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(P & (D | S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DPSnaa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (D & (P & ~S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SDPxon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(S | (D ^ P));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DSna(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (D & ~S);
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
    (void)P;
}

void REF2D_drawPixelR3_4_SPDnaon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(S | (P & ~D));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SPxDSxa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ((S ^ P) & (D ^ S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PDSPanaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(P ^ (D & ~(S & P)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SDPSaox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (S ^ (D | (P & S)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SDPSxnox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (S ^ (D | ~(P ^ S)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DPSxa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (D & (P ^ S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PSDPSaoxxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(P ^ (S ^ (D | (P & S))));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DPSana(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (D & ~(P & S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SSPxPDxaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(S ^ ((S ^ P) & (P ^ D)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SPDSoax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (S ^ (P & (D | S)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PSDnox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (P ^ (S | ~D));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PSDPxox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (P ^ (S | (D ^ P)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PSDnoan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(P & (S | ~D));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PSna(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    S = (P & ~S);
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SDPnaon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(S | (D & ~P));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SDPSoox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (S ^ (D | (P | S)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_Sn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    S = ~S;
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
    (void)P;
}

void REF2D_drawPixelR3_4_SPDSaox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (S ^ (P | (D & S)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SPDSxnox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (S ^ (P | ~(D ^ S)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SDPox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (S ^ (D | P));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SDPoan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(S & (D | P));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PSDPoax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (P ^ (S & (D | P)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SPDnox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (S ^ (P | ~D));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SPDSxox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (S ^ (P | (D ^ S)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SPDnoan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(S & (P | ~D));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PSx(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    S = (P ^ S);
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SPDSonox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (S ^ (P | ~(D | S)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SPDSnaox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (S ^ (P | (D & ~S)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PSan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    S = ~(P & S);
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PSDnaa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (P & (S & ~D));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DPSxon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(D | (P ^ S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SDxPDxa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ((S ^ D) & (P ^ D));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SPDSanaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(S ^ (P & ~(D & S)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SDna(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (S & ~D);
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
    (void)P;
}

void REF2D_drawPixelR3_4_DPSnaon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(D | (P & ~S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DSPDaox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (D ^ (S | (P & D)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PSDPxaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(P ^ (S & (D ^ P)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SDPxa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (S & (D ^ P));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PDSPDaoxxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(P ^ (D ^ (S | (P & D))));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DPSDoax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (D ^ (P & (S | D)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PDSnox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (P ^ (D | ~S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SDPana(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (S & ~(D & P));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SSPxDSxoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(S ^ ((S ^ P) | (D ^ S)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PDSPxox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (P ^ (D | (S ^ P)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PDSnoan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(P & (D | ~S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PDna(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (P & ~D);
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
    (void)S;
}

void REF2D_drawPixelR3_4_DSPnaon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(D | (S & ~P));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DPSDaox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (D ^ (P | (S & D)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SPDSxaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(S ^ (P & (D ^ S)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DPSonon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(D | ~(P | S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_Dn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~D;
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
    (void)S;
    (void)P;
}

void REF2D_drawPixelR3_4_DPSox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (D ^ (P | S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DPSoan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(D & (P | S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PDSPoax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (P ^ (D & (S | P)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DPSnox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (D ^ (P | ~S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DPx(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (D ^ P);
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
    (void)S;
}

void REF2D_drawPixelR3_4_DPSDonox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (D ^ (P | ~(S | D)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DPSDxox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (D ^ (P | (S ^ D)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DPSnoan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(D & (P | ~S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DPSDnaox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (D ^ (P | (S & ~D)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DPan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(D & P);
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
    (void)S;
}

void REF2D_drawPixelR3_4_PDSxa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (P & (D ^ S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DSPDSaoxxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(D ^ (S ^ (P | (D & S))));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DSPDoax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (D ^ (S & (P | D)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SDPnox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (S ^ (D | ~P));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SDPSoax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (S ^ (D & (P | S)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DSPnox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (D ^ (S | ~P));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DSx(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (D ^ S);
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
    (void)P;
}

void REF2D_drawPixelR3_4_SDPSonox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (S ^ (D | ~(P | S)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DSPDSonoxxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(D ^ (S ^ (P | ~(D | S))));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PDSxxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(P ^ (D ^ S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DPSax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (D ^ (P & S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PSDPSoaxxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(P ^ (S ^ (D & (P | S))));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SDPax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (S ^ (D & P));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PDSPDoaxxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(P ^ (D ^ (S & (P | D))));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SDPSnoax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (S ^ (D & (P | ~S)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PDSxnan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(P & ~(D ^ S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_PDSana(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (P & ~(D & S));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SSDxPDxaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(S ^ ((S ^ D) & (P ^ D)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SDPSxox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (S ^ (D | (P ^ S)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_SDPnoan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = ~(S & (D | ~P));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

void REF2D_drawPixelR3_4_DSPDxox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    S = (D ^ (S | (P ^ D)));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (S << 4));
}

#endif

