/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser functions
*               for the SciTech SNAP Graphics Architecture.
*
****************************************************************************/

#include "ref2dint.h"

// NOTE: This module is broken up to get it to compile properly with
//       the IBM VisualAge compiler.

/*----------------------------- Implementation ----------------------------*/

#ifdef SUPPORT_1BPP

void REF2D_drawPixelR3_1_DSPnoan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(D & (S | ~P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPSnaox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S ^ (D | (P & ~S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DSan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color D = *addr;
    GA_color mask = MONO_MASK(x);
    *addr = (uchar)((D & ~mask) | (~(D & (S ? mask : 0)) & mask));
    (void)P;
}

void REF2D_drawPixelR3_1_PDSax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P ^ (D & S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DSPDSoaxxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(D ^ (S ^ (P & (D | S))));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSDnoax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D ^ (P & (S | ~D)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPxnan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S & ~(D ^ P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPDSnoax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S ^ (P & (D | ~S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSxnan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(D & ~(P ^ S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPxDSxo(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ((S ^ P) | (D ^ S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSaan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(D & (P & S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSaa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D & (P & S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPxDSxon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~((S ^ P) | (D ^ S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSxna(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D & ~(P ^ S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPDSnoaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S ^ (P & (D | ~S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPxna(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S & ~(D ^ P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSPnoaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P ^ (D & (S | ~P)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DSPDSoaxx(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D ^ (S ^ (P & (D | S))));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P ^ (D & S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DSa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    *addr = *addr & (~mask | (S ? mask : 0));
    (void)P;
}

void REF2D_drawPixelR3_1_SDPSnaoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S ^ (D | (P & ~S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DSPnoa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D & (S | ~P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DSPDxoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(D ^ (S | (P ^ D)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPnoa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S & (D | ~P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPSxoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S ^ (D | (P ^ S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SSDxPDxax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S ^ ((S ^ D) & (P ^ D)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSanan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P & ~(D & S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSxna(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P & ~(D ^ S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPSnoaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S ^ (D & (P | ~S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSDPoaxx(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D ^ (P ^ (S & (D | P))));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPDaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S ^ (P & D));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PSDPSoaxx(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P ^ (S ^ (D & (P | S))));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(D ^ (P & S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSxx(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D ^ (P ^ S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PSDPSonoxx(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P ^ (S ^ (D | ~(P | S))));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPSonoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S ^ (D | ~(P | S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DSxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color D = *addr;
    GA_color mask = MONO_MASK(x);
    *addr = (uchar)((D & ~mask) | (~(D ^ (S ? mask : 0)) & mask));
    (void)P;
}

void REF2D_drawPixelR3_1_DPSnax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D ^ (P & ~S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPSoaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S ^ (D & (P | S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPDnax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S ^ (P & ~D));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DSPDoaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(D ^ (S & (P | D)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DSPDSaoxx(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D ^ (S ^ (P | (D & S))));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSxan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P & (D ^ S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    P = (P ? mask : 0);
    S = (D & P);
    *addr = (uchar)((dest & ~mask) | (S & mask));
    (void)S;
}

void REF2D_drawPixelR3_1_PDSPnaoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P ^ (D | (S & ~P)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSnoa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D & (P | ~S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSDxoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(D ^ (P | (S ^ D)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSPonoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P ^ (D | ~(S | P)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    P = (P ? mask : 0);
    S = ~(P ^ D);
    *addr = (uchar)((dest & ~mask) | (S & mask));
    (void)S;
}

void REF2D_drawPixelR3_1_DSPnax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D ^ (S & ~P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSPoaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P ^ (D & (S | P)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSoa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D & (P | S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(D ^ (P | S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_D(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = D;
    *addr = (uchar)((dest & ~mask) | (S & mask));
    (void)S;
    (void)P;
}

void REF2D_drawPixelR3_1_DPSono(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D | ~(P | S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPDSxax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S ^ (P & (D ^ S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSDaoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(D ^ (P | (S & D)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DSPnao(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D | (S & ~P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPno(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    P = (P ? mask : 0);
    S = (D | ~P);
    *addr = (uchar)((dest & ~mask) | (S & mask));
    (void)S;
}

void REF2D_drawPixelR3_1_PDSnoa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P & (D | ~S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSPxoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P ^ (D | (S ^ P)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SSPxDSxox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S ^ ((S ^ P) | (D ^ S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPanan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S & ~(D & P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PSDnax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P ^ (S & ~D));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSDoaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(D ^ (P & (S | D)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSDPaoxx(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D ^ (P ^ (S | (D & P))));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPxan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S & (D ^ P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PSDPxax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P ^ (S & (D ^ P)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DSPDaoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(D ^ (S | (P & D)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSnao(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D | (P & ~S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DSno(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color D = *addr;
    GA_color mask = MONO_MASK(x);
    *addr = (uchar)((D & ~mask) | ((D | ~(S ? mask : 0)) & mask));
    (void)P;
}

void REF2D_drawPixelR3_1_SPDSanax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S ^ (P & ~(D & S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDxPDxan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~((S ^ D) & (P ^ D));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSxo(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D | (P ^ S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSano(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D | ~(P & S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PSa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P & S);
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPDSnaoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S ^ (P | (D & ~S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPDSonoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S ^ (P | ~(D | S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PSxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P ^ S);
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPDnoa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S & (P | ~D));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPDSxoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S ^ (P | (D ^ S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPnax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S ^ (D & ~P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PSDPoaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P ^ (S & (D | P)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPoa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S & (D | P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPDoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S ^ (P | D));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSDxax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D ^ (P & (S ^ D)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPDSaoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S ^ (P | (D & S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_S(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    *addr = (*addr & ~mask) | (S ? mask : 0);
    (void)P;
}

void REF2D_drawPixelR3_1_SDPono(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S | ~(D | P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPnao(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S | (D & ~P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPno(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S | ~P);
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PSDnoa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P & (S | ~D));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PSDPxoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P ^ (S | (D ^ P)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSnax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P ^ (D & ~S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPDSoaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S ^ (P & (D | S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SSPxPDxax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S ^ ((S ^ P) & (P ^ D)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSanan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(D & ~(P & S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PSDPSaoxx(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P ^ (S ^ (D | (P & S))));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSxan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(D & (P ^ S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSPxax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P ^ (D & (S ^ P)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPSaoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(S ^ (D | (P & S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSDanax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D ^ (P & ~(S & D)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPxDSxan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~((S ^ P) & (D ^ S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPDnao(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S | (P & ~D));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDno(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color D = *addr;
    GA_color mask = MONO_MASK(x);
    *addr = (uchar)((D & ~mask) | ((~D | (S ? mask : 0)) & mask));
    (void)P;
}

void REF2D_drawPixelR3_1_SDPxo(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S | (D ^ P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPano(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S | ~(D & P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSoa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P & (D | S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P ^ (D | S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DSPDxax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D ^ (S & (P ^ D)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PSDPaoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P ^ (S | (D & P)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPSxax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S ^ (D & (P ^ S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSPaoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(P ^ (D | (S & P)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPSanax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S ^ (D & ~(P & S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SPxPDxan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~((S ^ P) & (P ^ D));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SSPxDSxax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S ^ ((S ^ P) & (D ^ S)));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DSPDSanaxxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = ~(D ^ (S ^ (P & ~(D & S))));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSao(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D | (P & S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSxno(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D | ~(P ^ S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPao(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S | (D & P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_SDPxno(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S | ~(D ^ P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DSo(uchar *addr,int x,GA_color S,GA_color P)
{
    *addr |= S ? MONO_MASK(x) : 0;
    (void)P;
}

void REF2D_drawPixelR3_1_SDPnoo(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (S | (D | ~P));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_P(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    P = (P ? mask : 0);
    S = P;
    *addr = (uchar)((dest & ~mask) | (S & mask));
    (void)S;
}

void REF2D_drawPixelR3_1_PDSono(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P | ~(D | S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSnao(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P | (D & ~S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PSno(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P | ~S);
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PSDnao(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P | (S & ~D));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDno(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    P = (P ? mask : 0);
    S = (P | ~D);
    *addr = (uchar)((dest & ~mask) | (S & mask));
    (void)S;
}

void REF2D_drawPixelR3_1_PDSxo(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P | (D ^ S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSano(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P | ~(D & S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSao(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P | (D & S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PDSxno(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P | ~(D ^ S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPo(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    P = (P ? mask : 0);
    S = (D | P);
    *addr = (uchar)((dest & ~mask) | (S & mask));
    (void)S;
}

void REF2D_drawPixelR3_1_DPSnoo(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D | (P | ~S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PSo(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P | S);
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_PSDnoo(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (P | (S | ~D));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_DPSoo(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color mask = MONO_MASK(x);
    GA_color dest = *addr;
    GA_color D = (dest & mask);
    S = (S ? mask : 0);
    P = (P ? mask : 0);
    S = (D | (P | S));
    *addr = (uchar)((dest & ~mask) | (S & mask));
}

void REF2D_drawPixelR3_1_1(uchar *addr,int x,GA_color S,GA_color P)
{
    *addr |= MONO_MASK(x);
    (void)S;
    (void)P;
}

#endif
