/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser functions
*               for the SciTech SNAP Graphics Architecture.
*
****************************************************************************/

#include "ref2dint.h"

// NOTE: This module is broken up to get it to compile properly with
//       the IBM VisualAge compiler.

/*----------------------------- Implementation ----------------------------*/

#define BUF_INCR    3
#define BUF_INCR2   3
#define COLOR_TYPE  uchar

/****************************************************************************
REMARKS:
ROP3 masked pixel functions.
****************************************************************************/

void REF2D_drawPixelR3Masked_24_0(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = (D & DC.planeMaskInv) | 0;
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)S;
    (void)P;
}

void REF2D_drawPixelR3Masked_24_DPSoon(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D | (P | S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSona(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D & ~(P | S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PSon(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P | S) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPona(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S & ~(D | P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPon(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D | P) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)S;
}

void REF2D_drawPixelR3Masked_24_PDSxnon(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P | ~(D ^ S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSaon(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P | (D & S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPnaa(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S & (D & ~P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSxon(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P | (D ^ S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPna(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D & ~P) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)S;
}

void REF2D_drawPixelR3Masked_24_PSDnaon(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P | (S & ~D)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPna(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S & ~P) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSnaon(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P | (D & ~S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSonon(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P | ~(D | S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_Pn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~P & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)S;
}

void REF2D_drawPixelR3Masked_24_PDSona(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P & ~(D | S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DSon(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D | S) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)P;
}

void REF2D_drawPixelR3Masked_24_SDPxnon(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S | ~(D ^ P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPaon(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S | (D & P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSxnon(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D | ~(P ^ S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSaon(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D | (P & S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PSDPSanaxx(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P ^ (S ^ (D & ~(P & S)))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SSPxDSxaxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S ^ ((S ^ P) & (D ^ S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPxPDxa(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (((S ^ P) & (P ^ D)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPSanaxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S ^ (D & ~(P & S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSPaox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P ^ (D | (S & P))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPSxaxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S ^ (D & (P ^ S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PSDPaox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P ^ (S | (D & P))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DSPDxaxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D ^ (S & (P ^ D))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P ^ (D | S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSoan(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P & (D | S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSnaa(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D & (P & ~S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPxon(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S | (D ^ P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DSna(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D & ~S) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)P;
}

void REF2D_drawPixelR3Masked_24_SPDnaon(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S | (P & ~D)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPxDSxa(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (((S ^ P) & (D ^ S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSPanaxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P ^ (D & ~(S & P))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPSaox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S ^ (D | (P & S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPSxnox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S ^ (D | ~(P ^ S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSxa(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D & (P ^ S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PSDPSaoxxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P ^ (S ^ (D | (P & S)))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSana(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D & ~(P & S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SSPxPDxaxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S ^ ((S ^ P) & (P ^ D))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPDSoax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S ^ (P & (D | S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PSDnox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P ^ (S | ~D)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PSDPxox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P ^ (S | (D ^ P))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PSDnoan(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P & (S | ~D)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PSna(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P & ~S) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPnaon(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S | (D & ~P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPSoox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S ^ (D | (P | S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_Sn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~S & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)P;
}

void REF2D_drawPixelR3Masked_24_SPDSaox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S ^ (P | (D & S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPDSxnox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S ^ (P | ~(D ^ S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S ^ (D | P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPoan(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S & (D | P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PSDPoax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P ^ (S & (D | P))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPDnox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S ^ (P | ~D)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPDSxox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S ^ (P | (D ^ S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPDnoan(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S & (P | ~D)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PSx(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P ^ S) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPDSonox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S ^ (P | ~(D | S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPDSnaox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S ^ (P | (D & ~S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PSan(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P & S) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PSDnaa(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P & (S & ~D)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSxon(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D | (P ^ S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDxPDxa(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (((S ^ D) & (P ^ D)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPDSanaxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S ^ (P & ~(D & S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDna(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S & ~D) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)P;
}

void REF2D_drawPixelR3Masked_24_DPSnaon(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D | (P & ~S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DSPDaox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D ^ (S | (P & D))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PSDPxaxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P ^ (S & (D ^ P))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPxa(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S & (D ^ P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSPDaoxxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P ^ (D ^ (S | (P & D)))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSDoax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D ^ (P & (S | D))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSnox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P ^ (D | ~S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPana(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S & ~(D & P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SSPxDSxoxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S ^ ((S ^ P) | (D ^ S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSPxox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P ^ (D | (S ^ P))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSnoan(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P & (D | ~S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDna(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P & ~D) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)S;
}

void REF2D_drawPixelR3Masked_24_DSPnaon(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D | (S & ~P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSDaox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D ^ (P | (S & D))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPDSxaxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S ^ (P & (D ^ S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSonon(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D | ~(P | S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_Dn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~D & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)S;
    (void)P;
}

void REF2D_drawPixelR3Masked_24_DPSox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D ^ (P | S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSoan(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D & (P | S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSPoax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P ^ (D & (S | P))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSnox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D ^ (P | ~S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPx(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D ^ P) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)S;
}

void REF2D_drawPixelR3Masked_24_DPSDonox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D ^ (P | ~(S | D))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSDxox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D ^ (P | (S ^ D))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSnoan(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D & (P | ~S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSDnaox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D ^ (P | (S & ~D))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPan(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D & P) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)S;
}

void REF2D_drawPixelR3Masked_24_PDSxa(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P & (D ^ S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DSPDSaoxxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D ^ (S ^ (P | (D & S)))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DSPDoax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D ^ (S & (P | D))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPnox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S ^ (D | ~P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPSoax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S ^ (D & (P | S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DSPnox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D ^ (S | ~P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DSx(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D ^ S) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)P;
}

void REF2D_drawPixelR3Masked_24_SDPSonox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S ^ (D | ~(P | S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DSPDSonoxxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D ^ (S ^ (P | ~(D | S)))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSxxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P ^ (D ^ S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D ^ (P & S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PSDPSoaxxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P ^ (S ^ (D & (P | S)))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S ^ (D & P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSPDoaxxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P ^ (D ^ (S & (P | D)))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPSnoax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S ^ (D & (P | ~S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSxnan(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P & ~(D ^ S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSana(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P & ~(D & S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SSDxPDxaxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S ^ ((S ^ D) & (P ^ D))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPSxox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S ^ (D | (P ^ S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPnoan(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S & (D | ~P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DSPDxox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D ^ (S | (P ^ D))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}
