/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser functions
*               for the SciTech SNAP Graphics Architecture.
*
****************************************************************************/

#include "ref2dint.h"

// NOTE: This module is broken up to get it to compile properly with
//       the IBM VisualAge compiler.

/*----------------------------- Implementation ----------------------------*/

#ifdef SUPPORT_4BPP

void REF2D_drawPixelR3Masked_4_0(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | 0);
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)S;
    (void)P;
}

void REF2D_drawPixelR3Masked_4_DPSoon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D | (P | S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSona(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D & ~(P | S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PSon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P | S) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPona(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S & ~(D | P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D | P) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)S;
}

void REF2D_drawPixelR3Masked_4_PDSxnon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P | ~(D ^ S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSaon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P | (D & S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPnaa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S & (D & ~P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSxon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P | (D ^ S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPna(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D & ~P) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)S;
}

void REF2D_drawPixelR3Masked_4_PSDnaon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P | (S & ~D)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPna(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S & ~P) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSnaon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P | (D & ~S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSonon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P | ~(D | S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_Pn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~P & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)S;
}

void REF2D_drawPixelR3Masked_4_PDSona(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P & ~(D | S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DSon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D | S) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)P;
}

void REF2D_drawPixelR3Masked_4_SDPxnon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S | ~(D ^ P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPaon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S | (D & P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSxnon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D | ~(P ^ S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSaon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D | (P & S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PSDPSanaxx(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P ^ (S ^ (D & ~(P & S)))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SSPxDSxaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S ^ ((S ^ P) & (D ^ S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPxPDxa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (((S ^ P) & (P ^ D)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPSanaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S ^ (D & ~(P & S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSPaox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P ^ (D | (S & P))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPSxaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S ^ (D & (P ^ S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PSDPaox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P ^ (S | (D & P))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DSPDxaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D ^ (S & (P ^ D))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P ^ (D | S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSoan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P & (D | S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSnaa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D & (P & ~S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPxon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S | (D ^ P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DSna(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D & ~S) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)P;
}

void REF2D_drawPixelR3Masked_4_SPDnaon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S | (P & ~D)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPxDSxa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (((S ^ P) & (D ^ S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSPanaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P ^ (D & ~(S & P))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPSaox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S ^ (D | (P & S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPSxnox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S ^ (D | ~(P ^ S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSxa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D & (P ^ S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PSDPSaoxxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P ^ (S ^ (D | (P & S)))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSana(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D & ~(P & S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SSPxPDxaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S ^ ((S ^ P) & (P ^ D))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPDSoax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S ^ (P & (D | S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PSDnox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P ^ (S | ~D)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PSDPxox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P ^ (S | (D ^ P))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PSDnoan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P & (S | ~D)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PSna(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P & ~S) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPnaon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S | (D & ~P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPSoox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S ^ (D | (P | S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_Sn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~S & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)P;
}

void REF2D_drawPixelR3Masked_4_SPDSaox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S ^ (P | (D & S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPDSxnox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S ^ (P | ~(D ^ S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S ^ (D | P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPoan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S & (D | P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PSDPoax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P ^ (S & (D | P))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPDnox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S ^ (P | ~D)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPDSxox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S ^ (P | (D ^ S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPDnoan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S & (P | ~D)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PSx(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P ^ S) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPDSonox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S ^ (P | ~(D | S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPDSnaox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S ^ (P | (D & ~S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PSan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P & S) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PSDnaa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P & (S & ~D)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSxon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D | (P ^ S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDxPDxa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (((S ^ D) & (P ^ D)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPDSanaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S ^ (P & ~(D & S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDna(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S & ~D) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)P;
}

void REF2D_drawPixelR3Masked_4_DPSnaon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D | (P & ~S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DSPDaox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D ^ (S | (P & D))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PSDPxaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P ^ (S & (D ^ P))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPxa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S & (D ^ P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSPDaoxxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P ^ (D ^ (S | (P & D)))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSDoax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D ^ (P & (S | D))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSnox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P ^ (D | ~S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPana(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S & ~(D & P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SSPxDSxoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S ^ ((S ^ P) | (D ^ S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSPxox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P ^ (D | (S ^ P))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSnoan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P & (D | ~S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDna(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P & ~D) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)S;
}

void REF2D_drawPixelR3Masked_4_DSPnaon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D | (S & ~P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSDaox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D ^ (P | (S & D))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPDSxaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S ^ (P & (D ^ S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSonon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D | ~(P | S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_Dn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~D & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)S;
    (void)P;
}

void REF2D_drawPixelR3Masked_4_DPSox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D ^ (P | S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSoan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D & (P | S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSPoax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P ^ (D & (S | P))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSnox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D ^ (P | ~S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPx(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D ^ P) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)S;
}

void REF2D_drawPixelR3Masked_4_DPSDonox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D ^ (P | ~(S | D))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSDxox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D ^ (P | (S ^ D))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSnoan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D & (P | ~S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSDnaox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D ^ (P | (S & ~D))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D & P) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)S;
}

void REF2D_drawPixelR3Masked_4_PDSxa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P & (D ^ S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DSPDSaoxxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D ^ (S ^ (P | (D & S)))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DSPDoax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D ^ (S & (P | D))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPnox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S ^ (D | ~P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPSoax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S ^ (D & (P | S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DSPnox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D ^ (S | ~P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DSx(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D ^ S) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)P;
}

void REF2D_drawPixelR3Masked_4_SDPSonox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S ^ (D | ~(P | S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DSPDSonoxxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D ^ (S ^ (P | ~(D | S)))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSxxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P ^ (D ^ S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D ^ (P & S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PSDPSoaxxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P ^ (S ^ (D & (P | S)))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S ^ (D & P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSPDoaxxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P ^ (D ^ (S & (P | D)))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPSnoax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S ^ (D & (P | ~S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSxnan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P & ~(D ^ S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSana(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P & ~(D & S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SSDxPDxaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S ^ ((S ^ D) & (P ^ D))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPSxox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S ^ (D | (P ^ S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPnoan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S & (D | ~P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DSPDxox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D ^ (S | (P ^ D))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

#endif

