/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser functions
*               for the SciTech SNAP Graphics Architecture.
*
****************************************************************************/

#include "ref2dint.h"

// NOTE: This module is broken up to get it to compile properly with
//       the IBM VisualAge compiler.

/*----------------------------- Implementation ----------------------------*/

#define BUF_INCR    3
#define BUF_INCR2   3
#define COLOR_TYPE  uchar

/****************************************************************************
REMARKS:
ROP3 masked pixel functions.
****************************************************************************/

void REF2D_drawPixelR3Masked_24_DSPnoan(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D & (S | ~P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPSnaox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S ^ (D | (P & ~S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DSan(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D & S) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)P;
}

void REF2D_drawPixelR3Masked_24_PDSax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P ^ (D & S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DSPDSoaxxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D ^ (S ^ (P & (D | S)))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSDnoax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D ^ (P & (S | ~D))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPxnan(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S & ~(D ^ P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPDSnoax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S ^ (P & (D | ~S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSxnan(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D & ~(P ^ S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPxDSxo(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (((S ^ P) | (D ^ S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSaan(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D & (P & S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSaa(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D & (P & S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPxDSxon(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~((S ^ P) | (D ^ S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSxna(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D & ~(P ^ S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPDSnoaxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S ^ (P & (D | ~S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPxna(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S & ~(D ^ P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSPnoaxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P ^ (D & (S | ~P))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DSPDSoaxx(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D ^ (S ^ (P & (D | S)))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSaxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P ^ (D & S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DSa(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D & S) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)P;
}

void REF2D_drawPixelR3Masked_24_SDPSnaoxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S ^ (D | (P & ~S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DSPnoa(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D & (S | ~P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DSPDxoxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D ^ (S | (P ^ D))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPnoa(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S & (D | ~P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPSxoxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S ^ (D | (P ^ S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SSDxPDxax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S ^ ((S ^ D) & (P ^ D))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSanan(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P & ~(D & S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSxna(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P & ~(D ^ S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPSnoaxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S ^ (D & (P | ~S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSDPoaxx(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D ^ (P ^ (S & (D | P)))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPDaxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S ^ (P & D)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PSDPSoaxx(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P ^ (S ^ (D & (P | S)))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSaxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D ^ (P & S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSxx(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D ^ (P ^ S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PSDPSonoxx(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P ^ (S ^ (D | ~(P | S)))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPSonoxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S ^ (D | ~(P | S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DSxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D ^ S) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)P;
}

void REF2D_drawPixelR3Masked_24_DPSnax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D ^ (P & ~S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPSoaxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S ^ (D & (P | S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPDnax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S ^ (P & ~D)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DSPDoaxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D ^ (S & (P | D))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DSPDSaoxx(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D ^ (S ^ (P | (D & S)))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSxan(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P & (D ^ S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPa(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D & P) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)S;
}

void REF2D_drawPixelR3Masked_24_PDSPnaoxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P ^ (D | (S & ~P))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSnoa(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D & (P | ~S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSDxoxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D ^ (P | (S ^ D))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSPonoxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P ^ (D | ~(S | P))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P ^ D) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)S;
}

void REF2D_drawPixelR3Masked_24_DSPnax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D ^ (S & ~P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSPoaxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P ^ (D & (S | P))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSoa(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D & (P | S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSoxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D ^ (P | S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_D(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (D & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)S;
    (void)P;
}

void REF2D_drawPixelR3Masked_24_DPSono(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D | ~(P | S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPDSxax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S ^ (P & (D ^ S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSDaoxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D ^ (P | (S & D))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DSPnao(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D | (S & ~P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPno(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D | ~P) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)S;
}

void REF2D_drawPixelR3Masked_24_PDSnoa(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P & (D | ~S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSPxoxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P ^ (D | (S ^ P))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SSPxDSxox(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S ^ ((S ^ P) | (D ^ S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPanan(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S & ~(D & P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PSDnax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P ^ (S & ~D)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSDoaxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D ^ (P & (S | D))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSDPaoxx(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D ^ (P ^ (S | (D & P)))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPxan(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S & (D ^ P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PSDPxax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P ^ (S & (D ^ P))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DSPDaoxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D ^ (S | (P & D))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSnao(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D | (P & ~S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DSno(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D | ~S) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)P;
}

void REF2D_drawPixelR3Masked_24_SPDSanax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S ^ (P & ~(D & S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDxPDxan(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~((S ^ D) & (P ^ D)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSxo(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D | (P ^ S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSano(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D | ~(P & S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PSa(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P & S) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPDSnaoxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S ^ (P | (D & ~S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPDSonoxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S ^ (P | ~(D | S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PSxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P ^ S) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPDnoa(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S & (P | ~D)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPDSxoxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S ^ (P | (D ^ S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPnax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S ^ (D & ~P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PSDPoaxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P ^ (S & (D | P))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPoa(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S & (D | P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPDoxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S ^ (P | D)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSDxax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D ^ (P & (S ^ D))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPDSaoxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S ^ (P | (D & S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_S(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (S & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)P;
}

void REF2D_drawPixelR3Masked_24_SDPono(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S | ~(D | P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPnao(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S | (D & ~P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPno(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S | ~P) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PSDnoa(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P & (S | ~D)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PSDPxoxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P ^ (S | (D ^ P))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSnax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P ^ (D & ~S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPDSoaxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S ^ (P & (D | S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SSPxPDxax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S ^ ((S ^ P) & (P ^ D))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSanan(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D & ~(P & S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PSDPSaoxx(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P ^ (S ^ (D | (P & S)))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSxan(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D & (P ^ S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSPxax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P ^ (D & (S ^ P))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPSaoxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(S ^ (D | (P & S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSDanax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D ^ (P & ~(S & D))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPxDSxan(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~((S ^ P) & (D ^ S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPDnao(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S | (P & ~D)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDno(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S | ~D) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)P;
}

void REF2D_drawPixelR3Masked_24_SDPxo(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S | (D ^ P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPano(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S | ~(D & P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSoa(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P & (D | S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSoxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P ^ (D | S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DSPDxax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D ^ (S & (P ^ D))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PSDPaoxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P ^ (S | (D & P))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPSxax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S ^ (D & (P ^ S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSPaoxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(P ^ (D | (S & P))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPSanax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S ^ (D & ~(P & S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SPxPDxan(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~((S ^ P) & (P ^ D)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SSPxDSxax(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S ^ ((S ^ P) & (D ^ S))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DSPDSanaxxn(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (~(D ^ (S ^ (P & ~(D & S)))) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSao(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D | (P & S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSxno(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D | ~(P ^ S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPao(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S | (D & P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_SDPxno(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S | ~(D ^ P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DSo(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D | S) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)P;
}

void REF2D_drawPixelR3Masked_24_SDPnoo(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((S | (D | ~P)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_P(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | (P & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)S;
}

void REF2D_drawPixelR3Masked_24_PDSono(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P | ~(D | S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSnao(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P | (D & ~S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PSno(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P | ~S) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PSDnao(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P | (S & ~D)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDno(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P | ~D) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)S;
}

void REF2D_drawPixelR3Masked_24_PDSxo(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P | (D ^ S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSano(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P | ~(D & S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSao(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P | (D & S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PDSxno(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P | ~(D ^ S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPo(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D | P) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)S;
}

void REF2D_drawPixelR3Masked_24_DPSnoo(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D | (P | ~S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PSo(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P | S) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_PSDnoo(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((P | (S | ~D)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_DPSoo(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = ((D & DC.planeMaskInv) | ((D | (P | S)) & DC.planeMask));
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelR3Masked_24_1(uchar *addr DCL_X_ARG,GA_color S,GA_color P)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = (D & DC.planeMaskInv) | DC.planeMask;
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)S;
    (void)P;
}
