/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser functions
*               for the SciTech SNAP Graphics Architecture.
*
****************************************************************************/

#include "ref2dint.h"

// NOTE: This module is broken up to get it to compile properly with
//       the IBM VisualAge compiler.

/*----------------------------- Implementation ----------------------------*/

#ifdef SUPPORT_4BPP

void REF2D_drawPixelR3Masked_4_DSPnoan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D & (S | ~P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPSnaox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S ^ (D | (P & ~S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DSan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D & S) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)P;
}

void REF2D_drawPixelR3Masked_4_PDSax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P ^ (D & S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DSPDSoaxxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D ^ (S ^ (P & (D | S)))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSDnoax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D ^ (P & (S | ~D))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPxnan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S & ~(D ^ P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPDSnoax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S ^ (P & (D | ~S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSxnan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D & ~(P ^ S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPxDSxo(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (((S ^ P) | (D ^ S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSaan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D & (P & S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSaa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D & (P & S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPxDSxon(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~((S ^ P) | (D ^ S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSxna(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D & ~(P ^ S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPDSnoaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S ^ (P & (D | ~S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPxna(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S & ~(D ^ P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSPnoaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P ^ (D & (S | ~P))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DSPDSoaxx(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D ^ (S ^ (P & (D | S)))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P ^ (D & S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DSa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D & S) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)P;
}

void REF2D_drawPixelR3Masked_4_SDPSnaoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S ^ (D | (P & ~S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DSPnoa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D & (S | ~P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DSPDxoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D ^ (S | (P ^ D))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPnoa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S & (D | ~P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPSxoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S ^ (D | (P ^ S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SSDxPDxax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S ^ ((S ^ D) & (P ^ D))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSanan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P & ~(D & S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSxna(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P & ~(D ^ S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPSnoaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S ^ (D & (P | ~S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSDPoaxx(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D ^ (P ^ (S & (D | P)))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPDaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S ^ (P & D)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PSDPSoaxx(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P ^ (S ^ (D & (P | S)))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D ^ (P & S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSxx(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D ^ (P ^ S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PSDPSonoxx(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P ^ (S ^ (D | ~(P | S)))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPSonoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S ^ (D | ~(P | S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DSxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D ^ S) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)P;
}

void REF2D_drawPixelR3Masked_4_DPSnax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D ^ (P & ~S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPSoaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S ^ (D & (P | S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPDnax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S ^ (P & ~D)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DSPDoaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D ^ (S & (P | D))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DSPDSaoxx(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D ^ (S ^ (P | (D & S)))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSxan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P & (D ^ S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D & P) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)S;
}

void REF2D_drawPixelR3Masked_4_PDSPnaoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P ^ (D | (S & ~P))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSnoa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D & (P | ~S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSDxoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D ^ (P | (S ^ D))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSPonoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P ^ (D | ~(S | P))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P ^ D) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)S;
}

void REF2D_drawPixelR3Masked_4_DSPnax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D ^ (S & ~P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSPoaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P ^ (D & (S | P))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSoa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D & (P | S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D ^ (P | S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_D(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (D & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)S;
    (void)P;
}

void REF2D_drawPixelR3Masked_4_DPSono(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D | ~(P | S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPDSxax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S ^ (P & (D ^ S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSDaoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D ^ (P | (S & D))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DSPnao(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D | (S & ~P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPno(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D | ~P) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)S;
}

void REF2D_drawPixelR3Masked_4_PDSnoa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P & (D | ~S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSPxoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P ^ (D | (S ^ P))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SSPxDSxox(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S ^ ((S ^ P) | (D ^ S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPanan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S & ~(D & P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PSDnax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P ^ (S & ~D)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSDoaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D ^ (P & (S | D))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSDPaoxx(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D ^ (P ^ (S | (D & P)))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPxan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S & (D ^ P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PSDPxax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P ^ (S & (D ^ P))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DSPDaoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D ^ (S | (P & D))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSnao(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D | (P & ~S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DSno(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D | ~S) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)P;
}

void REF2D_drawPixelR3Masked_4_SPDSanax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S ^ (P & ~(D & S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDxPDxan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~((S ^ D) & (P ^ D)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSxo(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D | (P ^ S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSano(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D | ~(P & S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PSa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P & S) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPDSnaoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S ^ (P | (D & ~S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPDSonoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S ^ (P | ~(D | S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PSxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P ^ S) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPDnoa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S & (P | ~D)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPDSxoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S ^ (P | (D ^ S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPnax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S ^ (D & ~P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PSDPoaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P ^ (S & (D | P))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPoa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S & (D | P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPDoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S ^ (P | D)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSDxax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D ^ (P & (S ^ D))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPDSaoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S ^ (P | (D & S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_S(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (S & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)P;
}

void REF2D_drawPixelR3Masked_4_SDPono(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S | ~(D | P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPnao(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S | (D & ~P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPno(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S | ~P) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PSDnoa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P & (S | ~D)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PSDPxoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P ^ (S | (D ^ P))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSnax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P ^ (D & ~S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPDSoaxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S ^ (P & (D | S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SSPxPDxax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S ^ ((S ^ P) & (P ^ D))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSanan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D & ~(P & S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PSDPSaoxx(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P ^ (S ^ (D | (P & S)))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSxan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D & (P ^ S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSPxax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P ^ (D & (S ^ P))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPSaoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(S ^ (D | (P & S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSDanax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D ^ (P & ~(S & D))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPxDSxan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~((S ^ P) & (D ^ S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPDnao(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S | (P & ~D)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDno(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S | ~D) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)P;
}

void REF2D_drawPixelR3Masked_4_SDPxo(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S | (D ^ P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPano(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S | ~(D & P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSoa(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P & (D | S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P ^ (D | S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DSPDxax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D ^ (S & (P ^ D))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PSDPaoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P ^ (S | (D & P))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPSxax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S ^ (D & (P ^ S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSPaoxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(P ^ (D | (S & P))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPSanax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S ^ (D & ~(P & S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SPxPDxan(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~((S ^ P) & (P ^ D)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SSPxDSxax(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S ^ ((S ^ P) & (D ^ S))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DSPDSanaxxn(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D ^ (S ^ (P & ~(D & S)))) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSao(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D | (P & S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSxno(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D | ~(P ^ S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPao(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S | (D & P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_SDPxno(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S | ~(D ^ P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DSo(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D | S) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)P;
}

void REF2D_drawPixelR3Masked_4_SDPnoo(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S | (D | ~P)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_P(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (P & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)S;
}

void REF2D_drawPixelR3Masked_4_PDSono(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P | ~(D | S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSnao(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P | (D & ~S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PSno(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P | ~S) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PSDnao(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P | (S & ~D)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDno(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P | ~D) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)S;
}

void REF2D_drawPixelR3Masked_4_PDSxo(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P | (D ^ S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSano(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P | ~(D & S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSao(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P | (D & S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PDSxno(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P | ~(D ^ S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPo(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D | P) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)S;
}

void REF2D_drawPixelR3Masked_4_DPSnoo(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D | (P | ~S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PSo(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P | S) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_PSDnoo(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((P | (S | ~D)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_DPSoo(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D | (P | S)) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelR3Masked_4_1(uchar *addr,int x,GA_color S,GA_color P)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | DC.planeMask);
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)S;
    (void)P;
}

#endif

