/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser functions
*               for the SciTech SNAP Graphics Architecture.
*
****************************************************************************/

#include "ref2dint.h"

/*----------------------------- Implementation ----------------------------*/

#ifdef SUPPORT_1BPP

/****************************************************************************
REMARKS:
Pixel reading functions
****************************************************************************/

GA_color REF2D_readPixel_1(uchar *addr,int x)
{
    return *addr >> MONO_SHIFT(x);
}

/****************************************************************************
REMARKS:
Source Transparent color keyed pixel functions
****************************************************************************/

void REF2D_drawPixelSrcColorKey_1(
    uchar *addr DCL_X_ARG,
    GA_color S)
{
    // TODO
    GA_color mask = MONO_MASK(x);
    *addr = (*addr & ~mask) | (S ? mask : 0);
}

void REF2D_drawPixelSrcColorKeyRange_1(
    uchar *addr DCL_X_ARG,
    GA_color S)
{
    // TODO
    GA_color mask = MONO_MASK(x);
    *addr = (*addr & ~mask) | (S ? mask : 0);
}

/****************************************************************************
REMARKS:
ROP2 pixel functions.
****************************************************************************/

void REF2D_drawPixel_1_0(uchar *addr,int x,GA_color S)
{
    *addr &= ~MONO_MASK(x);
}

void REF2D_drawPixel_1_DSon(uchar *addr,int x,GA_color S)
{
    GA_color D = *addr;
    GA_color mask = MONO_MASK(x);
    *addr = (uchar)((D & ~mask) | ((~D | (S ? mask : 0)) & mask));
}

void REF2D_drawPixel_1_DSna(uchar *addr,int x,GA_color S)
{
    GA_color D = *addr;
    GA_color mask = MONO_MASK(x);
    *addr = (uchar)((D & ~mask) | ((D & ~(S ? mask : 0)) & mask));
}

void REF2D_drawPixel_1_Sn(uchar *addr,int x,GA_color S)
{
    GA_color D = *addr;
    GA_color mask = MONO_MASK(x);
    *addr = (uchar)((D & ~mask) | (~(S ? mask : 0) & mask));
}

void REF2D_drawPixel_1_SDna(uchar *addr,int x,GA_color S)
{
    GA_color D = *addr;
    GA_color mask = MONO_MASK(x);
    *addr = (uchar)((D & ~mask) | ((~D & (S ? mask : 0)) & mask));
}

void REF2D_drawPixel_1_Dn(uchar *addr,int x,GA_color S)
{
    *addr ^= MONO_MASK(x);
    (void)S;
}

void REF2D_drawPixel_1_DSx(uchar *addr,int x,GA_color S)
{
    *addr ^= S ? MONO_MASK(x) : 0;
}

void REF2D_drawPixel_1_DSan(uchar *addr,int x,GA_color S)
{
    GA_color D = *addr;
    GA_color mask = MONO_MASK(x);
    *addr = (uchar)((D & ~mask) | (~(D & (S ? mask : 0)) & mask));
}

void REF2D_drawPixel_1_DSa(uchar *addr,int x,GA_color S)
{
    GA_color mask = MONO_MASK(x);
    *addr = *addr & (~mask | (S ? mask : 0));
}

void REF2D_drawPixel_1_DSxn(uchar *addr,int x,GA_color S)
{
    GA_color D = *addr;
    GA_color mask = MONO_MASK(x);
    *addr = (uchar)((D & ~mask) | (~(D ^ (S ? mask : 0)) & mask));
}

void REF2D_drawPixel_1_DSno(uchar *addr,int x,GA_color S)
{
    GA_color D = *addr;
    GA_color mask = MONO_MASK(x);
    *addr = (uchar)((D & ~mask) | ((D | ~(S ? mask : 0)) & mask));
}

void REF2D_drawPixel_1_S(uchar *addr,int x,GA_color S)
{
    GA_color mask = MONO_MASK(x);
    *addr = (*addr & ~mask) | (S ? mask : 0);
}

void REF2D_drawPixel_1_SDno(uchar *addr,int x,GA_color S)
{
    GA_color D = *addr;
    GA_color mask = MONO_MASK(x);
    *addr = (uchar)((D & ~mask) | ((~D | (S ? mask : 0)) & mask));
}

void REF2D_drawPixel_1_DSo(uchar *addr,int x,GA_color S)
{
    *addr |= S ? MONO_MASK(x) : 0;
}

void REF2D_drawPixel_1_1(uchar *addr,int x,GA_color S)
{
    *addr |= MONO_MASK(x);
    (void)S;
}

#endif
