/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser functions
*               for the SciTech SNAP Graphics Architecture.
*
****************************************************************************/

#include "ref2dint.h"

/*----------------------------- Implementation ----------------------------*/

#define BUF_INCR    2
#define COLOR_TYPE  ushort
#define COLOR_WHITE 0xFFFF
#define PATT_TYPE   b16

/****************************************************************************
REMARKS:
Pixel reading functions
****************************************************************************/

GA_color REF2D_readPixel_16(
    uchar *addr DCL_X_ARG)
{
    return *((COLOR_TYPE*)addr);
}

/****************************************************************************
REMARKS:
Pixel color conversion functions
****************************************************************************/

GA_color REF2D_convertPixel16_15(
    uchar *addr DCL_X_ARG)
{
    GA_color c = *((ushort*)addr);
    return ((c & 0xFFC0) >> 1) | (c & 0x1F);
}

GA_color REF2D_convertPixel15_16(
    uchar *addr DCL_X_ARG)
{
    GA_color c = *((ushort*)addr);
    return ((c & 0x7FE0) << 1) | (c & 0x1F);
}

GA_color REF2D_convertPixel24RGB_15(
    uchar *addr DCL_X_ARG)
{
    return PACK_COLOR_15(addr[2],addr[1],addr[0]);
}

GA_color REF2D_convertPixel24BGR_15(
    uchar *addr DCL_X_ARG)
{
    return PACK_COLOR_15(addr[0],addr[1],addr[2]);
}

GA_color REF2D_convertPixel24RGB_16(
    uchar *addr DCL_X_ARG)
{
    return PACK_COLOR_16(addr[2],addr[1],addr[0]);
}

GA_color REF2D_convertPixel24BGR_16(
    uchar *addr DCL_X_ARG)
{
    return PACK_COLOR_16(addr[0],addr[1],addr[2]);
}

GA_color REF2D_convertPixel32RGBA_15(
    uchar *addr DCL_X_ARG)
{
    return PACK_COLOR_15(addr[3],addr[2],addr[1]);
}

GA_color REF2D_convertPixel32BGRA_15(
    uchar *addr DCL_X_ARG)
{
    return PACK_COLOR_15(addr[1],addr[2],addr[3]);
}

GA_color REF2D_convertPixel32RGBA_16(
    uchar *addr DCL_X_ARG)
{
    return PACK_COLOR_16(addr[3],addr[2],addr[1]);
}

GA_color REF2D_convertPixel32BGRA_16(
    uchar *addr DCL_X_ARG)
{
    return PACK_COLOR_16(addr[1],addr[2],addr[3]);
}

GA_color REF2D_convertPixel32ARGB_15_Blend(
    uchar *addr DCL_X_ARG)
{
    DC.alphaValue = addr[3];
    return PACK_COLOR_15(addr[2],addr[1],addr[0]);
}

GA_color REF2D_convertPixel32ABGR_15_Blend(
    uchar *addr DCL_X_ARG)
{
    DC.alphaValue = addr[3];
    return PACK_COLOR_15(addr[0],addr[1],addr[2]);
}

GA_color REF2D_convertPixel32RGBA_15_Blend(
    uchar *addr DCL_X_ARG)
{
    DC.alphaValue = addr[0];
    return PACK_COLOR_15(addr[3],addr[2],addr[1]);
}

GA_color REF2D_convertPixel32BGRA_15_Blend(
    uchar *addr DCL_X_ARG)
{
    DC.alphaValue = addr[0];
    return PACK_COLOR_15(addr[1],addr[2],addr[3]);
}

GA_color REF2D_convertPixel32ARGB_16_Blend(
    uchar *addr DCL_X_ARG)
{
    DC.alphaValue = addr[3];
    return PACK_COLOR_16(addr[2],addr[1],addr[0]);
}

GA_color REF2D_convertPixel32ABGR_16_Blend(
    uchar *addr DCL_X_ARG)
{
    DC.alphaValue = addr[3];
    return PACK_COLOR_16(addr[0],addr[1],addr[2]);
}

GA_color REF2D_convertPixel32RGBA_16_Blend(
    uchar *addr DCL_X_ARG)
{
    DC.alphaValue = addr[0];
    return PACK_COLOR_16(addr[3],addr[2],addr[1]);
}

GA_color REF2D_convertPixel32BGRA_16_Blend(
    uchar *addr DCL_X_ARG)
{
    DC.alphaValue = addr[0];
    return PACK_COLOR_16(addr[1],addr[2],addr[3]);
}

/****************************************************************************
REMARKS:
8bit + alpha pixel conversion functions
****************************************************************************/

GA_color REF2D_convertPixel32ARGB_8A(
    uchar *addr DCL_X_ARG)
{
    return REF2D_findClosest_24RGB(*((ulong*)addr)) | ((ulong)addr[3] << 8);
}

GA_color REF2D_convertPixel32ABGR_8A(
    uchar *addr DCL_X_ARG)
{
    return REF2D_findClosest_24BGR(*((ulong*)addr)) | ((ulong)addr[3] << 8);
}

GA_color REF2D_convertPixel32RGBA_8A(
    uchar *addr DCL_X_ARG)
{
    return REF2D_findClosest_32RGBA(*((ulong*)addr)) | ((ulong)addr[0] << 8);
}

GA_color REF2D_convertPixel32BGRA_8A(
    uchar *addr DCL_X_ARG)
{
    return REF2D_findClosest_32BGRA(*((ulong*)addr)) | ((ulong)addr[0] << 8);
}

void REF2D_setupDither_8A(
    N_int32 y)
{
    DC.ditherRow = &_REF2D_dither8x8[(y & 7) << 3];
}

GA_color REF2D_convertPixel32ARGB_8A_Dithered(
    uchar *addr DCL_X_ARG)
{
    GA_color    c;
    uchar       R,G,B;

    B = addr[0];
    G = addr[1];
    R = addr[2];
    PACK_COLOR_DITHER8_EX(DC.ditherRow,c,DC.dstX,R,G,B);
    DC.dstX++;
    return c | ((ulong)addr[3] << 8);
}

GA_color REF2D_convertPixel32ABGR_8A_Dithered(
    uchar *addr DCL_X_ARG)
{
    GA_color    c;
    uchar       R,G,B;

    R = addr[0];
    G = addr[1];
    B = addr[2];
    PACK_COLOR_DITHER8_EX(DC.ditherRow,c,DC.dstX,R,G,B);
    DC.dstX++;
    return c | ((ulong)addr[3] << 8);
}

GA_color REF2D_convertPixel32RGBA_8A_Dithered(
    uchar *addr DCL_X_ARG)
{
    GA_color    c;
    uchar       R,G,B;

    B = addr[1];
    G = addr[2];
    R = addr[3];
    PACK_COLOR_DITHER8_EX(DC.ditherRow,c,DC.dstX,R,G,B);
    DC.dstX++;
    return c | ((ulong)addr[0] << 8);
}

GA_color REF2D_convertPixel32BGRA_8A_Dithered(
    uchar *addr DCL_X_ARG)
{
    GA_color    c;
    uchar       R,G,B;

    R = addr[1];
    G = addr[2];
    B = addr[3];
    PACK_COLOR_DITHER8_EX(DC.ditherRow,c,DC.dstX,R,G,B);
    DC.dstX++;
    return c | ((ulong)addr[0] << 8);
}

/****************************************************************************
REMARKS:
Dithered pixel functions
****************************************************************************/

void REF2D_setupDither_16(
    N_int32 y)
{
    DC.ditherRow = &_REF2D_dither4x4[(y & 3) << 2];
}

GA_color REF2D_convertPixel24RGB_15_Dithered(
    uchar *addr DCL_X_ARG)
{
    GA_color    c;
    uchar       R,G,B;

    B = addr[0];
    G = addr[1];
    R = addr[2];
    PACK_COLOR_DITHER15_EX(DC.ditherRow,c,DC.dstX,R,G,B);
    DC.dstX++;
    return c;
}

GA_color REF2D_convertPixel24BGR_15_Dithered(
    uchar *addr DCL_X_ARG)
{
    GA_color    c;
    uchar       R,G,B;

    R = addr[0];
    G = addr[1];
    B = addr[2];
    PACK_COLOR_DITHER15_EX(DC.ditherRow,c,DC.dstX,R,G,B);
    DC.dstX++;
    return c;
}

GA_color REF2D_convertPixel24RGB_16_Dithered(
    uchar *addr DCL_X_ARG)
{
    GA_color    c;
    uchar       R,G,B;

    B = addr[0];
    G = addr[1];
    R = addr[2];
    PACK_COLOR_DITHER16_EX(DC.ditherRow,c,DC.dstX,R,G,B);
    DC.dstX++;
    return c;
}

GA_color REF2D_convertPixel24BGR_16_Dithered(
    uchar *addr DCL_X_ARG)
{
    GA_color    c;
    uchar       R,G,B;

    R = addr[0];
    G = addr[1];
    B = addr[2];
    PACK_COLOR_DITHER16_EX(DC.ditherRow,c,DC.dstX,R,G,B);
    DC.dstX++;
    return c;
}

GA_color REF2D_convertPixel32RGBA_15_Dithered(
    uchar *addr DCL_X_ARG)
{
    GA_color    c;
    uchar       R,G,B;

    B = addr[1];
    G = addr[2];
    R = addr[3];
    PACK_COLOR_DITHER15_EX(DC.ditherRow,c,DC.dstX,R,G,B);
    DC.dstX++;
    return c;
}

GA_color REF2D_convertPixel32BGRA_15_Dithered(
    uchar *addr DCL_X_ARG)
{
    GA_color    c;
    uchar       R,G,B;

    R = addr[1];
    G = addr[2];
    B = addr[3];
    PACK_COLOR_DITHER15_EX(DC.ditherRow,c,DC.dstX,R,G,B);
    DC.dstX++;
    return c;
}

GA_color REF2D_convertPixel32RGBA_16_Dithered(
    uchar *addr DCL_X_ARG)
{
    GA_color    c;
    uchar       R,G,B;

    B = addr[1];
    G = addr[2];
    R = addr[3];
    PACK_COLOR_DITHER16_EX(DC.ditherRow,c,DC.dstX,R,G,B);
    DC.dstX++;
    return c;
}

GA_color REF2D_convertPixel32BGRA_16_Dithered(
    uchar *addr DCL_X_ARG)
{
    GA_color    c;
    uchar       R,G,B;

    R = addr[1];
    G = addr[2];
    B = addr[3];
    PACK_COLOR_DITHER16_EX(DC.ditherRow,c,DC.dstX,R,G,B);
    DC.dstX++;
    return c;
}

GA_color REF2D_convertPixel32ARGB_15_Dithered_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color    c;
    uchar       R,G,B;

    B = addr[0];
    G = addr[1];
    R = addr[2];
    DC.alphaValue = addr[3];
    PACK_COLOR_DITHER15_EX(DC.ditherRow,c,DC.dstX,R,G,B);
    DC.dstX++;
    return c;
}

GA_color REF2D_convertPixel32ABGR_15_Dithered_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color    c;
    uchar       R,G,B;

    R = addr[0];
    G = addr[1];
    B = addr[2];
    DC.alphaValue = addr[3];
    PACK_COLOR_DITHER15_EX(DC.ditherRow,c,DC.dstX,R,G,B);
    DC.dstX++;
    return c;
}

GA_color REF2D_convertPixel32RGBA_15_Dithered_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color    c;
    uchar       R,G,B;

    DC.alphaValue = addr[0];
    B = addr[1];
    G = addr[2];
    R = addr[3];
    PACK_COLOR_DITHER15_EX(DC.ditherRow,c,DC.dstX,R,G,B);
    DC.dstX++;
    return c;
}

GA_color REF2D_convertPixel32BGRA_15_Dithered_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color    c;
    uchar       R,G,B;

    DC.alphaValue = addr[0];
    R = addr[1];
    G = addr[2];
    B = addr[3];
    PACK_COLOR_DITHER15_EX(DC.ditherRow,c,DC.dstX,R,G,B);
    DC.dstX++;
    return c;
}

GA_color REF2D_convertPixel32ARGB_16_Dithered_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color    c;
    uchar       R,G,B;

    B = addr[0];
    G = addr[1];
    R = addr[2];
    DC.alphaValue = addr[3];
    PACK_COLOR_DITHER16_EX(DC.ditherRow,c,DC.dstX,R,G,B);
    DC.dstX++;
    return c;
}

GA_color REF2D_convertPixel32ABGR_16_Dithered_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color    c;
    uchar       R,G,B;

    R = addr[0];
    G = addr[1];
    B = addr[2];
    DC.alphaValue = addr[3];
    PACK_COLOR_DITHER16_EX(DC.ditherRow,c,DC.dstX,R,G,B);
    DC.dstX++;
    return c;
}

GA_color REF2D_convertPixel32RGBA_16_Dithered_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color    c;
    uchar       R,G,B;

    DC.alphaValue = addr[0];
    B = addr[1];
    G = addr[2];
    R = addr[3];
    PACK_COLOR_DITHER16_EX(DC.ditherRow,c,DC.dstX,R,G,B);
    DC.dstX++;
    return c;
}

GA_color REF2D_convertPixel32BGRA_16_Dithered_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color    c;
    uchar       R,G,B;

    DC.alphaValue = addr[0];
    R = addr[1];
    G = addr[2];
    B = addr[3];
    PACK_COLOR_DITHER16_EX(DC.ditherRow,c,DC.dstX,R,G,B);
    DC.dstX++;
    return c;
}

/****************************************************************************
REMARKS:
Source Transparent color keyed pixel functions
****************************************************************************/

#define DRAW_PIXEL_SRC_COLOR_KEY_PROTO() \
    void REF2D_drawPixelSrcColorKey_16(uchar *addr DCL_X_ARG,GA_color color)
#define DRAW_PIXEL_SRC_COLOR_KEY_RANGE_PROTO() \
    void REF2D_drawPixelSrcColorKeyRange_16(uchar *addr DCL_X_ARG,GA_color color)
#include "raster/transpix.c"

/****************************************************************************
REMARKS:
ROP2 pixel functions.
****************************************************************************/

#define DRAW_PIXEL_PROTO(rop) \
    void REF2D_drawPixel_16##rop (uchar *addr DCL_X_ARG,GA_color color)
#include "raster/pixel.c"

#define DRAW_PIXEL_PROTO(rop) \
    void REF2D_drawPixelMasked_16##rop (uchar *addr DCL_X_ARG,GA_color color)
#include "raster/pixelm.c"

