/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser functions
*               for the SciTech SNAP Graphics Architecture.
*
****************************************************************************/

#include "ref2dint.h"

/*----------------------------- Implementation ----------------------------*/

#define BUF_INCR    3
#define BUF_INCR2   3
#define COLOR_TYPE  uchar
#define FULL_COLOR  GA_color

/****************************************************************************
REMARKS:
Pixel reading functions
****************************************************************************/

GA_color REF2D_readPixel_24(uchar *addr DCL_X_ARG)
{
    GA_color D = 0;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    return D;
}

/****************************************************************************
REMARKS:
Pixel color conversion functions
****************************************************************************/

GA_color REF2D_convertPixel15_24RGB(
    uchar *addr DCL_X_ARG)
{
    GA_color D,c = *((ushort*)addr);
    BYTE0(D) = UNPACK_B_15(c);
    BYTE1(D) = UNPACK_G_15(c);
    BYTE2(D) = UNPACK_R_15(c);
    return D;
}

GA_color REF2D_convertPixel15_24BGR(
    uchar *addr DCL_X_ARG)
{
    GA_color D,c = *((ushort*)addr);
    BYTE0(D) = UNPACK_R_15(c);
    BYTE1(D) = UNPACK_G_15(c);
    BYTE2(D) = UNPACK_B_15(c);
    return D;
}

GA_color REF2D_convertPixel16_24RGB(
    uchar *addr DCL_X_ARG)
{
    GA_color D,c = *((ushort*)addr);
    BYTE0(D) = UNPACK_B_16(c);
    BYTE1(D) = UNPACK_G_16(c);
    BYTE2(D) = UNPACK_R_16(c);
    return D;
}

GA_color REF2D_convertPixel16_24BGR(
    uchar *addr DCL_X_ARG)
{
    GA_color D,c = *((ushort*)addr);
    BYTE0(D) = UNPACK_R_16(c);
    BYTE1(D) = UNPACK_G_16(c);
    BYTE2(D) = UNPACK_B_16(c);
    return D;
}

GA_color REF2D_convertPixel24RGB_24BGR(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE2(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE0(D) = addr[2];
    return D;
}

GA_color REF2D_convertPixel32ABGR_24RGB(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE2(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE0(D) = addr[2];
    return D;
}

GA_color REF2D_convertPixel32BGRA_24RGB(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE2(D) = addr[1];
    BYTE1(D) = addr[2];
    BYTE0(D) = addr[3];
    return D;
}

GA_color REF2D_convertPixel32ARGB_24RGB(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    return D;
}

GA_color REF2D_convertPixel32RGBA_24RGB(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE0(D) = addr[1];
    BYTE1(D) = addr[2];
    BYTE2(D) = addr[3];
    return D;
}

GA_color REF2D_convertPixel32ABGR_24BGR(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    return D;
}

GA_color REF2D_convertPixel32BGRA_24BGR(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE0(D) = addr[1];
    BYTE1(D) = addr[2];
    BYTE2(D) = addr[3];
    return D;
}

GA_color REF2D_convertPixel32ARGB_24BGR(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE2(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE0(D) = addr[2];
    return D;
}

GA_color REF2D_convertPixel32RGBA_24BGR(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE2(D) = addr[1];
    BYTE1(D) = addr[2];
    BYTE0(D) = addr[3];
    return D;
}

GA_color REF2D_convertPixel32ABGR_24RGB_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE2(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE0(D) = addr[2];
    BYTE3(D) = addr[3];
    return D;
}

GA_color REF2D_convertPixel32BGRA_24RGB_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE3(D) = addr[0];
    BYTE2(D) = addr[1];
    BYTE1(D) = addr[2];
    BYTE0(D) = addr[3];
    return D;
}

GA_color REF2D_convertPixel32ARGB_24RGB_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    BYTE3(D) = addr[3];
    return D;
}

GA_color REF2D_convertPixel32RGBA_24RGB_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE3(D) = addr[0];
    BYTE0(D) = addr[1];
    BYTE1(D) = addr[2];
    BYTE2(D) = addr[3];
    return D;
}

GA_color REF2D_convertPixel32ABGR_24BGR_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    BYTE3(D) = addr[3];
    return D;
}

GA_color REF2D_convertPixel32BGRA_24BGR_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE3(D) = addr[0];
    BYTE0(D) = addr[1];
    BYTE1(D) = addr[2];
    BYTE2(D) = addr[3];
    return D;
}

GA_color REF2D_convertPixel32ARGB_24BGR_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE2(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE0(D) = addr[2];
    BYTE3(D) = addr[3];
    return D;
}

GA_color REF2D_convertPixel32RGBA_24BGR_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE3(D) = addr[0];
    BYTE2(D) = addr[1];
    BYTE1(D) = addr[2];
    BYTE0(D) = addr[3];
    return D;
}

/****************************************************************************
REMARKS:
Source Transparent color keyed pixel functions
****************************************************************************/

void REF2D_drawPixelSrcColorKey_24(
    uchar *addr DCL_X_ARG,
    GA_color color)
{
    if (color != DC.colorKeyLo) {
        PLOT_S(addr[0],BYTE0(color));
        PLOT_S(addr[1],BYTE1(color));
        PLOT_S(addr[2],BYTE2(color));
        }
}

void REF2D_drawPixelSrcColorKeyRange_24(
    uchar *addr DCL_X_ARG,
    GA_color color)
{
    if (color < DC.colorKeyLo || color > DC.colorKeyHi) {
        PLOT_S(addr[0],BYTE0(color));
        PLOT_S(addr[1],BYTE1(color));
        PLOT_S(addr[2],BYTE2(color));
        }
}

/****************************************************************************
REMARKS:
ROP2 pixel functions.
****************************************************************************/

void REF2D_drawPixel_24_0(uchar *addr DCL_X_ARG,GA_color color)
{
    addr[0] = 0;
    addr[1] = 0;
    addr[2] = 0;
    (void)color;
}

void REF2D_drawPixel_24_DSon(uchar *addr DCL_X_ARG,GA_color color)
{
    PLOT_DSon(addr[0],BYTE0(color));
    PLOT_DSon(addr[1],BYTE1(color));
    PLOT_DSon(addr[2],BYTE2(color));
}

void REF2D_drawPixel_24_DSna(uchar *addr DCL_X_ARG,GA_color color)
{
    PLOT_DSna(addr[0],BYTE0(color));
    PLOT_DSna(addr[1],BYTE1(color));
    PLOT_DSna(addr[2],BYTE2(color));
}

void REF2D_drawPixel_24_Sn(uchar *addr DCL_X_ARG,GA_color color)
{
    PLOT_Sn(addr[0],BYTE0(color));
    PLOT_Sn(addr[1],BYTE1(color));
    PLOT_Sn(addr[2],BYTE2(color));
}

void REF2D_drawPixel_24_SDna(uchar *addr DCL_X_ARG,GA_color color)
{
    PLOT_SDna(addr[0],BYTE0(color));
    PLOT_SDna(addr[1],BYTE1(color));
    PLOT_SDna(addr[2],BYTE2(color));
}

void REF2D_drawPixel_24_Dn(uchar *addr DCL_X_ARG,GA_color color)
{
    addr[0] = ~addr[0];
    addr[1] = ~addr[1];
    addr[2] = ~addr[2];
    (void)color;
}

void REF2D_drawPixel_24_DSx(uchar *addr DCL_X_ARG,GA_color color)
{
    PLOT_DSx(addr[0],BYTE0(color));
    PLOT_DSx(addr[1],BYTE1(color));
    PLOT_DSx(addr[2],BYTE2(color));
}

void REF2D_drawPixel_24_DSan(uchar *addr DCL_X_ARG,GA_color color)
{
    PLOT_DSan(addr[0],BYTE0(color));
    PLOT_DSan(addr[1],BYTE1(color));
    PLOT_DSan(addr[2],BYTE2(color));
}

void REF2D_drawPixel_24_DSa(uchar *addr DCL_X_ARG,GA_color color)
{
    PLOT_DSa(addr[0],BYTE0(color));
    PLOT_DSa(addr[1],BYTE1(color));
    PLOT_DSa(addr[2],BYTE2(color));
}

void REF2D_drawPixel_24_DSxn(uchar *addr DCL_X_ARG,GA_color color)
{
    PLOT_DSxn(addr[0],BYTE0(color));
    PLOT_DSxn(addr[1],BYTE1(color));
    PLOT_DSxn(addr[2],BYTE2(color));
}

void REF2D_drawPixel_24_DSno(uchar *addr DCL_X_ARG,GA_color color)
{
    PLOT_DSno(addr[0],BYTE0(color));
    PLOT_DSno(addr[1],BYTE1(color));
    PLOT_DSno(addr[2],BYTE2(color));
}

void REF2D_drawPixel_24_S(uchar *addr DCL_X_ARG,GA_color color)
{
    PLOT_S(addr[0],BYTE0(color));
    PLOT_S(addr[1],BYTE1(color));
    PLOT_S(addr[2],BYTE2(color));
}

void REF2D_drawPixel_24_SDno(uchar *addr DCL_X_ARG,GA_color color)
{
    PLOT_SDno(addr[0],BYTE0(color));
    PLOT_SDno(addr[1],BYTE1(color));
    PLOT_SDno(addr[2],BYTE2(color));
}

void REF2D_drawPixel_24_DSo(uchar *addr DCL_X_ARG,GA_color color)
{
    PLOT_DSo(addr[0],BYTE0(color));
    PLOT_DSo(addr[1],BYTE1(color));
    PLOT_DSo(addr[2],BYTE2(color));
}

void REF2D_drawPixel_24_1(uchar *addr DCL_X_ARG,GA_color color)
{
    addr[0] = 0xFF;
    addr[1] = 0xFF;
    addr[2] = 0xFF;
    (void)color;
}

/****************************************************************************
REMARKS:
ROP2 masked pixel functions.
****************************************************************************/

void REF2D_drawPixelMasked_24_0(uchar *addr DCL_X_ARG,GA_color color)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = (D & DC.planeMaskInv) | 0;
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)color;
}

void REF2D_drawPixelMasked_24_DSon(uchar *addr DCL_X_ARG,GA_color color)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = (D & DC.planeMaskInv) | (~(D | (color)) & DC.planeMask);
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelMasked_24_DSna(uchar *addr DCL_X_ARG,GA_color color)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = (D & DC.planeMaskInv) | ((D & ~(color)) & DC.planeMask);
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelMasked_24_Sn(uchar *addr DCL_X_ARG,GA_color color)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = (D & DC.planeMaskInv) | (~(color) & DC.planeMask);
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelMasked_24_SDna(uchar *addr DCL_X_ARG,GA_color color)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = (D & DC.planeMaskInv) | ((~D & (color)) & DC.planeMask);
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelMasked_24_Dn(uchar *addr DCL_X_ARG,GA_color color)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = (D & DC.planeMaskInv) | ((~D) & DC.planeMask);
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)color;
}

void REF2D_drawPixelMasked_24_DSx(uchar *addr DCL_X_ARG,GA_color color)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = (D & DC.planeMaskInv) | ((D ^ (color)) & DC.planeMask);
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelMasked_24_DSan(uchar *addr DCL_X_ARG,GA_color color)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = (D & DC.planeMaskInv) | (~(D & (color)) & DC.planeMask);
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelMasked_24_DSa(uchar *addr DCL_X_ARG,GA_color color)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = (D & DC.planeMaskInv) | ((D & (color)) & DC.planeMask);
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelMasked_24_DSxn(uchar *addr DCL_X_ARG,GA_color color)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = (D & DC.planeMaskInv) | (~(D ^ (color)) & DC.planeMask);
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelMasked_24_DSno(uchar *addr DCL_X_ARG,GA_color color)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = (D & DC.planeMaskInv) | ((D | ~(color)) & DC.planeMask);
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelMasked_24_S(uchar *addr DCL_X_ARG,GA_color color)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = (D & DC.planeMaskInv) | ((color) & DC.planeMask);
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelMasked_24_SDno(uchar *addr DCL_X_ARG,GA_color color)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = (D & DC.planeMaskInv) | ((~D | (color)) & DC.planeMask);
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelMasked_24_DSo(uchar *addr DCL_X_ARG,GA_color color)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = (D & DC.planeMaskInv) | ((D | (color)) & DC.planeMask);
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
}

void REF2D_drawPixelMasked_24_1(uchar *addr DCL_X_ARG,GA_color color)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    D = (D & DC.planeMaskInv) | DC.planeMask;
    addr[0] = BYTE0(D);
    addr[1] = BYTE1(D);
    addr[2] = BYTE2(D);
    (void)color;
}

