/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser functions
*               for the SciTech SNAP Graphics Architecture.
*
****************************************************************************/

#include "ref2dint.h"

/*----------------------------- Implementation ----------------------------*/

#define BUF_INCR    4
#define COLOR_TYPE  ulong
#define COLOR_WHITE 0xFFFFFFFF
#define PATT_TYPE   b32

/****************************************************************************
REMARKS:
Pixel reading functions
****************************************************************************/

GA_color REF2D_readPixel_32(
    uchar *addr DCL_X_ARG)
{
    return *((COLOR_TYPE*)addr);
}

/****************************************************************************
REMARKS:
Pixel color conversion functions
****************************************************************************/

GA_color REF2D_convertPixel15_32ARGB(
    uchar *addr DCL_X_ARG)
{
    GA_color D,c = *((ushort*)addr);
    BYTE2(D) = UNPACK_R_15(c);
    BYTE1(D) = UNPACK_G_15(c);
    BYTE0(D) = UNPACK_B_15(c);
    return D;
}

GA_color REF2D_convertPixel15_32ABGR(
    uchar *addr DCL_X_ARG)
{
    GA_color D,c = *((ushort*)addr);
    BYTE0(D) = UNPACK_R_15(c);
    BYTE1(D) = UNPACK_G_15(c);
    BYTE2(D) = UNPACK_B_15(c);
    return D;
}

GA_color REF2D_convertPixel15_32RGBA(
    uchar *addr DCL_X_ARG)
{
    GA_color D,c = *((ushort*)addr);
    BYTE3(D) = UNPACK_R_15(c);
    BYTE2(D) = UNPACK_G_15(c);
    BYTE1(D) = UNPACK_B_15(c);
    return D;
}

GA_color REF2D_convertPixel15_32BGRA(
    uchar *addr DCL_X_ARG)
{
    GA_color D,c = *((ushort*)addr);
    BYTE1(D) = UNPACK_R_15(c);
    BYTE2(D) = UNPACK_G_15(c);
    BYTE3(D) = UNPACK_B_15(c);
    return D;
}

GA_color REF2D_convertPixel16_32ARGB(
    uchar *addr DCL_X_ARG)
{
    GA_color D,c = *((ushort*)addr);
    BYTE2(D) = UNPACK_R_16(c);
    BYTE1(D) = UNPACK_G_16(c);
    BYTE0(D) = UNPACK_B_16(c);
    return D;
}

GA_color REF2D_convertPixel16_32ABGR(
    uchar *addr DCL_X_ARG)
{
    GA_color D,c = *((ushort*)addr);
    BYTE0(D) = UNPACK_R_16(c);
    BYTE1(D) = UNPACK_G_16(c);
    BYTE2(D) = UNPACK_B_16(c);
    return D;
}

GA_color REF2D_convertPixel16_32RGBA(
    uchar *addr DCL_X_ARG)
{
    GA_color D,c = *((ushort*)addr);
    BYTE3(D) = UNPACK_R_16(c);
    BYTE2(D) = UNPACK_G_16(c);
    BYTE1(D) = UNPACK_B_16(c);
    return D;
}

GA_color REF2D_convertPixel16_32BGRA(
    uchar *addr DCL_X_ARG)
{
    GA_color D,c = *((ushort*)addr);
    BYTE1(D) = UNPACK_R_16(c);
    BYTE2(D) = UNPACK_G_16(c);
    BYTE3(D) = UNPACK_B_16(c);
    return D;
}

GA_color REF2D_convertPixel15_32ARGB_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color D,c = *((ushort*)addr);
    BYTE3(D) = UNPACK_A_15(c);
    BYTE2(D) = UNPACK_R_15(c);
    BYTE1(D) = UNPACK_G_15(c);
    BYTE0(D) = UNPACK_B_15(c);
    return D;
}

GA_color REF2D_convertPixel15_32ABGR_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color D,c = *((ushort*)addr);
    BYTE3(D) = UNPACK_A_15(c);
    BYTE0(D) = UNPACK_R_15(c);
    BYTE1(D) = UNPACK_G_15(c);
    BYTE2(D) = UNPACK_B_15(c);
    return D;
}

GA_color REF2D_convertPixel15_32RGBA_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color D,c = *((ushort*)addr);
    BYTE0(D) = UNPACK_A_15(c);
    BYTE3(D) = UNPACK_R_15(c);
    BYTE2(D) = UNPACK_G_15(c);
    BYTE1(D) = UNPACK_B_15(c);
    return D;
}

GA_color REF2D_convertPixel15_32BGRA_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color D,c = *((ushort*)addr);
    BYTE0(D) = UNPACK_A_15(c);
    BYTE1(D) = UNPACK_R_15(c);
    BYTE2(D) = UNPACK_G_15(c);
    BYTE3(D) = UNPACK_B_15(c);
    return D;
}

GA_color REF2D_convertPixel24RGB_32ARGB(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    return D;
}

GA_color REF2D_convertPixel24RGB_32ABGR(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE2(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE0(D) = addr[2];
    return D;
}

GA_color REF2D_convertPixel24RGB_32RGBA(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE1(D) = addr[0];
    BYTE2(D) = addr[1];
    BYTE3(D) = addr[2];
    return D;
}

GA_color REF2D_convertPixel24RGB_32BGRA(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE3(D) = addr[0];
    BYTE2(D) = addr[1];
    BYTE1(D) = addr[2];
    return D;
}

GA_color REF2D_convertPixel24BGR_32ARGB(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE2(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE0(D) = addr[2];
    return D;
}

GA_color REF2D_convertPixel24BGR_32ABGR(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE2(D) = addr[2];
    return D;
}

GA_color REF2D_convertPixel24BGR_32RGBA(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE3(D) = addr[0];
    BYTE2(D) = addr[1];
    BYTE1(D) = addr[2];
    return D;
}

GA_color REF2D_convertPixel24BGR_32BGRA(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE1(D) = addr[0];
    BYTE2(D) = addr[1];
    BYTE3(D) = addr[2];
    return D;
}

GA_color REF2D_convertPixel32ABGR_32ARGB(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE2(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE0(D) = addr[2];
    return D;
}

GA_color REF2D_convertPixel32BGRA_32ARGB(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE2(D) = addr[1];
    BYTE1(D) = addr[2];
    BYTE0(D) = addr[3];
    return D;
}

GA_color REF2D_convertPixel32RGBA_32ARGB(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE0(D) = addr[1];
    BYTE1(D) = addr[2];
    BYTE2(D) = addr[3];
    return D;
}

GA_color REF2D_convertPixel32BGRA_32ABGR(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE0(D) = addr[1];
    BYTE1(D) = addr[2];
    BYTE2(D) = addr[3];
    return D;
}

GA_color REF2D_convertPixel32ARGB_32ABGR(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE2(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE0(D) = addr[2];
    return D;
}

GA_color REF2D_convertPixel32RGBA_32ABGR(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE2(D) = addr[1];
    BYTE1(D) = addr[2];
    BYTE0(D) = addr[3];
    return D;
}

GA_color REF2D_convertPixel32ARGB_32RGBA(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE1(D) = addr[0];
    BYTE2(D) = addr[1];
    BYTE3(D) = addr[2];
    return D;
}

GA_color REF2D_convertPixel32ABGR_32RGBA(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE3(D) = addr[0];
    BYTE2(D) = addr[1];
    BYTE1(D) = addr[2];
    return D;
}

GA_color REF2D_convertPixel32BGRA_32RGBA(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE3(D) = addr[1];
    BYTE2(D) = addr[2];
    BYTE1(D) = addr[3];
    return D;
}

GA_color REF2D_convertPixel32ARGB_32BGRA(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE3(D) = addr[0];
    BYTE2(D) = addr[1];
    BYTE1(D) = addr[2];
    return D;
}

GA_color REF2D_convertPixel32ABGR_32BGRA(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE1(D) = addr[0];
    BYTE2(D) = addr[1];
    BYTE3(D) = addr[2];
    return D;
}

GA_color REF2D_convertPixel32RGBA_32BGRA(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE3(D) = addr[1];
    BYTE2(D) = addr[2];
    BYTE1(D) = addr[3];
    return D;
}

GA_color REF2D_convertPixel32ABGR_32ARGB_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE2(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE0(D) = addr[2];
    BYTE3(D) = addr[3];
    return D;
}

GA_color REF2D_convertPixel32BGRA_32ARGB_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE3(D) = addr[0];
    BYTE2(D) = addr[1];
    BYTE1(D) = addr[2];
    BYTE0(D) = addr[3];
    return D;
}

GA_color REF2D_convertPixel32RGBA_32ARGB_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE3(D) = addr[0];
    BYTE0(D) = addr[1];
    BYTE1(D) = addr[2];
    BYTE2(D) = addr[3];
    return D;
}

GA_color REF2D_convertPixel32BGRA_32ABGR_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE3(D) = addr[0];
    BYTE0(D) = addr[1];
    BYTE1(D) = addr[2];
    BYTE2(D) = addr[3];
    return D;
}

GA_color REF2D_convertPixel32ARGB_32ABGR_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE2(D) = addr[0];
    BYTE1(D) = addr[1];
    BYTE0(D) = addr[2];
    BYTE3(D) = addr[3];
    return D;
}

GA_color REF2D_convertPixel32RGBA_32ABGR_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE3(D) = addr[0];
    BYTE2(D) = addr[1];
    BYTE1(D) = addr[2];
    BYTE0(D) = addr[3];
    return D;
}

GA_color REF2D_convertPixel32ARGB_32RGBA_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE1(D) = addr[0];
    BYTE2(D) = addr[1];
    BYTE3(D) = addr[2];
    BYTE0(D) = addr[3];
    return D;
}

GA_color REF2D_convertPixel32ABGR_32RGBA_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE3(D) = addr[0];
    BYTE2(D) = addr[1];
    BYTE1(D) = addr[2];
    BYTE0(D) = addr[3];
    return D;
}

GA_color REF2D_convertPixel32BGRA_32RGBA_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE3(D) = addr[1];
    BYTE2(D) = addr[2];
    BYTE1(D) = addr[3];
    return D;
}

GA_color REF2D_convertPixel32ARGB_32BGRA_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE3(D) = addr[0];
    BYTE2(D) = addr[1];
    BYTE1(D) = addr[2];
    BYTE0(D) = addr[3];
    return D;
}

GA_color REF2D_convertPixel32ABGR_32BGRA_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE1(D) = addr[0];
    BYTE2(D) = addr[1];
    BYTE3(D) = addr[2];
    BYTE0(D) = addr[3];
    return D;
}

GA_color REF2D_convertPixel32RGBA_32BGRA_Blend(
    uchar *addr DCL_X_ARG)
{
    GA_color D;
    BYTE0(D) = addr[0];
    BYTE3(D) = addr[1];
    BYTE2(D) = addr[2];
    BYTE1(D) = addr[3];
    return D;
}

/****************************************************************************
REMARKS:
Source Transparent color keyed pixel functions
****************************************************************************/

#define DRAW_PIXEL_SRC_COLOR_KEY_PROTO() \
    void REF2D_drawPixelSrcColorKey_32(uchar *addr DCL_X_ARG,GA_color color)
#define DRAW_PIXEL_SRC_COLOR_KEY_RANGE_PROTO() \
    void REF2D_drawPixelSrcColorKeyRange_32(uchar *addr DCL_X_ARG,GA_color color)
#include "raster/transpix.c"

/****************************************************************************
REMARKS:
ROP2 pixel functions.
****************************************************************************/

#define DRAW_PIXEL_PROTO(rop) \
    void REF2D_drawPixel_32##rop (uchar *addr DCL_X_ARG,GA_color color)
#include "raster/pixel.c"

#define DRAW_PIXEL_PROTO(rop) \
    void REF2D_drawPixelMasked_32##rop (uchar *addr DCL_X_ARG,GA_color color)
#include "raster/pixelm.c"

