/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser functions
*               for the SciTech SNAP Graphics Architecture.
*
****************************************************************************/

#include "ref2dint.h"

/*----------------------------- Implementation ----------------------------*/

#ifdef SUPPORT_4BPP

/****************************************************************************
REMARKS:
Pixel reading functions
****************************************************************************/

GA_color REF2D_readPixel_4(uchar *addr,int x)
{
    if (x & 1)
        return (*addr & 0x0F);
    else
        return (*addr >> 4);
}

/****************************************************************************
REMARKS:
Source Transparent color keyed pixel functions
****************************************************************************/

void REF2D_drawPixelSrcColorKey_4(
    uchar *addr DCL_X_ARG,
    GA_color S)
{
    // TODO
    if (x & 1)
        *addr = (uchar)((*addr & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((*addr & 0x0F) | (S << 4));
}

void REF2D_drawPixelSrcColorKeyRange_4(
    uchar *addr DCL_X_ARG,
    GA_color S)
{
    // TODO
    if (x & 1)
        *addr = (uchar)((*addr & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((*addr & 0x0F) | (S << 4));
}

/****************************************************************************
REMARKS:
ROP2 pixel functions.
****************************************************************************/

void REF2D_drawPixel_4_0(uchar *addr,int x,GA_color S)
{
    if (x & 1)
        *addr = (uchar)(*addr & 0xF0);
    else
        *addr = (uchar)(*addr & 0x0F);
    (void)S;
}

void REF2D_drawPixel_4_DSon(uchar *addr,int x,GA_color S)
{
    GA_color D = *addr;
    if (x & 1)
        *addr = (uchar)((D & 0xF0) | (~((D & 0x0F) | S) & 0x0F));
    else
        *addr = (uchar)((D & 0x0F) | (~((D >> 4) | S) << 4));
}

void REF2D_drawPixel_4_DSna(uchar *addr,int x,GA_color S)
{
    GA_color D = *addr;
    if (x & 1)
        *addr = (uchar)((D & 0xF0) | (((D & 0x0F) & ~S) & 0x0F));
    else
        *addr = (uchar)((D & 0x0F) | (((D >> 4) & ~S) << 4));
}

void REF2D_drawPixel_4_Sn(uchar *addr,int x,GA_color S)
{
    if (x & 1)
        *addr = (uchar)((*addr & 0xF0) | ((~S) & 0x0F));
    else
        *addr = (uchar)((*addr & 0x0F) | ((~S) << 4));
}

void REF2D_drawPixel_4_SDna(uchar *addr,int x,GA_color S)
{
    GA_color D = *addr;
    if (x & 1)
        *addr = (uchar)((D & 0xF0) | ((~(D & 0x0F) & S) & 0x0F));
    else
        *addr = (uchar)((D & 0x0F) | ((~(D >> 4) & S) << 4));
}

void REF2D_drawPixel_4_Dn(uchar *addr,int x,GA_color S)
{
    if (x & 1)
        *addr = (uchar)(*addr ^ 0x0F);
    else
        *addr = (uchar)(*addr ^ 0xF0);
    (void)S;
}

void REF2D_drawPixel_4_DSx(uchar *addr,int x,GA_color S)
{
    GA_color D = *addr;
    if (x & 1)
        *addr = (uchar)((D & 0xF0) | (((D & 0x0F) ^ S) & 0x0F));
    else
        *addr = (uchar)((D & 0x0F) | (((D >> 4) ^ S) << 4));
}

void REF2D_drawPixel_4_DSan(uchar *addr,int x,GA_color S)
{
    GA_color D = *addr;
    if (x & 1)
        *addr = (uchar)((D & 0xF0) | ((~((D & 0x0F) & S)) & 0x0F));
    else
        *addr = (uchar)((D & 0x0F) | ((~((D & 0x0F) & S)) << 4));
}

void REF2D_drawPixel_4_DSa(uchar *addr,int x,GA_color S)
{
    GA_color D = *addr;
    if (x & 1)
        *addr = (uchar)((D & 0xF0) | (((D & 0x0F) & S) & 0x0F));
    else
        *addr = (uchar)((D & 0x0F) | (((D >> 4) & S) << 4));
}

void REF2D_drawPixel_4_DSxn(uchar *addr,int x,GA_color S)
{
    GA_color D = *addr;
    if (x & 1)
        *addr = (uchar)((D & 0xF0) | (~((D & 0x0F) ^ S) & 0x0F));
    else
        *addr = (uchar)((D & 0x0F) | (~((D >> 4) ^ S) << 4));
}

void REF2D_drawPixel_4_DSno(uchar *addr,int x,GA_color S)
{
    GA_color D = *addr;
    if (x & 1)
        *addr = (uchar)((D & 0xF0) | (((D & 0x0F) | ~S) & 0x0F));
    else
        *addr = (uchar)((D & 0x0F) | (((D >> 4) | ~S) << 4));
}

void REF2D_drawPixel_4_S(uchar *addr,int x,GA_color S)
{
    if (x & 1)
        *addr = (uchar)((*addr & 0xF0) | (S & 0x0F));
    else
        *addr = (uchar)((*addr & 0x0F) | (S << 4));
}

void REF2D_drawPixel_4_SDno(uchar *addr,int x,GA_color S)
{
    GA_color D = *addr;
    if (x & 1)
        *addr = (uchar)((D & 0xF0) | ((~(D & 0x0F) | S) & 0x0F));
    else
        *addr = (uchar)((D & 0x0F) | ((~(D >> 4) | S) << 4));
}

void REF2D_drawPixel_4_DSo(uchar *addr,int x,GA_color S)
{
    GA_color D = *addr;
    if (x & 1)
        *addr = (uchar)((D & 0xF0) | (((D & 0x0F) | S) & 0x0F));
    else
        *addr = (uchar)((D & 0x0F) | (((D >> 4) | S) << 4));
}

void REF2D_drawPixel_4_1(uchar *addr,int x,GA_color S)
{
    if (x & 1)
        *addr = (uchar)(*addr | 0x0F);
    else
        *addr = (uchar)(*addr | 0xF0);
    (void)S;
}

/****************************************************************************
REMARKS:
ROP2 masked pixel functions.
****************************************************************************/

void REF2D_drawPixelMasked_4_0(uchar *addr,int x,GA_color S)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | 0);
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)S;
}

void REF2D_drawPixelMasked_4_DSon(uchar *addr,int x,GA_color S)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D | S) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelMasked_4_DSna(uchar *addr,int x,GA_color S)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D & ~S) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelMasked_4_Sn(uchar *addr,int x,GA_color S)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~S & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelMasked_4_SDna(uchar *addr,int x,GA_color S)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S & ~D) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelMasked_4_Dn(uchar *addr,int x,GA_color S)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~D & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)S;
}

void REF2D_drawPixelMasked_4_DSx(uchar *addr,int x,GA_color S)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D ^ S) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelMasked_4_DSan(uchar *addr,int x,GA_color S)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D & S) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelMasked_4_DSa(uchar *addr,int x,GA_color S)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D & S) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelMasked_4_DSxn(uchar *addr,int x,GA_color S)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (~(D ^ S) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelMasked_4_DSno(uchar *addr,int x,GA_color S)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D | ~S) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelMasked_4_S(uchar *addr,int x,GA_color S)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | (S & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelMasked_4_SDno(uchar *addr,int x,GA_color S)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((S | ~D) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelMasked_4_DSo(uchar *addr,int x,GA_color S)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | ((D | S) & DC.planeMask));
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
}

void REF2D_drawPixelMasked_4_1(uchar *addr,int x,GA_color S)
{
    GA_color dest = *addr;
    GA_color D = (x & 1) ? (dest & 0x0F) : (dest >> 4);
    D = ((D & DC.planeMaskInv) | DC.planeMask);
    if (x & 1)
        *addr = (uchar)((dest & 0xF0) | (D & 0x0F));
    else
        *addr = (uchar)((dest & 0x0F) | (D << 4));
    (void)S;
}

#endif

