/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser functions
*               for the SciTech SNAP Graphics Architecture.
*
****************************************************************************/

#include "ref2dint.h"

/*----------------------------- Implementation ----------------------------*/

#define BUF_INCR    1
#define COLOR_TYPE  uchar
#define COLOR_WHITE 0xFF
#define PATT_TYPE   b8

/****************************************************************************
REMARKS:
Pixel reading functions
****************************************************************************/

GA_color REF2D_readPixel_8(
    uchar *addr DCL_X_ARG)
{
    return *((COLOR_TYPE*)addr);
}

/****************************************************************************
REMARKS:
Pixel color conversion functions
****************************************************************************/

GA_color REF2D_convertPixel16_8(
    uchar *addr DCL_X_ARG)
{
    return DC.findClosest(*((ushort*)addr));
}

GA_color REF2D_convertPixel24_8(
    uchar *addr DCL_X_ARG)
{
    GA_color    c;

    BYTE0(c) = addr[0];
    BYTE1(c) = addr[1];
    BYTE2(c) = addr[2];
    return DC.findClosest(c);
}

GA_color REF2D_convertPixel32_8(
    uchar *addr DCL_X_ARG)
{
    return DC.findClosest(*((ulong*)addr));
}

/****************************************************************************
REMARKS:
Dithered pixel functions
****************************************************************************/

void REF2D_setupDither_8(
    N_int32 y)
{
    DC.ditherRow = &_REF2D_dither8x8[(y & 7) << 3];
}

GA_color REF2D_convertPixel15_8_Dithered(
    uchar *addr DCL_X_ARG)
{
    GA_color    c = *((ushort*)addr);
    uchar       R,G,B;

    UNPACK_COLOR_15(c,R,G,B);
    PACK_COLOR_DITHER8_EX(DC.ditherRow,c,DC.dstX,R,G,B);
    DC.dstX++;
    return c;
}

GA_color REF2D_convertPixel16_8_Dithered(
    uchar *addr DCL_X_ARG)
{
    GA_color    c = *((ushort*)addr);
    uchar       R,G,B;

    UNPACK_COLOR_16(c,R,G,B);
    PACK_COLOR_DITHER8_EX(DC.ditherRow,c,DC.dstX,R,G,B);
    DC.dstX++;
    return c;
}

GA_color REF2D_convertPixel24RGB_8_Dithered(
    uchar *addr DCL_X_ARG)
{
    GA_color    c;
    uchar       R,G,B;

    B = addr[0];
    G = addr[1];
    R = addr[2];
    PACK_COLOR_DITHER8_EX(DC.ditherRow,c,DC.dstX,R,G,B);
    DC.dstX++;
    return c;
}

GA_color REF2D_convertPixel24BGR_8_Dithered(
    uchar *addr DCL_X_ARG)
{
    GA_color    c;
    uchar       R,G,B;

    R = addr[0];
    G = addr[1];
    B = addr[2];
    PACK_COLOR_DITHER8_EX(DC.ditherRow,c,DC.dstX,R,G,B);
    DC.dstX++;
    return c;
}

GA_color REF2D_convertPixel32RGBA_8_Dithered(
    uchar *addr DCL_X_ARG)
{
    GA_color    c;
    uchar       R,G,B;

    B = addr[1];
    G = addr[2];
    R = addr[3];
    PACK_COLOR_DITHER8_EX(DC.ditherRow,c,DC.dstX,R,G,B);
    DC.dstX++;
    return c;
}

GA_color REF2D_convertPixel32BGRA_8_Dithered(
    uchar *addr DCL_X_ARG)
{
    GA_color    c;
    uchar       R,G,B;

    R = addr[1];
    G = addr[2];
    B = addr[3];
    PACK_COLOR_DITHER8_EX(DC.ditherRow,c,DC.dstX,R,G,B);
    DC.dstX++;
    return c;
}

/****************************************************************************
REMARKS:
Source Transparent color keyed pixel functions
****************************************************************************/

#define DRAW_PIXEL_SRC_COLOR_KEY_PROTO() \
    void REF2D_drawPixelSrcColorKey_8(uchar *addr DCL_X_ARG,GA_color color)
#define DRAW_PIXEL_SRC_COLOR_KEY_RANGE_PROTO() \
    void REF2D_drawPixelSrcColorKeyRange_8(uchar *addr DCL_X_ARG,GA_color color)
#include "raster/transpix.c"

/****************************************************************************
REMARKS:
ROP2 pixel functions.
****************************************************************************/

#define DRAW_PIXEL_PROTO(rop) \
    void REF2D_drawPixel_8##rop (uchar *addr DCL_X_ARG,GA_color color)
#include "raster/pixel.c"

#define DRAW_PIXEL_PROTO(rop) \
    void REF2D_drawPixelMasked_8##rop (uchar *addr DCL_X_ARG,GA_color color)
#include "raster/pixelm.c"

