/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the portrait display support
*               routines in the SNAP Reference Rasteriser.
*
****************************************************************************/

#include "portint.h"

// TODO: Clipped lines!

// TODO: BitBltMasked!

//==========================================================================
// IMPLEMENTATION NOTES:
//==========================================================================
//
// 1. The 1bpp bitmap rotation is for MSB and needs to be changed to
//    support LSB bitmaps! Since LSB mono bitmaps are not useful, we
//    have not implemented this function for the time being.
//
//==========================================================================

/*--------------------------- Global variables ----------------------------*/

PORTRAIT_state  PORTRAIT_devCtx;

REF2D_driver PORTRAIT_drv = {
    sizeof(PORTRAIT_drv),
    PORTRAIT_QueryFunctions,
    PORTRAIT_SetDrawBufferSW,
    PORTRAIT_SetDrawSurface,
    PORTRAIT_RotateBitmap,
    PORTRAIT_UnloadDriver,
    PORTRAIT_InitFilter,
    PORTRAIT_DeInitFilter,
    PORTRAIT_GlobalOptionsUpdated,
    PORTRAIT_GetDrawBuffer,
    PORTRAIT_DrawRectExtSW,
    PORTRAIT_ForceSoftwareOnly,
    PORTRAIT_SetColorCompareMask,
    PORTRAIT_PostSwitchPhysicalResolution,
    };

static GA_initFuncs PORTRAIT_initPT = {
    sizeof(GA_initFuncs),
    NULL,                       // PORTRAIT_GetConfigInfo,
    NULL,                       // Reserved
    NULL,                       // PORTRAIT_SetModeProfile,
    NULL,                       // PORTRAIT_GetOptions,
    PORTRAIT_SetOptions,
    PORTRAIT_GetVideoModeInfo,
    PORTRAIT_SetVideoMode,
    NULL,                       // PORTRAIT_GetVideoMode,
    PORTRAIT_GetCustomVideoModeInfo,
    PORTRAIT_SetCustomVideoMode,
    NULL,                       // PORTRAIT_GetClosestPixelClock,
    NULL,                       // PORTRAIT_GetCRTCTimings,
    NULL,                       // PORTRAIT_SetCRTCTimings,
    NULL,                       // PORTRAIT_SetGlobalRefresh,
    NULL,                       // PORTRAIT_SaveRestoreState,
    NULL,                       // PORTRAIT_SetDisplayOutput,
    NULL,                       // PORTRAIT_GetDisplayOutput,
    NULL,                       // PORTRAIT_SetSoftwareRenderFuncs,
    NULL,                       // PORTRAIT_GetUniqueFilename,
    NULL,                       // Reserved
    NULL,                       // PORTRAIT_GetMonitorInfo,
    NULL,                       // PORTRAIT_SetMonitorInfo,
    PORTRAIT_GetCurrentVideoModeInfo,
    NULL,                       // PORTRAIT_GetCertifyInfo,
    NULL,                       // PORTRAIT_SetCRTCTimings,
    NULL,                       // PORTRAIT_AlignLinearBuffer,
    NULL,                       // PORTRAIT_GetIORegisterTraps,
    NULL,                       // PORTRAIT_GetCurrentRefreshRate,
    NULL,                       // PORTRAIT_PollForDisplaySwitch,
    NULL,                       // PORTRAIT_PerformDisplaySwitch,
    NULL,                       // TODO: PORTRAIT_GetVideoModeInfoExt,
    NULL,                       // TODO: PORTRAIT_GetCustomVideoModeInfoExt,
    NULL,                       // PORTRAIT_SwitchPhysicalResolution
    NULL,                       // PORTRAIT_GetNumberOfHeads
    NULL,                       // PORTRAIT_SetActiveHead
    NULL,                       // PORTRAIT_GetActiveHead
    NULL,                       // PORTRAIT_SetRef2dPointer
    };

static GA_cursorFuncs PORTRAIT_cursorPT = {
    sizeof(GA_cursorFuncs),
    PORTRAIT_SetMonoCursor,
    NULL,                       // PORTRAIT_SetMonoCursorColor,
    PORTRAIT_SetColorCursor,
    PORTRAIT_SetCursorPos,
    NULL,                       // PORTRAIT_ShowCursor,
    PORTRAIT_BeginAccess,
    NULL,                       // PORTRAIT_EndAccess,
    PORTRAIT_SetColorCursor256,
    PORTRAIT_SetColorCursorRGBA,
    NULL,                       // PORTRAIT_IsHardwareCursor,
    PORTRAIT_SetColorCursorRGB,
    };

static GA_2DStateFuncs PORTRAIT_state2dPT = {
    sizeof(GA_2DStateFuncs),
    PORTRAIT_SetDrawBuffer,
    NULL,                       // REF2D_IsIdle,
    NULL,                       // REF2D_WaitTillIdle,
    NULL,                       // REF2D_EnableDirectAccess
    NULL,                       // REF2D_DisableDirectAccess
    NULL,                       // REF2D_SetMix,
    NULL,                       // REF2D_SetForeColor,
    NULL,                       // REF2D_SetBackColor,
    PORTRAIT_Set8x8MonoPattern,
    NULL,                       // REF2D_Use8x8MonoPattern,
    NULL,                       // REF2D_Use8x8TransMonoPattern,
    PORTRAIT_Set8x8ColorPattern,
    NULL,                       // REF2D_Use8x8ColorPattern,
    NULL,                       // REF2D_Use8x8TransColorPattern,
    NULL,                       // REF2D_SetLineStipple,
    NULL,                       // REF2D_SetLineStippleCount,
    NULL,                       // REF2D_SetPlaneMask,
    NULL,                       // REF2D_SetSrcBlendFunc,
    NULL,                       // REF2D_SetDstBlendFunc,
    NULL,                       // REF2D_SetAlphaValue,
    NULL,                       // DRIVER_SetupForWindows
    NULL,                       // DRIVER_ResetWindows
    NULL,                       // REF2D_SetLineStyle,
    NULL,                       // REF2D_BuildTranslateVector,
    };

static GA_2DRenderFuncs PORTRAIT_draw2dPT = {
    sizeof(GA_2DRenderFuncs),
    PORTRAIT_GetPixel,
    PORTRAIT_PutPixel,
    PORTRAIT_DrawScanList,
    PORTRAIT_DrawPattScanList,
    PORTRAIT_DrawColorPattScanList,
    PORTRAIT_DrawEllipseList,
    PORTRAIT_DrawPattEllipseList,
    PORTRAIT_DrawColorPattEllipseList,
    PORTRAIT_DrawFatEllipseList,
    PORTRAIT_DrawPattFatEllipseList,
    PORTRAIT_DrawColorPattFatEllipseList,
    PORTRAIT_DrawRect,
    NULL,
    PORTRAIT_DrawPattRect,
    PORTRAIT_DrawColorPattRect,
    PORTRAIT_DrawTrap,
    PORTRAIT_DrawPattTrap,
    PORTRAIT_DrawColorPattTrap,
    PORTRAIT_DrawLineInt,
    PORTRAIT_DrawBresenhamLine,
    PORTRAIT_DrawStippleLineInt,
    PORTRAIT_DrawBresenhamStippleLine,
    PORTRAIT_DrawEllipse,
    PORTRAIT_ClipEllipse,
    PORTRAIT_PutMonoImageMSBSys,
    NULL,                       // PORTRAIT_PutMonoImageMSBLin,
    NULL,                       // PORTRAIT_PutMonoImageMSBBM,
    NULL,                       // PORTRAIT_PutMonoImageLSBSys,
    NULL,                       // PORTRAIT_PutMonoImageLSBLin,
    NULL,                       // PORTRAIT_PutMonoImageLSBBM,
    PORTRAIT_ClipMonoImageMSBSys,
    NULL,                       // PORTRAIT_ClipMonoImageMSBLin,
    NULL,                       // PORTRAIT_ClipMonoImageMSBBM,
    NULL,                       // PORTRAIT_ClipMonoImageLSBSys,
    NULL,                       // PORTRAIT_ClipMonoImageLSBLin,
    NULL,                       // PORTRAIT_ClipMonoImageLSBBM,
    PORTRAIT_BitBlt,
    NULL,                       // PORTRAIT_BitBltLin,
    PORTRAIT_BitBltSys,
    NULL,                       // PORTRAIT_BitBltBM,
    PORTRAIT_BitBltPatt,
    NULL,                       // PORTRAIT_BitBltPattLin,
    PORTRAIT_BitBltPattSys,
    NULL,                       // PORTRAIT_BitBltPattBM,
    PORTRAIT_BitBltColorPatt,
    NULL,                       // PORTRAIT_BitBltColorPattLin,
    PORTRAIT_BitBltColorPattSys,
    NULL,                       // PORTRAIT_BitBltColorPattBM,
    PORTRAIT_SrcTransBlt,
    NULL,                       // PORTRAIT_SrcTransBltLin,
    PORTRAIT_SrcTransBltSys,
    NULL,                       // PORTRAIT_SrcTransBltBM,
    PORTRAIT_DstTransBlt,
    NULL,                       // PORTRAIT_DstTransBltLin,
    PORTRAIT_DstTransBltSys,
    NULL,                       // PORTRAIT_DstTransBltBM,
    PORTRAIT_StretchBlt,
    NULL,                       // PORTRAIT_StretchBltLin,
    PORTRAIT_StretchBltSys,
    NULL,                       // PORTRAIT_StretchBltBM,
    PORTRAIT_ConvertBltSys_Obsolete,
    NULL,                       // PORTRAIT_ConvertBltBM_Obsolete,
    PORTRAIT_StretchConvertBltSys_Obsolete,
    NULL,                       // PORTRAIT_StretchConvertBltBM_Obsolete,
    NULL,                       // PORTRAIT_BitBltFxTest,
    NULL,                       // PORTRAIT_BitBltFx,
    NULL,                       // PORTRAIT_BitBltFxLin,
    NULL,                       // PORTRAIT_BitBltFxSys,
    NULL,                       // PORTRAIT_BitBltFxBM,
    PORTRAIT_GetBitmapSys,
    NULL,                       // PORTRAIT_GetBitmapBM,
    NULL,                       // PORTRAIT_UpdateScreen,
    NULL,                       // PORTRAIT_DrawClippedLineInt,
    NULL,                       // PORTRAIT_DrawClippedBresenhamLine,
    NULL,                       // PORTRAIT_DrawClippedStippleLineInt,
    NULL,                       // PORTRAIT_DrawClippedBresenhamStippleLine,
    NULL,                       // PORTRAIT_BitBltPlaneMasked,
    NULL,                       // PORTRAIT_BitBltPlaneMaskedLin,
    NULL,                       // PORTRAIT_BitBltPlaneMaskedSys,
    NULL,                       // PORTRAIT_BitBltPlaneMaskedBM,
    NULL,                       // PORTRAIT_DrawRectLin,
    PORTRAIT_DrawRectExt,
    PORTRAIT_DrawStyleLineInt,
    PORTRAIT_DrawBresenhamStyleLine,
    NULL,                       // PORTRAIT_DrawClippedStyleLineInt,
    NULL,                       // PORTRAIT_DrawClippedBresenhamStyleLine,
    };

static GA_2DRenderFuncs PORTRAIT_draw2dStub = {
    sizeof(GA_2DRenderFuncs),
    PTSTUB_GetPixel,
    PTSTUB_PutPixel,
    PTSTUB_DrawScanList,
    PTSTUB_DrawPattScanList,
    PTSTUB_DrawColorPattScanList,
    PTSTUB_DrawEllipseList,
    PTSTUB_DrawPattEllipseList,
    PTSTUB_DrawColorPattEllipseList,
    PTSTUB_DrawFatEllipseList,
    PTSTUB_DrawPattFatEllipseList,
    PTSTUB_DrawColorPattFatEllipseList,
    PTSTUB_DrawRect,
    NULL,
    PTSTUB_DrawPattRect,
    PTSTUB_DrawColorPattRect,
    PTSTUB_DrawTrap,
    PTSTUB_DrawPattTrap,
    PTSTUB_DrawColorPattTrap,
    PTSTUB_DrawLineInt,
    PTSTUB_DrawBresenhamLine,
    PTSTUB_DrawStippleLineInt,
    PTSTUB_DrawBresenhamStippleLine,
    PTSTUB_DrawEllipse,
    PTSTUB_ClipEllipse,
    PTSTUB_PutMonoImageMSBSys,
    NULL,                       // PTSTUB_PutMonoImageMSBLin,
    NULL,                       // PTSTUB_PutMonoImageMSBBM,
    NULL,                       // PTSTUB_PutMonoImageLSBSys,
    NULL,                       // PTSTUB_PutMonoImageLSBLin,
    NULL,                       // PTSTUB_PutMonoImageLSBBM,
    PTSTUB_ClipMonoImageMSBSys,
    NULL,                       // PTSTUB_ClipMonoImageMSBLin,
    NULL,                       // PTSTUB_ClipMonoImageMSBBM,
    NULL,                       // PTSTUB_ClipMonoImageLSBSys,
    NULL,                       // PTSTUB_ClipMonoImageLSBLin,
    NULL,                       // PTSTUB_ClipMonoImageLSBBM,
    PTSTUB_BitBlt,
    NULL,                       // PTSTUB_BitBltLin,
    PTSTUB_BitBltSys,
    NULL,                       // PTSTUB_BitBltBM,
    PTSTUB_BitBltPatt,
    NULL,                       // PTSTUB_BitBltPattLin,
    PTSTUB_BitBltPattSys,
    NULL,                       // PTSTUB_BitBltPattBM,
    PTSTUB_BitBltColorPatt,
    NULL,                       // PTSTUB_BitBltColorPattLin,
    PTSTUB_BitBltColorPattSys,
    NULL,                       // PTSTUB_BitBltColorPattBM,
    PTSTUB_SrcTransBlt,
    NULL,                       // PTSTUB_SrcTransBltLin,
    PTSTUB_SrcTransBltSys,
    NULL,                       // PTSTUB_SrcTransBltBM,
    PTSTUB_DstTransBlt,
    NULL,                       // PTSTUB_DstTransBltLin,
    PTSTUB_DstTransBltSys,
    NULL,                       // PTSTUB_DstTransBltBM,
    PTSTUB_StretchBlt,
    NULL,                       // PTSTUB_StretchBltLin,
    PTSTUB_StretchBltSys,
    NULL,                       // PTSTUB_StretchBltBM,
    PTSTUB_ConvertBltSys_Obsolete,
    NULL,                       // PTSTUB_ConvertBltBM_Obsolete,
    PTSTUB_StretchConvertBltSys_Obsolete,
    NULL,                       // PTSTUB_StretchConvertBltBM_Obsolete,
    NULL,                       // PTSTUB_BitBltFxTest,
    NULL,                       // PTSTUB_BitBltFx,
    NULL,                       // PTSTUB_BitBltFxLin,
    NULL,                       // PTSTUB_BitBltFxSys,
    NULL,                       // PTSTUB_BitBltFxBM,
    PTSTUB_GetBitmapSys,
    NULL,                       // PTSTUB_GetBitmapBM,
    NULL,                       // PTSTUB_UpdateScreen,
    NULL,                       // PTSTUB_DrawClippedLineInt,
    NULL,                       // PTSTUB_DrawClippedBresenhamLine,
    NULL,                       // PTSTUB_DrawClippedStippleLineInt,
    NULL,                       // PTSTUB_DrawClippedBresenhamStippleLine,
    NULL,                       // PTSTUB_BitBltPlaneMasked,
    NULL,                       // PTSTUB_BitBltPlaneMaskedLin,
    NULL,                       // PTSTUB_BitBltPlaneMaskedSys,
    NULL,                       // PTSTUB_BitBltPlaneMaskedBM,
    NULL,                       // PTSTUB_DrawRectLin,
    PTSTUB_DrawRectExt,
    PTSTUB_DrawStyleLineInt,
    PTSTUB_DrawBresenhamStyleLine,
    NULL,                       // PTSTUB_DrawClippedStyleLineInt,
    NULL,                       // PTSTUB_DrawClippedBresenhamStyleLine,
    };

GA_initFuncs        PORTRAIT_initREF;
GA_cursorFuncs      PORTRAIT_cursorREF;
GA_2DStateFuncs     PORTRAIT_state2dREF;
GA_2DRenderFuncs    PORTRAIT_draw2dREF;
GA_2DRenderFuncs    _VARAPI PORTRAIT_draw2dCnt;

#ifdef  __DRIVER__
PM_imports _VARAPI _PM_imports;
#endif

/* Declare a table for performing 90 degree bitmap rotations */

#define genTable(n)                                                         \
    {   0x00000000UL<<n,0x00000001UL<<n,0x00000100UL<<n,0x00000101UL<<n,    \
        0x00010000UL<<n,0x00010001UL<<n,0x00010100UL<<n,0x00010101UL<<n,    \
        0x01000000UL<<n,0x01000001UL<<n,0x01000100UL<<n,0x01000101UL<<n,    \
        0x01010000UL<<n,0x01010001UL<<n,0x01010100UL<<n,0x01010101UL<<n,    \
    }

static ulong rotTable[8][16] = {
    genTable(0),
    genTable(1),
    genTable(2),
    genTable(3),
    genTable(4),
    genTable(5),
    genTable(6),
    genTable(7),
    };

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
REMARKS:
Function to initialise the portrait display filter driver.
****************************************************************************/
static ibool PORTRAIT_init(
    GA_devCtx *hwCtx,
    GA_globalOptions *gOpt)
{
    /* Initialise the internal state of the filter driver */
    DC.hwCtx = hwCtx;
    DC.gOpt = gOpt;
    DC.enabled = false;
    PORTRAIT_initREF.dwSize = sizeof(PORTRAIT_initREF);
    if (!hwCtx->loader.QueryFunctions(GA_GET_INITFUNCS,-2,&PORTRAIT_initREF))
        return false;

    // Hook the device context QueryFunctions function so we can return
    // our hooked functions.
    DC.QueryFunctionsDC = hwCtx->loader.QueryFunctions;
    hwCtx->loader.QueryFunctions = PORTRAIT_QueryFunctionsDC;

    // Obtain a copy of the options structure so we can determine if we
    // are enabled or not.
    DC.opt.dwSize = sizeof(DC.opt);
    PORTRAIT_initREF.GetOptions(&DC.opt);

    // Initialise all the filter driver vectors
    INIT_INIT_VECTORS;
    return true;
}

#ifdef  __DRIVER__
/****************************************************************************
REMARKS:
Fatal error handler called when a non-imported function is called by the
driver. We leave this to a runtime error so that older applications and
shell drivers will work with newer bpd drivers provided no newer functions
are required by the driver itself. If they are, the application or shell
driver needs to be recompiled.
****************************************************************************/
static void _PM_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported PM_imports import function called! Please re-compile!\n");
}

/****************************************************************************
REMARKS:
Function entry point for the reference rasteriser library.
****************************************************************************/
REF2D_driver * _CEXPORT PORTRAIT_initLibrary(
    GA_devCtx *hwCtx,
    GA_globalOptions *gOpt,
    PM_imports *pmImp)
{
    int     i,max;
    ulong   *p;

    // Initialize all default imports to point to fatal error handler
    // for upwards compatibility.
    max = sizeof(_PM_imports)/sizeof(GA_initLibrary_t);
    for (i = 0,p = (ulong*)&_PM_imports; i < max; i++)
        *p++ = (ulong)_PM_fatalErrorHandler;
    memcpy(&_PM_imports,pmImp,MIN(sizeof(_PM_imports),pmImp->dwSize));
    if (!PORTRAIT_init(hwCtx,gOpt))
        return NULL;
    return &DRV;
}
#else
/****************************************************************************
REMARKS:
Function entry point to fake loading of a driver when linked into app.
****************************************************************************/
ibool NAPI PORTRAIT_loadDriver(
    GA_devCtx *hwCtx,
    GA_globalOptions *gOpt,
    REF2D_driver **drv,
    PE_MODULE **hModule,
    ulong *size)
{
    if (!PORTRAIT_init(hwCtx,gOpt))
        return false;
    *drv = &DRV;
    *hModule = NULL;
    *size = 0;
    return true;
}
#endif

/****************************************************************************
REMARKS:
Enabled or disabled the invert driver internally.
****************************************************************************/
static void PortraitEnable(void)
{
    if (DC.enabled)
        PORTRAIT_draw2dCnt = PORTRAIT_draw2dPT;
    else
        PORTRAIT_draw2dCnt = PORTRAIT_draw2dREF;
}

/****************************************************************************
REMARKS:
Function to initialise the INVERT display filter driver. This function
gets called long after the driver has actually been loaded, at the
time when the reference rasteriser module is loaded and initialsied (ie:
at mode init time).
****************************************************************************/
ibool NAPI PORTRAIT_InitFilter(
    GA_modeInfo *modeInfo,
    N_int32 transferStart,
    struct _REF2D_driver *ref2d,
    struct _REF2D_driver **drv)
{
    /* Initialise the internal state of the filter driver */
    DC.mi = *modeInfo;
    DC.ref2d = ref2d;
    DC.refState = ref2d->DriverState;
    DC.enabled = IS_PORTRAIT();
    PORTRAIT_initREF.dwSize = sizeof(PORTRAIT_initREF);
    if (!ref2d->QueryFunctions(GA_GET_INITFUNCS,&PORTRAIT_initREF))
        return false;
    PORTRAIT_cursorREF.dwSize = sizeof(PORTRAIT_cursorREF);
    if (!ref2d->QueryFunctions(GA_GET_CURSORFUNCS,&PORTRAIT_cursorREF))
        return false;
    PORTRAIT_state2dREF.dwSize = sizeof(PORTRAIT_state2dREF);
    if (!ref2d->QueryFunctions(GA_GET_2DSTATEFUNCS,&PORTRAIT_state2dREF))
        return false;
    PORTRAIT_draw2dREF.dwSize = sizeof(PORTRAIT_draw2dREF);
    if (!ref2d->QueryFunctions(GA_GET_2DRENDERFUNCS,&PORTRAIT_draw2dREF))
        return false;
    switch (DC.mi.BitsPerPixel) {
        case 8:     DC.bytesPerPixel = 1;   break;
        case 15:

        case 16:    DC.bytesPerPixel = 2;   break;
        case 24:    DC.bytesPerPixel = 3;   break;
        case 32:    DC.bytesPerPixel = 4;   break;
        }

    /* Allocate memory for the rotation buffer big enough to hold a color
     * bitmap the size of the display screen for rotations.
     */
    DC.bufferSize = DC.mi.XResolution * DC.mi.BytesPerScanLine;
    if ((DC.sysBuffer = malloc(DC.bufferSize)) == NULL)
        return false;
    DC.buffer = DC.sysBuffer;
    DC.transferStart = transferStart;

    /* Initialise all the filter driver vectors */
    INIT_INIT_VECTORS;
    INIT_CURSOR_VECTORS;
    INIT_FILTER_VECTORS;
    PortraitEnable();

    /* Return address of our filter driver as top of filter chain */
    *drv = &DRV;
    return true;
}

/****************************************************************************
REMARKS:
Function to clean up when unloading the driver.
****************************************************************************/
void NAPI PORTRAIT_UnloadDriver(void)
{
}

/****************************************************************************
REMARKS:
Function to clean up when the filter driver needs to de-initialise (ie:
going back to text mode when the ref2d driver is unloaded).
****************************************************************************/
void NAPI PORTRAIT_DeInitFilter(void)
{
    // Free our system memory blit buffer
    free(DC.sysBuffer);

    // Unhook our QueryFunctions hook, and restore the init vectors
//    DC.hwCtx->loader.QueryFunctions = DC.QueryFunctionsDC;
    PORTRAIT_initREF.dwSize = sizeof(PORTRAIT_initREF);
    DC.hwCtx->loader.QueryFunctions(GA_GET_INITFUNCS,-2,&PORTRAIT_initREF);
    INIT_INIT_VECTORS;
}

/****************************************************************************
PARAMETERS:
funcs       - Pointer to the function list to make valid
size        - Size of the function list in bytes

RETURNS:
True if valid, false if not.
****************************************************************************/
static ibool MakeValid(
    ulong *funcs,
    int size)
{
    int i,empty = true;

    for (i = 0; i < size/4; i++) {
        if (funcs[i] != 0)
            empty = false;
        }
    return !empty;
}

/****************************************************************************
DESCRIPTION:
Returns the device driver entry points for the specific function group.

PARAMETERS:
id          - Identifier for the function group to get pointers for
safetyLevel - Safety level being requested.
funcs       - Place to store the function block

RETURNS:
True if functions available, false if not.

REMARKS:
This function allows the filter driver to hook the device driver functions
that are called before any modes are initialised. This is used to hook
functions such as GetVideoModeInfo and SetVideoMode as necessary in the
filter driver.

NOTE: If we get a safetyLevel value of less than zero, it indicates that
      we a being called by some other filter driver and should simply
      pass the call onto the low level hardware driver. Otherwise we may
      end up getting into an infinite loop!
****************************************************************************/
ibool NAPI PORTRAIT_QueryFunctionsDC(
    N_uint32 id,
    N_int32 safetyLevel,
    void *funcs)
{
    int     i;
    ulong   dwSize = *((ulong*)funcs);
    ulong   *p,*q;

    memset(funcs,0,dwSize);
    if (safetyLevel >= 0) {
        switch (id) {
            case GA_GET_INITFUNCS:      RETURN_VALID(PORTRAIT_initPT);
            case GA_GET_2DRENDERFUNCS:
                /* We need to disable driver functions that this filter cannot
                 * support, such as the linear drawing functions.
                 */
                *((ulong*)funcs) = dwSize;
                if (!DC.QueryFunctionsDC(id,safetyLevel,funcs))
                    return false;
                p = funcs;
                q = (void*)&PORTRAIT_draw2dPT;
                for (i = (dwSize/sizeof(void*))-1; i > 0; i--) {
                    if (q[i] == 0)
                        p[i] = 0;
                    }
                return true;
            }
        }
    *((ulong*)funcs) = dwSize;
    return DC.QueryFunctionsDC(id,safetyLevel,funcs);
}

/****************************************************************************
DESCRIPTION:
Returns the reference rasterisers for the specified function group.

PARAMETERS:
id      - Identifier for the function group to get pointers for
funcs   - Place to store the function block

RETURNS:
True if functions available, false if not.

REMARKS:
This function is the main function that is used by the graphics application
code to get a block of function pointers for a specified function group.
The function groups are defined GA_funcGroupsType enumeration, and breaks
up the functions in the device driver API into groups of logically similar
functions.
****************************************************************************/
ibool NAPI PORTRAIT_QueryFunctions(
    N_uint32 id,
    void *funcs)
{
    ulong   dwSize = *((ulong*)funcs);

    memset(funcs,0,dwSize);
    switch (id) {
        case GA_GET_INITFUNCS:      RETURN_VALID(PORTRAIT_initPT);
        case GA_GET_CURSORFUNCS:    RETURN_VALID(PORTRAIT_cursorPT);
        case GA_GET_2DSTATEFUNCS:   RETURN_VALID(PORTRAIT_state2dPT);
        case GA_GET_2DRENDERFUNCS:  RETURN_VALID(PORTRAIT_draw2dStub);
        }
    *((ulong*)funcs) = dwSize;
    return DC.ref2d->QueryFunctions(id,funcs);
}

/****************************************************************************
PARAMETERS:
drawBuf             - Buffer to make the active drawing buffer

RETURNS:
True on success, false on failure.

REMARKS:
This function rotates the draw buffer such that it will work correctly
for the portrait display mode when doing page flipping. We do not support
rendering to offscreen buffers unless the code above is aware that it
needs to do some translation of coordinates.
****************************************************************************/
static ibool RotateDrawBuffer(void)
{
    /* We dont support any draw buffers except full sized pages which allows
     * us to support page flipping in hardware in portrait modes.
     */
    if (DC.buf.Width != DC.mi.XResolution || DC.buf.Height != DC.mi.YResolution)
        return false;
    SWAP(DC.buf.Width,DC.buf.Height);
    DC.buf.Stride = DC.mi.BytesPerScanLine;
    DC.maxY = DC.buf.Height-1;
    DC.maxYFX = N_TOFIX(DC.maxY);
    return true;
}

/****************************************************************************
DESCRIPTION:
Sets a display buffer as the active drawing buffer.

PARAMETERS:
drawBuf             - Buffer to make the active drawing buffer
frameBuffer         - Pointer to the start of frameBuffer memory
bitsPerPixel        - Color depth for the buffer
beginDirectAccess   - Direct frameBuffer arbitration function (NULL if none)
endDirectAccess     - Direct frameBuffer arbitration function (NULL if none)

RETURNS:
0 on success, -1 on failure.

REMARKS:
This function allows the application to tell make a display memory
buffer the active rendering buffer for all subsequent rendering commands.
****************************************************************************/
N_int32 NAPI PORTRAIT_SetDrawBufferSW(
    GA_buffer *drawBuf,
    void *frameBuffer,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pf,
    GA_devCtx *hwCtx,
    N_int32 forceSoftwareOnly)
{
    DC.buf = *drawBuf;
    DC.hwCtx = hwCtx;
    DC.enabled = (IS_PORTRAIT() && frameBuffer == DC.hwCtx->LinearMem);
    if (hwCtx && DC.enabled && !RotateDrawBuffer())
        return -1;
    if (DC.enabled) {
        DC.VRAMBufferStart = DC.VRAMBufferStartY = 0;
        if (DC2D.hwDraw.BitBltLin) {
            if ((DC.mi.MaxLinearOffset - DC.transferStart) > DC.bufferSize) {
                /* Buffers are all allocated in video memory and blitted
                 * with the linear blit functions for performance.
                 */
                DC.VRAMBufferStart = DC.transferStart;
                DC.VRAMBufferStartY = 0;
                }
            }
        else if (DC2D.hwDraw.BitBlt) {
            int startY = (DC.transferStart - DC.buf.Offset) / DC.mi.BytesPerScanLine;
            int sizeY = DC.bufferSize / DC.mi.BytesPerScanLine;
            if ((int)(DC.mi.MaxScanLines - startY) > sizeY) {
                /* Buffers are all allocated in video memory and blitted
                 * with the rectangular blit functions.
                 */
                DC.VRAMBufferStart = DC.transferStart;
                DC.VRAMBufferStartY = startY;
                }
            }
        if (DC.VRAMBufferStart)
            DC.buffer = (uchar*)frameBuffer + DC.VRAMBufferStart;
        else
            DC.buffer = DC.sysBuffer;
        }
    PortraitEnable();
    return DC.ref2d->SetDrawBuffer(&DC.buf,frameBuffer,bitsPerPixel,pf,hwCtx,forceSoftwareOnly);
}

/****************************************************************************
PARAMETERS:
drawBuf - New hardware draw buffer to make active

RETURNS:
0 on success, -1 on failure.

REMARKS:
This function changes the hardware draw buffer for both the software
rasteriser and the hardware device. This can also be used to change the
draw buffer for a software only buffer, but is generally only used when
using hardware.
****************************************************************************/
N_int32 NAPI PORTRAIT_SetDrawBuffer(
    GA_buffer *drawBuf)
{
    DC.buf = *drawBuf;
    if (DC.enabled && !RotateDrawBuffer())
        return -1;
    return STATE2D.SetDrawBuffer(&DC.buf);
}

/****************************************************************************
PARAMETERS:
surface         - Pointer to the start of the memory buffer for drawing
xRes            - X resolution for the surface
yRes            - Y resolution for the surface
bytesPerLine    - Scanline width for the memory buffer for drawing
bitsPerPixel    - New color depth for the surface
pf              - New pixel format for the surface

REMARKS:
This function changes the software draw buffer quickly, to allow the
software renderer to render to bitmaps in different locations in memory.
We do in here for speed, since we want to avoid the overhead of setting
the full draw buffer for the reference rasteriser code. If the color
depth or pixel format changes, this function will switch all the internal
rendering vectors to put to the new vectors for the appropriate functions.
This allows the reference rasteriser to work with all color depths active
in a single driver (and for our debug library to work properly).
****************************************************************************/
void NAPI PORTRAIT_SetDrawSurface(
    void *surface,
    N_int32 xRes,
    N_int32 yRes,
    N_int32 bytesPerLine,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pf)
{
    // Determine if we need to disable the portrait features
    DC.enabled = (IS_PORTRAIT() && DC.refState->haveHWDraw && surface == (DC.refState->frameBuffer + DC.refState->drawOffset));
    DC.buf.Width = xRes;
    DC.buf.Height = yRes;
    DC.buf.Stride = bytesPerLine;
    if (DC.enabled)
        RotateDrawBuffer();
    DC.ref2d->SetDrawSurface(surface,DC.buf.Width,DC.buf.Height,DC.buf.Stride,bitsPerPixel,pf);
}

/****************************************************************************
REMARKS:
Handles updating of the internals when the options structures are updated.
This allows inversion mode to be toggled automatically at runtime using
either the lobal device driver or global SNAP options functions.
****************************************************************************/
static void UpdateOptions(void)
{
    DC.enabled = IS_PORTRAIT();
    PortraitEnable();
}

/****************************************************************************
REMARKS:
Handles overriding of the options in the driver. This allows us to know
when the invert option has been disabled or enabled at runtime.
****************************************************************************/
void NAPI PORTRAIT_SetOptions(
    GA_options *options)
{
    PORTRAIT_initREF.SetOptions(options);
    PORTRAIT_initREF.GetOptions(&DC.opt);
    UpdateOptions();
}

/****************************************************************************
REMARKS:
This function gets called when the global options are changed at runtime,
allowing the filter driver to update the internals based on the changed
global options.
****************************************************************************/
void NAPI PORTRAIT_GlobalOptionsUpdated(void)
{
    UpdateOptions();
}

/****************************************************************************
REMARKS:
This returns the address of the currently active draw buffer
****************************************************************************/
GA_buffer * NAPI PORTRAIT_GetDrawBuffer(void)
{
    return DC.ref2d->GetDrawBuffer();
}

/****************************************************************************
PARAMETERS:
left        - Left coordinate of the rectangle to draw
top         - Top coordinate of the rectangle to draw
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
color       - Color to fill the rectangle with
mix         - Mix to use for drawing the rectangle
****************************************************************************/
void NAPI PORTRAIT_DrawRectExtSW(
    void *buffer,
    N_int32 dstPitch,
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    GA_color color,
    N_int32 mix)
{
    DC.ref2d->DrawRectExtSW(buffer, dstPitch, left, top, width, height, color, mix);
}

/****************************************************************************
DESCRIPTION:
Forces software only rendering on or off.
****************************************************************************/
N_int32 NAPI PORTRAIT_ForceSoftwareOnly(
    N_int32 enable)
{
    return DC.ref2d->ForceSoftwareOnly(enable);
}

/****************************************************************************
PARAMETERS:
mask    - New color compare mask

REMARKS:
This function is used to override the internal software rendering color
compare mask used for source and destination transparent blit functions.
****************************************************************************/
N_uint32 NAPI PORTRAIT_SetColorCompareMask(
    N_uint32 mask)
{
    return DC.ref2d->SetColorCompareMask(mask);
}

/****************************************************************************
DESCRIPTION:
Fix up the software rasteriser after a mode switch has occured
****************************************************************************/
void NAPI PORTRAIT_PostSwitchPhysicalResolution(void)
{
    DC.ref2d->PostSwitchPhysicalResolution();
}

/****************************************************************************
PARAMETERS:
xRes            - Physical pixel X resolution of the display mode
yRes            - Physical line Y resolution of the display mode
virtualX        - Virtual X resolution of the display mode
virtualY        - Virtual Y resolution of the display mode
bitsPerPixel    - Color depth of the display mode
modeInfo        - Structure to fill in with the mode information

RETURNS:
0 on success, -1 on failure

REMARKS:
Filter driver hook to enable portrait display modes.
****************************************************************************/
N_int32 NAPI PORTRAIT_GetCustomVideoModeInfo(
    N_int32 xRes,
    N_int32 yRes,
    N_int32 virtualX,
    N_int32 virtualY,
    N_int32 bitsPerPixel,
    GA_modeInfo *modeInfo)
{
    N_int32 retVal;

    if (IS_PORTRAIT_OPT() && bitsPerPixel >= 8) {
        SWAP(xRes,yRes);
        SWAP(virtualX,virtualY);
        }
    if ((retVal = PORTRAIT_initREF.GetCustomVideoModeInfo(xRes,yRes,virtualX,virtualY,bitsPerPixel,modeInfo)) != 0)
        return retVal;
    if (IS_PORTRAIT_OPT() && modeInfo->BitsPerPixel >= 8) {
        SWAP(modeInfo->XResolution,modeInfo->YResolution);
        SWAP(modeInfo->PhysicalXResolution,modeInfo->PhysicalYResolution);
        }
    return 0;
}

/****************************************************************************
PARAMETERS:
mode        - SNAP display mode to get info for
modeInfo    - Structure to fill in with the mode information

RETURNS:
0 on success, -1 on failure

REMARKS:
Filter driver hook to enable portrait display modes.
****************************************************************************/
N_int32 NAPI PORTRAIT_GetVideoModeInfo(
    N_uint32 mode,
    GA_modeInfo *modeInfo)
{
    N_int32 retVal;

    if ((retVal = PORTRAIT_initREF.GetVideoModeInfo(mode,modeInfo)) != 0)
        return retVal;
    if (IS_PORTRAIT_OPT() && modeInfo->BitsPerPixel >= 8) {
        SWAP(modeInfo->XResolution,modeInfo->YResolution);
        SWAP(modeInfo->PhysicalXResolution,modeInfo->PhysicalYResolution);
        }
    else if (IS_PORTRAIT_OPT() && modeInfo->BitsPerPixel == 4)
        return -1;   //get rid of 4bpp modes when running portrait
    return 0;
}

/****************************************************************************
PARAMETERS:
modeInfo    - Structure to fill in with the mode information

REMARKS:
Filter driver hook to enable portrait display modes.
****************************************************************************/
void NAPI PORTRAIT_GetCurrentVideoModeInfo(
    GA_modeInfo *modeInfo)
{
    PORTRAIT_initREF.GetCurrentVideoModeInfo(modeInfo);
    if (IS_PORTRAIT_OPT() && modeInfo->BitsPerPixel >= 8) {
        SWAP(modeInfo->XResolution,modeInfo->YResolution);
        SWAP(modeInfo->PhysicalXResolution,modeInfo->PhysicalYResolution);
        }
}

/****************************************************************************
DESCRIPTION:
Sets a custom display mode.

PARAMETERS:
xRes            - X resolution for custom display mode
yRes            - Y resolution for custom display mode
bitsPerPixel    - Color depth for custom display mode (0 = text mode)
virtualX        - Requested virtual display X resolution (-1 to use default)
virtualY        - Returns the addressable display memory Y resolution
bytesPerLine    - Returns the scanline width for the mode
maxMem          - Returns the maximum addressable display memory limit
crtc            - CRTC information block containing CRTC timings (required)

RETURNS:
0 on success, -1 on failure

REMARKS:
This function is used to initialize a custom display mode using the
specified X and Y resolution, color depth and CRTC timings.
****************************************************************************/
N_int32 NAPI PORTRAIT_SetCustomVideoMode(
    N_int32 xRes,
    N_int32 yRes,
    N_int32 bitsPerPixel,
    N_uint32 flags,
    N_int32 *virtualX,
    N_int32 *virtualY,
    N_int32 *bytesPerLine,
    N_int32 *maxMem,
    GA_CRTCInfo *crtc)
{
    if (IS_PORTRAIT_OPT() && bitsPerPixel >= 8) {
        // Dial-a-Mode is not supported in portrait modes
        return -1;
        }
    return PORTRAIT_initREF.SetCustomVideoMode(xRes,yRes,bitsPerPixel,flags,virtualX,virtualY,bytesPerLine,maxMem,crtc);
}

/****************************************************************************
DESCRIPTION:
Sets a specified display mode.

PARAMETERS:
mode            - Mode number to set, including flags
virtualX        - Requesyed virtual display X resolution (-1 to use default)
virtualY        - Returns the addressable display memory Y resolution
bytesPerLine    - Returns the scanline width for the mode
maxMem          - Returns the maximum addressable display memory limit
crtc            - CRTC information block if afRefreshCtrl specified

RETURNS:
0 on success, -1 on failure

REMARKS:
This function is used to initialize a specified display mode.
****************************************************************************/
N_int32 NAPI PORTRAIT_SetVideoMode(
    N_uint32 mode,
    N_int32 *virtualX,
    N_int32 *virtualY,
    N_int32 *bytesPerLine,
    N_int32 *maxMem,
    N_int32 refreshRate,
    GA_CRTCInfo *crtc)
{
    N_int32     retVal;
    GA_modeInfo mi;

    mi.dwSize = sizeof(mi);
    if ((retVal = PORTRAIT_initREF.GetVideoModeInfo(mode,&mi)) != 0)
        return retVal;
    if (IS_PORTRAIT_OPT() && mi.BitsPerPixel >= 8) {
        // Portrait does not support virtual display modes or generic
        // refresh rate control, so filter this out.
        if (*bytesPerLine != -1 || *virtualX != -1 || (mode & gaRefreshCtrl))
            return -1;
        }
    if ((retVal = PORTRAIT_initREF.SetVideoMode(mode,virtualX,virtualY,bytesPerLine,maxMem,refreshRate,crtc)) != 0)
        return retVal;
    if (IS_PORTRAIT_OPT() && mi.BitsPerPixel >= 8)
        SWAP(*virtualX,*virtualY);
    return 0;
}

/* Define macros to extract part of the 32 bit result from the tables */

#define extract(n,srcStep)          \
    lo |= rotTable[n][*src & 0xF];  \
    hi |= rotTable[n][*src >> 4];   \
    src += srcStep;                 \
    (void)src;

#define extractPartial(n,srcStep)   \
    if (--height < 0) goto Done;    \
    extract(n,srcStep)

/* Define macros to unpack 32 bits into the output bytes    */

#define unpackAClockwise(w,dstStep)             \
    *dst = w & 0xFF;            dst += dstStep; \
    *dst = (w >> 8) & 0xFF;     dst += dstStep; \
    *dst = (w >> 16) & 0xFF;    dst += dstStep; \
    *dst = (w >> 24) & 0xFF;    dst += dstStep; \
    (void)dst;

/****************************************************************************
PARAMETERS:
dst     - Starting address is detination bitmap
dstStep - Difference in bytes between rows in dest bitmap
src     - Starting address is source bitmap
srcStep - Difference in bytes between rows in source bitmap

REMARKS:
Rotates a full 8x8 bitmap tile anti-clockwise by using table lookup. The bit
extraction is done a nybble at a time to reduce the table sizes.

Original idea was presented in Graphics Gems II.
****************************************************************************/
static void RotateFull8x8AClockwise(
    uchar *dst,
    int dstStep,
    uchar *src,
    int srcStep)
{
    ulong   lo,hi;

// TODO: We have assembler versions of these that might be faster...
//       We should code those up for Watcom C++ specifically with inline
//       assembler and then the functions will be really fast...

    lo = hi = 0;
    extract(7,srcStep); extract(6,srcStep);
    extract(5,srcStep); extract(4,srcStep);
    extract(3,srcStep); extract(2,srcStep);
    extract(1,srcStep); extract(0,srcStep);
    unpackAClockwise(lo,dstStep);
    unpackAClockwise(hi,dstStep);
}

/****************************************************************************
PARAMETERS:
dst     - Starting address is detination bitmap
dstStep - Difference in bytes between rows in dest bitmap
src     - Starting address is source bitmap
srcStep - Difference in bytes between rows in source bitmap

REMARKS:
Rotates a partial 8x8 bitmap tile anti-clockwise by using table lookup. The
bit extraction is done a nybble at a time to reduce the table sizes. In this
version we check to ensure that we do not go past the end of the height of
the source bitmap.

Original idea was presented in Graphics Gems II.
****************************************************************************/
static void RotatePartial8x8AClockwise(
    uchar *dst,
    int dstStep,
    uchar *src,
    int srcStep,
    int height)
{
    ulong   lo,hi;

// TODO: We have assembler versions of these that might be faster...
//       We should code those up for Watcom C++ specifically with inline
//       assembler and then the functions will be really fast...

    lo = hi = 0;
    extractPartial(7,srcStep);  extractPartial(6,srcStep);
    extractPartial(5,srcStep);  extractPartial(4,srcStep);
    extractPartial(3,srcStep);  extractPartial(2,srcStep);
    extractPartial(1,srcStep);  extractPartial(0,srcStep);

Done:
    unpackAClockwise(lo,dstStep);
    unpackAClockwise(hi,dstStep);
}

/****************************************************************************
PARAMETERS:
dst         - Destination mono bitmap buffer
src         - Source mono bitmap buffer
byteWidth   - Width of the source mono bitmap in bytes
height      - Height of the source mono bitmap in scanlines

REMARKS:
This function rotates a mono bitmap image, and stores the value in the
destination buffer. The source buffer is not modified, and the rotated
bitmap may possibly be larger than the source bitmap. The resulting width and
height of the destination bitmap is returned.
****************************************************************************/
static void RotateBitmap1BPP(
    uchar *dst,
    uchar *src,
    N_int32 *pitch,
    N_int32 width,
    N_int32 height)
{
    int i,j,byteWidth,oldWidth,newWidth,maxFull,partialHeight;

    byteWidth       = *pitch;
    oldWidth        = (width + 7) >> 3;
    newWidth        = ROUNDUP((height + 7) >> 3,4);
    maxFull         = height >> 3;
    partialHeight   = height & 0x7;

    for (j = 0; j < oldWidth; j++) {
        for (i = 0; i < maxFull; i++) {
            /* Rotate each full 8x8 bitmap tile */
            RotateFull8x8AClockwise(
                dst + (((oldWidth-1-j) << 3) * newWidth) + i,
                newWidth,
                src + ((i * byteWidth) << 3) + j,
                byteWidth);
            }
        /* Now rotate the last partial 8x8 bitmap tile */
        if (partialHeight) {
            RotatePartial8x8AClockwise(
                dst + (((oldWidth-1-j) << 3) * newWidth) + i,
                newWidth,
                src + ((i * byteWidth) << 3) + j,
                byteWidth, partialHeight);
            }
        }
    *pitch = newWidth;
}

/****************************************************************************
PARAMETERS:
dst     - Destination bitmap buffer
src     - Source bitmap buffer
pitch   - Source bitmap pitch in bytes
width   - Width of the bitmap in pixels
height  - Height of the bitmap in scanlines

REMARKS:
This function rotates an 4BPP bitmap image, and stores the value in the
destination buffer. The source buffer is not modified.
****************************************************************************/
static void RotateBitmap4BPP(
    uchar *dst,
    uchar *src,
    N_int32 srcPitch,
    N_int32 dstPitch,
    N_int32 width,
    N_int32 height,
    N_int32 flipY)
{
    int y,yFull,count;

    BINC(dst,dstPitch * (width-1));
    yFull = height & ~1;
    width = (width+1) / 2;
    if (flipY) {
        BINC(src,srcPitch * (height-1));
        srcPitch = -srcPitch;
        }
    for (y = 0; y < yFull; y += 2) {
        uchar *psrc = src;
        uchar *pdst = dst + y/2;
        count = width;
        while (count--) {
            uchar byte0 = psrc[0];
            uchar byte1 = psrc[srcPitch];
            pdst[0] = (byte0 & 0xF0) | ((byte1 & 0xF0) >> 4);
            pdst[-dstPitch] = ((byte0 & 0x0F) << 4) | (byte1 & 0x0F);
            BDEC(pdst,dstPitch*2);
            BINC(psrc,1);
            }
        BINC(src,srcPitch*2);
        }

    /* Handle last odd scanline if there is one */
    if (height & 1) {
        uchar *psrc = src;
        uchar *pdst = dst + height/2 + 1;
        count = width;
        while (count--) {
            uchar byte0 = psrc[0];
            pdst[0] = byte0 & 0xF0;
            pdst[-dstPitch] = (byte0 & 0x0F) << 4;
            BDEC(pdst,dstPitch*2);
            BINC(psrc,1);
            }
        }
}

/****************************************************************************
PARAMETERS:
dst     - Destination bitmap buffer
src     - Source bitmap buffer
pitch   - Source bitmap stride in bytes
width   - Width of the source bitmap in pixels
height  - Height of the source bitmap in scanlines

REMARKS:
This function rotates an 8bpp bitmap image, and stores the value in the
destination buffer. The source buffer is not modified.
****************************************************************************/
static void RotateBitmap8BPP(
    uchar *dst,
    uchar *src,
    N_int32 srcPitch,
    N_int32 dstPitch,
    N_int32 width,
    N_int32 height,
    N_int32 flipY)
{
    int y,count,fullWidth,partialWidth;

    fullWidth = width / 8;
    partialWidth = width % 8;
    BINC(dst,dstPitch * (width-1));
    if (flipY) {
        BINC(src,srcPitch * (height-1));
        srcPitch = -srcPitch;
        }
    for (y = 0; y < height; y++) {
        uchar *psrc = src;
        uchar *pdst = dst + y;
        count = fullWidth;
        while (count--) {
            *pdst = psrc[0];    BDEC(pdst,dstPitch);
            *pdst = psrc[1];    BDEC(pdst,dstPitch);
            *pdst = psrc[2];    BDEC(pdst,dstPitch);
            *pdst = psrc[3];    BDEC(pdst,dstPitch);
            *pdst = psrc[4];    BDEC(pdst,dstPitch);
            *pdst = psrc[5];    BDEC(pdst,dstPitch);
            *pdst = psrc[6];    BDEC(pdst,dstPitch);
            *pdst = psrc[7];    BDEC(pdst,dstPitch);
            psrc += 8;
            }
        count = partialWidth;
        while (count--) {
            *pdst = *psrc++;
            BDEC(pdst,dstPitch);
            }
        BINC(src,srcPitch);
        }
}

/****************************************************************************
PARAMETERS:
dst     - Destination bitmap buffer
src     - Source bitmap buffer
pitch   - Source bitmap stride in bytes
width   - Width of the source bitmap in pixels
height  - Height of the source bitmap in scanlines

REMARKS:
This function rotates an 8bpp bitmap image, and stores the value in the
destination buffer. The source buffer is not modified. For blitting from
rotated buffer into non-rotated buffer.
****************************************************************************/
static void RotateBitmap8BPPReverse(
    uchar *dst,
    uchar *src,
    N_int32 srcPitch,
    N_int32 dstPitch,
    N_int32 width,
    N_int32 height,
    N_int32 flipY)
{
    int y,count,fullWidth,partialWidth;

    fullWidth = width / 8;
    partialWidth = width % 8;
    if (flipY) {
        BINC(src, srcPitch * (height-1));
        srcPitch = -srcPitch;
        }
    for (y = height - 1; y >= 0; y--) {
        uchar *psrc = src;
        uchar *pdst = dst + y;
        count = fullWidth;
        while (count--) {
            *pdst = psrc[0];    BINC(pdst,dstPitch);
            *pdst = psrc[1];    BINC(pdst,dstPitch);
            *pdst = psrc[2];    BINC(pdst,dstPitch);
            *pdst = psrc[3];    BINC(pdst,dstPitch);
            *pdst = psrc[4];    BINC(pdst,dstPitch);
            *pdst = psrc[5];    BINC(pdst,dstPitch);
            *pdst = psrc[6];    BINC(pdst,dstPitch);
            *pdst = psrc[7];    BINC(pdst,dstPitch);
            psrc += 8;
            }
        count = partialWidth;
        while (count--) {
            *pdst = *psrc++;
            BINC(pdst,dstPitch);
            }
        BINC(src,srcPitch);
        }
}

/****************************************************************************
PARAMETERS:
dst     - Destination bitmap buffer
src     - Source bitmap buffer
pitch   - Source bitmap pitch in bytes
width   - Width of the bitmap in pixels
height  - Height of the bitmap in scanlines

REMARKS:
This function rotates an 16BPP bitmap image, and stores the value in the
destination buffer. The source buffer is not modified.
****************************************************************************/
static void RotateBitmap16BPP(
    ushort *dst,
    ushort *src,
    N_int32 srcPitch,
    N_int32 dstPitch,
    N_int32 width,
    N_int32 height,
    N_int32 flipY)
{
    int y,count,fullWidth,partialWidth;

    fullWidth = width / 8;
    partialWidth = width % 8;
    BINC(dst,dstPitch * (width-1));
    if (flipY) {
        BINC(src,srcPitch * (height-1));
        srcPitch = -srcPitch;
        }
    for (y = 0; y < height; y++) {
        ushort *psrc = src;
        ushort *pdst = dst + y;
        count = fullWidth;
        while (count--) {
            *pdst = psrc[0];    BDEC(pdst,dstPitch);
            *pdst = psrc[1];    BDEC(pdst,dstPitch);
            *pdst = psrc[2];    BDEC(pdst,dstPitch);
            *pdst = psrc[3];    BDEC(pdst,dstPitch);
            *pdst = psrc[4];    BDEC(pdst,dstPitch);
            *pdst = psrc[5];    BDEC(pdst,dstPitch);
            *pdst = psrc[6];    BDEC(pdst,dstPitch);
            *pdst = psrc[7];    BDEC(pdst,dstPitch);
            psrc += 8;
            }
        count = partialWidth;
        while (count--) {
            *pdst = *psrc++;
            BDEC(pdst,dstPitch);
            }
        BINC(src,srcPitch);
        }
}

/****************************************************************************
PARAMETERS:
dst     - Destination bitmap buffer
src     - Source bitmap buffer
pitch   - Source bitmap pitch in bytes
width   - Width of the bitmap in pixels
height  - Height of the bitmap in scanlines

REMARKS:
This function rotates an 16BPP bitmap image, and stores the value in the
destination buffer. The source buffer is not modified. For blitting from
rotated buffer into non-rotated buffer.
****************************************************************************/
static void RotateBitmap16BPPReverse(
    ushort *dst,
    ushort *src,
    N_int32 srcPitch,
    N_int32 dstPitch,
    N_int32 width,
    N_int32 height,
    N_int32 flipY)
{
    int y,count,fullWidth,partialWidth;

    fullWidth = width / 8;
    partialWidth = width % 8;
    if (flipY) {
        BINC(src,srcPitch * (height-1));
        srcPitch = -srcPitch;
        }
    for (y = height - 1; y >= 0; y--) {
        ushort *psrc = src;
        ushort *pdst = dst + y;
        count = fullWidth;
        while (count--) {
            *pdst = psrc[0];    BINC(pdst,dstPitch);
            *pdst = psrc[1];    BINC(pdst,dstPitch);
            *pdst = psrc[2];    BINC(pdst,dstPitch);
            *pdst = psrc[3];    BINC(pdst,dstPitch);
            *pdst = psrc[4];    BINC(pdst,dstPitch);
            *pdst = psrc[5];    BINC(pdst,dstPitch);
            *pdst = psrc[6];    BINC(pdst,dstPitch);
            *pdst = psrc[7];    BINC(pdst,dstPitch);
            psrc += 8;
            }
        count = partialWidth;
        while (count--) {
            *pdst = *psrc++;
            BINC(pdst,dstPitch);
            }
        BINC(src,srcPitch);
        }
}

/****************************************************************************
PARAMETERS:
dst     - Destination bitmap buffer
src     - Source bitmap buffer
pitch   - Source bitmap pitch in bytes
width   - Width of the bitmap in pixels
height  - Height of the bitmap in scanlines

REMARKS:
This function rotates an 24BPP bitmap image, and stores the value in the
destination buffer. The source buffer is not modified.
****************************************************************************/
static void RotateBitmap24BPP(
    uchar *dst,
    uchar *src,
    N_int32 srcPitch,
    N_int32 dstPitch,
    N_int32 width,
    N_int32 height,
    N_int32 flipY)
{
    int y,count;

    BINC(dst,dstPitch * (width-1));
    if (flipY) {
        BINC(src,srcPitch * (height-1));
        srcPitch = -srcPitch;
        }
    for (y = 0; y < height; y++) {
        uchar *psrc = src;
        uchar *pdst = dst + y*3;
        count = width;
        while (count--) {
            pdst[0] = *psrc++;
            pdst[1] = *psrc++;
            pdst[2] = *psrc++;
            BDEC(pdst,dstPitch);
            }
        BINC(src,srcPitch);
        }
}

/****************************************************************************
PARAMETERS:
dst     - Destination bitmap buffer
src     - Source bitmap buffer
pitch   - Source bitmap pitch in bytes
width   - Width of the bitmap in pixels
height  - Height of the bitmap in scanlines

REMARKS:
This function rotates an 24BPP bitmap image, and stores the value in the
destination buffer. The source buffer is not modified. For blitting from
rotated buffer into non-rotated buffer.
****************************************************************************/
static void RotateBitmap24BPPReverse(
    uchar *dst,
    uchar *src,
    N_int32 srcPitch,
    N_int32 dstPitch,
    N_int32 width,
    N_int32 height,
    N_int32 flipY)
{
    int y,count;

    if (flipY) {
        BINC(src,srcPitch * (height-1));
        srcPitch = -srcPitch;
        }
    for (y = height - 1; y >= 0; y--) {
        uchar *psrc = src;
        uchar *pdst = dst + y*3;
        count = width;
        while (count--) {
            pdst[0] = *psrc++;
            pdst[1] = *psrc++;
            pdst[2] = *psrc++;
            BINC(pdst,dstPitch);
            }
        BINC(src,srcPitch);
        }
}

/****************************************************************************
PARAMETERS:
dst     - Destination bitmap buffer
src     - Source bitmap buffer
pitch   - Source bitmap pitch in bytes
width   - Width of the bitmap in pixels
height  - Height of the bitmap in scanlines

REMARKS:
This function rotates an 32BPP bitmap image, and stores the value in the
destination buffer. The source buffer is not modified.
****************************************************************************/
static void RotateBitmap32BPP(
    ulong *dst,
    ulong *src,
    N_int32 srcPitch,
    N_int32 dstPitch,
    N_int32 width,
    N_int32 height,
    N_int32 flipY)
{
    int y,count,fullWidth,partialWidth;

    fullWidth = width / 8;
    partialWidth = width % 8;
    BINC(dst,dstPitch * (width-1));
    if (flipY) {
        BINC(src,srcPitch * (height-1));
        srcPitch = -srcPitch;
        }
    for (y = 0; y < height; y++) {
        ulong *psrc = src;
        ulong *pdst = dst + y;
        count = fullWidth;
        while (count--) {
            *pdst = psrc[0];    BDEC(pdst,dstPitch);
            *pdst = psrc[1];    BDEC(pdst,dstPitch);
            *pdst = psrc[2];    BDEC(pdst,dstPitch);
            *pdst = psrc[3];    BDEC(pdst,dstPitch);
            *pdst = psrc[4];    BDEC(pdst,dstPitch);
            *pdst = psrc[5];    BDEC(pdst,dstPitch);
            *pdst = psrc[6];    BDEC(pdst,dstPitch);
            *pdst = psrc[7];    BDEC(pdst,dstPitch);
            psrc += 8;
            }
        count = partialWidth;
        while (count--) {
            *pdst = *psrc++;
            BDEC(pdst,dstPitch);
            }
        BINC(src,srcPitch);
        }
}

/****************************************************************************
PARAMETERS:
dst     - Destination bitmap buffer
src     - Source bitmap buffer
pitch   - Source bitmap pitch in bytes
width   - Width of the bitmap in pixels
height  - Height of the bitmap in scanlines

REMARKS:
This function rotates an 32BPP bitmap image, and stores the value in the
destination buffer. The source buffer is not modified. For blitting from
rotated buffer into non-rotated buffer.
****************************************************************************/
static void RotateBitmap32BPPReverse(
    ulong *dst,
    ulong *src,
    N_int32 srcPitch,
    N_int32 dstPitch,
    N_int32 width,
    N_int32 height,
    N_int32 flipY)
{
    int y,count,fullWidth,partialWidth;

    fullWidth = width / 8;
    partialWidth = width % 8;
    if (flipY) {
        BINC(src,srcPitch * (height-1));
        srcPitch = -srcPitch;
        }
    for (y = height - 1; y >= 0; y--) {
        ulong *psrc = src;
        ulong *pdst = dst + y;
        count = fullWidth;
        while (count--) {
            *pdst = psrc[0];    BINC(pdst,dstPitch);
            *pdst = psrc[1];    BINC(pdst,dstPitch);
            *pdst = psrc[2];    BINC(pdst,dstPitch);
            *pdst = psrc[3];    BINC(pdst,dstPitch);
            *pdst = psrc[4];    BINC(pdst,dstPitch);
            *pdst = psrc[5];    BINC(pdst,dstPitch);
            *pdst = psrc[6];    BINC(pdst,dstPitch);
            *pdst = psrc[7];    BINC(pdst,dstPitch);
            psrc += 8;
            }
        count = partialWidth;
        while (count--) {
            *pdst = *psrc++;
            BINC(pdst,dstPitch);
            }
        BINC(src,srcPitch);
        }
}

/****************************************************************************
PARAMETERS:
src             - Pointer to the source bitmap to rotate
bitsPerPixel    - Pixel depth of the source bitmap in bits
pitch           - Stride of the source bitmap in bytes
width           - Width of the source bitmap
height          - Height of the source bitmap
****************************************************************************/
void * NAPI PORTRAIT_RotateBitmap(
    void *src,
    N_int32 bitsPerPixel,
    N_int32 *pitch,
    N_int32 width,
    N_int32 height)
{
    int dstPitch;

    if (width > DC.mi.XResolution || height > DC.mi.YResolution)
        return NULL;
    switch (bitsPerPixel) {
        case 1:
            RotateBitmap1BPP(DC.sysBuffer,src,pitch,width,height);
            break;
        case 4:
            dstPitch = ROUNDUP((height+1)/2,4);
            RotateBitmap4BPP(DC.sysBuffer,src,*pitch,dstPitch,width,height,false);
            break;
        case 8:
            dstPitch = ROUNDUP(height,4);
            RotateBitmap8BPP(DC.sysBuffer,src,*pitch,dstPitch,width,height,false);
            break;
        case 15:
        case 16:
            dstPitch = ROUNDUP(height*2,4);
            RotateBitmap16BPP(DC.sysBuffer,src,*pitch,dstPitch,width,height,false);
            break;
        case 24:
            dstPitch = ROUNDUP(height*3,4);
            RotateBitmap24BPP(DC.sysBuffer,src,*pitch,dstPitch,width,height,false);
            break;
        case 32:
            dstPitch = ROUNDUP(height*4,4);
            RotateBitmap32BPP(DC.sysBuffer,src,*pitch,dstPitch,width,height,false);
            break;
        default:
            return NULL;
        }
    *pitch = dstPitch;
    return DC.sysBuffer;
}

/****************************************************************************
PARAMETERS:
index   - Index of the pattern to download
pattern - Pointer to pattern data to download
****************************************************************************/
void NAPI PORTRAIT_Set8x8MonoPattern(
    N_int32 index,
    GA_pattern *pattern)
{
    if (DC.enabled) {
        GA_pattern  pat;
        uchar       *dst = pat.p;
        uchar       *src = pattern->p;
        ulong       lo,hi;

        /* Rotate the 8x8 mono pattern before downloading to hardware */
        lo = hi = 0;
        extract(7,1); extract(6,1); extract(5,1); extract(4,1);
        extract(3,1); extract(2,1); extract(1,1); extract(0,1);
        unpackAClockwise(lo,1);
        unpackAClockwise(hi,1);
        STATE2D.Set8x8MonoPattern(index,&pat);
        }
    else
        STATE2D.Set8x8MonoPattern(index,pattern);
}

/****************************************************************************
PARAMETERS:
index   - Index of the pattern to download
pattern - Pointer to pattern data to download
****************************************************************************/
void NAPI PORTRAIT_Set8x8ColorPattern(
    N_int32 index,
    GA_colorPattern *pattern)
{
    if (DC.enabled) {
        int             x,y;
        GA_colorPattern pat;

        switch (DC.mi.BitsPerPixel) {
            case 8: {
                uchar *p = (uchar*)pattern->b8.p;
                for (y = 0; y < 8; y++) {
                    for (x = 0; x < 8; x++)
                        pat.b8.p[x][y] = *p++;
                    }
                } break;
            case 15:
            case 16: {
                ushort *p = (ushort*)pattern->b16.p;
                for (y = 0; y < 8; y++) {
                    for (x = 0; x < 8; x++)
                        pat.b16.p[x][y] = *p++;
                    }
                } break;
            case 24: {
                uchar *p = (uchar*)pattern->b8.p;
                for (y = 0; y < 8; y++) {
                    for (x = 0; x < 8; x++)
                        pat.b24.p[x][y][0] = *p++;
                        pat.b24.p[x][y][1] = *p++;
                        pat.b24.p[x][y][2] = *p++;
                    }
                }  break;
            case 32: {
                ulong *p = (ulong*)pattern->b32.p;
                for (y = 0; y < 8; y++) {
                    for (x = 0; x < 8; x++)
                        pat.b32.p[x][y] = *p++;
                    }
                }  break;
            }
        STATE2D.Set8x8ColorPattern(index,&pat);
        }
    else
        STATE2D.Set8x8ColorPattern(index,pattern);
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
x1      - First X coordinate of scanline
x2      - Second X coordinate of scanline
****************************************************************************/
static void DrawScan(
    N_int32 y,
    N_int32 x1,
    N_int32 x2)
{
    int width;

    if ((width = (x2 - x1)) == 0)
        return;
    if (width < 0) {
        x1 = x2;
        width = -width;
        }
    ROTATE(x1,y,width);
    DRAW2D.DrawRect(x1,y,1,width);
}

/****************************************************************************
PARAMETERS:
y           - Y coordinate for scanline
x1          - First X coordinate of scanline
x2          - Second X coordinate of scanline
****************************************************************************/
static void DrawPattScan(
    N_int32 y,
    N_int32 x1,
    N_int32 x2)
{
    int width;

    if ((width = (x2 - x1)) == 0)
        return;
    if (width < 0) {
        x1 = x2;
        width = -width;
        }
    ROTATE(x1,y,width);
    DRAW2D.DrawPattRect(x1,y,1,width);
}

/****************************************************************************
PARAMETERS:
y           - Y coordinate for scanline
x1          - First X coordinate of scanline
x2          - Second X coordinate of scanline
****************************************************************************/
static void DrawColorPattScan(
    N_int32 y,
    N_int32 x1,
    N_int32 x2)
{
    int width;

    if ((width = (x2 - x1)) == 0)
        return;
    if (width < 0) {
        x1 = x2;
        width = -width;
        }
    ROTATE(x1,y,width);
    DRAW2D.DrawColorPattRect(x1,y,1,width);
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI PORTRAIT_DrawScanList(
    N_int32 y,
    N_int32 length,
    N_int16 *scans)
{
    while (length--) {
        int width;
        int x1 = *scans++;
        int x2 = *scans++;
        int y1 = y;
        if ((width = (x2 - x1)) != 0) {
            if (width < 0) {
                x1 = x2;
                width = -width;
                }
            ROTATE(x1,y1,width);
            DRAW2D.DrawRect(x1,y1,1,width);
            }
        y++;
        }
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI PORTRAIT_DrawPattScanList(
    N_int32 y,
    N_int32 length,
    N_int16 *scans)
{
    while (length--) {
        int width;
        int x1 = *scans++;
        int x2 = *scans++;
        int y1 = y;
        if ((width = (x2 - x1)) != 0) {
            if (width < 0) {
                x1 = x2;
                width = -width;
                }
            ROTATE(x1,y1,width);
            DRAW2D.DrawPattRect(x1,y1,1,width);
            }
        y++;
        }
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI PORTRAIT_DrawColorPattScanList(
    N_int32 y,
    N_int32 length,
    N_int16 *scans)
{
    while (length--) {
        int width;
        int x1 = *scans++;
        int x2 = *scans++;
        int y1 = y;
        if ((width = (x2 - x1)) != 0) {
            if (width < 0) {
                x1 = x2;
                width = -width;
                }
            ROTATE(x1,y1,width);
            DRAW2D.DrawColorPattRect(x1,y1,1,width);
            }
        y++;
        }
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI PORTRAIT_DrawEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int i,j,maxIndex;

    // Render the scanlines in the ellipse
    maxIndex = y+length-1;
    for (i = y,j = y+height; i < maxIndex; i++,j--,scans += 2) {
        DrawScan(i,scans[0],scans[1]);
        DrawScan(j,scans[0],scans[1]);
        }
    if (!(height & 1))
        DrawScan(i,scans[0],scans[1]);
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI PORTRAIT_DrawPattEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int i,j,maxIndex;

    // Render the scanlines in the ellipse
    maxIndex = y+length-1;
    for (i = y,j = y+height; i < maxIndex; i++,j--,scans += 2) {
        DrawPattScan(i,scans[0],scans[1]);
        DrawPattScan(j,scans[0],scans[1]);
        }
    if (!(height & 1))
        DrawPattScan(i,scans[0],scans[1]);
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI PORTRAIT_DrawColorPattEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int i,j,maxIndex;

    // Render the scanlines in the ellipse
    maxIndex = y+length-1;
    for (i = y,j = y+height; i < maxIndex; i++,j--,scans += 2) {
        DrawColorPattScan(i,scans[0],scans[1]);
        DrawColorPattScan(j,scans[0],scans[1]);
        }
    if (!(height & 1))
        DrawColorPattScan(i,scans[0],scans[1]);
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis + pen height adjustment
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI PORTRAIT_DrawFatEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int i,j,maxIndex;

    // Render the scanlines in the ellipse
    maxIndex = y+length;
    for (i = y,j = y+height; i < maxIndex; i++,j--,scans += 4) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            DrawScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            DrawScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            DrawScan(j,scans[I_LEFTL],scans[I_LEFTR]);
            DrawScan(j,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            DrawScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            DrawScan(j,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
    if ((height+1) & 1) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            DrawScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            DrawScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            DrawScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis + pen height adjustment
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI PORTRAIT_DrawPattFatEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int i,j,maxIndex;

    // Render the scanlines in the ellipse
    maxIndex = y+length;
    for (i = y,j = y+height; i < maxIndex; i++,j--,scans += 4) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            DrawPattScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            DrawPattScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            DrawPattScan(j,scans[I_LEFTL],scans[I_LEFTR]);
            DrawPattScan(j,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            DrawPattScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            DrawPattScan(j,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
    if ((height+1) & 1) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            DrawPattScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            DrawPattScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            DrawPattScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis + pen height adjustment
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI PORTRAIT_DrawColorPattFatEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int i,j,maxIndex;

    // Render the scanlines in the ellipse
    maxIndex = y+length;
    for (i = y,j = y+height; i < maxIndex; i++,j--,scans += 4) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            DrawColorPattScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            DrawColorPattScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            DrawColorPattScan(j,scans[I_LEFTL],scans[I_LEFTR]);
            DrawColorPattScan(j,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            DrawColorPattScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            DrawColorPattScan(j,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
    if ((height+1) & 1) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            DrawColorPattScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            DrawColorPattScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            DrawColorPattScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the rectangle to draw
top     - Top coordinate of the rectangle to draw
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
****************************************************************************/
void NAPI PORTRAIT_DrawRect(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height)
{
    ROTATE(left,top,width);
    SWAP(width,height);
    DRAW2D.DrawRect(left,top,width,height);
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the rectangle to draw
top     - Top coordinate of the rectangle to draw
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
color   - Foreground color to use
mix     - Mix to draw with
****************************************************************************/
void NAPI PORTRAIT_DrawRectExt(
    N_int32  left,
    N_int32  top,
    N_int32  width,
    N_int32  height,
    GA_color color,
    N_int32  mix)
{
    ROTATE(left,top,width);
    SWAP(width,height);
    DRAW2D.DrawRectExt(left,top,width,height,color,mix);
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the rectangle to draw
top     - Top coordinate of the rectangle to draw
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
****************************************************************************/
void NAPI PORTRAIT_DrawPattRect(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height)
{
    ROTATE(left,top,width);
    SWAP(width,height);
    DRAW2D.DrawPattRect(left,top,width,height);
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the rectangle to draw
top     - Top coordinate of the rectangle to draw
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
****************************************************************************/
void NAPI PORTRAIT_DrawColorPattRect(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height)
{
    ROTATE(left,top,width);
    SWAP(width,height);
    DRAW2D.DrawColorPattRect(left,top,width,height);
}

/****************************************************************************
PARAMETERS:
trap    - Pointer to the GA_trap structure describing the trapezoid
****************************************************************************/
void NAPI PORTRAIT_DrawTrap(
    GA_trap *trap)
{
    int y,x1,x2,count = trap->count;

    if (!count)
        return;
    y = trap->y;
    x1 = trap->x1;
    x2 = trap->x2;
    trap->y += count;
    while (count--) {
        int width;
        int ix1 = (x1 + 0x8000) >> 16;
        int ix2 = (x2 + 0x8000) >> 16;
        int iy1 = y;
        if ((width = (ix2 - ix1)) != 0) {
            if (width < 0) {
                ix1 = ix2;
                width = -width;
                }
            ROTATE(ix1,iy1,width);
            DRAW2D.DrawRect(ix1,iy1,1,width);
            }
        x1 += trap->slope1;
        x2 += trap->slope2;
        y++;
        }
    trap->x1 = x1;
    trap->x2 = x2;
}

/****************************************************************************
PARAMETERS:
trap    - Pointer to the GA_trap structure describing the trapezoid
****************************************************************************/
void NAPI PORTRAIT_DrawPattTrap(
    GA_trap *trap)
{
    int y,x1,x2,count = trap->count;

    if (!count)
        return;
    y = trap->y;
    x1 = trap->x1;
    x2 = trap->x2;
    trap->y += count;
    while (count--) {
        int width;
        int ix1 = (x1 + 0x8000) >> 16;
        int ix2 = (x2 + 0x8000) >> 16;
        int iy1 = y;
        if ((width = (ix2 - ix1)) != 0) {
            if (width < 0) {
                ix1 = ix2;
                width = -width;
                }
            ROTATE(ix1,iy1,width);
            DRAW2D.DrawPattRect(ix1,iy1,1,width);
            }
        x1 += trap->slope1;
        x2 += trap->slope2;
        y++;
        }
    trap->x1 = x1;
    trap->x2 = x2;
}

/****************************************************************************
PARAMETERS:
trap    - Pointer to the GA_trap structure describing the trapezoid
****************************************************************************/
void NAPI PORTRAIT_DrawColorPattTrap(
    GA_trap *trap)
{
    int y,x1,x2,count = trap->count;

    if (!count)
        return;
    y = trap->y;
    x1 = trap->x1;
    x2 = trap->x2;
    trap->y += count;
    while (count--) {
        int width;
        int ix1 = (x1 + 0x8000) >> 16;
        int ix2 = (x2 + 0x8000) >> 16;
        int iy1 = y;
        if ((width = (ix2 - ix1)) != 0) {
            if (width < 0) {
                ix1 = ix2;
                width = -width;
                }
            ROTATE(ix1,iy1,width);
            DRAW2D.DrawColorPattRect(ix1,iy1,1,width);
            }
        x1 += trap->slope1;
        x2 += trap->slope2;
        y++;
        }
    trap->x1 = x1;
    trap->x2 = x2;
}

/****************************************************************************
PARAMETERS:
x1          - X1 coordinate
y1          - Y1 coordinate
x2          - X2 coordinate
y2          - Y2 coordinate
drawLast    - 1 to draw last pixel, 0 to skip it
****************************************************************************/
void NAPI PORTRAIT_DrawLineInt(
    N_int32 x1,
    N_int32 y1,
    N_int32 x2,
    N_int32 y2,
    N_int32 drawLast)
{
    ROTATE_PT(x1,y1);
    ROTATE_PT(x2,y2);
    DRAW2D.DrawLineInt(x1,y1,x2,y2,drawLast);
}

/****************************************************************************
PARAMETERS:
x1          - X1 coordinate
y1          - Y1 coordinate
x2          - X2 coordinate
y2          - Y2 coordinate
drawLast    - 1 to draw last pixel, 0 to skip it
transparent - 1 if the line is transparent, 0 if opaque
****************************************************************************/
void NAPI PORTRAIT_DrawStippleLineInt(
    N_int32 x1,
    N_int32 y1,
    N_int32 x2,
    N_int32 y2,
    N_int32 drawLast,
    N_int32 transparent)
{
    ROTATE_PT(x1,y1);
    ROTATE_PT(x2,y2);
    DRAW2D.DrawStippleLineInt(x1,y1,x2,y2,drawLast,transparent);
}

/****************************************************************************
PARAMETERS:
x1              - X1 coordinate
y1              - Y1 coordinate
initialError    - Initial error term
majorInc        - Major increment factor for error term
diagInc         - Diagonal increment factor for error term
count           - Number of pixels to draw
flags           - Flags to control drawing
****************************************************************************/
void NAPI PORTRAIT_DrawBresenhamLine(
    N_int32 x1,
    N_int32 y1,
    N_int32 initialError,
    N_int32 majorInc,
    N_int32 diagInc,
    N_int32 count,
    N_int32 flags)
{
    N_int32 newFlags;

    ROTATE_PT(x1,y1);
    newFlags = 0;
    if (flags & gaLineYPositive)
        newFlags |= gaLineXPositive;
    if (!(flags & gaLineXPositive))
        newFlags |= gaLineYPositive;
    if (!(flags & gaLineXMajor))
        newFlags |= gaLineXMajor;
    if (flags & gaLineDoLastPel)
        newFlags |= gaLineDoLastPel;
    DRAW2D.DrawBresenhamLine(x1,y1,initialError,majorInc,diagInc,count,newFlags);
}

/****************************************************************************
PARAMETERS:
x1              - X1 coordinate
y1              - Y1 coordinate
initialError    - Initial error term
majorInc        - Major increment factor for error term
diagInc         - Diagonal increment factor for error term
count           - Number of pixels to draw
flags           - Flags to control drawing
transparent     - 1 if the line is transparent, 0 if opaque
****************************************************************************/
void NAPI PORTRAIT_DrawBresenhamStippleLine(
    N_int32 x1,
    N_int32 y1,
    N_int32 initialError,
    N_int32 majorInc,
    N_int32 diagInc,
    N_int32 count,
    N_int32 flags,
    N_int32 transparent)
{
    N_int32 newFlags;

    ROTATE_PT(x1,y1);
    newFlags = 0;
    if (flags & gaLineYPositive)
        newFlags |= gaLineXPositive;
    if (!(flags & gaLineXPositive))
        newFlags |= gaLineYPositive;
    if (!(flags & gaLineXMajor))
        newFlags |= gaLineXMajor;
    if (flags & gaLineDoLastPel)
        newFlags |= gaLineDoLastPel;
    DRAW2D.DrawBresenhamStippleLine(x1,y1,initialError,majorInc,diagInc,count,newFlags,transparent);
}

/****************************************************************************
PARAMETERS:
x1          - X1 coordinate
y1          - Y1 coordinate
x2          - X2 coordinate
y2          - Y2 coordinate
drawLast    - 1 to draw last pixel, 0 to skip it
transparent - 1 if the line is transparent, 0 if opaque
****************************************************************************/
void NAPI PORTRAIT_DrawStyleLineInt(
    N_int32 x1,
    N_int32 y1,
    N_int32 x2,
    N_int32 y2,
    N_int32 drawLast,
    N_int32 transparent)
{
    ROTATE_PT(x1,y1);
    ROTATE_PT(x2,y2);
    DRAW2D.DrawStyleLineInt(x1,y1,x2,y2,drawLast,transparent);
}

/****************************************************************************
PARAMETERS:
x1              - X1 coordinate
y1              - Y1 coordinate
initialError    - Initial error term
majorInc        - Major increment factor for error term
diagInc         - Diagonal increment factor for error term
count           - Number of pixels to draw
flags           - Flags to control drawing
transparent     - 1 if the line is transparent, 0 if opaque
****************************************************************************/
void NAPI PORTRAIT_DrawBresenhamStyleLine(
    N_int32 x1,
    N_int32 y1,
    N_int32 initialError,
    N_int32 majorInc,
    N_int32 diagInc,
    N_int32 count,
    N_int32 flags,
    N_int32 transparent)
{
    N_int32 newFlags;

    ROTATE_PT(x1,y1);
    newFlags = 0;
    if (flags & gaLineYPositive)
        newFlags |= gaLineXPositive;
    if (!(flags & gaLineXPositive))
        newFlags |= gaLineYPositive;
    if (!(flags & gaLineXMajor))
        newFlags |= gaLineXMajor;
    if (flags & gaLineDoLastPel)
        newFlags |= gaLineDoLastPel;
    DRAW2D.DrawBresenhamStyleLine(x1,y1,initialError,majorInc,diagInc,count,newFlags,transparent);
}

/****************************************************************************
PARAMETERS:
x           - Destination X coordinate to draw the bitmap at
y           - Destination Y coordinate to draw the bitmap at
width       - Width of the bitmap in pixels
height      - Height of the bitmap in pixels
byteWidth   - Width of the bitmap in bytes
image       - Pointer to the bitmap image data to draw
transparent - 1 for transparent, 0 for opaque

REMARKS:
If the transparent parameter is set to 1, where a bit is 0 in the bitmap
image the destination pixel remains untouched. If the transparent
parameter is set to 0, where a bit is 0 in the bitmap image the
destination pixel is drawn in the background color. In all cases where a
bit in the bitmap image is 1, the pixel is drawn in the foreground color.

Note:   This function processes the bitmap data in MSB fashion, in that the
        first pixel drawn corrsponds to bit 7 of the first byte. The second
        pixel is bit 6 of the first byte, ...,the 8th pixel is bit 7 of
        the second byte etc. Both LSB and MSB versions are provided for
        performance.
****************************************************************************/
void NAPI PORTRAIT_PutMonoImageMSBSys(
    N_int32 x,
    N_int32 y,
    N_int32 width,
    N_int32 height,
    N_int32 byteWidth,
    N_uint8 *image,
    N_int32 transparent)
{
    ulong adjust;

    RotateBitmap1BPP(DC.sysBuffer,image,&byteWidth,width,height);
    ROTATE(x,y,width);
    SWAP(width,height);
    adjust = height & 7 ? (8 - (height & 7)) * byteWidth : 0;
    DRAW2D.PutMonoImageMSBSys(x,y,width,height,byteWidth,(uchar *)DC.sysBuffer + adjust,transparent);
}

/****************************************************************************
PARAMETERS:
x           - Destination X coordinate to draw the bitmap at
y           - Destination Y coordinate to draw the bitmap at
width       - Width of the bitmap in pixels
height      - Height of the bitmap in pixels
byteWidth   - Width of the bitmap in bytes
image       - Pointer to the bitmap image data to draw
transparent - 1 for transparent, 0 for opaque
clipLeft    - Left edge of clip rectangle (inclusive)
clipTop     - Top edge of clip rectangle (inclusive)
clipRight   - Right edge of clip rectangle (exclusive)
clipBottom  - Bottom edge of clip rectangle (exclusive)

REMARKS:
If the transparent parameter is set to 1, where a bit is 0 in the bitmap
image the destination pixel remains untouched. If the transparent
parameter is set to 0, where a bit is 0 in the bitmap image the
destination pixel is drawn in the background color. In all cases where a
bit in the bitmap image is 1, the pixel is drawn in the foreground color.

Note:   This function processes the bitmap data in MSB fashion, in that the
        first pixel drawn corrsponds to bit 7 of the first byte. The second
        pixel is bit 6 of the first byte, ...,the 8th pixel is bit 7 of
        the second byte etc. Both LSB and MSB versions are provided for
        performance.
****************************************************************************/
void NAPI PORTRAIT_ClipMonoImageMSBSys(
    N_int32 x,
    N_int32 y,
    N_int32 width,
    N_int32 height,
    N_int32 byteWidth,
    N_uint8 *image,
    N_int32 transparent,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
    int     /*bottom,right,*/clipWidth,clipHeight;
#if 0
    // Check if image needs to be clipped at all
    right = x + width;
    bottom = y + height;
    if ((bottom <= clipTop) || (y >= clipBottom) ||
        (right <= clipLeft) || (x >= clipRight))
        return;

    // Clip the top and bottom of the bitmap image from being drawn
    if (y < clipTop) {
        int skip = (clipTop - y);
        y = clipTop;
        image += skip * byteWidth;
        height -= skip;
        }
    if (bottom > clipBottom)
        height -= (bottom - clipBottom);

    // Now clip the left and right edges so that the clip rectangle lies
    // within the first and last bytes of the image.
    if (x < clipLeft) {
        int skip = (clipLeft - x) / 8;
        image += skip;
        x += skip * 8;
        width -= skip * 8;
        }
    else
        clipLeft = x;
    if (right > clipRight)
        width -= (right - clipRight);

    width += (clipLeft-x);
#endif
    RotateBitmap1BPP(DC.sysBuffer,image,&byteWidth,width,height);
    ROTATE(x,y,width);
    SWAP(width,height);
    clipWidth = clipRight - clipLeft;
    clipHeight = clipBottom - clipTop;
    ROTATE(clipLeft,clipTop,clipWidth);
    SWAP(clipWidth,clipHeight);
    clipRight = clipLeft + clipWidth;
    clipBottom = clipTop + clipHeight;
    DRAW2D.ClipMonoImageMSBSys(x,y,width,height,byteWidth,DC.sysBuffer,
        transparent,clipLeft,clipTop,clipRight,clipBottom);
}

/****************************************************************************
PARAMETERS:
srcLeft - Left coordinate of the source rectangle to copy
srcTop  - Top coordinate of the source rectangle to copy
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
dstLeft - Left coordinate of destination
dstTop  - Top coordinate of destination
mix     - Mix code for the copy (GA_mixCodesType)
****************************************************************************/
void NAPI PORTRAIT_BitBlt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix)
{
    ROTATE(srcLeft,srcTop,width);
    ROTATE(dstLeft,dstTop,width);
    SWAP(width,height);
    DRAW2D.BitBlt(srcLeft,srcTop,width,height,dstLeft,dstTop,mix);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI PORTRAIT_BitBltSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    N_int32 flipY)
{
    int dstPitch;

    ROTATE(dstLeft,dstTop,width);
    SWAP(width,height);
    DC2D.beginDirectAccess();
    switch (DC.mi.BitsPerPixel) {
        case 8:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft;
            if (DC.VRAMBufferStartY)
                dstPitch = DC.mi.BytesPerScanLine;
            else
                dstPitch = ROUNDUP(width,DC.mi.BitmapStridePad);
            RotateBitmap8BPP(DC.buffer,srcAddr,srcPitch,dstPitch,height,width,flipY);
            break;
        case 15:
        case 16:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*2;
            if (DC.VRAMBufferStartY)
                dstPitch = DC.mi.BytesPerScanLine;
            else
                dstPitch = ROUNDUP(width*2,DC.mi.BitmapStridePad);
            RotateBitmap16BPP(DC.buffer,srcAddr,srcPitch,dstPitch,height,width,flipY);
            break;
        case 24:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*3;
            if (DC.VRAMBufferStartY)
                dstPitch = DC.mi.BytesPerScanLine;
            else
                dstPitch = ROUNDUP(width*3,DC.mi.BitmapStridePad);
            RotateBitmap24BPP(DC.buffer,srcAddr,srcPitch,dstPitch,height,width,flipY);
            break;
        case 32:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*4;
            if (DC.VRAMBufferStartY)
                dstPitch = DC.mi.BytesPerScanLine;
            else
                dstPitch = ROUNDUP(width*4,DC.mi.BitmapStridePad);
            RotateBitmap32BPP(DC.buffer,srcAddr,srcPitch,dstPitch,height,width,flipY);
            break;
        }
    DC2D.endDirectAccess();
    if (DC.VRAMBufferStartY)
        DRAW2D.BitBlt(0,DC.VRAMBufferStartY,width,height,dstLeft,dstTop,mix);
    else if (DC.VRAMBufferStart)
        DRAW2D.BitBltLin(DC.VRAMBufferStart,dstPitch,0,0,width,height,dstLeft,dstTop,mix);
    else
        DRAW2D.BitBltSys(DC.buffer,dstPitch,0,0,width,height,dstLeft,dstTop,mix,false);
}

/****************************************************************************
PARAMETERS:
srcLeft - Left coordinate of the source rectangle to copy
srcTop  - Top coordinate of the source rectangle to copy
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
dstLeft - Left coordinate of destination
dstTop  - Top coordinate of destination
rop3    - ROP3 mix code for the copy
****************************************************************************/
void NAPI PORTRAIT_BitBltPatt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    ROTATE(srcLeft,srcTop,width);
    ROTATE(dstLeft,dstTop,width);
    SWAP(width,height);
    DRAW2D.BitBltPatt(srcLeft,srcTop,width,height,dstLeft,dstTop,rop3);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3        - Mix code for the copy (GA_mixCodesType)
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI PORTRAIT_BitBltPattSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3,
    N_int32 flipY)
{
    int dstPitch;

    ROTATE(dstLeft,dstTop,width);
    SWAP(width,height);
    DC2D.beginDirectAccess();
    switch (DC.mi.BitsPerPixel) {
        case 8:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft;
            if (DC.VRAMBufferStartY)
                dstPitch = DC.mi.BytesPerScanLine;
            else
                dstPitch = ROUNDUP(width,DC.mi.BitmapStridePad);
            RotateBitmap8BPP(DC.buffer,srcAddr,srcPitch,dstPitch,height,width,flipY);
            break;
        case 15:
        case 16:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*2;
            if (DC.VRAMBufferStartY)
                dstPitch = DC.mi.BytesPerScanLine;
            else
                dstPitch = ROUNDUP(width*2,DC.mi.BitmapStridePad);
            RotateBitmap16BPP(DC.buffer,srcAddr,srcPitch,dstPitch,height,width,flipY);
            break;
        case 24:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*3;
            if (DC.VRAMBufferStartY)
                dstPitch = DC.mi.BytesPerScanLine;
            else
                dstPitch = ROUNDUP(width*3,DC.mi.BitmapStridePad);
            RotateBitmap24BPP(DC.buffer,srcAddr,srcPitch,dstPitch,height,width,flipY);
            break;
        case 32:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*4;
            if (DC.VRAMBufferStartY)
                dstPitch = DC.mi.BytesPerScanLine;
            else
                dstPitch = ROUNDUP(width*4,DC.mi.BitmapStridePad);
            RotateBitmap32BPP(DC.buffer,srcAddr,srcPitch,dstPitch,height,width,flipY);
            break;
        }
    DC2D.endDirectAccess();
    if (DC.VRAMBufferStartY)
        DRAW2D.BitBltPatt(0,DC.VRAMBufferStartY,width,height,dstLeft,dstTop,rop3);
    else if (DC.VRAMBufferStart)
        DRAW2D.BitBltPattLin(DC.VRAMBufferStart,dstPitch,0,0,width,height,dstLeft,dstTop,rop3);
    else
        DRAW2D.BitBltPattSys(DC.buffer,dstPitch,0,0,width,height,dstLeft,dstTop,rop3,false);
}

/****************************************************************************
PARAMETERS:
srcLeft - Left coordinate of the source rectangle to copy
srcTop  - Top coordinate of the source rectangle to copy
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
dstLeft - Left coordinate of destination
dstTop  - Top coordinate of destination
rop3    - ROP3 mix code for the copy
****************************************************************************/
void NAPI PORTRAIT_BitBltColorPatt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    ROTATE(srcLeft,srcTop,width);
    ROTATE(dstLeft,dstTop,width);
    SWAP(width,height);
    DRAW2D.BitBltColorPatt(srcLeft,srcTop,width,height,dstLeft,dstTop,rop3);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3        - Mix code for the copy (GA_mixCodesType)
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI PORTRAIT_BitBltColorPattSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3,
    N_int32 flipY)
{
    int dstPitch;

    ROTATE(dstLeft,dstTop,width);
    SWAP(width,height);
    DC2D.beginDirectAccess();
    switch (DC.mi.BitsPerPixel) {
        case 8:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft;
            if (DC.VRAMBufferStartY)
                dstPitch = DC.mi.BytesPerScanLine;
            else
                dstPitch = ROUNDUP(width,DC.mi.BitmapStridePad);
            RotateBitmap8BPP(DC.buffer,srcAddr,srcPitch,dstPitch,height,width,flipY);
            break;
        case 15:
        case 16:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*2;
            if (DC.VRAMBufferStartY)
                dstPitch = DC.mi.BytesPerScanLine;
            else
                dstPitch = ROUNDUP(width*2,DC.mi.BitmapStridePad);
            RotateBitmap16BPP(DC.buffer,srcAddr,srcPitch,dstPitch,height,width,flipY);
            break;
        case 24:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*3;
            if (DC.VRAMBufferStartY)
                dstPitch = DC.mi.BytesPerScanLine;
            else
                dstPitch = ROUNDUP(width*3,DC.mi.BitmapStridePad);
            RotateBitmap24BPP(DC.buffer,srcAddr,srcPitch,dstPitch,height,width,flipY);
            break;
        case 32:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*4;
            if (DC.VRAMBufferStartY)
                dstPitch = DC.mi.BytesPerScanLine;
            else
                dstPitch = ROUNDUP(width*4,DC.mi.BitmapStridePad);
            RotateBitmap32BPP(DC.buffer,srcAddr,srcPitch,dstPitch,height,width,flipY);
            break;
        }
    DC2D.endDirectAccess();
    if (DC.VRAMBufferStartY)
        DRAW2D.BitBltColorPatt(0,DC.VRAMBufferStartY,width,height,dstLeft,dstTop,rop3);
    else if (DC.VRAMBufferStart)
        DRAW2D.BitBltColorPattLin(DC.VRAMBufferStart,dstPitch,0,0,width,height,dstLeft,dstTop,rop3);
    else
        DRAW2D.BitBltColorPattSys(DC.buffer,dstPitch,0,0,width,height,dstLeft,dstTop,rop3,false);
}

/****************************************************************************
PARAMETERS:
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
****************************************************************************/
void NAPI PORTRAIT_SrcTransBlt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    ROTATE(srcLeft,srcTop,width);
    ROTATE(dstLeft,dstTop,width);
    SWAP(width,height);
    DRAW2D.SrcTransBlt(srcLeft,srcTop,width,height,dstLeft,dstTop,mix,transparent);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI PORTRAIT_SrcTransBltSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent,
    N_int32 flipY)
{
    int dstPitch;

    ROTATE(dstLeft,dstTop,width);
    SWAP(width,height);
    DC2D.beginDirectAccess();
    switch (DC.mi.BitsPerPixel) {
        case 8:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft;
            if (DC.VRAMBufferStartY)
                dstPitch = DC.mi.BytesPerScanLine;
            else
                dstPitch = ROUNDUP(width,DC.mi.BitmapStridePad);
            RotateBitmap8BPP(DC.buffer,srcAddr,srcPitch,dstPitch,height,width,flipY);
            break;
        case 15:
        case 16:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*2;
            if (DC.VRAMBufferStartY)
                dstPitch = DC.mi.BytesPerScanLine;
            else
                dstPitch = ROUNDUP(width*2,DC.mi.BitmapStridePad);
            RotateBitmap16BPP(DC.buffer,srcAddr,srcPitch,dstPitch,height,width,flipY);
            break;
        case 24:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*3;
            if (DC.VRAMBufferStartY)
                dstPitch = DC.mi.BytesPerScanLine;
            else
                dstPitch = ROUNDUP(width*3,DC.mi.BitmapStridePad);
            RotateBitmap24BPP(DC.buffer,srcAddr,srcPitch,dstPitch,height,width,flipY);
            break;
        case 32:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*4;
            if (DC.VRAMBufferStartY)
                dstPitch = DC.mi.BytesPerScanLine;
            else
                dstPitch = ROUNDUP(width*4,DC.mi.BitmapStridePad);
            RotateBitmap32BPP(DC.buffer,srcAddr,srcPitch,dstPitch,height,width,flipY);
            break;
        }
    DC2D.endDirectAccess();
    if (DC.VRAMBufferStartY)
        DRAW2D.SrcTransBlt(0,DC.VRAMBufferStartY,width,height,dstLeft,dstTop,mix,transparent);
    else if (DC.VRAMBufferStart)
        DRAW2D.SrcTransBltLin(DC.VRAMBufferStart,dstPitch,0,0,width,height,dstLeft,dstTop,mix,transparent);
    else
        DRAW2D.SrcTransBltSys(DC.buffer,dstPitch,0,0,width,height,dstLeft,dstTop,mix,transparent,false);
}

/****************************************************************************
PARAMETERS:
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
****************************************************************************/
void NAPI PORTRAIT_DstTransBlt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    ROTATE(srcLeft,srcTop,width);
    ROTATE(dstLeft,dstTop,width);
    SWAP(width,height);
    DRAW2D.DstTransBlt(srcLeft,srcTop,width,height,dstLeft,dstTop,mix,transparent);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI PORTRAIT_DstTransBltSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent,
    N_int32 flipY)
{
    int dstPitch;

    ROTATE(dstLeft,dstTop,width);
    SWAP(width,height);
    DC2D.beginDirectAccess();
    switch (DC.mi.BitsPerPixel) {
        case 8:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft;
            if (DC.VRAMBufferStartY)
                dstPitch = DC.mi.BytesPerScanLine;
            else
                dstPitch = ROUNDUP(width,DC.mi.BitmapStridePad);
            RotateBitmap8BPP(DC.buffer,srcAddr,srcPitch,dstPitch,height,width,flipY);
            break;
        case 15:
        case 16:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*2;
            if (DC.VRAMBufferStartY)
                dstPitch = DC.mi.BytesPerScanLine;
            else
                dstPitch = ROUNDUP(width*2,DC.mi.BitmapStridePad);
            RotateBitmap16BPP(DC.buffer,srcAddr,srcPitch,dstPitch,height,width,flipY);
            break;
        case 24:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*3;
            if (DC.VRAMBufferStartY)
                dstPitch = DC.mi.BytesPerScanLine;
            else
                dstPitch = ROUNDUP(width*3,DC.mi.BitmapStridePad);
            RotateBitmap24BPP(DC.buffer,srcAddr,srcPitch,dstPitch,height,width,flipY);
            break;
        case 32:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*4;
            if (DC.VRAMBufferStartY)
                dstPitch = DC.mi.BytesPerScanLine;
            else
                dstPitch = ROUNDUP(width*4,DC.mi.BitmapStridePad);
            RotateBitmap32BPP(DC.buffer,srcAddr,srcPitch,dstPitch,height,width,flipY);
            break;
        }
    DC2D.endDirectAccess();
    if (DC.VRAMBufferStartY)
        DRAW2D.DstTransBlt(0,DC.VRAMBufferStartY,width,height,dstLeft,dstTop,mix,transparent);
    else if (DC.VRAMBufferStart)
        DRAW2D.DstTransBltLin(DC.VRAMBufferStart,dstPitch,0,0,width,height,dstLeft,dstTop,mix,transparent);
    else
        DRAW2D.DstTransBltSys(DC.buffer,dstPitch,0,0,width,height,dstLeft,dstTop,mix,transparent,false);
}

/****************************************************************************
PARAMETERS:
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstWidth    - Width of the destination rectangle in pixels
dstHeight   - Height of the destination rectangle in scanlines
doClip      - True if the blit should be clipped
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle
mix         - Mix code for the copy (GA_mixCodesType)
****************************************************************************/
void NAPI PORTRAIT_StretchBlt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    N_int32 doClip,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom,
    N_int32 mix)
{
    int clipWidth = clipRight - clipLeft;
    int clipHeight = clipBottom - clipTop;

    ROTATE(srcLeft,srcTop,srcWidth);
    ROTATE(dstLeft,dstTop,dstWidth);
    ROTATE(clipLeft,clipTop,clipWidth);
    SWAP(srcWidth,srcHeight);
    SWAP(dstWidth,dstHeight);
    SWAP(clipWidth,clipHeight);
    DRAW2D.StretchBlt(srcLeft,srcTop,srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,doClip,clipLeft,clipTop,clipLeft+clipWidth,clipTop+clipHeight,mix);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstWidth    - Width of the destination rectangle in pixels
dstHeight   - Height of the destination rectangle in scanlines
doClip      - True if the blit should be clipped
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle
mix         - Mix code for the copy (GA_mixCodesType)
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI PORTRAIT_StretchBltSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    N_int32 doClip,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom,
    N_int32 mix,
    N_int32 flipY)
{
    int dstPitch;
    int clipWidth = clipRight - clipLeft;
    int clipHeight = clipBottom - clipTop;

    ROTATE(dstLeft,dstTop,dstWidth);
    ROTATE(clipLeft,clipTop,clipWidth);
    SWAP(srcWidth,srcHeight);
    SWAP(dstWidth,dstHeight);
    SWAP(clipWidth,clipHeight);
    switch (DC.mi.BitsPerPixel) {
        case 8:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft;
            dstPitch = ROUNDUP(srcWidth,4);
            RotateBitmap8BPP(DC.sysBuffer,srcAddr,srcPitch,dstPitch,srcHeight,srcWidth,flipY);
            break;
        case 15:
        case 16:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*2;
            dstPitch = ROUNDUP(srcWidth*2,4);
            RotateBitmap16BPP(DC.sysBuffer,srcAddr,srcPitch,dstPitch,srcHeight,srcWidth,flipY);
            break;
        case 24:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*3;
            dstPitch = ROUNDUP(srcWidth*3,4);
            RotateBitmap24BPP(DC.sysBuffer,srcAddr,srcPitch,dstPitch,srcHeight,srcWidth,flipY);
            break;
        case 32:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*4;
            dstPitch = ROUNDUP(srcWidth*4,4);
            RotateBitmap32BPP(DC.sysBuffer,srcAddr,srcPitch,dstPitch,srcHeight,srcWidth,flipY);
            break;
        }
    DRAW2D.StretchBltSys(DC.sysBuffer,dstPitch,0,0,srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,doClip,clipLeft,clipTop,clipLeft+clipWidth,clipTop+clipHeight,mix,false);
}

/****************************************************************************
PARAMETERS:
srcAddr         - Address of source bitmap in system memory
srcPitch        - Pitch of source bitmap in bytes
srcLeft         - Left coordinate within source bitmap to copy
srcTop          - Top coordinate within source bitmap to copy
width           - Width of the rectangle in pixels
height          - Height of the rectangle in scanlines
dstLeft         - Left coordinate of destination
dstTop          - Top coordinate of destination
bitsPerPixel    - Pixel depth for the bitmap to copy
pixelFormat     - Pixel format for the bitmap to copy
dstPal          - Color palette for the destination device (if dest is 8bpp)
srcPal          - Color palette for the bitmap to copy (if source is 8bpp)
dither          - True if the output should be dithered
mix             - Mix code for the copy (GA_mixCodesType)
flipY           - True if the bitmap should be flipped in Y

RETURNS:
0 on success, -1 on failure.
****************************************************************************/
N_int32 NAPI PORTRAIT_ConvertBltSys_Obsolete(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pixelFormat,
    GA_palette *dstPal,
    GA_palette *srcPal,
    N_int32 dither,
    N_int32 mix,
    N_int32 flipY)
{
    int dstPitch;

    ROTATE(dstLeft,dstTop,width);
    SWAP(width,height);
    switch (bitsPerPixel) {
        case 4:
            // TODO: This is probably not correct, since we dont take into
            //       account the mid pixel point problems...
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft/2;
            srcLeft &= 1;
            srcTop = 0;
            dstPitch = ROUNDUP((width+1)/2,4);
            RotateBitmap4BPP(DC.sysBuffer,srcAddr,srcPitch,dstPitch,height+srcLeft,width,flipY);
            break;
        case 8:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft;
            srcLeft = srcTop = 0;
            dstPitch = ROUNDUP(width,4);
            RotateBitmap8BPP(DC.sysBuffer,srcAddr,srcPitch,dstPitch,height,width,flipY);
            break;
        case 15:
        case 16:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*2;
            srcLeft = srcTop = 0;
            dstPitch = ROUNDUP(width*2,4);
            RotateBitmap16BPP(DC.sysBuffer,srcAddr,srcPitch,dstPitch,height,width,flipY);
            break;
        case 24:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*3;
            srcLeft = srcTop = 0;
            dstPitch = ROUNDUP(width*3,4);
            RotateBitmap24BPP(DC.sysBuffer,srcAddr,srcPitch,dstPitch,height,width,flipY);
            break;
        case 32:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*4;
            srcLeft = srcTop = 0;
            dstPitch = ROUNDUP(width*4,4);
            RotateBitmap32BPP(DC.sysBuffer,srcAddr,srcPitch,dstPitch,height,width,flipY);
            break;
        }
    return DRAW2D.ConvertBltSys_Obsolete(DC.sysBuffer,dstPitch,srcLeft,srcTop,width,height,
        dstLeft,dstTop,bitsPerPixel,pixelFormat,dstPal,srcPal,dither,mix,false);
}

/****************************************************************************
PARAMETERS:
srcAddr         - Address of source bitmap in system memory
srcPitch        - Pitch of source bitmap in bytes
srcLeft         - Left coordinate of the source rectangle to copy
srcTop          - Top coordinate of the source rectangle to copy
srcWidth        - Width of the source rectangle in pixels
srcHeight       - Height of the source rectangle in scanlines
dstLeft         - Left coordinate of destination
dstTop          - Top coordinate of destination
dstWidth        - Width of the destination rectangle in pixels
dstHeight       - Height of the destination rectangle in scanlines
doClip          - True if the blit should be clipped
clipLeft        - Left coordinate of clip rectangle
clipTop         - Top coordinate of clip rectangle
clipRight       - Right coordinate of clip rectangle
clipBottom      - Bottom coordinate of clip rectangle
bitsPerPixel    - Pixel depth for the bitmap to copy
pixelFormat     - Pixel format for the bitmap to copy
dstPal          - Color palette for the destination device (if dest is 8bpp)
srcPal          - Color palette for the bitmap to copy (if source is 8bpp)
dither          - True if the output should be dithered
mix             - Mix code for the copy (GA_mixCodesType)
flipY           - True if the bitmap should be flipped in Y

RETURNS:
0 on success, -1 on failure.
****************************************************************************/
N_int32 NAPI PORTRAIT_StretchConvertBltSys_Obsolete(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    N_int32 doClip,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pixelFormat,
    GA_palette *dstPal,
    GA_palette *srcPal,
    N_int32 dither,
    N_int32 mix,
    N_int32 flipY)
{
    int dstPitch;
    int clipWidth = clipRight - clipLeft;
    int clipHeight = clipBottom - clipTop;

    ROTATE(dstLeft,dstTop,dstWidth);
    ROTATE(clipLeft,clipTop,clipWidth);
    SWAP(srcWidth,srcHeight);
    SWAP(dstWidth,dstHeight);
    SWAP(clipWidth,clipHeight);
    switch (bitsPerPixel) {
        case 4:
            // TODO: This is probably not correct, since we dont take into
            //       account the mid pixel point problems...
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft/2;
            srcLeft &= 1;
            srcTop = 0;
            dstPitch = ROUNDUP((srcWidth+1)/2,4);
            RotateBitmap4BPP(DC.sysBuffer,srcAddr,srcPitch,dstPitch,srcHeight+srcLeft,srcWidth,flipY);
            break;
        case 8:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft;
            srcLeft = srcTop = 0;
            dstPitch = ROUNDUP(srcWidth,4);
            RotateBitmap8BPP(DC.sysBuffer,srcAddr,srcPitch,dstPitch,srcHeight,srcWidth,flipY);
            break;
        case 15:
        case 16:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*2;
            srcLeft = srcTop = 0;
            dstPitch = ROUNDUP(srcWidth*2,4);
            RotateBitmap16BPP(DC.sysBuffer,srcAddr,srcPitch,dstPitch,srcHeight,srcWidth,flipY);
            break;
        case 24:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*3;
            srcLeft = srcTop = 0;
            dstPitch = ROUNDUP(srcWidth*3,4);
            RotateBitmap24BPP(DC.sysBuffer,srcAddr,srcPitch,dstPitch,srcHeight,srcWidth,flipY);
            break;
        case 32:
            srcAddr = (uchar*)srcAddr + (srcTop * srcPitch) + srcLeft*4;
            srcLeft = srcTop = 0;
            dstPitch = ROUNDUP(srcWidth*4,4);
            RotateBitmap32BPP(DC.sysBuffer,srcAddr,srcPitch,dstPitch,srcHeight,srcWidth,flipY);
            break;
        }
    return DRAW2D.StretchConvertBltSys_Obsolete(DC.sysBuffer,dstPitch,srcLeft,srcTop,
        srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,doClip,
        clipLeft,clipTop,clipLeft+clipWidth,clipTop+clipHeight,
        bitsPerPixel,pixelFormat,dstPal,srcPal,dither,mix,false);
}

/****************************************************************************
PARAMETERS:
dstAddr     - Address of destination bitmap in system memory
dstPhysAddr - Physical address of destination bitmap in system memory
dstPitch    - Pitch of destination bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
****************************************************************************/
void NAPI PORTRAIT_GetBitmapSys(
    void *dstAddr,
    N_int32 dstPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix)
{
    int     srcPitch;
    void    *src;

    // TODO: dstLeft and dstTop should be handled!!
    // TODO: Handle mixes!
    (void)dstLeft;
    (void)dstTop;
    ROTATE(srcLeft,srcTop,width);
    SWAP(width,height);
    switch (DC.mi.BitsPerPixel) {
        case 8:
            src = (uchar*)DC2D.surface + DC.buf.Offset + DC.buf.Stride * srcTop + srcLeft;
            srcPitch = DC.buf.Stride;
            RotateBitmap8BPPReverse(dstAddr,src,srcPitch,dstPitch,width,height,false);
            break;
        case 15:
        case 16:
            src = (uchar*)DC2D.surface + DC.buf.Offset + DC.buf.Stride * srcTop + srcLeft*2;
            srcPitch = DC.buf.Stride;
            RotateBitmap16BPPReverse(dstAddr,src,srcPitch,dstPitch,width,height,false);
            break;
        case 24:
            src = (uchar*)DC2D.surface + DC.buf.Offset + DC.buf.Stride * srcTop + srcLeft*3;
            srcPitch = DC.buf.Stride;
            RotateBitmap24BPPReverse(dstAddr,src,srcPitch,dstPitch,width,height,false);
            break;
        case 32:
            src = (uchar*)DC2D.surface + DC.buf.Offset + DC.buf.Stride * srcTop + srcLeft*4;
            srcPitch = DC.buf.Stride;
            RotateBitmap32BPPReverse(dstAddr,src,srcPitch,dstPitch,width,height,false);
            break;
        }
}

/****************************************************************************
PARAMETERS:
ileft   - Left coordinate of ellipse rectangle
itop    - Top coordinate of ellipse rectangle
A       - Major axis dimension
B       - Minor axis dimension
****************************************************************************/
void NAPI PORTRAIT_DrawEllipse(
    N_int32 left,
    N_int32 top,
    N_int32 A,
    N_int32 B)
{
    ROTATE(left,top,A);
    SWAP(A,B);
    DRAW2D.DrawEllipse(left,top,A,B);
}

/****************************************************************************
PARAMETERS:
left        - Left coordinate of ellipse rectangle
top         - Top coordinate of ellipse rectangle
A           - Major axis dimension
B           - Minor axis dimension
clipLeft    - Left coordinate for clip rectangle
clipTop     - Top coordinate for clip rectangle
clipRight   - Right coordinate for clip rectangle
clipBottom  - Bottom coordinate for clip rectangle
****************************************************************************/
void NAPI PORTRAIT_ClipEllipse(
    N_int32 left,
    N_int32 top,
    N_int32 A,
    N_int32 B,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
    int clipWidth = clipRight - clipLeft;
    int clipHeight = clipBottom - clipTop;

    ROTATE(left,top,A);
    ROTATE(clipLeft,clipTop,clipWidth);
    SWAP(A,B);
    SWAP(clipWidth,clipHeight);
    DRAW2D.ClipEllipse(left,top,A,B,clipLeft,clipTop,clipLeft+clipWidth,clipTop+clipHeight);
}

/****************************************************************************
PARAMETERS:
x   - X coordinate to read pixel from
y   - Y coordinate to read pixel from

RETURNS:
Color read from the pixel on the screen
****************************************************************************/
GA_color NAPI PORTRAIT_GetPixel(
    N_int32 x,
    N_int32 y)
{
    ROTATE_PT(x,y);
    return DRAW2D.GetPixel(x,y);
}

/****************************************************************************
PARAMETERS:
x   - X coordinate to plot pixel at
y   - Y coordinate to plot pixel at
****************************************************************************/
void NAPI PORTRAIT_PutPixel(
    N_int32 x,
    N_int32 y)
{
    ROTATE_PT(x,y);
    DRAW2D.PutPixel(x,y);
}

/****************************************************************************
PARAMETERS:
cursor      - Pointer to an GA_monoCursor structure for the cursor image

REMARKS:
This function downloads the specified monochrome cursor definition from the
application into the hardware cursor. The cursor data is passed by the
application in the GA_monoCursor structure.
****************************************************************************/
void NAPI PORTRAIT_SetMonoCursor(
    GA_monoCursor *cursor)
{
    N_int32         pitch = 8;

    if (DC.enabled) {
        DC.cursorHotX = cursor->HotX;
        DC.cursorHotY = cursor->HotY;
        DC.monoCursor.HotX = DC.monoCursor.HotY = 0;
        RotateBitmap1BPP(DC.monoCursor.XORMask,cursor->XORMask,&pitch,64,64);
        RotateBitmap1BPP(DC.monoCursor.ANDMask,cursor->ANDMask,&pitch,64,64);
        PORTRAIT_cursorREF.SetMonoCursor(&DC.monoCursor);
        }
    else
        PORTRAIT_cursorREF.SetMonoCursor(cursor);
}

/****************************************************************************
PARAMETERS:
cursor  - Pointer to an GA_colorCursor structure for the cursor image

REMARKS:
This function downloads the specified colorCurs cursor definition from the
application into the hardware cursor.
****************************************************************************/
void NAPI PORTRAIT_SetColorCursor(
    GA_colorCursor *cursor)
{
    N_int32         pitch = 8;

    if (DC.enabled) {
        DC.cursorHotX = cursor->HotX;
        DC.cursorHotY = cursor->HotY;
        DC.colorCursor.HotX = DC.colorCursor.HotY = 0;
        memcpy(DC.colorCursor.Palette,cursor->Palette,sizeof(DC.colorCursor.Palette));
        RotateBitmap4BPP(DC.colorCursor.ColorData,cursor->ColorData,32,32,64,64,false);
        RotateBitmap1BPP(DC.colorCursor.ANDMask,cursor->ANDMask,&pitch,64,64);
        PORTRAIT_cursorREF.SetColorCursor(&DC.colorCursor);
        }
    else
        PORTRAIT_cursorREF.SetColorCursor(cursor);
}

/****************************************************************************
PARAMETERS:
cursor  - Pointer to an GA_colorCursor256 structure for the cursor image

REMARKS:
This function downloads the specified colorCurs cursor definition from the
application into the hardware cursor.
****************************************************************************/
void NAPI PORTRAIT_SetColorCursor256(
    GA_colorCursor256 *cursor)
{
    N_int32             pitch = 8;

    if (DC.enabled) {
        DC.cursorHotX = cursor->HotX;
        DC.cursorHotY = cursor->HotY;
        DC.colorCursor256.HotX = DC.colorCursor256.HotY = 0;
        memcpy(DC.colorCursor256.Palette,cursor->Palette,sizeof(DC.colorCursor256.Palette));
        RotateBitmap8BPP(DC.colorCursor256.ColorData,cursor->ColorData,64,64,64,64,false);
        RotateBitmap1BPP(DC.colorCursor256.ANDMask,cursor->ANDMask,&pitch,64,64);
        PORTRAIT_cursorREF.SetColorCursor256(&DC.colorCursor256);
        }
    else
        PORTRAIT_cursorREF.SetColorCursor256(cursor);
}

/****************************************************************************
PARAMETERS:
cursor  - Pointer to an GA_colorCursorRGB structure for the cursor image

REMARKS:
This function downloads the specified colorCurs cursor definition from the
application into the hardware cursor.
****************************************************************************/
void NAPI PORTRAIT_SetColorCursorRGB(
    GA_colorCursorRGB *cursor)
{
    N_int32             pitch = 8;

    if (DC.enabled) {
        DC.cursorHotX = cursor->HotX;
        DC.cursorHotY = cursor->HotY;
        DC.colorCursorRGB.HotX = DC.colorCursorRGB.HotY = 0;
        RotateBitmap24BPP(DC.colorCursorRGB.ColorData,cursor->ColorData,192,192,64,64,false);
        RotateBitmap1BPP(DC.colorCursorRGB.ANDMask,cursor->ANDMask,&pitch,64,64);
        PORTRAIT_cursorREF.SetColorCursorRGB(&DC.colorCursorRGB);
        }
    else
        PORTRAIT_cursorREF.SetColorCursorRGB(cursor);
}

/****************************************************************************
PARAMETERS:
cursor  - Pointer to an GA_colorCursorRGBA structure for the cursor image

REMARKS:
This function downloads the specified colorCurs cursor definition from the
application into the hardware cursor.
****************************************************************************/
void NAPI PORTRAIT_SetColorCursorRGBA(
    GA_colorCursorRGBA *cursor)
{
    if (DC.enabled) {
        DC.cursorHotX = cursor->HotX;
        DC.cursorHotY = cursor->HotY;
        DC.colorCursorRGBA.HotX = DC.colorCursorRGBA.HotY = 0;
        RotateBitmap32BPP((ulong*)DC.colorCursorRGBA.ColorData,(ulong*)cursor->ColorData,256,256,64,64,false);
        PORTRAIT_cursorREF.SetColorCursorRGBA(&DC.colorCursorRGBA);
        }
    else
        PORTRAIT_cursorREF.SetColorCursorRGBA(cursor);
}

/****************************************************************************
PARAMETERS:
x   - X coordinate of the cursor position
y   - Y coordinate of the cursor position

REMARKS:
This function sets the location of the hardware cursor in display
coordinates. This function takes the X and Y coordinates of the new cursor
location. This function will place the cursor so that the hotspot of the
cursor image is located at the specified (X,Y) location, and will correctly
handle special cases where the cursor definition needs to be located off the
edges of the display screen (such as when X=0, Y=0 and HotX,HotY > 0).
****************************************************************************/
N_int32 NAPI PORTRAIT_SetCursorPos(
    N_int32 x,
    N_int32 y)
{
    if (DC.enabled) {
        x -= DC.cursorHotX;
        y -= DC.cursorHotY;
        ROTATE(x,y,64);
        return PORTRAIT_cursorREF.SetCursorPos(x,y);
        }
    return PORTRAIT_cursorREF.SetCursorPos(x,y);
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of display affected
top     - Top coordinate of display affected
right   - Right coordinate of display affected (exclusive)
bottom  - Bottom coordinate of display affected (exclusive)

REMARKS:
This function is used when the SNAP driver is using software monoCurs and
colorCurs cursor acceleration. The cursor images are stored in offscreen
memory buffers, and the hardware blitter is used to draw the cursor images.
This function is used to inform the SNAP driver that the shell display
driver is about to draw to a particular location of memory, and to remove
the cursor image from the screen temporarily while the drawing is taking
place if the drawing region and the cursor position overlap.
****************************************************************************/
void NAPI PORTRAIT_BeginAccess(
    N_int32 left,
    N_int32 top,
    N_int32 right,
    N_int32 bottom)
{
    if (DC.enabled) {
        int width = right - left;
        int height = bottom - top;
        ROTATE(left,top,width);
        SWAP(width,height);
        PORTRAIT_cursorREF.BeginAccess(left,top,left+width,top+height);
        }
    else
        PORTRAIT_cursorREF.BeginAccess(left,top,right,bottom);
}
