/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Include module to implement pixel color blending
*
****************************************************************************/

/****************************************************************************
REQUIRED MACROS:
COLOR_TYPE          - Type of color variable (ie: uchar, ushort, ulong)
DRAW_PIXEL_PROTO    - Macro to declare the draw pixel function prototype
PACK_COLOR          - Macro to pack color from 32-bit ARGB to destination
REPACK_COLOR        - Macro to repack color to 32-bit ARGB format

PARAMETERS:
addr    - Memory address to draw pixel at
color   - Color to draw the pixel in

REMARKS:
This function draws a single pixel at the specified address in the specified
blending mode. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/

#define dest    ((COLOR_TYPE*)addr)

/****************************************************************************
PARAMETERS:
dest    - Desintation address
color   - Pixel color to set

REMARKS:
Implements pixel blending by blending together the source and destination
pixels based on the current blend functions selected.
****************************************************************************/
DRAW_PIXEL_PROTO()
{
    GA_color D = REPACK_COLOR(*dest);
    GA_color F = REPACK_COLOR(DC.foreColor);
    GA_color S = REPACK_COLOR(color);
    D = DC.blendSrc(S,D,F,CONST_ALPHA) + DC.blendDst(S,D,F,CONST_ALPHA);
    *dest = PACK_COLOR(D);
}

#undef  dest
#undef  DRAW_PIXEL_PROTO
#undef  PACK_COLOR
#undef  REPACK_COLOR
#undef  CONST_ALPHA

