/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the ternary ROP scanline blits with
*               color patterning.
*
****************************************************************************/

/****************************************************************************
PARAMETERS:
din     - Destination memory address
sin     - Source memory address
x1      - First pixel in scanline
y       - Y coordinate for scanline
width   - Number of pixels to copy

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/

#define PREAMBLE \
    COLOR_TYPE  *d = (COLOR_TYPE*)din; \
    COLOR_TYPE  *s = (COLOR_TYPE*)sin; \
    COLOR_TYPE  *pattern = &DC.color8x8Patt[DC.color8x8Index].PATT_TYPE.p[y & 7][0]; \
    COLOR_TYPE  *p = &pattern[x1 & 7]; \

DRAW_SCAN_PROTO()
{
    #define PLOT()                                  \
    {                                               \
        DC.drawPixelR3((uchar*)d X_ARG(x1),*s,*p);  \
        INC_DEST(x1,d);                             \
        INC_DEST(srcLeft,s);                        \
        INC_X(srcLeft);                             \
        p++;                                        \
    }
    #include "raster/inner/bltcolor.c"
}

#undef  DRAW_SCAN_PROTO
#undef  PREAMBLE

