/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser functions
*               for the SciTech SNAP Graphics Architecture.
*
****************************************************************************/

/****************************************************************************
PARAMETERS:
din     - Destination memory address
sin     - Source memory address
width   - Number of pixels to copy

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/

DRAW_SCAN_PROTO(_Generic)
{
    int     bytes;

    if ((srcLeft & 1) == (dstLeft & 1)) {
        // Handle first odd pixel if any
        if (srcLeft & 1) {
            DC.drawPixel((uchar*)d++,dstLeft++,(*s++ & 0x0F));
            width--;
            }

        // Block copy the middle pixels two at a time. We could
        // make this faster if we have special 'two pixel' functions
        // that draw two pixels at a time with ROP's etc.
        bytes = width/2;
        while (bytes--) {
            DC.drawPixel((uchar*)d,dstLeft++,(*s >> 4));
            DC.drawPixel((uchar*)d++,dstLeft++,(*s++ & 0x0F));
            }

        // Handle last odd pixel if any
        if (width & 1)
            DC.drawPixel((uchar*)d,dstLeft,(*s >> 4));
        }
    else {
        // Handle odd source or destination pixels
        if (dstLeft & 1) {
            DC.drawPixel((uchar*)d++,dstLeft++,(*s >> 4));
            width--;
            }

        // Copy pixels for all middle bytes
        bytes = width/2;
        while (bytes--) {
            DC.drawPixel((uchar*)d,dstLeft++,(*s++ & 0x0F));
            DC.drawPixel((uchar*)d++,dstLeft++,(*s >> 4));
            }

        // Handle last odd source pixel
        if (width & 1)
            DC.drawPixel((uchar*)d,dstLeft,(*s & 0x0F));
        }
}

#undef  DRAW_SCAN_PROTO

