/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser functions
*               for the SciTech SNAP Graphics Architecture.
*
****************************************************************************/

/****************************************************************************
PARAMETERS:
ileft       - Left coorindate of ellipse rectangle
itop        - Top coodinate of ellipse rectangle
A           - Major axis
B           - Minor axis
clipLeft    - Left coordinate for clip rectangle
clipTop     - Top coordinate for clip rectangle
clipRight   - Right coordinate for clip rectangle
clipBottom  - Bottom coordinate for clip rectangle

REMARKS:
Clipped ellipse drawing functions for generic and solid mix modes.
****************************************************************************/

DRAW_ELLIPSE_PROTO(_Generic)
{
    #define SET4PIXELS()                                                    \
        {                                                                   \
            if (inc_x) {                                                    \
                left--;         right++;                                    \
                pTopLeft--;     pBotLeft--;                                 \
                pTopRight++;    pBotRight++;                                \
                }                                                           \
            if (inc_y) {                                                    \
                top++;          bottom--;                                   \
                BINC(pTopLeft,DC.mi.bytesPerLine);                          \
                BINC(pTopRight,DC.mi.bytesPerLine);                         \
                BDEC(pBotLeft,DC.mi.bytesPerLine);                          \
                BDEC(pBotRight,DC.mi.bytesPerLine);                         \
                }                                                           \
            if (INSIDE_CLIPRECT(left,top))                                  \
                DC.drawPixel((uchar*)pTopLeft X_ARG(ileft),color);          \
            if (INSIDE_CLIPRECT(right,top) && pTopLeft != pTopRight)        \
                DC.drawPixel((uchar*)pTopRight X_ARG(ileft),color);         \
            if (pTopLeft != pBotLeft) {                                     \
                if (INSIDE_CLIPRECT(left,bottom))                           \
                    DC.drawPixel((uchar*)pBotLeft X_ARG(ileft),color);      \
                if (INSIDE_CLIPRECT(right,bottom) && pBotLeft != pBotRight) \
                    DC.drawPixel((uchar*)pBotRight X_ARG(ileft),color);     \
                }                                                           \
        }
    #include "raster/inner/ellipse.c"
}

#undef  DRAW_ELLIPSE_PROTO
