/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser functions
*               for the SciTech SNAP Graphics Architecture.
*
****************************************************************************/

/****************************************************************************
PARAMETERS:
din         - Pointer to destination bitmap scanline
sin         - Pointer to source bitmap scanline
width       - Number of pixels to convert
odd         - 1 if the first pixel is odd, 0 if not
translate   - Palette translation vector for destination bitmap
repCounts   - replication counts if this is to be stretched NULL if not
skipLeft    - first replication count to begin on

REMARKS:
This function tranlates the bits in the 4bpp source bitmap to the destination
bitmap using the passed in translation vector.
****************************************************************************/

DRAW_SCAN_PROTO(_S)
{
    COLOR_TYPE  *d = (COLOR_TYPE*)din;
    uchar       *s = (uchar*)sin;

    if (odd) {
        *d++ = t[*s++ & 0xF];   // Handle first odd pixel
        width--;
        }
    while (width--) {
        *d++ = t[*s >> 4];
        if (width-- == 0) break;
        *d++ = t[*s++ & 0xF];
        }
}

#undef  DRAW_SCAN_PROTO

