/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser functions
*               for the SciTech SNAP Graphics Architecture.
*
****************************************************************************/

/****************************************************************************
REQUIRED MACROS:
BUF_INCR        - Value to increment framebuffer pointer for single pixel
COLOR_TYPE      - Type of color variable (ie: uchar, ushort, ulong)
PLOT            - Macro to plot a pixel given the color

PARAMETERS:
addr    - Memory address to draw pixel at
image   - Mono bitmap image to draw with
width   - Number of pixels to draw for the scanline
****************************************************************************/
{
    FULL_COLOR  color;
    COLOR_TYPE  *d = (COLOR_TYPE*)addr;
    int         i;
    ulong       byte = *image++ | 0x10000;

    // Handle first byte which may require clipping
    for (i = 0; i < 8 && width--; i++) {
        if (XINSIDE(x)) {
            color = (byte & PIXEL0_MASK) ? DC.foreColor : DC.backColor;
            PLOT(d[0],x,color);
            }
        if ((byte PIXEL_SHIFT 1) & PIXEL_MASK)
            byte = *image++ | 0x10000;
        if (x++ & 1) d++;
        }

    // Handle all remaining bytes which are always unclipped
    if (width > 0) {
        while (width--) {
            color = (byte & PIXEL0_MASK) ? DC.foreColor : DC.backColor;
            PLOT(d[0],x,color);
            if ((byte PIXEL_SHIFT 1) & PIXEL_MASK)
                byte = *image++ | 0x10000;
            if (x++ & 1) d++;
            }
        }
}

#undef  PLOT

