/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser functions
*               for the SciTech SNAP Graphics Architecture.
*
****************************************************************************/

#ifndef SET4PIXELS_R1
#define SET4PIXELS_R1   SET4PIXELS
#define SET4PIXELS_R2   SET4PIXELS
#endif

/****************************************************************************
MACROS:
SET4PIXELS      - User supplied macro to plot 4 pixels
SET4PIXELS_R1   - Optional: Same as SET4PIXELS but for region 1 only
SET4PIXELS_R2   - Optional: Same as SET4PIXELS but for region 2 only

REMARKS:
Internal loop for the ellipse rendering algorithm. For each set of 4 pixels
in the ellipse, we call SET4PIXELS, which optionally increments the x and y
coordinates of the four pixels.
****************************************************************************/
{
    ibool       inc_x,inc_y;
    float       d,dx,dy;
    float       FourAsquared,EightAsquared,FourBsquared,EightBsquared;
    float       Asquared,Bsquared;
    int         left,top,right,bottom;
    COLOR_TYPE  *pTopLeft,*pTopRight,*pBotLeft,*pBotRight;
    GA_color    color = DC.foreColor;

    /* Compute static variables before entering any of the loops. Note that
     * all the setup calculations are done in floating point as the numbers
     * get very large.
     */
    Asquared        = A*A;
    FourAsquared    = 4*Asquared;
    EightAsquared   = 2*FourAsquared;
    Bsquared        = B*B;
    FourBsquared    = 4*Bsquared;
    EightBsquared   = 2*FourBsquared;

    /* Check to see if the semi-major axis is integer, and compute the
     * initial coordinates and decision variables depending on the outcome.
     */
    dx = FourAsquared * B;
    right = left = ileft + (A >> 1);
    top = itop;
    bottom = itop + B;
    if (A & 1) {
        /* We have a non-integer semi-major axis    */
        d = Asquared - 2*Asquared*B + 9*Bsquared;
        dy = FourBsquared;
        right++;
        }
    else {
        /* We have an integer semi-major axis       */
        d = Asquared - 2*Asquared*B + FourBsquared;
        dy = 0;
        }
    pTopLeft = pTopRight = (COLOR_TYPE*)Y_ADDR(top);
    BINC(pTopLeft,X_ADDR(left));
    BINC(pTopRight,X_ADDR(right));
    pBotLeft = pBotRight = (COLOR_TYPE*)Y_ADDR(bottom);
    BINC(pBotLeft,X_ADDR(left));
    BINC(pBotRight,X_ADDR(right));
    DC.beginDirectAccess();

    /* REGION 1: Draw the points on the ellipse where dy/dx > -1    */
    inc_x = inc_y = false;
    while (dx > dy) {
        SET4PIXELS_R1();
        inc_x = true;
        inc_y = false;
        if (d >= 0) {               /* Select pixel SE  */
            inc_y = true;
            dx -= EightAsquared;
            d -= dx;
            }
        dy += EightBsquared;
        d += FourBsquared + dy;
        }

    /* REGION 2: Draw the points on the ellipse where dy/dx < -1
     *
     * Compute new decision variable for region 2.
     */
    d += 3*(Asquared - Bsquared) - ((dx + dy)/2);
    while (dx >= 0) {
        SET4PIXELS_R2();
        inc_x = false;
        inc_y = true;
        if (d < 0) {                /* Select pixel SE  */
            inc_x = true;
            dy += EightBsquared;
            d += dy;
            }
        dx -= EightAsquared;
        d += FourAsquared - dx;
        }
    DC.endDirectAccess();
}

#undef  SET4PIXELS
#undef  SET4PIXELS_R1
#undef  SET4PIXELS_R2
