/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Include module to implement the inner loop for color
*               pattern scanline drawing
*
****************************************************************************/

#ifndef FULL_COLOR
#define FULL_COLOR  COLOR_TYPE
#endif
#ifndef BUF_INCR2
#define BUF_INCR2   1
#endif

/****************************************************************************
REQUIRED MACROS:
BUF_INCR        - Value to increment framebuffer pointer for single pixel
COLOR_TYPE      - Type of color variable (ie: uchar, ushort, ulong)
PLOT            - Macro to plot a pixel given the color

PARAMETERS:
addr    - Memory address to draw pixel at
image   - Mono bitmap image to draw with
width   - Number of pixels to draw for the scanline
****************************************************************************/
{
    COLOR_TYPE  *d = (COLOR_TYPE*)addr;
    COLOR_TYPE  *dMidEnd = d + ((width & ~7)*BUF_INCR2);
    COLOR_TYPE  *dEnd = d + (width*BUF_INCR2);
    FULL_COLOR  color = DC.foreColor;
    int         rightClip = ((width & 7) != 0);

    // Draw full middle bytes
    while (d < dMidEnd) {
        uchar byte = *image++;
        if (byte & PIXEL0_MASK) PLOT(d[0*BUF_INCR2],color);
        if (byte & PIXEL1_MASK) PLOT(d[1*BUF_INCR2],color);
        if (byte & PIXEL2_MASK) PLOT(d[2*BUF_INCR2],color);
        if (byte & PIXEL3_MASK) PLOT(d[3*BUF_INCR2],color);
        if (byte & PIXEL4_MASK) PLOT(d[4*BUF_INCR2],color);
        if (byte & PIXEL5_MASK) PLOT(d[5*BUF_INCR2],color);
        if (byte & PIXEL6_MASK) PLOT(d[6*BUF_INCR2],color);
        if (byte & PIXEL7_MASK) PLOT(d[7*BUF_INCR2],color);
        BINC(d,(8 * BUF_INCR));
        }

    // Draw clipped ending byte
    if (rightClip) {
        uchar byte = *image;
        while (d < dEnd) {
            if (byte & PIXEL0_MASK) PLOT(d[0],color);
            byte PIXEL_SHIFT 1;
            BINC(d,BUF_INCR);
            }
        }
}

#undef  PLOT
