/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser functions
*               for the SciTech SNAP Graphics Architecture.
*
****************************************************************************/

/****************************************************************************
PARAMETERS:
****************************************************************************/

DRAW_LINELO_PROTO()
{
    COLOR_TYPE          *src = (COLOR_TYPE*)srcAddr;
    COLOR_TYPE          *dst = (COLOR_TYPE*)(dstBase + (dstAddr & 0xFFFF));
    int                 bank;

    // Draw the body of the line
    while (count--) {
#if BUF_INCR == 3
        bank = dstAddr >> 16;
        if (((dstAddr+1) >> 16) != bank) {
            // 24bpp pixel is split across a bank boundary
            dst[0] = src[0];
            DC.SetBank(DC.cntBank = bank = (dstAddr+1) >> 16);
            dst = dstBase-1;
            dst[1] = src[1];
            dst[2] = src[2];
            }
        else if (((dstAddr+2) >> 16) != bank) {
            // 24bpp pixel is split across a bank boundary
            dst[0] = src[0];
            dst[1] = src[1];
            DC.SetBank(DC.cntBank = bank = (dstAddr+2) >> 16);
            dst = dstBase-2;
            dst[2] = src[2];
            }
        else {
            // Draw 24bpp pixel normally
            if (bank != DC.cntBank) {
                DC.SetBank(DC.cntBank = bank);
                dst = (COLOR_TYPE*)(dstBase + (dstAddr & 0xFFFF));
                }
            dst[0] = src[0];
            dst[1] = src[1];
            dst[2] = src[2];
            }
#else
        if ((bank = dstAddr >> 16) != DC.cntBank) {
            DC.SetBank(DC.cntBank = bank);
            dst = (COLOR_TYPE*)(dstBase + (dstAddr & 0xFFFF));
            }
        *dst = *src;
#endif
        BINC(src,INCR_SIGN BUF_INCR);
        BINC(dst,INCR_SIGN BUF_INCR);
        dstAddr INCR_OPE BUF_INCR;
        if (d < 0) {
            d += Eincr;             // Choose the Eastern Pixel
            }
        else {
            d += NEincr;            // Choose the North Eastern Pixel
            BINC(src,fincr);        // Increment framebuffer
            BINC(dst,fincr);        // Increment framebuffer
            dstAddr += fincr;
            }
        }
}

DRAW_LINEHI_PROTO()
{
    COLOR_TYPE          *src = (COLOR_TYPE*)srcAddr;
    COLOR_TYPE          *dst = (COLOR_TYPE*)(dstBase + (dstAddr & 0xFFFF));
    int                 bank;

    // Draw the body of the line
    while (count--) {
#if BUF_INCR == 3
        bank = dstAddr >> 16;
        if (((dstAddr+1) >> 16) != bank) {
            // 24bpp pixel is split across a bank boundary
            dst[0] = src[0];
            DC.SetBank(DC.cntBank = bank = (dstAddr+1) >> 16);
            dst = dstBase-1;
            dst[1] = src[1];
            dst[2] = src[2];
            }
        else if (((dstAddr+2) >> 16) != bank) {
            // 24bpp pixel is split across a bank boundary
            dst[0] = src[0];
            dst[1] = src[1];
            DC.SetBank(DC.cntBank = bank = (dstAddr+2) >> 16);
            dst = dstBase-2;
            dst[2] = src[2];
            }
        else {
            // Draw 24bpp pixel normally
            if (bank != DC.cntBank) {
                DC.SetBank(DC.cntBank = bank);
                dst = (COLOR_TYPE*)(dstBase + (dstAddr & 0xFFFF));
                }
            dst[0] = src[0];
            dst[1] = src[1];
            dst[2] = src[2];
            }
#else
        if ((bank = dstAddr >> 16) != DC.cntBank) {
            DC.SetBank(DC.cntBank = bank);
            dst = (COLOR_TYPE*)(dstBase + (dstAddr & 0xFFFF));
            }
        *dst = *src;
#endif
        BINC(src,fincr);        // Increment framebuffer
        BINC(dst,fincr);        // Increment framebuffer
        dstAddr += fincr;
        if (d < 0) {
            d += Eincr;         // Choose the Eastern Pixel
            }
        else {
            d += NEincr;        // Choose the North Eastern Pixel
            BINC(src,INCR_SIGN BUF_INCR);
            BINC(dst,INCR_SIGN BUF_INCR);
            dstAddr INCR_OPE BUF_INCR;
            }
        }
}

#undef INCR_SIGN
#undef INCR_OP
#undef INCR_OPE
#undef DRAW_LINELO_PROTO
#undef DRAW_LINEHI_PROTO
