/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser functions
*               for the SciTech SNAP Graphics Architecture.
*
****************************************************************************/

/****************************************************************************
PARAMETERS:
****************************************************************************/

#ifndef NO_SOLID_LINE
DRAW_LINELO_PROTO(_S)
{
    COLOR_TYPE  *dest = (COLOR_TYPE*)addr;
    COLOR_TYPE  foreColor = DC.foreColor;
    COLOR_TYPE  backColor = DC.backColor;

    // Draw the body of the line
    if (transparent) {
        while (count--) {
            if ((styleMask >> ((styleValue >> 8) & 0x1F)) & 1) {
                REPLACE_PIXEL(*dest,x,foreColor);
                }
            styleValue += styleStep;
#if INCR_POS == 1
            INC_DEST(x,dest);
            INC_X(x);
#else
            DEC_DEST(x,dest);
            DEC_X(x);
#endif
            if (d < 0) {
                d += Eincr;         // Choose the Eastern Pixel
                }
            else {
                d += NEincr;        // Choose the North Eastern Pixel
                BINC(dest,fincr);   // Increment framebuffer
                }
            }
        }
    else {
        while (count--) {
            if ((styleMask >> ((styleValue >> 8) & 0x1F)) & 1) {
                REPLACE_PIXEL(*dest,x,foreColor);
                }
            else {
                REPLACE_PIXEL(*dest,x,backColor);
                }
            styleValue += styleStep;
#if INCR_POS == 1
            INC_DEST(x,dest);
            INC_X(x);
#else
            DEC_DEST(x,dest);
            DEC_X(x);
#endif
            if (d < 0) {
                d += Eincr;         // Choose the Eastern Pixel
                }
            else {
                d += NEincr;        // Choose the North Eastern Pixel
                BINC(dest,fincr);   // Increment framebuffer
                }
            }
        }
}

DRAW_LINEHI_PROTO(_S)
{
    COLOR_TYPE  *dest = (COLOR_TYPE*)addr;
    COLOR_TYPE  foreColor = DC.foreColor;
    COLOR_TYPE  backColor = DC.backColor;

    // Draw the body of the line
    if (transparent) {
        while (count--) {
            if ((styleMask >> ((styleValue >> 8) & 0x1F)) & 1) {
                REPLACE_PIXEL(*dest,x,foreColor);
                }
            styleValue += styleStep;
            BINC(dest,fincr);       // Increment framebuffer
            if (d < 0) {
                d += Eincr;         // Choose the Eastern Pixel
                }
            else {
                d += NEincr;        // Choose the North Eastern Pixel
#if INCR_POS == 1
                INC_DEST(x,dest);
                INC_X(x);
#else
                DEC_DEST(x,dest);
                DEC_X(x);
#endif
                }
            }
        }
    else {
        while (count--) {
            if ((styleMask >> ((styleValue >> 8) & 0x1F)) & 1) {
                REPLACE_PIXEL(*dest,x,foreColor);
                }
            else {
                REPLACE_PIXEL(*dest,x,backColor);
                }
            styleValue += styleStep;
            BINC(dest,fincr);       // Increment framebuffer
            if (d < 0) {
                d += Eincr;         // Choose the Eastern Pixel
                }
            else {
                d += NEincr;        // Choose the North Eastern Pixel
#if INCR_POS == 1
                INC_DEST(x,dest);
                INC_X(x);
#else
                DEC_DEST(x,dest);
                DEC_X(x);
#endif
                }
            }
        }
}
#endif

DRAW_LINELO_PROTO(_Generic)
{
    COLOR_TYPE  *dest = (COLOR_TYPE*)addr;
    GA_color    foreColor = DC.foreColor;
    GA_color    backColor = DC.backColor;

    // Draw the body of the line
    while (count--) {
        if ((styleMask >> ((styleValue >> 8) & 0x1F)) & 1)
            DC.drawPixel((uchar*)dest X_ARG(x),foreColor);
        else if (!transparent)
            DC.drawPixel((uchar*)dest X_ARG(x),backColor);
        styleValue += styleStep;
#if INCR_POS == 1
        INC_DEST(x,dest);
        INC_X(x);
#else
        DEC_DEST(x,dest);
        DEC_X(x);
#endif
        if (d < 0) {
            d += Eincr;         // Choose the Eastern Pixel
            }
        else {
            d += NEincr;        // Choose the North Eastern Pixel
            BINC(dest,fincr);   // Increment framebuffer
            }
        }
}

DRAW_LINEHI_PROTO(_Generic)
{
    COLOR_TYPE  *dest = (COLOR_TYPE*)addr;
    GA_color    foreColor = DC.foreColor;
    GA_color    backColor = DC.backColor;

    // Draw the body of the line
    while (count--) {
        if ((styleMask >> ((styleValue >> 8) & 0x1F)) & 1)
            DC.drawPixel((uchar*)dest X_ARG(x),foreColor);
        else if (!transparent)
            DC.drawPixel((uchar*)dest X_ARG(x),backColor);
        styleValue += styleStep;
        BINC(dest,fincr);       // Increment framebuffer
        if (d < 0) {
            d += Eincr;         // Choose the Eastern Pixel
            }
        else {
            d += NEincr;        // Choose the North Eastern Pixel
#if INCR_POS == 1
            INC_DEST(x,dest);
            INC_X(x);
#else
            DEC_DEST(x,dest);
            DEC_X(x);
#endif
            }
        }
}

#undef INCR_POS
#undef DRAW_LINELO_PROTO
#undef DRAW_LINEHI_PROTO

