/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Include module to implement monochrome bitmap drawing
*               functions.
*
****************************************************************************/

#include "ref2dint.h"

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
PARAMETERS:
x           - Destination X coordinate to draw the bitmap at
y           - Destination Y coordinate to draw the bitmap at
width       - Width of the bitmap in pixels
height      - Height of the bitmap in pixels
byteWidth   - Width of the bitmap in bytes
image       - Pointer to the bitmap image data to draw
transparent - 1 for transparent, 0 for opaque

REMARKS:
If the transparent parameter is set to 1, where a bit is 0 in the bitmap
image the destination pixel remains untouched. If the transparent
parameter is set to 0, where a bit is 0 in the bitmap image the
destination pixel is drawn in the background color. In all cases where a
bit in the bitmap image is 1, the pixel is drawn in the foreground color.

Note:   This function processes the bitmap data in MSB fashion, in that the
        first pixel drawn corrsponds to bit 7 of the first byte. The second
        pixel is bit 6 of the first byte, ...,the 8th pixel is bit 7 of
        the second byte etc. Both LSB and MSB versions are provided for
        performance.
****************************************************************************/
{
    uchar *base = PIXEL_ADDR(x,y);

    DC.beginDirectAccess();
    if (transparent) {
        // Draw transparent mono bitmap
        while (height--) {
            DC.MONO_TRANS(base,image X_ARG(x),width);
            image += byteWidth;
            BINC(base,DC.mi.bytesPerLine);
            }
        }
    else {
        // Draw opaque mono bitmap
        while (height--) {
            DC.MONO_OPAQUE(base,image X_ARG(x),width);
            image += byteWidth;
            BINC(base,DC.mi.bytesPerLine);
            }
        }
    DC.endDirectAccess();
}

#undef MONO_TRANS
#undef MONO_OPAQUE
