/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Include module to implement masked pixel drawing
*
****************************************************************************/

/****************************************************************************
REQUIRED MACROS:
COLOR_TYPE          - Type of color variable (ie: uchar, ushort, ulong)
COLOR_WHITE         - Macro for the color white constant
DRAW_PIXEL_PROTO    - Macro to declare the draw pixel function prototype

PARAMETERS:
addr    - Memory address to draw pixel at
color   - Color to draw the pixel in

REMARKS:
This function draws a single pixel at the specified address in the specified
color. We abstract this function so that we can keep the inner loops of
our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.

NOTE: This version does a masked pixel update based on the current plane
      mask.
****************************************************************************/

#define dest    ((COLOR_TYPE*)addr)

DRAW_PIXEL_PROTO(_0)
{
    GA_color D = *dest;
    (void)color;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | 0);
}

DRAW_PIXEL_PROTO(_DSon)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D | (color)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSna)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D & ~(color)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_Sn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(color) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDna)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((~D & (color)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_Dn)
{
    GA_color D = *dest;
    (void)color;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((~D) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSx)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D ^ (color)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSan)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D & (color)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSa)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D & (color)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D ^ (color)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSno)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D | ~(color)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_S)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((color) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDno)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((~D | (color)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSo)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D | (color)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_1)
{
    GA_color D = *dest;
    (void)color;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | DC.planeMask);
}

#undef  dest
#undef  DRAW_PIXEL_PROTO
