/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Include module to implement ternary ROP pixel drawing
*
****************************************************************************/

/****************************************************************************
REQUIRED MACROS:
COLOR_TYPE          - Type of color variable (ie: uchar, ushort, ulong)
COLOR_WHITE         - Macro for the color white constant
DRAW_PIXEL_PROTO    - Macro to declare the draw pixel function prototype

PARAMETERS:
addr    - Memory address to draw pixel at
S       - Source color to draw the pixel in
P       - Pattern color to draw the pixel in

REMARKS:
This function draws a single pixel at the specified address in the specified
source color and pattern color, combined together using the desired tenrary
raster operation code. We abstract this function so that we can keep the
inner loops of our reference rasteriser color depth 'clean', however we
pass the address of the pixel to draw to avoid address computations per
pixel to give us performance close to an optimised rasteriser function.
****************************************************************************/

#define dest    ((COLOR_TYPE*)addr)

DRAW_PIXEL_PROTO(_0)
{
    *dest = (*dest & (COLOR_TYPE)DC.planeMaskInv) | 0;
    (void)S;
    (void)P;
}

DRAW_PIXEL_PROTO(_DPSoon)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D | (P | S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSona)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D & ~(P | S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PSon)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P | S) & DC.planeMask));
    (void)S;
}

DRAW_PIXEL_PROTO(_SDPona)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S & ~(D | P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPon)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D | P) & DC.planeMask));
    (void)S;
}

DRAW_PIXEL_PROTO(_PDSxnon)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P | ~(D ^ S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSaon)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P | (D & S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPnaa)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S & (D & ~P)) & DC.planeMask));
    (void)S;
}

DRAW_PIXEL_PROTO(_PDSxon)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P | (D ^ S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPna)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D & ~P) & DC.planeMask));
    (void)S;
}

DRAW_PIXEL_PROTO(_PSDnaon)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P | (S & ~D)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPna)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S & ~P) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSnaon)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P | (D & ~S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSonon)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P | ~(D | S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_Pn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~P & DC.planeMask));
    (void)S;
}

DRAW_PIXEL_PROTO(_PDSona)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P & ~(D | S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSon)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D | S) & DC.planeMask));
    (void)P;
}

DRAW_PIXEL_PROTO(_SDPxnon)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S | ~(D ^ P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPaon)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S | (D & P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSxnon)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D | ~(P ^ S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSaon)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D | (P & S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PSDPSanaxx)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P ^ (S ^ (D & ~(P & S)))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SSPxDSxaxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S ^ ((S ^ P) & (D ^ S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPxPDxa)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (((S ^ P) & (P ^ D)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPSanaxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S ^ (D & ~(P & S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSPaox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P ^ (D | (S & P))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPSxaxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S ^ (D & (P ^ S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PSDPaox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P ^ (S | (D & P))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSPDxaxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D ^ (S & (P ^ D))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P ^ (D | S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSoan)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P & (D | S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSnaa)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D & (P & ~S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPxon)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S | (D ^ P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSna)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D & ~S) & DC.planeMask));
    (void)P;
}

DRAW_PIXEL_PROTO(_SPDnaon)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S | (P & ~D)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPxDSxa)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (((S ^ P) & (D ^ S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSPanaxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P ^ (D & ~(S & P))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPSaox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S ^ (D | (P & S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPSxnox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S ^ (D | ~(P ^ S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSxa)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D & (P ^ S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PSDPSaoxxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P ^ (S ^ (D | (P & S)))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSana)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D & ~(P & S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SSPxPDxaxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S ^ ((S ^ P) & (P ^ D))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPDSoax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S ^ (P & (D | S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PSDnox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P ^ (S | ~D)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PSDPxox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P ^ (S | (D ^ P))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PSDnoan)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P & (S | ~D)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PSna)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P & ~S) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPnaon)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S | (D & ~P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPSoox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S ^ (D | (P | S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_Sn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~S & DC.planeMask));
    (void)P;
}

DRAW_PIXEL_PROTO(_SPDSaox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S ^ (P | (D & S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPDSxnox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S ^ (P | ~(D ^ S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S ^ (D | P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPoan)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S & (D | P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PSDPoax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P ^ (S & (D | P))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPDnox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S ^ (P | ~D)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPDSxox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S ^ (P | (D ^ S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPDnoan)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S & (P | ~D)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PSx)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P ^ S) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPDSonox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S ^ (P | ~(D | S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPDSnaox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S ^ (P | (D & ~S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PSan)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P & S) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PSDnaa)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P & (S & ~D)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSxon)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D | (P ^ S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDxPDxa)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (((S ^ D) & (P ^ D)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPDSanaxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S ^ (P & ~(D & S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDna)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S & ~D) & DC.planeMask));
    (void)P;
}

DRAW_PIXEL_PROTO(_DPSnaon)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D | (P & ~S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSPDaox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D ^ (S | (P & D))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PSDPxaxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P ^ (S & (D ^ P))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPxa)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S & (D ^ P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSPDaoxxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P ^ (D ^ (S | (P & D)))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSDoax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D ^ (P & (S | D))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSnox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P ^ (D | ~S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPana)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S & ~(D & P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SSPxDSxoxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S ^ ((S ^ P) | (D ^ S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSPxox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P ^ (D | (S ^ P))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSnoan)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P & (D | ~S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDna)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P & ~D) & DC.planeMask));
    (void)S;
}

DRAW_PIXEL_PROTO(_DSPnaon)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D | (S & ~P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSDaox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D ^ (P | (S & D))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPDSxaxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S ^ (P & (D ^ S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSonon)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D | ~(P | S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_Dn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~D & DC.planeMask));
    (void)S;
    (void)P;
}

DRAW_PIXEL_PROTO(_DPSox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D ^ (P | S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSoan)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D & (P | S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSPoax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P ^ (D & (S | P))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSnox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D ^ (P | ~S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPx)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D ^ P) & DC.planeMask));
    (void)S;
}

DRAW_PIXEL_PROTO(_DPSDonox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D ^ (P | ~(S | D))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSDxox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D ^ (P | (S ^ D))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSnoan)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D & (P | ~S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSDnaox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D ^ (P | (S & ~D))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPan)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D & P) & DC.planeMask));
    (void)S;
}

DRAW_PIXEL_PROTO(_PDSxa)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P & (D ^ S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSPDSaoxxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D ^ (S ^ (P | (D & S)))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSPDoax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D ^ (S & (P | D))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPnox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S ^ (D | ~P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPSoax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S ^ (D & (P | S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSPnox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D ^ (S | ~P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSx)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D ^ S) & DC.planeMask));
    (void)P;
}

DRAW_PIXEL_PROTO(_SDPSonox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S ^ (D | ~(P | S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSPDSonoxxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D ^ (S ^ (P | ~(D | S)))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSxxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P ^ (D ^ S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D ^ (P & S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PSDPSoaxxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P ^ (S ^ (D & (P | S)))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S ^ (D & P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSPDoaxxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P ^ (D ^ (S & (P | D)))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPSnoax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S ^ (D & (P | ~S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSxnan)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P & ~(D ^ S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSana)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P & ~(D & S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SSDxPDxaxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S ^ ((S ^ D) & (P ^ D))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPSxox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S ^ (D | (P ^ S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPnoan)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S & (D | ~P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSPDxox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D ^ (S | (P ^ D))) & DC.planeMask));
}

#undef  dest
#undef  DRAW_PIXEL_PROTO
