/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Include module to implement ternary ROP pixel drawing
*
****************************************************************************/

/****************************************************************************
REQUIRED MACROS:
COLOR_TYPE          - Type of color variable (ie: uchar, ushort, ulong)
COLOR_WHITE         - Macro for the color white constant
DRAW_PIXEL_PROTO    - Macro to declare the draw pixel function prototype

PARAMETERS:
addr    - Memory address to draw pixel at
S       - Source color to draw the pixel in
P       - Pattern color to draw the pixel in

REMARKS:
This function draws a single pixel at the specified address in the specified
source color and pattern color, combined together using the desired tenrary
raster operation code. We abstract this function so that we can keep the
inner loops of our reference rasteriser color depth 'clean', however we
pass the address of the pixel to draw to avoid address computations per
pixel to give us performance close to an optimised rasteriser function.
****************************************************************************/

#define dest    ((COLOR_TYPE*)addr)

DRAW_PIXEL_PROTO(_DSPnoan)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D & (S | ~P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPSnaox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S ^ (D | (P & ~S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSan)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D & S) & DC.planeMask));
    (void)P;
}

DRAW_PIXEL_PROTO(_PDSax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P ^ (D & S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSPDSoaxxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D ^ (S ^ (P & (D | S)))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSDnoax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D ^ (P & (S | ~D))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPxnan)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S & ~(D ^ P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPDSnoax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S ^ (P & (D | ~S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSxnan)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D & ~(P ^ S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPxDSxo)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (((S ^ P) | (D ^ S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSaan)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D & (P & S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSaa)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D & (P & S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPxDSxon)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~((S ^ P) | (D ^ S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSxna)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D & ~(P ^ S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPDSnoaxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S ^ (P & (D | ~S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPxna)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S & ~(D ^ P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSPnoaxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P ^ (D & (S | ~P))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSPDSoaxx)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D ^ (S ^ (P & (D | S)))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSaxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P ^ (D & S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSa)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D & S) & DC.planeMask));
    (void)P;
}

DRAW_PIXEL_PROTO(_SDPSnaoxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S ^ (D | (P & ~S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSPnoa)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D & (S | ~P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSPDxoxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D ^ (S | (P ^ D))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPnoa)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S & (D | ~P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPSxoxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S ^ (D | (P ^ S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SSDxPDxax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S ^ ((S ^ D) & (P ^ D))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSanan)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P & ~(D & S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSxna)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P & ~(D ^ S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPSnoaxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S ^ (D & (P | ~S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSDPoaxx)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D ^ (P ^ (S & (D | P)))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPDaxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S ^ (P & D)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PSDPSoaxx)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P ^ (S ^ (D & (P | S)))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSaxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D ^ (P & S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSxx)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D ^ (P ^ S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PSDPSonoxx)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P ^ (S ^ (D | ~(P | S)))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPSonoxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S ^ (D | ~(P | S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D ^ S) & DC.planeMask));
    (void)P;
}

DRAW_PIXEL_PROTO(_DPSnax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D ^ (P & ~S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPSoaxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S ^ (D & (P | S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPDnax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S ^ (P & ~D)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSPDoaxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D ^ (S & (P | D))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSPDSaoxx)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D ^ (S ^ (P | (D & S)))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSxan)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P & (D ^ S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPa)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D & P) & DC.planeMask));
    (void)S;
}

DRAW_PIXEL_PROTO(_PDSPnaoxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P ^ (D | (S & ~P))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSnoa)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D & (P | ~S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSDxoxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D ^ (P | (S ^ D))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSPonoxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P ^ (D | ~(S | P))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P ^ D) & DC.planeMask));
    (void)S;
}

DRAW_PIXEL_PROTO(_DSPnax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D ^ (S & ~P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSPoaxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P ^ (D & (S | P))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSoa)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D & (P | S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSoxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D ^ (P | S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_D)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (D & DC.planeMask));
    (void)S;
    (void)P;
}

DRAW_PIXEL_PROTO(_DPSono)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D | ~(P | S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPDSxax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S ^ (P & (D ^ S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSDaoxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D ^ (P | (S & D))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSPnao)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D | (S & ~P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPno)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D | ~P) & DC.planeMask));
    (void)S;
}

DRAW_PIXEL_PROTO(_PDSnoa)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P & (D | ~S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSPxoxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P ^ (D | (S ^ P))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SSPxDSxox)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S ^ ((S ^ P) | (D ^ S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPanan)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S & ~(D & P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PSDnax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P ^ (S & ~D)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSDoaxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D ^ (P & (S | D))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSDPaoxx)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D ^ (P ^ (S | (D & P)))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPxan)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S & (D ^ P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PSDPxax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P ^ (S & (D ^ P))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSPDaoxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D ^ (S | (P & D))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSnao)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D | (P & ~S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSno)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D | ~S) & DC.planeMask));
    (void)P;
}

DRAW_PIXEL_PROTO(_SPDSanax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S ^ (P & ~(D & S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDxPDxan)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~((S ^ D) & (P ^ D)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSxo)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D | (P ^ S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSano)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D | ~(P & S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PSa)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P & S) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPDSnaoxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S ^ (P | (D & ~S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPDSonoxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S ^ (P | ~(D | S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PSxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P ^ S) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPDnoa)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S & (P | ~D)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPDSxoxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S ^ (P | (D ^ S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPnax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S ^ (D & ~P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PSDPoaxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P ^ (S & (D | P))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPoa)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S & (D | P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPDoxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S ^ (P | D)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSDxax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D ^ (P & (S ^ D))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPDSaoxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S ^ (P | (D & S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_S)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (S & DC.planeMask));
    (void)P;
}

DRAW_PIXEL_PROTO(_SDPono)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S | ~(D | P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPnao)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S | (D & ~P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPno)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S | ~P) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PSDnoa)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P & (S | ~D)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PSDPxoxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P ^ (S | (D ^ P))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSnax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P ^ (D & ~S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPDSoaxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S ^ (P & (D | S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SSPxPDxax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S ^ ((S ^ P) & (P ^ D))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSanan)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D & ~(P & S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PSDPSaoxx)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P ^ (S ^ (D | (P & S)))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSxan)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D & (P ^ S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSPxax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P ^ (D & (S ^ P))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPSaoxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(S ^ (D | (P & S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSDanax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D ^ (P & ~(S & D))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPxDSxan)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~((S ^ P) & (D ^ S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPDnao)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S | (P & ~D)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDno)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S | ~D) & DC.planeMask));
    (void)P;
}

DRAW_PIXEL_PROTO(_SDPxo)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S | (D ^ P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPano)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S | ~(D & P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSoa)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P & (D | S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSoxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P ^ (D | S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSPDxax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D ^ (S & (P ^ D))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PSDPaoxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P ^ (S | (D & P))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPSxax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S ^ (D & (P ^ S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSPaoxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(P ^ (D | (S & P))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPSanax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S ^ (D & ~(P & S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SPxPDxan)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~((S ^ P) & (P ^ D)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SSPxDSxax)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S ^ ((S ^ P) & (D ^ S))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSPDSanaxxn)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (~(D ^ (S ^ (P & ~(D & S)))) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSao)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D | (P & S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSxno)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D | ~(P ^ S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPao)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S | (D & P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_SDPxno)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S | ~(D ^ P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DSo)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D | S) & DC.planeMask));
    (void)P;
}

DRAW_PIXEL_PROTO(_SDPnoo)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((S | (D | ~P)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_P)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | (P & DC.planeMask));
    (void)S;
}

DRAW_PIXEL_PROTO(_PDSono)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P | ~(D | S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSnao)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P | (D & ~S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PSno)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P | ~S) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PSDnao)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P | (S & ~D)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDno)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P | ~D) & DC.planeMask));
    (void)S;
}

DRAW_PIXEL_PROTO(_PDSxo)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P | (D ^ S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSano)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P | ~(D & S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSao)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P | (D & S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PDSxno)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P | ~(D ^ S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPo)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D | P) & DC.planeMask));
    (void)S;
}

DRAW_PIXEL_PROTO(_DPSnoo)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D | (P | ~S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PSo)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P | S) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_PSDnoo)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((P | (S | ~D)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_DPSoo)
{
    GA_color D = *dest;
    *dest = (COLOR_TYPE)((D & DC.planeMaskInv) | ((D | (P | S)) & DC.planeMask));
}

DRAW_PIXEL_PROTO(_1)
{
    *dest = (*dest & (COLOR_TYPE)DC.planeMaskInv) | (COLOR_TYPE)DC.planeMask;
    (void)S;
    (void)P;
}

#undef  dest
#undef  DRAW_PIXEL_PROTO
