/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Include module to implement ternary ROP solid color and
*               mono pattern scanline drawing
*
****************************************************************************/

/****************************************************************************
REQUIRED MACROS:
COLOR_TYPE  - Type of color variable (ie: uchar, ushort, ulong)
COLOR_WHITE - Macro for the color white constant
DRAW_SCAN_PROTO - Macro to declare the draw scan function prototype

PARAMETERS:
addr    - Memory address to draw pixel at
x1      - First pixel in scanline
y       - Y coordinate for scanline
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a monochrome patterned scanline at the specified address
in the current color. We abstract this function so that we can keep the inner
loops of our reference rasteriser color depth 'clean', however we pass the
address of the pixel to draw to avoid address computations per pixel to give
us performance close to an optimised rasteriser function.
****************************************************************************/

#define PREAMBLE \
    COLOR_TYPE  *dest = (COLOR_TYPE*)addr; \
    uchar       pattern = DC.mono8x8Patt[DC.mono8x8Index].p[y & 7]; \
    uchar       p = pattern << (x1 & 7)

DRAW_SCAN_PROTO()
{
    #define PLOT(d,c,p) DC.drawPixelR3((uchar*)&(d) X_ARG(x1),c,p)
    #include "raster/inner/scanmono.c"
}

#undef  DRAW_SCAN_PROTO
#undef  PREAMBLE
