/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser functions
*               for the SciTech SNAP Graphics Architecture.
*
****************************************************************************/

/****************************************************************************
PARAMETERS:
din         - Destination memory address
sin         - Source memory address
width       - Number of pixels to copy
transparent - Transaparent source color to skip

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
DRAW_SCAN_PROTO(_S)
{
    COLOR_TYPE  *d = (COLOR_TYPE*)din;
    COLOR_TYPE  *s = (COLOR_TYPE*)sin;
    COLOR_TYPE  c;
    COLOR_TYPE  transparent = (COLOR_TYPE)DC.transparent;

#ifndef NO_UNROLLED_PIXELS
    while (width >= 8) {
        if (s[0] != transparent) d[0] = s[0];
        if (s[1] != transparent) d[1] = s[1];
        if (s[2] != transparent) d[2] = s[2];
        if (s[3] != transparent) d[3] = s[3];
        if (s[4] != transparent) d[4] = s[4];
        if (s[5] != transparent) d[5] = s[5];
        if (s[6] != transparent) d[6] = s[6];
        if (s[7] != transparent) d[7] = s[7];
        d += 8;
        s += 8;
        width -= 8;
        }
#endif
    while (width--) {
        if ((c = READ_PIXEL(*s,srcLeft)) != transparent)
            REPLACE_PIXEL(*d,dstLeft,c);
        INC_DEST(dstLeft,d);
        INC_DEST(srcLeft,s);
        INC_X(dstLeft);
        INC_X(srcLeft);
        }
}

DRAW_SCAN_PROTO(_Generic)
{
    COLOR_TYPE  *d = (COLOR_TYPE*)din;
    COLOR_TYPE  *s = (COLOR_TYPE*)sin;
    COLOR_TYPE  c;
    COLOR_TYPE  transparent = (COLOR_TYPE)DC.transparent;

    while (width--) {
        if ((c = READ_PIXEL(*s,srcLeft)) != transparent)
            DC.drawPixel((uchar*)d X_ARG(dstLeft),c);
        INC_DEST(dstLeft,d);
        INC_DEST(srcLeft,s);
        INC_X(dstLeft);
        INC_X(srcLeft);
        }
}

DRAW_SCAN_PROTO(_Masked)
{
    COLOR_TYPE  *d = (COLOR_TYPE*)din;
    COLOR_TYPE  *s = (COLOR_TYPE*)sin;
    COLOR_TYPE  c;
    COLOR_TYPE  transparent = (COLOR_TYPE)DC.transparent;

    while (width--) {
        if ((c = (READ_PIXEL(*s,srcLeft) & DC.colorCmpMask)) != transparent)
            DC.drawPixel((uchar*)d X_ARG(dstLeft),c);
        INC_DEST(dstLeft,d);
        INC_DEST(srcLeft,s);
        INC_X(dstLeft);
        INC_X(srcLeft);
        }
}

#undef  DRAW_SCAN_PROTO

