/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the 2D reference rasteriser functions
*               for the SciTech SNAP Graphics Architecture.
*
****************************************************************************/

#include "ref2dint.h"

/*----------------------------- Implementation ----------------------------*/

#define BUF_INCR                    1
#define COLOR_TYPE                  N_uint8
#define FULL_COLOR                  N_uint8
#define COLOR_WHITE                 0xFF
#define PATT_TYPE                   b8
#define INC_DEST(x,addr)            (addr)++
#define INC_X(x)
#define DEC_DEST(x,addr)            (addr)--
#define DEC_X(x)
#define REPLACE_PIXEL(dest,x,color) PLOT_S(dest,color)
#define READ_PIXEL(src,x)           (src)

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void REF2D_fillScan_8_S(
    uchar *addr DCL_X_ARG,
    int width,
    GA_color color)
{
    DWORD_FILL(addr,width,color);
}

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_fillScan_8##rop (uchar *addr DCL_X_ARG,int width,GA_color color)
#include "raster/fillscan.c"

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void REF2D_drawScan_8_S(
    uchar *addr DCL_X_ARG,
    int width)
{
    ulong color = DC.foreColor;
    DWORD_FILL(addr,width,color);
}

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawScan_8##rop (uchar *addr DCL_X_ARG,int width)
#include "raster/scan.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawPattScan_8##rop (uchar *addr,int x1,int y,int width)
#include "raster/monoscan.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawColorPattScan_8##rop (uchar *addr,int x1,int y,int width)
#include "raster/colrscan.c"

#define DRAW_SCAN_PROTO() \
    void REF2D_drawScanPatt_8 (uchar *addr,int x1,int y,int width)
#include "raster/scanmono.c"

#define DRAW_SCAN_PROTO() \
    void REF2D_drawScanColorPatt_8 (uchar *addr,int x1,int y,int width)
#include "raster/scancolr.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawMonoMSBTrans_8##rop (uchar *addr,uchar *image DCL_X_ARG,int width)
#include "raster/msbtran.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawMonoMSBOpaque_8##rop (uchar *addr,uchar *image DCL_X_ARG,int width)
#include "raster/msbopaq.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawClipMonoMSBTrans_8##rop (int x,uchar *addr,uchar *image,int width,int clipLeft)
#include "raster/cmsbtran.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawClipMonoMSBOpaque_8##rop (int x,uchar *addr,uchar *image,int width,int clipLeft)
#include "raster/cmsbopaq.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawMonoLSBTrans_8##rop (uchar *addr,uchar *image DCL_X_ARG,int width)
#include "raster/lsbtran.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawMonoLSBOpaque_8##rop (uchar *addr,uchar *image DCL_X_ARG,int width)
#include "raster/lsbopaq.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawClipMonoLSBTrans_8##rop (int x,uchar *addr,uchar *image,int width,int clipLeft)
#include "raster/clsbtran.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_drawClipMonoLSBOpaque_8##rop (int x,uchar *addr,uchar *image,int width,int clipLeft)
#include "raster/clsbopaq.c"

#define DRAW_SCAN_PROTO() \
    void REF2D_drawMonoCursor_8(uchar *addr,uchar *maskPtr,uchar *imagePtr DCL_X_ARG,int width)
#include "raster/monocurs.c"

#define DRAW_SCAN_PROTO() \
    void REF2D_drawColorCursor_8(uchar *addr,uchar *maskPtr,uchar *imagePtr DCL_X_ARG,int width)
#include "raster/colrcurs.c"

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void REF2D_readScan_8(
    uchar *d,
    uchar *s DCL_X_ARG2,
    int width)
{
    // Force DWORD source alignment
    while (width && ((uint)s & 3) != 0) {
        *d++ = *s++;
        width--;
        }
    if (width)
        memcpy(d,s,width);
}

/****************************************************************************
PARAMETERS:
addr    - Memory address to draw pixel at
width   - Number of pixels to draw for the scanline

REMARKS:
This function draws a solid scanline at the specified address in the
current color. We abstract this function so that we can keep the inner loops
of our reference rasteriser color depth 'clean', however we pass the address
of the pixel to draw to avoid address computations per pixel to give us
performance close to an optimised rasteriser function.
****************************************************************************/
void REF2D_bltScan_8_S(
    uchar *d,
    uchar *s DCL_X_ARG2,
    int width)
{
    if (width >= 8) {
        // Force DWORD destination alignment
        while (((uint)d & 3) != 0) {
            *d++ = *s++;
            width--;
            }
        memcpy(d,s,width);
        }
    else
        memcpy(d,s,width);
}

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_bltScan_8##rop (uchar *din,uchar *sin DCL_X_ARG2,int width)
#include "raster/bltscan.c"

#define DRAW_SCAN_PROTO() \
    void REF2D_bltScanPatt_8 (uchar *din,uchar *sin DCL_X_ARG3,int x1,int y,int width)
#include "raster/bltmono.c"

#define DRAW_SCAN_PROTO() \
    void REF2D_bltScanColorPatt_8 (uchar *din,uchar *sin DCL_X_ARG3,int x1,int y,int width)
#include "raster/bltcolor.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_srcTransBltScan_8##rop (uchar *din,uchar *sin DCL_X_ARG2,int width)
#include "raster/transcan.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_dstTransBltScan_8##rop (uchar *din,uchar *sin DCL_X_ARG2,int width)
#include "raster/dstscan.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_stretchScan2x_8##rop (uchar *din,uchar *sin DCL_X_ARG2,int width)
#include "raster/stretch2.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_stretchScan_8##rop (uchar *din,uchar *sin DCL_X_ARG2,int width,int *repCounts)
#include "raster/stretch.c"

#define DRAW_LINEXPOSLO_PROTO(rop) \
    void REF2D_drawLineXPosLo_8##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr)
#define DRAW_LINEXPOSHI_PROTO(rop) \
    void REF2D_drawLineXPosHi_8##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr)
#define DRAW_LINEXNEGLO_PROTO(rop) \
    void REF2D_drawLineXNegLo_8##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr)
#define DRAW_LINEXNEGHI_PROTO(rop) \
    void REF2D_drawLineXNegHi_8##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr)
#include "raster/line.c"

#define DRAW_LINEXPOSLO_PROTO(rop) \
    void REF2D_drawStippleLineXPosLo_8##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent)
#define DRAW_LINEXPOSHI_PROTO(rop) \
    void REF2D_drawStippleLineXPosHi_8##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent)
#define DRAW_LINEXNEGLO_PROTO(rop) \
    void REF2D_drawStippleLineXNegLo_8##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent)
#define DRAW_LINEXNEGHI_PROTO(rop) \
    void REF2D_drawStippleLineXNegHi_8##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong mask,ulong stipple,int transparent)
#include "raster/stipline.c"

#define DRAW_LINEXPOSLO_PROTO(rop) \
    void REF2D_drawStyleLineXPosLo_8##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent)
#define DRAW_LINEXPOSHI_PROTO(rop) \
    void REF2D_drawStyleLineXPosHi_8##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent)
#define DRAW_LINEXNEGLO_PROTO(rop) \
    void REF2D_drawStyleLineXNegLo_8##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent)
#define DRAW_LINEXNEGHI_PROTO(rop) \
    void REF2D_drawStyleLineXNegHi_8##rop (uchar *addr DCL_X_ARG,int d,int Eincr,int NEincr,int count,int fincr,ulong styleMask,ulong styleStep,ulong styleValue,int transparent)
#include "raster/stylline.c"

/****************************************************************************
PARAMETERS:
srcPal      - Source bitmap palette to build translation vector for
dstPal      - Destination bitmap palette to build translation vector for
srcColor    - Number of colors in source palette

REMARKS:
This function builds a translation vector for translating 4bpp or 8bpp
color values to the destination device color values, by finding the closest
color in the destination palette that matches the color in the source
palette.
****************************************************************************/
GA_color * REF2D_buildTranslateVector_8(
    GA_color *t,
    GA_palette *dstPal,
    GA_palette *srcPal,
    int srcColors)
{
    GA_color    *p = t;

    while (srcColors--) {
        int         j,closest = 0,d,distance = 0x7FFFL;
        GA_palette  *dp = dstPal;

        for (j = 0; j < 256; j++,dp++) {
            if ((d = ABS((int)srcPal->Red - (int)dp->Red)
                   + ABS((int)srcPal->Green - (int)dp->Green)
                   + ABS((int)srcPal->Blue - (int)dp->Blue)) < distance) {
                distance = d;
                closest = j;
                }
            }
        *p++ = closest;
        srcPal++;
        }
    return t;
}

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_convScan4_8##rop (uchar *din,uchar *sin DCL_X_ARG,int width,int odd,GA_color *t)
#include "raster/cvscan4.c"

#define DRAW_SCAN_PROTO(rop) \
    void REF2D_convScan8_8##rop (uchar *din,uchar *sin DCL_X_ARG,int width,GA_color *t)
#include "raster/cvscan8.c"

#define DRAW_ELLIPSE_PROTO(rop) \
    void REF2D_drawEllipse_8##rop(int ileft,int itop,int A,int B)
#include "raster/ellipse.c"

#define DRAW_ELLIPSE_PROTO(rop) \
    void REF2D_clipEllipse_8##rop(int ileft,int itop,int A,int B,int clipLeft,int clipTop,int clipRight,int clipBottom)
#include "raster/cellipse.c"

