/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Header file for the SciTech SNAP Graphics 2D region manager
*               internal functions.
*
****************************************************************************/

#ifndef __REGINT_H
#define __REGINT_H

#include "snap/common/list.h"
#include "snap/common/mem.h"
#include "snap/graphics.h"

/*---------------------- Macros and type definitions ----------------------*/

#pragma pack(4)

/* Region manipulation macros for maximum speed */

#define RGN_tmpRectRegion(r)                \
{                                           \
    RG.tmpY1.y = (r)->rect.top;             \
    RG.tmpY2.y = (r)->rect.bottom;          \
    RG.tmpX1.x = (r)->rect.left;            \
    RG.tmpX2.x = (r)->rect.right;           \
    RG.rtmp.rect = (r)->rect;               \
    (r) = &RG.rtmp;                         \
}

#define RGN_tmpRectRegion2(r,rt)            \
{                                           \
    RG.rtmp2.rect = rt;                     \
    (r) = &RG.rtmp2;                        \
}

/****************************************************************************
REMARKS:
Structure for block list of segments.

MEMBERS:
bl          - Memory pool containg objects
freeList    - Start of free list for pool
****************************************************************************/
typedef struct {
    N_blockList bl;
    GA_segment  *freeList;
    } segmentList_t;

/****************************************************************************
REMARKS:
Structure for block list of spans.

MEMBERS:
bl          - Memory pool containg objects
freeList    - Start of free list for pool
****************************************************************************/
typedef struct {
    N_blockList bl;
    GA_span     *freeList;
    } spanList_t;

/****************************************************************************
REMARKS:
Structure for block list of regions.

MEMBERS:
bl          - Memory pool containg objects
freeList    - Start of free list for pool
****************************************************************************/
typedef struct {
    N_blockList bl;
    GA_region   *freeList;
    } regionList_t;

/****************************************************************************
REMARKS:
Structure to store the current state of the 2D region manager functions.

MEMBERS:
****************************************************************************/
typedef struct {
    GA_segment          tmpX1;
    GA_segment          tmpX2;
    GA_span             tmpY1;
    GA_span             tmpY2;
    GA_region           rtmp;
    GA_region           rtmp2;
    segmentList_t       segList;
    spanList_t          spanList;
    regionList_t        regionList;
    } RGN_state;

#pragma pack()

/*---------------------------- Global Variables ---------------------------*/

extern RGN_state        RGN_devCtx;

#define RG  RGN_devCtx

/*-------------------------- Function Prototypes --------------------------*/

#ifdef  __cplusplus
extern "C" {                        /* Use "C" linkage when in C++ mode */
#endif

/* Internal region management routines */

ibool       RGN_resizeRegionList(void);
ibool       RGN_resizeSpanList(void);
ibool       RGN_resizeSegmentList(void);
GA_segment *RGN_newSegment(void);
GA_segment *RGN_copySegment(GA_segment *s);
GA_segment *RGN_copySegmentOfs(GA_segment *s,int xOffset);
void        RGN_freeSegment(GA_segment *s);
GA_span *   RGN_newSpan(void);
GA_span *   RGN_createSpan(int y,int x1,int x2);
GA_span *   RGN_createSpan2(int y,int x1,int x2,int x3,int x4);
GA_span *   RGN_copySpan(GA_span *s);
GA_span *   RGN_copySpanOfs(GA_span *s,int xOffset,int yOffset);
GA_span *   RGN_copySpans(GA_span *s);
GA_span *   RGN_copySpansOfs(GA_span *s,int xOffset,int yOffset);
void        RGN_freeSpan(GA_span *s);
GA_span *   RGN_rectRegion(GA_region *r);

/* Region management and algegra functions */

GA_region * NAPI RGN_NewRegion(void);
GA_region * NAPI RGN_NewRectRegion(N_int32 left,N_int32 top,N_int32 right,N_int32 bottom);
GA_region * NAPI RGN_CopyRegion(const GA_region *s);
void        NAPI RGN_CopyIntoRegion(GA_region *d,const GA_region *s);
void        NAPI RGN_ClearRegion(GA_region *r);
void        NAPI RGN_FreeRegion(GA_region *r);
GA_region * NAPI RGN_SectRegion(const GA_region *r1,const GA_region *r2);
GA_region * NAPI RGN_SectRegionRect(const GA_region *r1,const GA_rect *r);
ibool       NAPI RGN_UnionRegion(GA_region *r1,const GA_region *r2);
ibool       NAPI RGN_UnionRegionRect(GA_region *r1,const GA_rect *r);
ibool       NAPI RGN_UnionRegionOfs(GA_region *r1,const GA_region *r2,N_int32 dx,N_int32 dy);
ibool       NAPI RGN_DiffRegion(GA_region *r1,const GA_region *r2);
ibool       NAPI RGN_DiffRegionRect(GA_region *r1,const GA_rect *r);
void        NAPI RGN_OptimizeRegion(GA_region *r);
void        NAPI RGN_OffsetRegion(GA_region *r,N_int32 dx,N_int32 dy);
ibool       NAPI RGN_IsEmptyRegion(const GA_region *r);
ibool       NAPI RGN_IsEqualRegion(const GA_region *r1,const GA_region *r2);
ibool       NAPI RGN_PtInRegion(const GA_region *r,N_int32 x,N_int32 y);
void        NAPI RGN_TraverseRegion(GA_region *rgn,GA_regionCallback doRect,void *parms);

#ifdef  __cplusplus
}                                   /* End of "C" linkage for C++       */
#endif

#endif  /* __REGINT_H */

