/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Region traversal routines
*
****************************************************************************/

#include "ref2dint.h"

/*------------------------- Implementation --------------------------------*/

/****************************************************************************
PARAMETERS:
rgn     - Region to traverse
doRect  - Callback function to call for every rectangle processed
parms   - Optional parameters passed to region callback function

REMARKS:
This function traverses the definition of the region, calling the supplied
callback function once for every rectangle in union of rectangles that
make up the complex region.
****************************************************************************/
void NAPI RGN_TraverseRegion(
    GA_region *rgn,
    GA_regionCallback doRect,
    void *parms)
{
    GA_segment  *seg;
    GA_span     *s = rgn->spans;
    GA_rect     r;

    if (!s)
        doRect(&rgn->rect,parms);
    else {
        while (s->next) {
            r.top = s->y;
            r.bottom = s->next->y;
            for (seg = s->seg; seg; seg = seg->next->next) {
                r.left = seg->x;
                r.right = seg->next->x;
                doRect(&r,parms);
                }
            s = s->next;
            }
        }
}

/****************************************************************************
PARAMETERS:
rgn - Region to test
x   - x coordinate to test for inclusion
y   - y coordinate to test for inclusion

RETURNS:
True if the point is contained in the region, false if not.

REMARKS:
This function determines if a specified point is contained within a
particular region. Note that if the region has a hole it in, and the
point lies within the hole, then the point is classified as not being
included in the region.
****************************************************************************/
ibool NAPI RGN_PtInRegion(
    const GA_region *rgn,
    N_int32 x,
    N_int32 y)
{
    int         topy,boty;
    GA_segment  *seg;
    GA_span     *s = rgn->spans;

    if (!s)
        return GA_ptInRect(x,y,rgn->rect);
    else {
        while (s->next) {
            topy = s->y;
            boty = s->next->y;
            for (seg = s->seg; seg; seg = seg->next->next)
                if (topy <= y && y < boty && seg->x <= x && x < seg->next->x)
                    return true;
            s = s->next;
            }
        }
    return false;
}
