/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the shadow buffer filter driver support
*               routines in the SNAP Reference Rasteriser.
*
****************************************************************************/

#include "shadint.h"

/*----------------------------- Implementation ----------------------------*/

#define BUF_INCR    2
#define COLOR_TYPE  ushort

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the rectangle to draw
top     - Top coordinate of the rectangle to draw
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
****************************************************************************/
void NAPI SHADOW_DrawRect_16_Solid(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height)
{
    int     bytes,startAddr,endAddr,newAddr,bank,bytesStart,bytesEnd;
    uchar   *d,*bankedMem;

    bytes = width*2;
    startAddr = (DC.mi.BytesPerScanLine * top) + left*2;
    bank = (startAddr += DC.buf.Offset) >> 16;
    bankedMem = DC.hwCtx->BankedMem;
    d = bankedMem + (startAddr & 0xFFFF);

    if (DC.cntBank != bank)
        DC.SetBank(DC.cntBank = bank);
    while (height--) {
        newAddr = startAddr + DC.mi.BytesPerScanLine;
        if ((bank = newAddr >> 16) != DC.cntBank) {
            // Bank boundary is present before next scanline
            if ((endAddr = startAddr + bytes) >> 16 != DC.cntBank) {
                // Data to copy is split by bank boundary
                bytesStart = (endAddr & ~0xFFFF) - startAddr;
                bytesEnd = endAddr & 0xFFFF;
                rep_stosdw(d,bytesStart>>1,DC.foreColor);
                DC.SetBank(DC.cntBank = bank);
                rep_stosdw(bankedMem,bytesEnd>>1,DC.foreColor);
                }
            else {
                // Data to copy is contained within the current bank
                rep_stosdw(d,bytes>>1,DC.foreColor);
                DC.SetBank(DC.cntBank = bank);
                }
            d = bankedMem + (newAddr & 0xFFFF);
            }
        else {
            // Scanline is contained within the current bank
            rep_stosdw(d,bytes>>1,DC.foreColor);
            d += DC.mi.BytesPerScanLine;
            }
        startAddr += DC.mi.BytesPerScanLine;
        }
}

#define DRAW_LINEXPOSLO_PROTO() \
    void SHADOW_drawLineXPosLo_16 (uchar *dstBase,int dstAddr,uchar *srcAddr,int d,int Eincr,int NEincr,int count,int fincr)
#define DRAW_LINEXPOSHI_PROTO() \
    void SHADOW_drawLineXPosHi_16 (uchar *dstBase,int dstAddr,uchar *srcAddr,int d,int Eincr,int NEincr,int count,int fincr)
#define DRAW_LINEXNEGLO_PROTO() \
    void SHADOW_drawLineXNegLo_16 (uchar *dstBase,int dstAddr,uchar *srcAddr,int d,int Eincr,int NEincr,int count,int fincr)
#define DRAW_LINEXNEGHI_PROTO() \
    void SHADOW_drawLineXNegHi_16 (uchar *dstBase,int dstAddr,uchar *srcAddr,int d,int Eincr,int NEincr,int count,int fincr)
#include "raster/shadline.c"
