/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Header file for the SciTech SNAP Graphics 2D Shadow Buffer
*               filter driver internal functions.
*
****************************************************************************/

#ifndef __SHADINT_H
#define __SHADINT_H

#include "ref2dint.h"

/*---------------------- Macros and type definitions ----------------------*/

#pragma pack(4)

/****************************************************************************
REMARKS:
Structure to store the current state of the 2D portrait filter driver.

MEMBERS:
hwCtx       - Hardware device context bound to this filter driver
shared      - True if loaded into shared memory
enabled     - True if shadow buffer is enabled
cntMode     - Current display mode for driver
mi          - Mode information block for the current hardware mode
byteShift   - Byte shift for address arithmetic for display mode
buf         - Current draw buffer
cntBank     - Current bank value for framebuffer
ref2d       - Pointer to the ref2d driver
refState    - Pointer to the ref2d driver internal state
bufferSize  - Size of the shadow buffer in bytes
buffer      - Shadow buffer pointer
****************************************************************************/
typedef struct {
    GA_devCtx       *hwCtx;
    ibool           enabled;
    N_uint32        cntMode;
    GA_modeInfo     mi;
    N_int32         byteShift;
    N_int32         mix;
    GA_color        foreColor;
    N_uint32        planeMask;
    N_int32         srcBlendFunc;
    N_int32         dstBlendFunc;
    N_int32         cntBank;
    GA_buffer       buf;
    REF2D_driver    *ref2d;
    REF2D_state     *refState;
    PE_MODULE       *hModBuffer;
    ulong           bufferSize;
    void            *LinearMem;
    void            *buffer;
    void            (*UpdateScreen)(int offset,int left,int top,int width,int height);
    void            (*drawLineXPosLo)(uchar *dstBase,int dstAddr,uchar *srcAddr,int d,int Eincr,int NEincr,int count,int fincr);
    void            (*drawLineXPosHi)(uchar *dstBase,int dstAddr,uchar *srcAddr,int d,int Eincr,int NEincr,int count,int fincr);
    void            (*drawLineXNegLo)(uchar *dstBase,int dstAddr,uchar *srcAddr,int d,int Eincr,int NEincr,int count,int fincr);
    void            (*drawLineXNegHi)(uchar *dstBase,int dstAddr,uchar *srcAddr,int d,int Eincr,int NEincr,int count,int fincr);
    void            (NAPIP SetBank)(N_int32 bank);
    void            (NAPIP DrawRect)(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
    void            (NAPIP DrawRectExt)(N_int32 left,N_int32 top,N_int32 width,N_int32 height,GA_color color,N_int32 mix);
    void            (NAPIP DrawLineInt)(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast);
    void            (NAPIP DrawBresenhamLine)(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags);
    void            (NAPIP DrawStippleLineInt)(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast);
    void            (NAPIP DrawBresenhamStippleLine)(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags);
    void            (NAPIP BitBlt)(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
    } SHADOW_state;

/****************************************************************************
REMARKS:
Macros to set a state 2d vector if not overriden by the filter driver
****************************************************************************/
#define SET_STATE2D_VECTOR(name)    \
    if (!SHADOW_state2dPT.name) SHADOW_state2dPT.name = SHADOW_state2dREF.name

/****************************************************************************
REMARKS:
Macros to set a draw 2d vector if not overriden by the filter driver
****************************************************************************/
#define SET_DRAW2D_VECTOR(name) \
    if (SHADOW_draw2dStub.name && !SHADOW_draw2dPT.name) SHADOW_draw2dPT.name = SHADOW_draw2dREF.name

#pragma pack()

/*---------------------------- Global Variables ---------------------------*/

extern SHADOW_state SHADOW_devCtx;
extern REF2D_driver     SHADOW_drv;
extern GA_cursorFuncs   SHADOW_cursorREF;
extern GA_2DStateFuncs  SHADOW_state2dREF;
extern GA_2DRenderFuncs SHADOW_draw2dREF;
extern GA_2DRenderFuncs _VARAPI SHADOW_draw2dCnt;

#undef  DC
#undef  DRV
#define DC      SHADOW_devCtx
#define DC2D    (*SHADOW_devCtx.refState)
#define DRV     SHADOW_drv
#define CURSOR  SHADOW_cursorREF
#define STATE2D SHADOW_state2dREF
#define DRAW2D  SHADOW_draw2dREF

/*-------------------------- Function Prototypes --------------------------*/

#ifdef  __cplusplus
extern "C" {                        /* Use "C" linkage when in C++ mode */
#endif

/* Library initialisation entry points */

ibool       NAPI SHADOW_QueryFunctions(N_uint32 id,void *funcs);
N_int32     NAPI SHADOW_SetDrawBufferSW(GA_buffer *drawBuf,void *framebuffer,N_int32 bitsPerPixel,GA_pixelFormat *pf,GA_devCtx *hwCtx,N_int32 softwareOnly);
void        NAPI SHADOW_SetDrawSurface(void *surface,N_int32 xRes,N_int32 yRes,N_int32 bytesPerLine,N_int32 bitsPerPixel,GA_pixelFormat *pf);
void        NAPI SHADOW_UnloadDriver(void);
ibool       NAPI SHADOW_InitFilter(GA_modeInfo *modeInfo,N_int32 transferStart,struct _REF2D_driver *ref2d,struct _REF2D_driver **drv);
void        NAPI SHADOW_DeInitFilter(void);
GA_buffer * NAPI SHADOW_GetDrawBuffer(void);
void        NAPI SHADOW_DrawRectExtSW(void *buffer,N_int32 dstPitch,N_int32 left,N_int32 top,N_int32 width,N_int32 height,GA_color color,N_int32 mix);
N_int32     NAPI SHADOW_ForceSoftwareOnly(N_int32 enable);
N_uint32    NAPI SHADOW_SetColorCompareMask(N_uint32 mask);
void        NAPI SHADOW_PostSwitchPhysicalResolution(void);

/* Functions to update the screen */

void        SHADOW_UpdateScreenBnk(int offset,int left,int top,int width,int height);
void        SHADOW_UpdateScreen4(int offset,int left,int top,int width,int height);
void        SHADOW_UpdateCursor4(int offset,int left,int top,int width,int height);

/* Generic shadow buffer functions */

void        NAPI SHADOW_DrawRect_Generic(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI SHADOW_DrawRectExt_Generic(N_int32 left,N_int32 top,N_int32 width,N_int32 height,GA_color color,N_int32 mix);
void        NAPI SHADOW_DrawLineInt_Generic(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast);
void        NAPI SHADOW_DrawBresenhamLine_Generic(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags);

/* 4bpp shadow buffer functions */

void        NAPI SHADOW_DrawRect_4_Solid(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI SHADOW_DrawRectExt_4_Solid(N_int32 left,N_int32 top,N_int32 width,N_int32 height,GA_color color,N_int32 mix);
void        NAPI SHADOW_DrawLineInt_4_Solid(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast);
void        NAPI SHADOW_DrawLineInt_4_Generic(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast);
void        NAPI SHADOW_DrawBresenhamLine_4_Solid(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags);
void        NAPI SHADOW_DrawBresenhamLine_4_Generic(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags);

/* 8bpp shadow buffer functions */

void        NAPI SHADOW_DrawRect_8_Solid(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        SHADOW_drawLineXPosLo_8(uchar *dstBase,int dstAddr,uchar *srcAddr,int d,int Eincr,int NEincr,int count,int fincr);
void        SHADOW_drawLineXPosHi_8(uchar *dstBase,int dstAddr,uchar *srcAddr,int d,int Eincr,int NEincr,int count,int fincr);
void        SHADOW_drawLineXNegLo_8(uchar *dstBase,int dstAddr,uchar *srcAddr,int d,int Eincr,int NEincr,int count,int fincr);
void        SHADOW_drawLineXNegHi_8(uchar *dstBase,int dstAddr,uchar *srcAddr,int d,int Eincr,int NEincr,int count,int fincr);

/* 16bpp shadow buffer functions */

void        NAPI SHADOW_DrawRect_16_Solid(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        SHADOW_drawLineXPosLo_16(uchar *dstBase,int dstAddr,uchar *srcAddr,int d,int Eincr,int NEincr,int count,int fincr);
void        SHADOW_drawLineXPosHi_16(uchar *dstBase,int dstAddr,uchar *srcAddr,int d,int Eincr,int NEincr,int count,int fincr);
void        SHADOW_drawLineXNegLo_16(uchar *dstBase,int dstAddr,uchar *srcAddr,int d,int Eincr,int NEincr,int count,int fincr);
void        SHADOW_drawLineXNegHi_16(uchar *dstBase,int dstAddr,uchar *srcAddr,int d,int Eincr,int NEincr,int count,int fincr);

/* 24bpp shadow buffer functions */

void        SHADOW_drawLineXPosLo_24(uchar *dstBase,int dstAddr,uchar *srcAddr,int d,int Eincr,int NEincr,int count,int fincr);
void        SHADOW_drawLineXPosHi_24(uchar *dstBase,int dstAddr,uchar *srcAddr,int d,int Eincr,int NEincr,int count,int fincr);
void        SHADOW_drawLineXNegLo_24(uchar *dstBase,int dstAddr,uchar *srcAddr,int d,int Eincr,int NEincr,int count,int fincr);
void        SHADOW_drawLineXNegHi_24(uchar *dstBase,int dstAddr,uchar *srcAddr,int d,int Eincr,int NEincr,int count,int fincr);

/* 32bpp shadow buffer functions */

void        NAPI SHADOW_DrawRect_32_Solid(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        SHADOW_drawLineXPosLo_32(uchar *dstBase,int dstAddr,uchar *srcAddr,int d,int Eincr,int NEincr,int count,int fincr);
void        SHADOW_drawLineXPosHi_32(uchar *dstBase,int dstAddr,uchar *srcAddr,int d,int Eincr,int NEincr,int count,int fincr);
void        SHADOW_drawLineXNegLo_32(uchar *dstBase,int dstAddr,uchar *srcAddr,int d,int Eincr,int NEincr,int count,int fincr);
void        SHADOW_drawLineXNegHi_32(uchar *dstBase,int dstAddr,uchar *srcAddr,int d,int Eincr,int NEincr,int count,int fincr);

/* 2D state reference functions */

N_int32     NAPI SHADOW_SetDrawBuffer(GA_buffer *drawBuf);
void        NAPI SHADOW_EnableDirectAccess(void);
void        NAPI SHADOW_DisableDirectAccess(void);
N_int32     NAPI SHADOW_SetMix(N_int32 mix);
void        NAPI SHADOW_SetForeColor(GA_color color);
void        NAPI SHADOW_SetPlaneMask(N_uint32 mask);
void        NAPI SHADOW_SetSrcBlendFunc(N_int32 blendFunc);
void        NAPI SHADOW_SetDstBlendFunc(N_int32 blendFunc);
void        NAPI SHADOW_SetBlendFunc(N_int32 srcBlendFunc,N_int32 dstBlendFunc);

/* 2D generic reference functions */

void        NAPI SHADOW_PutPixel(N_int32 x,N_int32 y);
void        NAPI SHADOW_DrawEllipse(N_int32 left,N_int32 top,N_int32 A,N_int32 B);
void        NAPI SHADOW_ClipEllipse(N_int32 left,N_int32 top,N_int32 A,N_int32 B,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
void        NAPI SHADOW_DrawScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI SHADOW_DrawPattScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI SHADOW_DrawColorPattScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI SHADOW_DrawEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI SHADOW_DrawPattEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI SHADOW_DrawColorPattEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI SHADOW_DrawFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI SHADOW_DrawPattFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI SHADOW_DrawColorPattFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI SHADOW_DrawRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI SHADOW_DrawPattRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI SHADOW_DrawColorPattRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI SHADOW_DrawRectPatt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,N_int32 rop3);
void        NAPI SHADOW_DrawRectColorPatt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,N_int32 rop3);
void        NAPI SHADOW_DrawTrap(GA_trap *trap);
void        NAPI SHADOW_DrawPattTrap(GA_trap *trap);
void        NAPI SHADOW_DrawColorPattTrap(GA_trap *trap);
void        NAPI SHADOW_DrawLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast);
void        NAPI SHADOW_DrawBresenhamLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags);
void        NAPI SHADOW_DrawStippleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent);
void        NAPI SHADOW_DrawBresenhamStippleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent);
void        NAPI SHADOW_PutMonoImageMSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent);
void        NAPI SHADOW_PutMonoImageLSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent);
void        NAPI SHADOW_ClipMonoImageMSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipWidth,N_int32 clipHeight);
void        NAPI SHADOW_ClipMonoImageLSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipWidth,N_int32 clipHeight);
void        NAPI SHADOW_BitBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI SHADOW_BitBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,N_int32 flipY);
void        NAPI SHADOW_BitBltPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI SHADOW_BitBltPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void        NAPI SHADOW_BitBltColorPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI SHADOW_BitBltColorPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void        NAPI SHADOW_SrcTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI SHADOW_SrcTransBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent,N_int32 flipY);
void        NAPI SHADOW_DstTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI SHADOW_DstTransBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent,N_int32 flipY);
void        NAPI SHADOW_StretchBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix);
void        NAPI SHADOW_StretchBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix,N_int32 flipY);
N_int32     NAPI SHADOW_ConvertBltSys_Obsolete(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix,N_int32 flipY);
N_int32     NAPI SHADOW_StretchConvertBltSys_Obsolete(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix,N_int32 flipY);
void        NAPI SHADOW_BitBltFx(N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,GA_bltFx *fx);
void        NAPI SHADOW_BitBltFxSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,GA_bltFx *fx);
void        NAPI SHADOW_UpdateScreen(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
N_int32     NAPI SHADOW_DrawClippedLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32     NAPI SHADOW_DrawClippedBresenhamLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32     NAPI SHADOW_DrawClippedStippleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32     NAPI SHADOW_DrawClippedBresenhamStippleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
void        NAPI SHADOW_DrawRectExt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,GA_color color,N_int32 mix);
void        NAPI SHADOW_DrawStyleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent);
void        NAPI SHADOW_DrawBresenhamStyleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent);
N_int32     NAPI SHADOW_DrawClippedStyleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32     NAPI SHADOW_DrawClippedBresenhamStyleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);

/* Assembler stubs to call the functions via the global REF_draw2d */

GA_color    NAPI SSTUB_GetPixel(N_int32 x,N_int32 y);
void        NAPI SSTUB_PutPixel(N_int32 x,N_int32 y);
void        NAPI SSTUB_DrawEllipse(N_int32 left,N_int32 top,N_int32 A,N_int32 B);
void        NAPI SSTUB_ClipEllipse(N_int32 left,N_int32 top,N_int32 A,N_int32 B,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
void        NAPI SSTUB_DrawScan(N_int32 y,N_int32 x1,N_int32 x2);
void        NAPI SSTUB_DrawPattScan(N_int32 y,N_int32 x1,N_int32 x2);
void        NAPI SSTUB_DrawColorPattScan(N_int32 y,N_int32 x1,N_int32 x2);
void        NAPI SSTUB_DrawScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI SSTUB_DrawPattScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI SSTUB_DrawColorPattScanList(N_int32 y,N_int32 length,N_int16 *scans);
void        NAPI SSTUB_DrawEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI SSTUB_DrawPattEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI SSTUB_DrawColorPattEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI SSTUB_DrawFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI SSTUB_DrawPattFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI SSTUB_DrawColorPattFatEllipseList(N_int32 y,N_int32 length,N_int32 height,N_int16 *scans);
void        NAPI SSTUB_DrawRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI SSTUB_DrawPattRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI SSTUB_DrawColorPattRect(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
void        NAPI SSTUB_DrawRectPatt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,N_int32 rop3);
void        NAPI SSTUB_DrawRectColorPatt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,N_int32 rop3);
void        NAPI SSTUB_DrawTrap(GA_trap *trap);
void        NAPI SSTUB_DrawPattTrap(GA_trap *trap);
void        NAPI SSTUB_DrawColorPattTrap(GA_trap *trap);
void        NAPI SSTUB_DrawLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast);
void        NAPI SSTUB_DrawBresenhamLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags);
void        NAPI SSTUB_DrawStippleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent);
void        NAPI SSTUB_DrawBresenhamStippleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent);
void        NAPI SSTUB_PutMonoImageMSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent);
void        NAPI SSTUB_PutMonoImageLSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent);
void        NAPI SSTUB_ClipMonoImageMSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipWidth,N_int32 clipHeight);
void        NAPI SSTUB_ClipMonoImageLSBSys(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipWidth,N_int32 clipHeight);
void        NAPI SSTUB_BitBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI SSTUB_BitBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,N_int32 flipY);
void        NAPI SSTUB_BitBltPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI SSTUB_BitBltPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void        NAPI SSTUB_BitBltColorPatt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3);
void        NAPI SSTUB_BitBltColorPattSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 rop3,N_int32 flipY);
void        NAPI SSTUB_SrcTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI SSTUB_SrcTransBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent,N_int32 flipY);
void        NAPI SSTUB_DstTransBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
void        NAPI SSTUB_DstTransBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent,N_int32 flipY);
void        NAPI SSTUB_StretchBlt(N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix);
void        NAPI SSTUB_StretchBltSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 mix,N_int32 flipY);
N_int32     NAPI SSTUB_ConvertBltSys_Obsolete(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix,N_int32 flipY);
N_int32     NAPI SSTUB_StretchConvertBltSys_Obsolete(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,N_int32 doClip,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom,N_int32 bitsPerPixel,GA_pixelFormat *pixelFormat,GA_palette *dstPal,GA_palette *srcPal,N_int32 dither,N_int32 mix,N_int32 flipY);
void        NAPI SSTUB_BitBltFx(N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,GA_bltFx *fx);
void        NAPI SSTUB_BitBltFxSys(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 srcWidth,N_int32 srcHeight,N_int32 dstLeft,N_int32 dstTop,N_int32 dstWidth,N_int32 dstHeight,GA_bltFx *fx);
void        NAPI SSTUB_GetBitmapSys(void *dstAddr,N_int32 dstPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
void        NAPI SSTUB_UpdateScreen(N_int32 left,N_int32 top,N_int32 width,N_int32 height);
N_int32     NAPI SSTUB_DrawClippedLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32     NAPI SSTUB_DrawClippedBresenhamLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32     NAPI SSTUB_DrawClippedStippleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32     NAPI SSTUB_DrawClippedBresenhamStippleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
void        NAPI SSTUB_DrawRectExt(N_int32 left,N_int32 top,N_int32 width,N_int32 height,GA_color color,N_int32 mix);
void        NAPI SSTUB_DrawStyleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent);
void        NAPI SSTUB_DrawBresenhamStyleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent);
N_int32     NAPI SSTUB_DrawClippedStyleLineInt(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
N_int32     NAPI SSTUB_DrawClippedBresenhamStyleLine(N_int32 x1,N_int32 y1,N_int32 initialError,N_int32 majorInc,N_int32 diagInc,N_int32 count,N_int32 flags,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);

#ifdef  __cplusplus
}                                   /* End of "C" linkage for C++       */
#endif

#endif  /* __SHADINT_H */

