/****************************************************************************
*
*                    SciTech SNAP Graphics Architecture
*
*  ========================================================================
*
*   Copyright (C) 1991-2004 SciTech Software, Inc. All rights reserved.
*
*   This file may be distributed and/or modified under the terms of the
*   GNU General Public License version 2 as published by the Free
*   Software Foundation and appearing in the file LICENSE.GPL included
*   in the packaging of this file.
*
*   Licensees holding a valid Commercial License for this product from
*   SciTech Software, Inc. may use this file in accordance with the
*   Commercial License Agreement provided with the Software.
*
*   This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING
*   THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
*   PURPOSE.
*
*   See http://www.scitechsoft.com/license/ for information about
*   the licensing options available and how to purchase a Commercial
*   License Agreement.
*
*   Contact license@scitechsoft.com if any conditions of this licensing
*   are not clear to you, or you have questions about licensing options.
*
*  ========================================================================
*
* Language:     ANSI C
* Environment:  Any
*
* Description:  Module to implement the shadow buffer filter driver support
*               routines in the SNAP Reference Rasteriser.
*
****************************************************************************/

#include "shadint.h"

// TODO: Delayed and cached updates every 20ms for fast performance. If we
//       do this, it needs to be optional so that we can run without it
//       for environments where we can't initialise a high priority thread
//       to handle the 20ms intervals.
//
//       Also if we do this, it will probaby make a shadow buffer version
//       work faster than for direct linear buffer modes for GUI
//       environments!
//
//       One catch if we do this. We need a way to flush to cached updates
//       automatically if a period of time elapses with no drawing commands
//       being sent to the hardware. Could be done with an OS specific
//       thread that calls a special update function every 20ms or something.

// TODO: Clipped lines!

// TODO: BitBltMasked!

/*--------------------------- Global variables ----------------------------*/

SHADOW_state    SHADOW_devCtx;

REF2D_driver SHADOW_drv = {
    sizeof(SHADOW_drv),
    SHADOW_QueryFunctions,
    SHADOW_SetDrawBufferSW,
    SHADOW_SetDrawSurface,
    NULL,
    SHADOW_UnloadDriver,
    SHADOW_InitFilter,
    SHADOW_DeInitFilter,
    NULL,
    SHADOW_GetDrawBuffer,
    SHADOW_DrawRectExtSW,
    SHADOW_ForceSoftwareOnly,
    SHADOW_SetColorCompareMask,
    SHADOW_PostSwitchPhysicalResolution,
    };

static GA_2DStateFuncs SHADOW_state2dPT = {
    sizeof(GA_2DStateFuncs),
    SHADOW_SetDrawBuffer,
    NULL,                       // SHADOW_IsIdle,
    NULL,                       // SHADOW_WaitTillIdle,
    SHADOW_EnableDirectAccess,
    SHADOW_DisableDirectAccess,
    SHADOW_SetMix,
    SHADOW_SetForeColor,
    NULL,                       // SHADOW_SetBackColor,
    NULL,                       // SHADOW_Set8x8MonoPattern,
    NULL,                       // SHADOW_Use8x8MonoPattern,
    NULL,                       // SHADOW_Use8x8TransMonoPattern,
    NULL,                       // SHADOW_Set8x8ColorPattern,
    NULL,                       // SHADOW_Use8x8ColorPattern,
    NULL,                       // SHADOW_Use8x8TransColorPattern,
    NULL,                       // SHADOW_SetLineStipple,
    NULL,                       // SHADOW_SetLineStippleCount,
    SHADOW_SetPlaneMask,
    SHADOW_SetSrcBlendFunc,
    SHADOW_SetDstBlendFunc,
    NULL,                       // SHADOW_SetAlphaValue,
    NULL,                       // DRIVER_SetupForWindows
    NULL,                       // DRIVER_ResetWindows
    NULL,                       // SHADOW_SetLineStyle,
    NULL,                       // SHADOW_BuildTranslateVector
    SHADOW_SetBlendFunc,
    };

static GA_2DRenderFuncs SHADOW_draw2dPT = {
    sizeof(GA_2DRenderFuncs),
    NULL,                       // SHADOW_GetPixel,
    SHADOW_PutPixel,
    SHADOW_DrawScanList,
    SHADOW_DrawPattScanList,
    SHADOW_DrawColorPattScanList,
    SHADOW_DrawEllipseList,
    SHADOW_DrawPattEllipseList,
    SHADOW_DrawColorPattEllipseList,
    SHADOW_DrawFatEllipseList,
    SHADOW_DrawPattFatEllipseList,
    SHADOW_DrawColorPattFatEllipseList,
    SHADOW_DrawRect,
    NULL,
    SHADOW_DrawPattRect,
    SHADOW_DrawColorPattRect,
    SHADOW_DrawTrap,
    SHADOW_DrawPattTrap,
    SHADOW_DrawColorPattTrap,
    SHADOW_DrawLineInt,
    SHADOW_DrawBresenhamLine,
    SHADOW_DrawStippleLineInt,
    SHADOW_DrawBresenhamStippleLine,
    SHADOW_DrawEllipse,
    SHADOW_ClipEllipse,
    SHADOW_PutMonoImageMSBSys,
    NULL,                       // SHADOW_PutMonoImageMSBLin,
    NULL,                       // SHADOW_PutMonoImageMSBBM,
    SHADOW_PutMonoImageLSBSys,
    NULL,                       // SHADOW_PutMonoImageLSBLin,
    NULL,                       // SHADOW_PutMonoImageLSBBM,
    SHADOW_ClipMonoImageMSBSys,
    NULL,                       // SHADOW_ClipMonoImageMSBLin,
    NULL,                       // SHADOW_ClipMonoImageMSBBM,
    SHADOW_ClipMonoImageLSBSys,
    NULL,                       // SHADOW_ClipMonoImageLSBLin,
    NULL,                       // SHADOW_ClipMonoImageLSBBM,
    SHADOW_BitBlt,
    NULL,                       // SHADOW_BitBltLin,
    SHADOW_BitBltSys,
    NULL,                       // SHADOW_BitBltBM,
    SHADOW_BitBltPatt,
    NULL,                       // SHADOW_BitBltPattLin,
    SHADOW_BitBltPattSys,
    NULL,                       // SHADOW_BitBltPattBM,
    SHADOW_BitBltColorPatt,
    NULL,                       // SHADOW_BitBltColorPattLin,
    SHADOW_BitBltColorPattSys,
    NULL,                       // SHADOW_BitBltColorPattBM,
    SHADOW_SrcTransBlt,
    NULL,                       // SHADOW_SrcTransBltLin,
    SHADOW_SrcTransBltSys,
    NULL,                       // SHADOW_SrcTransBltBM,
    SHADOW_DstTransBlt,
    NULL,                       // SHADOW_DstTransBltLin,
    SHADOW_DstTransBltSys,
    NULL,                       // SHADOW_DstTransBltBM,
    SHADOW_StretchBlt,
    NULL,                       // SHADOW_StretchBltLin,
    SHADOW_StretchBltSys,
    NULL,                       // SHADOW_StretchBltBM,
    SHADOW_ConvertBltSys_Obsolete,
    NULL,                       // SHADOW_ConvertBltBM_Obsolete,
    SHADOW_StretchConvertBltSys_Obsolete,
    NULL,                       // SHADOW_StretchConvertBltBM_Obsolete,
    NULL,                       // SHADOW_BitBltFxTest,
    SHADOW_BitBltFx,
    NULL,                       // SHADOW_BitBltFxLin,
    SHADOW_BitBltFxSys,
    NULL,                       // SHADOW_BitBltFxBM,
    NULL,                       // SHADOW_GetBitmapSys,
    NULL,                       // SHADOW_GetBitmapBM,
    SHADOW_UpdateScreen,
    SHADOW_DrawClippedLineInt,
    SHADOW_DrawClippedBresenhamLine,
    SHADOW_DrawClippedStippleLineInt,
    SHADOW_DrawClippedBresenhamStippleLine,
    NULL,                       // SHADOW_BitBltPlaneMasked,
    NULL,                       // SHADOW_BitBltPlaneMaskedLin,
    NULL,                       // SHADOW_BitBltPlaneMaskedSys,
    NULL,                       // SHADOW_BitBltPlaneMaskedBM,
    NULL,                       // SHADOW_DrawRectLin,
    SHADOW_DrawRectExt,
    SHADOW_DrawStyleLineInt,
    SHADOW_DrawBresenhamStyleLine,
    SHADOW_DrawClippedStyleLineInt,
    SHADOW_DrawClippedBresenhamStyleLine,
    };

static GA_2DRenderFuncs SHADOW_draw2dStub = {
    sizeof(GA_2DRenderFuncs),
    SSTUB_GetPixel,
    SSTUB_PutPixel,
    SSTUB_DrawScanList,
    SSTUB_DrawPattScanList,
    SSTUB_DrawColorPattScanList,
    SSTUB_DrawEllipseList,
    SSTUB_DrawPattEllipseList,
    SSTUB_DrawColorPattEllipseList,
    SSTUB_DrawFatEllipseList,
    SSTUB_DrawPattFatEllipseList,
    SSTUB_DrawColorPattFatEllipseList,
    SSTUB_DrawRect,
    NULL,
    SSTUB_DrawPattRect,
    SSTUB_DrawColorPattRect,
    SSTUB_DrawTrap,
    SSTUB_DrawPattTrap,
    SSTUB_DrawColorPattTrap,
    SSTUB_DrawLineInt,
    SSTUB_DrawBresenhamLine,
    SSTUB_DrawStippleLineInt,
    SSTUB_DrawBresenhamStippleLine,
    SSTUB_DrawEllipse,
    SSTUB_ClipEllipse,
    SSTUB_PutMonoImageMSBSys,
    NULL,                       // SSTUB_PutMonoImageMSBLin,
    NULL,                       // SSTUB_PutMonoImageMSBBM,
    SSTUB_PutMonoImageLSBSys,
    NULL,                       // SSTUB_PutMonoImageLSBLin,
    NULL,                       // SSTUB_PutMonoImageLSBBM,
    SSTUB_ClipMonoImageMSBSys,
    NULL,                       // SSTUB_ClipMonoImageMSBLin,
    NULL,                       // SSTUB_ClipMonoImageMSBBM,
    SSTUB_ClipMonoImageLSBSys,
    NULL,                       // SSTUB_ClipMonoImageLSBLin,
    NULL,                       // SSTUB_ClipMonoImageLSBBM,
    SSTUB_BitBlt,
    NULL,                       // SSTUB_BitBltLin,
    SSTUB_BitBltSys,
    NULL,                       // SSTUB_BitBltBM,
    SSTUB_BitBltPatt,
    NULL,                       // SSTUB_BitBltPattLin,
    SSTUB_BitBltPattSys,
    NULL,                       // SSTUB_BitBltPattBM,
    SSTUB_BitBltColorPatt,
    NULL,                       // SSTUB_BitBltColorPattLin,
    SSTUB_BitBltColorPattSys,
    NULL,                       // SSTUB_BitBltColorPattBM,
    SSTUB_SrcTransBlt,
    NULL,                       // SSTUB_SrcTransBltLin,
    SSTUB_SrcTransBltSys,
    NULL,                       // SSTUB_SrcTransBltBM,
    SSTUB_DstTransBlt,
    NULL,                       // SSTUB_DstTransBltLin,
    SSTUB_DstTransBltSys,
    NULL,                       // SSTUB_DstTransBltBM,
    SSTUB_StretchBlt,
    NULL,                       // SSTUB_StretchBltLin,
    SSTUB_StretchBltSys,
    NULL,                       // SSTUB_StretchBltBM,
    SSTUB_ConvertBltSys_Obsolete,
    NULL,                       // SSTUB_ConvertBltBM_Obsolete,
    SSTUB_StretchConvertBltSys_Obsolete,
    NULL,                       // SSTUB_StretchConvertBltBM_Obsolete,
    NULL,                       // SSTUB_BitBltFxTest,
    SSTUB_BitBltFx,
    NULL,                       // SSTUB_BitBltFxLin,
    SSTUB_BitBltFxSys,
    NULL,                       // SSTUB_BitBltFxBM,
    SSTUB_GetBitmapSys,
    NULL,                       // SSTUB_GetBitmapBM,
    SSTUB_UpdateScreen,
    SSTUB_DrawClippedLineInt,
    SSTUB_DrawClippedBresenhamLine,
    SSTUB_DrawClippedStippleLineInt,
    SSTUB_DrawClippedBresenhamStippleLine,
    NULL,                       // SSTUB_BitBltPlaneMasked,
    NULL,                       // SSTUB_BitBltPlaneMaskedLin,
    NULL,                       // SSTUB_BitBltPlaneMaskedSys,
    NULL,                       // SSTUB_BitBltPlaneMaskedBM,
    NULL,                       // SSTUB_DrawRectLin,
    SSTUB_DrawRectExt,
    SSTUB_DrawStyleLineInt,
    SSTUB_DrawBresenhamStyleLine,
    SSTUB_DrawClippedStyleLineInt,
    SSTUB_DrawClippedBresenhamStyleLine,
    };

GA_2DStateFuncs     SHADOW_state2dREF;
GA_2DRenderFuncs    SHADOW_draw2dREF;
GA_2DRenderFuncs    _VARAPI SHADOW_draw2dCnt;
GA_bufferFuncs      SHADOW_buffer = {0};

#ifdef  __DRIVER__
PM_imports _VARAPI _PM_imports;
#endif

/*----------------------------- Implementation ----------------------------*/

/****************************************************************************
REMARKS:
Function to initialise the shadow buffer filter driver.
****************************************************************************/
static ibool SHADOW_init(
    GA_devCtx *hwCtx)
{
    /* Initialise the internal state of the filter driver */
    DC.hwCtx = hwCtx;
    return true;
}

#ifdef  __DRIVER__
/****************************************************************************
REMARKS:
Fatal error handler called when a non-imported function is called by the
driver. We leave this to a runtime error so that older applications and
shell drivers will work with newer bpd drivers provided no newer functions
are required by the driver itself. If they are, the application or shell
driver needs to be recompiled.
****************************************************************************/
static void _PM_fatalErrorHandler(void)
{
    PM_fatalError("Unsupported PM_imports import function called! Please re-compile!\n");
}

/****************************************************************************
REMARKS:
Function entry point for the reference rasteriser library.
****************************************************************************/
REF2D_driver * _CEXPORT SHADOW_initLibrary(
    GA_devCtx *hwCtx,
    GA_globalOptions *gOpt,
    PM_imports *pmImp)
{
    int     i,max;
    ulong   *p;

    // Initialize all default imports to point to fatal error handler
    // for upwards compatibility.
    max = sizeof(_PM_imports)/sizeof(GA_initLibrary_t);
    for (i = 0,p = (ulong*)&_PM_imports; i < max; i++)
        *p++ = (ulong)_PM_fatalErrorHandler;
    memcpy(&_PM_imports,pmImp,MIN(sizeof(_PM_imports),pmImp->dwSize));
    if (!SHADOW_init(hwCtx))
        return NULL;
    (void)gOpt;
    return &DRV;
}
#else
/****************************************************************************
REMARKS:
Function entry point to fake loading of a driver when linked into app.
****************************************************************************/
ibool NAPI SHADOW_loadDriver(
    GA_devCtx *hwCtx,
    GA_globalOptions *gOpt,
    REF2D_driver **drv,
    PE_MODULE **hModule,
    ulong *size)
{
    if (!SHADOW_init(hwCtx))
        return false;
    *drv = &DRV;
    *hModule = NULL;
    *size = 0;
    (void)gOpt;
    return true;
}
#endif

/****************************************************************************
DESCRIPTION:
Switches to the current set of rendering vectors

REMARKS:
This function is called in multiple places in the reference rasteriser
where the rendering vectors may change, and quickly switches to the new
set of rendering vectors depending on whether we require software or
hardware rendering for the current rasteriser state.
****************************************************************************/
static void SwitchVectors(void)
{
    ibool   isSolidMix = IS_SOLID_MIX(DC.mix);

    if (DC.mi.BitsPerPixel == 4) {
        // 4bpp planar vectors
        if (isSolidMix) {
            DC.DrawRect                 = SHADOW_DrawRect_4_Solid;
            DC.DrawRectExt              = SHADOW_DrawRectExt_4_Solid;
            DC.DrawLineInt              = SHADOW_DrawLineInt_4_Solid;
            DC.DrawBresenhamLine        = SHADOW_DrawBresenhamLine_4_Solid;
            }
        else {
            DC.DrawRect                 = SHADOW_DrawRect_Generic;
            DC.DrawRectExt              = SHADOW_DrawRectExt_Generic;
            DC.DrawLineInt              = SHADOW_DrawLineInt_4_Generic;
            DC.DrawBresenhamLine        = SHADOW_DrawBresenhamLine_4_Generic;
            }
        DC.DrawStippleLineInt           = SHADOW_DrawLineInt_4_Generic;
        DC.DrawBresenhamStippleLine     = SHADOW_DrawBresenhamLine_4_Generic;
        }
    else {
        // All color depths vectors
        DC.DrawRect                 = SHADOW_DrawRect_Generic;
        DC.DrawLineInt              = SHADOW_DrawLineInt_Generic;
        DC.DrawBresenhamLine        = SHADOW_DrawBresenhamLine_Generic;
        DC.DrawStippleLineInt       = SHADOW_DrawLineInt_Generic;
        DC.DrawBresenhamStippleLine = SHADOW_DrawBresenhamLine_Generic;
        switch (DC.mi.BitsPerPixel) {
            case 8:
                if (isSolidMix)
                    DC.DrawRect     = SHADOW_DrawRect_8_Solid;
                DC.drawLineXPosLo   = SHADOW_drawLineXPosLo_8;
                DC.drawLineXPosHi   = SHADOW_drawLineXPosHi_8;
                DC.drawLineXNegLo   = SHADOW_drawLineXNegLo_8;
                DC.drawLineXNegHi   = SHADOW_drawLineXNegHi_8;
                break;
            case 15:
            case 16:
                if (isSolidMix)
                    DC.DrawRect     = SHADOW_DrawRect_16_Solid;
                DC.drawLineXPosLo   = SHADOW_drawLineXPosLo_16;
                DC.drawLineXPosHi   = SHADOW_drawLineXPosHi_16;
                DC.drawLineXNegLo   = SHADOW_drawLineXNegLo_16;
                DC.drawLineXNegHi   = SHADOW_drawLineXNegHi_16;
                break;
            case 24:
                DC.drawLineXPosLo   = SHADOW_drawLineXPosLo_24;
                DC.drawLineXPosHi   = SHADOW_drawLineXPosHi_24;
                DC.drawLineXNegLo   = SHADOW_drawLineXNegLo_24;
                DC.drawLineXNegHi   = SHADOW_drawLineXNegHi_24;
                break;
            case 32:
                if (isSolidMix)
                    DC.DrawRect     = SHADOW_DrawRect_32_Solid;
                DC.drawLineXPosLo   = SHADOW_drawLineXPosLo_32;
                DC.drawLineXPosHi   = SHADOW_drawLineXPosHi_32;
                DC.drawLineXNegLo   = SHADOW_drawLineXNegLo_32;
                DC.drawLineXNegHi   = SHADOW_drawLineXNegHi_32;
                break;
            }
        }
}

/****************************************************************************
REMARKS:
Function to initialise the portrait display filter driver.
****************************************************************************/
ibool NAPI SHADOW_InitFilter(
    GA_modeInfo *modeInfo,
    N_int32 transferStart,
    struct _REF2D_driver *ref2d,
    struct _REF2D_driver **drv)
{
    GA_initFuncs        init;
    GA_driverFuncs      driver;
    GA_2DStateFuncs     state2d;

    // Initialise the internal state of the filter driver
    init.dwSize = sizeof(init);
    if (!DC.hwCtx->loader.QueryFunctions(GA_GET_INITFUNCS,0,&init))
        return false;
    driver.dwSize = sizeof(driver);
    if (!DC.hwCtx->loader.QueryFunctions(GA_GET_DRIVERFUNCS,0,&driver))
        return false;
    state2d.dwSize = sizeof(state2d);
    DC.hwCtx->loader.QueryFunctions(GA_GET_2DSTATEFUNCS,0,&state2d);
    DC.cntMode = init.GetVideoMode();
    DC.mi = *modeInfo;
    DC.ref2d = ref2d;
    DC.refState = ref2d->DriverState;
    DC.SetBank = driver.SetBank;
    DC.cntBank = -1;
    DC.mix = GA_REPLACE_MIX;
    DC.planeMask = 0xFFFFFFFF;
    DC.srcBlendFunc = gaBlendNone;
    DC.dstBlendFunc = gaBlendNone;
    SHADOW_state2dREF.dwSize = sizeof(SHADOW_state2dREF);
    if (!ref2d->QueryFunctions(GA_GET_2DSTATEFUNCS,&SHADOW_state2dREF))
        return false;
    SHADOW_draw2dREF.dwSize = sizeof(SHADOW_draw2dREF);
    if (!ref2d->QueryFunctions(GA_GET_2DRENDERFUNCS,&SHADOW_draw2dREF))
        return false;

    // Export the existing software rasteriser draw rect ext function
    SHADOW_drv.GetDrawBuffer = ref2d->GetDrawBuffer;
    SHADOW_drv.DrawRectExtSW = ref2d->DrawRectExtSW;

    // Initialise direct framebuffer access functions
    if (DC.cntMode & gaLinearBuffer)
        PM_fatalError("Shadow buffer only supports banked modes!");
    DC.UpdateScreen                     = SHADOW_UpdateScreenBnk;
    DC.refState->updateCursor           = SHADOW_UpdateScreenBnk;

    // Intialise internal rendering vectors
    switch (DC.mi.BitsPerPixel) {
        case 4:
            DC.byteShift = 0;
            DC.bufferSize = DC.mi.YResolution * (DC.mi.XResolution / 2);
            DC.UpdateScreen                     = SHADOW_UpdateScreen4;
            DC.refState->updateCursor           = SHADOW_UpdateCursor4;
            break;
        case 8:
            DC.byteShift = 0;
            DC.bufferSize = DC.mi.YResolution * DC.mi.BytesPerScanLine;
            break;
        case 15:
        case 16:
            DC.byteShift = 1;
            DC.bufferSize = DC.mi.YResolution * DC.mi.BytesPerScanLine;
            break;
        case 24:
            DC.byteShift = 3;
            DC.bufferSize = DC.mi.YResolution * DC.mi.BytesPerScanLine;
            break;
        case 32:
            DC.byteShift = 2;
            DC.bufferSize = DC.mi.YResolution * DC.mi.BytesPerScanLine;
            break;
        }
    SwitchVectors();

    // Allocate memory for the shadow buffer, and clear it
    if ((DC.buffer = malloc(DC.bufferSize)) == NULL)
        return false;
    memset(DC.buffer,0,DC.bufferSize);

    // Save the original hardware linear framebuffer pointer and
    // replace it with a pointer to our shadow buffer
    DC.LinearMem = DC.hwCtx->LinearMem;
    DC.hwCtx->LinearMem = DC.buffer;

    // Change the cursor drawing code to always draw to the shadow buffer by default
    DC.refState->cursorDrawAddr = DC.buffer;

    // Initialise all the filter driver vectors
    INIT_FILTER_VECTORS;

    // Return address of our filter driver as top of filter chain
    *drv = &DRV;
    (void)transferStart;
    return true;
}

/****************************************************************************
REMARKS:
Function to clean up when unloading the driver.
****************************************************************************/
void NAPI SHADOW_UnloadDriver(void)
{
}

/****************************************************************************
REMARKS:
Function to clean up when the filter driver needs to de-initialise (ie:
going back to text mode when the ref2d driver is unloaded).
****************************************************************************/
void NAPI SHADOW_DeInitFilter(void)
{
    DC.hwCtx->LinearMem = DC.LinearMem;
    free(DC.buffer);
    if (SHADOW_buffer.dwSize) {
        DC.refState->UnloadBufferMgr(DC.hModBuffer);
        memset(&SHADOW_buffer,0,sizeof(SHADOW_buffer));
        }
}

/****************************************************************************
REMARKS:
This returns the address of the currently active draw buffer
****************************************************************************/
GA_buffer * NAPI SHADOW_GetDrawBuffer(void)
{
    return DC.ref2d->GetDrawBuffer();
}

/****************************************************************************
PARAMETERS:
left        - Left coordinate of the rectangle to draw
top         - Top coordinate of the rectangle to draw
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
color       - Color to fill the rectangle with
mix         - Mix to use for drawing the rectangle
****************************************************************************/
void NAPI SHADOW_DrawRectExtSW(
    void *buffer,
    N_int32 dstPitch,
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    GA_color color,
    N_int32 mix)
{
    DC.ref2d->DrawRectExtSW(buffer, dstPitch, left, top, width, height, color, mix);
}

/****************************************************************************
DESCRIPTION:
Forces software only rendering on or off.
****************************************************************************/
N_int32 NAPI SHADOW_ForceSoftwareOnly(
    N_int32 enable)
{
    return DC.ref2d->ForceSoftwareOnly(enable);
}

/****************************************************************************
PARAMETERS:
mask    - New color compare mask

REMARKS:
This function is used to override the internal software rendering color
compare mask used for source and destination transparent blit functions.
****************************************************************************/
N_uint32 NAPI SHADOW_SetColorCompareMask(
    N_uint32 mask)
{
    return DC.ref2d->SetColorCompareMask(mask);
}

/****************************************************************************
DESCRIPTION:
Fix up the software rasteriser after a mode switch has occured
****************************************************************************/
void NAPI SHADOW_PostSwitchPhysicalResolution(void)
{
    DC.ref2d->PostSwitchPhysicalResolution();
}

/****************************************************************************
PARAMETERS:
funcs       - Pointer to the function list to make valid
size        - Size of the function list in bytes

RETURNS:
True if valid, false if not.
****************************************************************************/
static ibool MakeValid(
    ulong *funcs,
    int size)
{
    int i,empty = true;

    for (i = 0; i < size/4; i++) {
        if (funcs[i] != 0)
            empty = false;
        }
    return !empty;
}

/****************************************************************************
DESCRIPTION:
Returns the reference rasterisers for the specified function group.

PARAMETERS:
id      - Identifier for the function group to get pointers for
funcs   - Place to store the function block

RETURNS:
True if functions available, false if not.

REMARKS:
This function is the main function that is used by the graphics application
code to get a block of function pointers for a specified function group.
The function groups are defined GA_funcGroupsType enumeration, and breaks
up the functions in the device driver API into groups of logically similar
functions.
****************************************************************************/
ibool NAPI SHADOW_QueryFunctions(
    N_uint32 id,
    void *funcs)
{
    ulong   dwSize = *((ulong*)funcs),size;

    memset(funcs,0,dwSize);
    switch (id) {
        case GA_GET_2DSTATEFUNCS:   RETURN_VALID(SHADOW_state2dPT);
        case GA_GET_2DRENDERFUNCS:  RETURN_VALID(SHADOW_draw2dStub);
        case GA_GET_BUFFERFUNCS: {
            // We override loading the buffer manager functions here
            if (SHADOW_buffer.dwSize == 0) {
                SHADOW_buffer.dwSize = sizeof(SHADOW_buffer);
                if (!DC.refState->LoadBufferMgr(DC.hwCtx,&SHADOW_drv,&SHADOW_buffer,&DC.hModBuffer,&size))
                    return false;
                }
            RETURN_VALID(SHADOW_buffer);
            }
        }
    *((ulong*)funcs) = dwSize;
    return DC.ref2d->QueryFunctions(id,funcs);
}

/****************************************************************************
DESCRIPTION:
Sets a display buffer as the active drawing buffer.

PARAMETERS:
drawBuf             - Buffer to make the active drawing buffer
framebuffer         - Pointer to the start of framebuffer memory
bitsPerPixel        - Color depth for the buffer
beginDirectAccess   - Direct framebuffer arbitration function (NULL if none)
endDirectAccess     - Direct framebuffer arbitration function (NULL if none)

RETURNS:
0 on success, -1 on failure.

REMARKS:
This function allows the application to tell make a display memory
buffer the active rendering buffer for all subsequent rendering commands.
****************************************************************************/
N_int32 NAPI SHADOW_SetDrawBufferSW(
    GA_buffer *drawBuf,
    void *framebuffer,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pf,
    GA_devCtx *hwCtx,
    N_int32 forceSoftwareOnly)
{
    (void)forceSoftwareOnly;
    DC.buf = *drawBuf;
    if (DC.hwCtx && framebuffer == DC.hwCtx->LinearMem) {
        DC.enabled = true;
        SHADOW_draw2dCnt = SHADOW_draw2dPT;
        drawBuf->Offset = 0;
        if (DC.mi.BitsPerPixel == 4)
            drawBuf->Stride = DC.mi.XResolution/2;
        return DC.ref2d->SetDrawBuffer(drawBuf,DC.buffer,bitsPerPixel,pf,hwCtx,true);
        }
    else {
        DC.enabled = false;
        SHADOW_draw2dCnt = SHADOW_draw2dREF;
        return DC.ref2d->SetDrawBuffer(drawBuf,framebuffer,bitsPerPixel,pf,hwCtx,true);
        }
}

/****************************************************************************
PARAMETERS:
surface         - Pointer to the start of the memory buffer for drawing
xRes            - X resolution for the surface
yRes            - Y resolution for the surface
bytesPerLine    - Scanline width for the memory buffer for drawing
bitsPerPixel    - New color depth for the surface
pf              - New pixel format for the surface

REMARKS:
This function changes the software draw buffer quickly, to allow the
software renderer to render to bitmaps in different locations in memory.
We do in here for speed, since we want to avoid the overhead of setting
the full draw buffer for the reference rasteriser code. If the color
depth or pixel format changes, this function will switch all the internal
rendering vectors to put to the new vectors for the appropriate functions.
This allows the reference rasteriser to work with all color depths active
in a single driver (and for our debug library to work properly).
****************************************************************************/
void NAPI SHADOW_SetDrawSurface(
    void *surface,
    N_int32 xRes,
    N_int32 yRes,
    N_int32 bytesPerLine,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pf)
{
    // Determine if the shadow buffer should be enabled
    if (DC.hwCtx && surface == DC.hwCtx->LinearMem) {
        DC.enabled = true;
        SHADOW_draw2dCnt = SHADOW_draw2dPT;
        if (DC.mi.BitsPerPixel == 4)
            bytesPerLine = DC.mi.XResolution/2;
        DC.ref2d->SetDrawSurface(DC.buffer,xRes,yRes,bytesPerLine,bitsPerPixel,pf);
        }
    else {
        DC.enabled = false;
        SHADOW_draw2dCnt = SHADOW_draw2dREF;
        DC.ref2d->SetDrawSurface(surface,xRes,yRes,bytesPerLine,bitsPerPixel,pf);
        }
}

/****************************************************************************
PARAMETERS:
drawBuf - New hardware draw buffer to make active

RETURNS:
0 on success, -1 on failure.

REMARKS:
This function changes the hardware draw buffer for both the software
rasteriser and the hardware device. This can also be used to change the
draw buffer for a software only buffer, but is generally only used when
using hardware.
****************************************************************************/
N_int32 NAPI SHADOW_SetDrawBuffer(
    GA_buffer *drawBuf)
{
    DC.buf = *drawBuf;
    if (DC.enabled) {
        drawBuf->Offset = 0;
        if (DC.mi.BitsPerPixel == 4)
            drawBuf->Stride = DC.mi.XResolution/2;
        }
    return STATE2D.SetDrawBuffer(drawBuf);
}

/****************************************************************************
PARAMETERS:
color   - New foreground color to set
****************************************************************************/
void NAPI SHADOW_SetForeColor(
    GA_color color)
{
    switch (DC.mi.BitsPerPixel) {
        case 4:
            DC.foreColor = color;
            break;
        case 8:
            DC.foreColor = mkcolor8(color);
            break;
        case 15:
        case 16:
            DC.foreColor = mkcolor16(color);
            break;
        case 24:
        case 32:
            DC.foreColor = color;
            break;
        }
    STATE2D.SetForeColor(color);
}

/****************************************************************************
PARAMETERS:
mix     - New mix code to set (GA_mixCodesType)
****************************************************************************/
N_int32 NAPI SHADOW_SetMix(
    N_int32 mix)
{
    DC.mix = mix;
    SwitchVectors();
    return STATE2D.SetMix(mix);
}

/****************************************************************************
PARAMETERS:
mask    - New 32-bit plane mask to set
****************************************************************************/
void NAPI SHADOW_SetPlaneMask(
    N_uint32 mask)
{
    DC.planeMask = mask;
    SwitchVectors();
    STATE2D.SetPlaneMask(mask);
}

/****************************************************************************
PARAMETERS:
blendFunc   - New source blending function to make active

REMARKS:
Obsolete, use SetBlendFunc instead.
****************************************************************************/
void NAPI SHADOW_SetSrcBlendFunc(
    N_int32 blendFunc)
{
    SHADOW_SetBlendFunc(blendFunc,DC.dstBlendFunc);
}

/****************************************************************************
PARAMETERS:
blendFunc   - New destination blending function to make active

REMARKS:
Obsolete, use SetBlendFunc instead.
****************************************************************************/
void NAPI SHADOW_SetDstBlendFunc(
    N_int32 blendFunc)
{
    SHADOW_SetBlendFunc(DC.srcBlendFunc,blendFunc);
}

/****************************************************************************
PARAMETERS:
blendFunc   - New source blending function to make active
****************************************************************************/
void NAPI SHADOW_SetBlendFunc(
    N_int32 srcBlendFunc,
    N_int32 dstBlendFunc)
{
    DC.srcBlendFunc = srcBlendFunc;
    DC.dstBlendFunc = dstBlendFunc;
    SwitchVectors();
    STATE2D.SetBlendFunc(srcBlendFunc,dstBlendFunc);
}

/****************************************************************************
REMARKS:
Updates the screen by copying a portion of the shadow buffer
to a banked framebuffer mode.
****************************************************************************/
void SHADOW_UpdateScreenBnk(
    int offset,
    int left,
    int top,
    int width,
    int height)
{
    int     bytes,startAddr,endAddr,newAddr,bank,bytesStart,bytesEnd;
    uchar   *s,*d,*bankedMem;

    bytes = LIN_X_ADDR(width,DC.byteShift);
    startAddr = (DC.mi.BytesPerScanLine * top) + LIN_X_ADDR(left,DC.byteShift);
    s = (uchar*)DC.buffer + startAddr;
    bank = (startAddr += offset) >> 16;
    bankedMem = DC.hwCtx->BankedMem;
    d = bankedMem + (startAddr & 0xFFFF);

    if (DC.cntBank != bank)
        DC.SetBank(DC.cntBank = bank);
    while (height--) {
        newAddr = startAddr + DC.mi.BytesPerScanLine;
        if ((bank = newAddr >> 16) != DC.cntBank) {
            // Bank boundary is present before next scanline
            if ((endAddr = startAddr + bytes) >> 16 != DC.cntBank) {
                // Data to copy is split by bank boundary
                bytesStart = (endAddr & ~0xFFFF) - startAddr;
                bytesEnd = endAddr & 0xFFFF;
                memcpy(d,s,bytesStart);
                DC.SetBank(DC.cntBank = bank);
                memcpy(bankedMem,s + bytesStart,bytesEnd);
                }
            else {
                // Data to copy is contained within the current bank
                memcpy(d,s,bytes);
                DC.SetBank(DC.cntBank = bank);
                }
            d = bankedMem + (newAddr & 0xFFFF);
            }
        else {
            // Scanline is contained within the current bank
            memcpy(d,s,bytes);
            d += DC.mi.BytesPerScanLine;
            }
        startAddr += DC.mi.BytesPerScanLine;
        s += DC.mi.BytesPerScanLine;
        }
}

/****************************************************************************
REMARKS:
Enables direct access to the shadow buffer.
****************************************************************************/
void NAPI SHADOW_EnableDirectAccess(void)
{
}

/****************************************************************************
REMARKS:
Disables direct access to the shadow buffer. This indicates custom rendering
has been completed, so we need to blit the entire buffer to the screen.
****************************************************************************/
void NAPI SHADOW_DisableDirectAccess(void)
{
    DC.UpdateScreen(DC.buf.Offset,0,0,DC.mi.XResolution,DC.mi.YResolution);
}

/****************************************************************************
PARAMETERS:
x   - X coordinate to plot pixel at
y   - Y coordinate to plot pixel at

REMARKS:
Writes a pixel to the framebuffer.

NOTE:   This function *assumes* the caller has already enabled direct
        framebuffer access for performance!!
****************************************************************************/
void NAPI SHADOW_PutPixel(
    N_int32 x,
    N_int32 y)
{
    DRAW2D.PutPixel(x,y);
    DC.UpdateScreen(DC.buf.Offset,x,y,1,1);
}

/****************************************************************************
PARAMETERS:
ileft   - Left coordinate of ellipse rectangle
itop    - Top coordinate of ellipse rectangle
A       - Major axis dimension
B       - Minor axis dimension
****************************************************************************/
void NAPI SHADOW_DrawEllipse(
    N_int32 left,
    N_int32 top,
    N_int32 A,
    N_int32 B)
{
    DRAW2D.DrawEllipse(left,top,A,B);
    DC.UpdateScreen(DC.buf.Offset,left,top,A,B);
}

/****************************************************************************
PARAMETERS:
left        - Left coordinate of ellipse rectangle
top         - Top coordinate of ellipse rectangle
A           - Major axis dimension
B           - Minor axis dimension
clipLeft    - Left coordinate for clip rectangle
clipTop     - Top coordinate for clip rectangle
clipRight   - Right coordinate for clip rectangle
clipBottom  - Bottom coordinate for clip rectangle
****************************************************************************/
void NAPI SHADOW_ClipEllipse(
    N_int32 left,
    N_int32 top,
    N_int32 A,
    N_int32 B,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
    DRAW2D.ClipEllipse(left,top,A,B,clipLeft,clipTop,clipRight,clipBottom);
    DC.UpdateScreen(DC.buf.Offset,left,top,A,B);
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
x1      - First X coordinate of scanline
x2      - Second X coordinate of scanline
****************************************************************************/
static void DrawScan(
    N_int32 y,
    N_int32 x1,
    N_int32 x2)
{
    int width;

    if ((width = (x2 - x1)) == 0)
        return;
    if (width < 0) {
        x1 = x2;
        width = -width;
        }
    DRAW2D.DrawRect(x1,y,width,1);
    DC.DrawRect(x1,y,width,1);
}

/****************************************************************************
PARAMETERS:
y           - Y coordinate for scanline
x1          - First X coordinate of scanline
x2          - Second X coordinate of scanline
****************************************************************************/
static void DrawPattScan(
    N_int32 y,
    N_int32 x1,
    N_int32 x2)
{
    int width;

    if ((width = (x2 - x1)) == 0)
        return;
    if (width < 0) {
        x1 = x2;
        width = -width;
        }
    DRAW2D.DrawPattRect(x1,y,width,1);
    DC.UpdateScreen(DC.buf.Offset,x1,y,width,1);
}

/****************************************************************************
PARAMETERS:
y           - Y coordinate for scanline
x1          - First X coordinate of scanline
x2          - Second X coordinate of scanline
****************************************************************************/
static void DrawColorPattScan(
    N_int32 y,
    N_int32 x1,
    N_int32 x2)
{
    int width;

    if ((width = (x2 - x1)) == 0)
        return;
    if (width < 0) {
        x1 = x2;
        width = -width;
        }
    DRAW2D.DrawColorPattRect(x1,y,width,1);
    DC.UpdateScreen(DC.buf.Offset,x1,y,width,1);
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI SHADOW_DrawScanList(
    N_int32 y,
    N_int32 length,
    N_int16 *scans)
{
    while (length--) {
        int width;
        int x1 = *scans++;
        int x2 = *scans++;
        if ((width = (x2 - x1)) != 0) {
            if (width < 0) {
                x1 = x2;
                width = -width;
                }
            DRAW2D.DrawRect(x1,y,width,1);
            DC.DrawRect(x1,y,width,1);
            }
        y++;
        }
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI SHADOW_DrawPattScanList(
    N_int32 y,
    N_int32 length,
    N_int16 *scans)
{
    while (length--) {
        int width;
        int x1 = *scans++;
        int x2 = *scans++;
        if ((width = (x2 - x1)) != 0) {
            if (width < 0) {
                x1 = x2;
                width = -width;
                }
            DRAW2D.DrawPattRect(x1,y,width,1);
            DC.UpdateScreen(DC.buf.Offset,x1,y,width,1);
            }
        y++;
        }
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI SHADOW_DrawColorPattScanList(
    N_int32 y,
    N_int32 length,
    N_int16 *scans)
{
    while (length--) {
        int width;
        int x1 = *scans++;
        int x2 = *scans++;
        if ((width = (x2 - x1)) != 0) {
            if (width < 0) {
                x1 = x2;
                width = -width;
                }
            DRAW2D.DrawColorPattRect(x1,y,width,1);
            DC.UpdateScreen(DC.buf.Offset,x1,y,width,1);
            }
        y++;
        }
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI SHADOW_DrawEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int i,j,maxIndex;

    // Render the scanlines in the ellipse
    maxIndex = y+length-1;
    for (i = y,j = y+height; i < maxIndex; i++,j--,scans += 2) {
        DrawScan(i,scans[0],scans[1]);
        DrawScan(j,scans[0],scans[1]);
        }
    if (!(height & 1))
        DrawScan(i,scans[0],scans[1]);
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI SHADOW_DrawPattEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int i,j,maxIndex;

    // Render the scanlines in the ellipse
    maxIndex = y+length-1;
    for (i = y,j = y+height; i < maxIndex; i++,j--,scans += 2) {
        DrawPattScan(i,scans[0],scans[1]);
        DrawPattScan(j,scans[0],scans[1]);
        }
    if (!(height & 1))
        DrawPattScan(i,scans[0],scans[1]);
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI SHADOW_DrawColorPattEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int i,j,maxIndex;

    // Render the scanlines in the ellipse
    maxIndex = y+length-1;
    for (i = y,j = y+height; i < maxIndex; i++,j--,scans += 2) {
        DrawColorPattScan(i,scans[0],scans[1]);
        DrawColorPattScan(j,scans[0],scans[1]);
        }
    if (!(height & 1))
        DrawColorPattScan(i,scans[0],scans[1]);
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis + pen height adjustment
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI SHADOW_DrawFatEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int i,j,maxIndex;

    // Render the scanlines in the ellipse
    maxIndex = y+length;
    for (i = y,j = y+height; i < maxIndex; i++,j--,scans += 4) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            DrawScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            DrawScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            DrawScan(j,scans[I_LEFTL],scans[I_LEFTR]);
            DrawScan(j,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            DrawScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            DrawScan(j,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
    if ((height+1) & 1) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            DrawScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            DrawScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            DrawScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis + pen height adjustment
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI SHADOW_DrawPattFatEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int i,j,maxIndex;

    // Render the scanlines in the ellipse
    maxIndex = y+length;
    for (i = y,j = y+height; i < maxIndex; i++,j--,scans += 4) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            DrawPattScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            DrawPattScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            DrawPattScan(j,scans[I_LEFTL],scans[I_LEFTR]);
            DrawPattScan(j,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            DrawPattScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            DrawPattScan(j,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
    if ((height+1) & 1) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            DrawPattScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            DrawPattScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            DrawPattScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
}

/****************************************************************************
PARAMETERS:
y       - Y coordinate for scanline
length  - Number of scanlines in the list
height  - Height of the ellipse minor axis + pen height adjustment
scans   - Pointer to an array of scanline data
****************************************************************************/
void NAPI SHADOW_DrawColorPattFatEllipseList(
    N_int32 y,
    N_int32 length,
    N_int32 height,
    N_int16 *scans)
{
    int i,j,maxIndex;

    // Render the scanlines in the ellipse
    maxIndex = y+length;
    for (i = y,j = y+height; i < maxIndex; i++,j--,scans += 4) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            DrawColorPattScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            DrawColorPattScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            DrawColorPattScan(j,scans[I_LEFTL],scans[I_LEFTR]);
            DrawColorPattScan(j,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            DrawColorPattScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            DrawColorPattScan(j,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
    if ((height+1) & 1) {
        if (scans[I_LEFTR] < scans[I_RIGHTL]) {
            DrawColorPattScan(i,scans[I_LEFTL],scans[I_LEFTR]);
            DrawColorPattScan(i,scans[I_RIGHTL],scans[I_RIGHTR]);
            }
        else {
            DrawColorPattScan(i,scans[I_LEFTL],scans[I_RIGHTR]);
            }
        }
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the rectangle to draw
top     - Top coordinate of the rectangle to draw
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
****************************************************************************/
void NAPI SHADOW_DrawRect(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height)
{
    DRAW2D.DrawRect(left,top,width,height);
    DC.DrawRect(left,top,width,height);
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the rectangle to draw
top     - Top coordinate of the rectangle to draw
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
****************************************************************************/
void NAPI SHADOW_DrawRectExt(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    GA_color color,
    N_int32 mix)
{
    DRAW2D.DrawRectExt(left,top,width,height,color,mix);
    DC.DrawRectExt(left,top,width,height,color,mix);
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the rectangle to draw
top     - Top coordinate of the rectangle to draw
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
****************************************************************************/
void NAPI SHADOW_DrawPattRect(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height)
{
    DRAW2D.DrawPattRect(left,top,width,height);
    DC.UpdateScreen(DC.buf.Offset,left,top,width,height);
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the rectangle to draw
top     - Top coordinate of the rectangle to draw
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
****************************************************************************/
void NAPI SHADOW_DrawColorPattRect(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height)
{
    DRAW2D.DrawColorPattRect(left,top,width,height);
    DC.UpdateScreen(DC.buf.Offset,left,top,width,height);
}

/****************************************************************************
PARAMETERS:
trap    - Pointer to the GA_trap structure describing the trapezoid
****************************************************************************/
void NAPI SHADOW_DrawTrap(
    GA_trap *trap)
{
    int y,x1,x2,count = trap->count;

    if (!count)
        return;
    y = trap->y;
    x1 = trap->x1;
    x2 = trap->x2;
    trap->y += count;
    while (count--) {
        int width;
        int ix1 = (x1 + 0x8000) >> 16;
        int ix2 = (x2 + 0x8000) >> 16;
        if ((width = (ix2 - ix1)) != 0) {
            if (width < 0) {
                ix1 = ix2;
                width = -width;
                }
            DRAW2D.DrawRect(ix1,y,width,1);
            DC.DrawRect(ix1,y,width,1);
            }
        x1 += trap->slope1;
        x2 += trap->slope2;
        y++;
        }
    trap->x1 = x1;
    trap->x2 = x2;
}

/****************************************************************************
PARAMETERS:
trap    - Pointer to the GA_trap structure describing the trapezoid
****************************************************************************/
void NAPI SHADOW_DrawPattTrap(
    GA_trap *trap)
{
    int y,x1,x2,count = trap->count;

    if (!count)
        return;
    y = trap->y;
    x1 = trap->x1;
    x2 = trap->x2;
    trap->y += count;
    while (count--) {
        int width;
        int ix1 = (x1 + 0x8000) >> 16;
        int ix2 = (x2 + 0x8000) >> 16;
        if ((width = (ix2 - ix1)) != 0) {
            if (width < 0) {
                ix1 = ix2;
                width = -width;
                }
            DRAW2D.DrawPattRect(ix1,y,width,1);
            DC.UpdateScreen(DC.buf.Offset,ix1,y,width,1);
            }
        x1 += trap->slope1;
        x2 += trap->slope2;
        y++;
        }
    trap->x1 = x1;
    trap->x2 = x2;
}

/****************************************************************************
PARAMETERS:
trap    - Pointer to the GA_trap structure describing the trapezoid
****************************************************************************/
void NAPI SHADOW_DrawColorPattTrap(
    GA_trap *trap)
{
    int y,x1,x2,count = trap->count;

    if (!count)
        return;
    y = trap->y;
    x1 = trap->x1;
    x2 = trap->x2;
    trap->y += count;
    while (count--) {
        int width;
        int ix1 = (x1 + 0x8000) >> 16;
        int ix2 = (x2 + 0x8000) >> 16;
        if ((width = (ix2 - ix1)) != 0) {
            if (width < 0) {
                ix1 = ix2;
                width = -width;
                }
            DRAW2D.DrawColorPattRect(ix1,y,width,1);
            DC.UpdateScreen(DC.buf.Offset,ix1,y,width,1);
            }
        x1 += trap->slope1;
        x2 += trap->slope2;
        y++;
        }
    trap->x1 = x1;
    trap->x2 = x2;
}

/****************************************************************************
PARAMETERS:
x1          - X1 coordinate
y1          - Y1 coordinate
x2          - X2 coordinate
y2          - Y2 coordinate
drawLast    - 1 to draw last pixel, 0 to skip it
****************************************************************************/
void NAPI SHADOW_DrawLineInt(
    N_int32 x1,
    N_int32 y1,
    N_int32 x2,
    N_int32 y2,
    N_int32 drawLast)
{
    DRAW2D.DrawLineInt(x1,y1,x2,y2,drawLast);
    DC.DrawLineInt(x1,y1,x2,y2,drawLast);
}

/****************************************************************************
PARAMETERS:
x1              - X1 coordinate
y1              - Y1 coordinate
initialError    - Initial error term
majorInc        - Major increment factor for error term
diagInc         - Diagonal increment factor for error term
count           - Number of pixels to draw
flags           - Flags to control drawing
****************************************************************************/
void NAPI SHADOW_DrawBresenhamLine(
    N_int32 x1,
    N_int32 y1,
    N_int32 initialError,
    N_int32 majorInc,
    N_int32 diagInc,
    N_int32 count,
    N_int32 flags)
{
    DRAW2D.DrawBresenhamLine(x1,y1,initialError,majorInc,diagInc,count,flags);
    DC.DrawBresenhamLine(x1,y1,initialError,majorInc,diagInc,count,flags);
}

/****************************************************************************
PARAMETERS:
x1          - X1 coordinate
y1          - Y1 coordinate
x2          - X2 coordinate
y2          - Y2 coordinate
drawLast    - 1 to draw last pixel, 0 to skip it
transparent - 1 if the line is transparent, 0 if opaque
****************************************************************************/
void NAPI SHADOW_DrawStippleLineInt(
    N_int32 x1,
    N_int32 y1,
    N_int32 x2,
    N_int32 y2,
    N_int32 drawLast,
    N_int32 transparent)
{
    DRAW2D.DrawStippleLineInt(x1,y1,x2,y2,drawLast,transparent);
    DC.DrawStippleLineInt(x1,y1,x2,y2,drawLast);
}

/****************************************************************************
PARAMETERS:
x1              - X1 coordinate
y1              - Y1 coordinate
initialError    - Initial error term
majorInc        - Major increment factor for error term
diagInc         - Diagonal increment factor for error term
count           - Number of pixels to draw
flags           - Flags to control drawing
transparent     - 1 if the line is transparent, 0 if opaque
****************************************************************************/
void NAPI SHADOW_DrawBresenhamStippleLine(
    N_int32 x1,
    N_int32 y1,
    N_int32 initialError,
    N_int32 majorInc,
    N_int32 diagInc,
    N_int32 count,
    N_int32 flags,
    N_int32 transparent)
{
    DRAW2D.DrawBresenhamStippleLine(x1,y1,initialError,majorInc,diagInc,count,flags,transparent);
    DC.DrawBresenhamStippleLine(x1,y1,initialError,majorInc,diagInc,count,flags);
}

/****************************************************************************
PARAMETERS:
x1          - X1 coordinate
y1          - Y1 coordinate
x2          - X2 coordinate
y2          - Y2 coordinate
drawLast    - 1 to draw last pixel, 0 to skip it
transparent - 1 if the line is transparent, 0 if opaque
****************************************************************************/
void NAPI SHADOW_DrawStyleLineInt(
    N_int32 x1,
    N_int32 y1,
    N_int32 x2,
    N_int32 y2,
    N_int32 drawLast,
    N_int32 transparent)
{
    DRAW2D.DrawStyleLineInt(x1,y1,x2,y2,drawLast,transparent);
    DC.DrawStippleLineInt(x1,y1,x2,y2,drawLast);
}

/****************************************************************************
PARAMETERS:
x1              - X1 coordinate
y1              - Y1 coordinate
initialError    - Initial error term
majorInc        - Major increment factor for error term
diagInc         - Diagonal increment factor for error term
count           - Number of pixels to draw
flags           - Flags to control drawing
transparent     - 1 if the line is transparent, 0 if opaque
****************************************************************************/
void NAPI SHADOW_DrawBresenhamStyleLine(
    N_int32 x1,
    N_int32 y1,
    N_int32 initialError,
    N_int32 majorInc,
    N_int32 diagInc,
    N_int32 count,
    N_int32 flags,
    N_int32 transparent)
{
    DRAW2D.DrawBresenhamStyleLine(x1,y1,initialError,majorInc,diagInc,count,flags,transparent);
    DC.DrawBresenhamStippleLine(x1,y1,initialError,majorInc,diagInc,count,flags);
}

/****************************************************************************
PARAMETERS:
x1          - X coordinate of first endpoint to clip
y1          - Y coordinate of first endpoint to clip
x2          - X coordinate of second endpoint to clip
y2          - Y coordinate of second endpoint to clip
drawLast    - True if last pixel should be drawn
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle

REMARKS:
Internal function to clip and draw an integer line, such that the clipped
line segment draws the *exact* same set of pixels that the unclipped line
would have drawn. We also correctly handle arbitrarily large lines, while
still allowing the clipped line segment to be drawn in hardware.
****************************************************************************/
N_int32 NAPI SHADOW_DrawClippedLineInt(
    N_int32 x1,
    N_int32 y1,
    N_int32 x2,
    N_int32 y2,
    N_int32 drawLast,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
#define DRAW_LINE_INT           SHADOW_DrawLineInt
#define DRAW_BRESENHAM_LINE     SHADOW_DrawBresenhamLine
#include "raster/lineclip.c"
}

/****************************************************************************
PARAMETERS:
x1          - X coordinate of first endpoint to clip
y1          - Y coordinate of first endpoint to clip
x2          - X coordinate of second endpoint to clip
y2          - Y coordinate of second endpoint to clip
drawLast    - True if last pixel should be drawn
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle

REMARKS:
Internal function to clip and draw an integer line, such that the clipped
line segment draws the *exact* same set of pixels that the unclipped line
would have drawn. We also correctly handle arbitrarily large lines, while
still allowing the clipped line segment to be drawn in hardware.
****************************************************************************/
N_int32 NAPI SHADOW_DrawClippedBresenhamLine(
    N_int32 x1,
    N_int32 y1,
    N_int32 initialError,
    N_int32 majorInc,
    N_int32 diagInc,
    N_int32 count,
    N_int32 flags,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
#define DRAW_BRESENHAM_LINE     SHADOW_DrawBresenhamLine
#include "raster/blineclp.c"
}

/****************************************************************************
PARAMETERS:
x1          - X coordinate of first endpoint to clip
y1          - Y coordinate of first endpoint to clip
x2          - X coordinate of second endpoint to clip
y2          - Y coordinate of second endpoint to clip
drawLast    - True if last pixel should be drawn
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle

REMARKS:
Internal function to clip and draw an integer line, such that the clipped
line segment draws the *exact* same set of pixels that the unclipped line
would have drawn. We also correctly handle arbitrarily large lines, while
still allowing the clipped line segment to be drawn in hardware.
****************************************************************************/
N_int32 NAPI SHADOW_DrawClippedStippleLineInt(
    N_int32 x1,
    N_int32 y1,
    N_int32 x2,
    N_int32 y2,
    N_int32 drawLast,
    N_int32 transparent,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
#define DRAW_LINE_INT(a,b,c,d,e)            SHADOW_DrawStippleLineInt(a,b,c,d,e,transparent)
#define DRAW_BRESENHAM_LINE(a,b,c,d,e,f,g)  SHADOW_DrawBresenhamStippleLine(a,b,c,d,e,f,g,transparent)
#include "raster/lineclip.c"
}

/****************************************************************************
PARAMETERS:
x1          - X coordinate of first endpoint to clip
y1          - Y coordinate of first endpoint to clip
x2          - X coordinate of second endpoint to clip
y2          - Y coordinate of second endpoint to clip
drawLast    - True if last pixel should be drawn
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle

REMARKS:
Internal function to clip and draw an integer line, such that the clipped
line segment draws the *exact* same set of pixels that the unclipped line
would have drawn. We also correctly handle arbitrarily large lines, while
still allowing the clipped line segment to be drawn in hardware.
****************************************************************************/
N_int32 NAPI SHADOW_DrawClippedBresenhamStippleLine(
    N_int32 x1,
    N_int32 y1,
    N_int32 initialError,
    N_int32 majorInc,
    N_int32 diagInc,
    N_int32 count,
    N_int32 flags,
    N_int32 transparent,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
#define DRAW_BRESENHAM_LINE(a,b,c,d,e,f,g)  SHADOW_DrawBresenhamStippleLine(a,b,c,d,e,f,g,transparent)
#include "raster/blineclp.c"
}

/****************************************************************************
PARAMETERS:
x1          - X coordinate of first endpoint to clip
y1          - Y coordinate of first endpoint to clip
x2          - X coordinate of second endpoint to clip
y2          - Y coordinate of second endpoint to clip
drawLast    - True if last pixel should be drawn
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle

REMARKS:
Internal function to clip and draw an integer line, such that the clipped
line segment draws the *exact* same set of pixels that the unclipped line
would have drawn. We also correctly handle arbitrarily large lines, while
still allowing the clipped line segment to be drawn in hardware.
****************************************************************************/
N_int32 NAPI SHADOW_DrawClippedStyleLineInt(
    N_int32 x1,
    N_int32 y1,
    N_int32 x2,
    N_int32 y2,
    N_int32 drawLast,
    N_int32 transparent,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
#define DRAW_LINE_INT(a,b,c,d,e)            SHADOW_DrawStyleLineInt(a,b,c,d,e,transparent)
#define DRAW_BRESENHAM_LINE(a,b,c,d,e,f,g)  SHADOW_DrawBresenhamStyleLine(a,b,c,d,e,f,g,transparent)
#include "raster/lineclip.c"
}

/****************************************************************************
PARAMETERS:
x1          - X coordinate of first endpoint to clip
y1          - Y coordinate of first endpoint to clip
x2          - X coordinate of second endpoint to clip
y2          - Y coordinate of second endpoint to clip
drawLast    - True if last pixel should be drawn
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle

REMARKS:
Internal function to clip and draw an integer line, such that the clipped
line segment draws the *exact* same set of pixels that the unclipped line
would have drawn. We also correctly handle arbitrarily large lines, while
still allowing the clipped line segment to be drawn in hardware.
****************************************************************************/
N_int32 NAPI SHADOW_DrawClippedBresenhamStyleLine(
    N_int32 x1,
    N_int32 y1,
    N_int32 initialError,
    N_int32 majorInc,
    N_int32 diagInc,
    N_int32 count,
    N_int32 flags,
    N_int32 transparent,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
#define DRAW_BRESENHAM_LINE(a,b,c,d,e,f,g)  SHADOW_DrawBresenhamStyleLine(a,b,c,d,e,f,g,transparent)
#include "raster/blineclp.c"
}

/****************************************************************************
PARAMETERS:
x           - Destination X coordinate to draw the bitmap at
y           - Destination Y coordinate to draw the bitmap at
width       - Width of the bitmap in pixels
height      - Height of the bitmap in pixels
byteWidth   - Width of the bitmap in bytes
image       - Pointer to the bitmap image data to draw
transparent - 1 for transparent, 0 for opaque

REMARKS:
If the transparent parameter is set to 1, where a bit is 0 in the bitmap
image the destination pixel remains untouched. If the transparent
parameter is set to 0, where a bit is 0 in the bitmap image the
destination pixel is drawn in the background color. In all cases where a
bit in the bitmap image is 1, the pixel is drawn in the foreground color.

Note:   This function processes the bitmap data in MSB fashion, in that the
        first pixel drawn corrsponds to bit 7 of the first byte. The second
        pixel is bit 6 of the first byte, ...,the 8th pixel is bit 7 of
        the second byte etc. Both LSB and MSB versions are provided for
        performance.
****************************************************************************/
void NAPI SHADOW_PutMonoImageMSBSys(
    N_int32 x,
    N_int32 y,
    N_int32 width,
    N_int32 height,
    N_int32 byteWidth,
    N_uint8 *image,
    N_int32 transparent)
{
    DRAW2D.PutMonoImageMSBSys(x,y,width,height,byteWidth,image,transparent);
    DC.UpdateScreen(DC.buf.Offset,x,y,width,height);
}

/****************************************************************************
PARAMETERS:
x           - Destination X coordinate to draw the bitmap at
y           - Destination Y coordinate to draw the bitmap at
width       - Width of the bitmap in pixels
height      - Height of the bitmap in pixels
byteWidth   - Width of the bitmap in bytes
image       - Pointer to the bitmap image data to draw
transparent - 1 for transparent, 0 for opaque

REMARKS:
This function is identical to the PutMonoImageMSBSys function, except
that it processes the bitmap data in LSB fashion. This means that the
first pixel drawn corrsponds to bit 0 of the first byte. The second
pixel is bit 1 of the first byte, ...,the 8th pixel is bit 0 of the second
byte etc. Both LSB and MSB versions are provided for performance.
****************************************************************************/
void NAPI SHADOW_PutMonoImageLSBSys(
    N_int32 x,
    N_int32 y,
    N_int32 width,
    N_int32 height,
    N_int32 byteWidth,
    N_uint8 *image,
    N_int32 transparent)
{
    DRAW2D.PutMonoImageLSBSys(x,y,width,height,byteWidth,image,transparent);
    DC.UpdateScreen(DC.buf.Offset,x,y,width,height);
}

/****************************************************************************
PARAMETERS:
x           - Destination X coordinate to draw the bitmap at
y           - Destination Y coordinate to draw the bitmap at
width       - Width of the bitmap in pixels
height      - Height of the bitmap in pixels
byteWidth   - Width of the bitmap in bytes
image       - Pointer to the bitmap image data to draw
transparent - 1 for transparent, 0 for opaque
clipLeft    - Left edge of clip rectangle (inclusive)
clipTop     - Top edge of clip rectangle (inclusive)
clipRight   - Right edge of clip rectangle (exclusive)
clipBottom  - Bottom edge of clip rectangle (exclusive)

REMARKS:
If the transparent parameter is set to 1, where a bit is 0 in the bitmap
image the destination pixel remains untouched. If the transparent
parameter is set to 0, where a bit is 0 in the bitmap image the
destination pixel is drawn in the background color. In all cases where a
bit in the bitmap image is 1, the pixel is drawn in the foreground color.

Note:   This function processes the bitmap data in MSB fashion, in that the
        first pixel drawn corrsponds to bit 7 of the first byte. The second
        pixel is bit 6 of the first byte, ...,the 8th pixel is bit 7 of
        the second byte etc. Both LSB and MSB versions are provided for
        performance.
****************************************************************************/
void NAPI SHADOW_ClipMonoImageMSBSys(
    N_int32 x,
    N_int32 y,
    N_int32 width,
    N_int32 height,
    N_int32 byteWidth,
    N_uint8 *image,
    N_int32 transparent,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
    int bottom,right,clipX,clipY,clipWidth,clipHeight;

    // Check if image needs to be clipped at all
    right = x + width;
    bottom = y + height;
    if ((bottom <= clipTop) || (y >= clipBottom) ||
        (right <= clipLeft) || (x >= clipRight))
        return;

    // Clip the top and bottom of the bitmap image from being drawn
    if (y < clipTop) {
        clipY = clipTop;
        clipHeight = height - (clipTop - y);
        }
    else {
        clipY = y;
        clipHeight = height;
        }
    if (bottom > clipBottom)
        clipHeight -= (bottom - clipBottom);

    // Clip the left and right edges
    if (x < clipLeft) {
        clipX = clipLeft;
        clipWidth = width - (clipLeft - x);
        }
    else {
        clipX = x;
        clipWidth = width;
        }
    if (right > clipRight)
        clipWidth -= (right - clipRight);

    // Now draw the bitmap
    DRAW2D.ClipMonoImageMSBSys(x,y,width,height,byteWidth,image,transparent,clipLeft,clipTop,clipRight,clipBottom);
    DC.UpdateScreen(DC.buf.Offset,clipX,clipY,clipWidth,clipHeight);
}

/****************************************************************************
PARAMETERS:
x           - Destination X coordinate to draw the bitmap at
y           - Destination Y coordinate to draw the bitmap at
width       - Width of the bitmap in pixels
height      - Height of the bitmap in pixels
byteWidth   - Width of the bitmap in bytes
image       - Pointer to the bitmap image data to draw
transparent - 1 for transparent, 0 for opaque
clipLeft    - Left edge of clip rectangle (inclusive)
clipTop     - Top edge of clip rectangle (inclusive)
clipRight   - Right edge of clip rectangle (exclusive)
clipBottom  - Bottom edge of clip rectangle (exclusive)

REMARKS:
This function is identical to the ClipMonoImageMSBSys function, except
that it processes the bitmap data in LSB fashion. This means that the
first pixel drawn corrsponds to bit 0 of the first byte. The second
pixel is bit 1 of the first byte, ...,the 8th pixel is bit 0 of the second
byte etc. Both LSB and MSB versions are provided for performance.
****************************************************************************/
void NAPI SHADOW_ClipMonoImageLSBSys(
    N_int32 x,
    N_int32 y,
    N_int32 width,
    N_int32 height,
    N_int32 byteWidth,
    N_uint8 *image,
    N_int32 transparent,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom)
{
    int bottom,right,clipX,clipY,clipWidth,clipHeight;

    // Check if image needs to be clipped at all
    right = x + width;
    bottom = y + height;
    if ((bottom <= clipTop) || (y >= clipBottom) ||
        (right <= clipLeft) || (x >= clipRight))
        return;

    // Clip the top and bottom of the bitmap image from being drawn
    if (y < clipTop) {
        clipY = clipTop;
        clipHeight = height - (clipTop - y);
        }
    else {
        clipY = y;
        clipHeight = height;
        }
    if (bottom > clipBottom)
        clipHeight -= (bottom - clipBottom);

    // Clip the left and right edges
    if (x < clipLeft) {
        clipX = clipLeft;
        clipWidth = width - (clipLeft - x);
        }
    else {
        clipX = x;
        clipWidth = width;
        }
    if (right > clipRight)
        clipWidth -= (right - clipRight);

    // Now draw the bitmap
    DRAW2D.ClipMonoImageLSBSys(x,y,width,height,byteWidth,image,transparent,clipLeft,clipTop,clipRight,clipBottom);
    DC.UpdateScreen(DC.buf.Offset,clipX,clipY,clipWidth,clipHeight);
}

/****************************************************************************
PARAMETERS:
srcLeft - Left coordinate of the source rectangle to copy
srcTop  - Top coordinate of the source rectangle to copy
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
dstLeft - Left coordinate of destination
dstTop  - Top coordinate of destination
mix     - Mix code for the copy (GA_mixCodesType)
****************************************************************************/
void NAPI SHADOW_BitBlt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix)
{
    DRAW2D.BitBlt(srcLeft,srcTop,width,height,dstLeft,dstTop,mix);
    if (DC.BitBlt)
        DC.BitBlt(srcLeft,srcTop,width,height,dstLeft,dstTop,mix);
    else
        DC.UpdateScreen(DC.buf.Offset,dstLeft,dstTop,width,height);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI SHADOW_BitBltSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    N_int32 flipY)
{
    DRAW2D.BitBltSys(srcAddr,srcPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,mix,flipY);
    DC.UpdateScreen(DC.buf.Offset,dstLeft,dstTop,width,height);
}

/****************************************************************************
PARAMETERS:
srcLeft - Left coordinate of the source rectangle to copy
srcTop  - Top coordinate of the source rectangle to copy
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
dstLeft - Left coordinate of destination
dstTop  - Top coordinate of destination
rop3    - ROP3 mix code for the copy
****************************************************************************/
void NAPI SHADOW_BitBltPatt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    DRAW2D.BitBltPatt(srcLeft,srcTop,width,height,dstLeft,dstTop,rop3);
    DC.UpdateScreen(DC.buf.Offset,dstLeft,dstTop,width,height);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3        - Mix code for the copy (GA_mixCodesType)
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI SHADOW_BitBltPattSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3,
    N_int32 flipY)
{
    DRAW2D.BitBltPattSys(srcAddr,srcPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,rop3,flipY);
    DC.UpdateScreen(DC.buf.Offset,dstLeft,dstTop,width,height);
}

/****************************************************************************
PARAMETERS:
srcLeft - Left coordinate of the source rectangle to copy
srcTop  - Top coordinate of the source rectangle to copy
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
dstLeft - Left coordinate of destination
dstTop  - Top coordinate of destination
rop3    - ROP3 mix code for the copy
****************************************************************************/
void NAPI SHADOW_BitBltColorPatt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3)
{
    DRAW2D.BitBltColorPatt(srcLeft,srcTop,width,height,dstLeft,dstTop,rop3);
    DC.UpdateScreen(DC.buf.Offset,dstLeft,dstTop,width,height);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
rop3        - Mix code for the copy (GA_mixCodesType)
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI SHADOW_BitBltColorPattSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 rop3,
    N_int32 flipY)
{
    DRAW2D.BitBltColorPattSys(srcAddr,srcPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,rop3,flipY);
    DC.UpdateScreen(DC.buf.Offset,dstLeft,dstTop,width,height);
}

/****************************************************************************
PARAMETERS:
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
****************************************************************************/
void NAPI SHADOW_SrcTransBlt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    DRAW2D.SrcTransBlt(srcLeft,srcTop,width,height,dstLeft,dstTop,mix,transparent);
    DC.UpdateScreen(DC.buf.Offset,dstLeft,dstTop,width,height);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI SHADOW_SrcTransBltSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent,
    N_int32 flipY)
{
    DRAW2D.SrcTransBltSys(srcAddr,srcPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,mix,transparent,flipY);
    DC.UpdateScreen(DC.buf.Offset,dstLeft,dstTop,width,height);
}

/****************************************************************************
PARAMETERS:
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
****************************************************************************/
void NAPI SHADOW_DstTransBlt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent)
{
    DRAW2D.DstTransBlt(srcLeft,srcTop,width,height,dstLeft,dstTop,mix,transparent);
    DC.UpdateScreen(DC.buf.Offset,dstLeft,dstTop,width,height);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate within source bitmap to copy
srcTop      - Top coordinate within source bitmap to copy
width       - Width of the rectangle in pixels
height      - Height of the rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
mix         - Mix code for the copy (GA_mixCodesType)
transparent - Transparent color value
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI SHADOW_DstTransBltSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 mix,
    GA_color transparent,
    N_int32 flipY)
{
    DRAW2D.DstTransBltSys(srcAddr,srcPitch,srcLeft,srcTop,width,height,dstLeft,dstTop,mix,transparent,flipY);
    DC.UpdateScreen(DC.buf.Offset,dstLeft,dstTop,width,height);
}

/****************************************************************************
PARAMETERS:
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstWidth    - Width of the destination rectangle in pixels
dstHeight   - Height of the destination rectangle in scanlines
doClip      - True if the blit should be clipped
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle
mix         - Mix code for the copy (GA_mixCodesType)
****************************************************************************/
void NAPI SHADOW_StretchBlt(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    N_int32 doClip,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom,
    N_int32 mix)
{
    DRAW2D.StretchBlt(srcLeft,srcTop,srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,doClip,clipLeft,clipTop,clipRight,clipBottom,mix);
    if (doClip) {
        int dstLeft2 = MAX(dstLeft, clipLeft);
        int dstTop2 = MAX(dstTop, clipTop);
        int dstRight = MIN((dstLeft + dstWidth), clipRight);
        int dstBottom = MIN((dstTop + dstHeight), clipBottom);
        DC.UpdateScreen(DC.buf.Offset,dstLeft2,dstTop2,dstRight-dstLeft2,dstBottom-dstTop2);
        }
    else
        DC.UpdateScreen(DC.buf.Offset,dstLeft,dstTop,dstWidth,dstHeight);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstWidth    - Width of the destination rectangle in pixels
dstHeight   - Height of the destination rectangle in scanlines
doClip      - True if the blit should be clipped
clipLeft    - Left coordinate of clip rectangle
clipTop     - Top coordinate of clip rectangle
clipRight   - Right coordinate of clip rectangle
clipBottom  - Bottom coordinate of clip rectangle
mix         - Mix code for the copy (GA_mixCodesType)
flipY       - True if the bitmap should be flipped in Y
****************************************************************************/
void NAPI SHADOW_StretchBltSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    N_int32 doClip,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom,
    N_int32 mix,
    N_int32 flipY)
{
    DRAW2D.StretchBltSys(srcAddr,srcPitch,srcLeft,srcTop,srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,doClip,clipLeft,clipTop,clipRight,clipBottom,mix,flipY);
    if (doClip) {
        int dstLeft2 = MAX(dstLeft, clipLeft);
        int dstTop2 = MAX(dstTop, clipTop);
        int dstRight = MIN((dstLeft + dstWidth), clipRight);
        int dstBottom = MIN((dstTop + dstHeight), clipBottom);
        DC.UpdateScreen(DC.buf.Offset,dstLeft2,dstTop2,dstRight-dstLeft2,dstBottom-dstTop2);
        }
    else
        DC.UpdateScreen(DC.buf.Offset,dstLeft,dstTop,dstWidth,dstHeight);
}

/****************************************************************************
PARAMETERS:
srcAddr         - Address of source bitmap in system memory
srcPitch        - Pitch of source bitmap in bytes
srcLeft         - Left coordinate within source bitmap to copy
srcTop          - Top coordinate within source bitmap to copy
width           - Width of the rectangle in pixels
height          - Height of the rectangle in scanlines
dstLeft         - Left coordinate of destination
dstTop          - Top coordinate of destination
bitsPerPixel    - Pixel depth for the bitmap to copy
pixelFormat     - Pixel format for the bitmap to copy
dstPal          - Color palette for the destination device (if dest is 8bpp)
srcPal          - Color palette for the bitmap to copy (if source is 8bpp)
dither          - True if the output should be dithered
mix             - Mix code for the copy (GA_mixCodesType)
flipY           - True if the bitmap should be flipped in Y

RETURNS:
0 on success, -1 on failure.
****************************************************************************/
N_int32 NAPI SHADOW_ConvertBltSys_Obsolete(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 width,
    N_int32 height,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pixelFormat,
    GA_palette *dstPal,
    GA_palette *srcPal,
    N_int32 dither,
    N_int32 mix,
    N_int32 flipY)
{
    if (DRAW2D.ConvertBltSys_Obsolete(srcAddr,srcPitch,srcLeft,srcTop,width,height,
            dstLeft,dstTop,bitsPerPixel,pixelFormat,dstPal,srcPal,dither,mix,flipY) == 0) {
        DC.UpdateScreen(DC.buf.Offset,dstLeft,dstTop,width,height);
        return 0;
        }
    return -1;
}

/****************************************************************************
PARAMETERS:
srcAddr         - Address of source bitmap in system memory
srcPitch        - Pitch of source bitmap in bytes
srcLeft         - Left coordinate of the source rectangle to copy
srcTop          - Top coordinate of the source rectangle to copy
srcWidth        - Width of the source rectangle in pixels
srcHeight       - Height of the source rectangle in scanlines
dstLeft         - Left coordinate of destination
dstTop          - Top coordinate of destination
dstWidth        - Width of the destination rectangle in pixels
dstHeight       - Height of the destination rectangle in scanlines
doClip          - True if the blit should be clipped
clipLeft        - Left coordinate of clip rectangle
clipTop         - Top coordinate of clip rectangle
clipRight       - Right coordinate of clip rectangle
clipBottom      - Bottom coordinate of clip rectangle
bitsPerPixel    - Pixel depth for the bitmap to copy
pixelFormat     - Pixel format for the bitmap to copy
dstPal          - Color palette for the destination device (if dest is 8bpp)
srcPal          - Color palette for the bitmap to copy (if source is 8bpp)
dither          - True if the output should be dithered
mix             - Mix code for the copy (GA_mixCodesType)
flipY           - True if the bitmap should be flipped in Y

RETURNS:
0 on success, -1 on failure.
****************************************************************************/
N_int32 NAPI SHADOW_StretchConvertBltSys_Obsolete(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    N_int32 doClip,
    N_int32 clipLeft,
    N_int32 clipTop,
    N_int32 clipRight,
    N_int32 clipBottom,
    N_int32 bitsPerPixel,
    GA_pixelFormat *pixelFormat,
    GA_palette *dstPal,
    GA_palette *srcPal,
    N_int32 dither,
    N_int32 mix,
    N_int32 flipY)
{
    if (DRAW2D.StretchConvertBltSys_Obsolete(srcAddr,srcPitch,srcLeft,srcTop,
            srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,doClip,clipLeft,clipTop,clipRight,clipBottom,
            bitsPerPixel,pixelFormat,dstPal,srcPal,dither,mix,flipY) == 0) {
        if (doClip) {
            int dstLeft2 = MAX(dstLeft, clipLeft);
            int dstTop2 = MAX(dstTop, clipTop);
            int dstRight = MIN((dstLeft + dstWidth), clipRight);
            int dstBottom = MIN((dstTop + dstHeight), clipBottom);
            DC.UpdateScreen(DC.buf.Offset,dstLeft2,dstTop2,dstRight-dstLeft2,dstBottom-dstTop2);
            }
        else
            DC.UpdateScreen(DC.buf.Offset,dstLeft,dstTop,dstWidth,dstHeight);
        return 0;
        }
    return -1;
}

/****************************************************************************
PARAMETERS:
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstWidth    - Width of the destination rectangle in pixels
dstHeight   - Height of the destination rectangle in scanlines
fx          - GA_bltFx structure describing the requested effects
****************************************************************************/
void NAPI SHADOW_BitBltFx(
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    GA_bltFx *fx)
{
    DRAW2D.BitBltFx(srcLeft,srcTop,srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,fx);
    DC.UpdateScreen(DC.buf.Offset,dstLeft,dstTop,dstWidth,dstHeight);
}

/****************************************************************************
PARAMETERS:
srcAddr     - Address of source bitmap in system memory
srcPitch    - Pitch of source bitmap in bytes
srcLeft     - Left coordinate of the source rectangle to copy
srcTop      - Top coordinate of the source rectangle to copy
srcWidth    - Width of the source rectangle in pixels
srcHeight   - Height of the source rectangle in scanlines
dstLeft     - Left coordinate of destination
dstTop      - Top coordinate of destination
dstWidth    - Width of the destination rectangle in pixels
dstHeight   - Height of the destination rectangle in scanlines
fx          - GA_bltFx structure describing the requested effects
****************************************************************************/
void NAPI SHADOW_BitBltFxSys(
    void *srcAddr,
    N_int32 srcPitch,
    N_int32 srcLeft,
    N_int32 srcTop,
    N_int32 srcWidth,
    N_int32 srcHeight,
    N_int32 dstLeft,
    N_int32 dstTop,
    N_int32 dstWidth,
    N_int32 dstHeight,
    GA_bltFx *fx)
{
    DRAW2D.BitBltFxSys(srcAddr,srcPitch,srcLeft,srcTop,srcWidth,srcHeight,dstLeft,dstTop,dstWidth,dstHeight,fx);
    DC.UpdateScreen(DC.buf.Offset,dstLeft,dstTop,dstWidth,dstHeight);
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the rectangle to draw
top     - Top coordinate of the rectangle to draw
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
****************************************************************************/
void NAPI SHADOW_DrawRect_Generic(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height)
{
    DC.UpdateScreen(DC.buf.Offset,left,top,width,height);
}

/****************************************************************************
PARAMETERS:
left    - Left coordinate of the rectangle to draw
top     - Top coordinate of the rectangle to draw
width   - Width of the rectangle in pixels
height  - Height of the rectangle in scanlines
****************************************************************************/
void NAPI SHADOW_DrawRectExt_Generic(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height,
    GA_color color,
    N_int32 mix)
{
    (void)color;
    (void)mix;
    DC.UpdateScreen(DC.buf.Offset,left,top,width,height);
}

/****************************************************************************
PARAMETERS:
x1          - X1 coordinate
y1          - Y1 coordinate
x2          - X2 coordinate
y2          - Y2 coordinate
drawLast    - 1 to draw last pixel, 0 to skip it
****************************************************************************/
void NAPI SHADOW_DrawLineInt_Generic(
    N_int32 x1,
    N_int32 y1,
    N_int32 x2,
    N_int32 y2,
    N_int32 drawLast)
{
    int     absDeltaX,absDeltaY,majorInc,diagInc;
    int     initialError,flags,fincr,count,dstAddr;
    uchar   *dstBase,*srcAddr;

    flags = gaLineXPositive | gaLineYPositive | gaLineXMajor;
    if ((absDeltaX = x2 - x1) < 0) {
        absDeltaX = -absDeltaX;
        flags &= ~gaLineXPositive;
        }
    if ((absDeltaY = y2 - y1) < 0) {
        fincr = -((int)DC.mi.BytesPerScanLine);
        absDeltaY = -absDeltaY;
        flags &= ~gaLineYPositive;
        }
    else
        fincr = (int)DC.mi.BytesPerScanLine;
    if (absDeltaY > absDeltaX) {
        SWAP(absDeltaX,absDeltaY);
        flags &= ~gaLineXMajor;
        }
    majorInc = 2 * absDeltaY;               // 2 * dy
    initialError = majorInc - absDeltaX;    // 2 * dy - dx
    diagInc = initialError - absDeltaX;     // 2 * (dy - dx)
    dstBase = DC.hwCtx->BankedMem;
    dstAddr = (DC.mi.BytesPerScanLine * y1) + LIN_X_ADDR(x1,DC.byteShift);
    srcAddr = (uchar*)DC.buffer + dstAddr;
    dstAddr += DC.buf.Offset;
    if (drawLast)
        count = absDeltaX+1;
    else
        count = absDeltaX;
    if (flags & gaLineXMajor) {
        if (flags & gaLineXPositive)
            DC.drawLineXPosLo(dstBase,dstAddr,srcAddr,initialError,majorInc,diagInc,count,fincr);
        else
            DC.drawLineXNegLo(dstBase,dstAddr,srcAddr,initialError,majorInc,diagInc,count,fincr);
        }
    else {
        if (flags & gaLineXPositive)
            DC.drawLineXPosHi(dstBase,dstAddr,srcAddr,initialError,majorInc,diagInc,count,fincr);
        else
            DC.drawLineXNegHi(dstBase,dstAddr,srcAddr,initialError,majorInc,diagInc,count,fincr);
        }
}

/****************************************************************************
PARAMETERS:
x1              - X1 coordinate
y1              - Y1 coordinate
initialError    - Initial error term
majorInc        - Major increment factor for error term
diagInc         - Diagonal increment factor for error term
count           - Number of pixels to draw
flags           - Flags to control drawing
****************************************************************************/
void NAPI SHADOW_DrawBresenhamLine_Generic(
    N_int32 x1,
    N_int32 y1,
    N_int32 initialError,
    N_int32 majorInc,
    N_int32 diagInc,
    N_int32 count,
    N_int32 flags)
{
    int     fincr,dstAddr;
    uchar   *dstBase,*srcAddr;

    if (flags & gaLineYPositive)
        fincr = (int)DC.mi.BytesPerScanLine;
    else
        fincr = -((int)DC.mi.BytesPerScanLine);
    dstBase = DC.hwCtx->BankedMem;
    dstAddr = (DC.mi.BytesPerScanLine * y1) + LIN_X_ADDR(x1,DC.byteShift);
    srcAddr = (uchar*)DC.buffer + dstAddr;
    dstAddr += DC.buf.Offset;
    if (!(flags & gaLineDoLastPel))
        count--;
    if (flags & gaLineXMajor) {
        if (flags & gaLineXPositive)
            DC.drawLineXPosLo(dstBase,dstAddr,srcAddr,initialError,majorInc,diagInc,count,fincr);
        else
            DC.drawLineXNegLo(dstBase,dstAddr,srcAddr,initialError,majorInc,diagInc,count,fincr);
        }
    else {
        if (flags & gaLineXPositive)
            DC.drawLineXPosHi(dstBase,dstAddr,srcAddr,initialError,majorInc,diagInc,count,fincr);
        else
            DC.drawLineXNegHi(dstBase,dstAddr,srcAddr,initialError,majorInc,diagInc,count,fincr);
        }
}

/****************************************************************************
REMARKS:
Function to update a region of the screen from the shadow buffer. This
is called via OS shell drivers that need to do custom rendering to the
shadow buffer, and then update the contents of that to the screen.
****************************************************************************/
void NAPI SHADOW_UpdateScreen(
    N_int32 left,
    N_int32 top,
    N_int32 width,
    N_int32 height)
{
    DC.UpdateScreen(DC.buf.Offset,left,top,width,height);
}
